
import java.io.*;
import java.sql.*;
/**
 * Communication PostgresSql with java,  ecriture et recuperation de donnes de la BD
 * @author Ouachani Mourad
 * @version 1.0
 */

public class Pgsql {

    private Connection connect;
    private String  driverName = null;
    private Statement stQuery;
    private Statement stQuerySelect;



    /**
     * Chargement du pilote et connection a la base de donnes
     * @param db le nom de la base de donnes
     * @param user le nom de l'utilisateur
     * @param passwd le password
     * @exception  ClassNotFoundException
     * @exception SQLException Indique une erreur d'ouverture de connection et tracage
     */
    public Pgsql(String db,String user, String passwd) throws SQLException,ClassNotFoundException {
	try{
	    String driver = "jdbc:postgresql:";
	    //chargement du pilote pour le pont jdbc/odbc
	    Class.forName("org.postgresql.Driver");
	    //connection a la base de donnee
	    this.connect = DriverManager.getConnection(driver+db,user,passwd);
	}
	//erreur de chargement du driver
	catch(ClassNotFoundException e) {
	    System.err.println("Driver non charg !");
	    e.printStackTrace();
	    throw e;
	} //fin du catch du chargement du driver
	//erreur de connection
	catch(SQLException sqle) {
	    System.err.println("Connection non etablie !");
	    sqle.printStackTrace();
	    throw sqle;
	}//fin du catch de la connection a la bd
    }//fin du constructeur


    /**
     * CloseConnection  ferme la connection a la base de donnees
     * @param  aucun
     * @return  rien
     * @exception  SQLException  Indique une erreur de fermeture de connection et tracage de l'erreur
     */
    public void closeConnection(){
   	//fermeture de la connection a la BD
   	try{
   	    this.connect.close();
   	}catch(SQLException sqle){
   	    System.err.println("Fermeture de la connection non etablie !");
	    sqle.printStackTrace();
	}//fin du catch
    }

    /**
     * query   executer une requete sql (insert ou update ou delete)
     * @param request  requete SQL a executer
     * @return int le nombre de colonne inserer, mis a jour ou supprimer
     * @exception SQLException l'exception n'est pas traiter localement
     */
    public int query (String request) throws SQLException{
	int rowCount;
	//creation de l'objet enregistrement
	this.stQuery = this.connect.createStatement();
	//execution de la requete pour la mise a jour
	rowCount = this.stQuery.executeUpdate(request);
	return rowCount;
    }

    /**
     * querySelect   executer une requete sql de selection
     * @param request  requete SQL a executer
     * @return ResultSet  ensemble d'objets contenant l'execution de la requete (SELECT)
     * @exception SQLException l'exception n'est pas traiter localement
     */
    public ResultSet querySelect(String requestSelect) throws SQLException{
   	ResultSet rsQuerySelection;
	//creation de l'objet enregistrement
   	this.stQuerySelect = this.connect.createStatement();
   	//execution de la requete de selection
   	rsQuerySelection = this.stQuerySelect.executeQuery(requestSelect);
	/*Ses lignes sont juste pour tester la fonction */
	//	while(rsQuerySelection.next()){
	//  System.out.println(rsQuerySelection.getString(1));
	//}
	/* fin du test */
   	return rsQuerySelection;
    }

    /* Main pour tester les fonctions 
    public static void main(String[] args){

	Pgsql connec = new Pgsql(args[0],args[1],args[2]);

	try{
	    int test = connec.query("insert into  mothttp (mot) values('kartoush')");
	    System.out.println("ca marche .."+test);
	}catch(SQLException sqle){
	    sqle.printStackTrace();
	}
	try{
	    ResultSet rs = connec.querySelect("select * from mothttp");
	}catch(SQLException sql){
	    sql.printStackTrace();
	}
	connec.closeConnection();
    }
    /* fin du Main */
}
