import java.io.*;
import java.lang.*;
import java.sql.*;
import java.net.*;

/**
 * Cette classe gnre un processus qui pour toutes les entres de la tables <code>Http</code>
 * (c'est  dire tous les liens trouvs) vrifie que la page correspondante existe et l'insre
 * dans la table <code>Page</code> si elle n'y est pas dj. Dans le
 * cas contraire ce lien est supprim de la base de donnes.<p>
 * Rem: La suppression d'une lien entraine galement la suppression de toutes les entres associes
 *  ce lien dans la table <code>AppHttp</code>.<p>
 * Rem: Les processus engendrs par cette classe sont appels dans le reste de cette documentation
 * 'processus de vrification des liens'.<p>
 * Les processus de vrificatioin de liens n'ont pas t conu pour fonctionner en parallle avec
 * les autres processus. Pour cela, il faudrait seuleemnt rajouter aux processus de navigation la
 * capacit d'indiquer a ces processus qu'un nouveau lien http  t insr dans la bd (comme
 * c'est le cas pour les processus de vrification de page).
 * @author Martin Ludovic
 * @see "Pour plus d'information sur le droulement d'une suppression d'un lien, voir la fonction
 * <code>supp_lien(INTEGER)</code> dans bd/crawler.sql."
 */
public class HttpVerifThread extends Thread {
    /** Nombre maximum de tentatives de rptitions de requtes avant ne soit considre comme choue.
     *  En effet, il est possible pour divers raisons qu'une requte choue alors qu'elle devrait russir.*/
    protected final int BOUCLE_MAX = 5;
    /** Drapeau signalant que le processus doit s'arrter. */
    protected boolean to_finish;
    /** Flux d'entre sur le fichier de journalisation. */
    protected PrintStream outLog;
    /** Identificateur du lien  vrifier et ventuellement  supprimer. */
    protected int id_lien;
    /** Le plus grand identificateur de page utilis dans la base de donnes. */
    protected int id_max;
    /** Drapeau indiquant si une requte <code>SELECT currval('page_id_page_seq')</code> a dj t 
     *  envoye. En effet, tant que cela n'a pas t fait on ne peut pas faire de requte <code>SELECT
     *  nextval('page_id_page_seq')</code>*/
    protected boolean first_currval;
    /** Objet permettant de communiquer avec la base de donnes du crawler. */
    protected Pgsql bd;

    /**
     * Construit un nouvel objet <code>HttpVerifThread</code>.
     * @param nom le nom du processus
     * @param outLog le flux permettant d'crire dans le fichier de journalisation
     * @see #outLog
     */
    public HttpVerifThread (String nom, PrintStream outLog) {
	super(nom);
	this.outLog = outLog;
	to_finish = false;
	id_lien = 0;
	first_currval = true;
	bd = null; // L'initialisation de la bd se fait ds la mthode run
    }

    /**
     * Permet un arrt propre du processus.<p>
     * L'arrt n'est pas immdiat, il s'agit plutt d'une demande d'arrt permettant au processus
     * de terminer ce qu'il est en train de faire et l'empchant d'enchainer sur une nouvelle tche.
     * Il existe cependant un cas ou l'arrt est quasiment immdiat: celui o le processus est bloqu
     * en attente qu'au moins une page soit insre dans la base de donnes. Le processus est alors sorti
     * de sont attente et est termin.
     */
    public void finish () {
	to_finish = true;
	if (id_max == 0) {
	    try {
		interrupt();
	    } catch (SecurityException e) {}
	}
    }

    /**
     * Recherche le plus grand identificateur de page utilis dans la base de donnes.
     * @return Le plus grand identificateur de page utilis dans la base de donnes.
     * @throws SQLException En cas d'erreur engendrant l'interruption du processus
     */
    protected int maxId () throws SQLException {
	String query,erreur;
	int max;
	ResultSet rs;
	boolean exist;

	query = "SELECT max(id_http),count(id_http) FROM Http";
	try {
	    // Envoie la requte  la bd
	    rs = bd.querySelect(query);
	    // Regarde si le rsultat n'est pas vide
	    exist = rs.next();
	    if (!exist) {
		// Si la requte s'est bien passe, le rsultat contient obligatoirement 1 ligne
		throw new SQLException();
	    }
	    // Lit le rsultat
	    if (rs.getInt(2) != 0)
		max = rs.getInt(1);
	    else // aucune page n'a encore t insre dans la bd
		max = 0;
	}
	catch (SQLException e) {
	    System.out.println("Erreur de requete ds " + getName() + " -> fermeture du proc");
	    erreur = "ERROR: " + getName() + ": echecs a " + query;
	    erreur += " ou lors de la lecture du rsultat => fermeture du proc";
	    outLog.println(erreur);
	    throw e;
	}
	return max;
    }

    /**
     * Recherche le nouvel identificateur maximum d'une page dans la base de donnes.<p>
     * Cette mthode permet de mettre  jour le champ <code>id_max</code>.
     * @throws SQLException En cas d'erreur de communication avec la base de donnes
     * @see #id_max
     */
    protected void newMaxId () throws SQLException {
	String query;
	ResultSet rs=null;
	boolean exist,error=true;
	int boucle;

	// Verifie qu'on peut envoyer la requte SELECT currval('http_id_http_seq')
	if (first_currval) {
	    boucle = 1;
	    query = "SELECT nextval('http_id_http_seq')";
	    do {
		try {
		    bd.querySelect(query);
		    // Le resultat de la requte ne nous interesse pas
		    error = false;
		}
		catch (SQLException e) {
		    /* Il est possible que la requte choue pour divers raisons. Dans ce cas, 
		       on la recommence jusqu' BOUCLE_MAX fois. */
		    error = true;
		    boucle++;
		}
	    } while (error && boucle<BOUCLE_MAX);
	    if (boucle == BOUCLE_MAX) {
		System.out.println("Erreur de requete ds " + getName() + " -> fermeture du proc");
		outLog.println("ERROR: " + getName() + ": " + BOUCLE_MAX + " echecs a " + query + " => fermeture du proc");
		throw new SQLException();
	    }
	    first_currval = false;
	}
	// Envoie de la requte...
	boucle = 1;
	query = "SELECT currval('http_id_http_seq')";
	do {
	    try {
		rs = bd.querySelect(query);
		error = false;
	    }
	    catch (SQLException e) {
		/* Il est possible que la requte choue pour divers raisons. Dans ce cas, 
		   on la recommence jusqu' BOUCLE_MAX fois. */
		error = true;
		boucle++;
	    }
	} while (error && boucle<BOUCLE_MAX);
	if (boucle == BOUCLE_MAX) {
	    System.out.println("Erreur de requete ds " + getName() + " -> fermeture du proc");
	    outLog.println("ERROR: " + getName() + ": " + BOUCLE_MAX + " echecs a " + query + " => fermeture du proc");
	    throw new SQLException();
	}
	// Lecture de la requte...
	try {
	    exist = rs.next();
	}
	catch (SQLException e) {
	    // Erreur inattendue
	    System.out.println("Erreur inattendue ds " + getName() + " -> fermeture du proc");
	    outLog.println("ERROR: " + getName() + ": " + e + " => fermeture du proc");
	    throw e;
	}
	if (!exist) {
	    // La requte n'a retourne aucun rsultat -> impossible
	    System.out.println("Erreur inattendue ds " + getName() + " -> fermeture du proc");
	    outLog.println("ERROR: " + getName() + ": " + query + ":aucun resultat => fermeture du proc");
	    throw new SQLException();
	}
	try {
	    // Mise  jour de id_max
	    id_max = rs.getInt(1);
	}
	catch (SQLException e) {
	    System.out.println("Erreur inattendue ds " + getName() + " -> fermeture du proc");
	    outLog.println("ERROR: " + getName() + ": " + e + " => fermeture du proc");
	    throw e;
	}
    }

    /**
     * Vrifie que l'identificateur de lien pass en argument existe bien dans la base de donnes du
     * crawler, c'est  dire qu'il correspond  un lien existant.
     * @param id l'identificateur de lien dont on veut vrifier l'existance.
     * @return <ul>
     *           <li> <code>true</code> si l'identificateur existe.
     *           <li> <code>false</code> dans le cas contraire.
     *         </ul>
     * @throws SQLException En cas d'erreur de communication avec la base de donnes
     */
    protected boolean existId (int id) throws SQLException {
	String query;
	ResultSet rs=null;
	boolean exist,error=true;
	int boucle;

	boucle = 1;
	query = "SELECT id_http FROM Http WHERE id_http=" + id;
	do {
	    try {
		rs = bd.querySelect(query);
		error = false;
	    }
	    catch (SQLException e) {
		/* Il est possible que la requte choue pour divers raisons. Dans ce cas, 
		   on la recommence jusqu' BOUCLE_MAX fois. */
		error = true;
		boucle++;
	    }
	} while (error && boucle<BOUCLE_MAX);
	if (boucle == BOUCLE_MAX) {
	    System.out.println("Erreur de requete ds " + getName() + " -> fermeture du proc");
	    outLog.println("ERROR: " + getName() + ": " + BOUCLE_MAX + " echecs a " + query + " => fermeture du proc");
	    throw new SQLException();
	}
	// Lecture de la requte...
	try {
	    exist = rs.next();
	}
	catch (SQLException e) {
	    // Erreur inattendue
	    System.out.println("Erreur inattendue ds " + getName() + " -> fermeture du proc");
	    outLog.println("ERROR: " + getName() + ": " + e + " => fermeture du proc");
	    throw e;
	}
	if (!exist) {
	    System.out.println("L'id "+id +" est inutilise");
	    return false;
	}
	else {
	    System.out.println("L'id "+id +" est utilise");
	    return true;
	}
    }

    /**
     * Retourne le plus petit identificateur de lien valide dans la base de donnes du crawler.
     * @return <ul>
     *           <li>le plus petit identificateur de lien valide en cas de succs.
     *           <li>-1 en cas d'chec
     *         </ul>
     */
    protected synchronized int nextId () {
	int next;
	boolean error;

	// Vrifie s'il y a des liens dans la base de donnes
	if (id_max == 0) {
	    // Il n'y a aucune lien dans la bd, le processus se met en attente
	    try {
		wait();
	    } catch (InterruptedException e) {return -2;}
	}
	// Vrifie si de nouveaux liens ont t rajouts rcemment dans la bd
	/* A faire ...*/
	// Recherche l'identificateur du prochain lien  vrifier
	next = id_lien;
	do {
	    next++;
	    if (next > id_max) {
		next = 1;
		System.out.println("VerifLien: Retour au 1er lien (max_id="+id_max+")");
	    }
	    // Vrifie que l'identificateur correspond  une page existante
	    try {
		error = !existId(next);
	    } catch (SQLException e) {return -1;}
	} while (error);
	return next;
    }

    /**
     * Supprime un lien n'existant plus sur le net de la base de donnes du crawler.<p>
     * Le lien supprim est le lien correspondant  l'identificateur de lien courant stock
     * dans le champ <code>id_lien</code>.
     * Cette mthode fait appel  la fonction <code>INTEGER supp_lien(INTEGER)</code> de la bd pour 
     * supprimer le lien et ses entres dans la table <code>AppHttp</code>.
     * @param sLien l'url du lien  supprimer (utilis uniquement pour le fichier de journalisation).
     * @return La valeur retourne par la fonction <code>supp_lien</code> de la base de donnes.
     * @throws SQLException En cas de problme de communication avec la base de donnes.
     * @see "Pour en savoir plus sur la suppression d'une lien, voir la fonction supp_lien
     * dans le fichier crawler.sql."
     */
    protected int suppLien (String sLien) throws SQLException {
	String query;
	ResultSet rs=null;
	boolean exist,error=true;
	int boucle,res;
	String log;
	String fichier;

	System.out.println("SUPP_DEB: " + getName() + " suppression du lien " + sLien);

	/// Suppression dans la bd
	outLog.println("SUPP_DEB: " + getName() + " suppression du lien " + sLien);
	boucle = 1;
	// Appel  la fonction supp_lien
	query = "SELECT supp_lien(" + id_lien + ")";
	do {
	    try {
		rs = bd.querySelect(query);
		error = false;
	    }
	    catch (SQLException e) {
		/* Il est possible que la requte choue pour divers raisons. Dans ce cas, 
		   on la recommence jusqu' BOUCLE_MAX fois. */
		error = true;
		boucle++;
	    }
	} while (error && boucle<BOUCLE_MAX);
	if (boucle == BOUCLE_MAX) {
	    System.out.println("Erreur de requete ds " + getName() + " -> fermeture du proc");
	    outLog.println("ERROR: " + getName() + ": " + BOUCLE_MAX + " echecs a " + query + " => fermeture du proc");
	    throw new SQLException();
	}
	try {
	    exist = rs.next();
	}
	catch (SQLException e) {
	    // Erreur inattendue
	    System.out.println("Erreur inattendue ds " + getName() + " -> fermeture du proc");
	    outLog.println("ERROR: " + getName() + ": " + e + " => fermeture du proc");
	    throw e;
	}
	if (!exist) {
	    // Erreur inattendue
	    System.out.println("Erreur inattendue ds " + getName() + " -> fermeture du proc");
	    outLog.println("ERROR: " + getName() + ": " + query + "->aucun resultat => fermeture du proc");
	    throw new SQLException();
	}
	try {
	    res = rs.getInt(1);
	}
	catch (SQLException e) {
	    System.out.println("Erreur inattendue ds " + getName() + " -> fermeture du proc");
	    outLog.println("ERROR: " + getName() + ": " + e + " => fermeture du proc");
	    throw e;
	}
	/// Fin
	log = "SUPP_FIN: " + getName() + " suppression de la lien " + sLien + " terminee avec succes (";
	log += res + " liens et/ou mots supprimes avec cette lien)";
	outLog.println(log);
	return res;
    }

    /**
     * Vrifie que la page correspondant  l'identificateur courant (<code>id_lien</code>) existe
     * sur le net et supprime le lien de la base de donnes si ce n'est pas le cas.
     * @return La valeur retourne par la fonction <code>supp_lien</code> de la base de donnes.
     * @throws SQLException En cas de problme de communication avec la base de donnes.
     * @see #suppLien(java.lang.String)
     * @see "Pour en savoir plus sur la suppression d'une lien, voir la fonction supp_lien
     * dans le fichier crawler.sql."
     */
    protected int verifLien () throws SQLException {
	String query;
	ResultSet rs=null;
	boolean exist,error=true;
	int boucle;
	String url_lien;
	URL url;
	URLConnection connect;
	InputStream is;
	byte buf[];
	int octets_lus;

	// Rcupre l'url correspondant  l'identificateur de lien
	boucle = 1;
	query = "SELECT lien FROM Http WHERE id_http=" + id_lien;
	do {
	    try {
		rs = bd.querySelect(query);
		error = false;
	    }
	    catch (SQLException e) {
		/* Il est possible que la requte choue pour divers raisons. Dans ce cas, 
		   on la recommence jusqu' BOUCLE_MAX fois. */
		error = true;
		boucle++;
	    }
	} while (error && boucle<BOUCLE_MAX);
	if (boucle == BOUCLE_MAX) {
	    System.out.println("Erreur de requete ds " + getName() + " -> fermeture du proc");
	    outLog.println("ERROR: " + getName() + ": " + BOUCLE_MAX + " echecs a " + query + " => fermeture du proc");
	    throw new SQLException();
	}
	try {
	    exist = rs.next();
	}
	catch (SQLException e) {
	    // Erreur inattendue
	    System.out.println("Erreur inattendue ds " + getName() + " -> fermeture du proc");
	    outLog.println("ERROR: " + getName() + ": " + e + " => fermeture du proc");
	    throw e;
	}
	if (!exist) {
	    /* Normalement la requte ne devrait pas renvoyer 0 rsultat car l'identificateur pass
	       en argument est sens tre valide. Cependant si ce n'est pas le cas, il ne s'agit pas
	       d'une erreur grave. On se contente donc de le signaler  la mthode appelante mais pas
	        l'utilisateur. */
	    return -1;
	}
	try {
	    url_lien = rs.getString(1);
	}
	catch (SQLException e) {
	    System.out.println("Erreur inattendue ds " + getName() + " -> fermeture du proc");
	    outLog.println("ERROR: " + getName() + ": " + e + " => fermeture du proc");
	    throw e;
	}
	// Vrifie que cette lien existe encore sur le net
	buf = new byte[1];
	try {
	    System.out.println("Verif du lien " + url_lien);
	    url = new URL(url_lien);
	    connect = url.openConnection();
	    is = connect.getInputStream();
	    octets_lus = is.read(buf,0,1);
	    if (octets_lus == 0) {
		return suppLien(url_lien);
	    }
	    System.out.println("Lien " + url_lien + " ok");
	} 
	catch (Exception e) {
	    // Si une erreur survient, cela signifie que le lien n'existe plus
	    // Suppression du lien dans la base de donnes du crawler
	    try {
		return suppLien(url_lien);
	    } catch (SQLException ex) {throw ex;}
	}
	return 0;
    }

    /**
     * Mthode dterminant le fonctionnement du processus.<p>
     * La majorit des champs de l'objet sont initialiss dans le constructeur, cependant certains le sont
     * dans cette mthode:
     * <ul>
     *   <li>la connection  la base de donnes (<code>bd</code>)
     *   <li>le plus grand identificateur de lien utilis dans la base de donnes (<code>id_max</code>)
     * </ul>
     * La tche de ce processus se dcompose en trois tapes:
     * <ul>
     *   <li>Recupration de l'identificateur d'une lien  parser
     *   <li>Vrification si la lien existe toujours sur le net
     *   <li>Suppression de la lien si elle n'existe plus
     * </ul>
     */
    public void run () {
	Page page;
	String erreur,log;

	outLog.println("Demarrage de " + getName());
	// Connection  la base de donnes
	try {
	    bd = new Pgsql(Crawler.name_db,Crawler.user_db,Crawler.passwd_db);
	}
	catch(ClassNotFoundException e) {
	    System.out.println("Erreur de connection a la bd ds " + getName() + " -> fermeture du proc");
	    erreur = "ERROR: " + getName();
	    erreur += ": connection bd impossible (chargement du pilote jdbc impossible): ";
	    erreur = erreur + e + " => fermeture du proc";
	    outLog.println(erreur);
	    return;
	}
	catch(SQLException sqle) {
	    System.out.println("Erreur de connection a la bd ds " + getName() + " -> fermeture du proc");
	    erreur = "ERROR: " + getName();
	    erreur += ": connection bd impossible (acces impossible): ";
	    erreur = erreur + sqle + " => fermeture du proc";
	    outLog.println(erreur);
	    return;
	}
	// Rcupre le plus grand identificateur de page utilis dans la base de donnes
	try {
	    id_max = maxId();
	}
	catch (SQLException e) {
	    return;
	}
	while (!to_finish) { // traitement d'une page
	    // Recupration d'une page  parser
	    id_lien = nextId();
	    if (id_lien < 0)
		return;
	    // Suppression ventuelle de la page
	    try {
		verifLien();
	    } catch (SQLException e) {return;}
	}
	log = getName() + ": fermeture du proc (demande de l'utilisateur)";
	outLog.println(log);
	System.out.println(log);
	bd.closeConnection();
	return;
    }
}
