/**
 * Classe (particulire  l'application du crawler) implmentant une pile FIFO spcialise dans la 
 * manipulation d'objet de type <code>PageWeb</code>.<p>
 * Un seul objet de ce type est partag entre tous les processus de navigation (<code>NavThread</code>).
 * Il est donc important que ceux-ci puissent accder  la pile de faon concurente. De plus losrque la
 * pile est vide, les processus demandant une nouvelle page web sont mis en attente.<p>
 * Remarque: Dans cette page on appelle 'page web' un ensemble constitu par l'url d'une page html et
 * sa profondeur par rapport  une page de dmarrage du crawler. Concrtement ces informations sont 
 * contenues dans des objets de type <code>PageWeb</code>.
 * @author Martin Ludovic
 * @see PageWeb
 */
public class WebFifo extends LockedFifo {
    /** Stocke la profondeur maximale de recherche autorise.<br>
     *  Si cette profondeur est de 0, alors la profondeur maximale est illimite. */
    protected int depth_max;

    /**
     * Construit un nouvel objet <code>WebFifo</code> ayant la taille par dfaut.<p>
     * Ce constructeur se contente de faire appel au constructeur sans paramtre de la classe mre
     * et ne pose aucune limite de profondeur d'exploration.
     */
    public WebFifo () {
	super();
	depth_max = 0;
    }

    /**
     * Construit un nouvel objet <code>WebFifo</code> ayant la taille par dfaut.<p>
     * @param depth_max profondeur maximale d'exploration  partir d'une des pages de dmarrage du crawler.
     * @see #depth_max
     * @throws IllegalArgumentException Si jamais la valeur passe en argument est < 0.
     */
    public WebFifo (int depth_max) throws IllegalArgumentException {
	super();
	if (depth_max < 0)
	    throw new IllegalArgumentException("depth_max ne peut prendre de valeurs ngatives.");
	this.depth_max = depth_max;
    }

    /**
     * Construit un nouvel objet <code>WebFifo</code> ayant la taille spcifie en argument.<p>
     * @param depth_max profondeur maximale d'exploration  partir d'une des pages de dmarrage du crawler.
     * @param taille_max la capacit maximale de la pile FIFO
     * @throws IllegalArgumentException Si:
     *   <ul><li>la valeur passe  <code>depth_max</code> en argument est < 0.
     *   <li>la valeur passe  <code>taille_max</code> en argument est < 1.</ul>
     * @see #depth_max
     */
    public WebFifo (int depth_max, int taille_max) throws IllegalArgumentException {
	super(taille_max);
	if (depth_max < 0)
	    throw new IllegalArgumentException("depth_max ne peut prendre de valeurs ngatives.");
	this.depth_max = depth_max;
    }

    /**
     * Retourne la page web en bas de la pile et la supprime de la pile.<p>
     * Remarque: cette mthode se contente de caster le rsultat de la mthode <code>Object get()</code>
     * de la classe mre.
     * @param nom_proc Le nom du processus voulant accder  la pile (util pour connatre l'tat de
     * celui-ci).<br> Si le paramtre vaut <code>null</code> aucun message n'est affich.
     * @return La page web en bas de la pile.
     * @see LockedFifo#get(String)
     */
    public PageWeb getPage (String nom_proc) {
	return (PageWeb)super.get(nom_proc);
    }

    /**
     * Insre une page web en haut de la pile.<p>
     * Avant d'insrer rellement la page dans la pile (ie de faire appel  la mthode <code>put(Object)
     * </code> de la classe mre), cette mthode fait deux vrifications:
     * <ul>
     *   <li>la profondeur de la page insre est infrieure  la profondeur maximale autorise.
     *   <li>la page insre ne se trouve pas dj dans la pile
     * </ul>
     * Except pour ces vrifications, cette mthode est identique  la mthode correspondante dans
     * la classe mre.
     * @param pw la nouvelle page web  insrer
     * @return <ul>
     *           <li>0 si l'objet a t insre correctement
     *           <li>1 si la profondeur de la page insre est trop grande
     *           <li>2 si la pile contient dj la page  insrer
     *           <li>-1 si la pile est pleine (la page web est ignore)
     *         <ul>
     * @see LockedFifo#put(Object)
     */
    public int put (PageWeb pw) {
	// Vrifie si la profondeur de la page n'est pas suprieur  la profondeur max autorise
	/*
	if ((pw.depth>depth_max) && (depth_max!=0))
	    return 1;
	*/
	// Vrifie que la page web n'est pas dj prsente dans la pile
	/*
	if (liste.contains(pw))
	    return 2;
	*/
	return super.put(pw);
    }
}

