import java.io.*;
import javax.swing.text.html.*;
import java.util.*;
import java.net.*;

/**
 * @author M. Sighireanu
 * @author Martin Ludovic
 */
public class CrawlerCallbackHandler extends HTMLEditorKit.ParserCallback {
    protected boolean isInsideLink = false;
    /** Tableau des tables de hachage dans lesquelles vont tre stocks tous les liens. */
    protected Hashtable[] ht_tab;
    /** Tableau des dictionnaires des mots  ignors lors du parsage. */
    protected Properties[] dicos;
    /** Le chemin du rpertoire contenant le fichier source de la page  parser.<p>
     *  Par exemple si la page  parser est <code>http://www.server.fr/rep1/rep2/fic.html</code> alors
     *  <code>root</code> contiendra <code>http://www.server.fr/rep1/rep2/</code>.<p>
     *  Ce champ est ncessaire pour transformer les chemins relatifs rencontrs dans cette page
     *  en chemin absolus. */
    protected String root;
    /** Le chemin du rpertoire contenant le fichier source de la page  parser sans l'url du serveur.<p>
     *  Par exemple si la page  parser est <code>http://www.server.fr/rep1/rep2/fic.html</code> alors
     *  <code>short_root</code> contiendra <code>/rep1/rep2/</code>.<p>
     *  Ce champ est ncessaire pour transformer les chemins relatifs rencontrs dans cette page
     *  en chemin absolus pour les images. */
    protected String short_root;
    /** Le lien en cours de traitement.<p>
     *  Cette valeur n'est valide que lorsque <code>isInsideLink</code> vaut <code>true</code>.*/
    protected String cur_link;

    /**
     * Construit un nouvel objet <code>CrawlerCallbackHandler</code>.
     * @param ht_tab Tableau des tables de hachage dans lesquelles vont tre stocks tous les liens.
     * @param dicos Tableau des dictionnaires des mots  ignors lors du parsage.
     * @param root Le chemin du rpertoire contenant le fichier source de la page  parser.
     * @see #ht_tab
     * @see #dicos
     * @see #root
     */
    public CrawlerCallbackHandler(Hashtable[] ht_tab,Properties[] dicos,String root) {
	this.ht_tab = ht_tab;
	this.dicos = dicos;
	this.root = root;
	URL url = null;
	try {
	    url = new URL(root+"a.html");
	} catch (MalformedURLException e) {/* Dj vrifi */}
	short_root = url.getPath();
	short_root = short_root.substring(0,short_root.length()-6);
    }

    /**
     * Rpre le dbut d'un tag double (c'est  dire un tag qu'on doit refermer).
     * Cette mthode ne traite que les tag <code>A</code> c'est  dire les hyperliens.<p>
     * Une fois qu'un tag de ce type est trouv, la mthode se dcompose en trois parties:
     * <ul>
     *   <li>elle regarde de quel type de lien il s'agit (<code>http</code> ou <code>mailto</code>)
     *   <li>dans le cas d'un lien <code>http</code>, elle reconstitue son url canonique grce  la
     *       mthode <code>String simplifyUrl(String)</code>
     *   <li>elle stocke le lien dans le champ <code>cur_link</code> pour le rendre accessible 
     *       la mthode <code>void handleText(char[],int)</code>
     * </ul>
     * @see #simplifyUrl(String)
     */
    public void handleStartTag(HTML.Tag tag, 
                               javax.swing.text.MutableAttributeSet set, 
                               int i) {
        // System.out.println("start: " + tag.toString());
        if (tag.equals(HTML.Tag.A)) {
            isInsideLink = true;
            Object link = set.getAttribute(HTML.Attribute.HREF);
            if (link != null) {
		// Dtermine le type de lien trouv
		if (((String)link).startsWith("http://")) {
		    // Lien absolu de type http
		    try {
			cur_link = simplifyUrl((String)link);
		    }
		    catch (MalformedURLException e) {
			cur_link = null;
		    }
		    catch (FileNotFoundException e) {
			cur_link = null;
		    }
		}
		else if (((String)link).startsWith("ftp://")) {
		    // Lien absolu de type http
		    try {
			cur_link = simplifyUrl((String)link);
		    }
		    catch (MalformedURLException e) {
			cur_link = null;
		    }
		    catch (FileNotFoundException e) {
			cur_link = null;
		    }
		}
		else if (((String)link).startsWith("mailto:")) {
		    // Lien de type mail
		    cur_link = (String)link;
		}
		else {
		    // Lien relatif de type http
		    try {
			cur_link = simplifyUrl(root + (String)link);
		    }
		    catch (MalformedURLException e) {
			cur_link = null;
		    }
		    catch (FileNotFoundException e) {
			cur_link = null;
		    }
		}
	    }
        }
    }

    /**
     * Rpre la fin d'un tag double (c'est  dire un tag qu'on doit refermer).
     * Cette mthode ne traite que les tag <code>/A</code> c'est  dire les hyperliens.<p>
     * Elle permet d'indiquer  la mthode <code>void handleText(char[],int)</code> qu'on est
     * sorti du lien.
     */
    public void handleEndTag(HTML.Tag tag, int i) {
        // System.out.println("end: " + tag.toString());
        if (tag.toString().compareToIgnoreCase("a") == 0) {
            isInsideLink = false;
        }
    }

    /**
     * Rpre les balises simples.<br>
     * Cette mthode ne considre que les balises de type <code>IMG</code> dont le champ alt n'est
     * pas vide.<p>
     * Lorsqu'une telle balise est trouve:
     * <ul>
     *   <li>Traitement du texte du champ <code>ALT</code> : transformation en minuscules, supprime les accents
     *       et dcoupage en tableau de mots.
     *   <li>Reconstitution du chemin ou de l'url de l'image:
     *     <ul>
     *       <li>Cas classique: l'image est stocke sur le serveur de la page<br>
     *           Dans ce cas, on reconstitue son chemin canonique depuis la racine du site grce  la 
     *           mthode <code>simplifyPath(String)</code>.
     *       <li>Cas rare: l'image est stocke sur un autre serveur (un autre site).<br>
     *           Dans ce cas, on reconstitue son url canonique grce  la mthode <code>
     *           simplifyUrl(String)</code>.
     *     </ul>
     *   <li>Insertion du lien dans la table de hachage des liens images grce  la 
     *       mthode <code>void insert (Hashtable,String,String[])</code>
     * </ul>
     * @see #simplifyUrl(String)
     * @see #simplifyPath(String)
     * @see #insert(Hashtable,String,String[])
     */
    public void handleSimpleTag(HTML.Tag tag, 
                                javax.swing.text.MutableAttributeSet set, 
                                int i) {
        // System.out.println("simple: " + tag.toString());
        if (tag.equals(HTML.Tag.IMG)) {
            Object src = set.getAttribute(HTML.Attribute.SRC);
            Object alt = set.getAttribute(HTML.Attribute.ALT);
	    // On ne considre pas les images dont le champ alt est vide
            if ((src != null) && (alt != null) && (alt != "")) {
		String mots[];
		String comment;
		// Tranformation du alt en minuscules
		comment = ((String)alt).toLowerCase();
		// Supprime les accents
		comment = comment.replaceAll("[]","e");
		comment = comment.replaceAll("[]","o");
		comment = comment.replaceAll("[]","i");
		comment = comment.replaceAll("[]","u");
		comment = comment.replaceAll("[]","a");
		// Dcoupage du alt en un tableau de mots
		comment = comment.replaceAll("[,:/]"," ");
		mots = comment.split("[ \\.'`\"|&~\\(\\)%\\*+\\{\\}$@]");
		// Rcupration du chemin absolu de l'image
		String lien;
		if (((String)src).startsWith("/")) {
		    try {
			lien = simplifyPath((String)src);
		    }
		    catch (FileNotFoundException e) {
			return;
		    }
		}
		else if (((String)src).startsWith("http://")) {
		    try {
			lien = simplifyUrl((String)src);
		    } 
		    catch (FileNotFoundException e) {return;}
		    catch (MalformedURLException e) {return;}
		}
		else if (((String)src).startsWith("ftp://")) {
		    try {
			lien = simplifyUrl((String)src);
		    } 
		    catch (FileNotFoundException e) {return;}
		    catch (MalformedURLException e) {return;}
		}
		else {
		    try {
			lien = simplifyPath(short_root + (String)src);
		    }
		    catch (FileNotFoundException e) {
			return;
		    }
		}
		// Insertion du lien courant associ  ces mots dans la table de hachage
		try {
		    //System.out.println("lien trouve:" + lien);
		    insert(ht_tab[2],lien,mots);
		} catch (NullPointerException e) {return;}
	    }
        }
    }

    /**
     * Lit le texte html associ  un lien.<p>
     * Cette mthode ne considre que le texte situ entre les tags doubles <code>A</code> et
     * <code>/A</code> grce au champ <code>isInsideLink</code>.<p>
     * Elle se dcompose en deux parties:
     * <ul>
     *   <li>traitement du texte entre les balises : transformation en minuscules, supprime les accents
     *       et dcoupage des phrases en tableau de mots.
     *   <li>insertion du lien dans la table de hachage correspondant  son type grce  la 
     *       mthode <code>void insert (Hashtable,String,String[])</code>
     * </ul>
     * @see #isInsideLink
     * @see #insert(Hashtable,String,String[])
     */
    public void handleText(char[] aChar, int i) {
        if (isInsideLink) {
	    String mots[];
	    String comment = new String(aChar);
	    if ((comment == null) || (comment.equals("")))
		return;
	    // Tranformation du commentaire du lien en minuscules
	    comment = comment.toLowerCase();
	    // Supprime les accents
	    comment = comment.replaceAll("[]","e");
	    comment = comment.replaceAll("[]","o");
	    comment = comment.replaceAll("[]","i");
	    comment = comment.replaceAll("[]","u");
	    comment = comment.replaceAll("[]","a");
	    // Dcoupage du alt en un tableau de mots
	    comment = comment.replaceAll("[,:/]"," ");
	    mots = comment.split("[ \\.'`\"|&~\\(\\)%\\*+\\{\\}$@]");
	    // Insertion du lien courant associ  ces mots dans la table de hachage
	    try {
		if ((cur_link.startsWith("http")) ||
		    (cur_link.startsWith("ftp")) ) {
		    //System.out.println("lien trouve:" + cur_link);
		    insert(ht_tab[0],cur_link,mots);
		}
		else {
		    //System.out.println("lien trouve:" + cur_link);
		    insert(ht_tab[1],cur_link,mots);
		}
	    } catch (NullPointerException e) {return;}
	}
    }

    /**
     * Transforme un chemin absolu de fichier en un chemin canonique, c'est  dire un chemin
     * commenant par '<code>/</code>' et ne comportant aucun '<code>.</code>' ou '<code>..</code>'.<p>
     * Par exemple:<br> Transforme<br> <code>/./rep1/rep2/../rep3/rep4/../../truc</code><br> en<br> 
     * <code>/rep1/truc</code>.
     * @param path Le chemin absolu  cononiser.
     * @return Le chemin canonique
     * @throws FileNotFoundException Si le chemin absolu pass en argument est incorrect (par ex <code>
     * /rep1/.././../truc</code>).
     */
    protected String simplifyPath (String path) throws FileNotFoundException {
	String[] parts;
	boolean[] valide;
	int i,j;

	// Supprime le premier '/'
	if (path.startsWith("/"))
	    path = path.substring(1);
	// Dcoupage du chemin en parties
	parts = path.split("/");
	valide = new boolean[parts.length];
	// Elimine tous les lments inutils du chemin
	for (i=0 ; i<parts.length ; i++) {
	    if (parts[i].startsWith(".")) {
		valide[i] = false;
		if (parts[i].equals("..")) {
		    // On remonte d'un niveau dans l'arborescence
		    if (i > 0)
			j = i-1;
		    else
			// On ne pas avoir '..' dans la racine
			throw new FileNotFoundException(".. a la racine");
		    while (valide[j] != true) {
			if (j > 0)
			    j--;
			else
			    // On est remont au-del de la racine
			    throw new FileNotFoundException("trop de ..");
		    }
		    valide[j] = false;
		}
		else if (!parts[i].equals(".")) {
		    // Il s'agit peut tre d'un fichier dont le nom commence par '.'
		    valide[i] = true;
		}
	    }
	    else if (parts[i].equals("")) {
		/* Cas trs particulier: Il peut arriver qu'un lien relatif soit absolu. C'est  dire
		 * qu'il est relatif au site dans lequel on se trouve mais que son chemin soit absolu.
		 * Par ex: <A HREF="/rep1/rep2/index.html">Blabla</A>. Or comme la valeur de la var
		 * path est parfois une concatnation entre le chemin d'une url et le lien, cela se
		 * traduit par path=/rep4/rep5//rep1/rep2/index.html ; ce qu'il faut traduire
		 * en /rep1/rep2/index.html */
		for (j=0 ; j<i+1 ; j++)
		    valide[j] = false;
	    }
	    else
		valide[i] = true;
	}
	// Reconstitue le chemin
	String simp_path = "";
	if (parts.length > 0) {
	    for (i=0 ; i<parts.length-1 ; i++) {
		if (valide[i])
		    simp_path += "/" + parts[i];
	    }
	    try {
		if (!(parts[parts.length-1].equals(".") || parts[parts.length-1].equals("..")))
		    simp_path += "/" + parts[parts.length-1];
	    }
	    catch (ArrayIndexOutOfBoundsException e) {
		System.out.println("ArrayIndexOutOfBoundsException: simp_path="+simp_path);
		System.out.println("ArrayIndexOutOfBoundsException: parts.length="+parts.length);
		for (i=0 ; i<parts.length ; i++) {
		    System.out.println("ArrayIndexOutOfBoundsException: parts["+i+"]="+parts[i]);
		}
		throw e;
	    }
	}
	if (simp_path.equals(""))
	    simp_path = "/";
	// Retourne le chemin simplifi
	return simp_path;
    }

    /**
     * Transforme une url absolue en une url canonique, c'est  dire une url dont le chemin est
     * canonique.<p>
     * Le chemin d'une url est par exemple pour <code>http://www.somewhere.com/rep1/rep2/index.html
     * </code>,  <code>/rep1/rep2/index.html</code>.<p>
     * Pour savoir ce qu'est un chemin canonique, voir <code>simplifyPath(String)</code>.
     * @param url L'url  canoniser.
     * @return L'url canonique
     * @throws FileNotFoundException Si le chemin de l'url passe en argument est incorrect (par ex <code>
     * /rep1/.././../truc</code>).
     * @throws MalformedURLException Si l'url est mal forme.
     * @see #simplifyPath(java.lang.String)
     */
    protected String simplifyUrl (String s_url) throws FileNotFoundException,MalformedURLException {
	String path,simp_path;
	URL url;

	url = new URL(s_url);
	path = url.getPath();
	simp_path = simplifyPath(path);
	if (url.getQuery() != null) {
	    /*
	    System.out.println("host = " + url.getHost());
	    System.out.println("path complet = " + url.getPath());
	    System.out.println("path simp = " + simp_path);
	    System.out.println("query = " + url.getQuery());
	    System.out.println("root = " + root);
	    System.out.println("short_root = " + short_root + "\n");
	    Rem : une adresse de type http://docs.sun.com/?q=kinit&p=/doc/806-0624/6j9vek58e&a=view est valide
	    */
	    return url.getProtocol() + "://" + url.getHost() + simp_path + "?" + url.getQuery();
	}
	else
	    return url.getProtocol() + "://" + url.getHost() + simp_path;
    }

    /**
     * Dtecte la langue employe dans la page qu'on est en train de parser.<p>
     * Cette mthode reste entirement  faire (dans une version ultrieure) et pour le moment le dictionnaire
     * franais est systmatiquement retourn.
     * @return Une constante symbolisant la langue de la page web.
     * @see Crawler#DICO_FR
     * @see Crawler#DICO_EN
     * @see Crawler#DICO_DE
     */
    protected int language () {
	// Lit normalement la page web...
	// ... et slectionne une langue
	return Crawler.DICO_FR;
    }

    /**
     * Insre dans une table de hachage un lien associ  un ou plusieurs mot(s)<p>
     * Tous les mots passs en argument  la mthode ne sont pas forcment insrs dans la table de
     * hachage. En effet, les mots trouvs dans le dictionnaire de la langue de la page sont
     * ignors.<p>
     * Cette fonction peut tre utilise aussi bien par <code>getPages()</code>, <code>getMails()</code>
	 * ou <code>getPict()</code>.<p>
     * Remarque: Le rsultat de plusieurs appels  cette mthode pour le mme lien est quivalent  un seul
     * appel avec dont le troisime paramtre serait la concatnation du troisime paramtre de chaque
     * appel spar.
     * @param ht la table de hachage dans laquelle on veut raliser l'insertion
     * @param lien le lien peut tre une url de page web, une adresse mail ou un nom (avec chemin absolu) d'image
     * @param mots un ensemble de mots rfrenant le lien
     * @throws NullPointerException si l'un des paramtres est null
     * @see #dicos
     */
    protected void insert (Hashtable ht, String lien, String[] mots) throws NullPointerException {
	Hashtable ht_val;
	String mot;
	int lang;

	/// Vrifie la validit des arguments
	if ((ht==null) || (lien==null) || (mots==null))
	    throw new NullPointerException();
	/// Dtecte la langue de la page
	lang = language();
	/// Insre les liens et les mots dans la table de hachage
	// Est-ce que le lien existe? (si ce n'est pas le cas: cration du lien dans la HT)
	if (!ht.containsKey(lien)) {
	    ht.put(lien,new Hashtable());
	}
	ht_val = (Hashtable)ht.get(lien);
	// Enumration de tous les mots associs au lien
	for (int i=0 ; i<mots.length ; i++) {
	    mot = mots[i];
	    // Le mot est-il vide ?
	    if ((mot == null) || mot.equals(""))
		continue;
	    // Le mot appartient-il au dictionnaire des mots  ignorer?
	    if (dicos[lang].containsKey(mot))
		continue;
	    // Le mot rfrence-t-il dj le lien?
	    if (ht_val.containsKey(mot)) {
		//Le mot rfrence au moins une fois dj le lien -> augmente le nombre d'occurence
		Entier e = (Entier)ht_val.get(mot);
		e.increase();
	    }
	    else {
		ht_val.put(mot,new Entier(1));
	    }
	}
    }
}

