/****************************************************************************
 * Les rgles
 ****************************************************************************/

% mere(page_mere,page_fille)
mere(M,F):-app_href(M,Id_href,_,_),href(2,F,Id_href).

% filles(+page,-ensemble_de_ces_filles)
filles(Page,Set):-
    findall(Y,mere(Page,Y),L),
    list_to_set(L,Set).

% descendant(+page_ancetre,-page_descendante)
descendant(Start,D) :-
    filles(Start,Filles),               % Filles={pages a visiter}
    merge_set([Start], Filles, Closed), % Closed={pages vues (avec 1 iteration d'avance)}
    descendant_iter(Filles,Closed,D).
% descendant_iter(liste_pages_a_visiter,liste_pages_vues)
descendant_iter([X|_],_,X).
descendant_iter([X|Open1],Closed,Y) :-
    filles(X,Filles),
    subtract(Filles,Closed,Jamais_vues), % Jamais_vues={pages non encore visitees}
    append(Jamais_vues,Open1,Open2),     % Open2={pages a visiter dans la prochaine iteration}
    merge_set(Closed,Filles,Closed1),    % Closed1={pages vues (avec 1 iteration d'avance)}
    descendant_iter(Open2,Closed1,Y).

% meres(+page,-ensemble_de_ces_meres)
meres(Page,Set):-
    findall(Y,mere(Y,Page),L),
    list_to_set(L,Set).

% ancetre(+page_ancetre,-page_ancetre)
ancetre(Start,A) :-
    meres(Start,Meres),                % Meres={pages a visiter}
    merge_set([Start], Meres, Closed), % Closed={pages vues (avec 1 iteration d'avance)}
    ancetre_iter(Meres,Closed,A).
% ancetre_iter(liste_pages_a_visiter,liste_pages_vues)
ancetre_iter([X|_],_,X).
ancetre_iter([X|Open1],Closed,Y) :-
    meres(X,Meres),
    subtract(Meres,Closed,Jamais_vues),  % Jamais_vues={pages non encore visitees}
    append(Jamais_vues,Open1,Open2),     % Open2={pages a visiter dans la prochaine iteration}
    merge_set(Closed,Meres,Closed1),     % Closed1={pages vues (avec 1 iteration d'avance)}
    ancetre_iter(Open2,Closed1,Y).


/****************************************************************************
 *   COMPOSITION DES PREDICATS:
 * serveur(url_serveur,id_serveur).
 * img(id_serveur,path,nom,id_img).
 * page(id_serveur,url_page,date,id_page).
 * fichier(path,nom,id_fichier).
 * doc_txt(id_fichier).
 * protocole(prefixe,id_proto).
 * heberge(id_fichier,id_proto,id_serveur).
 * mot_txt(mot,id_txt).
 * txt_doc(id_txt,id_fichier,occurrence).
 * txt_page(id_txt,id_page,occurrence).
 * mot_desc(mot,id_desc).
 * href(type_fic,id_fic,id_href).
 * app_href(id_page,id_href,id_desc,occurrence).
 * mail(mail,id_mail).
 * app_mail(id_page,id_mail,id_desc,occurrence).
 * app_img(id_page,id_img,id_desc,occurrence).
 ****************************************************************************/

/**
 * Table serveur
 */

serveur('java.sun.com',1).
serveur('www.sun.com',15).
serveur('developer.java.sun.com',39).
serveur('ftp.java.sun.com',97).
serveur('docs.sun.com',118).
serveur('www.w3.org',128).
serveur('servlet.java.sun.com',149).
serveur('wwws.sun.com',150).
serveur('search.java.sun.com',170).
serveur('www201.ikiosk.com',185).
serveur('www.jcp.org',190).
serveur('jcp.org',379).
serveur('www.ietf.org',385).
serveur('ftp.omg.org',475).
serveur('ftp.javasoft.com',510).
serveur('wireless.java.sun.com',513).
serveur('forum.java.sun.com',514).
serveur('research.sun.com',553).
serveur('www.confluent.fr',555).
serveur('www.winzip.com',559).
serveur('home.mcom.com',566).
serveur('developer.netscape.com',578).
serveur('home.netscape.com',579).
serveur('www.amazon.com',599).
serveur('cseng.aw.com',602).
serveur('www.awl.com',605).
serveur('12.31.179.2',615).
serveur('developer.apple.com',664).
serveur('cseng.awl.com',679).
serveur('archives.java.sun.com',712).
serveur('vig.pearsoned.com',791).
serveur('cgi.omg.org',866).
serveur('suned.sun.com',923).
serveur('www.informit.com',951).
serveur('www.digitalguru.com',963).
serveur('sunonedev.sun.com',999).
serveur('www.corba.org',1008).
serveur('www.nvidia.com',1031).
serveur('www.research.avayalabs.com',1037).
serveur('www.atitech.com',1057).
serveur('www.omg.org',1084).
serveur('www.jguru.com',1086).
serveur('sunsolve.sun.com',1116).
serveur('www.adobe.com',1166).
serveur('support.installshield.com',1199).
serveur('www.microsoft.com',1200).
serveur('support.microsoft.com',1201).
/* (47 entrees) */


/**
 * Table img
 */

img(1,'/j2se/1.4.1/docs/images/','download.arrow.gif',1).
img(1,'/j2se/1.4.1/docs/images/','javalogo52x88.gif',2).
img(1,'/j2se/1.4.1/docs/images/','sunlogo64x30.gif',3).
img(1,'/images/','v4_java_sun_com.gif',4).
img(1,'/images/','v3_sun_logo.gif',5).
img(1,'/j2se/1.4.1/ja/images/','Japanese12B.gif',6).
img(1,'/images/','v3_print.gif',7).
img(1,'/images/','v4_java_logo.gif',8).
img(15,'/im/home/','masthead_slogan.gif',9).
img(15,'/im/','sun_logo.gif',10).
img(15,'/pics/promos/','b5_nc_03_q2.gif',11).
img(15,'/2003-0422/images/','b3_nascar-java.jpg',12).
img(15,'/pics/promos/','b5_sun_fire_blade_platform.gif',13).
img(15,'/pics/promos/','b5_get_java.gif',14).
img(15,'/pics/promos/','b5_why_not.gif',15).
img(15,'/im/','a.gif',16).
img(1,'/j2se/1.4.1/images/','view.option.gif',17).
img(1,'/j2se/1.4.1/docs/guide/rmi/images/','javalogo52x88.gif',18).
img(1,'/j2se/1.4.1/images/','download.option.gif',19).
img(1,'/j2se/images/','j2se_sm.gif',20).
img(1,'/images/','Japanese12B.gif',21).
img(1,'/images/misc/','arrow.gif',22).
img(39,'/developer/onlineTraining/new2java/','professionals.gif',23).
img(39,'/developer/onlineTraining/new2java/','new.gif',24).
img(39,'/developer/onlineTraining/new2java/','step3.gif',25).
img(39,'/developer/onlineTraining/new2java/','step2.gif',26).
img(39,'/developer/onlineTraining/new2java/','step1.gif',27).
img(39,'/images/','v4_java_sun_com.gif',28).
img(39,'/images/','v4_java_logo.gif',29).
img(1,'/images/','j1sf.register.gif',30).
img(39,'/developer/onlineTraining/new2java/','customed.gif',31).
img(39,'/developer/onlineTraining/new2java/','library.gif',32).
img(39,'/images/','v3_print.gif',33).
img(39,'/images/','v3_sun_logo.gif',34).
img(1,'/images/','j2se_homepage_button.jpg',35).
img(39,'/images/','T7.gif',36).
img(39,'/developer/onlineTraining/new2java/','step4.gif',37).
img(1,'/images/','hostthebutton.gif',38).
img(1,'/images/','j2ee_homepage_button.jpg',39).
img(1,'/images/rollover/','bnr.socketjxta.gif',40).
img(1,'/images/','j2me_homepage_button.jpg',41).
img(1,'/images/misc/','bullet.gif',42).
img(1,'/docs/books/tutorial/images/','dukesign.GIF',43).
img(1,'/docs/books/tutorial/images/bookcovers/','3ebook_sm.gif',44).
img(1,'/docs/books/tutorial/images/bookcovers/','continued_sm.gif',45).
img(1,'/docs/books/tutorial/images/','javatut.gif',46).
img(1,'/docs/books/tutorial/images/','download.gif',47).
img(1,'/docs/books/tutorial/images/bookcovers/','swing_sm.gif',48).
img(1,'/images/','smi.logo.gif',49).
img(1,'/images/','JAVA.85.GIF',50).
img(1,'/images/','endicon.duke.gif',51).
img(1,'/images/','getacro.gif',52).
img(1,'/images/navbar/','side.tab.docs.gif',53).
img(1,'/images/navbar/','side.tab.developer.gif',54).
img(1,'/images/navbar/','side.tab.products.gif',55).
img(1,'/images/banners/','azindex.gif',56).
img(1,'/images/navbar/','side.tab.support.gif',57).
img(1,'/images/banners/','help.gif',58).
img(1,'/images/navbar/','side.tab.community.gif',59).
img(1,'/images/banners/','thesourceforjava.gif',60).
img(1,'/images/logos/','javalogo52x88.gif',61).
img(1,'/images/navbar/','side.tab.case.gif',62).
img(1,'/images/navbar/','side.tab.news.gif',63).
img(1,'/images/buttons/','lgsun.gif',64).
img(1,'/j2se/1.4.1/docs/guide/corba/images/','omg_memb.gif',65).
/* (65 entrees) */


/**
 * Table page
 */

page(1,'/j2se/1.4.1/docs/guide/xml/index.html',0,1).
page(1,'/j2se/1.4.1/docs/guide/serialization/index.html',0,2).
page(1,'/j2se/1.4.1/docs/guide/awt/index.html',0,3).
page(1,'/j2se/1.4.1/compatibility.html',0,4).
page(1,'/j2se/1.4.1/docs',0,5).
page(1,'/j2se/1.4.1/docs/guide/util/logging/index.html',0,6).
page(1,'/j2se/1.4.1/docs/guide/beans/index.html',0,7).
page(1,'/j2se/1.4.1/docs/relnotes/features.html',0,8).
page(1,'/j2se/1.4.1/docs/tooldocs/tools.html',0,9).
page(1,'/j2se/1.4.1/docs/guide/jdbc/index.html',0,10).
page(1,'/docs/books/jls/second_edition/html/j.title.doc.html',0,11).
page(1,'/j2se/1.4.1/docs/guide/2d/index.html',0,12).
page(1,'/j2se/1.4.1/docs/guide/security/index.html',0,13).
page(1,'/j2se/1.4.1/docs/relnotes/demos.html',0,14).
page(15,'/',0,15).
page(1,'/j2se/1.4.1/docs/guide/jni/index.html',0,16).
page(1,'/j2se/1.4.1/docs/guide/swing/index.html',0,17).
page(1,'/docs/windows_format.html',0,18).
page(1,'/j2se/1.4.1/fixedbugs/index.html',0,19).
page(1,'/j2se/1.4.1/docs/guide/jws/index.html',0,20).
page(1,'/j2se/1.4.1/changes.html',0,21).
page(1,'/j2se/1.4.1/docs/guide/imf/index.html',0,22).
page(1,'/j2se/1.4.1/docs/guide/collections/index.html',0,23).
page(1,'/',0,24).
page(1,'/docs/codeconv/index.html',0,25).
page(1,'/j2se/1.4.1/download.html',0,26).
page(1,'/j2se/1.4.1/docs/guide/standards/index.html',0,27).
page(1,'/j2se/1.4.1/docs/guide/sound/index.html',0,28).
page(1,'/products/jdk/faq.html',0,29).
page(1,'/j2se/1.4.1/docs/guide/jpda/index.html',0,30).
page(1,'/j2se/1.4.1/docs/api/index.html',0,31).
page(1,'/j2se/1.4.1/docs/relnotes/contacts.html',0,32).
page(1,'/j2se/1.4.1/docs/guide/net/index.html',0,33).
page(1,'/j2se/1.4.1/docs/guide/rmi/index.html',0,34).
page(1,'/j2se/1.4.1/docs/relnotes/SMICopyright.html',0,35).
page(1,'/cgi-bin/bugreport.cgi',0,36).
page(1,'/j2se/1.4.1/docs/api/overview-summary.html',0,37).
page(1,'/docs/books/effective/index.html',0,38).
page(39,'/developer/onlineTraining/new2java/index.html',0,39).
page(1,'/docs/books/jls/index.html',0,40).
page(1,'/j2se/1.4.1/docs/guide/jps/index.html',0,41).
page(1,'/docs/books/vmspec/index.html',0,42).
page(1,'/j2se/1.4.1/docs/guide/versioning/index.html',0,43).
page(1,'/j2se/1.4.1/docs/guide/intl/index.html',0,44).
page(1,'/j2se/1.4.1/docs/guide/vm/index.html',0,45).
page(1,'/j2se/1.4.1/docs/guide/extensions/index.html',0,46).
page(1,'/j2se/1.4.1/docs/guide/reflection/index.html',0,47).
page(1,'/docs/books/vmspec/2nd-edition/html/VMSpecTOC.doc.html',0,48).
page(1,'/docs/books/chanlee/second_edition/vol1/examples.html',0,49).
page(1,'/docs/books/tutorial/index.html',0,50).
page(1,'/j2se/1.4.1/docchanges.html',0,51).
page(39,'/developer/onlineTraining',0,52).
page(1,'/j2se/1.4.1/docs/guide/lang/index.html',0,53).
page(1,'/j2se/1.4.1/docs/guide/jndi/index.html',0,54).
page(1,'/docs/books/chanlee',0,55).
page(1,'/features/2001/09/goslingrev.html',0,56).
page(1,'/j2se/1.4.1/docs/guide/resources/index.html',0,57).
page(1,'/j2se/1.4.1/docs/guide/plugin/index.html',0,58).
page(1,'/j2se/1.4.1/docs/guide/performance/index.html',0,59).
page(1,'/docs/books/chanlee/supplement/examples.html',0,60).
page(1,'/products/jdk/faq/faq-sun-packages.html',0,61).
page(1,'/j2se/1.4.1/relnotes.html',0,62).
page(1,'/j2se/1.4.1/docs/guide/refobs/index.html',0,63).
page(1,'/j2se/1.4.1/docs/guide/misc/index.html',0,64).
page(1,'/j2se/1.4.1/docs/guide/imageio/index.html',0,65).
page(1,'/j2se/1.4.1/docs/guide/dragndrop/index.html',0,66).
page(1,'/j2se/1.4.1/install-solaris.html',0,67).
page(1,'/j2se/1.4.1/docs/guide/math/index.html',0,68).
page(1,'/j2se/1.4.1/docs/guide/rmi-iiop/index.html',0,69).
page(1,'/j2se/1.4.1/docs/guide/corba/index.html',0,70).
page(1,'/j2se/1.4.1/docs/guide/jvmpi/index.html',0,71).
page(1,'/j2se/1.4.1/download-pdf-ps.html',0,72).
page(1,'/j2se/1.4.1/docs/guide/nio/index.html',0,73).
page(1,'/docs/redist.html',0,74).
page(1,'/j2se/1.4.1/install-linux.html',0,75).
page(1,'/j2se/1.4.1/search.html',0,76).
page(1,'/docs/books/chanlee/second_edition/examples.html',0,77).
page(1,'/j2se/1.4.1/docs/guide/idl/index.html',0,78).
page(1,'/j2se/1.4.1/README.html',0,79).
page(1,'/j2se/1.4.1/install-windows.html',0,80).
page(1,'/aboutJava/training',0,81).
page(1,'/j2se/1.4.1/docs/guide/access/index.html',0,82).
page(1,'/j2se/1.4.1/docs/guide/jar/index.html',0,83).
page(1,'/j2se/1.4.1/docs/api/java/awt/event/package-summary.html',0,84).
page(1,'/j2se/1.4.1/docs/guide/awt/1.3/designspec/index.html',0,85).
page(1,'/j2se/1.4.1/docs/api/java/awt/doc-files/FocusSpec.html',0,86).
page(1,'/j2se/1.4.1/docs/index.html',0,87).
page(1,'/j2se/1.4.1/docs/guide/awt/AWTChanges.html',0,88).
page(1,'/j2se/1.4.1/docs/api/java/awt/datatransfer/package-summary.html',0,89).
page(1,'/j2se/1.4.1/docs/guide/serialization/relnotes.html',0,90).
page(1,'/j2se/1.4.1/docs/api/java/io/package-summary.html',0,91).
page(1,'/products/jdk/serialization/faq',0,92).
page(1,'/j2se/1.4.1/docs/guide/serialization/spec/serialTOC.doc.html',0,93).
page(1,'/j2se/1.4.1/docs/guide/serialization/examples/index.html',0,94).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/serialver.html',0,95).
page(1,'/j2se/1.4.1/docs/guide/serialization/relnotes14.html',0,96).
page(1,'/products/jfc',0,97).
page(1,'/products/jdk/awt',0,98).
page(1,'/docs/books/tutorial/ui',0,99).
page(1,'/j2se/1.4.1/docs/guide/awt/demos/index.html',0,100).
page(1,'/j2se/1.4.1/docs/api/java/awt/package-summary.html',0,101).
page(1,'/j2se/1.4.1/docs/guide/awt/1.3/enhancements.html',0,102).
page(1,'/j2se/1.4.1/docs/guide/util/logging/MAILTO:j2se-comments@java.sun.com',0,103).
page(1,'/j2se/1.4.1/docs/api/java/util/logging/package-summary.html',0,104).
page(1,'/j2se/1.4.1/docs/guide/util/logging/overview.html',0,105).
page(1,'/j2se/1.4.1/docs/api/java/beans/beancontext/package-summary.html',0,106).
page(1,'/beans/index.html',0,107).
page(1,'/j2se/1.4.1/docs/guide/beans/changes14.html',0,108).
page(1,'/docs/books/tutorial/javabeans/index.html',0,109).
page(1,'/products/jfc/tsc',0,110).
page(1,'/beans/spec.html',0,111).
page(1,'/j2se/1.4.1/docs/api/java/beans/package-summary.html',0,112).
page(1,'/j2se/1.4.1/docs/guide/beans/spec/beancontextTOC.fm.html',0,113).
page(118,'/?q=kinit&p=/doc/806-0624/6j9vek58e&a=view',0,114).
page(1,'/j2se/1.4.1/docs/tooldocs/linux/java.html',0,115).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/jarsigner.html',0,116).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/extcheck.html',0,117).
page(1,'/j2se/1.4.1/docs/tooldocs/findingclasses.html',0,118).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/javah.html',0,119).
page(1,'/xml/tutorial_intro.html',0,120).
page(1,'/j2se/1.4.1/docs/guide/xml/jaxp/index.html',0,121).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/rmid.html',0,122).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/jdb.html',0,123).
page(128,'/XML',0,124).
page(1,'/xml',0,125).
page(15,'/xml',0,126).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/jdkfiles.html',0,127).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/jar.html',0,128).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/rmiregistry.html',0,129).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/native2ascii.html',0,130).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/javap.html',0,131).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/jarsigner.html',0,132).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/java.html',0,133).
page(1,'/j2se/1.4.1/docs/tooldocs/tools-changes.html',0,134).
page(1,'/j2se/1.4.1/docs/tooldocs/javadoc/index.html',0,135).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/rmiregistry.html',0,136).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/classpath.html',0,137).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/appletviewer.html',0,138).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/serialver.html',0,139).
page(1,'/j2se/1.4.1/docs/guide/idl/servertool.html',0,140).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/rmid.html',0,141).
page(1,'/j2se/1.4.1/docs/tooldocs',0,142).
page(39,'/developer/onlineTraining/new2java',0,143).
page(1,'/PrintPage.jsp',0,144).
page(149,'/help/legal_and_licensing',0,145).
page(150,'/software/java2',0,146).
page(1,'/j2se/1.4.1/guide/collections/index.html',0,147).
page(15,'/suntrademarks',0,148).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/jar.html',0,149).
page(1,'/j2se/1.4.1/ja',0,150).
page(1,'/products/jdbc/index.html',0,151).
page(1,'/j2se/1.4.1/guide/serialization/index.html',0,152).
page(1,'/j2se/1.4.1/guide/rmi-iiop/index.html',0,153).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/javadoc.html',0,154).
page(1,'/j2se/1.4.1/guide/jni/index.html',0,155).
page(1,'/j2se/1.4.1/guide/misc/index.html',0,156).
page(1,'/javaone',0,157).
page(1,'/tools/index.html',0,158).
page(15,'/corp_emp',0,159).
page(1,'/downloads',0,160).
page(1,'/j2se',0,161).
page(1,'/blueprints',0,162).
page(1,'/j2se/vb.development.center.html',0,163).
page(15,'/smi/Press/sunflash/web_sunflash.html',0,164).
page(1,'/j2se/1.4.1',0,165).
page(170,'/search/java/advanced.jsp',0,166).
page(1,'/industry',0,167).
page(15,'/privacy',0,168).
page(1,'/products',0,169).
page(1,'/products/jdk/1.2/compatibility.html',0,170).
page(1,'/apis.html',0,171).
page(1,'/j2se/1.4/compatibility.html',0,172).
page(15,'/share/text/termsofuse.html',0,173).
page(39,'/berkeley_license.html',0,174).
page(1,'/products/jdk/1.1/compatibility.html',0,175).
page(1,'/j2se/1.3/compatibility.html',0,176).
page(149,'/help',0,177).
page(39,'/',0,178).
page(39,'/developer/infodocs',0,179).
page(15,'/company',0,180).
page(185,'/cgi-shl/index.cgi?shop=33293&date=2002%3A04%3A26%3A10%3A29%3A26',0,181).
page(1,'/j2se/downloads.html',0,182).
page(1,'/docs',0,183).
page(1,'/j2se/1.4.1/docs/api/deprecated-list.html',0,184).
page(1,'/docs/books/jls/clarify.html',0,185).
page(190,'/',0,186).
page(1,'/products/plugin/1.4.1/demos/applets/CardTest/example1.html',0,187).
page(1,'/docs/books/tutorial/together/index.html',0,188).
page(1,'/docs/books/tutorial/getStarted/index.html',0,189).
page(1,'/products/plugin/1.4.1/demos/applets/GraphLayout/example2.html',0,190).
page(1,'/products/plugin/1.4.1/demos/applets/GraphLayout/example4.html',0,191).
page(1,'/products/plugin/1.4.1/demos/applets/WireFrame/example1.html',0,192).
page(1,'/products/plugin/1.4.1/demos/applets/SortDemo/example1.html',0,193).
page(1,'/products/plugin/1.4.1/demos/applets/WireFrame/example3.html',0,194).
page(1,'/docs/books/tutorial/getStarted/cupojava/index.html',0,195).
page(1,'/products/plugin/1.4.1/demos/applets/Animator/example2.html',0,196).
page(1,'/products/plugin/1.4.1/demos/applets/Animator/example4.html',0,197).
page(1,'/products/plugin/1.4.1/demos/applets/MoleculeViewer/example2.html',0,198).
page(1,'/products/plugin/1.4.1/demos/applets/DrawTest/example1.html',0,199).
page(1,'/products/plugin/1.4.1/demos/applets/Fractal/example1.html',0,200).
page(1,'/products/plugin/1.4.1/demos/applets/SimpleGraph/example1.html',0,201).
page(1,'/products/plugin/1.4.1/demos/applets/ImageMap/example1.html',0,202).
page(1,'/products/plugin/1.4.1/demos/applets/Clock/example1.html',0,203).
page(1,'/products/plugin/1.4.1/demos/applets/TicTacToe/example1.html',0,204).
page(1,'/products/plugin/1.4.1/demos/applets/GraphicsTest/example1.html',0,205).
page(1,'/products/plugin/1.4.1/demos/applets/GraphLayout/example1.html',0,206).
page(1,'/products/plugin/1.4.1/demos/applets/ArcTest/example1.html',0,207).
page(1,'/products/plugin/1.4.1/demos/applets/BarChart/example1.html',0,208).
page(1,'/applets/index.html',0,209).
page(1,'/products/plugin/1.4.1/demos/applets/GraphLayout/example3.html',0,210).
page(1,'/products/plugin/1.4.1/demos/applets/WireFrame/example2.html',0,211).
page(1,'/products/plugin/1.4.1/demos/applets/Animator/example1.html',0,212).
page(1,'/products/plugin/1.4.1/demos/applets/WireFrame/example4.html',0,213).
page(1,'/products/plugin/1.4.1/demos/applets/Blink/example1.html',0,214).
page(1,'/products/plugin/1.4.1/demos/applets/Animator/example3.html',0,215).
page(1,'/products/plugin/1.4.1/demos/applets/NervousText/example1.html',0,216).
page(1,'/products/plugin/1.4.1/demos/applets/MoleculeViewer/example1.html',0,217).
page(1,'/products/plugin/1.4.1/demos/applets/DitherTest/example1.html',0,218).
page(1,'/products/plugin/1.4.1/demos/applets/MoleculeViewer/example3.html',0,219).
page(1,'/products/plugin/1.4.1/demos/applets/JumpingBox/example1.html',0,220).
page(1,'/products/plugin/1.4.1/demos/applets/SpreadSheet/example1.html',0,221).
page(15,'/products',0,222).
page(39,'/developer/community/chat',0,223).
page(15,'/developers',0,224).
page(15,'/bigadmin',0,225).
page(15,'/learnabout/n1',0,226).
page(15,'/aboutsun/investor',0,227).
page(15,'/www.sun.com/featurepromo/sunfireblade?AssociateId=bladepromo',0,228).
page(15,'/MySun',0,229).
page(15,'/worldwide',0,230).
page(15,'/howtobuy',0,231).
page(15,'/datacenter/superg/feature/index.html',0,232).
page(15,'/learnabout/java',0,233).
page(15,'/learnabout/webservices',0,234).
page(15,'/company/newsevents',0,235).
page(15,'/bignews/nc03q2',0,236).
page(15,'/supportraining',0,237).
page(1,'/j2se/1.4.1/guide/awt/index.html',0,238).
page(1,'/j2se/1.4.1/guide/util/logging/index.html',0,239).
page(1,'/j2se/1.4.1/guide/jvmpi/index.html',0,240).
page(1,'/j2se/1.4.1/guide/math/index.html',0,241).
page(1,'/j2se/1.4.1/guide/corba/index.html',0,242).
page(1,'/j2se/1.4.1/guide/imf/index.html',0,243).
page(1,'/j2se/1.4.1/guide/access/index.html',0,244).
page(1,'/j2se/1.4.1/guide/extensions/index.html',0,245).
page(1,'/j2se/1.4.1/guide/reflection/index.html',0,246).
page(1,'/j2se/1.4.1/guide/jws/index.html',0,247).
page(1,'/j2se/1.4.1/guide/rmi/index.html',0,248).
page(1,'/j2se/1.4.1/guide/imageio/index.html',0,249).
page(1,'/j2se/1.4.1/guide/versioning/index.html',0,250).
page(1,'/j2se/1.4.1/guide/jdbc/index.html',0,251).
page(1,'/j2se/1.4.1/guide/jps/index.html',0,252).
page(1,'/j2se/1.4.1/guide/net/index.html',0,253).
page(1,'/j2se/1.4.1/guide/beans/index.html',0,254).
page(1,'/j2se/1.4.1/guide/security/index.html',0,255).
page(1,'/j2se/1.4.1/relnotes/features.html',0,256).
page(1,'/j2se/1.4.1/guide/standards/index.html',0,257).
page(1,'/j2se/1.4.1/tooldocs/tools.html',0,258).
page(1,'/j2se/1.4.1/guide/swing/index.html',0,259).
page(1,'/j2se/1.4.1/relnotes/SMICopyright.html',0,260).
page(1,'/j2se/1.4.1/guide/performance/index.html',0,261).
page(1,'/j2se/1.4.1/guide/sound/index.html',0,262).
page(1,'/j2se/1.4.1/guide/2d/index.html',0,263).
page(1,'/j2se/1.4.1/relnotes/demos.html',0,264).
page(1,'/j2se/1.4.1/relnotes/contacts.html',0,265).
page(1,'/j2se/1.4.1/guide/jpda/index.html',0,266).
page(1,'/j2se/1.4.1/guide/intl/index.html',0,267).
page(1,'/j2se/1.4.1/api/overview-summary.html',0,268).
page(1,'/j2se/1.4.1/api/index.html',0,269).
page(1,'/j2se/1.4.1/guide/nio/index.html',0,270).
page(1,'/j2se/1.4.1/guide/plugin/index.html',0,271).
page(1,'/j2se/1.4.1/guide/lang/index.html',0,272).
page(1,'/j2se/1.4.1/guide/jar/index.html',0,273).
page(1,'/j2se/1.4.1/guide/resources/index.html',0,274).
page(1,'/j2se/1.4.1/guide/xml/index.html',0,275).
page(1,'/j2se/1.4.1/guide/jndi/index.html',0,276).
page(1,'/j2se/1.4.1/guide/dragndrop/index.html',0,277).
page(1,'/j2se/1.4.1/guide/vm/index.html',0,278).
page(1,'/j2se/1.4.1/guide/refobs/index.html',0,279).
page(1,'/j2se/1.4.1/guide/idl/index.html',0,280).
page(1,'/products/javawebstart/faq.html',0,281).
page(1,'/products/javawebstart/architecture.html',0,282).
page(1,'/features/2001/09/webstart.html',0,283).
page(1,'/j2se/1.4.1/docs/guide/jws/Readme.html',0,284).
page(1,'/products/javawebstart/developers.html',0,285).
page(1,'/j2se/1.4.1/docs/guide/jws/downloadservletguide.html',0,286).
page(1,'/j2se/1.4.1/docs/guide/jws/relnotes.html',0,287).
page(1,'/products/javawebstart/1.2/docs/l10n/index.html',0,288).
page(1,'/j2se/1.4.1/docs/guide/jws/developersguide.html',0,289).
page(1,'/docs/forma/javaws-sendusmail.html',0,290).
page(1,'/j2se/1.4.1/docs/guide/jws/jnlp/index.html',0,291).
page(1,'/products/java-media/sound/index.html',0,292).
page(1,'/j2se/1.4.1/docs/api/javax/sound/midi/package-summary.html',0,293).
page(1,'/j2se/1.4.1/docs/guide/sound/programmer_guide/contents.html',0,294).
page(1,'/products/java-media/sound',0,295).
page(1,'/j2se/1.4.1/docs/api/javax/sound/midi/spi/package-summary.html',0,296).
page(1,'/j2se/1.4.1/docs/api/javax/sound/sampled/spi/package-summary.html',0,297).
page(1,'/j2se/1.4.1/docs/api/javax/sound/sampled/package-summary.html',0,298).
page(1,'/j2se/1.4.1/docs/guide/jpda/jpda.html',0,299).
page(1,'/j2se/1.4.1/docs/guide/jpda/jvmdi-spec.html',0,300).
page(1,'/j2se/1.4.1/docs/guide/jpda/architecture.html',0,301).
page(1,'/j2se/1.4.1/docs/guide/jpda/jdi/index.html',0,302).
page(1,'/j2se/1.4.1/docs/guide/jpda/enhancements.html',0,303).
page(1,'/j2se/1.4.1/docs/guide/jpda/conninv.html',0,304).
page(1,'/j2se/1.4.1/docs/guide/jpda/jdwp-spec.html',0,305).
page(1,'/products/jpda',0,306).
page(1,'/docs/forms/sendusmail.html',0,307).
page(1,'/mail',0,308).
page(1,'/feedback/faq/assistance.html',0,309).
page(1,'/feedback/sitecomments.html',0,310).
page(1,'/j2se/1.4.1/docs/guide/net/overview/overview.html',0,311).
page(1,'/j2se/1.4.1/docs/guide/net/properties.html',0,312).
page(1,'/j2se/1.4.1/docs/guide/net/socketOpt.html',0,313).
page(1,'/j2se/1.4.1/networking-relnotes.html',0,314).
page(1,'/j2se/1.4.1/docs/api/java/net/package-summary.html',0,315).
page(1,'/j2se/1.4.1/docs/guide/net/ipv6_guide/index.html',0,316).
page(1,'/products/jdbc/download.html',0,317).
page(1,'/j2se/1.4.1/docs/guide/jdbc/bridge.html',0,318).
page(1,'/j2se/1.4.1/docs/api/javax/sql/package-summary.html',0,319).
page(1,'/j2se/1.4.1/docs/guide/jdbc/blob.html',0,320).
page(1,'/products/jdbc/jdbc.drivers.html',0,321).
page(1,'/j2se/1.4.1/docs/api/java/sql/package-summary.html',0,322).
page(1,'/docs/books/tutorial/jdbc/basics/index.html',0,323).
page(1,'/j2ee/docs.html',0,324).
page(39,'/developer/Books/JDBCTutorial/index.html',0,325).
page(1,'/docs/books/jdbc',0,326).
page(39,'/developer/Books/JDBCTutorial/chapter5.html',0,327).
page(1,'/j2se/1.4.1/docs/guide/jdbc/getstart/GettingStartedTOC.fm.html',0,328).
page(1,'/products/jdbc/jdbc-frequent.html',0,329).
page(1,'/j2se/1.4.1/docs/guide/jdbc/clob.html',0,330).
page(1,'/j2se/1.4.1/docs/guide/security/smPortGuide.html',0,331).
page(1,'/docs/books/tutorial/security1.2/index.html',0,332).
page(1,'/j2se/1.4.1/docs/api/org/ietf/jgss/package-summary.html',0,333).
page(1,'/j2se/1.4.1/docs/guide/security/HowToImplAProvider.html',0,334).
page(1,'/j2se/1.4.1/docs/guide/security/PolicyFiles.html',0,335).
page(1,'/j2se/1.4.1/docs/guide/security/permissions.html',0,336).
page(1,'/j2se/1.4.1/docs/api/javax/crypto/package-summary.html',0,337).
page(1,'/j2se/1.4.1/docs/api/javax/crypto/spec/package-summary.html',0,338).
page(1,'/j2se/1.4.1/docs/api/javax/crypto/interfaces/package-summary.html',0,339).
page(1,'/j2se/1.4.1/docs/api/javax/security/auth/kerberos/package-summary.html',0,340).
page(1,'/j2se/1.4.1/docs/guide/security/jaas/spec/com/sun/security/auth/login/package-summary.html',0,341).
page(1,'/j2se/1.4.1/docs/api/javax/security/auth/login/package-summary.html',0,342).
page(1,'/j2se/1.4.1/docs/guide/security/jce/JCERefGuide.html',0,343).
page(1,'/j2se/1.4.1/docs/guide/security/jgss/single-signon.html',0,344).
page(1,'/j2se/1.4.1/docs/api/java/security/package-summary.html',0,345).
page(1,'/j2se/1.4.1/docs/guide/security/spec/security-spec.doc.html',0,346).
page(1,'/j2se/1.4.1/docs/guide/security/jaas/spec/com/sun/security/auth/callback/package-summary.html',0,347).
page(1,'/j2se/1.4.1/docs/guide/security/jaas/tutorials/index.html',0,348).
page(1,'/j2se/1.4.1/docs/api/javax/security/auth/callback/package-summary.html',0,349).
page(1,'/j2se/1.4.1/docs/api/javax/security/cert/package-summary.html',0,350).
page(1,'/j2se/1.4.1/docs/guide/security/jaas/spec/com/sun/security/auth/module/package-summary.html',0,351).
page(1,'/j2se/1.4.1/docs/api/java/security/cert/package-summary.html',0,352).
page(1,'/j2se/1.4.1/docs/api/javax/net/ssl/package-summary.html',0,353).
page(1,'/j2se/1.4.1/docs/api/java/security/spec/package-summary.html',0,354).
page(1,'/security/index.html',0,355).
page(1,'/j2se/1.4.1/docs/api/javax/security/auth/x500/package-summary.html',0,356).
page(1,'/j2se/1.4.1/docs/api/javax/security/auth/package-summary.html',0,357).
page(1,'/docs/books/tutorial/trailmap.html',0,358).
page(1,'/j2se/1.4.1/docs/api/javax/net/package-summary.html',0,359).
page(1,'/j2se/1.4.1/docs/guide/security/doprivileged.html',0,360).
page(1,'/j2se/1.4.1/docs/guide/security/jgss/spec/com/sun/security/jgss/package-summary.html',0,361).
page(1,'/j2se/1.4.1/docs/guide/security/jaas/JAASLMDevGuide.html',0,362).
page(1,'/j2se/1.4.1/docs/guide/security/jce/HowToImplAJCEProvider.html',0,363).
page(1,'/security/jaas/doc/acsac.html',0,364).
page(1,'/j2se/1.4.1/docs/guide/security/SecurityToolsSummary.html',0,365).
page(1,'/j2se/1.4.1/docs/guide/security/cert3.html',0,366).
page(1,'/j2se/1.4.1/docs/api/javax/security/auth/spi/package-summary.html',0,367).
page(1,'/j2se/1.4.1/docs/api/java/security/interfaces/package-summary.html',0,368).
page(1,'/j2se/1.4.1/docs/guide/security/jaas/spec/com/sun/security/auth/package-summary.html',0,369).
page(1,'/j2se/1.4.1/docs/guide/security/CryptoSpec.html',0,370).
page(39,'/developer/bugParade/bugs/4532517.html',0,371).
page(39,'/developer/bugParade/bugs/4639735.html',0,372).
page(39,'/developer/bugParade/bugs/4513715.html',0,373).
page(379,'/aboutJava/communityprocess/maintenance/jsr059/index.html',0,374).
page(39,'/developer/bugParade/bugs/4668963.html',0,375).
page(39,'/developer/bugParade/bugs/4395157.html',0,376).
page(39,'/developer/bugParade/bugs/4659800.html',0,377).
page(39,'/developer/bugParade/bugs/4615396.html',0,378).
page(1,'/j2se/1.4.1/docs/api/java/rmi/server/RMISocketFactory.html',0,379).
page(1,'/j2se/1.4.1/docs/guide/plugin/developer_guide/control_panel.html',0,380).
page(39,'/developer/bugParade/bugs/4632782.html',0,381).
page(1,'/j2se/1.4.1/docs/guide/intl/encoding.doc.html',0,382).
page(39,'/developer/bugParade/bugs/4556882.html',0,383).
page(1,'/j2se/1.4.1/docs/guide/jndi/jndi-ldap.html',0,384).
page(1,'/j2se/1.4.1/docs/tooldocs/linux/ktab.html',0,385).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/policytool.html',0,386).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/javadoc.html',0,387).
page(1,'/j2se/1.4.1/docs/tooldocs/linux/klist.html',0,388).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/classpath.html',0,389).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/javah.html',0,390).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/javac.html',0,391).
page(1,'/j2se/1.4.1/docs/guide/idl/orbd.html',0,392).
page(1,'/j2se/1.4.1/docs/tooldocs/linux/jdkfiles.html',0,393).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/java.html',0,394).
page(1,'/j2se/1.4.1/docs/tooldocs/linux/kinit.html',0,395).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/klist.html',0,396).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/javap.html',0,397).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/jdkfiles.html',0,398).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/policytool.html',0,399).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/rmic.html',0,400).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/extcheck.html',0,401).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/native2ascii.html',0,402).
page(1,'/products/plugin/1.3/docs/script.html',0,403).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/kinit.html',0,404).
page(118,'/?q=kadmin&p=/doc/816-0211/6m6nc66th&a=view',0,405).
page(1,'/j2se/1.4.1/docs/guide/rmi-iiop/toJavaPortableUG.html',0,406).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/appletviewer.html',0,407).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/rmic.html',0,408).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/ktab.html',0,409).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/javac.html',0,410).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/jdb.html',0,411).
page(118,'/?q=kinit&p=/doc/806-0624/6j9vek58d&a=view',0,412).
page(1,'/j2se/1.4.1/docs/tooldocs/solaris/keytool.html',0,413).
page(1,'/j2se/1.4.1/docs/tooldocs/windows/keytool.html',0,414).
page(1,'/j2se/1.4.1/docs/guide/idl/jidlNaming.html',0,415).
page(1,'/docs/books/jls/second_edition/html/jcopyright.doc.html',0,416).
page(1,'/docs/books/jls/second_edition/html/jTOC.doc.html',0,417).
page(1,'/docs/books/jls/second_edition/html/jIX.fm.html',0,418).
page(1,'/docs/books/tutorial/2d/index.html',0,419).
page(1,'/j2se/1.4.1/docs/guide/2d/samples/README.html',0,420).
page(1,'/products/java-media/2D/perf_graphics.html',0,421).
page(1,'/j2se/1.4.1/docs/guide/2d/new_features.html',0,422).
page(1,'/products/java-media/2D/index.html',0,423).
page(1,'/j2se/1.4.1/docs/guide/2d/spec.html',0,424).
page(1,'/j2se/1.4.1/docs/guide/2d/spec/j2d-title.fm.html',0,425).
page(1,'/products/jfc/index.html',0,426).
page(1,'/products/jdk/faq/jnifaq.html',0,427).
page(1,'/j2se/1.4.1/docs/guide/jni/spec/jniTOC.doc.html',0,428).
page(1,'/docs/books/tutorial/native1.1/index.html',0,429).
page(1,'/j2se/1.4.1/docs/guide/awt/1.3/AWT_Native_Interface.html',0,430).
page(1,'/j2se/1.4.1/docs/guide/jni/jni-12.html',0,431).
page(1,'/j2se/1.4.1/docs/api/javax/swing/tree/package-summary.html',0,432).
page(1,'/j2se/1.4.1/docs/api/javax/swing/event/package-summary.html',0,433).
page(1,'/j2se/1.4.1/docs/api/javax/swing/plaf/basic/package-summary.html',0,434).
page(1,'/j2se/1.4.1/docs/api/javax/swing/table/package-summary.html',0,435).
page(1,'/j2se/1.4.1/docs/guide/swing/SwingChanges.html',0,436).
page(1,'/j2se/1.4.1/docs/guide/swing/1.3/SwingChanges.html',0,437).
page(1,'/j2se/1.4.1/docs/api/javax/swing/plaf/package-summary.html',0,438).
page(1,'/j2se/1.4.1/docs/api/javax/swing/undo/package-summary.html',0,439).
page(1,'/products/jfc/tsc/index.html',0,440).
page(1,'/j2se/1.4.1/docs/api/javax/swing/plaf/metal/package-summary.html',0,441).
page(1,'/j2se/1.4.1/docs/api/javax/swing/colorchooser/package-summary.html',0,442).
page(1,'/j2se/1.4.1/docs/api/javax/swing/filechooser/package-summary.html',0,443).
page(1,'/j2se/1.4.1/docs/api/javax/swing/package-summary.html',0,444).
page(1,'/j2se/1.4.1/docs/api/javax/swing/text/rtf/package-summary.html',0,445).
page(1,'/j2se/1.4.1/docs/api/javax/swing/border/package-summary.html',0,446).
page(1,'/docs/books/tutorial/uiswing/index.html',0,447).
page(1,'/j2se/1.4.1/docs/api/javax/swing/text/html/parser/package-summary.html',0,448).
page(1,'/j2se/1.4.1/docs/guide/swing/1.4/Post1.4.html',0,449).
page(1,'/j2se/1.4.1/docs/api/javax/swing/text/package-summary.html',0,450).
page(1,'/j2se/1.4.1/docs/api/javax/swing/text/html/package-summary.html',0,451).
page(1,'/j2se/1.4.1/docs/api/javax/swing/plaf/multi/package-summary.html',0,452).
page(1,'/j2se/1.4.1/docs/guide/collections/overview.html',0,453).
page(1,'/docs/books/tutorial/collections/index.html',0,454).
page(1,'/j2se/1.4.1/docs/guide/collections/designfaq.html',0,455).
page(1,'/j2se/1.4.1/docs/api/java/rmi/server/RMIClassLoader.html',0,456).
page(1,'/j2se/1.4.1/docs/guide/security/jgss/tutorials/index.html',0,457).
page(1,'/j2se/1.4.1/docs/api/java/net/NetworkInterface.html',0,458).
page(1,'/j2se/1.4.1/docs/api/java/lang/Class.html',0,459).
page(1,'/j2se/1.4.1/docs/api/java/lang/reflect/Field.html',0,460).
page(1,'/j2se/1.4.1/docs/api/java/util/regex/package-summary.html',0,461).
page(1,'/j2se/1.4.1/docs/guide/swing/1.4/dnd.html',0,462).
page(1,'/j2se/1.4.1/docs/guide/collections/changes4.html',0,463).
page(1,'/j2se/1.4.1/docs/api/java/rmi/server/RMIClassLoaderSpi.html',0,464).
page(1,'/j2se/1.4.1/docs/guide/performance/speed.html',0,465).
page(1,'/j2se/1.4.1/docs/api/java/lang/reflect/Constructor.html',0,466).
page(1,'/j2se/1.4.1/docs/guide/intl/locale.doc.html',0,467).
page(1,'/j2se/1.4.1/docs/guide/idl/jidlChanges.html',0,468).
page(1,'/j2se/1.4.1/docs/api/java/net/InetAddress.html',0,469).
page(1,'/j2se/1.4.1/docs/guide/security/jsse/JSSERefGuide.html',0,470).
page(1,'/j2se/1.4.1/docs/guide/jni/jni-14.html',0,471).
page(1,'/j2se/1.4.1/docs/api/java/lang/reflect/Method.html',0,472).
page(1,'/j2se/1.4.1/docs/guide/security/certpath/CertPathProgGuide.html',0,473).
page(1,'/j2se/1.4.1/docs/api/java/net/URI.html',0,474).
page(1,'/j2se/1.4.1/docs/guide/security/jaas/JAASRefGuide.html',0,475).
page(1,'/j2se/1.4.1/docs/guide/net/enhancements14.html',0,476).
page(1,'/j2se/1.4.1/docs/api/java/text/NumberFormat.html',0,477).
page(1,'/j2se/1.4.1/docs/guide/lang/preferences.html',0,478).
page(1,'/j2se/1.4.1/docs/guide/idl/POA.html',0,479).
page(1,'/j2se/1.4.1/docs/relnotes',0,480).
page(1,'/products/hotspot',0,481).
page(1,'/j2se/1.4.1/docs/guide/lang/chained-exceptions.html',0,482).
page(1,'/j2se/1.4.1/docs/api/java/util/Currency.html',0,483).
page(1,'/j2se/1.4.1/docs/tooldocs/javadoc/whatsnew-1.4.1.html',0,484).
page(1,'/j2se/1.4.1/docs/guide/lang/assert.html',0,485).
page(1,'/j2se/1.4.1/docs/guide/dragndrop',0,486).
page(1,'/jdc/bugParade/index.html',0,487).
page(1,'/j2se/1.4/docs/relnotes/SMICopyright.html',0,488).
page(1,'/j2se/1.4.1/fixedbugs/fixedbugs.html',0,489).
page(39,'/developer/bugParade/index.html',0,490).
page(1,'/j2se/1.4.1/docs/guide/imf/spi-tutorial.html',0,491).
page(1,'/j2se/1.4.1/docs/guide/imf/overview.html',0,492).
page(1,'/j2se/1.4.1/docs/guide/imf/api-reference.html',0,493).
page(1,'/j2se/1.4.1/docs/guide/imf/api-tutorial.html',0,494).
page(1,'/j2se/1.4.1/docs/guide/imf/spi-reference.html',0,495).
page(1,'/j2se/1.4.1/docs/guide/imf/spec.html',0,496).
page(39,'/developer/bugParade',0,497).
page(39,'/developer/earlyAccess',0,498).
page(1,'/j2me',0,499).
page(1,'/webservices',0,500).
page(39,'/developer/codesamples',0,501).
page(1,'/docs/codeconv/html/CodeConvTOC.doc.html',0,502).
page(513,'/',0,503).
page(514,'/',0,504).
page(1,'/products/OV_otherProduct.html',0,505).
page(39,'/developer/support',0,506).
page(1,'/j2ee',0,507).
page(1,'/j2se/1.4.1/relnotes-cobundle.html',0,508).
page(149,'/help/download',0,509).
page(1,'/j2se/docs/PerformanceFAQ.html',0,510).
page(1,'/j2se/1.4.1/docs/guide/vm/sample-output.html',0,511).
page(1,'/j2se/1.4.1/docs/guide/vm/signal-chaining.html',0,512).
page(1,'/j2se/docs/VMOptions.html',0,513).
page(1,'/products/hotspot/docs/whitepaper/Java_Hotspot_v1.4.1/Java_HSpot_WP_v1.4.1_1002_1.html',0,514).
page(1,'/docs/hotspot/index.html',0,515).
page(1,'/products/hotspot/docs/general/hs2.html',0,516).
page(1,'/j2se/1.4.1/docs/guide/vm/error-handling.html',0,517).
page(1,'/j2se/1.4.1/docs/guide/reflection/proxy.html',0,518).
page(1,'/j2se/1.4.1/docs/guide/reflection/reflection.html',0,519).
page(1,'/j2se/1.4.1/docs/guide/reflection/spec/java-reflectionTOC.doc.html',0,520).
page(1,'/j2se/1.4.1/docs/api/java/lang/ref/Reference.html',0,521).
page(1,'/j2se/1.4.1/docs/api/java/lang/ref/SoftReference.html',0,522).
page(1,'/j2se/1.4.1/docs/api/java/lang/ref/WeakReference.html',0,523).
page(1,'/j2se/1.4.1/docs/api/java/lang/ref/ReferenceQueue.html',0,524).
page(1,'/j2se/1.4.1/docs/api/java/lang/ref/PhantomReference.html',0,525).
page(1,'/j2se/1.4.1/docs/guide/misc/deprecation/index.html',0,526).
page(1,'/j2se/1.4.1/docs/guide/misc/applet.html',0,527).
page(1,'/j2se/1.4.1/docs/guide/imageio/spec/title.fm.html',0,528).
page(15,'/downloads',0,529).
page(1,'/getjava',0,530).
page(15,'/research',0,531).
page(15,'/executives',0,532).
page(15,'/partners',0,533).
page(15,'/emrkt/whynot',0,534).
page(15,'/documentation',0,535).
page(15,'/featured-articles',0,536).
page(15,'/learnabout/solaris',0,537).
page(15,'/contact',0,538).
page(15,'/learnabout/netid',0,539).
page(15,'/aboutsun/media',0,540).
page(553,'/techrep/2003/abstract-122.html',0,541).
page(15,'/solutions',0,542).
page(555,'/javadoc/javadoce.html',0,543).
page(555,'/javadoc/javadoc.html',0,544).
page(1,'/j2se/1.4.1/docs/api/java/io/File.html',0,545).
page(15,'/solaris/java/faq.html',0,546).
page(559,'/',0,547).
page(1,'/nav/new/index.html',0,548).
page(1,'/products/hotjava',0,549).
page(1,'/cgi-bin/feedback.pl',0,550).
page(1,'/cgi-bin/java-ports.cgi',0,551).
page(1,'/products/jdk',0,552).
page(1,'/j2se/1.3/docs/tooldocs/tools.html',0,553).
page(566,'/assist/support/client',0,554).
page(1,'/licensees.html',0,555).
page(1,'/mail/index.html',0,556).
page(1,'/index.html',0,557).
page(1,'/products/jdk/1.2/jre/download-solaris.html',0,558).
page(1,'/products/jdk/1.2/download-solaris.html',0,559).
page(15,'/solaris/java',0,560).
page(1,'/products/jdk/1.1/jre/download-jre-solaris.html',0,561).
page(15,'/sunservice/suned',0,562).
page(15,'/solaris/jre',0,563).
page(1,'/docs/index.html',0,564).
page(1,'/products/jdk/1.1/download-jdk-solaris.html',0,565).
page(578,'/tech/javascript/index.html',0,566).
page(579,'/',0,567).
page(1,'/nav/business/trademark_guidelines.html',0,568).
page(1,'/communitysource/index.html',0,569).
page(1,'/starter.html',0,570).
page(1,'/nav/business/index.html',0,571).
page(1,'/products/jdk/1.1/solaris-product-comparison.html',0,572).
page(39,'/developer/onlineTraining/new2java/overview.html',0,573).
page(39,'/developer/onlineTraining/new2java/supplements/2003/Feb03.html',0,574).
page(39,'/developer/onlineTraining/new2java/javaworkbook',0,575).
page(39,'/developer',0,576).
page(150,'/software/sundev/jde/buy',0,577).
page(39,'/developer/onlineTraining/new2java/reference.html',0,578).
page(150,'/software/products/appsrvr/appsrvr_download.html',0,579).
page(39,'/developer/onlineTraining/new2java/certification.html',0,580).
page(39,'/developer/onlineTraining/new2java/learningpaths.html',0,581).
page(39,'/developer/onlineTraining/new2java/supplements/quizzes/index.html',0,582).
page(39,'/developer/onlineTraining/new2java/gettingstartedjava.html',0,583).
page(39,'/developer/onlineTraining/new2java/programming/learn/unravelingjava.html',0,584).
page(1,'/webservices/downloads/webservicespack.html',0,585).
page(39,'/developer/onlineTraining/new2java/stepbystep.html',0,586).
page(599,'/exec/obidos/ASIN/0201310082/qid=959301911/sr=1-2/103-0036873-9502255/javasoftsunmicroA',0,587).
page(1,'/docs/books/jls/first_edition/html/index.html',0,588).
page(602,'/book/backcover/0,3828,0201634511,00.html',0,589).
page(605,'/cp/javaseries.html',0,590).
page(1,'/docs/forms/jls-sendusmail.html',0,591).
page(1,'/docs/books/jls/first_edition/html/1.1Update.html',0,592).
page(1,'/docs/books/jls/html/jcopyright.doc.html',0,593).
page(1,'/docs/books/jls/jls-maintenance.html',0,594).
page(615,'/DigitalGuru/product_detail.asp?catalog_name=Books&category_name=&product_id=0201310082&partner_id=87',0,595).
page(615,'/DigitalGuru/product_detail.asp?catalog_name=Books&category_name=&product_id=0201634511&partner_id=87',0,596).
page(599,'/exec/obidos/ISBN=0201634511/javasoftsunmicroA',0,597).
page(602,'/book/toc/0,3828,0201634511,00.html',0,598).
page(1,'/j2se/1.4.1/docs/guide/extensions/spec.html',0,599).
page(1,'/j2se/1.4.1/docs/api/java/lang/Thread.html',0,600).
page(1,'/j2se/1.4.1/docs/guide/collections/changes3.html',0,601).
page(1,'/j2se/1.4.1/docs/guide/collections/reference.html',0,602).
page(149,'/java.sun.com/apis.html',0,603).
page(149,'/java.sun.com/docs/books/tutorial',0,604).
page(149,'/java.sun.com/webservices/downloads/webservicespack.html',0,605).
page(149,'/java.sun.com/j2ee',0,606).
page(149,'/developer.java.sun.com/developer/codesamples',0,607).
page(149,'/java.sun.com/j2se/1.4.2/download.html',0,608).
page(149,'/java.sun.com/getjava/download.html',0,609).
page(149,'/developer.java.sun.com/developer/community',0,610).
page(39,'/developer/technicalArticles/Networking/jxta2.0/index.html',0,611).
page(149,'/java.sun.com/tools',0,612).
page(1,'/j2ee/download.html',0,613).
page(149,'/java.sun.com/webservices',0,614).
page(149,'/developer.java.sun.com/developer/support',0,615).
page(1,'/features/index.html',0,616).
page(149,'/wwws.sun.com/software/products/appsrvr/appsrvr_download.html',0,617).
page(149,'/developer.java.sun.com/developer/onlineTraining/new2java',0,618).
page(149,'/developer.java.sun.com/servlet/SessionServlet?url=/developer/',0,619).
page(149,'/javaone/sf2003/conf/sessions/index.en.jsp',0,620).
page(149,'/java.sun.com/whyregister.html',0,621).
page(149,'/forum.java.sun.com',0,622).
page(149,'/developer.java.sun.com/developer/infodocs',0,623).
page(149,'/developer.java.sun.com/developer/onlineTraining',0,624).
page(149,'/java.sun.com/j2me',0,625).
page(149,'/developer.java.sun.com/servlet/RegistrationServlet',0,626).
page(149,'/java.sun.com/products/OV_otherProduct.html',0,627).
page(149,'/java.sun.com/industry',0,628).
page(149,'/java.sun.com/j2se/1.4.1/download.html',0,629).
page(149,'/java.sun.com/j2se',0,630).
page(149,'/java.sun.com/jugs',0,631).
page(149,'/wwws.sun.com/software/sundev/jde/buy/index.html',0,632).
page(149,'/developer.java.sun.com/developer/community/chat',0,633).
page(149,'/java.sun.com/products',0,634).
page(39,'/developer/technicalArticles/Programming/TopCoder',0,635).
page(149,'/developer.java.sun.com/subscription',0,636).
page(149,'/java.sun.com/downloads',0,637).
page(149,'/java.sun.com/xml',0,638).
page(149,'/developer.java.sun.com/developer/earlyAccess',0,639).
page(149,'/developer.java.sun.com/developer/bugParade',0,640).
page(149,'/wireless.java.sun.com',0,641).
page(664,'/wwdc/index_j.html',0,642).
page(149,'/developer.java.sun.com',0,643).
page(149,'/java.sun.com/blueprints',0,644).
page(1,'/aboutJava/communityprocess/review/jsr006/index.html',0,645).
page(1,'/j2se/1.4.1/docs/api/javax/print/attribute/package-summary.html',0,646).
page(1,'/printing',0,647).
page(1,'/j2se/1.4.1/docs/guide/jps/spec/jpsTitle.fm.html',0,648).
page(1,'/docs/books/vmspec/html/VMSpecTOC.doc.html',0,649).
page(599,'/exec/obidos/ISBN=020163452X/javasoftsunmicroA',0,650).
page(599,'/exec/obidos/ISBN=0201432943/javasoftsunmicroA',0,651).
page(1,'/docs/books/vmspec/2nd-edition/html/Copyright.doc.html',0,652).
page(679,'/book/preface/0,3829,0201432943,00.html',0,653).
page(1,'/docs/books/vmspec/2nd-edition/jvms-maintenance.html',0,654).
page(1,'/docs/books/vmspec/2nd-edition/errata.html',0,655).
page(615,'/DigitalGuru/product_detail.asp?catalog_name=Books&category_name=&product_id=0201432943&partner_id=87',0,656).
page(1,'/docs/books/vmspec/jvmcopyr.html',0,657).
page(1,'/docs/books/vmspec/errata.html',0,658).
page(602,'/bookdetail.qry?ISBN=0-201-63452-X&ptype=1160',0,659).
page(615,'/DigitalGuru/product_detail.asp?catalog_name=Books&category_name=&product_id=020163452X&partner_id=87',0,660).
page(602,'/book/backcover/0,3828,0201432943,00.html',0,661).
page(602,'/bookdetail.qry?ISBN=0-201-63452-X&ptype=0',0,662).
page(1,'/j2se/1.4.1/docs/api/java/lang/System.html',0,663).
page(1,'/j2se/1.4.1/docs/api/java/lang/Package.html',0,664).
page(1,'/j2se/1.4.1/docs/guide/versioning/spec/VersioningTOC.html',0,665).
page(1,'/j2se/1.4.1/docs/api/java/lang/ClassLoader.html',0,666).
page(1,'/j2se/1.4.1/docs/guide/net/enhancements.html',0,667).
page(1,'/j2se/1.4.1/docs/guide/net/socketException.html',0,668).
page(1,'/j2se/1.4.1/docs/guide/net/extendingSocks.html',0,669).
page(1,'/j2se/1.4.1/docs/guide/net/miscNet.html',0,670).
page(1,'/j2se/1.4.1/docs/guide/rmi/javarmiproperties.html',0,671).
page(1,'/j2se/1.4.1/docs/guide/rmi/codebase.html',0,672).
page(1,'/j2se/1.4.1/docs/api/java/rmi/registry/package-summary.html',0,673).
page(1,'/j2se/1.4.1/docs/guide/rmi/relnotes.html',0,674).
page(1,'/j2se/1.4.1/docs/guide/rmi/socketfactory/SSLInfo.html',0,675).
page(1,'/docs/books/tutorial/rmi/index.html',0,676).
page(1,'/j2se/1.4.1/docs/guide/rmi/sunrmiproperties.html',0,677).
page(712,'/archives/rmi-users.html',0,678).
page(1,'/j2se/1.4.1/docs/guide/rmi/faq.html',0,679).
page(1,'/j2se/1.4.1/docs/guide/rmi/subscribe.html',0,680).
page(1,'/j2se/1.4.1/docs/guide/rmi/Factory.html',0,681).
page(1,'/j2se/1.4.1/docs/api/java/rmi/activation/package-summary.html',0,682).
page(1,'/j2se/1.4.1/docs/api/java/rmi/package-summary.html',0,683).
page(1,'/products/jdk/rmi',0,684).
page(1,'/j2se/1.4.1/docs/api/java/rmi/server/package-summary.html',0,685).
page(1,'/j2se/1.4.1/docs/guide/rmi/getstart.doc.html',0,686).
page(1,'/j2se/1.4.1/docs/guide/rmi/spec/rmiTOC.html',0,687).
page(1,'/j2se/1.4.1/docs/guide/rmi/activation.html',0,688).
page(1,'/j2se/1.4.1/docs/guide/rmi/socketfactory/index.html',0,689).
page(1,'/j2se/1.4.1/docs/api/java/rmi/dgc/package-summary.html',0,690).
page(1,'/j2se/1.4.1/docs/relnotes/license.html',0,691).
page(1,'/j2se/1.4.1/docs/relnotes/prop_rights.html',0,692).
page(1,'/j2se/1.4.1/docs/api/org/omg/DynamicAny/DynAnyPackage/package-summary.html',0,693).
page(1,'/j2se/1.4.1/docs/api/org/omg/stub/java/rmi/package-summary.html',0,694).
page(1,'/j2se/1.4.1/docs/api/org/omg/CosNaming/package-summary.html',0,695).
page(1,'/j2se/1.4.1/docs/api/java/lang/reflect/package-summary.html',0,696).
page(1,'/j2se/1.4.1/docs/api/java/nio/charset/package-summary.html',0,697).
page(1,'/j2se/1.4.1/docs/api/org/omg/CORBA/ORBPackage/package-summary.html',0,698).
page(1,'/j2se/1.4.1/docs/api/javax/naming/directory/package-summary.html',0,699).
page(1,'/j2se/1.4.1/docs/api/java/awt/im/package-summary.html',0,700).
page(1,'/j2se/1.4.1/docs/api/java/util/jar/package-summary.html',0,701).
page(1,'/j2se/1.4.1/docs/api/javax/print/attribute/standard/package-summary.html',0,702).
page(1,'/j2se/1.4.1/docs/api/java/awt/print/package-summary.html',0,703).
page(1,'/j2se/1.4.1/docs/api/javax/imageio/event/package-summary.html',0,704).
page(1,'/j2se/1.4.1/docs/api/java/awt/dnd/package-summary.html',0,705).
page(1,'/j2se/1.4.1/docs/api/org/omg/DynamicAny/DynAnyFactoryPackage/package-summary.html',0,706).
page(1,'/j2se/1.4.1/docs/api/org/omg/CORBA/portable/package-summary.html',0,707).
page(1,'/j2se/1.4.1/docs/api/org/omg/DynamicAny/package-summary.html',0,708).
page(1,'/j2se/1.4.1/docs/relnotes/devdocs-vs-specs.html',0,709).
page(1,'/j2se/1.4.1/docs/api/org/omg/PortableInterceptor/ORBInitInfoPackage/package-summary.html',0,710).
page(1,'/j2se/1.4.1/docs/api/javax/transaction/xa/package-summary.html',0,711).
page(1,'/j2se/1.4.1/docs/api/java/util/zip/package-summary.html',0,712).
page(1,'/j2se/1.4.1/docs/api/org/xml/sax/helpers/package-summary.html',0,713).
page(1,'/j2se/1.4.1/docs/api/org/omg/PortableServer/POAManagerPackage/package-summary.html',0,714).
page(1,'/j2se/1.4.1/docs/api/javax/print/package-summary.html',0,715).
page(1,'/j2se/1.4.1/docs/api/org/omg/IOP/package-summary.html',0,716).
page(1,'/j2se/1.4.1/docs/api/org/omg/PortableServer/portable/package-summary.html',0,717).
page(1,'/j2se/1.4.1/docs/api/java/nio/package-summary.html',0,718).
page(1,'/j2se/1.4.1/docs/api/javax/accessibility/package-summary.html',0,719).
page(1,'/j2se/1.4.1/docs/api/org/xml/sax/ext/package-summary.html',0,720).
page(1,'/j2se/1.4.1/docs/api/javax/naming/ldap/package-summary.html',0,721).
page(1,'/j2se/1.4.1/docs/api/javax/xml/transform/stream/package-summary.html',0,722).
page(1,'/j2se/1.4.1/docs/api/javax/naming/event/package-summary.html',0,723).
page(1,'/j2se/1.4.1/docs/api/org/omg/PortableServer/ServantLocatorPackage/package-summary.html',0,724).
page(1,'/j2se/1.4.1/docs/api/javax/xml/transform/dom/package-summary.html',0,725).
page(1,'/j2se/1.4.1/docs/api/java/util/prefs/package-summary.html',0,726).
page(1,'/j2se/1.4.1/docs/api/java/text/package-summary.html',0,727).
page(1,'/j2se/1.4.1/docs/api/org/omg/CORBA/package-summary.html',0,728).
page(1,'/j2se/1.4.1/docs/api/javax/print/event/package-summary.html',0,729).
page(1,'/j2se/1.4.1/docs/api/org/omg/CosNaming/NamingContextPackage/package-summary.html',0,730).
page(1,'/j2se/1.4.1/docs/api/javax/imageio/metadata/package-summary.html',0,731).
page(39,'/developer/bugParade/bugs/4627627.html',0,732).
page(39,'/developer/bugParade/bugs/4379138.html',0,733).
page(1,'/j2se/1.4.1/docs/guide/plugin/developer_guide/extensions.html',0,734).
page(39,'/developer/bugParade/bugs/4628933.html',0,735).
page(379,'/aboutJava/communityprocess/maintenance/jsr059/index2.html',0,736).
page(1,'/j2se/1.4.1/docs/guide/jndi',0,737).
page(39,'/developer/bugParade/bugs/4513638.html',0,738).
page(1,'/j2se/1.4.1/docs/guide/plugin/developer_guide/special_tags.html',0,739).
page(39,'/developer/bugParade/bugs/4030374.html',0,740).
page(1,'/j2se/1.4.1/docs/guide/plugin/developer_guide/using_tags.html',0,741).
page(1,'/products/plugin/versions.html',0,742).
page(1,'/j2se/1.4.1/sample-output.html',0,743).
page(39,'/developer/bugParade/bugs/4690831.html',0,744).
page(39,'/developer/bugParade/bugs/4218431.html',0,745).
page(1,'/j2se/1.4.1/docs/guide/plugin/developer_guide/java_js.html',0,746).
page(39,'/developer/bugParade/bugs/4636548.html',0,747).
page(1,'/j2se/1.4.1/docs/api/java/rmi/server/RemoteServer.html',0,748).
page(39,'/developer/bugParade/bugs/4627542.html',0,749).
page(1,'/j2se/1.4.1/docs/api/java/rmi/Naming.html',0,750).
page(39,'/developer/bugParade/bugs/4668511.html',0,751).
page(1,'/j2se/1.4.1/docs/api/java/rmi/server/RMIFailureHandler.html',0,752).
page(39,'/developer/bugParade/bugs/4669873.html',0,753).
page(1,'/docs/books/effective/chapters.html',0,754).
page(1,'/docs/books/effective/awards.html',0,755).
page(791,'/store/home/1,4158,store-5401_id-2,00.html',0,756).
page(599,'/exec/obidos/ASIN/0201310058/ref=ase_javasoftsunmicroA/002-3146347-8288839',0,757).
page(1,'/docs/books/effective/reviews.html',0,758).
page(1,'/docs/books/effective/errata.html',0,759).
page(1,'/docs/books/effective/translations.html',0,760).
page(615,'/DigitalGuru/product_detail.asp?catalog_name=Books&category_name=&product_id=0201310058&partner_id=87',0,761).
page(1,'/docs/books/effective/bio.html',0,762).
page(1,'/docs/books/effective/excursions.html',0,763).
page(1,'/docs/books/effective/praise.html',0,764).
page(1,'/docs/books/effective/toc.html',0,765).
page(1,'/docs/books/effective/interviews.html',0,766).
page(1,'/j2se/1.4.1/docs/guide/io/io-min.html',0,767).
page(1,'/j2se/1.4.1/docs/guide/io/troubleshooting.html',0,768).
page(1,'/j2se/1.4.1/docs/guide/io/io-maj.html',0,769).
page(1,'/j2se/1.4.1/docs/guide/io/io.html',0,770).
page(39,'/developer/onlineTraining/new2java/divelog/part3',0,771).
page(39,'/developer/onlineTraining/Beans/JBShortCourse',0,772).
page(39,'/developer/onlineTraining/Beans/bean01',0,773).
page(39,'/developer/onlineTraining/javaai',0,774).
page(39,'/developer/onlineTraining/J2EE/Intro',0,775).
page(39,'/developer/onlineTraining/J2EE',0,776).
page(39,'/developer/onlineTraining/Beans/beans3',0,777).
page(1,'/j2ee/1.3/docs/tutorial/doc',0,778).
page(39,'/developer/onlineTraining/Beans',0,779).
page(39,'/developer/onlineTraining/GUI/AWTLayoutMgr',0,780).
page(39,'/developer/onlineTraining/new2java/divelog',0,781).
page(1,'/products/jndi/tutorial',0,782).
page(39,'/developer/onlineTraining/Security/Fundamentals/abstract.html',0,783).
page(39,'/developer/onlineTraining/Media/2DText',0,784).
page(1,'/j2ee/tutorial/1_3-fcs',0,785).
page(39,'/developer/onlineTraining/Programming/BasicJava2',0,786).
page(39,'/developer/onlineTraining/Programming/BasicJava1',0,787).
page(39,'/developer/onlineTraining/GUI',0,788).
page(39,'/developer/onlineTraining/Beans/EJBTutorial',0,789).
page(39,'/developer/onlineTraining/distributed',0,790).
page(599,'/exec/obidos/ASIN/0201707209/o/qid%3D957218666/sr%3D8-1/ref%3Daps%5Fsr%5Fb%5F1%5F1/102-8347053-6316853',0,791).
page(39,'/developer/onlineTraining/Beans/beans02',0,792).
page(39,'/developer/onlineTraining/collections',0,793).
page(39,'/developer/onlineTraining/Database/JDBC20Intro',0,794).
page(39,'/developer/onlineTraining/forte4j3',0,795).
page(39,'/developer/onlineTraining/java3d',0,796).
page(39,'/developer/onlineTraining/Security',0,797).
page(39,'/developer/onlineTraining/Database/JDBCShortCourse',0,798).
page(1,'/webservices/tutorial.html',0,799).
page(39,'/developer/onlineTraining/new2java/divelog/part4/index.jsp',0,800).
page(39,'/developer/onlineTraining/Downloads',0,801).
page(39,'/developer/onlineTraining/corba',0,802).
page(599,'/exec/obidos/ASIN/0201715015/o/qid=963590813/sr=8-2/ref=aps_sr_b_1_2/002-0968593-5090437',0,803).
page(1,'/j2se/1.4.1/docs/api/javax/xml/transform/sax/package-summary.html',0,804).
page(1,'/j2se/1.4.1/docs/api/java/lang/ref/package-summary.html',0,805).
page(1,'/j2se/1.4.1/docs/api/org/omg/IOP/CodecFactoryPackage/package-summary.html',0,806).
page(1,'/j2se/1.4.1/docs/api/javax/transaction/package-summary.html',0,807).
page(1,'/j2se/1.4.1/docs/api/javax/rmi/package-summary.html',0,808).
page(1,'/j2se/1.4.1/docs/api/javax/naming/spi/package-summary.html',0,809).
page(1,'/j2se/1.4.1/docs/api/javax/imageio/package-summary.html',0,810).
page(1,'/j2se/1.4.1/docs/api/org/omg/IOP/CodecPackage/package-summary.html',0,811).
page(1,'/j2se/1.4.1/docs/api/javax/imageio/stream/package-summary.html',0,812).
page(1,'/j2se/1.4.1/docs/api/java/awt/geom/package-summary.html',0,813).
page(1,'/j2se/1.4.1/docs/api/org/omg/PortableServer/package-summary.html',0,814).
page(1,'/j2se/1.4.1/docs/api/overview-tree.html',0,815).
page(1,'/j2se/1.4.1/docs/api/java/nio/channels/package-summary.html',0,816).
page(1,'/j2se/1.4.1/docs/api/javax/naming/package-summary.html',0,817).
page(1,'/j2se/1.4.1/docs/api/java/awt/color/package-summary.html',0,818).
page(1,'/j2se/1.4.1/docs/api/org/omg/Dynamic/package-summary.html',0,819).
page(1,'/j2se/1.4.1/docs/api/help-doc.html',0,820).
page(1,'/j2se/1.4.1/docs/api/javax/imageio/spi/package-summary.html',0,821).
page(1,'/j2se/1.4.1/docs/api/java/applet/package-summary.html',0,822).
page(1,'/j2se/1.4.1/docs/api/org/omg/SendingContext/package-summary.html',0,823).
page(1,'/j2se/1.4.1/docs/api/org/omg/PortableInterceptor/package-summary.html',0,824).
page(1,'/j2se/1.4.1/docs/api/org/omg/CORBA_2_3/package-summary.html',0,825).
page(1,'/j2se/1.4.1/docs/api/java/math/package-summary.html',0,826).
page(1,'/j2se/1.4.1/docs/api/javax/rmi/CORBA/package-summary.html',0,827).
page(1,'/j2se/1.4.1/docs/api/org/omg/PortableServer/CurrentPackage/package-summary.html',0,828).
page(1,'/j2se/1.4.1/docs/api',0,829).
page(866,'/cgi-bin/doc?ptc/2000-08-06',0,830).
page(1,'/j2se/1.4.1/docs/api/java/security/acl/package-summary.html',0,831).
page(1,'/j2se/1.4.1/docs/api/java/lang/package-summary.html',0,832).
page(1,'/j2se/1.4.1/docs/api/javax/xml/parsers/package-summary.html',0,833).
page(1,'/j2se/1.4.1/docs/api/javax/xml/transform/package-summary.html',0,834).
page(1,'/j2se/1.4.1/docs/api/java/nio/charset/spi/package-summary.html',0,835).
page(1,'/j2se/1.4.1/docs/api/org/omg/CosNaming/NamingContextExtPackage/package-summary.html',0,836).
page(1,'/j2se/1.4.1/docs/api/java/awt/image/package-summary.html',0,837).
page(1,'/j2se/1.4.1/docs/api/java/util/package-summary.html',0,838).
page(1,'/j2se/1.4.1/docs/api/javax/imageio/plugins/jpeg/package-summary.html',0,839).
page(1,'/j2se/1.4.1/docs/api/org/omg/CORBA/DynAnyPackage/package-summary.html',0,840).
page(1,'/j2se/1.4.1/docs/api/java/awt/image/renderable/package-summary.html',0,841).
page(866,'/cgi-bin/doc?formal/99-10-07',0,842).
page(1,'/j2se/1.4.1/docs/api/index-files/index-1.html',0,843).
page(1,'/j2se/1.4.1/docs/api/org/omg/PortableServer/POAPackage/package-summary.html',0,844).
page(1,'/j2se/1.4.1/docs/api/allclasses-noframe.html',0,845).
page(1,'/j2se/1.4.1/docs/api/org/xml/sax/package-summary.html',0,846).
page(1,'/j2se/1.4.1/docs/api/java/awt/im/spi/package-summary.html',0,847).
page(1,'/j2se/1.4.1/docs/api/org/omg/CORBA_2_3/portable/package-summary.html',0,848).
page(1,'/j2se/1.4.1/docs/api/org/omg/Messaging/package-summary.html',0,849).
page(1,'/j2se/1.4.1/docs/api/java/awt/font/package-summary.html',0,850).
page(1,'/j2se/1.4.1/docs/api/java/nio/channels/spi/package-summary.html',0,851).
page(1,'/j2se/1.4.1/docs/api/org/w3c/dom/package-summary.html',0,852).
page(1,'/j2se/1.4.1/docs/api/org/omg/CORBA/TypeCodePackage/package-summary.html',0,853).
page(1,'/j2se/1.4.1/docs/guide/intl/font.html',0,854).
page(1,'/j2se/1.4.1/docs/guide/intl/demos/MessageFormat/example1.html',0,855).
page(1,'/j2se/1.4.1/docs/guide/intl/intlTOC.doc.html',0,856).
page(1,'/j2se/1.4.1/docs/guide/intl/demos/NumberFormat/example1.html',0,857).
page(1,'/j2se/1.4.1/docs/guide/intl/demos/TextBound/example1.html',0,858).
page(1,'/docs/books/tutorial/i18n/index.html',0,859).
page(1,'/j2se/1.4.1/docs/guide/intl/demos/Collate/example1.html',0,860).
page(39,'/developer/technicalArticles/Intl',0,861).
page(1,'/j2se/1.4.1/docs/guide/2d/spec/j2d-fonts.fm.html',0,862).
page(1,'/j2se/1.4.1/docs/guide/intl/faq.html',0,863).
page(1,'/j2se/1.4.1/docs/guide/intl/demos/DateTimeFormat/example1.html',0,864).
page(1,'/j2se/1.4.1/docs/guide/intl/intl.doc.html',0,865).
page(1,'/j2se/1.4.1/docs/guide/intl/fontprop.html',0,866).
page(1,'/docs/books/chanlee/first_edition/examples.html',0,867).
page(1,'/docs/books/chanlee/first_edition',0,868).
page(1,'/docs/books/chanlee/second_edition',0,869).
page(1,'/docs/books/chanlee/second_edition/vol1/index.html',0,870).
page(1,'/docs/books/vmspec/2nd-edition/html/Threads.doc.html',0,871).
page(1,'/docs/books/vmspec/2nd-edition/html/Instructions.doc.html',0,872).
page(1,'/docs/books/vmspec/2nd-edition/html/ChangesAppendix.doc.html',0,873).
page(1,'/docs/books/vmspec/2nd-edition/html/Concepts.doc.html',0,874).
page(1,'/docs/books/vmspec/2nd-edition/html/VMSpecIX.fm.html',0,875).
page(1,'/docs/books/vmspec/2nd-edition/html/Mnemonics.doc.html',0,876).
page(1,'/docs/books/vmspec/2nd-edition/html/Introduction.doc.html',0,877).
page(1,'/docs/books/vmspec/2nd-edition/html/Overview.doc.html',0,878).
page(149,'/feedback.jsp?qid=258',0,879).
page(1,'/docs/books/vmspec/2nd-edition/html/Compiling.doc.html',0,880).
page(1,'/docs/books/vmspec/2nd-edition/html/ClassFile.doc.html',0,881).
page(1,'/docs/books/vmspec/2nd-edition/html/ConstantPool.doc.html',0,882).
page(1,'/docs/books/vmspec/2nd-edition/html/Preface.doc.html',0,883).
page(1,'/j2se/1.4.1/docs/guide/plugin/developer_guide/contents.html',0,884).
page(1,'/docs/books/chanlee/supplement',0,885).
page(923,'/USA/catalog/options/index.html',0,886).
page(923,'/USA/consulting',0,887).
page(923,'/',0,888).
page(923,'/USA/certification',0,889).
page(923,'/USA/catalog/java.html',0,890).
page(923,'/usa',0,891).
page(15,'/access/index.html',0,892).
page(1,'/j2se/1.4.1/docs/guide/access/properties.html',0,893).
page(1,'/j2se/1.4.1/docs/api/javax/swing/doc-files/Key-Index.html',0,894).
page(1,'/j2se/1.4.1/docs/guide/access/new-features.html',0,895).
page(1,'/products/accessbridge',0,896).
page(1,'/products/jfc/jaccess-1.3/doc/index.html',0,897).
page(1,'/products/jfc/accessibility.html',0,898).
page(1,'/docs/books/tutorial/overview/index.html',0,899).
page(599,'/exec/obidos/ASIN/0201433214/javasoftsunmicroA',0,900).
page(1,'/docs/books/tutorial/java/index.html',0,901).
page(1,'/docs/books/tutorial/post1.0/index.html',0,902).
page(1,'/docs/books/tutorial/networking/index.html',0,903).
page(1,'/docs/books/tutorial/reflect/index.html',0,904).
page(1,'/docs/books/tutorial/extra/regex/index.html',0,905).
page(599,'/exec/obidos/ISBN%3D0201703939/javasoftsunmicroA',0,906).
page(1,'/docs/books/tutorial',0,907).
page(1,'/docs/books/tutorial/book.html',0,908).
page(1,'/docs/books/tutorial/sound/index.html',0,909).
page(1,'/docs/books/tutorial/information/copyright.html',0,910).
page(1,'/docs/books/tutorial/information/download.html',0,911).
page(1,'/docs/books/tutorial/extra/fullscreen/index.html',0,912).
page(951,'/promo/java_sun_books',0,913).
page(1,'/docs/books/tutorial/uiswing/mini/index.html',0,914).
page(1,'/docs/books/tutorial/getStarted/cupojava/unix.html',0,915).
page(1,'/docs/books/tutorial/jdbc/index.html',0,916).
page(1,'/docs/books/tutorial/information/FAQ.html',0,917).
page(1,'/docs/books/tutorial/information/announcelist.html',0,918).
page(1,'/docs/books/tutorial/security1.1/index.html',0,919).
page(1,'/docs/books/tutorial/information/resources.html',0,920).
page(1,'/docs/books/tutorial/ext/index.html',0,921).
page(1,'/docs/books/tutorial/getStarted/cupojava/win32.html',0,922).
page(1,'/docs/books/tutorial/applet/index.html',0,923).
page(599,'/exec/obidos/ISBN%3D0201485583/javasoftsunmicroA',0,924).
page(963,'/DigitalGuru/product_detail.asp?catalog_name=Books&product_id=0201485583&partner_id=87',0,925).
page(1,'/docs/books/tutorial/reallybigindex.html',0,926).
page(963,'/DigitalGuru/product_detail.asp?catalog_name=Books&product_id=0201703939&partner_id=87',0,927).
page(1,'/docs/books/tutorial/jar/index.html',0,928).
page(1,'/docs/books/tutorial/information/history.html',0,929).
page(1,'/docs/books/tutorial/getStarted/cupojava/mac.html',0,930).
page(1,'/docs/books/tutorial/essential/index.html',0,931).
page(1,'/docs/books/tutorial/idl/index.html',0,932).
page(1,'/docs/books/tutorial/dnd/index.html',0,933).
page(963,'/DigitalGuru/product_detail.asp?catalog_name=Books&product_id=0201433214&partner_id=87',0,934).
page(1,'/j2se/1.4.1/docs/guide/jndi/spec/jndi/index.html',0,935).
page(1,'/j2se/1.4.1/docs/guide/jndi/jndi-cos.html',0,936).
page(1,'/j2se/1.4.1/docs/guide/jndi/jndi-dns.html',0,937).
page(1,'/j2se/1.4.1/docs/guide/jndi/reference.html',0,938).
page(1,'/j2se/1.4.1/docs/guide/jndi/spec/spi/spicover.frame.html',0,939).
page(1,'/products/jndi/serviceproviders.html',0,940).
page(1,'/j2se/1.4.1/docs/api/javax/naming/ldap/StartTlsResponse.html',0,941).
page(1,'/products/jndi',0,942).
page(1,'/j2se/1.4.1/docs/guide/jndi/jndi-rmi.html',0,943).
page(1,'/products/jndi/tutorial/index.html',0,944).
page(1,'/j2se/1.4.1/docs/guide/jndi/jndi-ldap-gl.html',0,945).
page(1,'/products/jndi/tutorial/ldap/security',0,946).
page(1,'/j2se/1.4.1/docs/guide/jndi/spec/execsumm/jndiexecsumm.html',0,947).
page(1,'/j2se/1.4.1/docs/api/javax/naming/ldap/StartTlsRequest.html',0,948).
page(1,'/products/jndi/docs.html',0,949).
page(1,'/products/jndi/staytouch.html',0,950).
page(999,'/building/tech_articles/j2ee_soap_part2.html',0,951).
page(39,'/developer/Books/effectivejava',0,952).
page(602,'/book/0,3828,0201310058,00.html',0,953).
page(1,'/features/authors.html',0,954).
page(1,'/docs/books/effective',0,955).
page(1,'/j2se/1.4.1/docs/guide/rmi-iiop',0,956).
page(1,'/j2se/1.4.1/docs/guide/rmi',0,957).
page(866,'/cgi-bin/doc?ptc/00-01-08',0,958).
page(866,'/cgi-bin/doc?ptc/00-01-06',0,959).
page(1008,'/',0,960).
page(1,'/j2se/1.4.1/docs/guide/corba',0,961).
page(1,'/j2se/1.4.1/docs/guide/extensions/extensions.html',0,962).
page(1,'/j2se/1.4.1/docs/guide/extensions/versioning.html',0,963).
page(1,'/j2se/1.4.1/docs/api/java/security/SecureClassLoader.html',0,964).
page(1,'/docs/books/tutorial/ext',0,965).
page(1,'/j2se/1.4.1/docs/api/java/net/URLClassLoader.html',0,966).
page(1,'/j2se/1.4.1/docs/api/java/net/JarURLConnection.html',0,967).
page(1,'/j2se/1.4.1/docs/guide/collections',0,968).
page(1,'/j2se/1.4.1/docs/guide/lang/enhancements-1.3.html',0,969).
page(1,'/j2se/1.4.1/docs/guide/jar',0,970).
page(1,'/j2se/1.4.1/docs/guide/plugin/dom/index.html',0,971).
page(1,'/docs/forms/plugin-sendusmail.html',0,972).
page(1,'/docs/books/chanlee/second_edition/vol1',0,973).
page(1,'/docs/books/chanlee/supplement/index.html',0,974).
page(15,'/share/text/SMICopyright.html',0,975).
page(1,'/products/jdk/1.1/docs/index.html',0,976).
page(39,'/developer/bugParade/bugs/4693219.html',0,977).
page(39,'/developer/bugParade/bugs/4615751.html',0,978).
page(39,'/developer/bugParade/bugs/4697804.html',0,979).
page(1031,'/',0,980).
page(39,'/developer/bugParade/bugs/4648702.html',0,981).
page(39,'/developer/bugParade/bugs/4628281.html',0,982).
page(39,'/developer/bugParade/bugs/4682626.html',0,983).
page(39,'/developer/bugParade/bugs/4505899.html',0,984).
page(1037,'/project/libsafe',0,985).
page(1,'/j2se/1.4.1/docs/api/java/awt/TextArea.html',0,986).
page(1,'/j2se/1.4.1/ReleaseNotes.html',0,987).
page(39,'/developer/bugParade/bugs/4704500.html',0,988).
page(39,'/developer/bugParade/bugs/4634165.html',0,989).
page(39,'/developer/bugParade/bugs/4736093.html',0,990).
page(39,'/developer/bugParade/bugs/4654636.html',0,991).
page(39,'/developer/bugParade/bugs/4644270.html',0,992).
page(39,'/developer/bugParade/bugs/4460583.html',0,993).
page(39,'/developer/bugParade/bugs/4387291.html',0,994).
page(39,'/developer/bugParade/bugs/4503092.html',0,995).
page(39,'/developer/bugParade/bugs/4656697.html',0,996).
page(39,'/developer/bugParade/bugs/4456393.html',0,997).
page(39,'/developer/bugParade/bugs/4740901.html',0,998).
page(39,'/developer/bugParade/bugs/4672707.html',0,999).
page(1,'/j2se/javadoc/faq.html',0,1000).
page(39,'/developer/bugParade/bugs/4521661.html',0,1001).
page(39,'/developer/bugParade/bugs/4460065.html',0,1002).
page(1057,'/',0,1003).
page(1,'/products/java-media/jmf/index.html',0,1004).
page(39,'/developer/bugParade/bugs/4650839.html',0,1005).
page(1,'/j2se/1.4.1/nio/index.html',0,1006).
page(39,'/developer/bugParade/bugs/4682743.html',0,1007).
page(1,'/j2se/1.4.1/docs/guide/idl/limitations.html',0,1008).
page(39,'/developer/bugParade/bugs/4701394.html',0,1009).
page(1,'/j2se/1.4.1/docs/api/java/applet/Applet.html',0,1010).
page(39,'/developer/bugParade/bugs/4652655.html',0,1011).
page(1,'/j2se/1.4.1/docs/api/java/awt/Component.html',0,1012).
page(1,'/j2se/1.4.1/docs/api/java/awt/datatransfer/DataFlavor.html',0,1013).
page(1,'/j2se/1.4.1/docs/guide/dragndrop/spec/dnd1TOC.doc.html',0,1014).
page(1,'/j2se/1.4.1/docs/guide/math/enhancements.html',0,1015).
page(1,'/j2se/1.4.1/docs/guide/math/enhancements14.html',0,1016).
page(1,'/j2se/1.4.1/docs/api/org/omg/PortableServer/Servant.html',0,1017).
page(1,'/j2se/1.4.1/docs/guide/rmi-iiop/rmiiiopexample.html',0,1018).
page(1,'/j2se/1.4.1/docs/guide/idl/PI.html',0,1019).
page(1,'/j2se/1.4.1/docs/guide/rmi-iiop/interop.html',0,1020).
page(1,'/products/ejb/index.html',0,1021).
page(1,'/j2se/1.4.1/docs/api/org/omg/CORBA/doc-files/compliance.html',0,1022).
page(1,'/j2se/1.4.1/docs/guide/rmi-iiop/rmi_iiop_pg.html',0,1023).
page(1,'/products/rmi-iiop',0,1024).
page(1,'/j2se/1.4.1/docs/api/org/omg/CORBA_2_3/portable/ObjectImpl.html',0,1025).
page(1,'/j2se/1.4.1/docs/guide/rmi-iiop/tutorial.html',0,1026).
page(1084,'/',0,1027).
page(1,'/docs/books',0,1028).
page(1086,'/',0,1029).
page(39,'/developer/onlineTraining/awt',0,1030).
page(39,'/developer/onlineTraining/Servlets/Fundamentals',0,1031).
page(39,'/developer/onlineTraining/new2java/divelog/part2/index.jsp',0,1032).
page(1,'/docs/books/tutorial/uiswing',0,1033).
page(39,'/developer/onlineTraining/protocolhandlers',0,1034).
page(39,'/developer/onlineTraining/JavaIntro',0,1035).
page(1,'/j2ee/1.4/docs/tutorial',0,1036).
page(39,'/developer/onlineTraining/JSPIntro',0,1037).
page(39,'/developer/onlineTraining/JavaMail',0,1038).
page(39,'/developer/onlineTraining/Beans/JBeansAPI',0,1039).
page(39,'/developer/onlineTraining/GUI/Swing2',0,1040).
page(39,'/developer/onlineTraining/GUI/Swing1',0,1041).
page(39,'/developer/onlineTraining/webcasts',0,1042).
page(39,'/developer/onlineTraining/rmi',0,1043).
page(39,'/developer/onlineTraining/Media',0,1044).
page(1,'/products/jms/tutorial',0,1045).
page(39,'/developer/onlineTraining/EJBIntro',0,1046).
page(599,'/exec/obidos/ASIN/0201703939/o/qid=963596417/sr=8-3/ref=aps_sr_b_2_2/002-0968593-5090437',0,1047).
page(39,'/developer/onlineTraining/Programming/JDCBook',0,1048).
page(39,'/developer/onlineTraining/collections/index2.html',0,1049).
page(39,'/developer/onlineTraining/J2EE/Intro2/j2ee.html',0,1050).
page(1,'/j2se/1.4.1/docs/api/java/lang/SecurityManager.html',0,1051).
page(1,'/j2se/1.4.1/docs/guide/resources/resources.html',0,1052).
page(1,'/j2se/1.4.1/docs/api/java/util/ResourceBundle.html',0,1053).
page(1,'/j2se/1.4/performance.guide.html',0,1054).
page(1,'/j2se/1.4.1/docs/guide/performance/startup.html',0,1055).
page(1,'/docs/performance/index.html',0,1056).
page(1,'/j2se/1.4.1/install-solaris-64.html',0,1057).
page(1,'/j2se/1.4.1/font-requirements.html',0,1058).
page(1116,'/pub-cgi/show.pl?target=patches/J2SE',0,1059).
page(1,'/j2se/1.4/docs/guide/plugin/index.html',0,1060).
page(1,'/j2se/1.4.1/docs/guide/jvmpi/jvmpi.html',0,1061).
page(1,'/j2se/1.4.1/docs/api/java/net/InetSocketAddress.html',0,1062).
page(1,'/j2se/1.4.1/docs/api/java/net/ServerSocket.html',0,1063).
page(1,'/j2se/1.4.1/docs/guide/nio/example/index.html',0,1064).
page(1,'/j2se/1.4.1/docs/api/java/net/DatagramSocket.html',0,1065).
page(1,'/j2se/1.4.1/docs/api/java/io/FileOutputStream.html',0,1066).
page(1,'/j2se/1.4.1/docs/api/java/lang/StringBuffer.html',0,1067).
page(1,'/j2se/1.4.1/docs/guide/io/index.html',0,1068).
page(1,'/j2se/1.4.1/docs/api/java/io/RandomAccessFile.html',0,1069).
page(1,'/aboutJava/communityprocess/jsr/jsr_051_ioapis.html',0,1070).
page(1,'/j2se/1.4.1/docs/api/java/lang/CharSequence.html',0,1071).
page(1,'/j2se/1.4.1/docs/api/java/io/FileInputStream.html',0,1072).
page(1,'/j2se/1.4.1/docs/api/java/net/SocketAddress.html',0,1073).
page(1,'/j2se/1.4.1/docs/api/java/net/Socket.html',0,1074).
page(1,'/j2se/1.4.1/docs/api/java/io/OutputStreamWriter.html',0,1075).
page(1,'/j2se/1.4/nio/index.html',0,1076).
page(1,'/j2se/1.4.1/docs/api/java/lang/String.html',0,1077).
page(1,'/j2se/1.4.1/docs/api/java/io/InputStreamReader.html',0,1078).
page(1,'/j2se/1.4.1/docs/guide/idl/servantlocators.html',0,1079).
page(1,'/j2se/1.4.1/docs/guide/idl/jidlExample3.html',0,1080).
page(1,'/j2se/1.4.1/docs/guide/idl/jidlDSI.html',0,1081).
page(1,'/j2se/1.4.1/docs/guide/idl/jidlImplBaseServer.html',0,1082).
page(1,'/j2se/1.4.1/docs/guide/idl/servantactivator.html',0,1083).
page(1,'/j2se/1.4.1/docs/guide/idl/INStutorial.html',0,1084).
page(1,'/j2se/1.4.1/docs/api/org/omg/CORBA/ORB.html',0,1085).
page(39,'/developer/technicalArticles/releases/corba',0,1086).
page(1,'/j2se/1.4.1/docs/guide/idl/jidlUsingCORBA.html',0,1087).
page(1,'/j2se/1.4.1/docs/guide/idl/jidlExceptions.html',0,1088).
page(1,'/j2se/1.4.1/docs/guide/idl/jidlFAQ.html',0,1089).
page(1,'/j2se/1.4.1/docs/guide/idl/GShome.html',0,1090).
page(1,'/j2se/1.4.1/docs/guide/idl/mapping/jidlMapping.html',0,1091).
page(1,'/j2se/1.4.1/docs/guide/idl/jidlExample2.html',0,1092).
page(1,'/j2se/1.4.1/docs/guide/idl/jidlTieServer.html',0,1093).
page(1,'/j2se/1.4.1/docs/guide/idl/jidlDistApp.html',0,1094).
page(1,'/j2se/1.4.1/docs/guide/idl/jidlSampleCode.html',0,1095).
page(1,'/j2se/1.4.1/docs/guide/idl/jidlInitialization.html',0,1096).
page(1,'/j2se/1.4.1/docs/guide/idl/tnameserv.html',0,1097).
page(1,'/j2se/1.4.1/docs/guide/idl/jidlGlossary.html',0,1098).
page(1,'/j2se/1.4.1/docs/guide/idl/jidlExample.html',0,1099).
page(1,'/j2se/1.4.1/docs/guide/idl/adapteractivator.html',0,1100).
page(1,'/aboutJava/communityprocess/jsr/jsr_015_iio.html',0,1101).
page(1,'/products/java-media/jai/index.html',0,1102).
page(1,'/feedback/index.html',0,1103).
page(1166,'/acrobat/readstep.html',0,1104).
page(1,'/j2se/1.4/docs/relnotes/demos.html',0,1105).
page(1,'/download',0,1106).
page(1,'/j2se/1.4/ja/docs/ja/index.html',0,1107).
page(15,'/policies/trademarks',0,1108).
page(605,'/cseng/javaseries',0,1109).
page(1,'/j2se/1.4/docs/index.html',0,1110).
page(1,'/nav/whatis/index.html',0,1111).
page(1,'/j2se/tools.html',0,1112).
page(1,'/jdc',0,1113).
page(1,'/infodocs',0,1114).
page(1,'/support',0,1115).
page(1,'/community',0,1116).
page(1,'/products/jdk/1.2',0,1117).
page(15,'/software/communitysource/java2/index.html',0,1118).
page(1,'/products/jdk/1.1',0,1119).
page(1,'/products/archive/index.html',0,1120).
page(1,'/docs/glossary.html',0,1121).
page(1,'/j2se/1.4',0,1122).
page(1,'/casestudies',0,1123).
page(1,'/j2se/1.3',0,1124).
page(1,'/j2se/technologies.html',0,1125).
page(1,'/j2se/index.html',0,1126).
page(170,'/Search/java?help=tips',0,1127).
page(1,'/j2se/faq.html',0,1128).
page(1,'/docs/books/chanlee/second_edition/index.html',0,1129).
page(1199,'/kb/view.asp?pcode=ALL&articleid=Q104985',0,1130).
page(1200,'/Com/DCOM/Dcom98/relnotes.asp',0,1131).
page(1201,'/support/kb/articles/Q142/2/71.asp',0,1132).
page(1,'/j2se/1.4.1/docs/guide/plugin/developer_guide/silent.html',0,1133).
page(1,'/j2se/1.4.1/index.html',0,1134).
page(1,'/j2se/1.4.1/docs/api/java/util/jar/JarFile.html',0,1135).
page(1,'/j2se/1.4.1/docs/guide/jar/jarGuide.html',0,1136).
page(1,'/j2se/1.4.1/docs/api/java/util/zip/ZipFile.html',0,1137).
page(1,'/j2se/1.4.1/docs/guide/jar/jar.html',0,1138).
page(1,'/docs/books/tutorial/jar',0,1139).
/* (1139 entrees) */


/**
 * Table fichier
 */

fichier('/docs/j2se1.4/','serial-spec.pdf',1).
fichier('/docs/j2se1.4/','serial-spec.ps',2).
fichier('/products/javabeans/glasgow/','beancontext.ps',3).
fichier('/products/javabeans/glasgow/','beancontext.pdf',4).
fichier('/docs/j2se1.4/','programmer_guide.pdf',5).
fichier('/internet-drafts/','draft-ietf-ldapext-locate-08.txt',6).
fichier('/docs/j2se/1.4/','j2d-book.pdf',7).
fichier('/docs/j2se/1.4/','j2d-book.ps',8).
fichier('/pub/docs/ptc/','99-12-03.pdf',9).
fichier('/docs/codeconv/','codeconv.zip',10).
fichier('/docs/codeconv/','CodeConventions.pdf',11).
fichier('/docs/codeconv/','CodeConventions.ps',12).
fichier('/docs/specs/','langspec-2.0.html.tar.Z',13).
fichier('/docs/specs/','langspec-1.0.html.zip',14).
fichier('/docs/specs/','langspec-1.0.html.tar.Z',15).
fichier('/docs/specs/','langspec-1.0.ps.zip',16).
fichier('/docs/specs/','langspec-2.0.pdf',17).
fichier('/docs/specs/','langspec-1.0.pdf',18).
fichier('/docs/books/jls/second_edition/download/','langspec-2.0.html.tar.Z',19).
fichier('/docs/books/jls/second_edition/download/','langspec-2.0.html.zip',20).
fichier('/docs/specs/','langspec-2.0.html.zip',21).
fichier('/docs/specs/','langspec-1.0.ps.Z',22).
fichier('/docs/j2se1.4/','JPS_PDF.pdf',23).
fichier('/docs/specs/','vmspec.html.tar.Z',24).
fichier('/docs/specs/','vmspec.2nded.html.zip',25).
fichier('/docs/specs/','vmspec.2nded.html.tar.Z',26).
fichier('/docs/specs/','vmspec.html.zip',27).
fichier('/docs/specs/','vmspec.2nded.html.tar.gz',28).
fichier('/docs/specs/','vmspec.html.tar.gz',29).
fichier('/docs/j2se1.4/','versioning.pdf',30).
fichier('/docs/j2se1.4/','versioning.ps',31).
fichier('/docs/j2se1.4/','networking.pdf',32).
fichier('/docs/j2se1.4/','ipv6_guide.pdf',33).
fichier('/docs/j2se1.4/','rmi-spec-1.4.pdf',34).
fichier('/docs/j2se1.4/','rmi-spec-1.4.ps',35).
fichier('/docs/books/effective/','effective.zip',36).
fichier('/docs/books/chanlee/second_edition/vol1/','readme.txt',37).
fichier('/docs/books/chanlee/','JavaClassLibVol1Examples.zip',38).
fichier('/docs/j2se1.4/','jndi.ps',39).
fichier('/rfc/','rfc2829.txt',40).
fichier('/rfc/','rfc2222.txt',41).
fichier('/docs/j2se1.4/','jndi.pdf',42).
fichier('/docs/j2se1.4/','jndispi.pdf',43).
fichier('/rfc/','rfc2831.txt',44).
fichier('/docs/j2se1.4/','jndiexecsumm.pdf',45).
fichier('/docs/j2se1.4/','jndiexecsumm.ps',46).
fichier('/rfc/','rfc2830.txt',47).
fichier('/docs/j2se1.4/','jndispi.ps',48).
fichier('/docs/j2se1.4.1/','developer_guide.pdf',49).
fichier('/docs/books/chanlee/','sup1-egs.zip',50).
fichier('/docs/books/chanlee/supplement/','readme.txt',51).
fichier('/rfc/','rfc2246.txt',52).
fichier('/products/java-media/2D/','perf_graphics.pdf',53).
fichier('/internet-drafts/','draft-ietf-ldapext-locate-07.txt',54).
fichier('/docs/j2se1.4/','dnd1.ps',55).
fichier('/docs/j2se1.4/','dnd1.pdf',56).
fichier('/j2se/1.4.1/docs/guide/rmi-iiop/','interop.pdf',57).
fichier('/docs/j2se1.4/','imageio_guide.pdf',58).
fichier('/docs/j2se1.4/','imageio_guide.ps',59).
fichier('/docs/j2se1.4/','j2d-book.ps',60).
fichier('/docs/j2se1.4/','j2d-book.pdf',61).
fichier('/docs/j2se1.4/','rmi-spec-1.3.pdf',62).
fichier('/docs/j2se/1.4/','JPS_PDF.pdf',63).
fichier('/docs/j2se1.4/','VolatileImage.pdf',64).
fichier('/docs/j2se1.4/','rmi-spec-1.3.ps',65).
fichier('/docs/j2se1.4/','beancontext.pdf',66).
fichier('/docs/j2se1.4/','beancontext.ps',67).
fichier('/docs/books/chanlee/second_edition/','readme.txt',68).
fichier('/docs/books/chanlee/','JavaClassLibVol2Examples.zip',69).
/* (69 entrees) */


/**
 * Table protocole
 */

protocole('ftp://',1).
protocole('http://',2).
/* (2 entrees) */


/**
 * Table heberge
 */

heberge(1,1,97).
heberge(2,1,97).
heberge(3,2,1).
heberge(4,2,1).
heberge(5,1,97).
heberge(6,2,385).
heberge(7,1,97).
heberge(8,1,97).
heberge(9,1,475).
heberge(10,1,510).
heberge(11,1,510).
heberge(12,1,510).
heberge(13,1,510).
heberge(14,1,510).
heberge(15,1,510).
heberge(16,1,510).
heberge(17,1,510).
heberge(18,1,510).
heberge(19,2,1).
heberge(20,2,1).
heberge(21,1,510).
heberge(22,1,510).
heberge(23,1,97).
heberge(24,1,510).
heberge(25,1,510).
heberge(26,1,510).
heberge(27,1,510).
heberge(28,1,510).
heberge(29,1,510).
heberge(30,1,97).
heberge(31,1,97).
heberge(32,1,97).
heberge(33,1,97).
heberge(34,1,97).
heberge(35,1,97).
heberge(36,2,1).
heberge(37,2,1).
heberge(38,1,510).
heberge(39,1,97).
heberge(40,2,385).
heberge(41,2,385).
heberge(42,1,97).
heberge(43,1,97).
heberge(44,2,385).
heberge(45,1,97).
heberge(46,1,97).
heberge(47,2,385).
heberge(48,1,97).
heberge(49,1,97).
heberge(50,1,510).
heberge(51,2,1).
heberge(52,2,385).
heberge(53,2,1).
heberge(54,2,385).
heberge(55,1,97).
heberge(56,1,97).
heberge(57,2,1).
heberge(58,1,97).
heberge(59,1,97).
heberge(60,1,97).
heberge(61,1,97).
heberge(62,1,97).
heberge(63,1,97).
heberge(64,1,97).
heberge(65,1,97).
heberge(66,1,97).
heberge(67,1,97).
heberge(68,2,1).
heberge(69,1,510).
/* (69 entrees) */


/**
 * Table mot_desc
 */

mot_desc('>',1).
mot_desc('-',2).
mot_desc('--',3).
mot_desc('?',4).
mot_desc('',5).
mot_desc('',6).
mot_desc('',7).
mot_desc('0',8).
mot_desc('1',9).
mot_desc('10',10).
mot_desc('101',11).
mot_desc('1_02',12).
mot_desc('11',13).
mot_desc('1165634',14).
mot_desc('12',15).
mot_desc('13',16).
mot_desc('1321577',17).
mot_desc('14',18).
mot_desc('15',19).
mot_desc('16',20).
mot_desc('16-bit',21).
mot_desc('17',22).
mot_desc('18',23).
mot_desc('19',24).
mot_desc('1999',25).
mot_desc('2',26).
mot_desc('2000-08-06',27).
mot_desc('2002',28).
mot_desc('2003',29).
mot_desc('2-11',30).
mot_desc('2222',31).
mot_desc('2246-tlsv1',32).
mot_desc('2829',33).
mot_desc('2830',34).
mot_desc('2831',35).
mot_desc('2d',36).
mot_desc('3',37).
mot_desc('30',38).
mot_desc('32',39).
mot_desc('3d',40).
mot_desc('4',41).
mot_desc('4030374',42).
mot_desc('4218431',43).
mot_desc('4379138',44).
mot_desc('4387291',45).
mot_desc('4395157',46).
mot_desc('4456393',47).
mot_desc('4460065',48).
mot_desc('4460583',49).
mot_desc('4503092',50).
mot_desc('4505899',51).
mot_desc('4513638',52).
mot_desc('4513715',53).
mot_desc('4521661',54).
mot_desc('4532517',55).
mot_desc('4556882',56).
mot_desc('4615396',57).
mot_desc('4615751',58).
mot_desc('4627542',59).
mot_desc('4627627',60).
mot_desc('4628281',61).
mot_desc('4628933',62).
mot_desc('4632782',63).
mot_desc('4634165',64).
mot_desc('4636548',65).
mot_desc('4639735',66).
mot_desc('4644270',67).
mot_desc('4648702',68).
mot_desc('4650839',69).
mot_desc('4652655',70).
mot_desc('4654636',71).
mot_desc('4656697',72).
mot_desc('4659800',73).
mot_desc('4668511',74).
mot_desc('4668963',75).
mot_desc('4669873',76).
mot_desc('4672707',77).
mot_desc('4682626',78).
mot_desc('4682743',79).
mot_desc('4690831',80).
mot_desc('4693219',81).
mot_desc('4697804',82).
mot_desc('4701394',83).
mot_desc('4704500',84).
mot_desc('4736093',85).
mot_desc('4740901',86).
mot_desc('48',87).
mot_desc('5',88).
mot_desc('509',89).
mot_desc('6',90).
mot_desc('64-bit',91).
mot_desc('7',92).
mot_desc('71',93).
mot_desc('745698',94).
mot_desc('754',95).
mot_desc('8',96).
mot_desc('9',97).
mot_desc('95',98).
mot_desc('99-10-07',99).
mot_desc('99-12-03',100).
mot_desc('a1',101).
mot_desc('a10',102).
mot_desc('a11',103).
mot_desc('a12',104).
mot_desc('a13',105).
mot_desc('a14',106).
mot_desc('a2',107).
mot_desc('a3',108).
mot_desc('a4',109).
mot_desc('a5',110).
mot_desc('a6',111).
mot_desc('a7',112).
mot_desc('a8',113).
mot_desc('a9',114).
mot_desc('about',115).
mot_desc('abrupt',116).
mot_desc('abstract',117).
mot_desc('access',118).
mot_desc('accessibility',119).
mot_desc('accessing',120).
mot_desc('acl',121).
mot_desc('acrobat',122).
mot_desc('activation',123).
mot_desc('activators',124).
mot_desc('adapter',125).
mot_desc('addendum',126).
mot_desc('addison-wesley',127).
mot_desc('additional',128).
mot_desc('address',129).
mot_desc('addresses',130).
mot_desc('admins',131).
mot_desc('advanced',132).
mot_desc('affecting',133).
mot_desc('all',134).
mot_desc('allimant',135).
mot_desc('alpha',136).
mot_desc('amazon',137).
mot_desc('amendments',138).
mot_desc('an',139).
mot_desc('and',140).
mot_desc('annotated',141).
mot_desc('announcements',142).
mot_desc('answers',143).
mot_desc('any',144).
mot_desc('api',145).
mot_desc('apis',146).
mot_desc('app?',147).
mot_desc('appearing',148).
mot_desc('appendix',149).
mot_desc('apple',150).
mot_desc('applet',151).
mot_desc('applets',152).
mot_desc('applets?',153).
mot_desc('appletviewer?',154).
mot_desc('application',155).
mot_desc('applications',156).
mot_desc('applying',157).
mot_desc('april',158).
mot_desc('arbitrary-precision',159).
mot_desc('architecture',160).
mot_desc('archive',161).
mot_desc('archives',162).
mot_desc('arctest',163).
mot_desc('are',164).
mot_desc('area',165).
mot_desc('areas',166).
mot_desc('arguments',167).
mot_desc('arithmetic',168).
mot_desc('array',169).
mot_desc('arrays',170).
mot_desc('articleid=q104985',171).
mot_desc('articles',172).
mot_desc('ary',173).
mot_desc('asked',174).
mot_desc('asp',175).
mot_desc('asp?pcode=all',176).
mot_desc('assertion',177).
mot_desc('assignment',178).
mot_desc('assistive',179).
mot_desc('assumptions',180).
mot_desc('at',181).
mot_desc('atitech',182).
mot_desc('attribute',183).
mot_desc('attributes',184).
mot_desc('audiocasts',185).
mot_desc('auth',186).
mot_desc('authentication',187).
mot_desc('author',188).
mot_desc('authorization',189).
mot_desc('automatic',190).
mot_desc('available',191).
mot_desc('available?',192).
mot_desc('availableflavors',193).
mot_desc('avaya',194).
mot_desc('avayalabs',195).
mot_desc('awards',196).
mot_desc('awt',197).
mot_desc('a-z',198).
mot_desc('aquick',199).
mot_desc('b1',200).
mot_desc('b2',201).
mot_desc('b3',202).
mot_desc('b4',203).
mot_desc('b5',204).
mot_desc('b6',205).
mot_desc('b7',206).
mot_desc('b8',207).
mot_desc('b9',208).
mot_desc('barchart',209).
mot_desc('basic',210).
mot_desc('basics',211).
mot_desc('be',212).
mot_desc('beancontext',213).
mot_desc('beans',214).
mot_desc('before',215).
mot_desc('behavior',216).
mot_desc('below',217).
mot_desc('beta',218).
mot_desc('between',219).
mot_desc('beyond',220).
mot_desc('big',221).
mot_desc('bin',222).
mot_desc('binary',223).
mot_desc('binding',224).
mot_desc('bindings',225).
mot_desc('bingo!',226).
mot_desc('blade',227).
mot_desc('blink',228).
mot_desc('blob',229).
mot_desc('blocks',230).
mot_desc('blueprints',231).
mot_desc('book',232).
mot_desc('books',233).
mot_desc('boolean',234).
mot_desc('bootstrap',235).
mot_desc('border',236).
mot_desc('boundaries',237).
mot_desc('bridge',238).
mot_desc('broker',239).
mot_desc('browser',240).
mot_desc('browsers',241).
mot_desc('bug',242).
mot_desc('bugreport',243).
mot_desc('bugs',244).
mot_desc('build',245).
mot_desc('building',246).
mot_desc('bullet',247).
mot_desc('business',248).
mot_desc('button',249).
mot_desc('buy',250).
mot_desc('by',251).
mot_desc('bytecode',252).
mot_desc('bytes',253).
mot_desc('c1',254).
mot_desc('c2',255).
mot_desc('cacerts',256).
mot_desc('call',257).
mot_desc('callback',258).
mot_desc('can',259).
mot_desc('candidate',260).
mot_desc('cardtest',261).
mot_desc('carry',262).
mot_desc('case',263).
mot_desc('casting',264).
mot_desc('causes',265).
mot_desc('center',266).
mot_desc('cert',267).
mot_desc('certificate',268).
mot_desc('certificates',269).
mot_desc('certification',270).
mot_desc('cgi',271).
mot_desc('cgi-bin',272).
mot_desc('chained',273).
mot_desc('chaining',274).
mot_desc('change',275).
mot_desc('changed',276).
mot_desc('changes',277).
mot_desc('channels',278).
mot_desc('chapter',279).
mot_desc('chapters',280).
mot_desc('character',281).
mot_desc('charsequence',282).
mot_desc('charset',283).
mot_desc('chat',284).
mot_desc('claims',285).
mot_desc('clarification',286).
mot_desc('clarifications',287).
mot_desc('class',288).
mot_desc('classes',289).
mot_desc('classfile',290).
mot_desc('classloader',291).
mot_desc('client',292).
mot_desc('clients',293).
mot_desc('clob',294).
mot_desc('clock',295).
mot_desc('code',296).
mot_desc('code?',297).
mot_desc('codebase',298).
mot_desc('codecfactorypackage',299).
mot_desc('codecpackage',300).
mot_desc('coders',301).
mot_desc('collation',302).
mot_desc('collections',303).
mot_desc('collections-comments',304).
mot_desc('college',305).
mot_desc('college-age',306).
mot_desc('color',307).
mot_desc('colorchooser',308).
mot_desc('com',309).
mot_desc('come',310).
mot_desc('comformance',311).
mot_desc('comment',312).
mot_desc('comments',313).
mot_desc('common',314).
mot_desc('communication',315).
mot_desc('community',316).
mot_desc('company',317).
mot_desc('compatibility',318).
mot_desc('competition',319).
mot_desc('compiler',320).
mot_desc('compiling',321).
mot_desc('completion',322).
mot_desc('compliance',323).
mot_desc('component',324).
mot_desc('components',325).
mot_desc('computing',326).
mot_desc('concepts',327).
mot_desc('confluent',328).
mot_desc('conformance',329).
mot_desc('conjunction',330).
mot_desc('connection',331).
mot_desc('connectivity',332).
mot_desc('connector',333).
mot_desc('consistency',334).
mot_desc('constant',335).
mot_desc('constant_class_info',336).
mot_desc('constant_double_info',337).
mot_desc('constant_fieldref_info',338).
mot_desc('constant_float_info',339).
mot_desc('constant_integer_info',340).
mot_desc('constant_interfacemethodref_info',341).
mot_desc('constant_long_info',342).
mot_desc('constant_methodref_info',343).
mot_desc('constant_nameandtype_info',344).
mot_desc('constants',345).
mot_desc('constant_string_info',346).
mot_desc('constant_utf8_info',347).
mot_desc('constantvalue',348).
mot_desc('constraints',349).
mot_desc('constructor',350).
mot_desc('constructors',351).
mot_desc('constructs',352).
mot_desc('consulting',353).
mot_desc('contact',354).
mot_desc('contacting',355).
mot_desc('contain?',356).
mot_desc('content',357).
mot_desc('contents',358).
mot_desc('continued',359).
mot_desc('control',360).
mot_desc('controlling',361).
mot_desc('conventions',362).
mot_desc('conversion',363).
mot_desc('conversions',364).
mot_desc('copyright',365).
mot_desc('copyrights',366).
mot_desc('corba',367).
mot_desc('corba_2_3',368).
mot_desc('cos',369).
mot_desc('cosnaming',370).
mot_desc('course',371).
mot_desc('courses',372).
mot_desc('covered',373).
mot_desc('created',374).
mot_desc('creating',375).
mot_desc('creation',376).
mot_desc('crown',377).
mot_desc('crypto',378).
mot_desc('cryptography',379).
mot_desc('cup',380).
mot_desc('currency',381).
mot_desc('currently',382).
mot_desc('currentpackage',383).
mot_desc('curriculum',384).
mot_desc('custom',385).
mot_desc('cutting',386).
mot_desc('d1',387).
mot_desc('d2',388).
mot_desc('d3',389).
mot_desc('d4',390).
mot_desc('d5',391).
mot_desc('data',392).
mot_desc('database',393).
mot_desc('dataflavor[]',394).
mot_desc('datagramsocket',395).
mot_desc('datatransfer',396).
mot_desc('date',397).
mot_desc('dcom98',398).
mot_desc('debug',399).
mot_desc('debugger',400).
mot_desc('debugging',401).
mot_desc('decorative',402).
mot_desc('default',403).
mot_desc('defined',404).
mot_desc('defining',405).
mot_desc('demo',406).
mot_desc('demonstration',407).
mot_desc('demos',408).
mot_desc('deploying',409).
mot_desc('deprecated',410).
mot_desc('deriving',411).
mot_desc('description',412).
mot_desc('descriptions',413).
mot_desc('descriptors',414).
mot_desc('design',415).
mot_desc('desktop',416).
mot_desc('detailed',417).
mot_desc('details',418).
mot_desc('detection',419).
mot_desc('developer',420).
mot_desc('developers',421).
mot_desc('developing',422).
mot_desc('development',423).
mot_desc('developments?',424).
mot_desc('dgc',425).
mot_desc('did',426).
mot_desc('different',427).
mot_desc('digitalguru',428).
mot_desc('directory',429).
mot_desc('discount',430).
mot_desc('discovery',431).
mot_desc('discussion',432).
mot_desc('discussions',433).
mot_desc('displays',434).
mot_desc('distributed',435).
mot_desc('dithertest',436).
mot_desc('dnd',437).
mot_desc('dns',438).
mot_desc('do',439).
mot_desc('doc?formal',440).
mot_desc('doc?ptc',441).
mot_desc('docs',442).
mot_desc('document',443).
mot_desc('documentation',444).
mot_desc('documents',445).
mot_desc('does',446).
mot_desc('dom',447).
mot_desc('domain',448).
mot_desc('done',449).
mot_desc('dos',450).
mot_desc('double',451).
mot_desc('download',452).
mot_desc('downloading',453).
mot_desc('downloads',454).
mot_desc('draft-ietf-ldapext-locate-07',455).
mot_desc('draft-ietf-ldapext-locate-08',456).
mot_desc('drag',457).
mot_desc('drag-and-drop',458).
mot_desc('drawtest',459).
mot_desc('driver',460).
mot_desc('drivers',461).
mot_desc('drives',462).
mot_desc('drop',463).
mot_desc('duke',464).
mot_desc('[duke]',465).
mot_desc('dyanmic',466).
mot_desc('dynamic',467).
mot_desc('dynamicany',468).
mot_desc('dynanyfactorypackage',469).
mot_desc('dynanypackage',470).
mot_desc('e1',471).
mot_desc('e2',472).
mot_desc('e3',473).
mot_desc('e4',474).
mot_desc('e5',475).
mot_desc('eagerhttpfallback',476).
mot_desc('early',477).
mot_desc('ed',478).
mot_desc('edge',479).
mot_desc('edition',480).
mot_desc('education',481).
mot_desc('educational',482).
mot_desc('effective',483).
mot_desc('ejb',484).
mot_desc('email',485).
mot_desc('e-mail',486).
mot_desc('email?',487).
mot_desc('embed',488).
mot_desc('employment',489).
mot_desc('empower',490).
mot_desc('encoding',491).
mot_desc('encodings',492).
mot_desc('encountering',493).
mot_desc('end',494).
mot_desc('endorsed',495).
mot_desc('eng',496).
mot_desc('engine',497).
mot_desc('engineering',498).
mot_desc('engineers',499).
mot_desc('engines!',500).
mot_desc('english',501).
mot_desc('enhanced',502).
mot_desc('enhancements',503).
mot_desc('enterprise',504).
mot_desc('environment',505).
mot_desc('era',506).
mot_desc('errata',507).
mot_desc('error-reporting',508).
mot_desc('errors',509).
mot_desc('essential',510).
mot_desc('essentials',511).
mot_desc('event',512).
mot_desc('events',513).
mot_desc('every',514).
mot_desc('everywhere',515).
mot_desc('example',516).
mot_desc('examples',517).
mot_desc('exception',518).
mot_desc('exceptions',519).
mot_desc('exclusive',520).
mot_desc('excursions',521).
mot_desc('execution',522).
mot_desc('executive',523).
mot_desc('executives',524).
mot_desc('exit',525).
mot_desc('export',526).
mot_desc('expressions',527).
mot_desc('ext',528).
mot_desc('extendable',529).
mot_desc('extension',530).
mot_desc('extensions',531).
mot_desc('facility',532).
mot_desc('factility',533).
mot_desc('factory',534).
mot_desc('fans',535).
mot_desc('faq',536).
mot_desc('favorite',537).
mot_desc('fcs',538).
mot_desc('feature',539).
mot_desc('features',540).
mot_desc('february3',541).
mot_desc('februrary',542).
mot_desc('feedback',543).
mot_desc('field',544).
mot_desc('fields',545).
mot_desc('file',546).
mot_desc('filechooser',547).
mot_desc('fileinputstream',548).
mot_desc('fileoutputstream',549).
mot_desc('files',550).
mot_desc('finalization',551).
mot_desc('finally',552).
mot_desc('find',553).
mot_desc('fire',554).
mot_desc('first',555).
mot_desc('fixed',556).
mot_desc('fixes',557).
mot_desc('flags',558).
mot_desc('floating-point',559).
mot_desc('focus',560).
mot_desc('font',561).
mot_desc('fonts',562).
mot_desc('footprint',563).
mot_desc('for',564).
mot_desc('for?',565).
mot_desc('form',566).
mot_desc('formal',567).
mot_desc('format',568).
mot_desc('formats?',569).
mot_desc('formatting',570).
mot_desc('formerly',571).
mot_desc('forms',572).
mot_desc('forte',573).
mot_desc('forum',574).
mot_desc('forums',575).
mot_desc('found',576).
mot_desc('fp-strict',577).
mot_desc('fr',578).
mot_desc('fractal',579).
mot_desc('frames',580).
mot_desc('framework',581).
mot_desc('free?',582).
mot_desc('french',583).
mot_desc('frequently',584).
mot_desc('frequently-asked',585).
mot_desc('from',586).
mot_desc('from?',587).
mot_desc('front',588).
mot_desc('ftp',589).
mot_desc('full-screen',590).
mot_desc('full-speed',591).
mot_desc('fully',592).
mot_desc('functionality?',593).
mot_desc('fundamentals',594).
mot_desc('general',595).
mot_desc('geom',596).
mot_desc('get',597).
mot_desc('getbyname',598).
mot_desc('getcodebase',599).
mot_desc('getdocumentbase',600).
mot_desc('getreaderfortext',601).
mot_desc('getting',602).
mot_desc('glance',603).
mot_desc('glibc',604).
mot_desc('glossary',605).
mot_desc('gnome',606).
mot_desc('going',607).
mot_desc('gosling',608).
mot_desc('grammar',609).
mot_desc('graphical',610).
mot_desc('graphics',611).
mot_desc('graphicstest',612).
mot_desc('groups',613).
mot_desc('gss-api',614).
mot_desc('gui',615).
mot_desc('guide',616).
mot_desc('guidelines',617).
mot_desc('guis',618).
mot_desc('handlers',619).
mot_desc('handling',620).
mot_desc('hands-on',621).
mot_desc('has',622).
mot_desc('have',623).
mot_desc('haven',624).
mot_desc('heap',625).
mot_desc('heard',626).
mot_desc('hello',627).
mot_desc('helloworld',628).
mot_desc('help',629).
mot_desc('helpers',630).
mot_desc('helps',631).
mot_desc('hierarchy',632).
mot_desc('high',633).
mot_desc('home',634).
mot_desc('host',635).
mot_desc('hosts',636).
mot_desc('hotjava',637).
mot_desc('hotjava?',638).
mot_desc('hotspot',639).
mot_desc('hotspot[tm]',640).
mot_desc('how',641).
mot_desc('html',642).
mot_desc('http',643).
mot_desc('ide',644).
mot_desc('identification',645).
mot_desc('identifiers',646).
mot_desc('identity',647).
mot_desc('ides',648).
mot_desc('idl',649).
mot_desc('idlj',650).
mot_desc('ieee',651).
mot_desc('ietf',652).
mot_desc('ii',653).
mot_desc('iii',654).
mot_desc('iiop',655).
mot_desc('im',656).
mot_desc('image',657).
mot_desc('imageio',658).
mot_desc('imagemap',659).
mot_desc('imaging',660).
mot_desc('implement',661).
mot_desc('implementation',662).
mot_desc('implementations',663).
mot_desc('implementing',664).
mot_desc('import',665).
mot_desc('improvements',666).
mot_desc('in',667).
mot_desc('inc',668).
mot_desc('including',669).
mot_desc('incompatibilities',670).
mot_desc('in-depth',671).
mot_desc('index',672).
mot_desc('industry',673).
mot_desc('inetaddress',674).
mot_desc('inetaddresss',675).
mot_desc('inetsocketaddress',676).
mot_desc('info',677).
mot_desc('information',678).
mot_desc('informed',679).
mot_desc('inheritance',680).
mot_desc('initial',681).
mot_desc('initialization',682).
mot_desc('initializers',683).
mot_desc('initializing',684).
mot_desc('innerclasses',685).
mot_desc('input',686).
mot_desc('inputstreamreader',687).
mot_desc('ins',688).
mot_desc('install',689).
mot_desc('installation',690).
mot_desc('installshield',691).
mot_desc('instance',692).
mot_desc('instances',693).
mot_desc('instruction',694).
mot_desc('instructions',695).
mot_desc('int',696).
mot_desc('integral',697).
mot_desc('intel',698).
mot_desc('interaction',699).
mot_desc('interceptors',700).
mot_desc('interface',701).
mot_desc('interfaces',702).
mot_desc('internal',703).
mot_desc('internationalization',704).
mot_desc('internet',705).
mot_desc('internet-drafts',706).
mot_desc('interoperable',707).
mot_desc('inter-orb',708).
mot_desc('interviews',709).
mot_desc('introducing',710).
mot_desc('introduction',711).
mot_desc('introductory',712).
mot_desc('invest',713).
mot_desc('investor',714).
mot_desc('invocation',715).
mot_desc('invoke',716).
mot_desc('invoking',717).
mot_desc('io',718).
mot_desc('iop',719).
mot_desc('ipv6',720).
mot_desc('is',721).
mot_desc('it',722).
mot_desc('it?',723).
mot_desc('itanium',724).
mot_desc('j2ee',725).
mot_desc('j2me',726).
mot_desc('j2sdk',727).
mot_desc('j2se',728).
mot_desc('j2se-comments',729).
mot_desc('jaas',730).
mot_desc('james',731).
mot_desc('japanese',732).
mot_desc('jar',733).
mot_desc('jarfile',734).
mot_desc('jarurlconnection',735).
mot_desc('java',736).
mot_desc('java?',737).
mot_desc('java2d-comments',738).
mot_desc('java-awt',739).
mot_desc('javabeans',740).
mot_desc('java-beans',741).
mot_desc('javac',742).
mot_desc('javaclasslibvol1examples',743).
mot_desc('javaclasslibvol2examples',744).
mot_desc('java-debugger',745).
mot_desc('java-dnd',746).
mot_desc('javadoc',747).
mot_desc('javadoc?',748).
mot_desc('javadoctoolpage',749).
mot_desc('java-gc-comments',750).
mot_desc('javaidl',751).
mot_desc('java-intl',752).
mot_desc('java-io',753).
mot_desc('javamailapi',754).
mot_desc('javaone',755).
mot_desc('java-print',756).
mot_desc('javascript',757).
mot_desc('java-security',758).
mot_desc('javaserver',759).
mot_desc('javasound-comments',760).
mot_desc('java-technology',761).
mot_desc('java-to-javascript',762).
mot_desc('javax',763).
mot_desc('java2',764).
mot_desc('jce',765).
mot_desc('jdbc',766).
mot_desc('jdbc-odbc',767).
mot_desc('jdc',768).
mot_desc('jdk',769).
mot_desc('jdk-comments',770).
mot_desc('jeditorpane',771).
mot_desc('jfc',772).
mot_desc('jgss',773).
mot_desc('jguru',774).
mot_desc('jls',775).
mot_desc('jms',776).
mot_desc('jndi',777).
mot_desc('jndi-interest',778).
mot_desc('jni',779).
mot_desc('jnlp',780).
mot_desc('join',781).
mot_desc('jpda',782).
mot_desc('jpeg',783).
mot_desc('jre',784).
mot_desc('jre?',785).
mot_desc('jre-comments',786).
mot_desc('jsr-006',787).
mot_desc('jsr-015',788).
mot_desc('jsr015-comments',789).
mot_desc('jsr-51',790).
mot_desc('jsse',791).
mot_desc('jtabbedpane',792).
mot_desc('jumpingbox',793).
mot_desc('jvmpi',794).
mot_desc('jxta',795).
mot_desc('kadmin',796).
mot_desc('kb',797).
mot_desc('kde',798).
mot_desc('keep',799).
mot_desc('kerberos',800).
mot_desc('kernel',801).
mot_desc('keyboard',802).
mot_desc('kinit',803).
mot_desc('kit',804).
mot_desc('klist',805).
mot_desc('know',806).
mot_desc('known',807).
mot_desc('labs',808).
mot_desc('lang',809).
mot_desc('language',810).
mot_desc('language?',811).
mot_desc('large',812).
mot_desc('later',813).
mot_desc('latest',814).
mot_desc('launch',815).
mot_desc('launcher',816).
mot_desc('layout',817).
mot_desc('ldap',818).
mot_desc('learn',819).
mot_desc('learned',820).
mot_desc('learning',821).
mot_desc('lee',822).
mot_desc('legal',823).
mot_desc('let',824).
mot_desc('libjvm',825).
mot_desc('libraries',826).
mot_desc('library',827).
mot_desc('libsafe',828).
mot_desc('license',829).
mot_desc('licensing',830).
mot_desc('licensing?',831).
mot_desc('limitations',832).
mot_desc('linenumbertable',833).
mot_desc('linking',834).
mot_desc('links',835).
mot_desc('linux',836).
mot_desc('list',837).
mot_desc('listed',838).
mot_desc('lists',839).
mot_desc('literals',840).
mot_desc('live',841).
mot_desc('load',842).
mot_desc('loader',843).
mot_desc('loading',844).
mot_desc('local',845).
mot_desc('locale',846).
mot_desc('locales',847).
mot_desc('localized',848).
mot_desc('localvariabletable',849).
mot_desc('location',850).
mot_desc('location-independent',851).
mot_desc('locators',852).
mot_desc('locks',853).
mot_desc('logging',854).
mot_desc('login',855).
mot_desc('loginmodule',856).
mot_desc('logo',857).
mot_desc('logo?',858).
mot_desc('long',859).
mot_desc('long-term',860).
mot_desc('lookup',861).
mot_desc('lookups',862).
mot_desc('mac',863).
mot_desc('machine',864).
mot_desc('machines',865).
mot_desc('mail',866).
mot_desc('main',867).
mot_desc('maintenance',868).
mot_desc('make',869).
mot_desc('man',870).
mot_desc('management',871).
mot_desc('managers',872).
mot_desc('manipulation',873).
mot_desc('manner',874).
mot_desc('manpage',875).
mot_desc('mapping',876).
mot_desc('market?',877).
mot_desc('marketing',878).
mot_desc('masthead',879).
mot_desc('math',880).
mot_desc('me',881).
mot_desc('meaning',882).
mot_desc('mechanism',883).
mot_desc('media',884).
mot_desc('meet',885).
mot_desc('mega',886).
mot_desc('megabytes',887).
mot_desc('members',888).
mot_desc('memory',889).
mot_desc('message',890).
mot_desc('messaging',891).
mot_desc('metadata',892).
mot_desc('metal',893).
mot_desc('method',894).
mot_desc('methods',895).
mot_desc('micro',896).
mot_desc('microsoft',897).
mot_desc('microsoftwindows',898).
mot_desc('microsystems',899).
mot_desc('midi',900).
mot_desc('minimum',901).
mot_desc('miscellaneous',902).
mot_desc('mnemonics',903).
mot_desc('mode',904).
mot_desc('model',905).
mot_desc('modes',906).
mot_desc('modifiers',907).
mot_desc('modularity',908).
mot_desc('module',909).
mot_desc('modules',910).
mot_desc('more',911).
mot_desc('mounted',912).
mot_desc('multi',913).
mot_desc('multiple',914).
mot_desc('must',915).
mot_desc('my',916).
mot_desc('n1',917).
mot_desc('name',918).
mot_desc('named',919).
mot_desc('names',920).
mot_desc('namespace',921).
mot_desc('naming',922).
mot_desc('namingcontextextpackage',923).
mot_desc('namingcontextpackage',924).
mot_desc('narrowing',925).
mot_desc('nascar',926).
mot_desc('native',927).
mot_desc('native2ascii',928).
mot_desc('navigator',929).
mot_desc('nc03-q2',930).
mot_desc('need',931).
mot_desc('nervoustext',932).
mot_desc('nested',933).
mot_desc('net',934).
mot_desc('netbeans',935).
mot_desc('netscape',936).
mot_desc('netscape?',937).
mot_desc('network',938).
mot_desc('networking',939).
mot_desc('networkinterface',940).
mot_desc('new',941).
mot_desc('newinstance',942).
mot_desc('newly',943).
mot_desc('news',944).
mot_desc('newsgroups',945).
mot_desc('newsletter',946).
mot_desc('newsletters',947).
mot_desc('new-to-java',948).
mot_desc('next',949).
mot_desc('nio',950).
mot_desc('no',951).
mot_desc('noclassdeffounderror',952).
mot_desc('nonatomic',953).
mot_desc('normal',954).
mot_desc('not',955).
mot_desc('not?',956).
mot_desc('notation',957).
mot_desc('note',958).
mot_desc('notes',959).
mot_desc('notice',960).
mot_desc('notification',961).
mot_desc('nt',962).
mot_desc('number',963).
mot_desc('numberformat',964).
mot_desc('numeric',965).
mot_desc('nvidia',966).
mot_desc('object',967).
mot_desc('objectimpl',968).
mot_desc('objects',969).
mot_desc('of',970).
mot_desc('official',971).
mot_desc('omg',972).
mot_desc('on',973).
mot_desc('on?',974).
mot_desc('one',975).
mot_desc('online',976).
mot_desc('on-line',977).
mot_desc('onlineonly',978).
mot_desc('opcode',979).
mot_desc('opcodes',980).
mot_desc('open_delete',981).
mot_desc('open_read',982).
mot_desc('openwindows',983).
mot_desc('operand',984).
mot_desc('operating',985).
mot_desc('operations',986).
mot_desc('operators',987).
mot_desc('optimize',988).
mot_desc('optional',989).
mot_desc('optional-package',990).
mot_desc('options',991).
mot_desc('or',992).
mot_desc('orb',993).
mot_desc('orbd',994).
mot_desc('orbinitinfopackage',995).
mot_desc('orbpackage',996).
mot_desc('order',997).
mot_desc('org',998).
mot_desc('other',999).
mot_desc('our',1000).
mot_desc('out',1001).
mot_desc('out-of-order',1002).
mot_desc('output',1003).
mot_desc('outputstreamwriter',1004).
mot_desc('over',1005).
mot_desc('overloading',1006).
mot_desc('override',1007).
mot_desc('overriding',1008).
mot_desc('overview',1009).
mot_desc('pack',1010).
mot_desc('package',1011).
mot_desc('packages',1012).
mot_desc('packaging',1013).
mot_desc('page',1014).
mot_desc('pages',1015).
mot_desc('panel',1016).
mot_desc('paper',1017).
mot_desc('parade',1018).
mot_desc('parameters',1019).
mot_desc('parser',1020).
mot_desc('parsers',1021).
mot_desc('part',1022).
mot_desc('partners',1023).
mot_desc('patches',1024).
mot_desc('path',1025).
mot_desc('paths',1026).
mot_desc('pathseparatorchar',1027).
mot_desc('pattern',1028).
mot_desc('pc',1029).
mot_desc('pdf',1030).
mot_desc('pelegri',1031).
mot_desc('pentium',1032).
mot_desc('performance',1033).
mot_desc('permissions',1034).
mot_desc('persistence',1035).
mot_desc('persistent',1036).
mot_desc('phantomreference',1037).
mot_desc('plaf',1038).
mot_desc('planning',1039).
mot_desc('plans',1040).
mot_desc('platform',1041).
mot_desc('platform?',1042).
mot_desc('platform--jdc',1043).
mot_desc('platforms',1044).
mot_desc('please',1045).
mot_desc('plugin',1046).
mot_desc('plug-in',1047).
mot_desc('plugins',1048).
mot_desc('plugin-sendusmail',1049).
mot_desc('poa',1050).
mot_desc('poamanagerpackage',1051).
mot_desc('poapackage',1052).
mot_desc('policy',1053).
mot_desc('pool',1054).
mot_desc('pooling',1055).
mot_desc('portable',1056).
mot_desc('portableinterceptor',1057).
mot_desc('portableserver',1058).
mot_desc('porting',1059).
mot_desc('possible',1060).
mot_desc('postscript',1061).
mot_desc('praise',1062).
mot_desc('preface',1063).
mot_desc('preferences',1064).
mot_desc('prefs',1065).
mot_desc('premier',1066).
mot_desc('preparation',1067).
mot_desc('prescient',1068).
mot_desc('press',1069).
mot_desc('previous',1070).
mot_desc('price?',1071).
mot_desc('primitive',1072).
mot_desc('print',1073).
mot_desc('printable',1074).
mot_desc('printing',1075).
mot_desc('privacy',1076).
mot_desc('privacypolicy',1077).
mot_desc('private',1078).
mot_desc('privileged',1079).
mot_desc('problem',1080).
mot_desc('problems',1081).
mot_desc('procedure',1082).
mot_desc('process',1083).
mot_desc('processing',1084).
mot_desc('processors',1085).
mot_desc('product',1086).
mot_desc('production',1087).
mot_desc('products',1088).
mot_desc('professionals',1089).
mot_desc('profiler',1090).
mot_desc('program',1091).
mot_desc('programmer',1092).
mot_desc('programming',1093).
mot_desc('programs',1094).
mot_desc('project',1095).
mot_desc('promotion',1096).
mot_desc('promotions',1097).
mot_desc('properties',1098).
mot_desc('property',1099).
mot_desc('proprietary',1100).
mot_desc('protocol',1101).
mot_desc('provider',1102).
mot_desc('providers',1103).
mot_desc('proxy',1104).
mot_desc('ps',1105).
mot_desc('public',1106).
mot_desc('put',1107).
mot_desc('putting',1108).
mot_desc('q142',1109).
mot_desc('qualified',1110).
mot_desc('quality',1111).
mot_desc('questions',1112).
mot_desc('racing',1113).
mot_desc('ram',1114).
mot_desc('randomaccessfile',1115).
mot_desc('reader',1116).
mot_desc('readme',1117).
mot_desc('really',1118).
mot_desc('receiving',1119).
mot_desc('recommended',1120).
mot_desc('redistribution',1121).
mot_desc('ref',1122).
mot_desc('reference',1123).
mot_desc('referencequeue',1124).
mot_desc('references',1125).
mot_desc('reflect',1126).
mot_desc('reflection',1127).
mot_desc('reflection-comments',1128).
mot_desc('reflective',1129).
mot_desc('regex',1130).
mot_desc('register',1131).
mot_desc('register?',1132).
mot_desc('registered',1133).
mot_desc('registry',1134).
mot_desc('regular',1135).
mot_desc('related',1136).
mot_desc('relations',1137).
mot_desc('release',1138).
mot_desc('releases',1139).
mot_desc('reliable',1140).
mot_desc('remote',1141).
mot_desc('remoteserver',1142).
mot_desc('renderable',1143).
mot_desc('report',1144).
mot_desc('reports',1145).
mot_desc('representation',1146).
mot_desc('request',1147).
mot_desc('requestfocus',1148).
mot_desc('required',1149).
mot_desc('requirements',1150).
mot_desc('research',1151).
mot_desc('reserved',1152).
mot_desc('resolution',1153).
mot_desc('resourcebundle',1154).
mot_desc('resources',1155).
mot_desc('return',1156).
mot_desc('review',1157).
mot_desc('reviews',1158).
mot_desc('revocation',1159).
mot_desc('rfc',1160).
mot_desc('rights',1161).
mot_desc('rmi',1162).
mot_desc('rmic',1163).
mot_desc('rmiclassloader',1164).
mot_desc('rmiclassloaderspi',1165).
mot_desc('rmi-comments',1166).
mot_desc('rmifailurehandler',1167).
mot_desc('rmi-iiop',1168).
mot_desc('rmisocketfactory',1169).
mot_desc('rmi-users',1170).
mot_desc('rosanna',1171).
mot_desc('rowset',1172).
mot_desc('rpm',1173).
mot_desc('rtf',1174).
mot_desc('rules',1175).
mot_desc('run',1176).
mot_desc('running',1177).
mot_desc('runtime',1178).
mot_desc('runtimeexception',1179).
mot_desc('samba',1180).
mot_desc('sample',1181).
mot_desc('sampled',1182).
mot_desc('samples',1183).
mot_desc('save!',1184).
mot_desc('sax',1185).
mot_desc('scalability',1186).
mot_desc('scripting',1187).
mot_desc('scripts?',1188).
mot_desc('sdk',1189).
mot_desc('sdk-docl10nbugs',1190).
mot_desc('se',1191).
mot_desc('search',1192).
mot_desc('seat',1193).
mot_desc('second',1194).
mot_desc('section',1195).
mot_desc('secure',1196).
mot_desc('secureclassloader',1197).
mot_desc('security',1198).
mot_desc('securitymanager',1199).
mot_desc('see',1200).
mot_desc('selectbesttextflavor',1201).
mot_desc('selecting',1202).
mot_desc('self-extracting',1203).
mot_desc('send',1204).
mot_desc('sendingcontext',1205).
mot_desc('separately',1206).
mot_desc('serialization',1207).
mot_desc('serialver',1208).
mot_desc('series',1209).
mot_desc('servant',1210).
mot_desc('servantlocatorpackage',1211).
mot_desc('server',1212).
mot_desc('servers',1213).
mot_desc('server-side',1214).
mot_desc('serversocket',1215).
mot_desc('servertool',1216).
mot_desc('service',1217).
mot_desc('services',1218).
mot_desc('servlets',1219).
mot_desc('sessions',1220).
mot_desc('set',1221).
mot_desc('setfailurehandler',1222).
mot_desc('setlog',1223).
mot_desc('sets',1224).
mot_desc('shift-jis',1225).
mot_desc('short',1226).
mot_desc('should',1227).
mot_desc('sign',1228).
mot_desc('signal',1229).
mot_desc('signature',1230).
mot_desc('sign-on',1231).
mot_desc('silent',1232).
mot_desc('simple',1233).
mot_desc('simplegraph',1234).
mot_desc('since',1235).
mot_desc('single',1236).
mot_desc('site',1237).
mot_desc('sites',1238).
mot_desc('skeleton',1239).
mot_desc('skip',1240).
mot_desc('so',1241).
mot_desc('soap',1242).
mot_desc('socket',1243).
mot_desc('socketaddress',1244).
mot_desc('softreference',1245).
mot_desc('software',1246).
mot_desc('software?',1247).
mot_desc('solaris',1248).
mot_desc('solarisandlinux',1249).
mot_desc('solarisoperatingenvironment',1250).
mot_desc('solution',1251).
mot_desc('solutions',1252).
mot_desc('some',1253).
mot_desc('sortdemo',1254).
mot_desc('sound',1255).
mot_desc('source',1256).
mot_desc('sourcefile',1257).
mot_desc('sparc',1258).
mot_desc('sparcstation',1259).
mot_desc('spec',1260).
mot_desc('special',1261).
mot_desc('specialized',1262).
mot_desc('specially',1263).
mot_desc('specification',1264).
mot_desc('specifications',1265).
mot_desc('speed',1266).
mot_desc('spi',1267).
mot_desc('spreadsheet',1268).
mot_desc('sql',1269).
mot_desc('ssl',1270).
mot_desc('stack',1271).
mot_desc('stacks',1272).
mot_desc('stand',1273).
mot_desc('standard',1274).
mot_desc('standards',1275).
mot_desc('start',1276).
mot_desc('started',1277).
mot_desc('started?',1278).
mot_desc('starttlsrequest',1279).
mot_desc('starttlsresponse',1280).
mot_desc('startup',1281).
mot_desc('start-up',1282).
mot_desc('statement',1283).
mot_desc('static',1284).
mot_desc('staying',1285).
mot_desc('step',1286).
mot_desc('step-by-step',1287).
mot_desc('stock',1288).
mot_desc('stop',1289).
mot_desc('store',1290).
mot_desc('stream',1291).
mot_desc('streams',1292).
mot_desc('string',1293).
mot_desc('stringbuffer',1294).
mot_desc('structural',1295).
mot_desc('structure',1296).
mot_desc('structures',1297).
mot_desc('stub',1298).
mot_desc('studies',1299).
mot_desc('studio',1300).
mot_desc('styled',1301).
mot_desc('subclasses',1302).
mot_desc('submit',1303).
mot_desc('submitting',1304).
mot_desc('subscribe',1305).
mot_desc('subsequent',1306).
mot_desc('suggestions',1307).
mot_desc('suggestions?',1308).
mot_desc('summary',1309).
mot_desc('sun',1310).
mot_desc('sunsolve',1311).
mot_desc('sup1-egs',1312).
mot_desc('superclass',1313).
mot_desc('superclasses',1314).
mot_desc('superinterfaces',1315).
mot_desc('supplement',1316).
mot_desc('supplemental',1317).
mot_desc('support',1318).
mot_desc('supported',1319).
mot_desc('supporting',1320).
mot_desc('suppressing',1321).
mot_desc('swap',1322).
mot_desc('swing',1323).
mot_desc('swing-feedback',1324).
mot_desc('switches',1325).
mot_desc('synchronization',1326).
mot_desc('syntax',1327).
mot_desc('synthetic',1328).
mot_desc('system',1329).
mot_desc('table',1330).
mot_desc('tag',1331).
mot_desc('tags',1332).
mot_desc('team',1333).
mot_desc('tech',1334).
mot_desc('technical',1335).
mot_desc('technologies',1336).
mot_desc('technology',1337).
mot_desc('technology?',1338).
mot_desc('technology-enabled',1339).
mot_desc('terminology',1340).
mot_desc('terms',1341).
mot_desc('tester?',1342).
mot_desc('text',1343).
mot_desc('textarea',1344).
mot_desc('that',1345).
mot_desc('the',1346).
mot_desc('they',1347).
mot_desc('third',1348).
mot_desc('this',1349).
mot_desc('thomas',1350).
mot_desc('thread',1351).
mot_desc('threads',1352).
mot_desc('throwing',1353).
mot_desc('tictactoe',1354).
mot_desc('time',1355).
mot_desc('timeouts',1356).
mot_desc('tips',1357).
mot_desc('tm',1358).
mot_desc('tnameserv',1359).
mot_desc('to',1360).
mot_desc('together',1361).
mot_desc('tool',1362).
mot_desc('toolkit',1363).
mot_desc('tools',1364).
mot_desc('top',1365).
mot_desc('topcoders',1366).
mot_desc('topics',1367).
mot_desc('touch',1368).
mot_desc('trademark',1369).
mot_desc('trademarks',1370).
mot_desc('trail',1371).
mot_desc('training',1372).
mot_desc('transaction',1373).
mot_desc('transfer',1374).
mot_desc('transferable',1375).
mot_desc('transform',1376).
mot_desc('transient',1377).
mot_desc('translations',1378).
mot_desc('transport',1379).
mot_desc('treatment',1380).
mot_desc('tree',1381).
mot_desc('troubleshooting',1382).
mot_desc('try',1383).
mot_desc('tutorial',1384).
mot_desc('tutorials',1385).
mot_desc('txt',1386).
mot_desc('type',1387).
mot_desc('typecodepackage',1388).
mot_desc('types',1389).
mot_desc('ulrich',1390).
mot_desc('under',1391).
mot_desc('undo',1392).
mot_desc('unicode',1393).
mot_desc('unified',1394).
mot_desc('unix',1395).
mot_desc('unloading',1396).
mot_desc('update',1397).
mot_desc('uri',1398).
mot_desc('urlclassloader',1399).
mot_desc('urls',1400).
mot_desc('us',1401).
mot_desc('usage',1402).
mot_desc('use',1403).
mot_desc('used',1404).
mot_desc('useful',1405).
mot_desc('user',1406).
mot_desc('user-defined',1407).
mot_desc('using',1408).
mot_desc('util',1409).
mot_desc('utilities',1410).
mot_desc('utility',1411).
mot_desc('v1',1412).
mot_desc('v2',1413).
mot_desc('v8',1414).
mot_desc('value',1415).
mot_desc('values',1416).
mot_desc('variables',1417).
mot_desc('vb',1418).
mot_desc('ve',1419).
mot_desc('vendors',1420).
mot_desc('verification',1421).
mot_desc('verifier',1422).
mot_desc('verify',1423).
mot_desc('version',1424).
mot_desc('versioning',1425).
mot_desc('versions',1426).
mot_desc('via',1427).
mot_desc('vie',1428).
mot_desc('view',1429).
mot_desc('virtual',1430).
mot_desc('virus',1431).
mot_desc('virus-checker',1432).
mot_desc('vm',1433).
mot_desc('vol',1434).
mot_desc('volatile',1435).
mot_desc('volume',1436).
mot_desc('w3c',1437).
mot_desc('wait',1438).
mot_desc('walk-through',1439).
mot_desc('we',1440).
mot_desc('weakreference',1441).
mot_desc('wear',1442).
mot_desc('web',1443).
mot_desc('web-service-ready',1444).
mot_desc('website',1445).
mot_desc('what',1446).
mot_desc('when',1447).
mot_desc('where',1448).
mot_desc('where?',1449).
mot_desc('white',1450).
mot_desc('who',1451).
mot_desc('why',1452).
mot_desc('widening',1453).
mot_desc('will',1454).
mot_desc('window',1455).
mot_desc('windows',1456).
mot_desc('winhelp',1457).
mot_desc('wire',1458).
mot_desc('wireless',1459).
mot_desc('with',1460).
mot_desc('work',1461).
mot_desc('workbook',1462).
mot_desc('working',1463).
mot_desc('world',1464).
mot_desc('worldwide',1465).
mot_desc('write',1466).
mot_desc('writes',1467).
mot_desc('writing',1468).
mot_desc('wsdp',1469).
mot_desc('www',1470).
mot_desc('x500',1471).
mot_desc('xa',1472).
mot_desc('xml',1473).
mot_desc('xp',1474).
mot_desc('xxx',1475).
mot_desc('you',1476).
mot_desc('your',1477).
mot_desc('yyy?',1478).
mot_desc('zip',1479).
mot_desc('zipfile',1480).
mot_desc('zone',1481).
mot_desc('zzz',1482).
mot_desc('',1483).
mot_desc('2003',1484).
mot_desc('2sdk',1485).
mot_desc('printable',1486).
/* (1486 entrees) */


/**
 * Table href
 */

href(2,1,1).
href(2,2,2).
href(2,3,3).
href(2,4,4).
href(2,5,5).
href(2,6,6).
href(2,7,7).
href(2,8,8).
href(2,9,9).
href(2,10,10).
href(2,11,11).
href(2,12,12).
href(2,13,13).
href(2,14,14).
href(2,15,15).
href(2,16,16).
href(2,17,17).
href(2,18,18).
href(2,19,19).
href(2,20,20).
href(2,21,21).
href(2,22,22).
href(2,23,23).
href(2,24,24).
href(2,25,25).
href(2,26,26).
href(2,27,27).
href(2,28,28).
href(2,29,29).
href(2,30,30).
href(2,31,31).
href(2,32,32).
href(2,33,33).
href(2,34,34).
href(2,35,35).
href(2,36,36).
href(2,37,37).
href(2,38,38).
href(2,39,39).
href(2,40,40).
href(2,41,41).
href(2,42,42).
href(2,43,43).
href(2,44,44).
href(2,45,45).
href(2,46,46).
href(2,47,47).
href(2,48,48).
href(2,49,49).
href(2,50,50).
href(2,51,51).
href(2,52,52).
href(2,53,53).
href(2,54,54).
href(2,55,55).
href(2,56,56).
href(2,57,57).
href(2,58,58).
href(2,59,59).
href(2,60,60).
href(2,61,61).
href(2,62,62).
href(2,63,63).
href(2,64,64).
href(2,65,65).
href(2,66,66).
href(2,67,67).
href(2,68,68).
href(2,69,69).
href(2,70,70).
href(2,71,71).
href(2,72,72).
href(2,73,73).
href(2,74,74).
href(2,75,75).
href(2,76,76).
href(2,77,77).
href(2,78,78).
href(2,79,79).
href(2,80,80).
href(2,81,81).
href(2,82,82).
href(2,83,83).
href(2,84,84).
href(2,85,85).
href(2,86,86).
href(2,87,87).
href(2,88,88).
href(2,89,89).
href(2,90,90).
href(2,91,91).
href(2,92,92).
href(2,93,93).
href(2,94,94).
href(2,95,95).
href(2,96,96).
href(1,1,97).
href(1,2,98).
href(2,97,99).
href(2,98,100).
href(2,99,101).
href(2,100,102).
href(2,101,103).
href(2,102,104).
href(2,103,105).
href(2,104,106).
href(2,105,107).
href(2,106,108).
href(1,3,109).
href(2,107,110).
href(2,108,111).
href(2,109,112).
href(2,110,113).
href(1,4,114).
href(2,111,115).
href(2,112,116).
href(2,113,117).
href(2,114,118).
href(2,115,119).
href(2,116,120).
href(2,117,121).
href(2,118,122).
href(2,119,123).
href(2,120,124).
href(2,121,125).
href(2,122,126).
href(2,123,127).
href(2,124,128).
href(2,125,129).
href(2,126,130).
href(2,127,131).
href(2,128,132).
href(2,129,133).
href(2,130,134).
href(2,131,135).
href(2,132,136).
href(2,133,137).
href(2,134,138).
href(2,135,139).
href(2,136,140).
href(2,137,141).
href(2,138,142).
href(2,139,143).
href(2,140,144).
href(2,141,145).
href(2,142,146).
href(2,143,147).
href(2,144,148).
href(2,145,149).
href(2,146,150).
href(2,147,151).
href(2,148,152).
href(2,149,153).
href(2,150,154).
href(2,151,155).
href(2,152,156).
href(2,153,157).
href(2,154,158).
href(2,155,159).
href(2,156,160).
href(2,157,161).
href(2,158,162).
href(2,159,163).
href(2,160,164).
href(2,161,165).
href(2,162,166).
href(2,163,167).
href(2,164,168).
href(2,165,169).
href(2,166,170).
href(2,167,171).
href(2,168,172).
href(2,169,173).
href(2,170,174).
href(2,171,175).
href(2,172,176).
href(2,173,177).
href(2,174,178).
href(2,175,179).
href(2,176,180).
href(2,177,181).
href(2,178,182).
href(2,179,183).
href(2,180,184).
href(2,181,185).
href(2,182,186).
href(2,183,187).
href(2,184,188).
href(2,185,189).
href(2,186,190).
href(2,187,191).
href(2,188,192).
href(2,189,193).
href(2,190,194).
href(2,191,195).
href(2,192,196).
href(2,193,197).
href(2,194,198).
href(2,195,199).
href(2,196,200).
href(2,197,201).
href(2,198,202).
href(2,199,203).
href(2,200,204).
href(2,201,205).
href(2,202,206).
href(2,203,207).
href(2,204,208).
href(2,205,209).
href(2,206,210).
href(2,207,211).
href(2,208,212).
href(2,209,213).
href(2,210,214).
href(2,211,215).
href(2,212,216).
href(2,213,217).
href(2,214,218).
href(2,215,219).
href(2,216,220).
href(2,217,221).
href(2,218,222).
href(2,219,223).
href(2,220,224).
href(2,221,225).
href(2,222,226).
href(2,223,227).
href(2,224,228).
href(2,225,229).
href(2,226,230).
href(2,227,231).
href(2,228,232).
href(2,229,233).
href(2,230,234).
href(2,231,235).
href(2,232,236).
href(2,233,237).
href(2,234,238).
href(2,235,239).
href(2,236,240).
href(2,237,241).
href(2,238,242).
href(2,239,243).
href(2,240,244).
href(2,241,245).
href(2,242,246).
href(2,243,247).
href(2,244,248).
href(2,245,249).
href(2,246,250).
href(2,247,251).
href(2,248,252).
href(2,249,253).
href(2,250,254).
href(2,251,255).
href(2,252,256).
href(2,253,257).
href(2,254,258).
href(2,255,259).
href(2,256,260).
href(2,257,261).
href(2,258,262).
href(2,259,263).
href(2,260,264).
href(2,261,265).
href(2,262,266).
href(2,263,267).
href(2,264,268).
href(2,265,269).
href(2,266,270).
href(2,267,271).
href(2,268,272).
href(2,269,273).
href(2,270,274).
href(2,271,275).
href(2,272,276).
href(2,273,277).
href(2,274,278).
href(2,275,279).
href(2,276,280).
href(2,277,281).
href(2,278,282).
href(2,279,283).
href(2,280,284).
href(2,281,285).
href(2,282,286).
href(2,283,287).
href(2,284,288).
href(2,285,289).
href(2,286,290).
href(2,287,291).
href(2,288,292).
href(2,289,293).
href(2,290,294).
href(2,291,295).
href(2,292,296).
href(2,293,297).
href(2,294,298).
href(2,295,299).
href(2,296,300).
href(2,297,301).
href(1,5,302).
href(2,298,303).
href(2,299,304).
href(2,300,305).
href(2,301,306).
href(2,302,307).
href(2,303,308).
href(2,304,309).
href(2,305,310).
href(2,306,311).
href(2,307,312).
href(2,308,313).
href(2,309,314).
href(2,310,315).
href(2,311,316).
href(2,312,317).
href(2,313,318).
href(2,314,319).
href(2,315,320).
href(2,316,321).
href(2,317,322).
href(2,318,323).
href(2,319,324).
href(2,320,325).
href(2,321,326).
href(2,322,327).
href(2,323,328).
href(2,324,329).
href(2,325,330).
href(2,326,331).
href(2,327,332).
href(2,328,333).
href(2,329,334).
href(2,330,335).
href(2,331,336).
href(2,332,337).
href(2,333,338).
href(2,334,339).
href(2,335,340).
href(2,336,341).
href(2,337,342).
href(2,338,343).
href(2,339,344).
href(2,340,345).
href(2,341,346).
href(2,342,347).
href(2,343,348).
href(2,344,349).
href(2,345,350).
href(2,346,351).
href(2,347,352).
href(2,348,353).
href(2,349,354).
href(2,350,355).
href(2,351,356).
href(2,352,357).
href(2,353,358).
href(2,354,359).
href(2,355,360).
href(2,356,361).
href(2,357,362).
href(2,358,363).
href(2,359,364).
href(2,360,365).
href(2,361,366).
href(2,362,367).
href(2,363,368).
href(2,364,369).
href(2,365,370).
href(2,366,371).
href(2,367,372).
href(2,368,373).
href(2,369,374).
href(2,370,375).
href(2,371,376).
href(2,372,377).
href(2,373,378).
href(2,374,379).
href(2,375,380).
href(2,376,381).
href(2,377,382).
href(2,378,383).
href(2,379,384).
href(1,6,385).
href(2,380,386).
href(2,381,387).
href(2,382,388).
href(2,383,389).
href(2,384,390).
href(2,385,391).
href(2,386,392).
href(2,387,393).
href(2,388,394).
href(2,389,395).
href(2,390,396).
href(2,391,397).
href(2,392,398).
href(2,393,399).
href(2,394,400).
href(2,395,401).
href(2,396,402).
href(2,397,403).
href(2,398,404).
href(2,399,405).
href(2,400,406).
href(2,401,407).
href(2,402,408).
href(2,403,409).
href(2,404,410).
href(2,405,411).
href(2,406,412).
href(2,407,413).
href(2,408,414).
href(2,409,415).
href(2,410,416).
href(2,411,417).
href(2,412,418).
href(2,413,419).
href(2,414,420).
href(2,415,421).
href(2,416,422).
href(2,417,423).
href(2,418,424).
href(2,419,425).
href(2,420,426).
href(2,421,427).
href(1,7,428).
href(2,422,429).
href(2,423,430).
href(1,8,431).
href(2,424,432).
href(2,425,433).
href(2,426,434).
href(2,427,435).
href(2,428,436).
href(2,429,437).
href(2,430,438).
href(2,431,439).
href(2,432,440).
href(2,433,441).
href(2,434,442).
href(2,435,443).
href(2,436,444).
href(2,437,445).
href(2,438,446).
href(2,439,447).
href(2,440,448).
href(2,441,449).
href(2,442,450).
href(2,443,451).
href(2,444,452).
href(2,445,453).
href(2,446,454).
href(2,447,455).
href(2,448,456).
href(2,449,457).
href(2,450,458).
href(2,451,459).
href(2,452,460).
href(2,453,461).
href(2,454,462).
href(2,455,463).
href(2,456,464).
href(2,457,465).
href(2,458,466).
href(2,459,467).
href(2,460,468).
href(2,461,469).
href(2,462,470).
href(2,463,471).
href(2,464,472).
href(2,465,473).
href(2,466,474).
href(1,9,475).
href(2,467,476).
href(2,468,477).
href(2,469,478).
href(2,470,479).
href(2,471,480).
href(2,472,481).
href(2,473,482).
href(2,474,483).
href(2,475,484).
href(2,476,485).
href(2,477,486).
href(2,478,487).
href(2,479,488).
href(2,480,489).
href(2,481,490).
href(2,482,491).
href(2,483,492).
href(2,484,493).
href(2,485,494).
href(2,486,495).
href(2,487,496).
href(2,488,497).
href(2,489,498).
href(2,490,499).
href(2,491,500).
href(2,492,501).
href(2,493,502).
href(2,494,503).
href(2,495,504).
href(2,496,505).
href(2,497,506).
href(2,498,507).
href(2,499,508).
href(2,500,509).
href(1,10,510).
href(2,501,511).
href(2,502,512).
href(2,503,513).
href(2,504,514).
href(2,505,515).
href(1,11,516).
href(2,506,517).
href(2,507,518).
href(1,12,519).
href(2,508,520).
href(2,509,521).
href(2,510,522).
href(2,511,523).
href(2,512,524).
href(2,513,525).
href(2,514,526).
href(2,515,527).
href(2,516,528).
href(2,517,529).
href(2,518,530).
href(2,519,531).
href(2,520,532).
href(2,521,533).
href(2,522,534).
href(2,523,535).
href(2,524,536).
href(2,525,537).
href(2,526,538).
href(2,527,539).
href(2,528,540).
href(2,529,541).
href(2,530,542).
href(2,531,543).
href(2,532,544).
href(2,533,545).
href(2,534,546).
href(2,535,547).
href(2,536,548).
href(2,537,549).
href(2,538,550).
href(2,539,551).
href(2,540,552).
href(2,541,553).
href(2,542,554).
href(2,543,555).
href(2,544,556).
href(2,545,557).
href(2,546,558).
href(2,547,559).
href(2,548,560).
href(2,549,561).
href(2,550,562).
href(2,551,563).
href(2,552,564).
href(2,553,565).
href(2,554,566).
href(2,555,567).
href(2,556,568).
href(2,557,569).
href(2,558,570).
href(2,559,571).
href(2,560,572).
href(2,561,573).
href(2,562,574).
href(2,563,575).
href(2,564,576).
href(2,565,577).
href(2,566,578).
href(2,567,579).
href(2,568,580).
href(2,569,581).
href(2,570,582).
href(2,571,583).
href(2,572,584).
href(2,573,585).
href(2,574,586).
href(2,575,587).
href(2,576,588).
href(2,577,589).
href(2,578,590).
href(2,579,591).
href(2,580,592).
href(2,581,593).
href(2,582,594).
href(2,583,595).
href(2,584,596).
href(2,585,597).
href(2,586,598).
href(2,587,599).
href(1,13,600).
href(2,588,601).
href(2,589,602).
href(1,14,603).
href(1,15,604).
href(2,590,605).
href(1,16,606).
href(1,17,607).
href(1,18,608).
href(1,19,609).
href(2,591,610).
href(2,592,611).
href(2,593,612).
href(1,20,613).
href(2,594,614).
href(2,595,615).
href(1,21,616).
href(2,596,617).
href(2,597,618).
href(2,598,619).
href(1,22,620).
href(2,599,621).
href(2,600,622).
href(2,601,623).
href(2,602,624).
href(2,603,625).
href(2,604,626).
href(2,605,627).
href(2,606,628).
href(2,607,629).
href(2,608,630).
href(2,609,631).
href(2,610,632).
href(2,611,633).
href(2,612,634).
href(2,613,635).
href(2,614,636).
href(2,615,637).
href(2,616,638).
href(2,617,639).
href(2,618,640).
href(2,619,641).
href(2,620,642).
href(2,621,643).
href(2,622,644).
href(2,623,645).
href(2,624,646).
href(2,625,647).
href(2,626,648).
href(2,627,649).
href(2,628,650).
href(2,629,651).
href(2,630,652).
href(2,631,653).
href(2,632,654).
href(2,633,655).
href(2,634,656).
href(2,635,657).
href(2,636,658).
href(2,637,659).
href(2,638,660).
href(2,639,661).
href(2,640,662).
href(2,641,663).
href(2,642,664).
href(2,643,665).
href(2,644,666).
href(2,645,667).
href(1,23,668).
href(2,646,669).
href(2,647,670).
href(2,648,671).
href(1,24,672).
href(2,649,673).
href(1,25,674).
href(1,26,675).
href(2,650,676).
href(2,651,677).
href(2,652,678).
href(2,653,679).
href(2,654,680).
href(2,655,681).
href(2,656,682).
href(2,657,683).
href(2,658,684).
href(2,659,685).
href(2,660,686).
href(1,27,687).
href(1,28,688).
href(1,29,689).
href(2,661,690).
href(2,662,691).
href(1,30,692).
href(2,663,693).
href(2,664,694).
href(1,31,695).
href(2,665,696).
href(2,666,697).
href(1,32,698).
href(2,667,699).
href(1,33,700).
href(2,668,701).
href(2,669,702).
href(2,670,703).
href(1,34,704).
href(2,671,705).
href(2,672,706).
href(2,673,707).
href(2,674,708).
href(2,675,709).
href(2,676,710).
href(2,677,711).
href(2,678,712).
href(2,679,713).
href(2,680,714).
href(2,681,715).
href(2,682,716).
href(2,683,717).
href(2,684,718).
href(2,685,719).
href(2,686,720).
href(2,687,721).
href(2,688,722).
href(2,689,723).
href(1,35,724).
href(2,690,725).
href(2,691,726).
href(2,692,727).
href(2,693,728).
href(2,694,729).
href(2,695,730).
href(2,696,731).
href(2,697,732).
href(2,698,733).
href(2,699,734).
href(2,700,735).
href(2,701,736).
href(2,702,737).
href(2,703,738).
href(2,704,739).
href(2,705,740).
href(2,706,741).
href(2,707,742).
href(2,708,743).
href(2,709,744).
href(2,710,745).
href(2,711,746).
href(2,712,747).
href(2,713,748).
href(2,714,749).
href(2,715,750).
href(2,716,751).
href(2,717,752).
href(2,718,753).
href(2,719,754).
href(2,720,755).
href(2,721,756).
href(2,722,757).
href(2,723,758).
href(2,724,759).
href(2,725,760).
href(2,726,761).
href(2,727,762).
href(2,728,763).
href(2,729,764).
href(2,730,765).
href(2,731,766).
href(2,732,767).
href(2,733,768).
href(2,734,769).
href(2,735,770).
href(2,736,771).
href(2,737,772).
href(2,738,773).
href(2,739,774).
href(2,740,775).
href(2,741,776).
href(2,742,777).
href(2,743,778).
href(2,744,779).
href(2,745,780).
href(2,746,781).
href(2,747,782).
href(2,748,783).
href(2,749,784).
href(2,750,785).
href(2,751,786).
href(2,752,787).
href(2,753,788).
href(2,754,789).
href(2,755,790).
href(2,756,791).
href(2,757,792).
href(2,758,793).
href(1,36,794).
href(2,759,795).
href(2,760,796).
href(2,761,797).
href(2,762,798).
href(2,763,799).
href(2,764,800).
href(2,765,801).
href(2,766,802).
href(2,767,803).
href(2,768,804).
href(2,769,805).
href(2,770,806).
href(2,771,807).
href(2,772,808).
href(2,773,809).
href(2,774,810).
href(2,775,811).
href(2,776,812).
href(2,777,813).
href(2,778,814).
href(2,779,815).
href(2,780,816).
href(2,781,817).
href(2,782,818).
href(2,783,819).
href(2,784,820).
href(2,785,821).
href(2,786,822).
href(2,787,823).
href(2,788,824).
href(2,789,825).
href(2,790,826).
href(2,791,827).
href(2,792,828).
href(2,793,829).
href(2,794,830).
href(2,795,831).
href(2,796,832).
href(2,797,833).
href(2,798,834).
href(2,799,835).
href(2,800,836).
href(2,801,837).
href(2,802,838).
href(2,803,839).
href(2,804,840).
href(2,805,841).
href(2,806,842).
href(2,807,843).
href(2,808,844).
href(2,809,845).
href(2,810,846).
href(2,811,847).
href(2,812,848).
href(2,813,849).
href(2,814,850).
href(2,815,851).
href(2,816,852).
href(2,817,853).
href(2,818,854).
href(2,819,855).
href(2,820,856).
href(2,821,857).
href(2,822,858).
href(2,823,859).
href(2,824,860).
href(2,825,861).
href(2,826,862).
href(2,827,863).
href(2,828,864).
href(2,829,865).
href(2,830,866).
href(2,831,867).
href(2,832,868).
href(2,833,869).
href(2,834,870).
href(2,835,871).
href(2,836,872).
href(2,837,873).
href(2,838,874).
href(2,839,875).
href(2,840,876).
href(2,841,877).
href(2,842,878).
href(2,843,879).
href(2,844,880).
href(2,845,881).
href(2,846,882).
href(2,847,883).
href(2,848,884).
href(2,849,885).
href(2,850,886).
href(2,851,887).
href(2,852,888).
href(2,853,889).
href(2,854,890).
href(2,855,891).
href(2,856,892).
href(2,857,893).
href(2,858,894).
href(2,859,895).
href(2,860,896).
href(2,861,897).
href(2,862,898).
href(2,863,899).
href(2,864,900).
href(2,865,901).
href(2,866,902).
href(2,867,903).
href(2,868,904).
href(2,869,905).
href(2,870,906).
href(1,37,907).
href(2,871,908).
href(2,872,909).
href(2,873,910).
href(2,874,911).
href(2,875,912).
href(2,876,913).
href(2,877,914).
href(2,878,915).
href(2,879,916).
href(2,880,917).
href(2,881,918).
href(2,882,919).
href(2,883,920).
href(2,884,921).
href(2,885,922).
href(2,886,923).
href(2,887,924).
href(2,888,925).
href(2,889,926).
href(2,890,927).
href(2,891,928).
href(2,892,929).
href(2,893,930).
href(2,894,931).
href(2,895,932).
href(2,896,933).
href(2,897,934).
href(2,898,935).
href(1,38,936).
href(2,899,937).
href(2,900,938).
href(2,901,939).
href(2,902,940).
href(2,903,941).
href(2,904,942).
href(2,905,943).
href(2,906,944).
href(2,907,945).
href(2,908,946).
href(2,909,947).
href(2,910,948).
href(2,911,949).
href(2,912,950).
href(2,913,951).
href(2,914,952).
href(2,915,953).
href(2,916,954).
href(2,917,955).
href(2,918,956).
href(2,919,957).
href(2,920,958).
href(2,921,959).
href(2,922,960).
href(2,923,961).
href(2,924,962).
href(2,925,963).
href(2,926,964).
href(2,927,965).
href(2,928,966).
href(2,929,967).
href(2,930,968).
href(2,931,969).
href(2,932,970).
href(2,933,971).
href(2,934,972).
href(1,39,973).
href(2,935,974).
href(2,936,975).
href(2,937,976).
href(2,938,977).
href(2,939,978).
href(2,940,979).
href(1,40,980).
href(2,941,981).
href(1,41,982).
href(2,942,983).
href(2,943,984).
href(1,42,985).
href(2,944,986).
href(1,43,987).
href(1,44,988).
href(2,945,989).
href(2,946,990).
href(1,45,991).
href(2,947,992).
href(1,46,993).
href(2,948,994).
href(2,949,995).
href(1,47,996).
href(1,48,997).
href(2,950,998).
href(2,951,999).
href(2,952,1000).
href(2,953,1001).
href(2,954,1002).
href(2,955,1003).
href(2,956,1004).
href(2,957,1005).
href(2,958,1006).
href(2,959,1007).
href(2,960,1008).
href(2,961,1009).
href(2,962,1010).
href(2,963,1011).
href(2,964,1012).
href(2,965,1013).
href(2,966,1014).
href(2,967,1015).
href(2,968,1016).
href(2,969,1017).
href(2,970,1018).
href(1,49,1019).
href(2,971,1020).
href(2,972,1021).
href(1,50,1022).
href(2,973,1023).
href(1,51,1024).
href(2,974,1025).
href(2,975,1026).
href(2,976,1027).
href(2,977,1028).
href(2,978,1029).
href(2,979,1030).
href(2,980,1031).
href(2,981,1032).
href(1,52,1033).
href(2,982,1034).
href(2,983,1035).
href(2,984,1036).
href(2,985,1037).
href(2,986,1038).
href(2,987,1039).
href(2,988,1040).
href(2,989,1041).
href(2,990,1042).
href(2,991,1043).
href(2,992,1044).
href(1,53,1045).
href(2,993,1046).
href(2,994,1047).
href(2,995,1048).
href(2,996,1049).
href(1,54,1050).
href(2,997,1051).
href(2,998,1052).
href(2,999,1053).
href(2,1000,1054).
href(2,1001,1055).
href(2,1002,1056).
href(2,1003,1057).
href(2,1004,1058).
href(2,1005,1059).
href(2,1006,1060).
href(2,1007,1061).
href(2,1008,1062).
href(2,1009,1063).
href(2,1010,1064).
href(2,1011,1065).
href(2,1012,1066).
href(2,1013,1067).
href(1,55,1068).
href(1,56,1069).
href(2,1014,1070).
href(2,1015,1071).
href(2,1016,1072).
href(2,1017,1073).
href(2,1018,1074).
href(2,1019,1075).
href(2,1020,1076).
href(2,1021,1077).
href(2,1022,1078).
href(2,1023,1079).
href(2,1024,1080).
href(2,1025,1081).
href(1,57,1082).
href(2,1026,1083).
href(2,1027,1084).
href(2,1028,1085).
href(2,1029,1086).
href(2,1030,1087).
href(2,1031,1088).
href(2,1032,1089).
href(2,1033,1090).
href(2,1034,1091).
href(2,1035,1092).
href(2,1036,1093).
href(2,1037,1094).
href(2,1038,1095).
href(2,1039,1096).
href(2,1040,1097).
href(2,1041,1098).
href(2,1042,1099).
href(2,1043,1100).
href(2,1044,1101).
href(2,1045,1102).
href(2,1046,1103).
href(2,1047,1104).
href(2,1048,1105).
href(2,1049,1106).
href(2,1050,1107).
href(2,1051,1108).
href(2,1052,1109).
href(2,1053,1110).
href(2,1054,1111).
href(2,1055,1112).
href(2,1056,1113).
href(2,1057,1114).
href(2,1058,1115).
href(2,1059,1116).
href(2,1060,1117).
href(2,1061,1118).
href(2,1062,1119).
href(2,1063,1120).
href(2,1064,1121).
href(2,1065,1122).
href(2,1066,1123).
href(2,1067,1124).
href(2,1068,1125).
href(2,1069,1126).
href(2,1070,1127).
href(2,1071,1128).
href(2,1072,1129).
href(2,1073,1130).
href(2,1074,1131).
href(2,1075,1132).
href(2,1076,1133).
href(2,1077,1134).
href(2,1078,1135).
href(2,1079,1136).
href(2,1080,1137).
href(2,1081,1138).
href(2,1082,1139).
href(2,1083,1140).
href(2,1084,1141).
href(2,1085,1142).
href(2,1086,1143).
href(2,1087,1144).
href(2,1088,1145).
href(2,1089,1146).
href(2,1090,1147).
href(2,1091,1148).
href(2,1092,1149).
href(2,1093,1150).
href(2,1094,1151).
href(2,1095,1152).
href(2,1096,1153).
href(2,1097,1154).
href(2,1098,1155).
href(2,1099,1156).
href(2,1100,1157).
href(2,1101,1158).
href(1,58,1159).
href(2,1102,1160).
href(1,59,1161).
href(1,60,1162).
href(2,1103,1163).
href(1,61,1164).
href(1,62,1165).
href(2,1104,1166).
href(1,63,1167).
href(1,64,1168).
href(1,65,1169).
href(1,66,1170).
href(1,67,1171).
href(2,1105,1172).
href(2,1106,1173).
href(2,1107,1174).
href(2,1108,1175).
href(2,1109,1176).
href(2,1110,1177).
href(2,1111,1178).
href(2,1112,1179).
href(2,1113,1180).
href(2,1114,1181).
href(2,1115,1182).
href(2,1116,1183).
href(2,1117,1184).
href(2,1118,1185).
href(2,1119,1186).
href(2,1120,1187).
href(2,1121,1188).
href(2,1122,1189).
href(2,1123,1190).
href(2,1124,1191).
href(2,1125,1192).
href(2,1126,1193).
href(2,1127,1194).
href(2,1128,1195).
href(1,68,1196).
href(2,1129,1197).
href(1,69,1198).
href(2,1130,1199).
href(2,1131,1200).
href(2,1132,1201).
href(2,1133,1202).
href(2,1134,1203).
href(2,1135,1204).
href(2,1136,1205).
href(2,1137,1206).
href(2,1138,1207).
href(2,1139,1208).
/* (1208 entrees) */


/**
 * Table mail
 */

mail('mailto:java-io@java.sun.com',1).
mail('mailto:java-awt@java.sun.com',2).
mail('mailto:java-beans@java.sun.com',3).
mail('mailto:jdbc-odbc@sun.com',4).
mail('mailto:jdbc@sun.com',5).
mail('mailto:jls@java.sun.com',6).
mail('mailto:jdk-comments@java.sun.com',7).
mail('mailto:java2d-comments@sun.com',8).
mail('mailto:jni@java.sun.com',9).
mail('mailto:java-security@sun.com',10).
mail('mailto:swing-feedback@java.sun.com',11).
mail('mailto:collections-comments@java.sun.com',12).
mail('mailto:java-intl@java.sun.com',13).
mail('mailto:f.allimant@confluent.fr',14).
mail('mailto:j2se-comments@sun.com',15).
mail('mailto:javasound-comments@sun.com',16).
mail('mailto:java-debugger@sun.com',17).
mail('mailto:rmi-comments@java.sun.com',18).
mail('mailto:java-print@sun.com',19).
mail('mailto:j2se-comments@eng.sun.com',20).
mail('mailto:jre-comments@java.sun.com',21).
mail('mailto:jndi@java.sun.com',22).
mail('mailto:pelegri@eng.sun.com',23).
mail('mailto:reflection-comments@sun.com',24).
mail('mailto:java-gc-comments@eng.sun.com',25).
mail('mailto:java-dnd@java.sun.com',26).
mail('mailto:jsr015-comments@sun.com',27).
mail('mailto:jvmpi@eng.sun.com',28).
mail('mailto:rosanna.lee@sun.com',29).
mail('mailto:copyrights@sun.com',30).
mail('mailto:rmi-iiop@sun.com',31).
mail('mailto:sdk-docl10nbugs@java.sun.com',32).
mail('mailto:javaidl@sun.com',33).
mail('mailto:access@sun.com',34).
/* (34 entrees) */


/**
 * Table app_mail
 */

app_mail(2,1,753,1).
app_mail(2,1,309,1).
app_mail(2,1,1310,1).
app_mail(2,1,736,1).
app_mail(3,2,739,1).
app_mail(3,2,309,1).
app_mail(3,2,1310,1).
app_mail(3,2,736,1).
app_mail(7,3,309,1).
app_mail(7,3,1310,1).
app_mail(7,3,736,1).
app_mail(7,3,741,1).
app_mail(10,4,309,1).
app_mail(10,4,1310,1).
app_mail(10,4,767,1).
app_mail(10,5,309,1).
app_mail(10,5,1310,1).
app_mail(10,5,766,1).
app_mail(11,6,309,1).
app_mail(11,6,1310,1).
app_mail(11,6,736,1).
app_mail(11,6,775,1).
app_mail(14,7,309,1).
app_mail(14,7,1310,1).
app_mail(14,7,736,1).
app_mail(14,7,770,1).
app_mail(12,8,309,1).
app_mail(12,8,1310,1).
app_mail(12,8,738,1).
app_mail(16,9,309,1).
app_mail(16,9,1310,1).
app_mail(16,9,736,1).
app_mail(16,9,779,1).
app_mail(13,10,758,1).
app_mail(13,10,309,1).
app_mail(13,10,1310,1).
app_mail(17,11,309,1).
app_mail(17,11,1310,1).
app_mail(17,11,736,1).
app_mail(17,11,1324,1).
app_mail(23,12,304,1).
app_mail(23,12,309,1).
app_mail(23,12,1310,1).
app_mail(22,13,309,1).
app_mail(22,13,1310,1).
app_mail(23,12,736,1).
app_mail(22,13,736,1).
app_mail(22,13,752,1).
app_mail(18,14,578,1).
app_mail(18,14,135,1).
app_mail(18,14,328,1).
app_mail(27,15,309,1).
app_mail(27,15,1310,1).
app_mail(27,15,729,1).
app_mail(28,16,760,1).
app_mail(28,16,309,1).
app_mail(28,16,1310,1).
app_mail(30,17,309,1).
app_mail(30,17,745,1).
app_mail(30,17,1310,1).
app_mail(34,18,309,1).
app_mail(34,18,1310,1).
app_mail(34,18,736,1).
app_mail(34,18,1166,1).
app_mail(41,19,309,1).
app_mail(41,19,1310,1).
app_mail(41,19,756,1).
app_mail(45,20,309,1).
app_mail(45,20,1310,1).
app_mail(45,20,496,1).
app_mail(45,20,729,1).
app_mail(43,21,309,1).
app_mail(43,21,1310,1).
app_mail(43,21,736,1).
app_mail(43,21,786,1).
app_mail(44,13,309,1).
app_mail(44,13,1310,1).
app_mail(44,13,736,1).
app_mail(44,13,752,1).
app_mail(54,22,309,1).
app_mail(54,22,1310,1).
app_mail(54,22,736,1).
app_mail(54,22,777,1).
app_mail(57,23,309,1).
app_mail(57,23,1310,1).
app_mail(57,23,1031,1).
app_mail(57,23,496,1).
app_mail(47,24,309,1).
app_mail(47,24,1128,1).
app_mail(47,24,1310,1).
app_mail(47,24,496,1).
app_mail(63,25,309,1).
app_mail(63,25,1310,1).
app_mail(63,25,496,1).
app_mail(63,25,750,1).
app_mail(64,7,309,1).
app_mail(64,7,1310,1).
app_mail(64,7,736,1).
app_mail(64,7,770,1).
app_mail(59,20,309,1).
app_mail(59,20,1310,1).
app_mail(59,20,496,1).
app_mail(59,20,729,1).
app_mail(66,26,309,1).
app_mail(66,26,1310,1).
app_mail(66,26,736,1).
app_mail(66,26,746,1).
app_mail(68,15,309,1).
app_mail(68,15,1310,1).
app_mail(68,15,729,1).
app_mail(65,27,309,1).
app_mail(65,27,1310,1).
app_mail(65,27,789,1).
app_mail(71,28,309,1).
app_mail(71,28,1310,1).
app_mail(71,28,794,1).
app_mail(71,28,496,1).
app_mail(55,29,309,1).
app_mail(55,29,1310,1).
app_mail(55,29,1171,1).
app_mail(55,29,822,1).
app_mail(74,30,309,1).
app_mail(74,30,1310,1).
app_mail(74,30,366,1).
app_mail(73,1,753,1).
app_mail(73,1,309,1).
app_mail(73,1,1310,1).
app_mail(73,1,736,1).
app_mail(69,31,309,2).
app_mail(69,31,1168,2).
app_mail(69,31,1310,2).
app_mail(79,32,309,1).
app_mail(79,32,1190,1).
app_mail(79,32,1310,1).
app_mail(79,32,736,1).
app_mail(78,33,309,1).
app_mail(78,33,1310,1).
app_mail(78,33,751,1).
app_mail(78,31,309,1).
app_mail(78,31,1168,1).
app_mail(78,31,1310,1).
app_mail(82,34,118,1).
app_mail(82,34,309,1).
app_mail(82,34,1310,1).
/* (144 entrees) */


/**
 * Table app_img
 */

app_img(87,1,452,1).
app_img(87,1,444,1).
app_img(87,1,41,1).
app_img(87,1,9,2).
app_img(87,1,728,1).
app_img(87,2,736,1).
app_img(87,3,1310,1).
app_img(2,2,736,1).
app_img(2,3,1310,1).
app_img(3,2,736,1).
app_img(3,3,899,1).
app_img(3,3,1310,1).
app_img(3,3,668,1).
app_img(6,2,736,1).
app_img(6,3,1310,1).
app_img(7,2,736,1).
app_img(7,3,1310,1).
app_img(1,2,736,1).
app_img(1,3,1310,1).
app_img(9,2,736,1).
app_img(9,3,1310,1).
app_img(4,4,1337,1).
app_img(4,4,736,1).
app_img(4,4,1014,1).
app_img(10,3,1310,1).
app_img(4,4,634,1).
app_img(4,5,899,1).
app_img(4,5,1310,1).
app_img(4,5,668,1).
app_img(4,6,732,1).
app_img(4,7,1014,1).
app_img(4,7,1074,1).
app_img(4,8,1337,1).
app_img(4,8,736,1).
app_img(4,8,1014,1).
app_img(4,8,634,1).
app_img(14,2,736,1).
app_img(14,3,1310,1).
app_img(12,2,736,1).
app_img(12,3,899,1).
app_img(12,3,1310,1).
app_img(12,3,668,1).
app_img(16,2,736,1).
app_img(16,3,1310,1).
app_img(15,9,1461,1).
app_img(15,9,869,1).
app_img(15,9,1440,1).
app_img(15,9,1346,1).
app_img(15,9,934,1).
app_img(15,10,899,1).
app_img(15,10,1310,1).
app_img(15,10,857,1).
app_img(15,11,930,1).
app_img(15,11,886,1).
app_img(15,11,1310,1).
app_img(15,11,2,1).
app_img(15,11,815,1).
app_img(15,12,2,1).
app_img(15,12,1337,1).
app_img(15,12,1276,1).
app_img(15,12,926,1).
app_img(15,12,631,1).
app_img(15,12,1193,1).
app_img(15,12,309,1).
app_img(15,12,736,1).
app_img(15,12,535,1).
app_img(15,12,1107,1).
app_img(15,12,1346,1).
app_img(15,12,667,1).
app_img(15,12,500,1).
app_img(15,12,1113,1).
app_img(15,12,1477,1).
app_img(15,12,460,1).
app_img(15,13,1041,1).
app_img(15,13,386,1).
app_img(15,13,1310,1).
app_img(15,13,554,1).
app_img(15,13,941,1).
app_img(15,13,479,1).
app_img(15,13,227,1).
app_img(15,13,2,1).
app_img(15,14,597,1).
app_img(15,14,736,1).
app_img(15,15,1452,1).
app_img(15,15,956,1).
app_img(15,16,835,1).
app_img(15,16,1192,2).
app_img(15,16,357,1).
app_img(15,16,1240,2).
app_img(15,16,1303,1).
app_img(15,16,879,1).
app_img(15,16,1360,1).
app_img(8,2,736,1).
app_img(8,3,899,1).
app_img(8,3,1310,1).
app_img(8,3,668,1).
app_img(13,2,736,1).
app_img(13,3,1310,1).
app_img(17,2,736,1).
app_img(17,3,899,1).
app_img(17,3,1310,1).
app_img(17,3,668,1).
app_img(23,2,736,1).
app_img(23,3,1310,1).
app_img(22,2,736,1).
app_img(22,3,1310,1).
app_img(25,4,1337,1).
app_img(25,4,736,1).
app_img(25,4,1014,1).
app_img(25,4,634,1).
app_img(25,5,899,1).
app_img(25,5,1310,1).
app_img(25,5,668,1).
app_img(25,7,1014,1).
app_img(25,7,1074,1).
app_img(25,8,1337,1).
app_img(25,8,736,1).
app_img(25,8,1014,1).
app_img(25,8,634,1).
app_img(18,4,1337,1).
app_img(18,4,736,1).
app_img(18,4,1014,1).
app_img(18,4,634,1).
app_img(18,5,899,1).
app_img(18,5,1310,1).
app_img(18,5,668,1).
app_img(18,7,1014,1).
app_img(18,7,1074,1).
app_img(18,8,1337,1).
app_img(18,8,736,1).
app_img(18,8,1014,1).
app_img(18,8,634,1).
app_img(20,2,736,1).
app_img(20,3,1310,1).
app_img(27,2,736,1).
app_img(27,3,899,1).
app_img(27,3,1310,1).
app_img(27,3,668,1).
app_img(28,2,736,1).
app_img(28,3,1310,1).
app_img(30,2,736,1).
app_img(30,3,1310,1).
app_img(32,2,736,1).
app_img(32,3,1310,1).
app_img(33,2,736,1).
app_img(33,3,899,1).
app_img(33,3,1310,1).
app_img(33,3,668,1).
app_img(26,4,1337,1).
app_img(26,4,736,1).
app_img(26,4,1014,1).
app_img(26,4,634,1).
app_img(26,8,1337,1).
app_img(26,8,736,1).
app_img(26,8,1014,1).
app_img(26,8,634,1).
app_img(26,17,1117,2).
app_img(26,17,695,3).
app_img(26,17,134,2).
app_img(26,17,91,1).
app_img(26,17,784,4).
app_img(26,17,829,4).
app_img(26,17,1189,5).
app_img(26,17,836,1).
app_img(26,17,728,2).
app_img(26,17,1429,14).
app_img(26,17,959,3).
app_img(26,17,564,3).
app_img(26,17,694,1).
app_img(26,17,1138,4).
app_img(26,17,420,1).
app_img(26,17,690,4).
app_img(26,17,41,2).
app_img(26,17,1456,1).
app_img(26,17,9,4).
app_img(26,17,1044,2).
app_img(26,19,452,2).
app_img(26,19,1248,2).
app_img(26,19,1024,2).
app_img(26,19,698,1).
app_img(26,19,1258,1).
app_img(26,7,1014,1).
app_img(26,7,1074,1).
app_img(26,5,899,1).
app_img(26,5,1310,1).
app_img(26,5,668,1).
app_img(26,20,736,1).
app_img(26,20,1274,1).
app_img(26,20,480,1).
app_img(26,20,1424,1).
app_img(26,20,41,1).
app_img(26,20,1195,1).
app_img(26,20,26,1).
app_img(26,20,9,1).
app_img(26,20,12,1).
app_img(26,21,732,1).
app_img(45,2,736,1).
app_img(45,3,1310,1).
app_img(47,2,736,1).
app_img(47,3,1310,1).
app_img(63,2,736,1).
app_img(63,3,1310,1).
app_img(64,2,736,1).
app_img(64,3,1310,1).
app_img(65,2,736,1).
app_img(65,3,899,1).
app_img(65,3,1310,1).
app_img(65,3,668,1).
app_img(72,52,1116,1).
app_img(72,52,597,1).
app_img(72,52,122,1).
app_img(74,4,1337,1).
app_img(74,4,736,1).
app_img(74,4,1014,1).
app_img(74,4,634,1).
app_img(74,5,899,1).
app_img(74,5,1310,1).
app_img(74,5,668,1).
app_img(74,7,1014,1).
app_img(74,7,1074,1).
app_img(74,8,1337,1).
app_img(74,8,736,1).
app_img(74,8,1014,1).
app_img(74,8,634,1).
app_img(76,53,1385,1).
app_img(76,53,1334,1).
app_img(76,53,1372,1).
app_img(76,53,172,1).
app_img(76,53,442,1).
app_img(76,54,331,1).
app_img(76,54,420,1).
app_img(76,54,736,1).
app_img(76,55,444,1).
app_img(76,55,146,1).
app_img(76,55,454,1).
app_img(76,56,672,1).
app_img(76,56,198,1).
app_img(76,57,976,1).
app_img(76,57,1318,1).
app_img(76,58,1015,1).
app_img(76,58,629,1).
app_img(76,59,316,1).
app_img(76,59,432,1).
app_img(76,60,1337,1).
app_img(76,60,736,1).
app_img(76,60,1346,1).
app_img(76,60,564,1).
app_img(76,60,1256,1).
app_img(76,61,1337,1).
app_img(76,61,736,1).
app_img(76,61,1014,1).
app_img(76,61,634,1).
app_img(76,62,1465,1).
app_img(76,62,1337,1).
app_img(76,62,736,1).
app_img(76,62,1404,1).
app_img(76,62,641,1).
app_img(76,62,721,1).
app_img(76,63,513,1).
app_img(76,63,944,1).
app_img(76,63,515,1).
app_img(76,63,586,1).
app_img(76,64,899,1).
app_img(76,64,1310,1).
app_img(76,64,668,1).
app_img(80,4,1337,1).
app_img(80,4,736,1).
app_img(80,4,1014,1).
app_img(80,4,634,1).
app_img(80,5,899,1).
app_img(80,5,1310,1).
app_img(80,5,668,1).
app_img(80,6,732,1).
app_img(80,7,1014,1).
app_img(80,7,1074,1).
app_img(80,8,1337,1).
app_img(80,8,736,1).
app_img(80,8,1014,1).
app_img(80,8,634,1).
app_img(83,2,736,1).
app_img(83,3,1310,1).
app_img(34,18,736,1).
app_img(34,3,1310,1).
app_img(35,2,736,1).
app_img(35,3,1310,1).
app_img(44,2,736,1).
app_img(44,3,1310,1).
app_img(50,4,1337,1).
app_img(50,4,736,1).
app_img(50,4,1014,1).
app_img(50,4,634,1).
app_img(50,43,1228,1).
app_img(50,43,464,1).
app_img(50,8,1337,1).
app_img(50,8,736,1).
app_img(50,8,1014,1).
app_img(50,8,634,1).
app_img(50,44,1348,1).
app_img(50,44,736,1).
app_img(50,44,1346,1).
app_img(50,44,480,1).
app_img(50,44,1384,1).
app_img(50,45,736,1).
app_img(50,45,1346,1).
app_img(50,45,359,1).
app_img(50,45,1384,1).
app_img(50,46,1358,1).
app_img(50,46,736,1).
app_img(50,46,1346,1).
app_img(50,46,1384,1).
app_img(50,7,1014,1).
app_img(50,7,1074,1).
app_img(50,5,899,1).
app_img(50,5,1310,1).
app_img(50,5,668,1).
app_img(50,47,452,1).
app_img(50,48,1346,1).
app_img(50,48,1323,1).
app_img(50,48,1384,1).
app_img(50,48,772,1).
app_img(54,2,736,1).
app_img(54,3,899,1).
app_img(54,3,1310,1).
app_img(54,3,668,1).
app_img(56,4,1337,1).
app_img(56,4,736,1).
app_img(56,4,1014,1).
app_img(56,4,634,1).
app_img(56,5,899,1).
app_img(56,5,1310,1).
app_img(56,5,668,1).
app_img(56,7,1014,1).
app_img(56,7,1074,1).
app_img(56,51,494,1).
app_img(56,8,1337,1).
app_img(56,8,736,1).
app_img(56,8,1014,1).
app_img(56,8,634,1).
app_img(70,2,736,1).
app_img(70,3,1310,2).
app_img(70,65,972,1).
app_img(21,4,1337,1).
app_img(21,4,736,1).
app_img(21,4,1014,1).
app_img(21,4,634,1).
app_img(21,5,899,1).
app_img(21,5,1310,1).
app_img(21,5,668,1).
app_img(21,6,732,1).
app_img(21,7,1014,1).
app_img(21,7,1074,1).
app_img(21,8,1337,1).
app_img(21,8,736,1).
app_img(21,8,1014,1).
app_img(21,8,634,1).
app_img(38,42,247,10).
app_img(38,42,402,10).
app_img(38,4,1337,1).
app_img(38,4,736,1).
app_img(38,4,1014,1).
app_img(38,4,634,1).
app_img(38,5,899,1).
app_img(38,5,1310,1).
app_img(38,5,668,1).
app_img(38,7,1014,1).
app_img(38,7,1074,1).
app_img(38,8,1337,1).
app_img(38,8,736,1).
app_img(38,8,1014,1).
app_img(38,8,634,1).
app_img(1068,2,736,1).
app_img(1068,3,1310,1).
app_img(52,28,1337,1).
app_img(52,28,736,1).
app_img(52,28,1014,1).
app_img(52,28,634,1).
app_img(52,34,899,1).
app_img(52,34,1310,1).
app_img(52,34,668,1).
app_img(52,33,1014,1).
app_img(52,33,1074,1).
app_img(52,29,1337,1).
app_img(52,29,736,1).
app_img(52,29,1014,1).
app_img(52,29,634,1).
app_img(57,2,736,1).
app_img(57,3,899,1).
app_img(57,3,1310,1).
app_img(57,3,668,1).
app_img(59,2,736,1).
app_img(59,3,1310,1).
app_img(67,4,1337,1).
app_img(67,4,736,1).
app_img(67,4,1014,1).
app_img(67,4,634,1).
app_img(67,5,899,1).
app_img(67,5,1310,1).
app_img(67,5,668,1).
app_img(67,6,732,1).
app_img(67,7,1014,1).
app_img(67,7,1074,1).
app_img(67,8,1337,1).
app_img(67,8,736,1).
app_img(67,8,1014,1).
app_img(67,8,634,1).
app_img(71,2,1358,1).
app_img(71,2,736,1).
app_img(71,3,1310,1).
app_img(73,2,736,1).
app_img(73,3,1310,1).
app_img(78,53,1385,1).
app_img(78,53,1334,1).
app_img(78,53,1372,1).
app_img(78,53,172,1).
app_img(78,53,442,1).
app_img(78,54,331,1).
app_img(78,54,420,1).
app_img(78,54,736,1).
app_img(78,3,1310,2).
app_img(78,2,736,1).
app_img(78,55,444,1).
app_img(78,55,146,1).
app_img(78,55,454,1).
app_img(78,65,972,1).
app_img(78,57,976,1).
app_img(78,57,1318,1).
app_img(78,59,316,1).
app_img(78,59,432,1).
app_img(78,62,1465,1).
app_img(78,62,1337,1).
app_img(78,62,736,1).
app_img(78,62,1404,1).
app_img(78,62,641,1).
app_img(78,62,721,1).
app_img(78,63,513,1).
app_img(78,63,944,1).
app_img(78,63,515,1).
app_img(78,63,586,1).
app_img(29,4,1337,1).
app_img(29,4,736,1).
app_img(29,4,1014,1).
app_img(29,4,634,1).
app_img(29,5,899,1).
app_img(29,5,1310,1).
app_img(29,5,668,1).
app_img(29,7,1014,1).
app_img(29,7,1074,1).
app_img(29,8,1337,1).
app_img(29,8,736,1).
app_img(29,8,1014,1).
app_img(29,8,634,1).
app_img(29,22,1156,5).
app_img(29,22,1365,5).
app_img(29,22,1360,5).
app_img(39,23,1089,1).
app_img(39,24,941,1).
app_img(39,25,37,1).
app_img(39,25,1286,1).
app_img(39,26,26,1).
app_img(39,26,1286,1).
app_img(39,27,9,1).
app_img(39,27,1286,1).
app_img(39,28,1337,1).
app_img(39,28,736,1).
app_img(39,28,1014,1).
app_img(39,28,634,1).
app_img(39,29,1337,1).
app_img(39,29,736,1).
app_img(39,29,1014,1).
app_img(39,29,634,1).
app_img(39,31,821,1).
app_img(39,31,1026,1).
app_img(39,32,827,1).
app_img(39,33,1014,1).
app_img(39,33,1074,1).
app_img(39,34,899,1).
app_img(39,34,1310,1).
app_img(39,34,668,1).
app_img(39,36,465,1).
app_img(39,37,41,1).
app_img(39,37,1286,1).
app_img(40,4,1337,1).
app_img(40,4,736,1).
app_img(40,4,1014,1).
app_img(40,4,634,1).
app_img(40,5,899,1).
app_img(40,5,1310,1).
app_img(40,5,668,1).
app_img(40,7,1014,1).
app_img(40,7,1074,1).
app_img(40,8,1337,1).
app_img(40,8,736,1).
app_img(40,8,1014,1).
app_img(40,8,634,1).
app_img(46,2,736,1).
app_img(46,3,1310,1).
app_img(53,2,736,1).
app_img(53,3,1310,1).
app_img(58,2,736,1).
app_img(58,3,1310,1).
app_img(61,49,899,1).
app_img(61,49,1310,1).
app_img(61,49,668,1).
app_img(61,50,736,1).
app_img(62,4,1337,1).
app_img(62,4,736,1).
app_img(62,4,1014,1).
app_img(62,4,634,1).
app_img(62,5,899,1).
app_img(62,5,1310,1).
app_img(62,5,668,1).
app_img(62,6,732,1).
app_img(62,7,1014,1).
app_img(62,7,1074,1).
app_img(62,8,1337,1).
app_img(62,8,736,1).
app_img(62,8,1014,1).
app_img(62,8,634,1).
app_img(66,2,736,1).
app_img(66,3,899,1).
app_img(66,3,1310,1).
app_img(66,3,668,1).
app_img(68,2,736,1).
app_img(68,3,899,1).
app_img(68,3,1310,1).
app_img(68,3,668,1).
app_img(69,2,736,1).
app_img(69,3,1310,2).
app_img(69,65,972,1).
app_img(79,4,1337,1).
app_img(79,4,736,1).
app_img(79,4,1014,1).
app_img(79,4,634,1).
app_img(79,5,899,1).
app_img(79,5,1310,1).
app_img(79,5,668,1).
app_img(79,7,1014,1).
app_img(79,7,1074,1).
app_img(79,8,1337,1).
app_img(79,8,736,1).
app_img(79,8,1014,1).
app_img(79,8,634,1).
app_img(24,30,158,1).
app_img(24,30,1131,1).
app_img(24,30,1184,1).
app_img(24,30,215,1).
app_img(24,30,38,1).
app_img(24,30,1360,1).
app_img(24,30,29,1).
app_img(24,35,728,1).
app_img(24,4,1337,1).
app_img(24,4,736,1).
app_img(24,4,1014,1).
app_img(24,4,634,1).
app_img(24,8,1337,1).
app_img(24,8,736,1).
app_img(24,8,1014,1).
app_img(24,8,634,1).
app_img(24,38,3,1).
app_img(24,38,249,1).
app_img(24,38,736,1).
app_img(24,38,1346,1).
app_img(24,38,635,1).
app_img(24,39,725,1).
app_img(24,7,1014,1).
app_img(24,7,1074,1).
app_img(24,5,899,1).
app_img(24,5,1310,1).
app_img(24,5,668,1).
app_img(24,40,667,1).
app_img(24,40,1243,1).
app_img(24,40,26,1).
app_img(24,40,145,1).
app_img(24,40,1346,1).
app_img(24,40,8,1).
app_img(24,40,795,1).
app_img(24,41,726,1).
app_img(41,2,736,1).
app_img(41,3,899,1).
app_img(41,3,1310,1).
app_img(41,3,668,1).
app_img(42,4,1337,1).
app_img(42,4,736,1).
app_img(42,4,1014,1).
app_img(42,4,634,1).
app_img(42,5,899,1).
app_img(42,5,1310,1).
app_img(42,5,668,1).
app_img(42,7,1014,1).
app_img(42,7,1074,1).
app_img(42,8,1337,1).
app_img(42,8,736,1).
app_img(42,8,1014,1).
app_img(42,8,634,1).
app_img(43,2,736,1).
app_img(43,3,1310,1).
app_img(51,4,1337,1).
app_img(51,4,736,1).
app_img(51,4,1014,1).
app_img(51,4,634,1).
app_img(51,5,899,1).
app_img(51,5,1310,1).
app_img(51,5,668,1).
app_img(51,6,732,1).
app_img(51,7,1014,1).
app_img(51,7,1074,1).
app_img(51,8,1337,1).
app_img(51,8,736,1).
app_img(51,8,1014,1).
app_img(51,8,634,1).
app_img(55,4,1337,1).
app_img(55,4,736,1).
app_img(55,4,1014,1).
app_img(55,4,634,1).
app_img(55,5,899,1).
app_img(55,5,1310,1).
app_img(55,5,668,1).
app_img(55,7,1014,1).
app_img(55,7,1074,1).
app_img(55,8,1337,1).
app_img(55,8,736,1).
app_img(55,8,1014,1).
app_img(55,8,634,1).
app_img(75,4,1337,1).
app_img(75,4,736,1).
app_img(75,4,1014,1).
app_img(75,4,634,1).
app_img(75,5,899,1).
app_img(75,5,1310,1).
app_img(75,5,668,1).
app_img(75,6,732,1).
app_img(75,7,1014,1).
app_img(75,7,1074,1).
app_img(75,8,1337,1).
app_img(75,8,736,1).
app_img(75,8,1014,1).
app_img(75,8,634,1).
app_img(81,53,1385,1).
app_img(81,53,1334,1).
app_img(81,53,1372,1).
app_img(81,53,172,1).
app_img(81,53,442,1).
app_img(81,54,331,1).
app_img(81,54,420,1).
app_img(81,54,736,1).
app_img(81,55,444,1).
app_img(81,55,146,1).
app_img(81,55,454,1).
app_img(81,56,672,1).
app_img(81,56,198,1).
app_img(81,57,976,1).
app_img(81,57,1318,1).
app_img(81,58,1015,1).
app_img(81,58,629,1).
app_img(81,59,316,1).
app_img(81,59,432,1).
app_img(81,60,1337,1).
app_img(81,60,736,1).
app_img(81,60,1346,1).
app_img(81,60,564,1).
app_img(81,60,1256,1).
app_img(81,61,1337,1).
app_img(81,61,736,1).
app_img(81,61,1014,1).
app_img(81,61,634,1).
app_img(81,62,1465,1).
app_img(81,62,1337,1).
app_img(81,62,736,1).
app_img(81,62,1404,1).
app_img(81,62,641,1).
app_img(81,62,721,1).
app_img(81,63,513,1).
app_img(81,63,944,1).
app_img(81,63,515,1).
app_img(81,63,586,1).
app_img(81,64,899,1).
app_img(81,64,1310,1).
app_img(81,64,668,1).
app_img(82,2,736,1).
app_img(82,3,1310,1).
/* (680 entrees) */


/**
 * Table app_href
 */

app_href(87,1,1473,1).
app_href(87,2,1207,1).
app_href(87,2,967,1).
app_href(87,3,1363,1).
app_href(87,3,1455,1).
app_href(87,3,117,1).
app_href(87,4,1139,1).
app_href(87,4,1424,1).
app_href(87,4,318,1).
app_href(87,4,1070,1).
app_href(87,4,1460,1).
app_href(87,5,408,1).
app_href(87,5,540,1).
app_href(87,5,595,1).
app_href(87,5,616,1).
app_href(87,5,677,1).
app_href(87,5,1385,1).
app_href(87,5,810,1).
app_href(87,5,145,1).
app_href(87,5,850,1).
app_href(87,5,1360,1).
app_href(87,5,1192,1).
app_href(87,6,854,1).
app_href(87,7,1358,1).
app_href(87,7,324,1).
app_href(87,7,145,1).
app_href(87,7,740,1).
app_href(87,8,540,2).
app_href(87,8,1041,1).
app_href(87,8,941,2).
app_href(87,8,1309,1).
app_href(87,8,1235,1).
app_href(87,8,37,1).
app_href(87,8,728,1).
app_href(87,8,9,1).
app_href(87,8,970,1).
app_href(87,9,444,1).
app_href(87,9,1362,2).
app_href(87,9,442,1).
app_href(87,10,1358,1).
app_href(87,10,736,1).
app_href(87,10,332,1).
app_href(87,10,766,1).
app_href(87,10,393,1).
app_href(87,11,736,1).
app_href(87,11,1264,1).
app_href(87,11,1346,1).
app_href(87,11,810,1).
app_href(87,12,611,1).
app_href(87,12,660,1).
app_href(87,12,140,1).
app_href(87,12,36,1).
app_href(87,13,1198,1).
app_href(87,14,407,1).
app_href(87,14,152,1).
app_href(87,14,140,1).
app_href(87,14,156,1).
app_href(87,15,899,1).
app_href(87,15,1310,1).
app_href(87,15,668,1).
app_href(87,16,701,1).
app_href(87,16,736,1).
app_href(87,16,927,1).
app_href(87,16,779,1).
app_href(87,17,325,1).
app_href(87,17,1323,1).
app_href(87,17,1095,1).
app_href(87,18,667,1).
app_href(87,18,444,1).
app_href(87,18,736,1).
app_href(87,18,1457,1).
app_href(87,18,568,1).
app_href(87,19,556,1).
app_href(87,19,244,1).
app_href(87,20,1443,1).
app_href(87,20,1358,1).
app_href(87,20,736,1).
app_href(87,20,1276,1).
app_href(87,21,1041,1).
app_href(87,21,140,1).
app_href(87,21,667,1).
app_href(87,21,41,1).
app_href(87,21,277,1).
app_href(87,21,728,1).
app_href(87,21,9,2).
app_href(87,21,1358,1).
app_href(87,21,503,1).
app_href(87,22,581,1).
app_href(87,22,686,1).
app_href(87,22,894,1).
app_href(87,23,581,1).
app_href(87,23,303,1).
app_href(87,24,736,1).
app_href(87,24,1246,1).
app_href(87,24,1014,1).
app_href(87,24,634,1).
app_href(87,25,362,1).
app_href(87,25,736,1).
app_href(87,25,1346,1).
app_href(87,25,296,1).
app_href(87,25,1093,1).
app_href(87,25,564,1).
app_href(87,25,810,1).
app_href(87,26,1349,2).
app_href(87,26,452,3).
app_href(87,26,444,2).
app_href(87,26,26,1).
app_href(87,26,736,1).
app_href(87,26,1189,2).
app_href(87,26,1014,1).
app_href(87,27,1275,1).
app_href(87,27,1007,1).
app_href(87,27,883,1).
app_href(87,27,495,1).
app_href(87,28,1255,1).
app_href(87,29,536,1).
app_href(87,29,26,1).
app_href(87,29,736,2).
app_href(87,29,1346,1).
app_href(87,29,1189,1).
app_href(87,29,140,1).
app_href(87,30,400,1).
app_href(87,30,1041,1).
app_href(87,30,160,1).
app_href(87,30,736,1).
app_href(87,30,782,1).
app_href(87,31,1041,1).
app_href(87,31,26,1).
app_href(87,31,736,1).
app_href(87,31,1264,1).
app_href(87,31,145,1).
app_href(87,32,736,2).
app_href(87,32,1041,1).
app_href(87,32,181,1).
app_href(87,32,1401,1).
app_href(87,32,1204,1).
app_href(87,32,140,1).
app_href(87,32,1307,1).
app_href(87,32,313,1).
app_href(87,32,354,1).
app_href(87,32,1246,2).
app_href(87,32,728,1).
app_href(87,32,1358,2).
app_href(87,32,355,1).
app_href(87,32,603,1).
app_href(87,33,939,1).
app_href(87,34,1141,1).
app_href(87,34,1162,1).
app_href(87,34,715,1).
app_href(87,34,894,1).
app_href(87,35,444,1).
app_href(87,35,365,2).
app_href(87,35,6,1).
app_href(87,35,1341,1).
app_href(87,35,829,1).
app_href(87,35,140,1).
app_href(87,35,564,1).
app_href(87,36,1144,1).
app_href(87,36,242,1).
app_href(87,36,1304,1).
app_href(87,37,580,1).
app_href(87,37,951,1).
app_href(87,38,736,1).
app_href(87,38,483,1).
app_href(87,39,266,1).
app_href(87,39,1358,1).
app_href(87,39,1093,1).
app_href(87,39,948,1).
app_href(87,40,452,1).
app_href(87,41,1217,1).
app_href(87,41,1073,1).
app_href(87,42,452,1).
app_href(87,43,1424,1).
app_href(87,43,1011,1).
app_href(87,43,645,1).
app_href(87,44,704,1).
app_href(87,45,864,1).
app_href(87,45,1430,1).
app_href(87,46,883,1).
app_href(87,46,530,1).
app_href(87,47,1127,1).
app_href(87,48,736,1).
app_href(87,48,1264,1).
app_href(87,48,1346,1).
app_href(87,48,864,1).
app_href(87,48,1430,1).
app_href(87,49,1409,1).
app_href(87,49,736,5).
app_href(87,49,9,2).
app_href(87,49,880,1).
app_href(87,49,1012,1).
app_href(87,49,934,1).
app_href(87,49,1343,1).
app_href(87,49,809,1).
app_href(87,50,736,1).
app_href(87,50,1346,1).
app_href(87,50,1384,1).
app_href(87,51,444,1).
app_href(87,51,277,1).
app_href(87,52,372,1).
app_href(87,52,736,1).
app_href(87,52,977,1).
app_href(87,52,1346,1).
app_href(87,52,1093,1).
app_href(87,52,564,1).
app_href(87,52,810,1).
app_href(87,53,1411,1).
app_href(87,53,1012,1).
app_href(87,53,140,1).
app_href(87,53,810,1).
app_href(87,54,1358,1).
app_href(87,54,429,1).
app_href(87,54,701,1).
app_href(87,54,736,1).
app_href(87,54,140,1).
app_href(87,54,777,1).
app_href(87,54,922,1).
app_href(87,55,826,1).
app_href(87,55,736,1).
app_href(87,55,1346,1).
app_href(87,55,288,1).
app_href(87,56,1157,1).
app_href(87,56,608,1).
app_href(87,56,731,1).
app_href(87,56,251,1).
app_href(87,57,1155,1).
app_href(87,58,1358,1).
app_href(87,58,736,1).
app_href(87,58,1047,1).
app_href(87,59,503,1).
app_href(87,59,1033,1).
app_href(87,60,26,1).
app_href(87,60,9,1).
app_href(87,60,1316,1).
app_href(87,61,1310,1).
app_href(87,61,115,1).
app_href(87,61,958,1).
app_href(87,61,1012,1).
app_href(87,62,1138,1).
app_href(87,62,959,1).
app_href(87,63,1123,1).
app_href(87,63,969,1).
app_href(87,64,902,1).
app_href(87,64,540,1).
app_href(87,65,657,1).
app_href(87,66,1374,1).
app_href(87,66,392,1).
app_href(87,66,458,1).
app_href(87,67,505,1).
app_href(87,67,1358,1).
app_href(87,67,1248,1).
app_href(87,67,959,1).
app_href(87,67,985,1).
app_href(87,67,564,1).
app_href(87,67,690,1).
app_href(87,68,880,1).
app_href(87,68,159,1).
app_href(87,69,1168,1).
app_href(87,70,367,1).
app_href(87,71,1090,1).
app_href(87,71,701,1).
app_href(87,71,736,1).
app_href(87,71,864,1).
app_href(87,71,794,1).
app_href(87,71,1430,1).
app_href(87,72,1030,1).
app_href(87,72,452,1).
app_href(87,72,140,1).
app_href(87,72,1105,1).
app_href(87,73,941,1).
app_href(87,74,444,1).
app_href(87,74,1053,1).
app_href(87,74,1121,1).
app_href(87,75,959,1).
app_href(87,75,836,1).
app_href(87,75,564,1).
app_href(87,75,690,1).
app_href(87,76,444,1).
app_href(87,76,976,1).
app_href(87,76,1192,1).
app_href(87,76,1346,1).
app_href(87,77,214,1).
app_href(87,77,151,1).
app_href(87,77,736,3).
app_href(87,77,9,2).
app_href(87,77,1012,1).
app_href(87,77,197,1).
app_href(87,78,649,1).
app_href(87,78,736,1).
app_href(87,79,1117,1).
app_href(87,80,959,1).
app_href(87,80,897,1).
app_href(87,80,1456,1).
app_href(87,80,564,1).
app_href(87,80,690,1).
app_href(87,81,1372,1).
app_href(87,81,736,1).
app_href(87,81,1346,1).
app_href(87,81,1093,1).
app_href(87,81,564,1).
app_href(87,81,810,1).
app_href(87,82,119,1).
app_href(87,83,161,1).
app_href(87,83,736,1).
app_href(87,83,550,1).
app_href(87,83,733,1).
app_href(3,35,365,1).
app_href(3,35,6,1).
app_href(3,82,145,1).
app_href(3,82,119,1).
app_href(3,84,1011,1).
app_href(3,84,736,1).
app_href(3,84,512,1).
app_href(3,84,197,1).
app_href(3,85,1070,1).
app_href(3,85,970,1).
app_href(3,85,1009,1).
app_href(3,85,503,1).
app_href(3,15,899,1).
app_href(3,15,1310,1).
app_href(3,15,668,1).
app_href(3,86,560,1).
app_href(3,86,905,1).
app_href(3,86,1264,1).
app_href(3,87,444,1).
app_href(3,87,358,1).
app_href(3,88,736,1).
app_href(3,88,1189,1).
app_href(3,88,197,1).
app_href(3,88,1346,1).
app_href(3,88,667,1).
app_href(3,88,41,1).
app_href(3,88,1412,1).
app_href(3,88,26,1).
app_href(3,88,503,1).
app_href(3,89,1011,1).
app_href(3,89,736,1).
app_href(3,89,396,1).
app_href(2,90,1070,1).
app_href(2,90,503,1).
app_href(2,91,1011,1).
app_href(2,91,736,1).
app_href(2,91,718,1).
app_href(2,35,365,1).
app_href(2,35,6,1).
app_href(2,92,536,1).
app_href(3,89,197,1).
app_href(2,93,1264,1).
app_href(2,94,517,1).
app_href(2,95,1208,1).
app_href(2,96,667,1).
app_href(2,96,1358,1).
app_href(2,96,41,1).
app_href(2,96,9,1).
app_href(2,96,728,1).
app_href(2,96,503,1).
app_href(2,87,444,1).
app_href(2,87,358,1).
app_href(2,97,1030,1).
app_href(2,15,899,1).
app_href(2,15,1310,1).
app_href(2,15,668,1).
app_href(2,98,1105,1).
app_href(3,12,611,1).
app_href(3,12,660,1).
app_href(3,12,140,1).
app_href(3,12,2,1).
app_href(3,12,132,1).
app_href(3,12,36,1).
app_href(3,99,1346,1).
app_href(3,99,1014,1).
app_href(3,99,634,1).
app_href(3,99,772,1).
app_href(3,100,1346,1).
app_href(3,100,1014,1).
app_href(3,100,197,1).
app_href(3,100,634,1).
app_href(3,101,1406,1).
app_href(3,101,701,1).
app_href(3,101,375,1).
app_href(3,102,517,1).
app_href(3,102,1093,1).
app_href(3,103,1011,1).
app_href(3,103,736,1).
app_href(3,103,197,1).
app_href(3,17,325,1).
app_href(3,17,1323,1).
app_href(3,66,1374,1).
app_href(3,66,392,1).
app_href(3,66,458,1).
app_href(3,104,736,1).
app_href(3,104,1189,1).
app_href(3,104,197,1).
app_href(3,104,1346,1).
app_href(3,104,667,1).
app_href(3,104,37,1).
app_href(3,104,1412,1).
app_href(3,104,26,1).
app_href(3,104,503,1).
app_href(6,87,444,1).
app_href(6,87,358,1).
app_href(6,105,309,1).
app_href(6,105,1310,1).
app_href(6,105,736,1).
app_href(6,105,729,1).
app_href(6,106,1409,1).
app_href(6,106,854,1).
app_href(6,106,736,1).
app_href(6,107,854,1).
app_href(6,107,1009,1).
app_href(6,35,365,1).
app_href(6,35,6,1).
app_href(6,15,899,1).
app_href(6,15,1310,1).
app_href(6,15,668,1).
app_href(7,108,214,1).
app_href(7,108,736,1).
app_href(7,108,213,1).
app_href(7,109,1105,1).
app_href(7,110,740,1).
app_href(7,110,423,1).
app_href(7,110,140,1).
app_href(7,110,911,1).
app_href(7,110,1014,1).
app_href(7,110,1346,1).
app_href(7,110,634,1).
app_href(7,110,804,1).
app_href(7,110,678,1).
app_href(7,110,214,1).
app_href(7,35,365,1).
app_href(7,35,6,1).
app_href(7,111,667,1).
app_href(7,111,41,1).
app_href(7,111,145,1).
app_href(7,111,1412,1).
app_href(7,111,503,1).
app_href(7,112,1384,1).
app_href(7,112,740,1).
app_href(7,113,331,1).
app_href(7,113,1346,1).
app_href(7,113,1323,1).
app_href(7,87,444,1).
app_href(7,87,358,1).
app_href(7,114,1030,1).
app_href(7,115,1358,1).
app_href(7,115,1264,1).
app_href(7,115,740,1).
app_href(7,116,214,1).
app_href(7,116,736,1).
app_href(7,15,899,1).
app_href(7,15,1310,1).
app_href(7,15,668,1).
app_href(7,117,1264,1).
app_href(7,117,213,1).
app_href(9,118,1123,1).
app_href(9,118,805,1).
app_href(1,87,444,1).
app_href(9,118,1014,1).
app_href(1,87,358,1).
app_href(9,119,836,1).
app_href(9,120,1248,1).
app_href(9,120,836,1).
app_href(9,120,140,1).
app_href(9,121,1456,1).
app_href(9,122,1248,1).
app_href(9,122,140,1).
app_href(9,122,836,1).
app_href(9,122,1456,1).
app_href(9,123,1248,1).
app_href(9,123,836,1).
app_href(1,124,1473,1).
app_href(9,123,140,1).
app_href(9,126,1248,1).
app_href(9,126,836,1).
app_href(9,126,140,1).
app_href(9,127,1456,1).
app_href(9,95,1248,1).
app_href(9,95,836,1).
app_href(9,95,140,1).
app_href(9,131,1248,1).
app_href(9,15,899,1).
app_href(9,15,1310,1).
app_href(9,15,668,1).
app_href(9,132,1248,1).
app_href(9,132,836,1).
app_href(9,132,140,1).
app_href(9,133,1248,1).
app_href(9,133,836,1).
app_href(9,133,140,1).
app_href(9,134,1248,1).
app_href(9,134,836,1).
app_href(9,134,140,1).
app_href(9,135,1248,1).
app_href(9,135,836,1).
app_href(9,135,140,1).
app_href(9,136,1456,1).
app_href(9,137,1248,1).
app_href(9,138,1248,1).
app_href(9,138,140,1).
app_href(9,138,836,1).
app_href(9,138,1456,1).
app_href(9,139,749,1).
app_href(9,140,1456,1).
app_href(9,141,1248,1).
app_href(9,141,836,1).
app_href(9,141,140,1).
app_href(9,142,1456,1).
app_href(9,143,1456,1).
app_href(9,144,1216,1).
app_href(9,145,1456,1).
app_href(9,146,894,1).
app_href(9,146,1141,1).
app_href(9,146,736,2).
app_href(9,146,704,1).
app_href(9,146,1047,1).
app_href(9,146,1162,1).
app_href(9,146,649,1).
app_href(9,146,210,1).
app_href(9,146,140,1).
app_href(9,146,1364,6).
app_href(9,146,1168,1).
app_href(9,146,715,1).
app_href(9,146,1198,1).
app_href(9,146,1358,1).
app_href(9,153,1456,1).
app_href(9,158,1248,1).
app_href(9,158,836,1).
app_href(9,158,140,1).
app_href(9,391,836,1).
app_href(9,392,1456,1).
app_href(9,393,1456,1).
app_href(9,87,444,1).
app_href(9,87,358,1).
app_href(9,35,365,1).
app_href(9,394,836,1).
app_href(9,395,1456,1).
app_href(9,396,1456,1).
app_href(9,397,1249,1).
app_href(9,398,994,1).
app_href(9,399,836,1).
app_href(9,400,1456,1).
app_href(9,401,836,1).
app_href(9,402,1456,1).
app_href(9,403,1456,1).
app_href(9,404,1456,1).
app_href(9,405,1248,1).
app_href(9,405,836,1).
app_href(9,405,140,1).
app_href(9,406,1248,1).
app_href(9,406,836,1).
app_href(9,406,140,1).
app_href(9,407,1248,1).
app_href(9,407,836,1).
app_href(9,407,140,1).
app_href(9,408,1456,1).
app_href(9,409,1187,1).
app_href(9,409,736,1).
app_href(9,409,1047,1).
app_href(9,410,1456,1).
app_href(9,411,1123,1).
app_href(9,411,796,1).
app_href(9,411,1014,1).
app_href(9,412,650,1).
app_href(9,413,1248,1).
app_href(9,413,836,1).
app_href(9,413,140,1).
app_href(9,414,1456,1).
app_href(9,415,1456,1).
app_href(9,416,1456,1).
app_href(9,417,1248,1).
app_href(9,417,836,1).
app_href(9,417,140,1).
app_href(9,418,1123,1).
app_href(9,418,803,1).
app_href(9,418,1014,1).
app_href(9,419,1248,1).
app_href(9,419,836,1).
app_href(9,419,140,1).
app_href(9,420,1456,1).
app_href(9,421,1359,1).
app_href(11,422,365,1).
app_href(11,423,949,2).
app_href(11,423,358,2).
app_href(11,424,672,2).
app_href(12,425,611,1).
app_href(12,425,36,1).
app_href(12,426,1117,1).
app_href(12,35,365,1).
app_href(12,35,6,1).
app_href(12,427,1450,1).
app_href(12,427,611,1).
app_href(12,427,633,1).
app_href(12,427,1017,1).
app_href(12,427,1033,1).
app_href(12,87,444,1).
app_href(12,87,358,1).
app_href(12,428,1030,1).
app_href(12,429,736,1).
app_href(12,429,540,1).
app_href(12,429,941,1).
app_href(12,429,36,1).
app_href(12,430,736,1).
app_href(12,430,1346,1).
app_href(12,430,1014,1).
app_href(12,430,634,1).
app_href(12,430,36,1).
app_href(12,431,1061,1).
app_href(12,432,736,1).
app_href(12,432,1264,1).
app_href(12,432,145,1).
app_href(12,432,36,1).
app_href(12,433,660,1).
app_href(12,433,736,2).
app_href(12,433,616,1).
app_href(12,433,36,1).
app_href(12,433,502,1).
app_href(12,433,140,1).
app_href(12,433,145,1).
app_href(12,433,564,1).
app_href(12,433,611,1).
app_href(12,15,899,1).
app_href(12,15,1310,1).
app_href(12,15,668,1).
app_href(12,434,1346,1).
app_href(12,434,1014,1).
app_href(12,434,634,1).
app_href(12,434,772,1).
app_href(16,435,1357,1).
app_href(1,124,1384,1).
app_href(1,125,1358,1).
app_href(1,125,1473,1).
app_href(1,125,736,1).
app_href(1,125,145,1).
app_href(1,125,1346,1).
app_href(1,125,1084,1).
app_href(1,125,564,1).
app_href(1,128,1437,1).
app_href(1,129,1358,1).
app_href(1,129,1473,1).
app_href(1,129,1337,1).
app_href(1,129,736,1).
app_href(1,129,1014,1).
app_href(1,129,634,1).
app_href(1,130,181,1).
app_href(1,130,1473,1).
app_href(1,130,1310,1).
app_href(1,35,365,1).
app_href(1,15,899,1).
app_href(1,15,1310,1).
app_href(1,15,668,1).
app_href(4,147,736,1).
app_href(4,147,941,1).
app_href(4,147,2,1).
app_href(4,147,1360,1).
app_href(4,148,1486,1).
app_href(4,148,1014,1).
app_href(4,149,830,1).
app_href(4,40,318,1).
app_href(4,40,736,1).
app_href(4,40,1264,1).
app_href(4,40,223,1).
app_href(4,40,810,1).
app_href(4,150,830,1).
app_href(4,150,1256,1).
app_href(4,150,2,1).
app_href(4,61,1094,1).
app_href(4,61,955,1).
app_href(4,61,1452,1).
app_href(4,61,1310,1).
app_href(4,61,1012,1).
app_href(4,61,1345,1).
app_href(4,61,1227,1).
app_href(4,61,421,1).
app_href(4,61,1466,1).
app_href(4,61,257,1).
app_href(4,15,899,1).
app_href(4,15,1310,1).
app_href(4,15,668,1).
app_href(4,152,1370,3).
app_href(4,152,992,1).
app_href(4,152,1133,1).
app_href(10,155,309,1).
app_href(4,154,2,1).
app_href(10,155,736,1).
app_href(10,155,1310,1).
app_href(10,155,145,1).
app_href(10,155,564,1).
app_href(10,155,973,1).
app_href(10,155,1014,1).
app_href(10,155,634,1).
app_href(10,155,766,1).
app_href(10,35,365,1).
app_href(10,35,6,1).
app_href(10,322,452,1).
app_href(10,322,1014,1).
app_href(10,323,460,1).
app_href(4,161,755,1).
app_href(10,323,238,1).
app_href(10,323,767,1).
app_href(10,15,899,1).
app_href(10,15,1310,1).
app_href(10,15,668,1).
app_href(4,162,1364,1).
app_href(10,87,444,1).
app_href(4,162,2,1).
app_href(10,87,358,1).
app_href(4,163,489,1).
app_href(4,164,454,1).
app_href(4,165,867,1).
app_href(4,165,1336,1).
app_href(4,165,728,2).
app_href(10,324,1011,1).
app_href(4,166,736,1).
app_href(10,324,763,1).
app_href(10,324,1269,1).
app_href(10,325,229,1).
app_href(4,166,231,1).
app_href(10,325,812,1).
app_href(10,325,969,2).
app_href(4,167,1418,1).
app_href(10,325,1408,1).
app_href(4,167,423,1).
app_href(10,325,223,1).
app_href(10,326,461,1).
app_href(10,326,1391,1).
app_href(4,167,2,1).
app_href(4,168,1069,1).
app_href(4,169,1256,2).
app_href(10,326,382,1).
app_href(4,169,736,1).
app_href(4,169,1041,1).
app_href(4,169,1274,1).
app_href(4,169,223,1).
app_href(4,169,140,1).
app_href(4,169,670,3).
app_href(4,169,318,3).
app_href(4,169,480,1).
app_href(10,326,191,1).
app_href(10,326,423,1).
app_href(10,326,992,1).
app_href(10,326,1460,1).
app_href(10,326,1420,1).
app_href(4,169,41,2).
app_href(4,169,1412,2).
app_href(4,169,26,1).
app_href(10,327,1011,1).
app_href(4,169,9,1).
app_href(4,169,8,1).
app_href(4,169,219,1).
app_href(4,87,444,1).
app_href(4,87,1189,1).
app_href(4,170,1192,1).
app_href(4,170,132,1).
app_href(4,171,944,1).
app_href(4,171,673,1).
app_href(10,327,736,1).
app_href(10,327,1269,1).
app_href(10,328,210,1).
app_href(10,328,1384,1).
app_href(10,329,1443,1).
app_href(10,329,444,1).
app_href(4,172,1076,1).
app_href(10,329,26,1).
app_href(10,329,736,1).
app_href(10,329,1189,1).
app_href(10,329,1014,1).
app_href(10,329,480,1).
app_href(10,329,504,1).
app_href(4,172,1053,1).
app_href(10,330,1384,1).
app_href(10,330,132,1).
app_href(10,331,1194,1).
app_href(10,331,736,1).
app_href(10,331,1123,1).
app_href(10,331,232,1).
app_href(4,173,1336,1).
app_href(4,174,444,1).
app_href(4,174,318,1).
app_href(4,175,146,1).
app_href(4,175,2,1).
app_href(4,176,444,1).
app_href(4,176,318,1).
app_href(4,177,1403,1).
app_href(4,177,1341,1).
app_href(4,177,970,1).
app_href(5,151,581,1).
app_href(5,151,303,1).
app_href(5,4,1139,1).
app_href(5,4,1424,1).
app_href(5,4,318,1).
app_href(5,4,1070,1).
app_href(5,4,1460,1).
app_href(5,156,1207,1).
app_href(5,156,967,1).
app_href(5,157,1168,1).
app_href(5,159,701,1).
app_href(5,159,736,1).
app_href(5,159,927,1).
app_href(5,159,779,1).
app_href(5,160,902,1).
app_href(5,160,540,1).
app_href(5,242,1363,1).
app_href(5,242,1455,1).
app_href(5,242,117,1).
app_href(5,243,854,1).
app_href(5,11,736,1).
app_href(5,11,1264,1).
app_href(5,11,1346,1).
app_href(5,11,810,1).
app_href(5,244,1090,1).
app_href(5,244,701,1).
app_href(5,244,736,1).
app_href(5,244,864,1).
app_href(5,244,794,1).
app_href(5,244,1430,1).
app_href(5,245,880,1).
app_href(5,245,159,1).
app_href(5,246,367,1).
app_href(5,15,899,1).
app_href(5,15,1310,1).
app_href(5,15,668,1).
app_href(5,247,581,1).
app_href(5,247,686,1).
app_href(5,247,894,1).
app_href(5,18,667,1).
app_href(5,18,444,1).
app_href(5,18,736,1).
app_href(5,18,1457,1).
app_href(5,18,568,1).
app_href(5,248,119,1).
app_href(5,19,556,1).
app_href(5,19,244,1).
app_href(5,21,1041,1).
app_href(5,21,140,1).
app_href(5,21,667,1).
app_href(5,21,41,1).
app_href(5,21,277,1).
app_href(5,21,728,1).
app_href(5,21,9,2).
app_href(5,21,1358,1).
app_href(5,21,503,1).
app_href(5,24,736,1).
app_href(5,24,1246,1).
app_href(5,24,1014,1).
app_href(5,24,634,1).
app_href(5,249,883,1).
app_href(5,249,530,1).
app_href(5,25,362,1).
app_href(5,25,736,1).
app_href(5,25,1346,1).
app_href(5,25,296,1).
app_href(5,25,1093,1).
app_href(5,25,564,1).
app_href(5,25,810,1).
app_href(5,250,1127,1).
app_href(5,251,1443,1).
app_href(5,251,1358,1).
app_href(5,251,736,1).
app_href(5,251,1276,1).
app_href(5,26,1349,2).
app_href(5,26,452,3).
app_href(5,26,444,2).
app_href(5,26,26,1).
app_href(5,26,736,1).
app_href(5,26,1189,2).
app_href(5,26,1014,1).
app_href(5,29,536,1).
app_href(5,29,26,1).
app_href(5,29,736,2).
app_href(5,29,1346,1).
app_href(5,29,1189,1).
app_href(5,29,140,1).
app_href(5,252,1141,1).
app_href(5,252,1162,1).
app_href(5,252,715,1).
app_href(5,252,894,1).
app_href(5,253,657,1).
app_href(5,254,1424,1).
app_href(5,254,1011,1).
app_href(5,254,645,1).
app_href(5,255,1358,1).
app_href(5,255,736,1).
app_href(5,255,332,1).
app_href(5,255,766,1).
app_href(5,255,393,1).
app_href(5,256,1217,1).
app_href(5,256,1073,1).
app_href(5,36,1144,1).
app_href(5,36,242,1).
app_href(5,36,1304,1).
app_href(5,38,736,1).
app_href(5,38,483,1).
app_href(5,257,939,1).
app_href(5,39,266,1).
app_href(5,39,1358,1).
app_href(5,39,1093,1).
app_href(5,39,948,1).
app_href(5,40,452,1).
app_href(5,42,452,1).
app_href(5,258,1358,1).
app_href(5,258,324,1).
app_href(5,258,145,1).
app_href(5,258,740,1).
app_href(5,259,1198,1).
app_href(5,260,540,2).
app_href(5,260,1041,1).
app_href(5,260,941,2).
app_href(5,260,1309,1).
app_href(5,260,1235,1).
app_href(5,260,37,1).
app_href(5,260,728,1).
app_href(5,260,9,1).
app_href(5,260,970,1).
app_href(5,261,1275,1).
app_href(5,261,1007,1).
app_href(5,261,883,1).
app_href(5,261,495,1).
app_href(5,48,736,1).
app_href(5,48,1264,1).
app_href(5,48,1346,1).
app_href(5,48,864,1).
app_href(5,48,1430,1).
app_href(5,49,1409,1).
app_href(5,49,736,5).
app_href(5,49,9,2).
app_href(5,49,880,1).
app_href(5,49,1012,1).
app_href(5,49,934,1).
app_href(5,49,1343,1).
app_href(5,49,809,1).
app_href(5,262,444,1).
app_href(5,262,1362,2).
app_href(5,262,442,1).
app_href(5,50,736,1).
app_href(5,50,1346,1).
app_href(5,50,1384,1).
app_href(5,263,325,1).
app_href(5,263,1323,1).
app_href(5,263,1095,1).
app_href(5,51,444,1).
app_href(5,51,277,1).
app_href(5,264,444,1).
app_href(10,331,1384,1).
app_href(10,331,1209,1).
app_href(10,331,140,1).
app_href(10,331,145,1).
app_href(10,331,480,1).
app_href(10,331,766,1).
app_href(10,331,1358,1).
app_href(10,332,1384,1).
app_href(10,332,1172,1).
app_href(10,333,616,1).
app_href(10,333,1277,1).
app_href(10,333,1358,1).
app_href(10,333,1337,1).
app_href(10,333,602,1).
app_href(10,333,766,1).
app_href(10,334,585,1).
app_href(10,334,1112,1).
app_href(10,334,143,1).
app_href(10,334,1360,1).
app_href(10,335,812,1).
app_href(10,335,281,1).
app_href(10,335,969,2).
app_href(10,335,1408,1).
app_href(10,335,294,1).
app_href(13,465,1403,2).
app_href(13,465,614,2).
app_href(13,465,736,2).
app_href(13,465,1385,2).
app_href(13,465,140,2).
app_href(13,465,564,2).
app_href(13,465,730,2).
app_href(13,465,1460,2).
app_href(13,465,800,2).
app_href(13,336,1358,1).
app_href(13,336,26,1).
app_href(13,336,736,1).
app_href(13,336,1346,1).
app_href(13,336,1189,1).
app_href(13,336,140,1).
app_href(13,336,872,1).
app_href(13,336,1198,1).
app_href(13,337,1041,1).
app_href(13,337,1358,1).
app_href(13,337,26,1).
app_href(13,337,736,1).
app_href(13,337,1198,1).
app_href(13,120,1248,1).
app_href(13,120,564,1).
app_href(13,338,773,1).
app_href(13,338,1011,1).
app_href(13,338,652,1).
app_href(13,338,998,1).
app_href(13,339,661,1).
app_href(13,339,736,1).
app_href(13,339,160,1).
app_href(13,339,379,1).
app_href(13,339,564,1).
app_href(13,339,1346,1).
app_href(13,339,1102,1).
app_href(13,339,1360,1).
app_href(13,339,641,1).
app_href(13,340,662,1).
app_href(13,340,1327,1).
app_href(13,340,403,1).
app_href(13,340,546,1).
app_href(13,340,140,1).
app_href(13,340,1053,2).
app_href(13,341,1034,1).
app_href(13,341,1053,1).
app_href(13,342,1011,1).
app_href(13,342,763,1).
app_href(13,342,378,1).
app_href(13,15,899,1).
app_href(13,15,1310,1).
app_href(13,15,668,1).
app_href(13,343,1011,1).
app_href(13,343,763,1).
app_href(13,343,1260,1).
app_href(13,343,378,1).
app_href(13,344,1011,1).
app_href(13,344,763,1).
app_href(13,344,378,1).
app_href(13,344,702,1).
app_href(13,345,1011,1).
app_href(13,345,763,1).
app_href(13,345,800,1).
app_href(13,345,186,1).
app_href(13,345,1198,1).
app_href(13,136,1456,1).
app_href(13,136,564,1).
app_href(13,346,1011,1).
app_href(13,346,855,1).
app_href(13,346,309,1).
app_href(13,346,1310,1).
app_href(13,346,186,1).
app_href(13,346,1198,1).
app_href(13,347,1011,1).
app_href(13,347,763,1).
app_href(13,347,855,1).
app_href(13,347,186,1).
app_href(13,347,1198,1).
app_href(13,348,616,1).
app_href(13,348,1123,1).
app_href(13,348,765,1).
app_href(13,349,667,1).
app_href(13,349,1231,1).
app_href(13,349,800,1).
app_href(13,349,736,1).
app_href(13,349,1408,1).
app_href(13,349,1236,1).
app_href(13,350,1011,1).
app_href(13,350,736,1).
app_href(13,350,1198,1).
app_href(13,351,160,1).
app_href(13,351,1198,1).
app_href(13,352,1011,1).
app_href(13,352,309,1).
app_href(13,352,1310,1).
app_href(13,352,186,1).
app_href(13,352,1198,1).
app_href(13,352,258,1).
app_href(13,392,1456,1).
app_href(13,392,564,1).
app_href(13,87,444,1).
app_href(13,87,358,1).
app_href(13,353,1385,2).
app_href(13,353,730,2).
app_href(13,354,1011,1).
app_href(13,354,763,1).
app_href(13,354,186,1).
app_href(13,354,1198,1).
app_href(13,354,258,1).
app_href(13,35,365,1).
app_href(13,35,6,1).
app_href(13,355,1011,1).
app_href(13,355,763,1).
app_href(13,355,267,1).
app_href(13,355,1198,1).
app_href(13,482,616,1).
app_href(13,482,1092,1).
app_href(13,482,736,1).
app_href(13,482,145,1).
app_href(13,482,1025,1).
app_href(13,482,270,1).
app_href(13,356,1011,1).
app_href(13,356,309,1).
app_href(13,356,909,1).
app_href(13,356,1310,1).
app_href(13,356,186,1).
app_href(13,356,1198,1).
app_href(13,357,1011,2).
app_href(13,357,267,2).
app_href(13,357,736,2).
app_href(13,357,1198,2).
app_href(13,358,1011,1).
app_href(13,358,763,1).
app_href(13,358,1270,1).
app_href(4,178,829,1).
app_href(4,179,444,1).
app_href(4,179,318,1).
app_href(4,180,444,1).
app_href(4,180,318,1).
app_href(4,181,354,1).
app_href(4,182,1218,1).
app_href(4,182,420,1).
app_href(4,183,444,1).
app_href(4,184,317,1).
app_href(4,184,677,1).
app_href(4,185,1442,1).
app_href(4,185,736,1).
app_href(4,185,140,1).
app_href(4,185,233,1).
app_href(4,186,728,1).
app_href(4,186,454,1).
app_href(4,187,444,1).
app_href(4,187,728,1).
app_href(4,188,895,1).
app_href(4,188,410,1).
app_href(4,188,289,1).
app_href(4,188,140,1).
app_href(4,189,287,1).
app_href(4,189,140,1).
app_href(4,189,138,1).
app_href(4,190,316,1).
app_href(4,190,736,1).
app_href(4,190,1083,1).
app_href(14,191,261,1).
app_href(14,192,292,1).
app_href(14,192,226,1).
app_href(14,192,155,1).
app_href(14,192,1212,1).
app_href(14,193,155,1).
app_href(14,193,151,1).
app_href(14,193,627,1).
app_href(14,193,140,1).
app_href(14,193,1464,1).
app_href(14,194,26,1).
app_href(14,50,736,1).
app_href(14,50,1346,1).
app_href(14,50,1384,1).
app_href(14,195,41,1).
app_href(14,196,9,1).
app_href(14,197,1254,1).
app_href(14,15,899,1).
app_href(14,15,1310,1).
app_href(14,15,668,1).
app_href(14,198,37,1).
app_href(14,199,380,1).
app_href(14,199,736,1).
app_href(14,199,970,1).
app_href(14,199,555,1).
app_href(14,200,26,1).
app_href(14,201,41,1).
app_href(14,202,26,1).
app_href(14,203,459,1).
app_href(14,204,579,1).
app_href(14,205,1234,1).
app_href(14,87,616,1).
app_href(14,87,444,1).
app_href(14,87,540,1).
app_href(14,87,358,1).
app_href(14,87,1360,1).
app_href(14,206,659,1).
app_href(14,207,295,1).
app_href(14,35,365,1).
app_href(14,35,6,1).
app_href(14,208,1354,1).
app_href(14,209,612,1).
app_href(14,210,9,1).
app_href(14,211,163,1).
app_href(14,212,209,1).
app_href(14,213,152,1).
app_href(14,213,1014,1).
app_href(14,214,37,1).
app_href(14,215,26,1).
app_href(14,216,9,1).
app_href(14,217,41,1).
app_href(14,218,228,1).
app_href(14,219,37,1).
app_href(14,220,932,1).
app_href(14,221,9,1).
app_href(14,222,436,1).
app_href(14,223,37,1).
app_href(14,224,793,1).
app_href(14,225,1268,1).
app_href(14,9,1410,1).
app_href(14,9,26,1).
app_href(14,9,736,1).
app_href(14,9,140,1).
app_href(14,9,1189,1).
app_href(14,9,1364,1).
app_href(15,226,1218,1).
app_href(15,226,1088,1).
app_href(15,227,284,1).
app_href(15,227,976,1).
app_href(15,227,728,1).
app_href(15,228,421,1).
app_href(15,229,131,1).
app_href(15,229,1329,1).
app_href(15,230,917,1).
app_href(15,231,1137,1).
app_href(15,231,714,1).
app_href(15,232,1,1).
app_href(15,233,1310,1).
app_href(15,233,916,1).
app_href(15,15,309,1).
app_href(15,15,1,2).
app_href(15,15,1310,1).
app_href(15,152,1370,1).
app_href(15,163,489,1).
app_href(15,234,1465,1).
app_href(15,234,1238,1).
app_href(15,235,641,1).
app_href(15,235,250,1).
app_href(15,235,1360,1).
app_href(15,236,140,1).
app_href(15,236,819,1).
app_href(15,236,988,1).
app_href(15,237,736,1).
app_href(15,238,1443,1).
app_href(15,238,1218,1).
app_href(15,239,1200,1).
app_href(15,239,513,1).
app_href(15,239,944,1).
app_href(15,239,134,1).
app_href(15,239,140,1).
app_href(15,239,5,1).
app_href(15,240,1,1).
app_href(15,241,1372,1).
app_href(15,241,1318,1).
app_href(15,541,454,1).
app_href(15,172,1076,1).
app_href(15,542,1,1).
app_href(15,543,1151,1).
app_href(15,544,524,1).
app_href(15,545,1023,1).
app_href(15,546,1,1).
app_href(15,177,1403,1).
app_href(15,177,1341,1).
app_href(15,177,970,1).
app_href(15,547,444,1).
app_href(15,548,1200,1).
app_href(15,548,540,1).
app_href(15,548,134,1).
app_href(15,548,5,1).
app_href(15,549,1248,1).
app_href(15,184,317,1).
app_href(15,184,677,1).
app_href(15,550,354,1).
app_href(15,551,647,1).
app_href(15,551,938,1).
app_href(15,552,1069,1).
app_href(15,553,1144,1).
app_href(15,553,941,1).
app_href(8,464,1164,1).
app_href(8,465,1358,1).
app_href(8,465,614,1).
app_href(8,465,736,1).
app_href(8,466,940,1).
app_href(8,2,1207,1).
app_href(8,3,197,1).
app_href(8,467,942,1).
app_href(8,467,288,1).
app_href(8,6,854,1).
app_href(8,6,145,1).
app_href(8,7,325,1).
app_href(8,7,1035,1).
app_href(8,7,860,1).
app_href(8,7,740,1).
app_href(8,7,970,1).
app_href(8,468,1126,1).
app_href(8,468,736,1).
app_href(8,468,544,1).
app_href(8,468,809,1).
app_href(8,10,37,1).
app_href(8,10,145,1).
app_href(8,10,766,1).
app_href(8,10,8,1).
app_href(8,125,1473,1).
app_href(8,125,1084,1).
app_href(8,12,1358,1).
app_href(8,12,1337,1).
app_href(8,12,736,1).
app_href(8,12,36,1).
app_href(8,13,1198,1).
app_href(8,469,1409,1).
app_href(8,469,736,1).
app_href(8,469,1135,1).
app_href(8,469,1130,1).
app_href(8,469,527,1).
app_href(8,134,928,1).
app_href(8,16,701,1).
app_href(8,16,736,1).
app_href(8,16,927,1).
app_href(8,17,1323,1).
app_href(8,470,1374,1).
app_href(8,470,392,1).
app_href(8,470,1323,1).
app_href(8,20,1443,1).
app_href(8,20,1358,1).
app_href(8,20,736,1).
app_href(8,20,1086,1).
app_href(8,20,1276,1).
app_href(8,471,581,1).
app_href(8,471,303,1).
app_href(8,471,503,1).
app_href(8,21,667,1).
app_href(8,21,41,1).
app_href(8,21,9,2).
app_href(8,21,277,1).
app_href(8,21,728,1).
app_href(8,21,140,1).
app_href(8,21,503,1).
app_href(8,472,1165,1).
app_href(8,23,581,1).
app_href(8,23,303,1).
app_href(8,473,503,1).
app_href(8,473,1033,1).
app_href(8,27,1275,2).
app_href(8,27,1007,2).
app_href(8,27,883,2).
app_href(8,27,495,2).
app_href(8,474,1126,1).
app_href(8,474,942,1).
app_href(8,474,736,1).
app_href(8,474,350,1).
app_href(8,474,809,1).
app_href(8,30,400,1).
app_href(8,30,1041,1).
app_href(8,30,1358,1).
app_href(8,30,160,1).
app_href(8,30,736,1).
app_href(8,475,1217,1).
app_href(8,475,688,1).
app_href(8,475,707,1).
app_href(8,475,100,1).
app_href(8,475,1264,1).
app_href(8,475,922,1).
app_href(8,476,847,1).
app_href(8,476,1319,1).
app_href(8,33,720,1).
app_href(8,33,669,1).
app_href(8,33,1318,1).
app_href(8,33,939,1).
app_href(8,477,540,1).
app_href(8,477,140,1).
app_href(8,477,667,1).
app_href(8,477,41,1).
app_href(8,477,277,1).
app_href(8,477,37,1).
app_href(8,477,728,1).
app_href(8,477,9,2).
app_href(8,477,367,1).
app_href(8,477,219,1).
app_href(8,34,1162,1).
app_href(8,35,28,1).
app_href(8,35,365,1).
app_href(8,35,6,1).
app_href(8,87,616,1).
app_href(8,87,540,1).
app_href(8,87,358,1).
app_href(8,87,1360,1).
app_href(8,478,674,1).
app_href(8,479,616,1).
app_href(8,479,1123,1).
app_href(8,479,1243,1).
app_href(8,479,1358,1).
app_href(8,479,736,1).
app_href(8,479,1196,1).
app_href(8,479,791,2).
app_href(8,479,530,1).
app_href(8,480,779,1).
app_href(8,480,503,1).
app_href(8,41,1217,1).
app_href(8,41,1073,1).
app_href(8,41,736,1).
app_href(8,41,145,1).
app_href(8,44,704,1).
app_href(8,481,1126,1).
app_href(8,481,716,1).
app_href(8,481,736,1).
app_href(8,481,809,1).
app_href(8,481,894,1).
app_href(8,45,865,1).
app_href(8,45,736,1).
app_href(8,45,1430,1).
app_href(8,138,1362,1).
app_href(8,138,277,1).
app_href(8,47,1127,1).
app_href(8,482,1358,1).
app_href(8,482,736,1).
app_href(8,482,145,1).
app_href(8,482,1025,1).
app_href(8,482,270,1).
app_href(8,406,1163,1).
app_href(8,483,1398,1).
app_href(8,484,1217,1).
app_href(8,484,1358,1).
app_href(8,484,187,1).
app_href(8,484,736,1).
app_href(8,484,189,1).
app_href(8,484,140,1).
app_href(8,484,730,1).
app_href(8,485,540,1).
app_href(8,485,941,1).
app_href(8,486,964,1).
app_href(8,486,736,1).
app_href(8,486,1343,1).
app_href(8,137,736,1).
app_href(8,54,1358,1).
app_href(8,54,429,1).
app_href(5,264,365,2).
app_href(5,264,6,1).
app_href(5,264,1341,1).
app_href(5,264,829,1).
app_href(5,264,140,1).
app_href(5,264,564,1).
app_href(5,265,503,1).
app_href(5,265,1033,1).
app_href(5,52,372,1).
app_href(5,52,736,1).
app_href(5,52,977,1).
app_href(5,52,1346,1).
app_href(5,52,1093,1).
app_href(5,52,564,1).
app_href(5,52,810,1).
app_href(5,55,826,1).
app_href(5,55,736,1).
app_href(5,55,1346,1).
app_href(5,55,288,1).
app_href(5,56,1157,1).
app_href(5,56,608,1).
app_href(5,56,731,1).
app_href(5,56,251,1).
app_href(5,266,1255,1).
app_href(5,267,611,1).
app_href(5,267,660,1).
app_href(5,267,140,1).
app_href(5,267,36,1).
app_href(5,268,407,1).
app_href(5,268,152,1).
app_href(5,268,140,1).
app_href(5,268,156,1).
app_href(5,269,736,2).
app_href(5,269,1041,1).
app_href(5,269,181,1).
app_href(5,269,1401,1).
app_href(5,269,1204,1).
app_href(5,269,140,1).
app_href(5,269,1307,1).
app_href(5,269,313,1).
app_href(5,269,354,1).
app_href(5,269,1246,2).
app_href(5,269,728,1).
app_href(5,269,1358,2).
app_href(5,269,355,1).
app_href(5,269,603,1).
app_href(5,270,400,1).
app_href(5,270,1041,1).
app_href(5,270,160,1).
app_href(5,270,736,1).
app_href(5,270,782,1).
app_href(5,271,704,1).
app_href(5,60,26,1).
app_href(5,60,9,1).
app_href(5,60,1316,1).
app_href(5,272,580,1).
app_href(5,272,951,1).
app_href(5,61,1310,1).
app_href(5,61,115,1).
app_href(5,61,958,1).
app_href(5,61,1012,1).
app_href(5,62,1138,1).
app_href(5,62,959,1).
app_href(5,273,1041,1).
app_href(5,273,26,1).
app_href(5,273,736,1).
app_href(5,273,1264,1).
app_href(5,273,145,1).
app_href(5,67,505,1).
app_href(5,67,1358,1).
app_href(5,67,1248,1).
app_href(5,67,959,1).
app_href(5,67,985,1).
app_href(5,67,564,1).
app_href(5,67,690,1).
app_href(5,274,941,1).
app_href(5,275,1358,1).
app_href(5,275,736,1).
app_href(5,275,1047,1).
app_href(5,72,1030,1).
app_href(5,72,452,1).
app_href(5,72,140,1).
app_href(5,72,1105,1).
app_href(5,74,444,1).
app_href(5,74,1053,1).
app_href(5,74,1121,1).
app_href(5,75,959,1).
app_href(5,75,836,1).
app_href(5,75,564,1).
app_href(5,75,690,1).
app_href(5,76,444,1).
app_href(5,76,976,1).
app_href(5,76,1192,1).
app_href(5,76,1346,1).
app_href(5,77,214,1).
app_href(5,77,151,1).
app_href(5,77,736,3).
app_href(5,77,9,2).
app_href(5,77,1012,1).
app_href(5,77,197,1).
app_href(5,79,1117,1).
app_href(5,276,1411,1).
app_href(5,276,1012,1).
app_href(5,276,140,1).
app_href(5,276,810,1).
app_href(5,80,959,1).
app_href(5,80,897,1).
app_href(5,80,1456,1).
app_href(5,80,564,1).
app_href(5,80,690,1).
app_href(5,277,161,1).
app_href(5,277,736,1).
app_href(5,277,550,1).
app_href(5,277,733,1).
app_href(5,81,1372,1).
app_href(5,81,736,1).
app_href(5,81,1346,1).
app_href(5,81,1093,1).
app_href(5,81,564,1).
app_href(5,81,810,1).
app_href(5,278,1155,1).
app_href(5,279,1473,1).
app_href(5,280,1358,1).
app_href(5,280,429,1).
app_href(5,280,701,1).
app_href(5,280,736,1).
app_href(5,280,140,1).
app_href(5,280,777,1).
app_href(5,280,922,1).
app_href(5,281,1374,1).
app_href(5,281,392,1).
app_href(5,281,458,1).
app_href(5,282,864,1).
app_href(5,282,1430,1).
app_href(5,283,1123,1).
app_href(5,283,969,1).
app_href(5,284,649,1).
app_href(5,284,736,1).
app_href(5,169,408,1).
app_href(5,169,540,1).
app_href(5,169,595,1).
app_href(5,169,616,1).
app_href(5,169,677,1).
app_href(5,169,1385,1).
app_href(5,169,810,1).
app_href(5,169,145,1).
app_href(5,169,850,1).
app_href(5,169,1360,1).
app_href(5,169,1192,1).
app_href(20,285,536,1).
app_href(20,286,160,1).
app_href(20,287,1443,1).
app_href(20,287,1358,1).
app_href(20,287,736,1).
app_href(20,287,1477,1).
app_href(20,287,156,1).
app_href(20,287,490,1).
app_href(8,54,701,1).
app_href(8,54,736,1).
app_href(8,54,140,1).
app_href(8,54,777,2).
app_href(8,54,922,1).
app_href(8,54,503,1).
app_href(8,487,1064,1).
app_href(8,487,145,1).
app_href(8,488,1056,1).
app_href(8,488,905,1).
app_href(8,488,1050,1).
app_href(8,488,680,1).
app_href(8,488,1210,1).
app_href(8,58,1358,1).
app_href(8,58,736,1).
app_href(8,58,1047,1).
app_href(8,58,1086,1).
app_href(8,59,1033,1).
app_href(8,190,316,1).
app_href(8,190,736,1).
app_href(8,190,1083,1).
app_href(8,489,1041,1).
app_href(8,489,777,1).
app_href(8,489,883,1).
app_href(8,489,941,1).
app_href(8,489,720,1).
app_href(8,489,1127,1).
app_href(8,489,649,1).
app_href(8,489,657,1).
app_href(8,489,927,1).
app_href(8,489,939,1).
app_href(8,489,1217,1).
app_href(8,489,457,1).
app_href(8,489,145,3).
app_href(8,489,1162,1).
app_href(8,489,1337,1).
app_href(8,489,1207,1).
app_href(8,489,518,1).
app_href(8,489,669,1).
app_href(8,489,1443,1).
app_href(8,489,1323,1).
app_href(8,489,160,1).
app_href(8,489,367,1).
app_href(8,489,146,1).
app_href(8,489,1473,1).
app_href(8,489,740,1).
app_href(8,489,860,1).
app_href(8,489,429,1).
app_href(8,489,1033,1).
app_href(8,489,463,1).
app_href(8,489,37,1).
app_href(8,489,1410,1).
app_href(8,489,303,1).
app_href(8,489,1275,1).
app_href(8,489,8,1).
app_href(8,489,325,1).
app_href(8,489,1084,1).
app_href(8,489,1276,1).
app_href(8,489,1364,1).
app_href(8,489,854,1).
app_href(8,489,704,1).
app_href(8,489,1064,1).
app_href(8,489,880,1).
app_href(8,489,1430,1).
app_href(8,489,865,1).
app_href(8,489,1168,1).
app_href(8,489,177,1).
app_href(8,489,36,1).
app_href(8,489,1047,1).
app_href(8,489,532,2).
app_href(8,489,119,1).
app_href(8,489,1035,1).
app_href(8,489,970,1).
app_href(8,489,1086,2).
app_href(8,489,197,1).
app_href(8,489,1318,1).
app_href(8,489,273,1).
app_href(8,489,736,9).
app_href(8,489,140,4).
app_href(8,489,922,1).
app_href(8,489,400,1).
app_href(8,489,701,2).
app_href(8,489,1198,1).
app_href(8,489,1007,1).
app_href(8,489,495,1).
app_href(8,489,1073,1).
app_href(8,489,1135,1).
app_href(8,489,581,2).
app_href(8,489,766,1).
app_href(8,489,527,1).
app_href(8,489,1358,7).
app_href(8,65,581,1).
app_href(8,65,657,1).
app_href(8,68,880,1).
app_href(8,490,640,1).
app_href(8,490,736,1).
app_href(8,490,864,1).
app_href(8,490,1430,1).
app_href(8,69,1168,1).
app_href(8,491,532,1).
app_href(8,491,518,1).
app_href(8,491,273,1).
app_href(8,492,381,1).
app_href(8,492,1409,1).
app_href(8,492,736,1).
app_href(8,398,994,1).
app_href(8,412,650,1).
app_href(8,73,146,1).
app_href(8,73,941,1).
app_href(8,96,277,1).
app_href(8,96,140,1).
app_href(8,96,503,1).
app_href(8,493,667,1).
app_href(8,493,41,1).
app_href(8,493,747,1).
app_href(8,493,9,2).
app_href(8,493,941,1).
app_href(8,493,1446,1).
app_href(8,144,1216,1).
app_href(8,348,1358,1).
app_href(8,348,765,1).
app_href(8,348,736,1).
app_href(8,348,379,1).
app_href(8,348,530,1).
app_href(8,78,649,1).
app_href(8,78,736,1).
app_href(8,494,177,1).
app_href(8,494,532,1).
app_href(8,495,463,1).
app_href(8,495,457,1).
app_href(8,495,140,1).
app_href(8,82,119,1).
app_href(8,158,747,1).
app_href(19,496,331,1).
app_href(19,496,420,1).
app_href(19,496,736,1).
app_href(19,497,365,1).
app_href(19,498,556,1).
app_href(19,498,244,1).
app_href(19,499,242,1).
app_href(19,499,1018,1).
app_href(19,15,899,1).
app_href(19,15,1310,1).
app_href(19,15,668,1).
app_href(19,21,41,1).
app_href(19,21,9,1).
app_href(19,21,277,1).
app_href(19,21,8,1).
app_href(19,21,1235,1).
app_href(22,500,686,1).
app_href(22,500,497,1).
app_href(22,500,1384,1).
app_href(22,500,1267,1).
app_href(22,500,894,1).
app_href(22,501,1009,1).
app_href(22,35,365,1).
app_href(22,35,6,1).
app_href(16,435,701,1).
app_href(16,435,736,1).
app_href(16,435,927,1).
app_href(16,436,1264,1).
app_href(16,436,9,2).
app_href(16,436,779,1).
app_href(16,437,701,1).
app_href(16,437,736,1).
app_href(16,437,927,1).
app_href(16,437,1384,1).
app_href(16,480,667,1).
app_href(16,480,41,1).
app_href(16,480,9,1).
app_href(16,480,779,1).
app_href(16,480,503,1).
app_href(16,438,701,1).
app_href(16,438,927,1).
app_href(16,438,197,1).
app_href(16,35,365,1).
app_href(16,439,667,1).
app_href(16,439,26,1).
app_href(16,439,9,1).
app_href(16,439,779,1).
app_href(16,439,503,1).
app_href(16,87,444,1).
app_href(16,87,358,1).
app_href(16,15,899,1).
app_href(16,15,1310,1).
app_href(16,15,668,1).
app_href(17,440,1381,1).
app_href(17,440,763,1).
app_href(17,440,1323,1).
app_href(17,35,365,1).
app_href(17,35,6,1).
app_href(17,15,899,1).
app_href(17,15,1310,1).
app_href(17,15,668,1).
app_href(17,86,560,1).
app_href(17,86,905,1).
app_href(17,86,1264,1).
app_href(17,87,444,1).
app_href(17,87,358,1).
app_href(17,441,763,1).
app_href(17,441,512,1).
app_href(17,441,1323,1).
app_href(17,442,763,1).
app_href(17,442,210,1).
app_href(17,442,1038,1).
app_href(17,442,1323,1).
app_href(17,443,763,1).
app_href(17,443,1330,1).
app_href(17,443,1323,1).
app_href(17,444,1323,1).
app_href(17,444,736,1).
app_href(17,444,1189,1).
app_href(17,444,1346,1).
app_href(17,444,667,1).
app_href(17,444,41,1).
app_href(17,444,1412,1).
app_href(17,444,26,1).
app_href(17,444,503,1).
app_href(17,445,1323,1).
app_href(17,445,736,1).
app_href(17,445,1189,1).
app_href(17,445,1346,1).
app_href(17,445,667,1).
app_href(17,445,37,1).
app_href(17,445,1412,1).
app_href(17,445,26,1).
app_href(17,445,503,1).
app_href(17,446,763,1).
app_href(17,446,1038,1).
app_href(17,446,1323,1).
app_href(17,447,1392,1).
app_href(17,447,763,1).
app_href(17,447,1323,1).
app_href(17,448,331,1).
app_href(17,448,1346,1).
app_href(17,448,1323,1).
app_href(17,449,763,1).
app_href(17,449,893,1).
app_href(17,449,1038,1).
app_href(17,449,1323,1).
app_href(17,450,763,1).
app_href(17,450,308,1).
app_href(17,450,1323,1).
app_href(17,451,547,1).
app_href(17,451,763,1).
app_href(17,451,1323,1).
app_href(17,452,763,1).
app_href(17,452,1323,1).
app_href(17,453,763,1).
app_href(17,453,1174,1).
app_href(17,453,1323,1).
app_href(17,453,1343,1).
app_href(17,454,763,1).
app_href(17,454,236,1).
app_href(17,454,1323,1).
app_href(17,455,615,1).
app_href(17,455,736,1).
app_href(17,455,375,1).
app_href(17,455,1346,1).
app_href(17,455,1323,1).
app_href(17,455,1384,1).
app_href(17,455,772,1).
app_href(17,455,1460,1).
app_href(17,456,763,1).
app_href(17,456,642,1).
app_href(17,456,1323,1).
app_href(17,456,1343,1).
app_href(17,456,1020,1).
app_href(17,434,1346,1).
app_href(17,434,1014,1).
app_href(17,434,634,1).
app_href(17,434,772,1).
app_href(17,457,1323,1).
app_href(17,457,736,1).
app_href(17,457,1189,1).
app_href(17,457,1346,1).
app_href(17,457,667,1).
app_href(17,457,41,1).
app_href(17,457,1412,1).
app_href(17,457,26,1).
app_href(17,457,9,1).
app_href(17,457,503,1).
app_href(17,458,763,1).
app_href(17,458,1323,1).
app_href(17,458,1343,1).
app_href(17,459,763,1).
app_href(17,459,642,1).
app_href(17,459,1323,1).
app_href(17,459,1343,1).
app_href(17,460,763,1).
app_href(17,460,1038,1).
app_href(17,460,1323,1).
app_href(17,460,913,1).
app_href(23,461,1009,1).
app_href(23,35,365,1).
app_href(23,35,6,1).
app_href(23,462,1384,1).
app_href(23,463,536,1).
app_href(23,463,415,1).
app_href(23,623,1138,1).
app_href(23,623,667,1).
app_href(23,623,1346,1).
app_href(23,623,145,1).
app_href(23,623,1070,1).
app_href(23,623,503,1).
app_href(23,87,444,1).
app_href(23,87,358,1).
app_href(23,624,1123,1).
app_href(23,624,145,1).
app_href(23,471,667,1).
app_href(23,471,41,1).
app_href(23,471,9,1).
app_href(23,471,145,1).
app_href(23,471,727,1).
app_href(15,553,1335,1).
app_href(15,554,1252,1).
app_href(15,554,248,1).
app_href(15,554,673,1).
app_href(18,506,242,1).
app_href(18,506,393,1).
app_href(18,506,2,1).
app_href(18,148,1486,1).
app_href(18,148,1014,1).
app_href(18,149,830,1).
app_href(18,15,899,1).
app_href(18,15,1310,1).
app_href(18,15,668,1).
app_href(18,152,1370,3).
app_href(18,152,992,1).
app_href(18,152,1133,1).
app_href(18,161,755,1).
app_href(18,129,1473,1).
app_href(18,129,2,1).
app_href(18,507,118,1).
app_href(18,507,477,1).
app_href(18,507,2,1).
app_href(18,163,489,1).
app_href(18,508,726,1).
app_href(18,508,2,1).
app_href(18,164,454,1).
app_href(18,165,728,1).
app_href(18,165,2,1).
app_href(18,166,736,1).
app_href(18,166,231,1).
app_href(18,168,1069,1).
app_href(18,509,1443,1).
app_href(18,509,1218,1).
app_href(18,509,2,1).
app_href(18,170,1192,1).
app_href(18,170,132,1).
app_href(18,171,944,1).
app_href(18,171,673,1).
app_href(18,172,1076,1).
app_href(18,172,1053,1).
app_href(18,173,1336,1).
app_href(18,555,695,1).
app_href(18,555,452,1).
app_href(18,555,501,1).
app_href(18,511,1183,1).
app_href(18,511,296,1).
app_href(18,511,2,1).
app_href(18,175,146,1).
app_href(18,175,2,1).
app_href(18,177,1403,1).
app_href(18,177,1341,1).
app_href(18,177,970,1).
app_href(18,556,695,1).
app_href(18,556,583,1).
app_href(18,556,452,1).
app_href(18,178,829,1).
app_href(18,513,1459,1).
app_href(18,513,2,1).
app_href(18,181,354,1).
app_href(18,182,1200,1).
app_href(18,182,1218,1).
app_href(18,182,420,1).
app_href(18,182,134,1).
app_href(18,182,2,1).
app_href(18,514,575,1).
app_href(18,514,2,1).
app_href(18,183,444,1).
app_href(18,183,1200,1).
app_href(18,183,134,1).
app_href(18,183,2,1).
app_href(18,515,999,1).
app_href(18,515,2,1).
app_href(18,184,317,1).
app_href(18,184,677,1).
app_href(18,185,1442,1).
app_href(18,185,736,1).
app_href(18,185,140,1).
app_href(18,185,233,1).
app_href(18,52,1385,1).
app_href(18,52,2,1).
app_href(18,517,1318,1).
app_href(18,517,2,1).
app_href(18,518,725,1).
app_href(18,518,2,1).
app_href(18,190,316,1).
app_href(18,190,736,1).
app_href(18,190,1083,1).
app_href(27,87,444,1).
app_href(27,87,358,1).
app_href(27,557,1027,1).
app_href(27,557,546,1).
app_href(27,35,365,1).
app_href(27,35,6,1).
app_href(27,15,899,1).
app_href(27,15,1310,1).
app_href(27,15,668,1).
app_href(27,190,316,1).
app_href(27,190,736,1).
app_href(27,190,1083,1).
app_href(29,558,536,1).
app_href(29,558,1042,1).
app_href(29,558,1424,1).
app_href(29,558,115,1).
app_href(29,558,723,1).
app_href(29,558,1447,1).
app_href(29,558,769,1).
app_href(29,558,1346,1).
app_href(29,558,259,1).
app_href(29,558,916,1).
app_href(29,558,537,1).
app_href(29,558,1248,1).
app_href(29,558,564,2).
app_href(29,558,1446,1).
app_href(29,558,1138,1).
app_href(29,558,597,1).
app_href(29,558,109,1).
app_href(29,558,1087,1).
app_href(29,147,736,1).
app_href(29,147,941,1).
app_href(29,147,2,1).
app_href(29,147,1360,1).
app_href(29,148,1486,1).
app_href(29,148,1014,1).
app_href(29,149,830,1).
app_href(29,559,1452,1).
app_href(29,559,1091,1).
app_href(29,559,1383,1).
app_href(29,559,1237,1).
app_href(29,559,628,2).
app_href(29,559,916,2).
app_href(29,559,576,2).
app_href(29,559,439,2).
app_href(29,559,1431,1).
app_href(29,559,1443,1).
app_href(29,559,607,1).
app_href(29,559,1432,1).
app_href(29,559,288,1).
app_href(29,559,586,1).
app_href(29,559,1289,1).
app_href(29,559,365,1).
app_href(29,559,285,1).
app_href(29,559,742,1).
app_href(29,559,1176,1).
app_href(29,559,475,1).
app_href(29,559,960,1).
app_href(29,559,641,1).
app_href(29,559,474,1).
app_href(29,559,1276,1).
app_href(29,559,473,1).
app_href(29,559,550,1).
app_href(29,559,472,1).
app_href(29,559,1246,1).
app_href(29,559,1456,1).
app_href(29,559,1346,3).
app_href(29,559,667,1).
app_href(29,559,974,1).
app_href(29,559,148,1).
app_href(13,358,934,1).
app_href(13,359,1011,1).
app_href(13,359,1260,1).
app_href(13,359,736,1).
app_href(13,359,1198,1).
app_href(13,360,1014,1).
app_href(13,360,1198,1).
app_href(13,360,634,1).
app_href(13,405,1248,1).
app_href(13,405,564,1).
app_href(13,361,1011,1).
app_href(13,361,763,1).
app_href(13,361,1471,1).
app_href(13,361,186,1).
app_href(13,361,1198,1).
app_href(13,362,1011,1).
app_href(13,362,763,1).
app_href(13,362,186,1).
app_href(13,362,1198,1).
app_href(13,363,736,1).
app_href(13,363,1384,1).
app_href(13,364,1011,1).
app_href(13,364,763,1).
app_href(13,364,934,1).
app_href(13,365,145,1).
app_href(13,365,564,1).
app_href(13,365,1079,1).
app_href(13,365,230,1).
app_href(13,366,773,1).
app_href(13,366,1011,1).
app_href(13,366,309,1).
app_href(13,366,1310,1).
app_href(13,366,1198,1).
app_href(13,367,616,1).
app_href(13,367,420,1).
app_href(13,367,856,1).
app_href(13,367,730,1).
app_href(13,368,661,1).
app_href(13,368,736,1).
app_href(13,368,379,1).
app_href(13,368,564,1).
app_href(13,368,1346,1).
app_href(13,368,1102,1).
app_href(13,368,1360,1).
app_href(13,368,641,1).
app_href(13,368,530,1).
app_href(13,369,1450,1).
app_href(13,369,1017,1).
app_href(13,369,730,1).
app_href(13,370,1309,1).
app_href(13,370,1364,1).
app_href(13,370,1198,1).
app_href(13,371,839,1).
app_href(13,371,1159,1).
app_href(13,371,268,1).
app_href(13,371,89,1).
app_href(13,371,269,1).
app_href(13,371,140,1).
app_href(13,372,1011,1).
app_href(13,372,763,1).
app_href(13,372,186,1).
app_href(13,372,1198,1).
app_href(13,372,1267,1).
app_href(13,373,1011,1).
app_href(13,373,736,1).
app_href(13,373,1198,1).
app_href(13,373,702,1).
app_href(13,374,1011,1).
app_href(13,374,309,1).
app_href(13,374,1310,1).
app_href(13,374,186,1).
app_href(13,374,1198,1).
app_href(13,479,616,2).
app_href(13,479,1123,2).
app_href(13,479,791,2).
app_href(13,375,160,1).
app_href(13,375,379,1).
app_href(13,419,1248,1).
app_href(13,419,546,1).
app_href(13,419,269,1).
app_href(13,419,256,1).
app_href(13,419,1346,1).
app_href(13,419,564,1).
app_href(13,420,1456,1).
app_href(13,420,564,1).
app_href(13,9,1364,1).
app_href(13,9,1198,1).
app_href(13,484,616,1).
app_href(13,484,1123,1).
app_href(13,484,730,1).
app_href(21,376,55,1).
app_href(21,377,66,1).
app_href(21,291,320,1).
app_href(21,291,1276,1).
app_href(21,291,736,1).
app_href(21,291,546,1).
app_href(21,291,288,1).
app_href(21,291,140,1).
app_href(21,291,1443,1).
app_href(21,291,959,1).
app_href(21,291,277,1).
app_href(21,291,1138,1).
app_href(21,291,568,1).
app_href(21,378,53,1).
app_href(21,182,1218,1).
app_href(21,182,420,1).
app_href(21,308,591,1).
app_href(21,308,401,1).
app_href(21,379,1083,1).
app_href(21,379,736,1).
app_href(21,379,1264,1).
app_href(21,379,145,1).
app_href(21,379,564,1).
app_href(21,379,41,1).
app_href(21,379,218,1).
app_href(21,379,9,2).
app_href(21,379,316,1).
app_href(21,379,1157,1).
app_href(21,186,728,1).
app_href(21,186,454,1).
app_href(21,9,1364,1).
app_href(21,9,1198,1).
app_href(21,380,75,1).
app_href(21,15,899,1).
app_href(21,15,1310,1).
app_href(21,15,668,1).
app_href(21,181,354,1).
app_href(21,470,216,1).
app_href(21,470,276,1).
app_href(21,470,771,1).
app_href(21,470,564,1).
app_href(21,470,1346,1).
app_href(21,470,324,1).
app_href(21,470,526,1).
app_href(21,470,622,1).
app_href(21,470,665,1).
app_href(21,20,1443,1).
app_href(21,20,444,1).
app_href(21,20,736,1).
app_href(21,20,1276,1).
app_href(21,381,46,1).
app_href(21,382,73,1).
app_href(21,149,830,1).
app_href(21,168,1069,1).
app_href(21,167,1418,1).
app_href(21,167,423,1).
app_href(21,167,2,1).
app_href(21,711,476,1).
app_href(21,711,1104,1).
app_href(21,711,1310,1).
app_href(21,711,1162,1).
app_href(21,711,1379,1).
app_href(21,26,452,1).
app_href(21,26,1014,1).
app_href(21,152,1370,3).
app_href(21,152,992,1).
app_href(21,152,1133,1).
app_href(22,502,1123,1).
app_href(22,502,292,1).
app_href(22,502,686,1).
app_href(22,502,145,1).
app_href(22,502,894,1).
app_href(22,87,444,1).
app_href(22,87,358,1).
app_href(22,503,292,1).
app_href(22,503,686,1).
app_href(22,503,145,1).
app_href(22,503,1384,1).
app_href(22,503,894,1).
app_href(22,15,899,1).
app_href(22,15,1310,1).
app_href(22,15,668,1).
app_href(22,504,1123,1).
app_href(22,504,686,1).
app_href(22,504,497,1).
app_href(22,504,1267,1).
app_href(22,504,894,1).
app_href(22,505,1264,1).
app_href(25,506,242,1).
app_href(25,506,393,1).
app_href(25,506,2,1).
app_href(25,148,1486,1).
app_href(25,148,1014,1).
app_href(25,149,830,1).
app_href(25,15,899,1).
app_href(25,15,1310,1).
app_href(25,15,668,1).
app_href(25,152,1370,3).
app_href(25,152,992,1).
app_href(25,152,1133,1).
app_href(25,161,755,1).
app_href(25,129,1473,1).
app_href(25,129,2,1).
app_href(25,507,118,1).
app_href(25,507,477,1).
app_href(25,507,2,1).
app_href(25,163,489,1).
app_href(25,508,726,1).
app_href(25,508,2,1).
app_href(25,164,454,1).
app_href(25,165,728,1).
app_href(25,165,2,1).
app_href(25,166,736,1).
app_href(25,166,231,1).
app_href(25,168,1069,1).
app_href(25,509,1443,1).
app_href(25,509,1218,1).
app_href(25,509,2,1).
app_href(25,170,1192,1).
app_href(25,170,132,1).
app_href(25,171,944,1).
app_href(25,171,673,1).
app_href(25,172,1076,1).
app_href(25,172,1053,1).
app_href(25,173,1336,1).
app_href(25,510,452,1).
app_href(25,510,642,1).
app_href(25,511,1183,1).
app_href(25,511,296,1).
app_href(25,511,2,1).
app_href(25,512,1429,1).
app_href(25,512,642,1).
app_href(25,175,146,1).
app_href(25,175,2,1).
app_href(25,177,1403,1).
app_href(25,177,1341,1).
app_href(25,177,970,1).
app_href(25,178,829,1).
app_href(25,513,1459,1).
app_href(25,513,2,1).
app_href(25,181,354,1).
app_href(25,182,1200,1).
app_href(25,182,1218,1).
app_href(25,182,420,1).
app_href(25,182,134,1).
app_href(25,182,2,1).
app_href(25,514,575,1).
app_href(25,514,2,1).
app_href(25,183,444,1).
app_href(25,183,1200,1).
app_href(25,183,134,1).
app_href(25,183,2,1).
app_href(25,515,999,1).
app_href(25,515,2,1).
app_href(25,516,1030,1).
app_href(25,516,452,1).
app_href(25,184,317,1).
app_href(25,184,677,1).
app_href(25,185,1442,1).
app_href(25,185,736,1).
app_href(25,185,140,1).
app_href(25,185,233,1).
app_href(25,52,1385,1).
app_href(25,52,2,1).
app_href(25,517,1318,1).
app_href(25,517,2,1).
app_href(25,518,725,1).
app_href(25,518,2,1).
app_href(25,190,316,1).
app_href(25,190,736,1).
app_href(25,190,1083,1).
app_href(25,519,452,1).
app_href(25,519,1061,1).
app_href(26,165,867,1).
app_href(26,165,1336,1).
app_href(26,165,728,2).
app_href(26,185,1442,1).
app_href(26,185,736,1).
app_href(26,185,140,1).
app_href(26,185,233,1).
app_href(26,168,1069,1).
app_href(26,187,444,1).
app_href(26,187,728,1).
app_href(26,182,1218,1).
app_href(26,182,420,1).
app_href(26,163,489,1).
app_href(26,162,1364,1).
app_href(26,162,2,1).
app_href(26,164,454,1).
app_href(26,184,317,1).
app_href(26,184,677,1).
app_href(26,170,1192,1).
app_href(26,170,132,1).
app_href(26,15,899,1).
app_href(26,15,1310,1).
app_href(26,15,668,1).
app_href(26,152,1370,3).
app_href(26,152,992,1).
app_href(26,152,1133,1).
app_href(26,183,444,1).
app_href(26,166,736,1).
app_href(26,166,231,1).
app_href(26,150,830,1).
app_href(26,150,1256,1).
app_href(26,150,2,1).
app_href(26,149,830,1).
app_href(26,173,1336,1).
app_href(26,148,1486,1).
app_href(26,148,1014,1).
app_href(26,190,316,1).
app_href(26,190,736,1).
app_href(26,190,1083,1).
app_href(26,181,354,1).
app_href(26,171,944,1).
app_href(26,171,673,1).
app_href(26,520,444,1).
app_href(26,520,41,1).
app_href(26,520,9,2).
app_href(26,520,728,1).
app_href(26,521,452,1).
app_href(26,521,629,1).
app_href(26,167,1418,1).
app_href(26,167,423,1).
app_href(26,167,2,1).
app_href(23,471,503,1).
app_href(23,15,899,1).
app_href(23,15,1310,1).
app_href(23,15,668,1).
app_href(24,625,146,1).
app_href(24,625,2,1).
app_href(24,227,736,2).
app_href(24,227,1041,1).
app_href(24,227,1274,1).
app_href(24,227,1443,1).
app_href(24,227,480,1).
app_href(24,227,911,1).
app_href(24,227,41,1).
app_href(24,227,1010,1).
app_href(24,227,1218,1).
app_href(24,227,218,1).
app_href(24,227,728,1).
app_href(24,227,26,2).
app_href(24,227,9,3).
app_href(24,227,420,1).
app_href(24,148,1486,1).
app_href(24,148,1014,1).
app_href(24,149,830,1).
app_href(24,626,736,1).
app_href(24,626,1384,1).
app_href(24,627,736,1).
app_href(24,627,9,1).
app_href(24,627,1469,1).
app_href(24,627,1412,1).
app_href(24,628,725,1).
app_href(24,628,2,1).
app_href(24,629,1183,1).
app_href(24,629,296,1).
app_href(24,629,2,1).
app_href(24,630,41,1).
app_href(24,630,26,1).
app_href(24,630,9,1).
app_href(24,630,1189,1).
app_href(24,630,728,1).
app_href(24,630,218,1).
app_href(24,631,736,1).
app_href(24,631,1433,1).
app_href(24,632,316,1).
app_href(24,632,736,1).
app_href(24,632,1346,1).
app_href(24,632,781,1).
app_href(24,633,516,1).
app_href(24,633,296,1).
app_href(24,633,1287,1).
app_href(24,15,899,1).
app_href(24,15,1310,1).
app_href(24,15,668,1).
app_href(24,634,736,1).
app_href(24,634,1364,1).
app_href(24,635,725,1).
app_href(24,635,41,1).
app_href(24,635,9,1).
app_href(24,635,218,1).
app_href(24,587,1462,1).
app_href(24,587,1337,1).
app_href(24,587,736,1).
app_href(24,587,710,1).
app_href(24,636,1443,1).
app_href(24,636,1218,1).
app_href(24,636,2,1).
app_href(24,152,1370,3).
app_href(24,152,992,1).
app_href(24,152,1133,1).
app_href(24,637,1318,1).
app_href(24,637,2,1).
app_href(24,161,755,1).
app_href(24,638,911,1).
app_href(24,639,155,1).
app_href(24,639,975,1).
app_href(24,639,1310,1).
app_href(24,639,1212,1).
app_href(24,640,821,1).
app_href(24,640,736,1).
app_href(24,163,489,1).
app_href(24,641,855,1).
app_href(24,642,1220,1).
app_href(24,168,1069,1).
app_href(24,643,1452,1).
app_href(24,643,1132,1).
app_href(24,644,575,1).
app_href(24,644,2,1).
app_href(24,645,444,1).
app_href(24,645,1200,1).
app_href(24,645,134,1).
app_href(24,645,2,1).
app_href(24,646,1385,1).
app_href(24,646,2,1).
app_href(24,647,726,1).
app_href(24,647,2,1).
app_href(24,648,1131,1).
app_href(24,170,1192,1).
app_href(24,170,132,1).
app_href(24,649,999,1).
app_href(24,649,2,1).
app_href(24,650,944,2).
app_href(24,650,673,2).
app_href(24,651,41,1).
app_href(24,651,9,2).
app_href(24,651,1189,1).
app_href(24,651,728,1).
app_href(24,652,728,1).
app_href(24,652,2,1).
app_href(24,653,613,1).
app_href(24,653,1406,1).
app_href(24,653,736,1).
app_href(24,172,1076,1).
app_href(24,172,1053,1).
app_href(24,654,975,1).
app_href(24,654,1310,1).
app_href(24,654,1300,1).
app_href(24,655,736,1).
app_href(24,655,841,1).
app_href(24,656,1336,1).
app_href(24,657,306,1).
app_href(24,657,1346,1).
app_href(24,657,564,1).
app_href(24,657,1066,1).
app_href(24,657,319,1).
app_href(24,657,301,1).
app_href(24,658,1305,1).
app_href(24,658,947,1).
app_href(24,658,1360,1).
app_href(24,659,454,1).
app_href(24,660,1473,1).
app_href(24,660,2,1).
app_href(24,177,1403,1).
app_href(24,177,1341,1).
app_href(24,177,970,1).
app_href(24,661,118,1).
app_href(24,661,477,1).
app_href(24,661,2,1).
app_href(24,662,242,1).
app_href(24,662,393,1).
app_href(24,662,2,1).
app_href(24,178,829,1).
app_href(24,663,1459,1).
app_href(24,663,2,1).
app_href(24,181,354,1).
app_href(24,664,499,1).
app_href(24,664,150,1).
app_href(24,664,885,1).
app_href(24,664,1460,1).
app_href(24,26,41,2).
app_href(24,26,146,1).
app_href(24,26,9,4).
app_href(24,26,728,2).
app_href(24,185,1442,1).
app_href(24,185,736,1).
app_href(24,185,140,1).
app_href(24,185,233,1).
app_href(24,184,317,1).
app_href(24,184,677,1).
app_href(20,287,1276,1).
app_href(20,287,1460,1).
app_href(20,288,1117,1).
app_href(20,289,420,1).
app_href(20,289,1195,1).
app_href(20,290,667,1).
app_href(20,290,1443,1).
app_href(20,290,161,1).
app_href(20,290,780,1).
app_href(20,290,156,1).
app_href(20,290,1013,1).
app_href(20,35,365,1).
app_href(20,35,6,1).
app_href(20,291,1138,1).
app_href(20,291,959,1).
app_href(20,292,848,1).
app_href(20,292,445,1).
app_href(20,87,444,1).
app_href(20,87,358,1).
app_href(20,293,616,1).
app_href(20,293,420,1).
app_href(20,294,543,1).
app_href(20,294,566,1).
app_href(20,15,899,1).
app_href(20,15,1310,1).
app_href(20,15,668,1).
app_href(20,295,1264,1).
app_href(20,295,145,1).
app_href(20,295,780,1).
app_href(28,296,736,1).
app_href(28,296,1346,1).
app_href(28,296,1014,1).
app_href(28,296,1255,1).
app_href(28,296,634,1).
app_href(28,297,1011,1).
app_href(28,297,763,1).
app_href(28,297,900,1).
app_href(28,297,1255,1).
app_href(28,298,616,1).
app_href(28,298,1092,1).
app_href(28,298,145,1).
app_href(28,35,365,1).
app_href(28,299,736,1).
app_href(28,299,1014,1).
app_href(28,299,1255,1).
app_href(28,299,634,1).
app_href(28,87,444,1).
app_href(28,87,358,1).
app_href(28,300,1011,1).
app_href(28,300,763,1).
app_href(28,300,900,1).
app_href(28,300,1255,1).
app_href(28,300,1267,1).
app_href(28,301,1011,1).
app_href(28,301,763,1).
app_href(28,301,1182,1).
app_href(28,301,1255,1).
app_href(28,301,1267,1).
app_href(28,302,1030,1).
app_href(28,303,1011,1).
app_href(28,303,763,1).
app_href(28,303,1182,1).
app_href(28,303,1255,1).
app_href(28,15,899,1).
app_href(28,15,1310,1).
app_href(28,15,668,1).
app_href(30,304,782,1).
app_href(30,304,1059,1).
app_href(30,304,1439,1).
app_href(30,304,908,1).
app_href(30,304,1009,1).
app_href(30,305,1264,1).
app_href(30,306,399,2).
app_href(30,306,1430,1).
app_href(30,306,736,3).
app_href(30,306,400,2).
app_href(30,306,160,2).
app_href(30,306,864,1).
app_href(30,306,701,2).
app_href(30,306,1458,1).
app_href(30,306,1358,3).
app_href(30,306,1101,1).
app_href(30,306,1009,1).
app_href(30,306,702,1).
app_href(30,307,1264,1).
app_href(30,87,444,1).
app_href(30,87,358,1).
app_href(30,308,667,1).
app_href(30,308,41,1).
app_href(30,308,26,1).
app_href(30,308,736,1).
app_href(30,308,1346,1).
app_href(30,308,1189,1).
app_href(30,308,1412,1).
app_href(30,308,503,1).
app_href(30,309,331,1).
app_href(30,309,418,1).
app_href(30,309,140,1).
app_href(30,309,715,1).
app_href(30,310,1264,1).
app_href(30,15,899,1).
app_href(30,15,1310,1).
app_href(30,15,668,1).
app_href(30,311,1014,1).
app_href(30,311,782,1).
app_href(30,311,634,1).
app_href(32,312,444,1).
app_href(32,312,1333,1).
app_href(32,5,616,1).
app_href(32,5,540,1).
app_href(32,5,1360,1).
app_href(32,313,129,1).
app_href(32,313,485,1).
app_href(32,313,736,1).
app_href(32,313,1246,1).
app_href(32,313,837,1).
app_href(32,35,365,1).
app_href(32,35,6,1).
app_href(32,36,1144,1).
app_href(32,36,242,1).
app_href(32,36,566,1).
app_href(32,314,911,1).
app_href(32,314,1335,1).
app_href(32,314,1155,1).
app_href(32,87,444,1).
app_href(32,87,358,1).
app_href(32,315,543,1).
app_href(32,315,1443,1).
app_href(32,315,1237,1).
app_href(32,315,566,1).
app_href(32,15,899,1).
app_href(32,15,1310,1).
app_href(32,15,668,1).
app_href(33,485,540,1).
app_href(33,485,941,1).
app_href(33,485,140,1).
app_href(33,485,503,1).
app_href(33,358,1011,1).
app_href(33,358,763,1).
app_href(33,358,1270,1).
app_href(33,358,934,1).
app_href(33,364,1011,1).
app_href(33,364,763,1).
app_href(33,364,934,1).
app_href(33,35,365,1).
app_href(33,35,6,1).
app_href(33,316,1009,1).
app_href(33,316,939,1).
app_href(33,317,1098,1).
app_href(33,317,939,1).
app_href(33,318,1243,1).
app_href(33,318,991,1).
app_href(33,319,1138,1).
app_href(33,319,959,1).
app_href(33,319,939,1).
app_href(33,15,899,1).
app_href(33,15,1310,1).
app_href(33,15,668,1).
app_href(33,320,1011,1).
app_href(33,320,736,1).
app_href(33,320,934,1).
app_href(33,87,444,1).
app_href(33,87,358,1).
app_href(33,321,720,1).
app_href(33,321,1318,1).
app_href(33,698,1030,1).
app_href(33,699,277,1).
app_href(33,699,140,1).
app_href(33,699,503,1).
app_href(33,700,1030,1).
app_href(33,701,1243,1).
app_href(33,701,519,1).
app_href(33,702,1243,1).
app_href(33,702,529,1).
app_href(33,702,140,1).
app_href(33,702,1215,1).
app_href(33,703,902,1).
app_href(34,704,1030,1).
app_href(34,406,505,1).
app_href(34,406,1358,1).
app_href(34,406,1248,1).
app_href(34,406,1346,1).
app_href(34,406,985,1).
app_href(34,406,564,1).
app_href(34,705,1405,1).
app_href(34,705,1098,1).
app_href(34,705,736,1).
app_href(34,705,1162,1).
app_href(34,414,897,1).
app_href(34,414,1456,1).
app_href(34,414,564,1).
app_href(34,35,365,1).
app_href(34,35,6,1).
app_href(34,706,296,1).
app_href(34,706,736,1).
app_href(34,706,467,1).
app_href(34,706,1162,2).
app_href(34,706,1212,1).
app_href(34,706,453,1).
app_href(34,706,298,1).
app_href(34,706,1346,1).
app_href(34,706,1099,1).
app_href(34,706,1408,2).
app_href(34,126,505,1).
app_href(34,126,1248,1).
app_href(34,126,1346,1).
app_href(34,126,985,1).
app_href(34,126,564,1).
app_href(34,707,1011,1).
app_href(34,707,1134,1).
app_href(34,707,736,1).
app_href(34,707,1162,1).
app_href(34,140,897,1).
app_href(34,140,1456,1).
app_href(34,140,564,1).
app_href(34,708,1138,1).
app_href(34,708,959,1).
app_href(34,708,1162,1).
app_href(34,145,897,1).
app_href(34,145,1456,1).
app_href(34,145,564,1).
app_href(34,709,1270,1).
app_href(34,709,1162,1).
app_href(34,709,1408,1).
app_href(34,709,1460,1).
app_href(34,133,505,1).
app_href(34,133,1248,1).
app_href(34,133,1346,1).
app_href(34,133,985,1).
app_href(34,133,564,1).
app_href(34,15,899,1).
app_href(34,15,1310,1).
app_href(34,15,668,1).
app_href(34,87,444,1).
app_href(34,87,358,1).
app_href(34,710,1358,1).
app_href(34,710,1371,1).
app_href(34,710,736,1).
app_href(34,710,1346,2).
app_href(34,710,1162,1).
app_href(34,710,359,1).
app_href(34,710,1384,1).
app_href(34,710,970,1).
app_href(34,711,1405,1).
app_href(34,711,1098,1).
app_href(34,711,1310,1).
app_href(34,711,1162,1).
app_href(34,712,1170,1).
app_href(34,712,162,1).
app_href(34,712,1192,1).
app_href(34,712,1346,2).
app_href(34,712,837,1).
app_href(34,712,970,1).
app_href(34,713,1207,1).
app_href(34,713,536,1).
app_href(34,713,967,1).
app_href(34,713,1346,1).
app_href(34,713,140,1).
app_href(34,713,1162,1).
app_href(34,714,1170,1).
app_href(34,714,1305,1).
app_href(34,714,1346,1).
app_href(34,714,837,1).
app_href(34,714,1360,1).
app_href(34,715,534,1).
app_href(34,715,1346,1).
app_href(34,715,1028,1).
app_href(34,715,1162,1).
app_href(34,715,157,1).
app_href(34,715,1360,1).
app_href(34,716,1011,1).
app_href(34,716,736,1).
app_href(34,716,123,1).
app_href(34,716,1162,1).
app_href(34,717,1011,1).
app_href(34,717,736,1).
app_href(34,717,1162,1).
app_href(34,718,736,1).
app_href(34,718,1346,1).
app_href(34,718,1014,1).
app_href(34,718,1141,1).
app_href(34,718,1162,1).
app_href(34,718,715,1).
app_href(34,718,634,1).
app_href(34,718,894,1).
app_href(34,719,1011,1).
app_href(34,719,736,1).
app_href(34,719,1212,1).
app_href(34,719,1162,1).
app_href(34,720,1277,1).
app_href(34,720,602,1).
app_href(34,721,1264,1).
app_href(34,721,1162,1).
app_href(34,722,1385,1).
app_href(34,722,1346,1).
app_href(34,722,123,1).
app_href(34,723,1243,1).
app_href(34,723,534,1).
app_href(34,723,1346,1).
app_href(34,723,385,1).
app_href(34,723,1384,1).
app_href(34,724,1061,1).
app_href(34,725,1011,1).
app_href(34,725,425,1).
app_href(34,725,736,1).
app_href(34,725,1162,1).
app_href(35,87,444,1).
app_href(35,87,358,1).
app_href(35,726,1264,1).
app_href(35,726,829,1).
app_href(35,74,444,1).
app_href(35,74,1053,1).
app_href(35,74,1121,1).
app_href(35,35,365,1).
app_href(26,175,146,1).
app_href(26,175,2,1).
app_href(26,161,755,1).
app_href(26,177,1403,1).
app_href(26,177,1341,1).
app_href(26,177,970,1).
app_href(26,186,728,1).
app_href(26,186,454,1).
app_href(26,154,2,1).
app_href(26,172,1076,1).
app_href(26,172,1053,1).
app_href(26,147,736,1).
app_href(26,147,941,1).
app_href(26,147,2,1).
app_href(26,147,1360,1).
app_href(26,178,829,1).
app_href(45,522,536,1).
app_href(45,522,1337,1).
app_href(45,522,736,1).
app_href(45,522,639,1).
app_href(45,522,564,1).
app_href(45,16,701,1).
app_href(45,16,736,1).
app_href(45,16,927,1).
app_href(45,523,1003,1).
app_href(45,523,1181,1).
app_href(45,524,274,1).
app_href(45,524,1229,1).
app_href(45,119,836,1).
app_href(45,137,1250,1).
app_href(45,137,1483,1).
app_href(45,35,365,1).
app_href(45,35,6,1).
app_href(45,400,898,1).
app_href(45,525,991,2).
app_href(45,525,736,2).
app_href(45,525,1433,2).
app_href(45,525,639,2).
app_href(45,526,1450,1).
app_href(45,526,736,1).
app_href(45,526,1335,1).
app_href(45,526,864,1).
app_href(45,526,639,1).
app_href(45,526,1430,1).
app_href(45,526,1017,1).
app_href(45,26,452,1).
app_href(45,26,1014,1).
app_href(45,527,444,1).
app_href(45,527,1033,1).
app_href(45,87,444,1).
app_href(45,87,358,1).
app_href(45,528,1196,1).
app_href(45,528,639,1).
app_href(45,528,736,1).
app_href(45,528,1433,1).
app_href(45,528,1140,1).
app_href(45,528,1212,1).
app_href(45,528,564,2).
app_href(45,528,1346,3).
app_href(45,528,1251,1).
app_href(45,528,1033,1).
app_href(45,528,504,1).
app_href(45,529,508,1).
app_href(45,529,883,1).
app_href(45,15,899,1).
app_href(45,15,1310,1).
app_href(45,15,668,1).
app_href(47,731,1011,1).
app_href(47,731,1126,1).
app_href(47,731,736,1).
app_href(47,731,809,1).
app_href(47,468,1126,1).
app_href(47,468,736,1).
app_href(47,468,544,1).
app_href(47,468,809,1).
app_href(47,35,365,1).
app_href(47,87,444,1).
app_href(47,87,358,1).
app_href(47,530,466,1).
app_href(47,530,289,1).
app_href(47,530,1104,1).
app_href(47,467,942,1).
app_href(47,467,288,1).
app_href(47,474,1126,1).
app_href(47,474,942,1).
app_href(47,474,736,1).
app_href(47,474,350,1).
app_href(47,474,809,1).
app_href(47,15,899,1).
app_href(47,15,1310,1).
app_href(47,15,668,1).
app_href(47,481,1126,1).
app_href(47,481,716,1).
app_href(47,481,736,1).
app_href(47,481,809,1).
app_href(47,481,894,1).
app_href(47,531,118,1).
app_href(47,531,360,1).
app_href(47,531,1321,1).
app_href(47,531,1129,1).
app_href(47,532,9,2).
app_href(47,532,1127,1).
app_href(47,532,1009,1).
app_href(63,533,1123,1).
app_href(63,533,736,1).
app_href(63,533,1122,1).
app_href(63,533,809,1).
app_href(63,35,365,1).
app_href(63,35,6,1).
app_href(63,841,736,1).
app_href(63,841,1122,1).
app_href(63,841,809,1).
app_href(63,534,1245,1).
app_href(63,534,736,1).
app_href(63,534,1122,1).
app_href(63,534,809,1).
app_href(63,87,444,1).
app_href(63,87,358,1).
app_href(63,535,736,1).
app_href(63,535,1122,1).
app_href(63,535,809,1).
app_href(63,535,1441,1).
app_href(63,15,899,1).
app_href(63,15,1310,1).
app_href(63,15,668,1).
app_href(63,536,736,1).
app_href(63,536,1124,1).
app_href(63,536,1122,1).
app_href(63,536,809,1).
app_href(63,537,1037,1).
app_href(63,537,736,1).
app_href(63,537,1122,1).
app_href(63,537,809,1).
app_href(64,87,444,1).
app_href(64,87,358,1).
app_href(64,538,1331,1).
app_href(64,538,410,1).
app_href(64,538,1346,1).
app_href(64,35,365,1).
app_href(64,539,1331,1).
app_href(64,539,151,1).
app_href(64,15,899,1).
app_href(64,15,1310,1).
app_href(64,15,668,1).
app_href(65,540,642,1).
app_href(65,1158,581,1).
app_href(65,1158,1264,1).
app_href(65,1158,788,2).
app_href(65,1158,657,1).
app_href(65,846,763,1).
app_href(65,846,658,1).
app_href(65,1159,1030,1).
app_href(65,1160,660,1).
app_href(65,1160,736,1).
app_href(65,1160,132,1).
app_href(65,35,365,1).
app_href(65,35,6,1).
app_href(29,559,514,1).
app_href(29,559,147,1).
app_href(29,559,1447,1).
app_href(29,559,970,3).
app_href(29,559,151,2).
app_href(29,559,1253,1).
app_href(29,559,736,1).
app_href(29,559,955,1).
app_href(29,559,1355,1).
app_href(29,559,98,1).
app_href(29,559,952,1).
app_href(29,559,1313,1).
app_href(29,559,962,1).
app_href(29,559,597,1).
app_href(29,559,623,1).
app_href(29,559,1424,1).
app_href(29,559,154,1).
app_href(29,559,1446,1).
app_href(29,559,769,1).
app_href(29,559,1360,2).
app_href(29,560,1288,1).
app_href(29,560,192,1).
app_href(29,560,1338,1).
app_href(29,560,130,1).
app_href(29,560,115,1).
app_href(29,560,203,1).
app_href(29,560,202,1).
app_href(29,560,164,1).
app_href(29,560,1246,2).
app_href(29,560,667,1).
app_href(29,560,1346,1).
app_href(29,560,973,1).
app_href(29,560,259,1).
app_href(29,560,181,2).
app_href(29,560,736,4).
app_href(29,560,877,1).
app_href(29,560,485,1).
app_href(29,560,1337,1).
app_href(29,560,941,2).
app_href(29,560,641,1).
app_href(29,560,1446,4).
app_href(29,560,262,1).
app_href(29,560,713,1).
app_href(29,560,678,1).
app_href(29,560,391,1).
app_href(29,560,945,1).
app_href(29,150,830,1).
app_href(29,150,1256,1).
app_href(29,150,2,1).
app_href(29,561,736,1).
app_href(29,561,597,1).
app_href(29,561,811,1).
app_href(29,561,637,1).
app_href(29,561,1346,1).
app_href(29,561,1277,1).
app_href(29,561,667,1).
app_href(29,561,1093,2).
app_href(29,561,641,1).
app_href(29,561,112,1).
app_href(29,561,259,1).
app_href(29,61,1094,1).
app_href(29,61,955,1).
app_href(29,61,1452,1).
app_href(29,61,1310,1).
app_href(29,61,1012,1).
app_href(29,61,1345,1).
app_href(29,61,1227,1).
app_href(29,61,421,1).
app_href(29,61,1466,1).
app_href(29,61,257,1).
app_href(29,15,899,1).
app_href(29,15,1310,1).
app_href(29,15,668,1).
app_href(29,152,1370,3).
app_href(29,152,992,1).
app_href(29,152,1133,1).
app_href(29,154,2,1).
app_href(29,161,755,1).
app_href(29,24,192,1).
app_href(29,24,1310,2).
app_href(29,24,207,1).
app_href(29,24,206,1).
app_href(29,24,212,1).
app_href(29,24,136,1).
app_href(29,24,309,2).
app_href(29,24,164,1).
app_href(29,24,992,1).
app_href(29,24,1342,1).
app_href(29,24,259,1).
app_href(29,24,372,1).
app_href(29,24,1071,1).
app_href(29,24,181,1).
app_href(29,24,736,3).
app_href(29,24,139,1).
app_href(29,24,810,1).
app_href(29,24,1093,1).
app_href(29,24,643,2).
app_href(29,24,1446,1).
app_href(29,24,218,1).
app_href(29,24,1449,1).
app_href(29,162,1364,1).
app_href(29,162,2,1).
app_href(29,163,489,1).
app_href(29,562,806,1).
app_href(29,562,312,1).
app_href(29,562,1045,1).
app_href(29,562,1338,1).
app_href(29,562,317,1).
app_href(29,562,1476,1).
app_href(29,562,1246,1).
app_href(29,562,1401,1).
app_href(29,562,1482,1).
app_href(29,562,259,1).
app_href(29,562,973,1).
app_href(29,562,736,2).
app_href(29,562,1408,1).
app_href(29,562,824,1).
app_href(29,562,830,1).
app_href(29,562,564,1).
app_href(29,562,355,1).
app_href(29,562,1040,1).
app_href(29,562,390,1).
app_href(29,562,140,1).
app_href(29,563,1337,2).
app_href(29,563,452,1).
app_href(29,563,1247,1).
app_href(29,563,689,1).
app_href(29,563,736,3).
app_href(29,563,439,2).
app_href(29,563,992,1).
app_href(29,563,1044,1).
app_href(29,563,1320,1).
app_href(29,563,140,1).
app_href(29,563,1189,1).
app_href(29,563,723,1).
app_href(29,563,26,1).
app_href(29,563,641,2).
app_href(29,563,110,1).
app_href(29,564,113,1).
app_href(29,564,112,1).
app_href(29,564,111,1).
app_href(29,564,154,1).
app_href(29,564,858,1).
app_href(29,564,110,1).
app_href(29,564,937,1).
app_href(29,564,109,1).
app_href(29,564,108,1).
app_href(29,564,452,1).
app_href(29,564,107,1).
app_href(29,564,101,2).
app_href(29,564,936,4).
app_href(29,564,1345,1).
app_href(29,564,1466,1).
app_href(29,564,737,1).
app_href(29,564,257,1).
app_href(29,564,372,1).
app_href(29,564,1087,1).
app_href(29,564,1448,6).
app_href(29,564,629,1).
app_href(29,564,597,7).
app_href(29,564,1476,4).
app_href(29,564,1337,10).
app_href(29,564,1277,3).
app_href(29,564,164,6).
app_href(29,564,514,1).
app_href(29,564,444,2).
app_href(29,564,289,1).
app_href(29,564,624,1).
app_href(29,564,1273,2).
app_href(29,564,1308,1).
app_href(29,564,1426,2).
app_href(29,564,218,1).
app_href(29,564,952,1).
app_href(29,564,582,2).
app_href(29,564,1419,1).
app_href(29,564,537,1).
app_href(29,564,722,2).
app_href(29,564,565,2).
app_href(29,564,1176,3).
app_href(29,564,1145,1).
app_href(29,564,721,6).
app_href(29,564,151,2).
app_href(29,564,130,1).
app_href(29,564,667,8).
app_href(29,564,1040,2).
app_href(29,564,98,1).
app_href(29,564,1346,15).
app_href(29,564,192,6).
app_href(29,564,678,7).
app_href(29,564,145,1).
app_href(29,564,569,1).
app_href(29,564,317,1).
app_href(29,564,1342,1).
app_href(29,564,1408,2).
app_href(29,564,799,2).
app_href(29,564,1454,2).
app_href(29,564,689,1).
app_href(29,564,1212,1).
app_href(29,564,1253,1).
app_href(29,564,628,2).
app_href(29,564,1360,3).
app_href(29,564,426,2).
app_href(29,564,1204,1).
app_href(29,564,623,1).
app_href(29,564,136,1).
app_href(29,564,242,1).
app_href(29,564,918,4).
app_href(29,564,1256,1).
app_href(29,564,1447,2).
app_href(29,564,1198,1).
app_href(29,564,607,1).
app_href(29,564,1139,2).
app_href(29,564,1424,2).
app_href(29,564,144,1).
app_href(29,564,1383,1).
app_href(29,564,974,2).
app_href(29,564,153,3).
app_href(29,564,356,2).
app_href(29,564,587,2).
app_href(29,564,960,1).
app_href(29,564,140,7).
app_href(29,564,866,1).
app_href(29,564,259,18).
app_href(29,564,539,1).
app_href(29,564,1039,2).
app_href(29,564,382,2).
app_href(29,564,553,3).
app_href(29,564,285,1).
app_href(29,564,1368,2).
app_href(29,564,1247,1).
app_href(29,564,288,1).
app_href(29,564,1285,2).
app_href(29,564,106,2).
app_href(29,564,105,1).
app_href(29,564,104,1).
app_href(29,564,103,1).
app_href(29,564,262,1).
app_href(29,564,1382,2).
app_href(29,564,102,1).
app_href(29,564,550,1).
app_href(29,564,662,1).
app_href(29,564,602,4).
app_href(29,564,945,1).
app_href(29,564,148,1).
app_href(29,564,576,2).
app_href(29,564,1432,1).
app_href(29,564,713,1).
app_href(29,564,1333,2).
app_href(29,564,814,2).
app_href(29,564,1289,1).
app_href(29,564,831,2).
app_href(29,564,134,2).
app_href(29,564,1246,6).
app_href(29,564,785,1).
app_href(29,564,1456,1).
app_href(29,564,1355,1).
app_href(29,564,877,1).
app_href(29,564,115,5).
app_href(29,564,832,2).
app_href(29,564,1261,1).
app_href(29,564,424,2).
app_href(29,564,911,1).
app_href(29,564,439,9).
app_href(29,564,312,1).
app_href(29,564,1135,1).
app_href(29,564,546,3).
app_href(29,564,191,1).
app_href(29,564,1478,1).
app_href(29,564,769,4).
app_href(29,564,1431,1).
app_href(29,564,1460,6).
app_href(29,564,1452,3).
app_href(29,564,446,2).
app_href(29,564,1446,16).
app_href(29,564,475,1).
app_href(29,564,626,1).
app_href(29,564,474,1).
app_href(29,564,473,1).
app_href(29,564,1482,1).
app_href(29,564,1310,3).
app_href(29,564,472,1).
app_href(29,564,471,2).
app_href(29,564,1347,2).
app_href(29,564,1248,1).
app_href(29,564,1451,1).
app_href(29,564,955,2).
app_href(29,564,1001,1).
app_href(29,564,593,1).
app_href(29,564,992,5).
app_href(29,564,1123,1).
app_href(29,564,1091,1).
app_href(29,564,878,2).
app_href(29,564,1071,1).
app_href(29,564,973,5).
app_href(29,564,638,1).
app_href(29,564,970,8).
app_href(29,564,251,2).
app_href(29,564,641,8).
app_href(29,564,365,1).
app_href(29,564,147,1).
app_href(29,564,1461,2).
app_href(29,564,391,1).
app_href(29,564,390,1).
app_href(29,564,1044,1).
app_href(29,564,389,1).
app_href(29,564,388,1).
app_href(29,564,387,2).
app_href(29,564,736,26).
app_href(29,564,564,3).
app_href(29,564,1313,1).
app_href(29,564,962,1).
app_href(29,564,1276,1).
app_href(29,564,586,1).
app_href(29,564,1227,2).
app_href(21,383,57,1).
app_href(21,319,1138,1).
app_href(21,319,959,1).
app_href(21,319,1162,1).
app_href(21,319,939,1).
app_href(21,384,1169,1).
app_href(21,384,736,1).
app_href(21,384,1212,1).
app_href(21,384,1162,1).
app_href(21,384,1222,1).
app_href(21,183,444,1).
app_href(21,187,444,1).
app_href(21,187,728,1).
app_href(21,177,1403,1).
app_href(21,177,1341,1).
app_href(21,177,970,1).
app_href(21,444,286,1).
app_href(21,444,560,1).
app_href(21,444,140,1).
app_href(21,444,792,1).
app_href(21,161,755,1).
app_href(21,385,1386,1).
app_href(21,385,643,1).
app_href(21,385,652,1).
app_href(21,385,1470,1).
app_href(21,385,456,1).
app_href(21,385,939,1).
app_href(21,385,706,1).
app_href(21,385,998,1).
app_href(21,386,616,1).
app_href(21,386,1016,1).
app_href(21,386,420,1).
app_href(21,386,736,1).
app_href(21,386,1047,1).
app_href(21,386,132,1).
app_href(21,173,1336,1).
app_href(21,178,829,1).
app_href(21,708,1138,1).
app_href(21,708,959,1).
app_href(21,708,277,1).
app_href(21,708,197,1).
app_href(21,708,1162,1).
app_href(21,387,63,1).
app_href(21,88,1138,1).
app_href(21,88,959,1).
app_href(21,88,277,1).
app_href(21,88,1323,1).
app_href(21,88,197,1).
app_href(21,388,1319,1).
app_href(21,388,1225,1).
app_href(21,388,491,1).
app_href(21,388,492,1).
app_href(21,388,281,1).
app_href(21,388,275,1).
app_href(21,389,56,1).
app_href(21,389,1198,1).
app_href(21,45,400,1).
app_href(21,45,1041,1).
app_href(21,45,160,1).
app_href(21,45,444,1).
app_href(21,45,736,2).
app_href(21,45,1433,1).
app_href(21,45,639,1).
app_href(21,150,830,1).
app_href(21,150,1256,1).
app_href(21,150,2,1).
app_href(21,390,1217,1).
app_href(21,390,444,1).
app_href(21,390,1346,1).
app_href(21,390,818,1).
app_href(21,390,1102,1).
app_href(21,767,60,1).
app_href(21,768,44,1).
app_href(21,171,944,1).
app_href(21,171,673,1).
app_href(21,154,2,1).
app_href(21,769,989,1).
app_href(21,769,409,1).
app_href(21,769,1012,1).
app_href(21,162,1364,1).
app_href(21,162,2,1).
app_href(21,172,1076,1).
app_href(21,172,1053,1).
app_href(21,770,62,1).
app_href(21,170,1192,1).
app_href(21,170,132,1).
app_href(21,771,1083,1).
app_href(21,771,736,1).
app_href(21,771,1264,1).
app_href(21,771,145,1).
app_href(21,771,564,1).
app_href(21,771,41,1).
app_href(21,771,1138,1).
app_href(21,771,9,2).
app_href(21,771,316,1).
app_href(21,771,1157,1).
app_href(21,771,260,1).
app_href(21,166,736,1).
app_href(21,166,231,1).
app_href(21,772,444,1).
app_href(21,772,777,1).
app_href(21,773,52,1).
app_href(21,774,184,1).
app_href(21,774,1261,1).
app_href(21,774,151,1).
app_href(21,148,1486,1).
app_href(21,148,1014,1).
app_href(21,62,1138,1).
app_href(21,62,959,1).
app_href(21,775,747,1).
app_href(21,775,277,1).
app_href(21,775,42,1).
app_href(21,184,317,1).
app_href(21,184,677,1).
app_href(21,776,151,1).
app_href(21,776,736,1).
app_href(21,776,1047,1).
app_href(21,776,967,1).
app_href(21,776,140,1).
app_href(21,776,488,1).
app_href(21,776,667,1).
app_href(21,776,1408,1).
app_href(21,776,1332,1).
app_href(21,190,316,1).
app_href(21,190,736,1).
app_href(21,190,1083,1).
app_href(21,777,427,1).
app_href(21,777,493,1).
app_href(21,777,151,1).
app_href(21,777,1426,1).
app_href(21,777,1047,1).
app_href(21,777,241,1).
app_href(21,777,967,1).
app_href(21,777,140,2).
app_href(21,777,488,1).
app_href(21,777,1460,1).
app_href(21,777,1332,1).
app_href(21,147,736,1).
app_href(21,147,941,1).
app_href(21,147,2,1).
app_href(21,147,1360,1).
app_href(21,778,1003,1).
app_href(21,778,7,2).
app_href(21,778,1414,1).
app_href(21,778,1259,1).
app_href(21,778,564,1).
app_href(21,778,1085,1).
app_href(21,778,1181,1).
app_href(21,778,1318,1).
app_href(21,778,1258,1).
app_href(21,419,1337,1).
app_href(21,419,269,1).
app_href(21,419,736,1).
app_href(21,419,1047,1).
app_href(21,419,546,1).
app_href(21,419,1346,1).
app_href(21,419,1360,1).
app_href(29,564,450,2).
app_href(29,564,212,1).
app_href(29,564,1093,3).
app_href(29,564,1339,2).
app_href(29,564,679,2).
app_href(29,564,4,2).
app_href(29,564,181,1).
app_href(29,564,830,4).
app_href(29,564,96,2).
app_href(29,564,1449,1).
app_href(29,564,310,2).
app_href(29,564,255,1).
app_href(29,564,254,2).
app_href(29,564,1403,3).
app_href(29,564,139,1).
app_href(29,564,916,3).
app_href(29,564,37,4).
app_href(29,564,1475,1).
app_href(29,564,1094,1).
app_href(29,564,1338,3).
app_href(29,564,297,1).
app_href(29,564,1188,1).
app_href(29,564,1288,1).
app_href(29,564,1042,1).
app_href(29,564,881,4).
app_href(29,564,208,1).
app_href(29,564,207,1).
app_href(29,564,206,1).
app_href(29,564,205,1).
app_href(29,564,204,1).
app_href(29,564,421,1).
app_href(29,564,203,1).
app_href(29,564,202,1).
app_href(29,564,1138,1).
app_href(29,564,201,2).
app_href(29,564,1477,1).
app_href(29,564,200,2).
app_href(29,564,487,1).
app_href(29,564,1012,2).
app_href(29,564,742,1).
app_href(29,564,810,2).
app_href(29,564,1081,1).
app_href(29,564,485,1).
app_href(29,564,931,1).
app_href(29,564,723,4).
app_href(29,564,449,1).
app_href(29,564,757,2).
app_href(29,564,114,1).
app_href(29,565,444,1).
app_href(29,565,26,1).
app_href(29,565,736,1).
app_href(29,565,1364,1).
app_href(29,164,454,1).
app_href(29,566,1144,1).
app_href(29,566,242,1).
app_href(29,566,936,1).
app_href(29,567,1337,1).
app_href(29,567,830,1).
app_href(29,567,192,1).
app_href(29,567,115,1).
app_href(29,567,1001,1).
app_href(29,567,439,1).
app_href(29,567,721,1).
app_href(29,567,553,1).
app_href(29,567,911,1).
app_href(29,567,737,1).
app_href(29,567,723,1).
app_href(29,567,641,1).
app_href(29,567,757,1).
app_href(29,567,1451,1).
app_href(29,567,102,1).
app_href(29,568,736,2).
app_href(29,568,597,1).
app_href(29,568,192,1).
app_href(29,568,130,2).
app_href(29,568,721,1).
app_href(29,568,486,2).
app_href(29,568,204,1).
app_href(29,568,569,1).
app_href(29,568,444,1).
app_href(29,568,667,1).
app_href(29,568,723,1).
app_href(29,568,1246,2).
app_href(29,568,1446,2).
app_href(29,568,641,1).
app_href(29,568,259,1).
app_href(29,569,736,1).
app_href(29,569,1246,1).
app_href(29,569,1014,1).
app_href(29,569,634,1).
app_href(29,165,1443,1).
app_href(29,165,1227,1).
app_href(29,165,1336,1).
app_href(29,165,1014,1).
app_href(29,165,721,1).
app_href(29,165,867,1).
app_href(29,165,1123,1).
app_href(29,165,992,1).
app_href(29,165,1246,3).
app_href(29,165,769,1).
app_href(29,165,1346,4).
app_href(29,165,728,2).
app_href(29,165,974,1).
app_href(29,165,1237,1).
app_href(29,165,736,2).
app_href(29,165,191,1).
app_href(29,165,1248,1).
app_href(29,165,970,1).
app_href(29,165,1403,1).
app_href(29,165,1446,1).
app_href(29,165,1138,1).
app_href(29,165,634,1).
app_href(29,165,785,1).
app_href(29,165,1087,1).
app_href(29,165,108,1).
app_href(29,165,26,2).
app_href(29,165,1044,1).
app_href(29,165,107,1).
app_href(29,165,140,1).
app_href(29,165,761,1).
app_href(29,165,1041,2).
app_href(29,165,662,1).
app_href(29,570,1123,1).
app_href(29,570,662,1).
app_href(29,570,26,1).
app_href(29,570,9,1).
app_href(29,570,784,1).
app_href(29,166,736,1).
app_href(29,166,231,1).
app_href(29,167,1418,1).
app_href(29,167,423,1).
app_href(29,167,2,1).
app_href(29,168,1069,1).
app_href(29,571,1123,1).
app_href(29,571,662,1).
app_href(29,571,26,1).
app_href(29,571,9,1).
app_href(29,571,769,1).
app_href(29,572,1138,1).
app_href(29,572,769,1).
app_href(29,572,1087,1).
app_href(29,573,1123,1).
app_href(29,573,662,1).
app_href(29,573,9,2).
app_href(29,573,784,1).
app_href(29,574,242,1).
app_href(29,574,546,1).
app_href(29,574,439,1).
app_href(29,574,992,1).
app_href(29,574,208,1).
app_href(29,574,1310,1).
app_href(29,574,1145,1).
app_href(29,574,641,1).
app_href(29,574,1218,1).
app_href(29,574,1308,1).
app_href(29,574,539,1).
app_href(29,574,482,1).
app_href(21,419,503,1).
app_href(21,419,256,1).
app_href(21,163,489,1).
app_href(21,779,80,1).
app_href(21,175,146,1).
app_href(21,175,2,1).
app_href(21,457,41,1).
app_href(21,457,9,1).
app_href(21,457,277,1).
app_href(21,457,8,1).
app_href(21,457,728,1).
app_href(21,457,1323,1).
app_href(21,457,704,1).
app_href(21,457,1235,1).
app_href(21,185,1442,1).
app_href(21,185,736,1).
app_href(21,185,140,1).
app_href(21,185,233,1).
app_href(21,493,736,1).
app_href(21,493,701,1).
app_href(21,493,140,1).
app_href(21,493,145,1).
app_href(21,493,429,1).
app_href(21,493,941,1).
app_href(21,493,747,1).
app_href(21,493,667,1).
app_href(21,493,41,1).
app_href(21,493,1446,1).
app_href(21,493,9,2).
app_href(21,493,922,1).
app_href(21,165,867,1).
app_href(21,165,1336,1).
app_href(21,165,728,2).
app_href(21,780,43,1).
app_href(21,781,918,1).
app_href(21,781,1430,1).
app_href(21,781,861,1).
app_href(21,781,315,1).
app_href(21,781,864,1).
app_href(21,781,140,1).
app_href(21,781,438,1).
app_href(21,781,1217,1).
app_href(21,781,762,1).
app_href(21,781,800,1).
app_href(21,782,65,1).
app_href(21,783,1142,1).
app_href(21,783,736,1).
app_href(21,783,1223,1).
app_href(21,783,1212,1).
app_href(21,783,1162,1).
app_href(21,784,59,1).
app_href(21,164,454,1).
app_href(21,785,736,1).
app_href(21,785,837,1).
app_href(21,785,922,1).
app_href(21,785,1162,1).
app_href(21,786,74,1).
app_href(21,787,736,1).
app_href(21,787,1167,1).
app_href(21,787,1212,1).
app_href(21,787,1162,1).
app_href(21,788,76,1).
app_href(21,169,1041,1).
app_href(21,169,1259,1).
app_href(21,169,747,1).
app_href(21,169,1225,1).
app_href(21,169,1258,1).
app_href(21,169,275,1).
app_href(21,169,320,1).
app_href(21,169,1217,1).
app_href(21,169,939,1).
app_href(21,169,145,1).
app_href(21,169,1337,1).
app_href(21,169,1162,1).
app_href(21,169,7,2).
app_href(21,169,503,1).
app_href(21,169,1443,1).
app_href(21,169,724,1).
app_href(21,169,1323,1).
app_href(21,169,160,1).
app_href(21,169,91,1).
app_href(21,169,288,1).
app_href(21,169,438,1).
app_href(21,169,1085,2).
app_href(21,169,546,1).
app_href(21,169,429,1).
app_href(21,169,277,4).
app_href(21,169,26,1).
app_href(21,169,800,1).
app_href(21,169,9,1).
app_href(21,169,1276,1).
app_href(21,169,704,1).
app_href(21,169,568,1).
app_href(21,169,711,1).
app_href(21,169,1430,1).
app_href(21,169,1139,1).
app_href(21,169,1047,1).
app_href(21,169,197,1).
app_href(21,169,1318,1).
app_href(21,169,736,4).
app_href(21,169,564,2).
app_href(21,169,420,1).
app_href(21,169,491,1).
app_href(21,169,862,1).
app_href(21,169,864,1).
app_href(21,169,140,3).
app_href(21,169,922,1).
app_href(21,169,400,1).
app_href(21,169,1198,1).
app_href(21,169,701,1).
app_href(21,169,698,1).
app_href(21,169,281,1).
app_href(21,169,1414,1).
app_href(21,169,1360,1).
app_href(38,789,280,1).
app_href(38,789,1181,1).
app_href(38,506,242,1).
app_href(38,506,393,1).
app_href(38,506,2,1).
app_href(38,148,1486,1).
app_href(38,148,1014,1).
app_href(38,149,830,1).
app_href(38,790,196,1).
app_href(38,791,1290,1).
app_href(38,791,1209,1).
app_href(38,791,736,1).
app_href(38,791,38,1).
app_href(38,791,430,1).
app_href(38,15,899,1).
app_href(38,15,1310,1).
app_href(38,15,668,1).
app_href(38,792,137,1).
app_href(38,792,309,1).
app_href(38,152,1370,3).
app_href(38,152,992,1).
app_href(38,152,1133,1).
app_href(38,161,755,1).
app_href(38,129,1473,1).
app_href(38,129,2,1).
app_href(38,163,489,1).
app_href(38,507,118,1).
app_href(38,507,477,1).
app_href(38,507,2,1).
app_href(38,508,726,1).
app_href(38,508,2,1).
app_href(38,164,454,1).
app_href(38,793,1158,1).
app_href(38,165,728,1).
app_href(38,165,2,1).
app_href(38,166,736,1).
app_href(38,166,231,1).
app_href(38,794,667,1).
app_href(38,794,1349,1).
app_href(38,794,452,1).
app_href(38,794,232,1).
app_href(38,794,1183,1).
app_href(38,794,1404,1).
app_href(35,35,6,1).
app_href(35,15,899,1).
app_href(35,15,1310,1).
app_href(35,15,668,1).
app_href(35,727,1100,1).
app_href(35,727,960,1).
app_href(35,727,1161,1).
app_href(37,347,763,1).
app_href(37,347,855,1).
app_href(37,347,186,1).
app_href(37,347,1198,1).
app_href(37,361,763,1).
app_href(37,361,1471,1).
app_href(37,361,186,1).
app_href(37,361,1198,1).
app_href(37,447,1392,1).
app_href(37,447,763,1).
app_href(37,447,1323,1).
app_href(37,717,736,1).
app_href(37,717,1162,1).
app_href(37,728,470,1).
app_href(37,728,972,1).
app_href(37,728,468,1).
app_href(37,728,998,1).
app_href(37,345,763,1).
app_href(37,345,800,1).
app_href(37,345,186,1).
app_href(37,345,1198,1).
app_href(37,729,972,1).
app_href(37,729,736,1).
app_href(37,729,1162,1).
app_href(37,729,1298,1).
app_href(37,729,998,1).
app_href(37,730,972,1).
app_href(37,730,370,1).
app_href(37,730,998,1).
app_href(37,91,736,1).
app_href(37,91,718,1).
app_href(37,731,1126,1).
app_href(37,731,736,1).
app_href(37,731,809,1).
app_href(37,732,736,2).
app_href(37,732,283,2).
app_href(37,732,950,2).
app_href(37,733,996,1).
app_href(37,733,972,1).
app_href(37,733,367,1).
app_href(37,733,998,1).
app_href(37,734,429,1).
app_href(37,734,763,1).
app_href(37,734,922,1).
app_href(37,735,656,1).
app_href(37,735,736,1).
app_href(37,735,197,1).
app_href(37,736,1409,1).
app_href(37,736,736,1).
app_href(37,736,733,1).
app_href(37,320,736,1).
app_href(37,320,934,1).
app_href(37,324,763,1).
app_href(37,324,1269,1).
app_href(37,446,763,1).
app_href(37,446,1038,1).
app_href(37,446,1323,1).
app_href(37,737,763,1).
app_href(37,737,1274,1).
app_href(37,737,1073,1).
app_href(37,737,183,1).
app_href(37,129,1473,4).
app_href(37,129,736,4).
app_href(37,129,145,4).
app_href(37,129,1084,4).
app_href(37,129,564,4).
app_href(37,738,1073,1).
app_href(37,738,736,1).
app_href(37,738,197,1).
app_href(37,362,763,1).
app_href(37,362,186,1).
app_href(37,362,1198,1).
app_href(37,739,763,1).
app_href(37,739,658,1).
app_href(37,739,512,1).
app_href(37,719,736,1).
app_href(37,719,1212,1).
app_href(37,719,1162,1).
app_href(37,740,736,1).
app_href(37,740,437,1).
app_href(37,740,197,1).
app_href(37,741,972,1).
app_href(37,741,468,1).
app_href(37,741,469,1).
app_href(37,741,998,1).
app_href(37,742,972,1).
app_href(37,742,1056,1).
app_href(37,742,367,1).
app_href(37,742,998,1).
app_href(37,743,972,1).
app_href(37,743,468,1).
app_href(37,743,998,1).
app_href(37,188,410,2).
app_href(37,297,763,1).
app_href(37,297,900,1).
app_href(37,297,1255,1).
app_href(37,744,444,1).
app_href(37,744,420,1).
app_href(37,744,26,1).
app_href(37,744,736,1).
app_href(37,744,1189,1).
app_href(37,744,1191,1).
app_href(37,373,736,1).
app_href(37,373,1198,1).
app_href(37,373,702,1).
app_href(37,343,763,1).
app_href(37,343,1260,1).
app_href(37,343,378,1).
app_href(37,745,972,1).
app_href(37,745,1057,1).
app_href(37,745,995,1).
app_href(37,745,998,1).
app_href(37,746,763,1).
app_href(37,746,1373,1).
app_href(37,746,1472,1).
app_href(37,747,1409,1).
app_href(37,747,736,1).
app_href(37,747,1479,1).
app_href(37,748,1473,1).
app_href(37,748,630,1).
app_href(37,748,1185,1).
app_href(37,748,998,1).
app_href(37,749,972,1).
app_href(37,749,1051,1).
app_href(37,749,1058,1).
app_href(37,749,998,1).
app_href(37,442,763,1).
app_href(37,442,210,1).
app_href(37,442,1038,1).
app_href(37,442,1323,1).
app_href(37,750,763,1).
app_href(37,750,1073,1).
app_href(37,751,972,1).
app_href(37,751,719,1).
app_href(37,751,998,1).
app_href(37,752,972,1).
app_href(37,752,1056,1).
app_href(37,752,1058,1).
app_href(37,752,998,1).
app_href(37,753,736,1).
app_href(37,753,950,1).
app_href(37,754,763,1).
app_href(37,754,119,1).
app_href(37,116,214,1).
app_href(37,116,736,1).
app_href(37,755,1473,1).
app_href(37,755,528,1).
app_href(37,755,1185,1).
app_href(37,755,998,1).
app_href(37,358,763,1).
app_href(29,170,1192,1).
app_href(29,170,132,1).
app_href(29,171,944,1).
app_href(29,171,673,1).
app_href(29,172,1076,1).
app_href(29,172,1053,1).
app_href(29,173,1336,1).
app_href(29,575,1138,1).
app_href(29,575,1087,1).
app_href(29,575,784,1).
app_href(29,576,1135,1).
app_href(29,576,1204,1).
app_href(29,576,992,1).
app_href(29,576,866,1).
app_href(29,576,205,1).
app_href(29,576,672,1).
app_href(29,576,1476,1).
app_href(29,576,251,2).
app_href(29,576,444,1).
app_href(29,576,881,1).
app_href(29,576,487,1).
app_href(29,576,678,1).
app_href(29,576,259,1).
app_href(29,577,1123,1).
app_href(29,577,662,1).
app_href(29,577,9,2).
app_href(29,577,769,1).
app_href(29,175,146,1).
app_href(29,175,2,1).
app_href(29,578,1337,1).
app_href(29,578,164,1).
app_href(29,578,1188,1).
app_href(29,578,1081,1).
app_href(29,578,115,1).
app_href(29,578,936,1).
app_href(29,578,1419,1).
app_href(29,578,1346,1).
app_href(29,578,626,1).
app_href(29,578,1446,1).
app_href(29,578,1198,1).
app_href(29,578,757,2).
app_href(29,578,103,1).
app_href(29,177,1403,1).
app_href(29,177,1341,1).
app_href(29,177,970,1).
app_href(29,178,829,1).
app_href(29,36,242,1).
app_href(29,36,539,1).
app_href(29,36,1147,1).
app_href(29,36,1304,1).
app_href(29,36,992,1).
app_href(29,181,354,1).
app_href(29,182,1218,1).
app_href(29,182,420,1).
app_href(29,579,1337,1).
app_href(29,579,937,1).
app_href(29,579,736,1).
app_href(29,579,255,1).
app_href(29,579,629,1).
app_href(29,579,936,1).
app_href(29,579,1476,1).
app_href(29,579,667,1).
app_href(29,579,881,1).
app_href(29,579,929,1).
app_href(29,579,1408,1).
app_href(29,579,1460,1).
app_href(29,579,259,1).
app_href(29,183,444,1).
app_href(29,580,164,1).
app_href(29,580,1475,1).
app_href(29,580,736,1).
app_href(29,580,449,1).
app_href(29,580,624,1).
app_href(29,580,1452,1).
app_href(29,580,617,1).
app_href(29,580,1476,1).
app_href(29,580,564,1).
app_href(29,580,1369,1).
app_href(29,580,1446,1).
app_href(29,580,1478,1).
app_href(29,580,1040,1).
app_href(29,580,1477,1).
app_href(29,580,593,1).
app_href(29,580,389,1).
app_href(29,185,1442,1).
app_href(29,185,736,1).
app_href(29,185,140,1).
app_href(29,185,233,1).
app_href(29,184,317,1).
app_href(29,184,677,1).
app_href(29,186,728,1).
app_href(29,186,454,1).
app_href(29,581,316,1).
app_href(29,581,830,1).
app_href(29,581,296,1).
app_href(29,581,1256,1).
app_href(29,187,444,1).
app_href(29,187,728,1).
app_href(29,582,1277,1).
app_href(29,582,1460,1).
app_href(29,582,1338,1).
app_href(29,582,1360,1).
app_href(29,582,1261,1).
app_href(29,582,721,1).
app_href(29,582,1246,1).
app_href(29,582,931,1).
app_href(29,582,736,2).
app_href(29,582,153,1).
app_href(29,582,1212,1).
app_href(29,582,830,1).
app_href(29,582,439,1).
app_href(29,582,1337,1).
app_href(29,582,1403,1).
app_href(29,582,114,1).
app_href(29,582,113,1).
app_href(29,582,1451,1).
app_href(29,582,602,1).
app_href(29,583,388,1).
app_href(29,583,1337,1).
app_href(29,583,830,1).
app_href(29,583,1403,1).
app_href(29,583,736,1).
app_href(29,583,1310,1).
app_href(29,583,858,1).
app_href(29,583,678,1).
app_href(29,583,259,1).
app_href(29,584,1248,1).
app_href(29,584,9,2).
app_href(29,584,564,1).
app_href(29,584,1088,1).
app_href(29,584,1009,1).
app_href(29,584,2,1).
app_href(29,584,769,1).
app_href(29,584,223,1).
app_href(29,190,316,1).
app_href(29,190,736,1).
app_href(29,190,1083,1).
app_href(39,506,242,1).
app_href(39,506,393,1).
app_href(39,506,2,1).
app_href(39,148,1486,1).
app_href(39,148,1014,1).
app_href(39,149,830,1).
app_href(39,585,1041,1).
app_href(39,585,736,1).
app_href(39,585,1009,1).
app_href(39,586,1337,1).
app_href(39,586,736,1).
app_href(39,586,542,1).
app_href(39,586,594,1).
app_href(39,586,946,1).
app_href(39,15,899,1).
app_href(39,15,1310,1).
app_href(39,15,668,1).
app_href(39,587,1462,1).
app_href(39,587,1337,1).
app_href(39,587,736,1).
app_href(39,152,1370,3).
app_href(39,152,992,1).
app_href(39,152,1133,1).
app_href(39,161,755,1).
app_href(39,129,1473,1).
app_href(39,129,2,1).
app_href(39,588,1200,1).
app_href(39,588,1218,1).
app_href(39,588,420,1).
app_href(39,588,134,1).
app_href(39,588,2,1).
app_href(39,163,489,1).
app_href(39,589,975,1).
app_href(39,589,1310,1).
app_href(39,589,480,1).
app_href(39,589,1300,1).
app_href(39,589,504,1).
app_href(39,507,118,1).
app_href(39,507,477,1).
app_href(39,507,2,1).
app_href(39,508,726,1).
app_href(39,508,2,1).
app_href(39,164,454,1).
app_href(39,590,1125,1).
app_href(39,590,1155,1).
app_href(39,165,728,1).
app_href(39,165,2,1).
app_href(39,591,155,1).
app_href(39,591,975,1).
app_href(39,591,1310,1).
app_href(39,591,1212,1).
app_href(39,166,736,1).
app_href(39,166,231,1).
app_href(39,168,1069,1).
app_href(39,509,1443,1).
app_href(39,509,1218,1).
app_href(39,509,2,1).
app_href(39,170,1192,1).
app_href(39,170,132,1).
app_href(39,592,372,1).
app_href(39,592,140,1).
app_href(39,592,270,1).
app_href(39,171,944,1).
app_href(39,171,673,1).
app_href(39,593,821,1).
app_href(39,593,1026,1).
app_href(39,594,1157,1).
app_href(39,594,1419,1).
app_href(39,594,1476,1).
app_href(39,594,1446,1).
app_href(39,594,820,1).
app_href(39,172,1076,1).
app_href(39,172,1053,1).
app_href(39,173,1336,1).
app_href(39,542,736,1).
app_href(39,542,1433,1).
app_href(39,595,1277,1).
app_href(39,595,602,1).
app_href(39,596,725,1).
app_href(39,596,1274,1).
app_href(39,596,896,1).
app_href(39,596,726,1).
app_href(39,596,728,1).
app_href(39,596,480,3).
app_href(39,596,504,1).
app_href(39,511,1183,1).
app_href(39,511,296,1).
app_href(39,511,2,1).
app_href(39,175,146,1).
app_href(39,175,2,1).
app_href(39,177,1403,1).
app_href(39,177,1341,1).
app_href(39,177,970,1).
app_href(39,178,829,1).
app_href(39,513,1459,1).
app_href(39,513,2,1).
app_href(39,181,354,1).
app_href(39,597,1443,1).
app_href(39,597,1010,1).
app_href(39,597,420,1).
app_href(39,597,1218,1).
app_href(39,597,736,1).
app_href(39,514,575,1).
app_href(39,514,2,1).
app_href(39,598,1287,1).
app_href(39,598,1093,1).
app_href(39,26,41,1).
app_href(39,26,9,2).
app_href(39,26,1189,1).
app_href(39,26,728,1).
app_href(39,183,444,1).
app_href(39,183,1200,1).
app_href(39,183,134,1).
app_href(39,183,2,1).
app_href(39,515,999,1).
app_href(39,515,2,1).
app_href(39,185,1442,1).
app_href(39,185,736,1).
app_href(39,185,140,1).
app_href(39,185,233,1).
app_href(39,184,317,1).
app_href(39,184,677,1).
app_href(39,52,1385,1).
app_href(39,52,2,1).
app_href(39,517,1318,1).
app_href(39,517,2,1).
app_href(39,518,725,1).
app_href(39,518,2,1).
app_href(39,190,316,1).
app_href(39,190,736,1).
app_href(39,190,1083,1).
app_href(40,599,137,1).
app_href(40,599,309,1).
app_href(40,600,589,1).
app_href(40,506,242,1).
app_href(40,506,393,1).
app_href(40,506,2,1).
app_href(40,148,1486,1).
app_href(40,148,1014,1).
app_href(40,149,830,1).
app_href(40,601,1429,1).
app_href(40,601,642,1).
app_href(40,15,899,1).
app_href(40,15,1310,1).
app_href(40,15,668,1).
app_href(40,152,1370,3).
app_href(40,152,992,1).
app_href(40,152,1133,1).
app_href(40,602,232,1).
app_href(40,602,412,1).
app_href(40,161,755,1).
app_href(40,129,1473,1).
app_href(40,129,2,1).
app_href(40,603,452,1).
app_href(40,603,642,1).
app_href(40,163,489,1).
app_href(40,507,118,1).
app_href(40,507,477,1).
app_href(40,507,2,1).
app_href(40,604,452,1).
app_href(40,604,642,1).
app_href(40,508,726,1).
app_href(40,508,2,1).
app_href(40,164,454,1).
app_href(40,165,728,1).
app_href(40,165,2,1).
app_href(40,166,736,1).
app_href(40,166,231,1).
app_href(40,168,1069,1).
app_href(40,605,127,1).
app_href(40,606,452,1).
app_href(40,606,1061,1).
app_href(40,607,1030,1).
app_href(40,607,452,1).
app_href(40,509,1443,1).
app_href(40,509,1218,1).
app_href(40,509,2,1).
app_href(24,665,1200,1).
app_href(24,665,1218,1).
app_href(24,665,420,1).
app_href(24,665,134,1).
app_href(24,665,2,1).
app_href(24,666,736,1).
app_href(24,666,231,1).
app_href(24,190,316,1).
app_href(24,190,736,1).
app_href(24,190,1083,1).
app_href(41,667,787,1).
app_href(41,667,145,1).
app_href(41,667,1394,1).
app_href(41,667,1075,1).
app_href(41,668,1030,1).
app_href(41,35,365,1).
app_href(41,35,6,1).
app_href(41,669,763,1).
app_href(41,669,1073,1).
app_href(41,669,183,1).
app_href(41,670,1041,1).
app_href(41,670,973,1).
app_href(41,670,736,1).
app_href(41,670,1346,1).
app_href(41,670,1075,1).
app_href(41,750,763,1).
app_href(41,750,1073,1).
app_href(41,87,444,1).
app_href(41,87,358,1).
app_href(41,430,736,1).
app_href(41,430,1346,1).
app_href(41,430,1014,1).
app_href(41,430,634,1).
app_href(41,430,36,1).
app_href(41,671,642,1).
app_href(41,737,763,1).
app_href(41,737,1274,1).
app_href(41,737,1073,1).
app_href(41,737,183,1).
app_href(41,15,899,1).
app_href(41,15,1310,1).
app_href(41,15,668,1).
app_href(41,434,1346,1).
app_href(41,434,1014,1).
app_href(41,434,634,1).
app_href(41,434,772,1).
app_href(41,764,763,1).
app_href(41,764,1073,1).
app_href(41,764,512,1).
app_href(42,506,242,1).
app_href(42,506,393,1).
app_href(42,506,2,1).
app_href(42,148,1486,1).
app_href(42,148,1014,1).
app_href(42,149,830,1).
app_href(42,672,452,1).
app_href(42,672,642,1).
app_href(42,15,899,1).
app_href(42,15,1310,1).
app_href(42,15,668,1).
app_href(42,673,1429,1).
app_href(42,673,642,1).
app_href(42,152,1370,3).
app_href(42,152,992,1).
app_href(42,152,1133,1).
app_href(42,674,452,1).
app_href(42,674,642,1).
app_href(42,161,755,1).
app_href(42,675,452,1).
app_href(42,675,642,1).
app_href(42,129,1473,1).
app_href(42,129,2,1).
app_href(42,676,137,1).
app_href(42,676,309,1).
app_href(42,163,489,1).
app_href(42,507,118,1).
app_href(42,507,477,1).
app_href(42,507,2,1).
app_href(42,677,137,1).
app_href(42,677,309,1).
app_href(42,678,365,1).
app_href(42,679,1063,1).
app_href(42,508,726,1).
app_href(42,508,2,1).
app_href(42,164,454,1).
app_href(42,680,868,1).
app_href(42,680,678,1).
app_href(42,165,728,1).
app_href(42,165,2,1).
app_href(42,681,507,1).
app_href(42,166,736,1).
app_href(42,166,231,1).
app_href(42,168,1069,1).
app_href(42,48,1429,1).
app_href(42,48,642,1).
app_href(42,509,1443,1).
app_href(42,509,1218,1).
app_href(42,509,2,1).
app_href(42,682,428,1).
app_href(42,170,1192,1).
app_href(42,170,132,1).
app_href(42,171,944,1).
app_href(42,171,673,1).
app_href(42,172,1076,1).
app_href(42,172,1053,1).
app_href(42,683,365,1).
app_href(42,173,1336,1).
app_href(42,684,507,1).
app_href(42,685,1063,1).
app_href(42,511,1183,1).
app_href(42,511,296,1).
app_href(42,511,2,1).
app_href(42,175,146,1).
app_href(42,175,2,1).
app_href(42,686,428,1).
app_href(42,177,1403,1).
app_href(42,177,1341,1).
app_href(42,177,970,1).
app_href(42,178,829,1).
app_href(42,687,452,1).
app_href(42,687,642,1).
app_href(42,513,1459,1).
app_href(42,513,2,1).
app_href(42,688,452,1).
app_href(42,688,642,1).
app_href(42,181,354,1).
app_href(42,182,1200,1).
app_href(42,182,1218,1).
app_href(42,182,420,1).
app_href(42,182,134,1).
app_href(42,182,2,1).
app_href(42,514,575,1).
app_href(42,514,2,1).
app_href(42,183,444,1).
app_href(42,183,1200,1).
app_href(42,183,134,1).
app_href(42,183,2,1).
app_href(42,515,999,1).
app_href(42,515,2,1).
app_href(42,185,1442,1).
app_href(42,185,736,1).
app_href(42,185,140,1).
app_href(42,185,233,1).
app_href(42,184,317,1).
app_href(42,184,677,1).
app_href(42,689,452,1).
app_href(42,689,642,1).
app_href(42,52,1385,1).
app_href(42,52,2,1).
app_href(42,517,1318,1).
app_href(42,517,2,1).
app_href(42,690,232,1).
app_href(42,690,412,1).
app_href(42,518,725,1).
app_href(42,518,2,1).
app_href(42,190,316,1).
app_href(42,190,736,1).
app_href(37,358,1270,1).
app_href(37,358,934,1).
app_href(37,106,1409,1).
app_href(37,106,854,1).
app_href(37,106,736,1).
app_href(37,756,763,1).
app_href(37,756,818,1).
app_href(37,756,922,1).
app_href(37,443,763,1).
app_href(37,443,1330,1).
app_href(37,443,1323,1).
app_href(37,757,763,1).
app_href(37,757,1376,1).
app_href(37,757,1473,1).
app_href(37,757,1291,1).
app_href(37,342,763,1).
app_href(37,342,378,1).
app_href(37,108,214,1).
app_href(37,108,736,1).
app_href(37,108,213,1).
app_href(37,338,773,1).
app_href(37,338,652,1).
app_href(37,338,998,1).
app_href(37,758,763,1).
app_href(37,758,512,1).
app_href(37,758,922,1).
app_href(37,759,972,1).
app_href(37,759,1211,1).
app_href(37,759,1058,1).
app_href(37,759,998,1).
app_href(37,760,763,1).
app_href(37,760,1376,1).
app_href(37,760,1473,1).
app_href(37,760,447,1).
app_href(37,350,736,1).
app_href(37,350,1198,1).
app_href(37,89,736,1).
app_href(37,89,396,1).
app_href(37,89,197,1).
app_href(37,761,1409,1).
app_href(37,761,1065,1).
app_href(37,761,736,1).
app_href(37,301,763,1).
app_href(37,301,1182,1).
app_href(37,301,1255,1).
app_href(37,301,1267,1).
app_href(37,762,736,1).
app_href(37,762,1343,1).
app_href(37,763,972,1).
app_href(37,763,367,1).
app_href(37,763,998,1).
app_href(37,764,763,1).
app_href(37,764,1073,1).
app_href(37,764,512,1).
app_href(37,441,763,1).
app_href(37,441,512,1).
app_href(37,441,1323,1).
app_href(37,765,972,1).
app_href(37,765,370,1).
app_href(37,765,924,1).
app_href(37,765,998,1).
app_href(37,766,763,1).
app_href(37,766,658,1).
app_href(37,766,892,1).
app_href(37,840,763,1).
app_href(37,840,1376,1).
app_href(37,840,1473,1).
app_href(37,840,1185,1).
app_href(37,841,736,1).
app_href(37,841,1122,1).
app_href(37,841,809,1).
app_href(37,842,972,1).
app_href(37,842,299,1).
app_href(37,842,719,1).
app_href(37,842,998,1).
app_href(37,37,580,2).
app_href(37,37,951,2).
app_href(37,843,763,1).
app_href(37,843,1373,1).
app_href(37,844,763,1).
app_href(37,844,1162,1).
app_href(37,845,763,1).
app_href(37,845,922,1).
app_href(37,845,1267,1).
app_href(37,846,763,1).
app_href(37,846,658,1).
app_href(37,300,763,1).
app_href(37,300,900,1).
app_href(37,300,1255,1).
app_href(37,300,1267,1).
app_href(37,847,972,1).
app_href(37,847,300,1).
app_href(37,847,719,1).
app_href(37,847,998,1).
app_href(37,359,1260,1).
app_href(37,359,736,1).
app_href(37,359,1198,1).
app_href(37,357,267,1).
app_href(37,357,736,1).
app_href(37,357,1198,1).
app_href(37,364,763,1).
app_href(37,364,934,1).
app_href(37,450,763,1).
app_href(37,450,308,1).
app_href(37,450,1323,1).
app_href(37,848,763,1).
app_href(37,848,658,1).
app_href(37,848,1291,1).
app_href(37,849,596,1).
app_href(37,849,736,1).
app_href(37,849,197,1).
app_href(37,850,972,1).
app_href(37,850,1058,1).
app_href(37,850,998,1).
app_href(37,851,1381,2).
app_href(37,327,736,1).
app_href(37,327,1269,1).
app_href(37,84,736,1).
app_href(37,84,512,1).
app_href(37,84,197,1).
app_href(37,852,278,2).
app_href(37,852,736,2).
app_href(37,852,950,2).
app_href(37,456,763,1).
app_href(37,456,642,1).
app_href(37,456,1323,1).
app_href(37,456,1343,1).
app_href(37,456,1020,1).
app_href(37,853,763,1).
app_href(37,853,922,1).
app_href(37,460,763,1).
app_href(37,460,1038,1).
app_href(37,460,1323,1).
app_href(37,460,913,1).
app_href(37,854,307,1).
app_href(37,854,736,1).
app_href(37,854,197,1).
app_href(37,855,972,1).
app_href(37,855,467,1).
app_href(37,855,998,1).
app_href(37,856,629,2).
app_href(37,857,763,1).
app_href(37,857,658,1).
app_href(37,857,1267,1).
app_href(37,440,1381,1).
app_href(37,440,763,1).
app_href(37,440,1323,1).
app_href(37,453,763,1).
app_href(37,453,1174,1).
app_href(37,453,1323,1).
app_href(37,453,1343,1).
app_href(37,858,151,1).
app_href(37,858,736,1).
app_href(37,372,763,1).
app_href(37,372,186,1).
app_href(37,372,1198,1).
app_href(37,372,1267,1).
app_href(37,859,1205,1).
app_href(37,859,972,1).
app_href(37,859,998,1).
app_href(37,860,972,1).
app_href(37,860,1057,1).
app_href(37,860,998,1).
app_href(37,861,368,1).
app_href(37,861,972,1).
app_href(37,861,998,1).
app_href(37,669,763,1).
app_href(37,669,1073,1).
app_href(37,669,183,1).
app_href(37,862,736,1).
app_href(37,862,880,1).
app_href(37,863,763,1).
app_href(37,863,367,1).
app_href(37,863,1162,1).
app_href(37,725,425,1).
app_href(37,725,736,1).
app_href(37,725,1162,1).
app_href(37,864,972,1).
app_href(37,864,383,1).
app_href(37,864,1058,1).
app_href(37,864,998,1).
app_href(37,344,763,1).
app_href(37,344,378,1).
app_href(37,344,702,1).
app_href(37,865,412,1).
app_href(37,866,972,2).
app_href(37,866,271,2).
app_href(37,866,643,2).
app_href(37,866,27,2).
app_href(37,866,272,2).
app_href(37,866,441,2).
app_href(37,866,998,2).
app_href(37,449,763,1).
app_href(37,449,893,1).
app_href(37,449,1038,1).
app_href(37,449,1323,1).
app_href(37,867,736,1).
app_href(37,867,1198,1).
app_href(37,867,121,1).
app_href(37,868,736,1).
app_href(37,868,809,1).
app_href(37,452,763,1).
app_href(37,452,1323,1).
app_href(37,31,580,2).
app_href(37,869,763,1).
app_href(37,869,1473,1).
app_href(37,869,1021,1).
app_href(37,458,763,1).
app_href(37,458,1323,1).
app_href(37,458,1343,1).
app_href(37,870,763,1).
app_href(37,870,1376,1).
app_href(37,870,1473,1).
app_href(37,871,736,1).
app_href(37,871,1267,1).
app_href(37,871,283,1).
app_href(37,871,950,1).
app_href(37,469,1409,1).
app_href(37,469,736,1).
app_href(37,469,1130,1).
app_href(37,872,972,1).
app_href(37,872,923,1).
app_href(37,872,370,1).
app_href(37,872,998,1).
app_href(37,873,736,1).
app_href(37,873,657,1).
app_href(37,873,197,1).
app_href(37,874,1409,1).
app_href(37,874,736,1).
app_href(37,875,1048,1).
app_href(37,875,763,1).
app_href(37,875,658,1).
app_href(37,875,783,1).
app_href(37,459,763,1).
app_href(37,459,642,1).
app_href(37,459,1323,1).
app_href(37,459,1343,1).
app_href(37,876,470,1).
app_href(37,876,972,1).
app_href(37,876,367,1).
app_href(37,876,998,1).
app_href(37,716,736,1).
app_href(37,716,123,1).
app_href(37,716,1162,1).
app_href(37,877,736,1).
app_href(37,877,1143,1).
app_href(37,877,657,1).
app_href(37,877,197,1).
app_href(37,878,972,4).
app_href(37,878,271,4).
app_href(37,878,643,4).
app_href(37,878,99,4).
app_href(37,878,272,4).
app_href(37,878,440,4).
app_href(37,878,998,4).
app_href(37,879,672,2).
app_href(37,880,972,1).
app_href(37,880,1052,1).
app_href(37,880,1058,1).
app_href(37,880,998,1).
app_href(37,707,1134,1).
app_href(37,707,736,1).
app_href(37,707,1162,1).
app_href(37,881,289,2).
app_href(37,881,134,2).
app_href(37,451,547,1).
app_href(37,451,763,1).
app_href(37,451,1323,1).
app_href(37,882,1473,1).
app_href(37,882,1185,1).
app_href(37,882,998,1).
app_href(37,303,763,1).
app_href(37,303,1182,1).
app_href(37,303,1255,1).
app_href(37,883,656,1).
app_href(37,883,736,1).
app_href(37,883,197,1).
app_href(37,883,1267,1).
app_href(37,103,736,1).
app_href(37,103,197,1).
app_href(37,884,368,1).
app_href(37,884,972,1).
app_href(37,884,1056,1).
app_href(37,884,998,1).
app_href(37,354,763,1).
app_href(37,354,186,1).
app_href(37,354,1198,1).
app_href(37,354,258,1).
app_href(37,885,972,1).
app_href(37,885,891,1).
app_href(37,885,998,1).
app_href(37,454,763,1).
app_href(37,454,236,1).
app_href(37,454,1323,1).
app_href(37,355,763,1).
app_href(37,355,267,1).
app_href(37,355,1198,1).
app_href(37,36,242,1).
app_href(37,36,539,1).
app_href(37,36,1303,1).
app_href(37,36,992,1).
app_href(37,886,561,1).
app_href(37,886,736,1).
app_href(37,886,197,1).
app_href(37,887,278,1).
app_href(37,887,736,1).
app_href(37,887,1267,1).
app_href(37,887,950,1).
app_href(37,888,447,1).
app_href(37,888,1437,1).
app_href(37,888,998,1).
app_href(37,726,1341,1).
app_href(37,726,829,1).
app_href(37,889,972,1).
app_href(42,190,1083,1).
app_href(42,691,232,1).
app_href(42,691,412,1).
app_href(43,692,1030,1).
app_href(43,693,736,1).
app_href(43,693,1329,1).
app_href(43,693,809,1).
app_href(43,35,365,1).
app_href(43,35,6,1).
app_href(43,694,1011,1).
app_href(43,694,736,1).
app_href(43,694,809,1).
app_href(43,87,444,1).
app_href(43,87,358,1).
app_href(43,695,1105,1).
app_href(43,696,642,1).
app_href(43,467,736,1).
app_href(43,467,288,1).
app_href(43,467,809,1).
app_href(43,697,736,1).
app_href(43,697,291,1).
app_href(43,697,809,1).
app_href(43,15,899,1).
app_href(43,15,1310,1).
app_href(43,15,668,1).
app_href(48,908,1002,1).
app_href(48,908,1290,1).
app_href(48,908,516,2).
app_href(48,908,1438,1).
app_href(48,908,1352,2).
app_href(48,908,1175,4).
app_href(48,908,334,1).
app_href(48,908,432,1).
app_href(48,908,1417,4).
app_href(48,908,699,1).
app_href(48,908,1060,1).
app_href(48,908,97,1).
app_href(48,908,96,16).
app_href(48,908,853,4).
app_href(48,908,92,1).
app_href(48,908,90,1).
app_href(48,908,986,1).
app_href(48,908,115,3).
app_href(48,908,961,1).
app_href(48,908,1068,1).
app_href(48,908,88,1).
app_href(48,908,1224,1).
app_href(48,908,41,1).
app_href(48,908,18,1).
app_href(48,908,37,1).
app_href(48,908,16,1).
app_href(48,908,26,1).
app_href(48,908,15,1).
app_href(48,908,9,1).
app_href(48,908,13,1).
app_href(48,908,859,1).
app_href(48,908,10,1).
app_href(48,908,1346,1).
app_href(48,908,1322,1).
app_href(48,908,451,1).
app_href(48,908,997,1).
app_href(48,908,970,2).
app_href(48,908,1467,1).
app_href(48,908,564,1).
app_href(48,908,1326,1).
app_href(48,908,953,1).
app_href(48,908,140,7).
app_href(48,908,1380,1).
app_href(48,908,522,1).
app_href(48,908,1435,1).
app_href(48,908,581,1).
app_href(48,908,1340,1).
app_href(48,909,864,2).
app_href(48,909,413,1).
app_href(48,909,180,1).
app_href(48,909,1152,1).
app_href(48,909,882,1).
app_href(48,909,509,1).
app_href(48,909,1346,2).
app_href(48,909,980,1).
app_href(48,909,1221,1).
app_href(48,909,736,1).
app_href(48,909,970,2).
app_href(48,909,915,1).
app_href(48,909,568,1).
app_href(48,909,1430,2).
app_href(48,909,694,2).
app_href(48,909,90,5).
app_href(48,909,41,1).
app_href(48,909,37,1).
app_href(48,909,26,1).
app_href(48,909,9,1).
app_href(48,910,287,1).
app_href(48,910,1309,1).
app_href(48,910,149,1).
app_href(48,910,140,1).
app_href(48,910,138,1).
app_href(48,910,970,1).
app_href(48,911,1093,1).
app_href(48,911,551,1).
app_href(48,911,559,2).
app_href(48,911,527,1).
app_href(48,911,888,3).
app_href(48,911,864,2).
app_href(48,911,736,1).
app_href(48,911,577,1).
app_href(48,911,681,1).
app_href(48,911,363,4).
app_href(48,911,264,1).
app_href(48,911,701,2).
app_href(48,911,680,1).
app_href(48,911,1096,1).
app_href(48,911,683,1).
app_href(48,911,702,4).
app_href(48,911,234,1).
app_href(48,911,1072,3).
app_href(48,911,894,3).
app_href(48,911,1082,1).
app_href(48,911,697,1).
app_href(48,911,525,1).
app_href(48,911,519,2).
app_href(48,911,567,1).
app_href(48,911,1123,4).
app_href(48,911,987,4).
app_href(48,911,1097,1).
app_href(48,911,1019,1).
app_href(48,911,592,1).
app_href(48,911,289,5).
app_href(48,911,1416,8).
app_href(48,911,544,1).
app_href(48,911,364,6).
app_href(48,911,170,1).
app_href(48,911,350,1).
app_href(48,911,351,1).
app_href(48,911,1179,1).
app_href(48,911,1293,1).
app_href(48,911,667,1).
app_href(48,911,920,6).
app_href(48,911,1415,2).
app_href(48,911,1221,1).
app_href(48,911,1153,1).
app_href(48,911,973,4).
app_href(48,911,1224,1).
app_href(48,911,1284,1).
app_href(48,911,682,3).
app_href(48,911,834,1).
app_href(48,911,1233,1).
app_href(48,911,1067,1).
app_href(48,911,970,6).
app_href(48,911,518,3).
app_href(48,911,1006,1).
app_href(48,911,647,1).
app_href(48,911,376,2).
app_href(48,911,965,1).
app_href(48,911,941,1).
app_href(48,911,1012,2).
app_href(48,911,545,2).
app_href(37,889,367,1).
app_href(37,889,1388,1).
app_href(37,889,998,1).
app_href(44,890,1319,1).
app_href(44,890,562,1).
app_href(44,891,570,1).
app_href(44,891,890,1).
app_href(44,22,581,1).
app_href(44,22,686,1).
app_href(44,22,894,1).
app_href(44,35,365,1).
app_href(44,35,6,1).
app_href(44,892,1009,1).
app_href(44,892,704,1).
app_href(44,893,570,1).
app_href(44,893,963,1).
app_href(44,15,899,1).
app_href(44,15,1310,1).
app_href(44,15,668,1).
app_href(44,894,237,1).
app_href(44,894,1343,1).
app_href(44,87,444,1).
app_href(44,87,358,1).
app_href(44,895,1371,1).
app_href(44,895,704,1).
app_href(44,134,1424,1).
app_href(44,134,1248,1).
app_href(44,896,302,1).
app_href(44,363,736,1).
app_href(44,363,1384,1).
app_href(44,897,172,1).
app_href(44,897,704,1).
app_href(44,388,1319,1).
app_href(44,388,492,1).
app_href(44,408,1424,1).
app_href(44,408,1456,1).
app_href(44,898,817,1).
app_href(44,898,140,1).
app_href(44,898,1343,1).
app_href(44,898,562,1).
app_href(44,433,616,1).
app_href(44,433,736,1).
app_href(44,433,145,1).
app_href(44,433,36,1).
app_href(44,899,584,1).
app_href(44,899,1112,1).
app_href(44,899,174,1).
app_href(44,900,570,1).
app_href(44,900,397,1).
app_href(44,900,140,1).
app_href(44,900,1355,1).
app_href(44,901,895,1).
app_href(44,901,686,1).
app_href(44,476,847,1).
app_href(44,476,1319,1).
app_href(44,902,561,1).
app_href(44,902,1098,1).
app_href(44,902,1346,1).
app_href(44,902,550,1).
app_href(49,77,517,1).
app_href(49,77,9,2).
app_href(49,77,769,1).
app_href(49,903,517,1).
app_href(49,903,26,1).
app_href(49,903,9,1).
app_href(49,903,8,1).
app_href(49,903,769,1).
app_href(49,904,480,1).
app_href(49,904,555,1).
app_href(49,905,1194,1).
app_href(49,905,26,1).
app_href(49,905,1436,1).
app_href(49,905,480,1).
app_href(49,906,232,1).
app_href(49,907,1386,1).
app_href(49,907,1117,1).
app_href(49,936,253,1).
app_href(49,936,1479,1).
app_href(49,936,743,1).
app_href(49,936,14,1).
app_href(50,337,667,1).
app_href(50,337,26,1).
app_href(50,337,9,1).
app_href(50,337,1198,1).
app_href(50,937,1009,1).
app_href(50,938,137,1).
app_href(50,938,309,1).
app_href(50,182,1218,1).
app_href(50,182,420,1).
app_href(50,455,615,1).
app_href(50,455,375,1).
app_href(50,455,1323,1).
app_href(50,455,772,1).
app_href(50,455,1460,1).
app_href(50,939,821,1).
app_href(50,939,736,1).
app_href(50,939,1346,1).
app_href(50,939,810,1).
app_href(50,940,3,1).
app_href(50,940,1358,1).
app_href(50,940,9,2).
app_href(50,940,140,1).
app_href(50,940,220,1).
app_href(50,940,769,1).
app_href(50,941,385,1).
app_href(50,941,939,1).
app_href(50,942,1127,1).
app_href(50,943,1135,1).
app_href(50,943,527,1).
app_href(50,15,899,1).
app_href(50,15,1310,1).
app_href(50,15,668,1).
app_href(50,181,354,1).
app_href(50,944,137,1).
app_href(50,944,309,1).
app_href(50,149,830,1).
app_href(50,168,1069,1).
app_href(50,542,736,1).
app_href(50,542,1433,1).
app_href(50,26,41,1).
app_href(50,26,9,2).
app_href(50,26,1189,1).
app_href(50,26,728,1).
app_href(50,710,1162,1).
app_href(50,152,1370,3).
app_href(50,152,992,1).
app_href(50,152,1133,1).
app_href(50,183,444,1).
app_href(50,177,1403,1).
app_href(50,177,1341,1).
app_href(50,177,970,1).
app_href(50,161,755,1).
app_href(50,173,1336,1).
app_href(50,945,211,1).
app_href(50,945,978,1).
app_href(50,945,736,1).
app_href(50,945,1262,1).
app_href(50,945,1384,1).
app_href(50,945,618,1).
app_href(50,178,829,1).
app_href(50,192,1108,1).
app_href(50,192,134,1).
app_href(50,192,722,1).
app_href(50,192,1361,1).
app_href(50,946,233,1).
app_href(50,946,2,1).
app_href(50,112,1358,1).
app_href(50,112,740,1).
app_href(50,597,1443,1).
app_href(50,597,1010,1).
app_href(50,597,420,1).
app_href(50,597,1218,1).
app_href(50,597,736,1).
app_href(50,947,1255,1).
app_href(50,948,365,1).
app_href(50,948,1283,1).
app_href(50,895,704,1).
app_href(40,608,1030,1).
app_href(40,608,452,1).
app_href(40,170,1192,1).
app_href(40,170,132,1).
app_href(40,171,944,1).
app_href(40,171,673,1).
app_href(40,609,643,1).
app_href(40,610,543,1).
app_href(40,610,566,1).
app_href(40,172,1076,1).
app_href(40,172,1053,1).
app_href(40,173,1336,1).
app_href(40,611,736,1).
app_href(40,611,9,2).
app_href(40,611,277,1).
app_href(40,611,564,1).
app_href(40,612,365,1).
app_href(40,511,1183,1).
app_href(40,511,296,1).
app_href(40,511,2,1).
app_href(40,613,643,1).
app_href(40,614,868,1).
app_href(40,614,678,1).
app_href(40,175,146,1).
app_href(40,175,2,1).
app_href(40,177,1403,1).
app_href(40,177,1341,1).
app_href(40,177,970,1).
app_href(40,615,428,1).
app_href(40,178,829,1).
app_href(40,513,1459,1).
app_href(40,513,2,1).
app_href(40,616,589,1).
app_href(40,181,354,1).
app_href(40,182,1200,1).
app_href(40,182,1218,1).
app_href(40,182,420,1).
app_href(40,182,134,1).
app_href(40,182,2,1).
app_href(40,514,575,1).
app_href(40,514,2,1).
app_href(40,183,444,1).
app_href(40,183,1200,1).
app_href(40,183,134,1).
app_href(40,183,2,1).
app_href(40,515,999,1).
app_href(40,515,2,1).
app_href(40,185,1442,1).
app_href(40,185,736,1).
app_href(40,185,140,1).
app_href(40,185,233,1).
app_href(40,184,317,1).
app_href(40,184,677,1).
app_href(40,11,1429,1).
app_href(40,11,642,1).
app_href(40,617,428,1).
app_href(40,52,1385,1).
app_href(40,52,2,1).
app_href(40,618,137,1).
app_href(40,618,309,1).
app_href(40,619,358,1).
app_href(40,619,1330,1).
app_href(40,619,970,1).
app_href(40,517,1318,1).
app_href(40,517,2,1).
app_href(40,518,725,1).
app_href(40,518,2,1).
app_href(40,189,287,1).
app_href(40,189,140,1).
app_href(40,189,138,1).
app_href(40,620,452,1).
app_href(40,620,1061,1).
app_href(40,190,316,1).
app_href(40,190,736,1).
app_href(40,190,1083,1).
app_href(46,621,160,1).
app_href(46,621,418,1).
app_href(46,621,145,1).
app_href(46,621,883,1).
app_href(46,621,530,1).
app_href(46,622,1351,1).
app_href(46,622,736,1).
app_href(46,622,809,1).
app_href(46,1010,1009,1).
app_href(46,35,365,1).
app_href(46,35,6,1).
app_href(46,1011,1425,1).
app_href(46,1011,990,1).
app_href(46,694,1011,1).
app_href(46,694,736,1).
app_href(46,694,809,1).
app_href(46,87,444,1).
app_href(46,87,358,1).
app_href(46,697,736,1).
app_href(46,697,291,1).
app_href(46,697,809,1).
app_href(46,1012,736,1).
app_href(46,1012,1198,1).
app_href(46,1012,1197,1).
app_href(46,1013,736,1).
app_href(46,1013,1346,1).
app_href(46,1013,883,1).
app_href(46,1013,530,1).
app_href(46,15,899,1).
app_href(46,15,1310,1).
app_href(46,15,668,1).
app_href(46,1014,736,1).
app_href(46,1014,1399,1).
app_href(46,1014,934,1).
app_href(46,1015,735,1).
app_href(46,1015,736,1).
app_href(46,1015,934,1).
app_href(53,35,365,1).
app_href(53,1016,581,1).
app_href(53,1016,1346,1).
app_href(53,1016,303,1).
app_href(53,15,899,1).
app_href(53,15,1310,1).
app_href(53,15,668,1).
app_href(53,87,444,1).
app_href(53,87,358,1).
app_href(53,1017,667,1).
app_href(53,1017,37,1).
app_href(53,1017,9,1).
app_href(53,1017,503,1).
app_href(53,841,736,1).
app_href(53,841,1122,1).
app_href(53,841,809,1).
app_href(53,731,1126,1).
app_href(53,731,736,1).
app_href(53,731,809,1).
app_href(53,868,736,1).
app_href(53,868,809,1).
app_href(53,388,1319,1).
app_href(53,388,492,1).
app_href(53,736,1409,1).
app_href(53,736,736,1).
app_href(53,736,733,1).
app_href(53,106,1409,1).
app_href(53,106,854,1).
app_href(53,106,736,1).
app_href(53,494,177,1).
app_href(53,494,533,1).
app_href(53,874,1409,1).
app_href(53,874,736,1).
app_href(53,491,532,1).
app_href(53,491,518,1).
app_href(53,491,273,1).
app_href(53,6,854,1).
app_href(53,6,145,1).
app_href(53,487,1064,1).
app_href(53,487,145,1).
app_href(53,469,1409,2).
app_href(53,469,736,2).
app_href(53,469,1130,2).
app_href(53,747,1409,1).
app_href(53,747,736,1).
app_href(48,911,1453,2).
app_href(48,911,840,1).
app_href(48,911,693,2).
app_href(48,911,1417,4).
app_href(48,911,933,1).
app_href(48,911,140,14).
app_href(48,911,969,3).
app_href(48,911,810,1).
app_href(48,911,265,1).
app_href(48,911,118,2).
app_href(48,911,620,1).
app_href(48,911,327,1).
app_href(48,911,907,5).
app_href(48,911,632,1).
app_href(48,911,844,1).
app_href(48,911,1282,1).
app_href(48,911,139,1).
app_href(48,911,1008,1).
app_href(48,911,178,1).
app_href(48,911,715,1).
app_href(48,911,97,6).
app_href(48,911,96,8).
app_href(48,911,92,9).
app_href(48,911,90,14).
app_href(48,911,88,7).
app_href(48,911,1230,1).
app_href(48,911,41,18).
app_href(48,911,37,10).
app_href(48,911,1315,1).
app_href(48,911,26,89).
app_href(48,911,169,4).
app_href(48,911,1302,1).
app_href(48,911,9,13).
app_href(48,911,24,1).
app_href(48,911,23,1).
app_href(48,911,1421,1).
app_href(48,911,22,10).
app_href(48,911,1110,3).
app_href(48,911,20,5).
app_href(48,911,19,5).
app_href(48,911,18,1).
app_href(48,911,967,1).
app_href(48,911,16,5).
app_href(48,911,360,1).
app_href(48,911,1352,1).
app_href(48,911,15,2).
app_href(48,911,1389,6).
app_href(48,911,13,1).
app_href(48,911,1314,1).
app_href(48,911,10,5).
app_href(48,911,925,2).
app_href(48,911,895,1).
app_href(48,911,1430,2).
app_href(48,911,288,7).
app_href(48,911,522,1).
app_href(48,911,1393,1).
app_href(48,911,417,1).
app_href(48,911,1396,1).
app_href(48,911,1346,6).
app_href(48,911,646,1).
app_href(48,911,623,2).
app_href(48,912,672,3).
app_href(48,913,903,1).
app_href(48,913,251,1).
app_href(48,913,979,2).
app_href(48,913,97,1).
app_href(48,678,25,1).
app_href(48,678,899,1).
app_href(48,678,1310,1).
app_href(48,678,365,1).
app_href(48,678,6,1).
app_href(48,678,668,1).
app_href(48,914,9,1).
app_href(48,914,711,1).
app_href(48,915,694,1).
app_href(48,915,1374,1).
app_href(48,915,559,3).
app_href(48,915,1106,1).
app_href(48,915,664,1).
app_href(48,915,1156,1).
app_href(48,915,864,4).
app_href(48,915,165,1).
app_href(48,915,736,4).
app_href(48,915,842,1).
app_href(48,915,363,2).
app_href(48,915,845,1).
app_href(48,915,116,1).
app_href(48,915,984,2).
app_href(48,915,168,3).
app_href(48,915,166,1).
app_href(48,915,1146,1).
app_href(48,915,894,5).
app_href(48,915,625,1).
app_href(48,915,906,1).
app_href(48,915,392,1).
app_href(48,915,519,2).
app_href(48,915,1123,1).
app_href(48,915,322,2).
app_href(48,915,1416,1).
app_href(48,915,1029,1).
app_href(48,915,826,1).
app_href(48,915,871,1).
app_href(48,915,1326,1).
app_href(48,915,1415,1).
app_href(48,915,1221,2).
app_href(48,915,95,1).
app_href(48,915,682,1).
app_href(48,915,695,6).
app_href(48,915,335,1).
app_href(48,915,834,1).
app_href(48,915,552,1).
app_href(48,915,970,2).
app_href(48,915,919,1).
app_href(48,915,954,1).
app_href(48,915,128,1).
app_href(48,915,376,1).
app_href(48,915,1387,1).
app_href(48,915,580,1).
app_href(48,915,1272,3).
app_href(48,915,1296,1).
app_href(48,915,1417,1).
app_href(48,915,140,6).
app_href(48,915,969,1).
app_href(48,915,651,1).
app_href(48,915,1271,1).
app_href(48,915,715,3).
app_href(48,915,97,2).
app_href(48,915,96,5).
app_href(48,915,92,2).
app_href(48,915,1178,2).
app_href(48,915,90,10).
app_href(48,915,88,10).
app_href(48,915,1078,1).
app_href(48,915,41,4).
app_href(48,915,37,41).
app_href(48,915,26,4).
app_href(48,915,9,4).
app_href(48,915,1263,1).
app_href(48,915,967,1).
app_href(48,915,16,1).
app_href(48,915,360,1).
app_href(48,915,15,1).
app_href(48,915,1389,2).
app_href(48,915,13,13).
app_href(48,915,10,2).
app_href(48,915,1309,1).
app_href(48,915,678,1).
app_href(48,915,895,1).
app_href(48,915,1290,1).
app_href(48,915,288,1).
app_href(48,915,1430,4).
app_href(48,915,1131,1).
app_href(48,915,1054,1).
app_href(48,915,873,1).
app_href(48,915,1353,1).
app_href(48,915,467,1).
app_href(48,915,662,1).
app_href(48,915,927,1).
app_href(48,915,1346,4).
app_href(48,915,415,1).
app_href(48,916,543,1).
app_href(48,916,566,1).
app_href(48,917,984,1).
app_href(48,917,120,1).
app_href(48,917,702,1).
app_href(48,917,552,1).
app_href(48,917,1325,1).
app_href(48,917,289,1).
app_href(48,917,1463,1).
app_href(48,917,288,1).
app_href(48,917,519,1).
app_href(48,917,1417,1).
app_href(48,917,1178,1).
app_href(48,917,97,1).
app_href(48,917,321,4).
app_href(48,917,96,1).
app_href(48,917,1119,1).
app_href(48,917,92,17).
app_href(48,917,986,1).
app_href(48,917,90,1).
app_href(48,917,88,1).
app_href(48,917,19,1).
app_href(48,917,41,1).
app_href(48,917,18,1).
app_href(48,917,37,1).
app_href(48,917,16,1).
app_href(48,917,26,1).
app_href(48,917,15,1).
app_href(48,917,1271,1).
app_href(48,917,9,1).
app_href(48,917,13,1).
app_href(48,917,693,1).
app_href(48,917,360,2).
app_href(48,917,1353,1).
app_href(48,917,10,1).
app_href(48,917,568,1).
app_href(48,917,1346,3).
app_href(48,917,1430,1).
app_href(48,917,335,1).
app_href(48,917,973,1).
app_href(48,917,911,1).
app_href(48,917,895,1).
app_href(48,917,970,2).
app_href(48,917,517,2).
app_href(48,917,170,1).
app_href(48,917,167,1).
app_href(48,917,1460,1).
app_href(48,917,345,1).
app_href(48,917,845,1).
app_href(48,917,736,1).
app_href(48,917,564,1).
app_href(48,917,1054,1).
app_href(48,917,168,1).
app_href(48,917,1326,1).
app_href(48,917,717,1).
app_href(48,917,140,3).
app_href(48,917,864,1).
app_href(48,917,1403,1).
app_href(48,917,933,1).
app_href(48,917,352,1).
app_href(48,917,620,1).
app_href(48,918,943,1).
app_href(48,918,619,1).
app_href(48,918,922,1).
app_href(48,918,374,1).
app_href(48,918,344,1).
app_href(48,918,864,2).
app_href(48,918,1083,1).
app_href(48,918,736,2).
app_href(48,918,566,1).
app_href(48,918,701,1).
app_href(48,918,290,1).
app_href(48,918,550,1).
app_href(48,918,609,1).
app_href(48,918,568,1).
app_href(48,918,894,1).
app_href(48,918,410,1).
app_href(48,918,337,1).
app_href(48,918,1295,1).
app_href(48,918,519,2).
app_href(48,918,339,1).
app_href(48,918,348,1).
app_href(48,918,592,1).
app_href(48,918,1416,1).
app_href(48,918,544,1).
app_href(48,918,546,1).
app_href(48,918,342,1).
app_href(48,918,414,3).
app_href(48,918,296,2).
app_href(48,918,338,1).
app_href(48,918,183,9).
app_href(48,918,343,1).
app_href(48,918,1422,1).
app_href(48,918,920,1).
app_href(48,918,973,1).
app_href(48,918,682,1).
app_href(48,918,1284,1).
app_href(48,918,335,1).
app_href(48,918,552,1).
app_href(48,918,970,4).
app_href(48,918,518,1).
app_href(48,918,252,1).
app_href(48,918,685,1).
app_href(48,918,957,1).
app_href(48,918,941,1).
app_href(48,918,1297,3).
app_href(48,918,545,1).
app_href(48,918,1296,5).
app_href(48,918,969,1).
app_href(48,918,140,8).
app_href(48,918,451,1).
app_href(48,918,346,1).
app_href(48,918,703,1).
app_href(48,918,341,1).
app_href(48,918,833,1).
app_href(48,918,349,3).
app_href(48,918,1328,1).
app_href(48,918,832,1).
app_href(48,918,97,8).
app_href(48,918,96,4).
app_href(48,918,92,13).
app_href(48,918,90,4).
app_href(48,918,88,4).
app_href(48,918,41,50).
app_href(48,918,37,8).
app_href(48,918,26,6).
app_href(48,918,9,6).
app_href(48,918,1421,2).
app_href(48,918,184,2).
app_href(48,918,1110,1).
app_href(48,918,1389,1).
app_href(48,918,10,2).
app_href(48,918,336,1).
app_href(48,918,1257,1).
app_href(48,918,895,2).
app_href(48,918,1430,2).
app_href(48,918,288,3).
app_href(48,918,1054,1).
app_href(48,918,849,1).
app_href(48,918,692,1).
app_href(48,918,347,1).
app_href(48,918,405,1).
app_href(48,918,859,1).
app_href(48,918,1346,23).
app_href(48,918,340,1).
app_href(48,919,349,1).
app_href(48,919,1408,2).
app_href(48,919,375,1).
app_href(48,919,927,1).
app_href(48,919,169,1).
app_href(48,919,1407,1).
app_href(48,919,834,2).
app_href(48,919,289,1).
app_href(48,919,118,1).
app_href(48,919,1282,1).
app_href(48,919,224,1).
app_href(48,919,684,1).
app_href(48,919,682,1).
app_href(48,919,288,4).
app_href(48,919,586,1).
app_href(48,919,1178,1).
app_href(48,919,546,1).
app_href(48,919,90,1).
app_href(48,919,88,18).
app_href(48,919,843,2).
app_href(48,919,41,7).
app_href(48,919,37,8).
app_href(48,919,26,3).
app_href(48,919,9,3).
app_href(48,919,411,1).
app_href(48,919,360,1).
app_href(48,919,663,1).
app_href(48,919,1346,2).
app_href(48,919,1430,1).
app_href(48,919,335,1).
app_href(48,919,1146,1).
app_href(48,919,376,1).
app_href(48,919,235,1).
app_href(48,919,1054,1).
app_href(48,919,1153,1).
app_href(48,919,140,2).
app_href(48,919,864,1).
app_href(48,919,894,1).
app_href(48,919,844,5).
app_href(48,919,1421,1).
app_href(48,919,1067,1).
app_href(48,920,1063,1).
app_href(48,920,949,2).
app_href(51,147,736,1).
app_href(51,147,941,1).
app_href(51,147,2,1).
app_href(51,147,1360,1).
app_href(51,148,1486,1).
app_href(51,148,1014,1).
app_href(51,149,830,1).
app_href(51,150,830,1).
app_href(51,150,1256,1).
app_href(51,150,2,1).
app_href(51,15,899,1).
app_href(51,15,1310,1).
app_href(51,15,668,1).
app_href(51,154,2,1).
app_href(51,152,1370,3).
app_href(51,152,992,1).
app_href(51,152,1133,1).
app_href(51,21,443,1).
app_href(51,21,277,1).
app_href(51,161,755,1).
app_href(51,162,1364,1).
app_href(51,162,2,1).
app_href(51,163,489,1).
app_href(51,164,454,1).
app_href(51,165,867,1).
app_href(51,165,1336,1).
app_href(51,165,728,2).
app_href(51,166,736,1).
app_href(51,166,231,1).
app_href(51,167,1418,1).
app_href(51,167,423,1).
app_href(51,167,2,1).
app_href(51,168,1069,1).
app_href(51,87,1485,1).
app_href(51,87,444,3).
app_href(51,87,736,1).
app_href(51,87,358,1).
app_href(51,87,1330,1).
app_href(51,87,588,1).
app_href(51,87,1014,1).
app_href(51,87,970,1).
app_href(51,170,1192,1).
app_href(51,170,132,1).
app_href(51,171,944,1).
app_href(51,171,673,1).
app_href(51,400,444,1).
app_href(51,400,155,1).
app_href(51,400,736,1).
app_href(51,400,816,1).
app_href(51,172,1076,1).
app_href(51,172,1053,1).
app_href(51,173,1336,1).
app_href(51,175,146,1).
app_href(51,175,2,1).
app_href(51,177,1403,1).
app_href(51,177,1341,1).
app_href(51,177,970,1).
app_href(51,178,829,1).
app_href(51,293,616,1).
app_href(51,293,1443,1).
app_href(51,293,420,1).
app_href(51,293,736,1).
app_href(51,293,1276,1).
app_href(51,181,354,1).
app_href(51,182,1218,1).
app_href(51,182,420,1).
app_href(51,183,444,1).
app_href(51,184,317,1).
app_href(51,184,677,1).
app_href(51,185,1442,1).
app_href(51,185,736,1).
app_href(51,185,140,1).
app_href(51,185,233,1).
app_href(51,186,728,1).
app_href(51,186,454,1).
app_href(51,62,1138,1).
app_href(51,62,959,1).
app_href(51,187,444,1).
app_href(51,187,728,1).
app_href(51,921,616,1).
app_href(51,921,420,1).
app_href(51,921,736,1).
app_href(51,921,1047,1).
app_href(51,190,316,1).
app_href(51,190,736,1).
app_href(51,190,1083,1).
app_href(55,904,1123,1).
app_href(55,904,826,1).
app_href(55,904,141,1).
app_href(55,904,736,1).
app_href(55,904,1346,1).
app_href(55,904,139,1).
app_href(55,904,288,1).
app_href(55,1023,826,1).
app_href(55,1023,1194,1).
app_href(55,1023,1434,1).
app_href(55,1023,736,1).
app_href(55,1023,9,1).
app_href(55,1023,1346,1).
app_href(55,1023,480,1).
app_href(55,1023,288,1).
app_href(55,506,242,1).
app_href(55,506,393,1).
app_href(55,506,2,1).
app_href(55,148,1486,1).
app_href(55,148,1014,1).
app_href(55,149,830,1).
app_href(55,15,899,1).
app_href(55,15,1310,1).
app_href(55,15,668,1).
app_href(55,152,1370,3).
app_href(55,152,992,1).
app_href(55,152,1133,1).
app_href(55,161,755,1).
app_href(55,129,1473,1).
app_href(55,129,2,1).
app_href(55,507,118,1).
app_href(55,507,477,1).
app_href(55,507,2,1).
app_href(55,163,489,1).
app_href(55,508,726,1).
app_href(55,508,2,1).
app_href(55,164,454,1).
app_href(38,794,1346,1).
app_href(38,794,296,1).
app_href(38,168,1069,1).
app_href(38,795,507,1).
app_href(38,796,1378,1).
app_href(38,509,1443,1).
app_href(38,509,1218,1).
app_href(38,509,2,1).
app_href(38,797,428,1).
app_href(38,170,1192,1).
app_href(38,170,132,1).
app_href(38,171,944,1).
app_href(38,171,673,1).
app_href(38,798,188,1).
app_href(38,798,1346,1).
app_href(38,798,115,1).
app_href(38,172,1076,1).
app_href(38,172,1053,1).
app_href(38,173,1336,1).
app_href(38,799,671,1).
app_href(38,799,1367,1).
app_href(38,799,128,1).
app_href(38,799,433,1).
app_href(38,799,232,1).
app_href(38,799,3,1).
app_href(38,799,373,1).
app_href(38,799,667,1).
app_href(38,799,521,1).
app_href(38,799,1349,1).
app_href(38,799,1253,1).
app_href(38,799,970,1).
app_href(38,511,1183,1).
app_href(38,511,296,1).
app_href(38,511,2,1).
app_href(38,800,1062,1).
app_href(38,175,146,1).
app_href(38,175,2,1).
app_href(38,177,1403,1).
app_href(38,177,1341,1).
app_href(38,177,970,1).
app_href(38,178,829,1).
app_href(38,513,1459,1).
app_href(38,513,2,1).
app_href(38,181,354,1).
app_href(38,801,232,1).
app_href(38,801,358,1).
app_href(38,801,970,1).
app_href(38,182,1200,1).
app_href(38,182,1218,1).
app_href(38,182,420,1).
app_href(38,182,134,1).
app_href(38,182,2,1).
app_href(38,514,575,1).
app_href(38,514,2,1).
app_href(38,183,444,1).
app_href(38,183,1200,1).
app_href(38,183,134,1).
app_href(38,183,2,1).
app_href(38,515,999,1).
app_href(38,515,2,1).
app_href(38,185,1442,1).
app_href(38,185,736,1).
app_href(38,185,140,1).
app_href(38,185,233,1).
app_href(38,184,317,1).
app_href(38,184,677,1).
app_href(38,52,1385,1).
app_href(38,52,2,1).
app_href(38,517,1318,1).
app_href(38,517,2,1).
app_href(38,518,725,1).
app_href(38,518,2,1).
app_href(38,802,188,1).
app_href(38,802,1346,1).
app_href(38,802,1460,1).
app_href(38,802,709,1).
app_href(38,190,316,1).
app_href(38,190,736,1).
app_href(38,190,1083,1).
app_href(1068,73,145,2).
app_href(1068,73,941,2).
app_href(1068,73,950,2).
app_href(1068,803,999,1).
app_href(1068,803,277,1).
app_href(1068,804,1357,1).
app_href(1068,804,1382,1).
app_href(1068,91,736,1).
app_href(1068,91,718,1).
app_href(1068,35,365,1).
app_href(1068,557,546,1).
app_href(1068,557,736,1).
app_href(1068,557,718,1).
app_href(1068,87,444,1).
app_href(1068,87,358,1).
app_href(1068,805,546,1).
app_href(1068,805,736,1).
app_href(1068,805,540,1).
app_href(1068,805,970,1).
app_href(1068,805,718,1).
app_href(1068,753,736,1).
app_href(1068,753,950,1).
app_href(1068,806,281,1).
app_href(1068,806,1292,1).
app_href(1068,15,899,1).
app_href(1068,15,1310,1).
app_href(1068,15,668,1).
app_href(52,807,37,1).
app_href(52,807,1022,1).
app_href(52,807,155,1).
app_href(52,807,246,1).
app_href(52,807,139,1).
app_href(52,808,371,1).
app_href(52,808,1226,1).
app_href(52,808,740,1).
app_href(52,809,1022,1).
app_href(52,809,11,1).
app_href(52,809,1384,1).
app_href(52,809,740,1).
app_href(52,810,660,1).
app_href(52,810,736,1).
app_href(52,810,1384,1).
app_href(52,810,132,1).
app_href(52,811,1468,1).
app_href(52,811,736,1).
app_href(52,811,662,1).
app_href(52,811,1123,1).
app_href(52,811,156,1).
app_href(52,811,564,1).
app_href(52,811,480,1).
app_href(52,811,1346,1).
app_href(52,811,26,1).
app_href(52,811,504,2).
app_href(52,812,26,1).
app_href(52,812,736,1).
app_href(52,812,480,1).
app_href(52,812,504,1).
app_href(52,813,654,1).
app_href(52,813,1022,1).
app_href(52,813,11,1).
app_href(52,813,1384,1).
app_href(52,813,740,1).
app_href(52,814,1041,1).
app_href(52,814,725,1).
app_href(52,814,975,1).
app_href(52,814,1310,1).
app_href(52,814,1346,2).
app_href(52,814,564,1).
app_href(52,814,1384,1).
app_href(52,815,214,1).
app_href(52,518,725,1).
app_href(52,518,2,1).
app_href(52,588,1200,1).
app_href(52,588,1218,1).
app_href(52,588,420,1).
app_href(52,588,134,1).
app_href(52,588,2,1).
app_href(52,816,871,1).
app_href(50,949,452,1).
app_href(50,949,2,1).
app_href(50,171,944,1).
app_href(50,171,673,1).
app_href(50,950,145,1).
app_href(50,950,904,1).
app_href(50,950,520,1).
app_href(50,950,590,1).
app_href(50,951,1290,1).
app_href(50,951,1209,1).
app_href(50,951,736,1).
app_href(50,951,1346,1).
app_href(50,952,616,1).
app_href(50,952,1406,1).
app_href(50,952,199,1).
app_href(50,952,1345,1).
app_href(50,952,1323,1).
app_href(50,952,702,1).
app_href(50,952,1276,1).
app_href(50,953,1395,1).
app_href(50,172,1076,1).
app_href(50,172,1053,1).
app_href(50,170,1192,1).
app_href(50,170,132,1).
app_href(50,589,478,1).
app_href(50,589,975,1).
app_href(50,589,1310,1).
app_href(50,589,1300,1).
app_href(50,589,504,1).
app_href(50,954,118,1).
app_href(50,954,1358,1).
app_href(50,954,766,1).
app_href(50,954,393,1).
app_href(50,166,736,1).
app_href(50,166,231,1).
app_href(50,437,701,1).
app_href(50,437,736,1).
app_href(50,437,927,1).
app_href(50,148,1486,1).
app_href(50,148,1014,1).
app_href(50,955,536,1).
app_href(50,955,2,1).
app_href(50,184,317,1).
app_href(50,184,677,1).
app_href(50,190,316,1).
app_href(50,190,736,1).
app_href(50,190,1083,1).
app_href(50,956,486,1).
app_href(50,956,142,1).
app_href(50,957,667,1).
app_href(50,957,9,2).
app_href(50,957,1198,1).
app_href(50,163,489,1).
app_href(50,958,976,1).
app_href(50,958,1155,1).
app_href(50,959,1346,1).
app_href(50,959,883,1).
app_href(50,959,530,1).
app_href(50,591,155,1).
app_href(50,591,975,1).
app_href(50,591,1310,1).
app_href(50,591,1212,1).
app_href(50,960,898,1).
app_href(50,961,152,1).
app_href(50,961,1468,1).
app_href(50,962,137,1).
app_href(50,962,309,1).
app_href(50,963,428,1).
app_href(50,964,221,1).
app_href(50,964,672,1).
app_href(50,964,1118,1).
app_href(50,185,1442,1).
app_href(50,185,736,1).
app_href(50,185,140,1).
app_href(50,185,233,1).
app_href(50,965,428,1).
app_href(50,966,550,1).
app_href(50,966,733,1).
app_href(50,967,1484,1).
app_href(50,967,541,1).
app_href(50,968,863,1).
app_href(50,969,510,1).
app_href(50,969,289,1).
app_href(50,969,736,1).
app_href(50,970,649,1).
app_href(50,462,303,1).
app_href(50,164,454,1).
app_href(50,193,1277,1).
app_href(50,193,602,1).
app_href(50,971,463,1).
app_href(50,971,457,1).
app_href(50,971,140,1).
app_href(50,425,611,1).
app_href(50,425,36,1).
app_href(50,199,380,1).
app_href(50,199,736,1).
app_href(50,199,1477,1).
app_href(50,199,970,1).
app_href(50,199,555,1).
app_href(50,972,428,1).
app_href(54,973,1061,1).
app_href(54,974,145,1).
app_href(54,974,777,1).
app_href(54,975,1217,2).
app_href(54,975,1346,1).
app_href(54,975,369,2).
app_href(54,975,922,2).
app_href(54,975,1102,2).
app_href(54,35,365,1).
app_href(54,35,6,1).
app_href(54,976,1356,1).
app_href(54,976,1329,1).
app_href(54,976,448,1).
app_href(54,976,190,1).
app_href(54,976,139,1).
app_href(54,976,438,3).
app_href(54,976,1217,3).
app_href(54,976,1346,1).
app_href(54,976,431,1).
app_href(54,976,705,1).
app_href(54,976,1102,2).
app_href(54,976,922,1).
app_href(54,976,970,1).
app_href(54,976,361,1).
app_href(54,977,1217,1).
app_href(54,977,1264,1).
app_href(54,977,145,1).
app_href(54,977,1103,1).
app_href(54,977,140,1).
app_href(54,977,777,2).
app_href(54,977,1267,1).
app_href(54,978,777,1).
app_href(54,978,1267,1).
app_href(54,979,1217,1).
app_href(54,979,1103,1).
app_href(54,15,899,1).
app_href(54,15,1310,1).
app_href(54,15,668,1).
app_href(54,87,444,1).
app_href(54,87,358,1).
app_href(54,980,1160,1).
app_href(54,980,33,1).
app_href(54,981,763,1).
app_href(54,981,818,1).
app_href(54,981,922,1).
app_href(54,981,1280,1).
app_href(54,982,1160,1).
app_href(54,982,31,1).
app_href(54,983,240,1).
app_href(54,983,407,1).
app_href(54,983,1206,1).
app_href(54,983,921,1).
app_href(54,983,1014,1).
app_href(54,983,777,1).
app_href(54,983,191,1).
app_href(54,983,634,1).
app_href(54,984,1217,1).
app_href(52,816,817,1).
app_href(52,816,483,1).
app_href(52,15,899,1).
app_href(52,15,1310,1).
app_href(52,15,668,1).
app_href(52,817,1022,1).
app_href(52,817,155,1).
app_href(52,817,246,1).
app_href(52,817,9,1).
app_href(52,817,139,1).
app_href(52,818,1346,1).
app_href(52,818,777,1).
app_href(52,818,1384,1).
app_href(52,181,354,1).
app_href(52,819,736,1).
app_href(52,819,594,1).
app_href(52,819,1198,1).
app_href(52,819,970,1).
app_href(52,820,736,1).
app_href(52,820,1343,1).
app_href(52,820,1301,1).
app_href(52,820,36,1).
app_href(52,149,830,1).
app_href(52,168,1069,1).
app_href(52,821,725,1).
app_href(52,821,1346,1).
app_href(52,821,1384,1).
app_href(52,822,1022,2).
app_href(52,822,736,2).
app_href(52,822,511,2).
app_href(52,822,616,2).
app_href(52,822,810,2).
app_href(52,822,621,2).
app_href(52,822,1346,2).
app_href(52,822,26,2).
app_href(52,822,1093,2).
app_href(52,822,970,2).
app_href(52,823,1022,2).
app_href(52,823,736,2).
app_href(52,823,511,2).
app_href(52,823,616,2).
app_href(52,823,810,2).
app_href(52,823,621,2).
app_href(52,823,1346,2).
app_href(52,823,1093,2).
app_href(52,823,9,2).
app_href(52,823,970,2).
app_href(52,824,1406,1).
app_href(52,824,610,1).
app_href(52,824,702,1).
app_href(52,824,618,1).
app_href(52,152,1370,3).
app_href(52,152,992,1).
app_href(52,152,1133,1).
app_href(52,825,1384,1).
app_href(52,825,740,1).
app_href(52,825,504,1).
app_href(52,511,1183,1).
app_href(52,511,296,1).
app_href(52,511,2,1).
app_href(52,826,435,1).
app_href(52,826,326,1).
app_href(52,183,444,1).
app_href(52,183,1200,1).
app_href(52,183,134,1).
app_href(52,183,2,1).
app_href(52,177,1403,1).
app_href(52,177,1341,1).
app_href(52,177,970,1).
app_href(52,827,511,1).
app_href(52,827,736,1).
app_href(52,827,1346,1).
app_href(52,827,1093,1).
app_href(52,827,970,1).
app_href(52,827,810,1).
app_href(52,161,755,1).
app_href(52,828,1022,1).
app_href(52,828,11,1).
app_href(52,828,653,1).
app_href(52,828,1384,1).
app_href(52,828,740,1).
app_href(52,173,1336,1).
app_href(52,945,736,1).
app_href(52,945,1346,1).
app_href(52,945,1384,1).
app_href(52,178,829,1).
app_href(52,829,581,1).
app_href(52,829,1346,1).
app_href(52,829,303,1).
app_href(52,829,711,1).
app_href(52,829,1360,1).
app_href(52,509,1443,1).
app_href(52,509,1218,1).
app_href(52,509,2,1).
app_href(52,830,371,1).
app_href(52,830,1226,1).
app_href(52,830,26,1).
app_href(52,830,594,1).
app_href(52,830,766,1).
app_href(52,830,8,1).
app_href(52,515,999,1).
app_href(52,515,2,1).
app_href(52,831,1300,1).
app_href(52,831,571,1).
app_href(52,831,736,1).
app_href(52,831,573,1).
app_href(52,831,910,1).
app_href(52,831,975,1).
app_href(52,831,1310,1).
app_href(52,831,564,2).
app_href(52,831,644,1).
app_href(52,831,422,1).
app_href(52,831,1346,1).
app_href(52,831,41,1).
app_href(52,831,37,1).
app_href(52,831,8,1).
app_href(52,832,40,1).
app_href(52,832,736,1).
app_href(52,832,145,1).
app_href(52,832,1384,1).
app_href(52,50,1209,1).
app_href(52,50,736,2).
app_href(52,50,1346,1).
app_href(52,50,1384,1).
app_href(52,833,1198,1).
app_href(52,514,575,1).
app_href(52,514,2,1).
app_href(52,834,371,1).
app_href(52,834,1226,1).
app_href(52,834,766,1).
app_href(52,171,944,1).
app_href(52,171,673,1).
app_href(52,129,1473,1).
app_href(52,129,2,1).
app_href(52,52,1385,1).
app_href(52,52,2,1).
app_href(52,835,1443,2).
app_href(52,835,1218,2).
app_href(52,835,736,1).
app_href(52,835,1346,1).
app_href(52,835,1384,2).
app_href(52,836,41,1).
app_href(52,836,1022,1).
app_href(52,836,155,1).
app_href(52,836,246,1).
app_href(52,836,139,1).
app_href(52,172,1076,1).
app_href(52,172,1053,1).
app_href(52,837,454,1).
app_href(52,517,1318,1).
app_href(52,517,2,1).
app_href(52,170,1192,1).
app_href(52,170,132,1).
app_href(52,166,736,1).
app_href(52,166,231,1).
app_href(52,838,367,1).
app_href(52,838,711,1).
app_href(54,984,1134,1).
app_href(54,984,1346,1).
app_href(54,984,1162,1).
app_href(54,984,1102,1).
app_href(54,985,1030,1).
app_href(54,986,777,1).
app_href(54,986,1384,1).
app_href(54,987,1030,1).
app_href(54,988,1160,1).
app_href(54,988,35,1).
app_href(54,772,838,1).
app_href(54,772,217,1).
app_href(54,989,914,1).
app_href(54,989,1403,1).
app_href(54,989,1400,1).
app_href(54,989,970,1).
app_href(54,990,777,1).
app_href(54,990,1384,1).
app_href(54,991,1030,1).
app_href(54,992,1309,1).
app_href(54,992,523,1).
app_href(54,993,1061,1).
app_href(54,994,1279,1).
app_href(54,994,763,1).
app_href(54,994,818,1).
app_href(54,994,922,1).
app_href(54,475,1217,1).
app_href(54,475,688,1).
app_href(54,475,707,1).
app_href(54,475,100,1).
app_href(54,475,1264,1).
app_href(54,475,922,1).
app_href(54,390,331,1).
app_href(54,390,438,1).
app_href(54,390,190,1).
app_href(54,390,1217,3).
app_href(54,390,1346,1).
app_href(54,390,444,1).
app_href(54,390,431,1).
app_href(54,390,1102,2).
app_href(54,390,1427,1).
app_href(54,390,970,1).
app_href(54,390,818,3).
app_href(54,390,1055,1).
app_href(54,995,445,1).
app_href(54,995,1136,1).
app_href(54,996,1160,1).
app_href(54,996,34,1).
app_href(54,997,1061,1).
app_href(54,998,778,1).
app_href(54,998,837,1).
app_href(56,506,242,1).
app_href(56,506,393,1).
app_href(56,506,2,1).
app_href(56,999,1022,1).
app_href(56,999,156,1).
app_href(56,999,160,1).
app_href(56,999,333,1).
app_href(56,999,140,1).
app_href(56,999,1242,1).
app_href(56,999,1444,1).
app_href(56,999,1346,1).
app_href(56,999,725,1).
app_href(56,999,26,1).
app_href(56,999,1360,1).
app_href(56,999,245,1).
app_href(56,999,1408,1).
app_href(56,999,504,1).
app_href(56,148,1486,1).
app_href(56,148,1014,1).
app_href(56,149,830,1).
app_href(56,633,667,1).
app_href(56,633,1243,1).
app_href(56,633,26,1).
app_href(56,633,145,1).
app_href(56,633,1346,1).
app_href(56,633,8,1).
app_href(56,633,795,1).
app_href(56,1000,331,1).
app_href(56,1000,420,1).
app_href(56,1000,736,1).
app_href(56,15,899,1).
app_href(56,15,1310,1).
app_href(56,15,668,1).
app_href(56,587,1462,1).
app_href(56,587,1337,1).
app_href(56,587,736,1).
app_href(56,152,1370,3).
app_href(56,152,992,1).
app_href(56,152,1133,1).
app_href(56,161,755,1).
app_href(56,129,1473,1).
app_href(56,129,2,1).
app_href(56,638,911,1).
app_href(56,638,540,1).
app_href(56,507,118,1).
app_href(56,507,477,1).
app_href(56,507,2,1).
app_href(56,163,489,1).
app_href(56,508,726,1).
app_href(56,508,2,1).
app_href(56,164,454,1).
app_href(56,165,728,1).
app_href(56,165,2,1).
app_href(56,166,736,1).
app_href(56,166,231,1).
app_href(56,168,1069,1).
app_href(56,1001,127,1).
app_href(56,509,1443,1).
app_href(56,509,1218,1).
app_href(56,509,2,1).
app_href(56,170,1192,1).
app_href(56,170,132,1).
app_href(56,171,944,1).
app_href(56,171,673,1).
app_href(56,1002,1350,1).
app_href(56,1002,1390,1).
app_href(56,172,1076,1).
app_href(56,172,1053,1).
app_href(56,173,1336,1).
app_href(56,657,377,1).
app_href(56,657,564,1).
app_href(56,657,305,1).
app_href(56,657,1366,1).
app_href(56,657,1428,1).
app_href(56,511,1183,1).
app_href(56,511,296,1).
app_href(56,511,2,1).
app_href(56,175,146,1).
app_href(56,175,2,1).
app_href(56,177,1403,1).
app_href(56,177,1341,1).
app_href(56,177,970,1).
app_href(56,178,829,1).
app_href(56,513,1459,1).
app_href(56,513,2,1).
app_href(56,1003,309,1).
app_href(56,1003,1310,1).
app_href(56,1003,736,1).
app_href(56,181,354,1).
app_href(56,181,1015,1).
app_href(56,181,629,1).
app_href(56,182,1200,1).
app_href(56,182,1218,1).
app_href(56,182,420,1).
app_href(56,182,134,1).
app_href(56,182,2,1).
app_href(56,514,575,1).
app_href(56,514,2,1).
app_href(56,183,444,1).
app_href(56,183,1200,1).
app_href(56,183,134,1).
app_href(56,183,2,1).
app_href(56,515,999,1).
app_href(56,515,2,1).
app_href(56,185,1442,1).
app_href(56,185,736,1).
app_href(52,838,1360,1).
app_href(52,148,1486,1).
app_href(52,148,1014,1).
app_href(52,839,1041,1).
app_href(52,839,26,1).
app_href(52,839,736,1).
app_href(52,839,1346,1).
app_href(52,839,564,1).
app_href(52,839,1093,1).
app_href(52,839,132,1).
app_href(52,1086,309,1).
app_href(52,1086,774,1).
app_href(52,184,317,1).
app_href(52,184,677,1).
app_href(52,1087,594,1).
app_href(52,1087,197,1).
app_href(52,1088,736,1).
app_href(52,1088,594,1).
app_href(52,1088,1219,1).
app_href(52,1088,970,1).
app_href(52,190,316,1).
app_href(52,190,736,1).
app_href(52,190,1083,1).
app_href(52,1089,1022,1).
app_href(52,1089,155,1).
app_href(52,1089,26,1).
app_href(52,1089,246,1).
app_href(52,1089,139,1).
app_href(52,1090,1346,1).
app_href(52,1090,1323,1).
app_href(52,1090,1384,1).
app_href(52,1091,506,1).
app_href(52,1091,619,1).
app_href(52,1091,736,1).
app_href(52,1091,941,1).
app_href(52,1091,564,1).
app_href(52,1091,1101,1).
app_href(52,1092,371,1).
app_href(52,1092,511,1).
app_href(52,1092,1226,1).
app_href(52,1092,810,1).
app_href(52,163,489,1).
app_href(52,1093,725,1).
app_href(52,1093,126,1).
app_href(52,1093,1346,1).
app_href(52,1093,1384,1).
app_href(52,1094,1015,1).
app_href(52,1094,594,1).
app_href(52,1094,759,1).
app_href(52,1095,754,1).
app_href(52,1095,1346,1).
app_href(52,1095,594,1).
app_href(52,1095,970,1).
app_href(52,175,146,1).
app_href(52,175,2,1).
app_href(52,1096,145,1).
app_href(52,1096,1346,1).
app_href(52,1096,740,1).
app_href(52,1096,711,1).
app_href(52,1096,1360,1).
app_href(52,1097,1022,1).
app_href(52,1097,653,1).
app_href(52,1097,594,1).
app_href(52,1097,1323,1).
app_href(52,1097,970,1).
app_href(52,1097,772,1).
app_href(52,1098,1022,1).
app_href(52,1098,594,1).
app_href(52,1098,1323,1).
app_href(52,1098,970,1).
app_href(52,1098,772,1).
app_href(52,185,1442,1).
app_href(52,185,736,1).
app_href(52,185,140,1).
app_href(52,185,233,1).
app_href(52,508,726,1).
app_href(52,508,2,1).
app_href(52,506,242,1).
app_href(52,506,393,1).
app_href(52,506,2,1).
app_href(52,1099,1337,1).
app_href(52,1099,1310,1).
app_href(52,1099,185,1).
app_href(52,1100,594,1).
app_href(52,1100,1162,1).
app_href(52,1100,970,1).
app_href(52,1101,611,1).
app_href(52,1101,140,1).
app_href(52,1101,884,1).
app_href(52,1102,1346,1).
app_href(52,1102,776,1).
app_href(52,1102,1384,1).
app_href(52,165,728,1).
app_href(52,165,2,1).
app_href(52,507,118,1).
app_href(52,507,477,1).
app_href(52,507,2,1).
app_href(52,1103,1337,1).
app_href(52,1103,594,1).
app_href(52,1103,484,1).
app_href(52,1103,740,1).
app_href(52,1103,504,1).
app_href(52,513,1459,1).
app_href(52,513,2,1).
app_href(52,164,454,1).
app_href(52,1104,736,1).
app_href(52,1104,1346,1).
app_href(52,1104,1384,1).
app_href(52,1105,1041,1).
app_href(52,1105,1043,1).
app_href(52,1105,26,2).
app_href(52,1105,736,2).
app_href(52,1105,1346,2).
app_href(52,1105,564,2).
app_href(52,1105,1093,2).
app_href(52,1105,132,2).
app_href(52,1106,303,1).
app_href(52,1107,1041,1).
app_href(52,1107,725,1).
app_href(52,1107,1346,1).
app_href(52,1107,156,1).
app_href(52,1107,422,1).
app_href(52,1107,1408,1).
app_href(52,1107,504,1).
app_href(57,1108,736,1).
app_href(57,1108,288,1).
app_href(57,1108,1199,1).
app_href(57,1108,809,1).
app_href(57,35,365,1).
app_href(57,35,6,1).
app_href(57,350,1011,1).
app_href(57,350,736,1).
app_href(57,350,1198,1).
app_href(57,87,444,1).
app_href(57,87,358,1).
app_href(57,1109,667,1).
app_href(57,1109,874,1).
app_href(57,1109,851,1).
app_href(57,1109,1155,1).
app_href(57,1109,120,1).
app_href(57,1110,1409,1).
app_href(57,1110,736,1).
app_href(57,1110,288,1).
app_href(57,1110,1154,1).
app_href(57,467,736,1).
app_href(57,467,288,2).
app_href(57,467,809,1).
app_href(57,697,736,1).
app_href(57,697,291,1).
app_href(57,697,288,1).
app_href(57,697,809,1).
app_href(57,15,899,1).
app_href(57,15,1310,1).
app_href(57,15,668,1).
app_href(57,1014,736,1).
app_href(57,1014,288,1).
app_href(57,1014,1399,1).
app_href(53,747,1479,1).
app_href(53,1018,161,1).
app_href(53,1018,736,1).
app_href(53,1018,540,1).
app_href(53,1018,733,1).
app_href(53,761,1409,2).
app_href(53,761,1065,2).
app_href(53,761,736,2).
app_href(58,87,444,1).
app_href(58,87,358,1).
app_href(58,1019,1030,1).
app_href(58,1020,145,1).
app_href(58,1020,314,1).
app_href(58,1020,447,1).
app_href(58,921,616,1).
app_href(58,921,420,1).
app_href(58,921,736,1).
app_href(58,921,1047,1).
app_href(58,1021,572,1).
app_href(58,1021,309,1).
app_href(58,1021,643,1).
app_href(58,1021,642,1).
app_href(58,1021,1310,1).
app_href(58,1021,736,1).
app_href(58,1021,442,1).
app_href(58,1021,1049,1).
app_href(58,35,365,1).
app_href(58,35,6,1).
app_href(58,15,899,1).
app_href(58,15,1310,1).
app_href(58,15,668,1).
app_href(60,903,517,1).
app_href(60,903,26,1).
app_href(60,903,9,1).
app_href(60,903,8,1).
app_href(60,903,769,1).
app_href(60,904,480,1).
app_href(60,904,555,1).
app_href(60,49,517,1).
app_href(60,49,9,2).
app_href(60,49,769,1).
app_href(60,1022,253,1).
app_href(60,1022,1312,1).
app_href(60,1022,94,1).
app_href(60,1022,1479,1).
app_href(60,1023,1194,1).
app_href(60,1023,9,1).
app_href(60,1023,1436,1).
app_href(60,1023,480,1).
app_href(60,1024,1386,1).
app_href(60,1024,1117,1).
app_href(60,1025,232,1).
app_href(61,1026,365,1).
app_href(61,1026,6,1).
app_href(61,1027,358,1).
app_href(61,1027,769,1).
app_href(62,1028,81,1).
app_href(62,1028,959,1).
app_href(62,1028,836,1).
app_href(62,182,1218,1).
app_href(62,182,420,1).
app_href(62,1029,242,1).
app_href(62,1029,58,1).
app_href(62,1030,82,1).
app_href(62,1030,146,1).
app_href(62,1030,941,1).
app_href(62,1030,950,1).
app_href(62,4,1337,1).
app_href(62,4,736,1).
app_href(62,4,1139,1).
app_href(62,4,1047,1).
app_href(62,4,727,1).
app_href(62,4,140,1).
app_href(62,4,1070,1).
app_href(62,4,318,1).
app_href(62,4,41,1).
app_href(62,4,9,2).
app_href(62,4,219,1).
app_href(62,1031,309,1).
app_href(62,1031,643,1).
app_href(62,1031,1470,1).
app_href(62,1031,966,1).
app_href(62,186,728,1).
app_href(62,186,454,1).
app_href(62,1032,1337,1).
app_href(62,1032,736,1).
app_href(62,1032,68,1).
app_href(62,1032,36,1).
app_href(62,1033,1160,1).
app_href(62,1033,649,1).
app_href(62,1033,367,1).
app_href(62,1033,1168,1).
app_href(62,1033,736,2).
app_href(62,1033,140,1).
app_href(62,1033,32,1).
app_href(62,1034,61,1).
app_href(62,1034,242,1).
app_href(62,1035,1198,1).
app_href(62,1035,78,1).
app_href(62,1036,151,1).
app_href(62,1036,1041,1).
app_href(62,1036,1355,1).
app_href(62,1036,51,1).
app_href(62,1036,959,1).
app_href(62,1036,973,1).
app_href(62,1036,1481,1).
app_href(62,1036,419,1).
app_href(62,1036,1456,1).
app_href(62,1036,897,1).
app_href(62,15,899,1).
app_href(62,15,1310,1).
app_href(62,15,668,1).
app_href(62,1037,309,1).
app_href(62,1037,1151,2).
app_href(62,1037,1470,1).
app_href(62,1037,1207,1).
app_href(62,1037,732,1).
app_href(62,1037,959,1).
app_href(62,1037,194,1).
app_href(62,1037,973,1).
app_href(62,1037,828,1).
app_href(62,1037,846,1).
app_href(62,1037,1095,1).
app_href(62,1037,808,1).
app_href(62,1037,195,1).
app_href(62,1037,643,1).
app_href(62,181,354,1).
app_href(62,19,1138,1).
app_href(62,19,667,1).
app_href(62,19,41,1).
app_href(62,19,557,1).
app_href(62,19,242,1).
app_href(62,19,681,1).
app_href(62,19,9,2).
app_href(62,19,1346,1).
app_href(62,1038,1344,1).
app_href(62,21,736,1).
app_href(62,21,140,1).
app_href(62,21,1189,1).
app_href(62,21,667,1).
app_href(62,21,41,1).
app_href(62,21,277,2).
app_href(62,21,26,1).
app_href(62,21,9,2).
app_href(62,21,503,1).
app_href(62,149,830,1).
app_href(62,168,1069,1).
app_href(62,1039,242,1).
app_href(62,1039,1426,1).
app_href(62,1039,1397,1).
app_href(62,1039,557,1).
app_href(62,1039,667,1).
app_href(62,1039,41,1).
app_href(62,1039,1138,1).
app_href(62,1039,9,2).
app_href(62,1039,1306,1).
app_href(62,167,1418,1).
app_href(57,1014,934,1).
app_href(59,522,536,1).
app_href(59,522,1033,1).
app_href(59,35,365,1).
app_href(59,35,6,1).
app_href(59,1111,616,1).
app_href(59,1111,1186,1).
app_href(59,1111,140,1).
app_href(59,1111,1033,1).
app_href(59,525,991,1).
app_href(59,525,736,1).
app_href(59,525,1433,1).
app_href(59,525,639,1).
app_href(59,473,1091,1).
app_href(59,473,666,1).
app_href(59,473,1266,1).
app_href(59,473,522,1).
app_href(59,473,1360,1).
app_href(59,1112,563,1).
app_href(59,1112,140,1).
app_href(59,1112,666,1).
app_href(59,1112,1355,1).
app_href(59,1112,889,1).
app_href(59,1112,1281,1).
app_href(59,1112,1360,1).
app_href(59,87,444,1).
app_href(59,87,358,1).
app_href(59,1113,1041,1).
app_href(59,1113,444,1).
app_href(59,1113,736,1).
app_href(59,1113,1346,1).
app_href(59,1113,564,1).
app_href(59,1113,1033,1).
app_href(59,15,899,1).
app_href(59,15,1310,1).
app_href(59,15,668,1).
app_href(67,147,736,1).
app_href(67,147,941,1).
app_href(67,147,2,1).
app_href(67,147,1360,1).
app_href(67,148,1486,1).
app_href(67,148,1014,1).
app_href(67,149,830,1).
app_href(67,1114,695,1).
app_href(67,1114,91,1).
app_href(67,1114,690,1).
app_href(67,150,830,1).
app_href(67,150,1256,1).
app_href(67,150,2,1).
app_href(67,15,899,1).
app_href(67,15,1310,1).
app_href(67,15,668,1).
app_href(67,154,2,1).
app_href(67,152,1370,3).
app_href(67,152,992,1).
app_href(67,152,1133,1).
app_href(67,161,755,1).
app_href(67,162,1364,1).
app_href(67,162,2,1).
app_href(67,163,489,1).
app_href(67,1115,1248,1).
app_href(67,1115,736,1).
app_href(67,1115,561,1).
app_href(67,1115,1150,1).
app_href(67,1115,695,1).
app_href(67,1115,690,1).
app_href(67,1115,1189,1).
app_href(67,1115,26,1).
app_href(67,1115,1011,1).
app_href(67,164,454,1).
app_href(67,165,867,1).
app_href(67,165,1336,1).
app_href(67,165,728,2).
app_href(67,166,736,1).
app_href(67,166,231,1).
app_href(67,167,1418,1).
app_href(67,167,423,1).
app_href(67,167,2,1).
app_href(67,168,1069,1).
app_href(67,169,223,4).
app_href(67,169,695,3).
app_href(67,169,1150,2).
app_href(67,169,1443,1).
app_href(67,169,1329,2).
app_href(67,169,1276,1).
app_href(67,169,1189,1).
app_href(67,169,1346,1).
app_href(67,169,736,4).
app_href(67,169,970,5).
app_href(67,169,1012,3).
app_href(67,169,403,1).
app_href(67,169,1202,1).
app_href(67,169,1047,1).
app_href(67,169,690,8).
app_href(67,169,26,1).
app_href(67,169,1203,4).
app_href(67,169,1041,1).
app_href(67,4,736,1).
app_href(67,4,309,1).
app_href(67,4,642,1).
app_href(67,4,1310,1).
app_href(67,4,318,1).
app_href(67,4,41,1).
app_href(67,4,728,1).
app_href(67,4,9,2).
app_href(67,4,643,1).
app_href(67,170,1192,1).
app_href(67,170,132,1).
app_href(67,1116,1445,1).
app_href(67,1116,1311,1).
app_href(67,487,1064,1).
app_href(67,487,145,1).
app_href(67,171,944,1).
app_href(67,171,673,1).
app_href(67,172,1076,1).
app_href(67,172,1053,1).
app_href(67,173,1336,1).
app_href(67,175,146,1).
app_href(67,175,2,1).
app_href(67,177,1403,1).
app_href(67,177,1341,1).
app_href(67,177,970,1).
app_href(67,178,829,1).
app_href(67,1117,850,1).
app_href(67,1117,825,1).
app_href(67,1117,1310,1).
app_href(67,1117,1046,1).
app_href(67,1117,309,1).
app_href(67,1117,550,1).
app_href(67,1117,672,1).
app_href(67,1117,1346,1).
app_href(67,1117,728,1).
app_href(67,1117,642,1).
app_href(67,1117,442,1).
app_href(67,1117,736,2).
app_href(67,1117,970,1).
app_href(67,1117,643,1).
app_href(67,1117,1241,1).
app_href(67,1117,403,1).
app_href(67,1117,1202,1).
app_href(67,1117,41,1).
app_href(67,1117,9,1).
app_href(67,1117,616,1).
app_href(67,1117,1041,1).
app_href(67,181,354,1).
app_href(67,182,1218,1).
app_href(67,182,420,1).
app_href(67,183,444,1).
app_href(67,184,317,1).
app_href(67,184,677,1).
app_href(67,185,1442,1).
app_href(67,185,736,1).
app_href(67,185,140,1).
app_href(67,185,233,1).
app_href(67,186,728,1).
app_href(67,186,454,1).
app_href(67,62,1138,1).
app_href(62,167,423,1).
app_href(62,167,2,1).
app_href(62,152,1370,3).
app_href(62,152,992,1).
app_href(62,152,1133,1).
app_href(62,1040,84,1).
app_href(62,1041,559,1).
app_href(62,1041,1319,1).
app_href(62,1041,1248,1).
app_href(62,1041,519,1).
app_href(62,1041,955,1).
app_href(62,1041,64,1).
app_href(62,1041,505,1).
app_href(62,1041,90,1).
app_href(62,1041,26,1).
app_href(62,1041,985,1).
app_href(62,1042,85,1).
app_href(62,319,1138,1).
app_href(62,319,959,1).
app_href(62,319,197,1).
app_href(62,319,939,1).
app_href(62,183,444,1).
app_href(62,187,444,1).
app_href(62,187,728,1).
app_href(62,177,1403,1).
app_href(62,177,1341,1).
app_href(62,177,970,1).
app_href(62,1043,242,1).
app_href(62,1043,71,1).
app_href(62,1044,67,1).
app_href(62,161,755,1).
app_href(62,173,1336,1).
app_href(62,178,829,1).
app_href(62,1045,1030,1).
app_href(62,1045,1323,1).
app_href(62,1046,49,1).
app_href(62,1047,45,1).
app_href(62,1048,50,1).
app_href(62,45,444,1).
app_href(62,45,736,1).
app_href(62,45,1433,1).
app_href(62,45,639,1).
app_href(62,1049,72,1).
app_href(62,150,830,1).
app_href(62,150,1256,1).
app_href(62,150,2,1).
app_href(62,1050,133,1).
app_href(62,1050,242,1).
app_href(62,1050,884,1).
app_href(62,1050,455,1).
app_href(62,1050,736,2).
app_href(62,1050,1470,1).
app_href(62,1050,706,1).
app_href(62,1050,1386,1).
app_href(62,1050,581,1).
app_href(62,1050,652,1).
app_href(62,1050,1255,1).
app_href(62,1050,643,1).
app_href(62,1050,998,1).
app_href(62,1051,47,1).
app_href(62,171,944,1).
app_href(62,171,673,1).
app_href(62,154,2,1).
app_href(62,1052,86,1).
app_href(62,1052,462,1).
app_href(62,1052,1180,1).
app_href(62,1052,1474,1).
app_href(62,1052,1456,1).
app_href(62,1052,912,1).
app_href(62,1052,1460,1).
app_href(62,162,1364,1).
app_href(62,162,2,1).
app_href(62,172,1076,1).
app_href(62,172,1053,1).
app_href(62,170,1192,1).
app_href(62,170,132,1).
app_href(62,1053,77,1).
app_href(62,166,736,1).
app_href(62,166,231,1).
app_href(62,148,1486,1).
app_href(62,148,1014,1).
app_href(62,184,317,1).
app_href(62,184,677,1).
app_href(62,1054,667,1).
app_href(62,1054,748,1).
app_href(62,1054,164,1).
app_href(62,1054,1446,1).
app_href(62,1054,244,1).
app_href(62,190,316,1).
app_href(62,190,736,1).
app_href(62,190,1083,1).
app_href(62,147,736,1).
app_href(62,147,941,1).
app_href(62,147,2,1).
app_href(62,147,1360,1).
app_href(62,1055,54,1).
app_href(62,1055,242,1).
app_href(62,1056,48,1).
app_href(62,163,489,1).
app_href(62,427,558,1).
app_href(62,427,611,1).
app_href(62,427,140,1).
app_href(62,427,633,1).
app_href(62,427,361,1).
app_href(62,427,564,1).
app_href(62,427,1111,1).
app_href(62,427,1033,2).
app_href(62,1057,309,1).
app_href(62,1057,643,1).
app_href(62,1057,1470,1).
app_href(62,1057,182,1).
app_href(62,175,146,1).
app_href(62,175,2,1).
app_href(62,1058,973,1).
app_href(62,1058,581,1).
app_href(62,1058,1248,1).
app_href(62,1058,736,1).
app_href(62,1058,983,1).
app_href(62,1058,884,1).
app_href(62,1058,1460,1).
app_href(62,1058,329,1).
app_href(62,185,1442,1).
app_href(62,185,736,1).
app_href(62,185,140,1).
app_href(62,185,233,1).
app_href(62,73,867,1).
app_href(62,1059,69,1).
app_href(62,493,667,1).
app_href(62,493,41,1).
app_href(62,493,747,1).
app_href(62,493,9,2).
app_href(62,493,941,1).
app_href(62,493,1446,1).
app_href(62,1060,1317,1).
app_href(62,1060,429,1).
app_href(62,1060,701,1).
app_href(62,1060,736,1).
app_href(62,1060,145,1).
app_href(62,1060,140,1).
app_href(62,1060,922,1).
app_href(62,165,867,1).
app_href(62,165,1336,1).
app_href(62,165,728,2).
app_href(62,1061,79,1).
app_href(62,75,1300,1).
app_href(62,75,571,1).
app_href(62,75,1081,1).
app_href(62,75,736,1).
app_href(62,75,573,1).
app_href(62,75,140,1).
app_href(62,75,975,1).
app_href(62,75,1310,1).
app_href(62,75,935,1).
app_href(62,75,564,1).
app_href(62,75,807,1).
app_href(62,1062,649,1).
app_href(62,1062,865,1).
app_href(62,1062,1168,1).
app_href(62,1062,736,2).
app_href(62,1062,140,1).
app_href(62,1062,564,1).
app_href(62,1062,1430,1).
app_href(62,1062,832,1).
app_href(62,1063,83,1).
app_href(62,1064,600,1).
app_href(62,1064,599,1).
app_href(62,1064,151,2).
app_href(62,1064,736,2).
app_href(62,164,454,1).
app_href(62,1065,242,1).
app_href(62,1065,70,1).
app_href(62,1066,1148,1).
app_href(62,1066,324,1).
app_href(62,169,1041,1).
app_href(62,169,571,1).
app_href(62,169,935,1).
app_href(62,169,941,1).
app_href(62,169,985,1).
app_href(62,169,649,1).
app_href(62,169,133,1).
app_href(62,169,939,2).
app_href(62,169,145,1).
app_href(62,169,1337,2).
app_href(62,169,1207,1).
app_href(62,169,1248,2).
app_href(62,169,975,1).
app_href(62,169,1319,1).
app_href(62,169,897,1).
app_href(62,169,1323,1).
app_href(62,169,912,1).
app_href(62,169,367,1).
app_href(62,169,1481,1).
app_href(62,169,519,1).
app_href(62,169,1310,1).
app_href(62,169,836,1).
app_href(62,169,146,1).
app_href(62,169,242,1).
app_href(62,169,90,1).
app_href(62,169,429,1).
app_href(62,169,846,1).
app_href(62,169,26,1).
app_href(62,169,1410,1).
app_href(62,169,1364,1).
app_href(62,169,573,1).
app_href(62,169,1456,2).
app_href(62,169,1430,1).
app_href(62,169,865,1).
app_href(62,169,1255,1).
app_href(62,169,1168,1).
app_href(62,169,1300,1).
app_href(62,169,973,3).
app_href(62,169,36,1).
app_href(62,169,1047,1).
app_href(62,169,732,1).
app_href(62,169,1474,1).
app_href(62,169,462,1).
app_href(62,169,151,2).
app_href(62,169,311,1).
app_href(62,169,197,1).
app_href(62,169,1180,1).
app_href(62,169,1460,2).
app_href(62,169,736,8).
app_href(62,169,564,1).
app_href(62,169,884,1).
app_href(62,169,559,1).
app_href(62,169,955,1).
app_href(62,169,140,4).
app_href(62,169,505,1).
app_href(62,169,922,1).
app_href(62,169,1355,1).
app_href(62,169,950,1).
app_href(62,169,1198,1).
app_href(62,169,701,1).
app_href(62,169,959,4).
app_href(62,169,983,1).
app_href(62,169,581,1).
app_href(62,169,419,1).
app_href(62,169,1009,2).
app_href(66,1067,1375,2).
app_href(66,1067,1201,1).
app_href(66,1067,193,1).
app_href(66,1067,394,1).
app_href(66,1067,601,1).
app_href(66,35,365,1).
app_href(66,35,6,1).
app_href(66,470,1374,1).
app_href(66,470,392,1).
app_href(66,470,1323,1).
app_href(66,87,444,1).
app_href(66,87,358,1).
app_href(66,740,736,1).
app_href(66,740,437,1).
app_href(66,740,197,1).
app_href(66,1068,1061,1).
app_href(66,1069,1030,1).
app_href(66,15,899,1).
app_href(66,15,1310,1).
app_href(66,15,668,1).
app_href(66,434,1346,1).
app_href(66,434,1014,1).
app_href(66,434,634,1).
app_href(66,434,772,1).
app_href(66,1070,463,1).
app_href(66,1070,1264,1).
app_href(66,1070,457,1).
app_href(66,1070,140,1).
app_href(68,87,444,1).
app_href(68,87,358,1).
app_href(68,862,1011,1).
app_href(68,862,736,1).
app_href(68,862,880,1).
app_href(68,1071,736,1).
app_href(68,1071,880,1).
app_href(68,1071,1360,1).
app_href(68,1071,503,1).
app_href(68,35,365,1).
app_href(68,35,6,1).
app_href(68,15,899,1).
app_href(68,15,1310,1).
app_href(68,15,668,1).
app_href(68,1072,736,1).
app_href(68,1072,880,1).
app_href(68,1072,1360,1).
app_href(68,1072,503,1).
app_href(69,1073,972,1).
app_href(69,1073,1210,1).
app_href(69,1073,1058,1).
app_href(69,1073,998,1).
app_href(69,1074,655,1).
app_href(69,1074,905,1).
app_href(69,1074,1050,1).
app_href(69,1074,1005,1).
app_href(69,1074,1162,1).
app_href(69,1074,1214,1).
app_href(69,1074,1384,1).
app_href(69,1074,1460,1).
app_href(69,406,1248,1).
app_href(69,406,1163,1).
app_href(69,406,736,1).
app_href(69,406,320,1).
app_href(69,406,1346,1).
app_href(69,406,1162,1).
app_href(69,406,2,1).
app_href(69,414,1456,1).
app_href(69,34,1337,1).
app_href(69,34,736,1).
app_href(69,34,1162,1).
app_href(69,863,763,1).
app_href(69,863,367,1).
app_href(69,863,1162,1).
app_href(69,1075,1056,1).
app_href(69,1075,700,1).
app_href(69,1075,1408,1).
app_href(69,1076,293,2).
app_href(56,185,140,1).
app_href(56,185,233,1).
app_href(56,184,317,1).
app_href(56,184,677,1).
app_href(56,52,1385,1).
app_href(56,52,2,1).
app_href(56,517,1318,1).
app_href(56,517,2,1).
app_href(56,518,725,1).
app_href(56,518,2,1).
app_href(56,190,316,1).
app_href(56,190,736,1).
app_href(56,190,1083,1).
app_href(70,406,1163,1).
app_href(70,406,736,1).
app_href(70,406,320,1).
app_href(70,406,1346,1).
app_href(70,406,1298,1).
app_href(70,406,1162,1).
app_href(70,406,2,1).
app_href(70,34,444,1).
app_href(70,34,1162,1).
app_href(70,35,365,1).
app_href(70,35,6,1).
app_href(70,15,899,1).
app_href(70,15,1310,1).
app_href(70,15,668,1).
app_href(70,1077,740,1).
app_href(70,1077,504,1).
app_href(70,1004,655,1).
app_href(70,1078,323,1).
app_href(70,865,972,1).
app_href(70,865,998,1).
app_href(70,1147,649,1).
app_href(70,1147,1277,1).
app_href(70,1147,736,1).
app_href(70,1147,602,1).
app_href(70,1147,1384,1).
app_href(70,1147,1460,1).
app_href(70,1079,616,1).
app_href(70,1079,655,1).
app_href(70,1079,1092,1).
app_href(70,1079,1005,1).
app_href(70,1079,1162,1).
app_href(70,412,649,1).
app_href(70,412,1358,1).
app_href(70,412,736,1).
app_href(70,412,320,1).
app_href(70,477,540,1).
app_href(70,477,140,1).
app_href(70,477,667,1).
app_href(70,477,41,1).
app_href(70,477,277,1).
app_href(70,477,37,1).
app_href(70,477,728,1).
app_href(70,477,9,2).
app_href(70,477,367,1).
app_href(70,477,219,1).
app_href(70,78,649,2).
app_href(70,78,444,1).
app_href(70,78,736,1).
app_href(70,1142,993,1).
app_href(70,1083,655,1).
app_href(70,1083,1277,1).
app_href(70,1083,602,1).
app_href(70,1083,1005,1).
app_href(70,1083,1162,1).
app_href(70,1083,1384,1).
app_href(70,1083,1460,1).
app_href(70,1005,736,1).
app_href(70,1005,1162,1).
app_href(70,1006,972,1).
app_href(70,1006,649,1).
app_href(70,1006,876,1).
app_href(70,1006,736,1).
app_href(70,1006,1264,1).
app_href(70,1006,810,1).
app_href(70,1006,1360,1).
app_href(70,69,655,1).
app_href(70,69,444,1).
app_href(70,69,1005,1).
app_href(70,69,1162,1).
app_href(70,1007,972,1).
app_href(70,1007,649,1).
app_href(70,1007,876,1).
app_href(70,1007,736,1).
app_href(70,1007,1264,1).
app_href(70,1007,1360,1).
app_href(70,1007,810,1).
app_href(70,1008,367,1).
app_href(70,1008,160,1).
app_href(70,1008,239,1).
app_href(70,1008,967,1).
app_href(70,1008,314,1).
app_href(70,1008,1147,1).
app_href(70,1009,649,1).
app_href(70,1009,239,1).
app_href(70,1009,905,2).
app_href(70,1009,967,1).
app_href(70,1009,993,1).
app_href(70,1009,1147,1).
app_href(70,1009,1093,2).
app_href(70,1009,1162,1).
app_href(67,62,959,1).
app_href(67,187,444,1).
app_href(67,187,728,1).
app_href(67,190,316,1).
app_href(67,190,736,1).
app_href(67,190,1083,1).
app_href(71,87,444,1).
app_href(71,87,358,1).
app_href(71,35,365,1).
app_href(71,1118,1123,1).
app_href(71,1118,794,1).
app_href(71,15,899,1).
app_href(71,15,1310,1).
app_href(71,15,668,1).
app_href(73,1119,676,1).
app_href(73,35,365,1).
app_href(73,1120,1215,1).
app_href(73,1121,517,1).
app_href(73,1121,950,1).
app_href(73,1122,395,1).
app_href(73,15,899,1).
app_href(73,15,1310,1).
app_href(73,15,668,1).
app_href(73,87,444,1).
app_href(73,87,358,1).
app_href(73,1123,549,1).
app_href(73,1124,1294,1).
app_href(73,1125,1011,1).
app_href(73,1125,736,1).
app_href(73,1125,718,1).
app_href(73,480,667,1).
app_href(73,480,41,1).
app_href(73,480,9,1).
app_href(73,480,779,1).
app_href(73,480,503,1).
app_href(73,388,1319,1).
app_href(73,388,492,1).
app_href(73,871,736,1).
app_href(73,871,1267,1).
app_href(73,871,283,1).
app_href(73,871,950,1).
app_href(73,1126,1115,1).
app_href(73,1127,1041,1).
app_href(73,1127,736,1).
app_href(73,1127,146,1).
app_href(73,1127,1346,1).
app_href(73,1127,941,1).
app_href(73,1127,564,1).
app_href(73,1127,790,1).
app_href(73,1128,736,1).
app_href(73,1128,282,1).
app_href(73,1128,809,1).
app_href(73,1129,548,1).
app_href(73,1130,1244,1).
app_href(73,1131,1243,1).
app_href(73,622,1351,1).
app_href(73,1132,1004,1).
app_href(73,1133,128,1).
app_href(73,1133,444,1).
app_href(73,1133,941,1).
app_href(73,887,278,1).
app_href(73,887,736,1).
app_href(73,887,1267,1).
app_href(73,887,950,1).
app_href(73,852,278,1).
app_href(73,852,736,1).
app_href(73,852,950,1).
app_href(73,469,1409,1).
app_href(73,469,736,1).
app_href(73,469,1130,1).
app_href(73,1134,1293,1).
app_href(73,753,736,1).
app_href(73,753,950,1).
app_href(73,1135,687,1).
app_href(73,732,736,1).
app_href(73,732,283,1).
app_href(73,732,950,1).
app_href(78,1062,649,1).
app_href(78,1062,736,1).
app_href(78,1062,1086,1).
app_href(78,1062,832,1).
app_href(78,1136,852,1).
app_href(78,1136,1210,1).
app_href(78,1136,1408,1).
app_href(78,69,894,1).
app_href(78,69,1141,1).
app_href(78,69,736,1).
app_href(78,69,708,1).
app_href(78,69,1005,1).
app_href(78,69,705,1).
app_href(78,69,1168,1).
app_href(78,69,715,1).
app_href(78,69,1101,1).
app_href(78,1137,516,1).
app_href(78,1137,969,1).
app_href(78,1137,258,1).
app_href(78,730,972,1).
app_href(78,730,370,1).
app_href(78,730,998,1).
app_href(78,1138,701,1).
app_href(78,1138,1239,1).
app_href(78,1138,467,1).
app_href(78,1139,516,1).
app_href(78,1139,627,1).
app_href(78,1139,1464,1).
app_href(78,1140,124,1).
app_href(78,1140,1210,1).
app_href(78,1140,1408,1).
app_href(78,1141,1217,3).
app_href(78,1141,516,1).
app_href(78,1141,707,3).
app_href(78,1141,688,1).
app_href(78,1141,1346,1).
app_href(78,1141,922,3).
app_href(78,1141,1408,1).
app_href(78,1141,2,1).
app_href(78,1078,971,1).
app_href(78,1078,1265,1).
app_href(78,1078,323,1).
app_href(78,1078,564,1).
app_href(78,1078,667,1).
app_href(78,1078,41,1).
app_href(78,1078,728,1).
app_href(78,1078,1318,1).
app_href(78,1078,9,2).
app_href(78,1078,367,1).
app_href(78,1142,972,1).
app_href(78,1142,367,1).
app_href(78,1142,993,1).
app_href(78,1142,998,1).
app_href(78,1143,1036,1).
app_href(78,1143,1377,1).
app_href(78,1143,140,1).
app_href(78,1143,41,1).
app_href(78,1143,1093,2).
app_href(78,1143,728,1).
app_href(78,1143,1213,1).
app_href(78,1143,9,1).
app_href(78,1143,1460,1).
app_href(78,1143,367,1).
app_href(78,1144,649,1).
app_href(78,1144,367,1).
app_href(78,1144,736,1).
app_href(78,1144,140,1).
app_href(78,1144,1408,1).
app_href(78,860,972,1).
app_href(78,860,1057,1).
app_href(78,860,998,1).
app_href(78,1076,293,1).
app_href(78,1076,367,1).
app_href(78,1076,140,1).
app_href(78,1076,325,1).
app_href(78,1076,740,1).
app_href(78,1076,504,1).
app_href(78,15,899,1).
app_href(78,15,1310,1).
app_href(78,15,668,1).
app_href(65,739,763,1).
app_href(65,739,658,1).
app_href(65,739,512,1).
app_href(65,857,763,1).
app_href(65,857,658,1).
app_href(65,857,1267,1).
app_href(65,875,1048,1).
app_href(65,875,763,1).
app_href(65,875,658,1).
app_href(65,875,783,1).
app_href(65,87,444,1).
app_href(65,87,358,1).
app_href(65,430,736,1).
app_href(65,430,1346,1).
app_href(65,430,1014,1).
app_href(65,430,634,1).
app_href(65,430,36,1).
app_href(65,848,763,1).
app_href(65,848,658,1).
app_href(65,848,1291,1).
app_href(65,15,899,1).
app_href(65,15,1310,1).
app_href(65,15,668,1).
app_href(65,434,1346,1).
app_href(65,434,1014,1).
app_href(65,434,634,1).
app_href(65,434,772,1).
app_href(65,766,763,1).
app_href(65,766,658,1).
app_href(65,766,892,1).
app_href(65,1161,1105,1).
app_href(72,973,1105,1).
app_href(72,1068,1105,1).
app_href(72,1162,1105,1).
app_href(72,1163,543,1).
app_href(72,302,1030,1).
app_href(72,1159,1030,1).
app_href(72,1161,1105,1).
app_href(72,1164,1030,1).
app_href(72,87,444,1).
app_href(72,87,1189,1).
app_href(72,1165,1030,1).
app_href(72,1069,1030,1).
app_href(72,1019,1030,1).
app_href(72,97,1030,1).
app_href(72,985,1030,1).
app_href(72,1166,1116,1).
app_href(72,1166,452,1).
app_href(72,1166,122,1).
app_href(72,1166,1014,1).
app_href(72,1167,1030,1).
app_href(72,987,1030,1).
app_href(72,991,1030,1).
app_href(72,1168,1030,1).
app_href(72,1169,1105,1).
app_href(72,993,1105,1).
app_href(72,1170,1030,1).
app_href(72,692,1030,1).
app_href(72,700,1030,1).
app_href(72,98,1105,1).
app_href(72,997,1105,1).
app_href(72,1171,1105,1).
app_href(72,695,1105,1).
app_href(74,506,242,1).
app_href(74,506,393,1).
app_href(74,506,2,1).
app_href(74,148,1486,1).
app_href(74,148,1014,1).
app_href(74,149,830,2).
app_href(74,149,536,1).
app_href(74,149,140,1).
app_href(74,149,823,1).
app_href(74,1172,152,1).
app_href(74,1172,1000,1).
app_href(74,1172,140,1).
app_href(74,1172,406,1).
app_href(74,1172,156,1).
app_href(74,15,899,1).
app_href(74,15,1310,1).
app_href(74,15,668,1).
app_href(74,152,1370,3).
app_href(74,152,992,1).
app_href(74,152,1133,1).
app_href(74,161,755,1).
app_href(74,129,1473,1).
app_href(74,129,2,1).
app_href(74,507,118,1).
app_href(74,507,477,1).
app_href(74,507,2,1).
app_href(74,163,489,1).
app_href(74,508,726,1).
app_href(74,508,2,1).
app_href(74,164,454,1).
app_href(74,1173,452,1).
app_href(74,1173,309,1).
app_href(74,1173,643,1).
app_href(74,1173,1310,1).
app_href(74,1173,736,1).
app_href(74,165,728,1).
app_href(74,165,2,1).
app_href(74,166,736,1).
app_href(74,166,231,1).
app_href(74,168,1069,1).
app_href(74,509,1443,1).
app_href(74,509,1218,1).
app_href(74,509,2,1).
app_href(74,170,1192,1).
app_href(74,170,132,1).
app_href(74,171,944,1).
app_href(74,171,673,1).
app_href(74,1174,732,1).
app_href(74,172,1076,1).
app_href(74,172,1053,1).
app_href(74,213,152,1).
app_href(74,213,736,1).
app_href(74,213,1000,1).
app_href(74,173,1336,1).
app_href(74,511,1183,1).
app_href(74,511,296,1).
app_href(74,511,2,1).
app_href(74,175,146,1).
app_href(74,175,2,1).
app_href(74,177,1403,1).
app_href(74,177,1341,1).
app_href(74,177,970,1).
app_href(74,178,829,1).
app_href(74,1175,1310,1).
app_href(74,1175,1150,1).
app_href(74,1175,140,1).
app_href(74,1175,1402,1).
app_href(74,1175,857,1).
app_href(74,1175,1369,1).
app_href(74,513,1459,1).
app_href(74,513,2,1).
app_href(74,181,354,1).
app_href(74,182,1200,1).
app_href(74,182,1218,1).
app_href(74,182,420,1).
app_href(74,182,134,1).
app_href(74,182,2,1).
app_href(74,514,575,1).
app_href(74,514,2,1).
app_href(74,183,444,1).
app_href(74,183,1200,1).
app_href(74,183,134,1).
app_href(74,183,2,1).
app_href(74,515,999,1).
app_href(74,515,2,1).
app_href(74,185,1442,1).
app_href(74,185,736,1).
app_href(74,185,140,1).
app_href(74,185,233,1).
app_href(74,184,317,1).
app_href(74,184,677,1).
app_href(74,52,1385,1).
app_href(74,52,2,1).
app_href(74,1176,1443,1).
app_href(55,165,728,1).
app_href(55,165,2,1).
app_href(55,166,736,1).
app_href(55,166,231,1).
app_href(55,168,1069,1).
app_href(55,509,1443,1).
app_href(55,509,1218,1).
app_href(55,509,2,1).
app_href(55,922,1316,1).
app_href(55,922,1041,1).
app_href(55,922,736,2).
app_href(55,922,1194,1).
app_href(55,922,288,1).
app_href(55,922,1434,1).
app_href(55,922,1274,1).
app_href(55,922,826,1).
app_href(55,922,564,1).
app_href(55,922,480,2).
app_href(55,922,1346,2).
app_href(55,922,1412,1).
app_href(55,922,26,2).
app_href(55,922,9,1).
app_href(55,170,1192,1).
app_href(55,170,132,1).
app_href(55,171,944,1).
app_href(55,171,673,1).
app_href(55,172,1076,1).
app_href(55,172,1053,1).
app_href(55,173,1336,1).
app_href(55,905,826,1).
app_href(55,905,1194,1).
app_href(55,905,26,1).
app_href(55,905,736,1).
app_href(55,905,1346,1).
app_href(55,905,1436,1).
app_href(55,905,480,1).
app_href(55,905,288,1).
app_href(55,511,1183,1).
app_href(55,511,296,1).
app_href(55,511,2,1).
app_href(55,175,146,1).
app_href(55,175,2,1).
app_href(55,177,1403,1).
app_href(55,177,1341,1).
app_href(55,177,970,1).
app_href(55,178,829,1).
app_href(55,513,1459,1).
app_href(55,513,2,1).
app_href(55,181,354,1).
app_href(55,182,1200,1).
app_href(55,182,1218,1).
app_href(55,182,420,1).
app_href(55,182,134,1).
app_href(55,182,2,1).
app_href(55,514,575,1).
app_href(55,514,2,1).
app_href(55,183,444,1).
app_href(55,183,1200,1).
app_href(55,183,134,1).
app_href(55,183,2,1).
app_href(55,515,999,1).
app_href(55,515,2,1).
app_href(55,184,317,1).
app_href(55,184,677,1).
app_href(55,185,1442,1).
app_href(55,185,736,1).
app_href(55,185,140,1).
app_href(55,185,233,1).
app_href(55,52,1385,1).
app_href(55,52,2,1).
app_href(55,517,1318,1).
app_href(55,517,2,1).
app_href(55,518,725,1).
app_href(55,518,2,1).
app_href(55,190,316,1).
app_href(55,190,736,1).
app_href(55,190,1083,1).
app_href(75,165,867,1).
app_href(75,165,1336,1).
app_href(75,165,728,2).
app_href(75,185,1442,1).
app_href(75,185,736,1).
app_href(75,185,140,1).
app_href(75,185,233,1).
app_href(75,168,1069,1).
app_href(75,1117,616,1).
app_href(75,1117,736,1).
app_href(75,1117,309,1).
app_href(75,1117,642,1).
app_href(75,1117,1046,1).
app_href(75,1117,1310,1).
app_href(75,1117,672,1).
app_href(75,1117,442,1).
app_href(75,1117,41,1).
app_href(75,1117,728,1).
app_href(75,1117,9,1).
app_href(75,1117,643,1).
app_href(75,187,444,1).
app_href(75,187,728,1).
app_href(75,182,1218,1).
app_href(75,182,420,1).
app_href(75,163,489,1).
app_href(75,162,1364,1).
app_href(75,162,2,1).
app_href(75,164,454,1).
app_href(75,184,317,1).
app_href(75,184,677,1).
app_href(75,170,1192,1).
app_href(75,170,132,1).
app_href(75,15,899,1).
app_href(75,15,1310,1).
app_href(75,15,668,1).
app_href(75,152,1370,3).
app_href(75,152,992,1).
app_href(75,152,1133,1).
app_href(75,62,695,1).
app_href(75,62,959,1).
app_href(75,62,836,1).
app_href(75,62,690,1).
app_href(75,183,444,1).
app_href(75,166,736,1).
app_href(75,166,231,1).
app_href(75,150,830,1).
app_href(75,150,1256,1).
app_href(75,150,2,1).
app_href(75,149,830,1).
app_href(75,478,1293,1).
app_href(75,478,736,1).
app_href(75,478,598,1).
app_href(75,478,675,1).
app_href(75,478,635,1).
app_href(75,478,934,1).
app_href(75,173,1336,1).
app_href(75,148,1486,1).
app_href(75,148,1014,1).
app_href(75,190,316,1).
app_href(75,190,736,1).
app_href(75,190,1083,1).
app_href(75,169,807,1).
app_href(75,169,690,10).
app_href(75,169,1120,1).
app_href(75,169,1173,3).
app_href(75,169,1349,1).
app_href(75,169,21,1).
app_href(75,169,904,1).
app_href(75,169,1319,1).
app_href(75,169,434,1).
app_href(75,169,222,1).
app_href(75,169,1443,2).
app_href(75,169,1203,4).
app_href(75,169,223,3).
app_href(75,169,173,1).
app_href(75,169,1150,2).
app_href(75,169,1149,1).
app_href(75,169,836,1).
app_href(75,169,1081,1).
app_href(75,169,546,3).
app_href(74,1176,1237,1).
app_href(74,1176,1209,1).
app_href(74,1176,736,1).
app_href(74,1176,127,1).
app_href(74,1026,899,1).
app_href(74,1026,1310,1).
app_href(74,1026,1341,1).
app_href(74,1026,823,1).
app_href(74,1026,668,1).
app_href(74,517,1318,1).
app_href(74,517,2,1).
app_href(74,518,725,1).
app_href(74,518,2,1).
app_href(74,1177,501,1).
app_href(74,190,316,1).
app_href(74,190,736,1).
app_href(74,190,1083,1).
app_href(74,312,1041,1).
app_href(74,312,444,2).
app_href(74,312,1333,2).
app_href(74,312,26,1).
app_href(74,312,736,1).
app_href(76,1178,737,1).
app_href(76,1178,721,1).
app_href(76,1178,1446,1).
app_href(76,187,444,1).
app_href(76,35,365,1).
app_href(76,1179,140,1).
app_href(76,1179,1364,1).
app_href(76,1179,648,1).
app_href(76,1180,331,1).
app_href(76,1180,420,1).
app_href(76,1181,1372,1).
app_href(76,1181,442,1).
app_href(76,15,899,1).
app_href(76,15,1310,1).
app_href(76,15,668,1).
app_href(76,1182,1318,1).
app_href(76,1183,316,1).
app_href(76,1183,432,1).
app_href(76,1184,26,2).
app_href(76,1184,9,1).
app_href(76,1185,296,1).
app_href(76,1185,1256,1).
app_href(76,1186,9,2).
app_href(76,173,146,1).
app_href(76,173,1088,1).
app_href(76,1187,1139,1).
app_href(76,1187,161,1).
app_href(76,1188,605,1).
app_href(76,169,41,1).
app_href(76,169,538,1).
app_href(76,169,9,2).
app_href(76,181,1015,2).
app_href(76,181,629,2).
app_href(76,171,944,1).
app_href(76,171,673,1).
app_href(76,1189,41,1).
app_href(76,1189,9,1).
app_href(76,1189,8,1).
app_href(76,1190,263,1).
app_href(76,1190,1299,1).
app_href(76,167,266,1).
app_href(76,167,1418,1).
app_href(76,167,423,1).
app_href(76,1191,37,1).
app_href(76,1191,9,2).
app_href(76,1192,128,1).
app_href(76,1192,1336,1).
app_href(76,1193,1041,1).
app_href(76,1193,1358,1).
app_href(76,1193,1274,1).
app_href(76,1193,26,1).
app_href(76,1193,736,1).
app_href(76,1193,480,1).
app_href(76,1194,1357,1).
app_href(76,1194,517,1).
app_href(76,1194,1192,1).
app_href(76,1194,140,1).
app_href(76,177,1403,1).
app_href(76,177,1341,1).
app_href(76,177,970,1).
app_href(76,172,1077,1).
app_href(76,147,439,1).
app_href(76,147,597,1).
app_href(76,147,1278,1).
app_href(76,147,641,1).
app_href(76,1195,536,1).
app_href(77,903,517,1).
app_href(77,903,26,1).
app_href(77,903,9,1).
app_href(77,903,8,1).
app_href(77,903,769,1).
app_href(77,904,480,1).
app_href(77,904,555,1).
app_href(77,1196,1386,1).
app_href(77,1196,1117,1).
app_href(77,1197,232,1).
app_href(77,49,517,1).
app_href(77,49,9,2).
app_href(77,49,769,1).
app_href(77,1198,253,1).
app_href(77,1198,17,1).
app_href(77,1198,1479,1).
app_href(77,1198,744,1).
app_href(77,1023,1194,1).
app_href(77,1023,9,1).
app_href(77,1023,1436,1).
app_href(77,1023,480,1).
app_href(80,147,736,1).
app_href(80,147,941,1).
app_href(80,147,2,1).
app_href(80,147,1360,1).
app_href(80,148,1486,1).
app_href(80,148,1014,1).
app_href(80,149,830,1).
app_href(80,150,830,1).
app_href(80,150,1256,1).
app_href(80,150,2,1).
app_href(80,1199,176,1).
app_href(80,1199,1429,1).
app_href(80,1199,171,1).
app_href(80,1199,309,1).
app_href(80,1199,643,1).
app_href(80,1199,691,1).
app_href(80,1199,797,1).
app_href(80,1199,1318,1).
app_href(80,15,899,1).
app_href(80,15,1310,1).
app_href(80,15,668,1).
app_href(80,154,2,1).
app_href(80,1200,398,1).
app_href(80,1200,1086,1).
app_href(80,152,1370,3).
app_href(80,152,992,1).
app_href(80,152,1133,1).
app_href(80,161,755,1).
app_href(80,162,1364,1).
app_href(80,162,2,1).
app_href(80,562,1144,1).
app_href(80,562,1080,1).
app_href(80,562,139,1).
app_href(80,562,690,1).
app_href(80,1201,309,1).
app_href(80,1201,175,1).
app_href(80,1201,172,1).
app_href(80,1201,797,1).
app_href(80,1201,93,1).
app_href(80,1201,897,1).
app_href(80,1201,26,1).
app_href(80,1201,1318,2).
app_href(80,1201,643,1).
app_href(80,1201,1109,1).
app_href(80,163,489,1).
app_href(80,164,454,1).
app_href(80,165,867,1).
app_href(80,165,1336,1).
app_href(69,1076,367,2).
app_href(69,1076,140,2).
app_href(69,1076,325,2).
app_href(69,1076,740,2).
app_href(69,1076,504,2).
app_href(69,35,365,1).
app_href(69,398,875,1).
app_href(69,398,1014,1).
app_href(69,398,870,1).
app_href(69,398,994,1).
app_href(69,144,1014,1).
app_href(69,144,870,1).
app_href(69,15,899,1).
app_href(69,15,1310,1).
app_href(69,15,668,1).
app_href(69,1077,1358,1).
app_href(69,1077,740,1).
app_href(69,1077,504,1).
app_href(69,1078,323,1).
app_href(69,87,444,1).
app_href(69,87,358,1).
app_href(69,412,1014,1).
app_href(69,412,870,1).
app_href(69,1079,616,1).
app_href(69,1079,1092,1).
app_href(69,1079,1168,1).
app_href(69,1079,736,1).
app_href(69,477,540,2).
app_href(69,477,736,1).
app_href(69,477,649,1).
app_href(69,477,140,2).
app_href(69,477,667,2).
app_href(69,477,41,2).
app_href(69,477,37,2).
app_href(69,477,277,2).
app_href(69,477,1168,1).
app_href(69,477,728,2).
app_href(69,477,9,4).
app_href(69,477,1358,1).
app_href(69,477,367,2).
app_href(69,477,219,2).
app_href(69,730,972,1).
app_href(69,730,370,1).
app_href(69,730,998,1).
app_href(69,1080,655,1).
app_href(69,1080,736,1).
app_href(69,1080,1014,1).
app_href(69,1080,1005,1).
app_href(69,1080,1162,1).
app_href(69,1080,634,1).
app_href(69,78,1015,1).
app_href(69,78,649,2).
app_href(69,78,1337,2).
app_href(69,78,736,2).
app_href(69,844,763,1).
app_href(69,844,1162,1).
app_href(69,1081,368,1).
app_href(69,1081,972,1).
app_href(69,1081,968,1).
app_href(69,1081,1056,1).
app_href(69,1081,998,1).
app_href(69,878,279,1).
app_href(69,878,567,2).
app_href(69,878,99,2).
app_href(69,878,1264,1).
app_href(69,878,37,1).
app_href(69,878,26,1).
app_href(69,878,13,1).
app_href(69,878,9,1).
app_href(69,878,367,1).
app_href(69,878,655,1).
app_href(69,1082,1030,1).
app_href(69,1083,655,1).
app_href(69,1083,1277,1).
app_href(69,1083,736,1).
app_href(69,1083,602,1).
app_href(69,1083,1005,1).
app_href(69,1083,1162,1).
app_href(69,1083,1408,1).
app_href(69,514,309,1).
app_href(69,514,574,1).
app_href(69,514,643,1).
app_href(69,514,1310,1).
app_href(69,514,736,1).
app_href(69,1084,972,1).
app_href(69,850,972,1).
app_href(69,850,998,1).
app_href(69,850,1058,2).
app_href(69,763,972,1).
app_href(69,763,367,1).
app_href(69,763,998,1).
app_href(69,421,1217,1).
app_href(69,421,922,1).
app_href(79,506,1443,1).
app_href(79,506,242,1).
app_href(79,506,1018,1).
app_href(79,506,1014,1).
app_href(79,147,736,1).
app_href(79,147,941,1).
app_href(79,147,2,1).
app_href(79,147,1360,1).
app_href(79,148,1486,1).
app_href(79,148,1014,1).
app_href(79,149,830,1).
app_href(79,150,830,1).
app_href(79,150,1256,1).
app_href(79,150,2,1).
app_href(79,15,899,1).
app_href(79,15,1310,1).
app_href(79,15,668,1).
app_href(79,71,1430,1).
app_href(79,71,736,2).
app_href(79,71,864,1).
app_href(79,71,505,1).
app_href(79,71,1178,1).
app_href(79,71,701,1).
app_href(79,71,1346,1).
app_href(79,71,26,1).
app_href(79,71,1090,1).
app_href(79,152,1370,3).
app_href(79,152,992,1).
app_href(79,152,1133,1).
app_href(79,154,2,1).
app_href(79,161,755,1).
app_href(79,24,309,1).
app_href(79,24,643,1).
app_href(79,24,1310,1).
app_href(79,24,736,1).
app_href(79,16,701,1).
app_href(79,16,736,1).
app_href(79,16,927,1).
app_href(79,162,1364,1).
app_href(79,162,2,1).
app_href(79,305,400,1).
app_href(79,305,701,1).
app_href(79,305,736,1).
app_href(79,305,864,1).
app_href(79,305,1430,1).
app_href(79,163,489,1).
app_href(79,164,454,1).
app_href(79,313,485,1).
app_href(79,313,1333,1).
app_href(79,313,498,1).
app_href(79,313,736,1).
app_href(79,313,1246,1).
app_href(79,313,130,1).
app_href(79,165,867,1).
app_href(79,165,1336,1).
app_href(79,165,728,2).
app_href(79,166,736,1).
app_href(79,166,231,1).
app_href(79,167,1418,1).
app_href(79,167,423,1).
app_href(79,167,2,1).
app_href(79,168,1069,1).
app_href(79,31,1041,1).
app_href(79,31,1264,1).
app_href(75,169,1032,1).
app_href(75,169,87,1).
app_href(75,169,26,4).
app_href(75,169,606,1).
app_href(75,169,9,1).
app_href(75,169,15,1).
app_href(75,169,1276,2).
app_href(75,169,1221,1).
app_href(75,169,330,1).
app_href(75,169,164,1).
app_href(75,169,901,1).
app_href(75,169,721,1).
app_href(75,169,307,1).
app_href(75,169,801,1).
app_href(75,169,1346,3).
app_href(75,169,667,1).
app_href(75,169,992,2).
app_href(75,169,1177,1).
app_href(75,169,973,1).
app_href(75,169,695,5).
app_href(75,169,1047,2).
app_href(75,169,813,1).
app_href(75,169,970,9).
app_href(75,169,636,1).
app_href(75,169,1460,2).
app_href(75,169,845,1).
app_href(75,169,736,6).
app_href(75,169,30,1).
app_href(75,169,140,2).
app_href(75,169,39,1).
app_href(75,169,698,1).
app_href(75,169,1424,1).
app_href(75,169,1329,2).
app_href(75,169,798,1).
app_href(75,169,887,2).
app_href(75,169,1189,2).
app_href(75,169,604,1).
app_href(75,169,1044,1).
app_href(75,169,1413,1).
app_href(75,169,1360,1).
app_href(75,169,416,1).
app_href(75,169,1114,2).
app_href(75,181,354,1).
app_href(75,171,944,1).
app_href(75,171,673,1).
app_href(75,167,1418,1).
app_href(75,167,423,1).
app_href(75,167,2,1).
app_href(75,175,146,1).
app_href(75,175,2,1).
app_href(75,161,755,1).
app_href(75,487,1064,1).
app_href(75,487,145,1).
app_href(75,177,1403,1).
app_href(75,177,1341,1).
app_href(75,177,970,1).
app_href(75,186,728,1).
app_href(75,186,454,1).
app_href(75,154,2,1).
app_href(75,172,1076,1).
app_href(75,172,1053,1).
app_href(75,147,736,1).
app_href(75,147,941,1).
app_href(75,147,2,1).
app_href(75,147,1360,1).
app_href(75,178,829,1).
app_href(81,923,991,1).
app_href(81,923,1372,1).
app_href(81,924,1218,1).
app_href(81,924,353,1).
app_href(81,924,481,1).
app_href(81,1180,331,1).
app_href(81,1180,420,1).
app_href(81,1181,1372,1).
app_href(81,1181,442,1).
app_href(81,15,899,1).
app_href(81,15,1310,1).
app_href(81,15,668,1).
app_href(81,925,372,1).
app_href(81,925,1465,1).
app_href(81,925,1310,1).
app_href(81,925,586,1).
app_href(81,1182,1318,1).
app_href(81,1183,316,1).
app_href(81,1183,432,1).
app_href(81,926,270,1).
app_href(81,173,146,1).
app_href(81,173,1088,1).
app_href(81,1188,605,1).
app_href(81,171,944,1).
app_href(81,171,673,1).
app_href(81,927,821,1).
app_href(81,927,384,1).
app_href(81,927,1337,1).
app_href(81,927,736,1).
app_href(81,181,1015,2).
app_href(81,181,629,2).
app_href(81,1190,263,1).
app_href(81,1190,1299,1).
app_href(81,928,1443,1).
app_href(81,928,1237,1).
app_href(81,928,1218,1).
app_href(81,928,1310,1).
app_href(81,928,482,1).
app_href(81,177,1403,1).
app_href(81,177,1341,1).
app_href(81,177,970,1).
app_href(81,172,1077,1).
app_href(82,929,1091,1).
app_href(82,929,899,1).
app_href(82,929,1310,1).
app_href(82,929,119,1).
app_href(82,35,365,1).
app_href(82,35,6,1).
app_href(82,930,1098,1).
app_href(82,930,119,1).
app_href(82,931,802,1).
app_href(82,931,225,1).
app_href(82,932,119,1).
app_href(82,932,736,1).
app_href(82,932,1041,1).
app_href(82,932,1346,1).
app_href(82,932,667,1).
app_href(82,932,41,1).
app_href(82,932,1412,1).
app_href(82,932,26,1).
app_href(82,932,503,1).
app_href(82,113,331,1).
app_href(82,113,1346,1).
app_href(82,113,1323,1).
app_href(82,87,26,1).
app_href(82,87,736,1).
app_href(82,87,358,1).
app_href(82,87,1189,1).
app_href(82,933,118,1).
app_href(82,933,736,1).
app_href(82,933,897,1).
app_href(82,933,1456,1).
app_href(82,933,564,1).
app_href(82,933,238,1).
app_href(82,934,1410,2).
app_href(82,934,736,2).
app_href(82,934,119,2).
app_href(82,15,899,1).
app_href(82,15,1310,1).
app_href(82,15,668,1).
app_href(82,434,1346,1).
app_href(82,434,1014,2).
app_href(82,434,634,2).
app_href(82,434,772,2).
app_href(82,754,736,1).
app_href(82,754,1264,1).
app_href(82,754,145,1).
app_href(82,754,119,1).
app_href(82,935,1337,1).
app_href(82,935,179,1).
app_href(82,935,404,1).
app_href(79,31,145,1).
app_href(79,31,764,1).
app_href(79,31,1346,1).
app_href(79,169,318,1).
app_href(79,169,1443,1).
app_href(79,169,543,1).
app_href(79,169,711,2).
app_href(79,169,1145,1).
app_href(79,169,1189,3).
app_href(79,169,1346,1).
app_href(79,169,1015,1).
app_href(79,169,959,1).
app_href(79,169,736,4).
app_href(79,169,970,1).
app_href(79,169,1178,1).
app_href(79,169,358,1).
app_href(79,169,1138,1).
app_href(79,169,242,1).
app_href(79,169,505,1).
app_href(79,169,26,4).
app_href(79,169,444,2).
app_href(79,169,140,1).
app_href(79,169,1121,1).
app_href(79,87,1041,1).
app_href(79,87,444,1).
app_href(79,87,764,1).
app_href(79,4,1139,1).
app_href(79,4,543,1).
app_href(79,4,1145,1).
app_href(79,4,242,1).
app_href(79,4,318,1).
app_href(79,4,140,1).
app_href(79,4,1070,1).
app_href(79,4,1460,1).
app_href(79,170,1192,1).
app_href(79,170,132,1).
app_href(79,171,944,1).
app_href(79,171,673,1).
app_href(79,172,1076,1).
app_href(79,172,1053,1).
app_href(79,173,1336,1).
app_href(79,173,309,1).
app_href(79,173,643,1).
app_href(79,173,1310,1).
app_href(79,173,736,1).
app_href(79,173,1088,1).
app_href(79,175,146,1).
app_href(79,175,2,1).
app_href(79,27,1275,1).
app_href(79,27,616,1).
app_href(79,27,736,1).
app_href(79,27,309,1).
app_href(79,27,1015,1).
app_href(79,27,1443,1).
app_href(79,27,1310,1).
app_href(79,27,442,1).
app_href(79,27,41,1).
app_href(79,27,728,1).
app_href(79,27,9,2).
app_href(79,27,643,1).
app_href(79,177,1403,1).
app_href(79,177,1341,1).
app_href(79,177,970,1).
app_href(79,178,829,1).
app_href(79,36,271,1).
app_href(79,36,309,1).
app_href(79,36,643,1).
app_href(79,36,1310,1).
app_href(79,36,736,1).
app_href(79,36,243,1).
app_href(79,36,272,1).
app_href(79,1085,1138,1).
app_href(79,1085,959,1).
app_href(79,1085,1209,1).
app_href(79,1085,736,1).
app_href(79,1085,1346,1).
app_href(79,181,354,1).
app_href(79,182,1218,1).
app_href(79,182,420,1).
app_href(79,26,1443,1).
app_href(79,26,452,2).
app_href(79,26,1237,1).
app_href(79,26,1014,1).
app_href(79,26,1121,1).
app_href(79,183,444,1).
app_href(79,185,1442,1).
app_href(79,185,736,1).
app_href(79,185,140,1).
app_href(79,185,233,1).
app_href(79,184,317,1).
app_href(79,184,677,1).
app_href(79,186,728,1).
app_href(79,186,454,1).
app_href(79,62,1138,1).
app_href(79,62,318,1).
app_href(79,62,959,1).
app_href(79,1180,309,1).
app_href(79,1180,643,1).
app_href(79,1180,1310,1).
app_href(79,1180,736,1).
app_href(79,1180,768,1).
app_href(79,1182,736,1).
app_href(79,1182,358,1).
app_href(79,1182,140,1).
app_href(79,1182,1189,1).
app_href(79,1182,1346,1).
app_href(79,1182,1218,1).
app_href(79,1182,26,1).
app_href(79,1182,1318,1).
app_href(79,1182,970,1).
app_href(79,187,309,1).
app_href(79,187,444,1).
app_href(79,187,643,1).
app_href(79,187,1310,1).
app_href(79,187,736,1).
app_href(79,187,728,1).
app_href(79,187,442,1).
app_href(79,55,826,1).
app_href(79,55,736,1).
app_href(79,55,1346,1).
app_href(79,55,288,1).
app_href(79,9,1296,1).
app_href(79,9,444,1).
app_href(79,9,1362,1).
app_href(79,9,546,1).
app_href(79,9,1189,1).
app_href(79,190,316,1).
app_href(79,190,736,1).
app_href(79,190,1083,1).
app_href(80,165,728,2).
app_href(80,166,736,1).
app_href(80,166,231,1).
app_href(80,167,1418,1).
app_href(80,167,423,1).
app_href(80,167,2,1).
app_href(80,168,1069,1).
app_href(80,169,695,1).
app_href(80,169,1382,2).
app_href(80,169,1150,2).
app_href(80,169,1329,2).
app_href(80,169,690,1).
app_href(80,170,1192,1).
app_href(80,170,132,1).
app_href(80,171,944,1).
app_href(80,171,673,1).
app_href(80,1202,1232,1).
app_href(80,1202,690,1).
app_href(80,172,1076,1).
app_href(80,172,1053,1).
app_href(80,173,1336,1).
app_href(80,945,736,1).
app_href(80,945,1346,1).
app_href(80,945,1384,1).
app_href(80,175,146,1).
app_href(80,175,2,1).
app_href(80,177,1403,1).
app_href(80,177,1341,1).
app_href(80,177,970,1).
app_href(80,178,829,1).
app_href(80,181,354,1).
app_href(80,182,1218,1).
app_href(80,182,420,1).
app_href(80,183,444,1).
app_href(80,1203,452,2).
app_href(80,1203,444,1).
app_href(80,1203,26,2).
app_href(80,1203,736,2).
app_href(80,1203,1346,2).
app_href(80,1203,1189,2).
app_href(80,1203,1014,1).
app_href(80,1203,586,1).
app_href(80,184,317,1).
app_href(80,184,677,1).
app_href(80,185,1442,1).
app_href(80,185,736,1).
app_href(80,185,140,1).
app_href(80,185,233,1).
app_href(80,186,728,1).
app_href(80,186,454,1).
app_href(80,187,444,1).
app_href(80,187,728,1).
app_href(80,29,536,1).
app_href(80,29,736,1).
app_href(80,190,316,1).
app_href(80,190,736,1).
app_href(80,190,1083,1).
app_href(83,1204,736,1).
app_href(83,1204,546,2).
app_href(83,1204,733,1).
app_href(83,1204,1409,1).
app_href(83,1204,904,1).
app_href(83,1204,696,1).
app_href(83,1204,734,1).
app_href(83,1204,1423,1).
app_href(83,1204,234,1).
app_href(83,621,1264,1).
app_href(83,621,531,1).
app_href(83,35,365,1).
app_href(83,35,6,1).
app_href(83,132,1123,1).
app_href(83,132,1248,1).
app_href(83,132,1362,1).
app_href(83,132,564,1).
app_href(83,132,1014,1).
app_href(83,132,733,1).
app_href(83,1205,546,1).
app_href(83,1205,733,1).
app_href(83,1205,1009,1).
app_href(83,87,444,1).
app_href(83,87,358,1).
app_href(83,1206,1479,3).
app_href(83,1206,736,3).
app_href(83,1206,982,1).
app_href(83,1206,546,2).
app_href(83,1206,981,1).
app_href(83,1206,1409,3).
app_href(83,1206,904,1).
app_href(83,1206,696,1).
app_href(83,1206,1480,1).
app_href(83,15,899,1).
app_href(83,15,1310,1).
app_href(83,15,668,1).
app_href(83,1207,546,1).
app_href(83,1207,1264,1).
app_href(83,1207,733,1).
app_href(83,736,1409,1).
app_href(83,736,736,1).
app_href(83,736,733,1).
app_href(83,1015,735,1).
app_href(83,1015,736,1).
app_href(83,1015,934,1).
app_href(83,1208,546,1).
app_href(83,1208,161,1).
app_href(83,1208,736,1).
app_href(83,1208,1346,1).
app_href(83,1208,733,1).
app_href(83,1208,568,1).
app_href(83,153,1123,1).
app_href(83,153,1362,1).
app_href(83,153,1456,1).
app_href(83,153,564,1).
app_href(83,153,1014,1).
app_href(83,153,733,1).
app_href(78,878,567,1).
app_href(78,878,99,1).
app_href(78,763,972,1).
app_href(78,763,367,1).
app_href(78,763,998,1).
app_href(78,488,125,1).
app_href(78,488,1056,1).
app_href(78,488,967,1).
app_href(78,488,1050,1).
app_href(78,488,2,1).
app_href(78,1082,1030,1).
app_href(78,1145,519,1).
app_href(78,1146,649,2).
app_href(78,1146,536,2).
app_href(78,1146,736,2).
app_href(78,144,1216,1).
app_href(78,1147,516,1).
app_href(78,1147,712,1).
app_href(78,1147,627,1).
app_href(78,1147,1464,1).
app_href(78,1075,1056,2).
app_href(78,1075,700,2).
app_href(78,87,1358,1).
app_href(78,87,444,1).
app_href(78,87,41,1).
app_href(78,87,9,2).
app_href(78,87,358,1).
app_href(78,87,728,1).
app_href(78,35,365,1).
app_href(78,1148,649,1).
app_href(78,1148,876,1).
app_href(78,1148,736,1).
app_href(78,1148,1346,1).
app_href(78,1148,1093,1).
app_href(78,1148,810,1).
app_href(78,1148,1360,1).
app_href(78,1149,516,1).
app_href(78,1149,627,1).
app_href(78,1149,1464,1).
app_href(78,398,994,1).
app_href(78,838,367,1).
app_href(78,838,331,1).
app_href(78,838,420,1).
app_href(78,838,736,1).
app_href(78,838,711,1).
app_href(78,838,1360,1).
app_href(78,1150,516,1).
app_href(78,1150,627,1).
app_href(78,1150,1464,1).
app_href(78,1151,155,1).
app_href(78,1151,435,1).
app_href(78,1151,327,1).
app_href(78,70,367,1).
app_href(78,70,736,1).
app_href(78,70,1346,1).
app_href(78,70,140,1).
app_href(78,70,1093,1).
app_href(78,70,810,1).
app_href(78,1152,627,1).
app_href(78,1152,296,1).
app_href(78,1152,1181,1).
app_href(78,1152,1464,1).
app_href(78,1152,564,1).
app_href(78,1084,972,1).
app_href(78,1153,682,1).
app_href(78,477,540,1).
app_href(78,477,736,2).
app_href(78,477,649,1).
app_href(78,477,140,1).
app_href(78,477,667,1).
app_href(78,477,41,1).
app_href(78,477,37,1).
app_href(78,477,277,1).
app_href(78,477,1168,1).
app_href(78,477,728,1).
app_href(78,477,9,3).
app_href(78,477,367,1).
app_href(78,477,1358,1).
app_href(78,477,219,1).
app_href(78,514,309,1).
app_href(78,514,574,1).
app_href(78,514,643,1).
app_href(78,514,1310,1).
app_href(78,514,736,1).
app_href(78,850,972,1).
app_href(78,850,1058,1).
app_href(78,850,998,1).
app_href(78,412,650,1).
app_href(78,1154,1359,1).
app_href(78,1155,605,1).
app_href(78,743,972,1).
app_href(78,743,468,1).
app_href(78,743,998,1).
app_href(78,1156,516,1).
app_href(78,1156,905,1).
app_href(78,1156,627,1).
app_href(78,1156,1050,1).
app_href(78,1156,1464,1).
app_href(78,1156,1384,1).
app_href(78,1156,1214,1).
app_href(78,1157,125,1).
app_href(78,1157,124,1).
app_href(78,1157,1408,1).
app_href(78,421,1217,1).
app_href(78,421,922,1).
/* (8458 entrees) */


/**
 * Table mot_txt
 */

mot_txt('introduction',1).
mot_txt('the',2).
mot_txt('java',3).
mot_txt('programming',4).
mot_txt('language',5).
mot_txt('and',6).
mot_txt('environment',7).
mot_txt('is',8).
mot_txt('designed',9).
mot_txt('to',10).
mot_txt('solve',11).
mot_txt('number',12).
mot_txt('of',13).
mot_txt('problems',14).
mot_txt('in',15).
mot_txt('modern',16).
mot_txt('practice',17).
mot_txt('started',18).
mot_txt('as',19).
mot_txt('part',20).
mot_txt('larger',21).
mot_txt('project',22).
mot_txt('develop',23).
mot_txt('advanced',24).
mot_txt('software',25).
mot_txt('for',26).
mot_txt('consumer',27).
mot_txt('electronics',28).
mot_txt('these',29).
mot_txt('devices',30).
mot_txt('are',31).
mot_txt('small',32).
mot_txt('reliable',33).
mot_txt('portable',34).
mot_txt('distributed',35).
mot_txt('real',36).
mot_txt('time',37).
mot_txt('embedded',38).
mot_txt('systems',39).
mot_txt('when',40).
mot_txt('we',41).
mot_txt('intended',42).
mot_txt('use',43).
mot_txt('but',44).
mot_txt('encountered',45).
mot_txt('initially',46).
mot_txt('were',47).
mot_txt('just',48).
mot_txt('compiler',49).
mot_txt('technology',50).
mot_txt('passed',51).
mot_txt('more',52).
mot_txt('emerged',53).
mot_txt('that',54).
mot_txt('best',55).
mot_txt('solved',56).
mot_txt('by',57).
mot_txt('changing',58).
mot_txt('simple',59).
mot_txt('object',60).
mot_txt('oriented',61).
mot_txt('network',62).
mot_txt('savvy',63).
mot_txt('interpreted',64).
mot_txt('robust',65).
mot_txt('secure',66).
mot_txt('architecture',67).
mot_txt('neutral',68).
mot_txt('high',69).
mot_txt('performance',70).
mot_txt('multithreaded',71).
mot_txt('dynamic',72).
mot_txt('one',73).
mot_txt('way',74).
mot_txt('characterize',75).
mot_txt('system',76).
mot_txt('with',77).
mot_txt('set',78).
mot_txt('buzzwords',79).
mot_txt('standard',80).
mot_txt('them',81).
mot_txt('describing',82).
mot_txt('here',83).
mot_txt('an',84).
mot_txt('explanation',85).
mot_txt('what',86).
mot_txt('mean',87).
mot_txt('those',88).
mot_txt('trying',89).
mot_txt('rchimedes',90).
mot_txt('inc',91).
mot_txt('fictitious',92).
mot_txt('company',93).
mot_txt('produces',94).
mot_txt('teach',95).
mot_txt('about',96).
mot_txt('basic',97).
mot_txt('physics',98).
mot_txt('this',99).
mot_txt('interact',100).
mot_txt('user',101).
mot_txt('providing',102).
mot_txt('not',103).
mot_txt('only',104).
mot_txt('text',105).
mot_txt('illustrations',106).
mot_txt('manner',107).
mot_txt('traditional',108).
mot_txt('textbook',109).
mot_txt('also',110).
mot_txt('lab',111).
mot_txt('benches',112).
mot_txt('on',113).
mot_txt('which',114).
mot_txt('experiments',115).
mot_txt('can',116).
mot_txt('be',117).
mot_txt('up',118).
mot_txt('their',119).
mot_txt('behavior',120).
mot_txt('simulated',121).
mot_txt('most',122).
mot_txt('experiment',123).
mot_txt('allows',124).
mot_txt('students',125).
mot_txt('put',126).
mot_txt('together',127).
mot_txt('levers',128).
mot_txt('pulleys',129).
mot_txt('see',130).
mot_txt('how',131).
mot_txt('they',132).
mot_txt('act',133).
mot_txt('italicized',134).
mot_txt('narrative',135).
mot_txt('trials',136).
mot_txt('tribulations',137).
mot_txt('archimedes',138).
mot_txt('designers',139).
mot_txt('used',140).
mot_txt('provide',141).
mot_txt('examples',142).
mot_txt('concepts',143).
mot_txt('wanted',144).
mot_txt('build',145).
mot_txt('could',146).
mot_txt('programmed',147).
mot_txt('easily',148).
mot_txt('without',149).
mot_txt('lot',150).
mot_txt('esoteric',151).
mot_txt('training',152).
mot_txt('leveraged',153).
mot_txt('today',154).
mot_txt('programmers',155).
mot_txt('working',156).
mot_txt('days',157).
mot_txt('doing',158).
mot_txt('so',159).
mot_txt('even',160).
mot_txt('though',161).
mot_txt('found',162).
mot_txt('was',163).
mot_txt('unsuitable',164).
mot_txt('closely',165).
mot_txt('possible',166).
mot_txt('order',167).
mot_txt('make',168).
mot_txt('comprehensible',169).
mot_txt('omits',170).
mot_txt('many',171).
mot_txt('rarely',172).
mot_txt('poorly',173).
mot_txt('understood',174).
mot_txt('confusing',175).
mot_txt('features',176).
mot_txt('our',177).
mot_txt('experience',178).
mot_txt('bring',179).
mot_txt('grief',180).
mot_txt('than',181).
mot_txt('benefit',182).
mot_txt('omitted',183).
mot_txt('primarily',184).
mot_txt('consist',185).
mot_txt('operator',186).
mot_txt('overloading',187).
mot_txt('although',188).
mot_txt('does',189).
mot_txt('have',190).
mot_txt('method',191).
mot_txt('multiple',192).
mot_txt('inheritance',193).
mot_txt('extensive',194).
mot_txt('automatic',195).
mot_txt('coercions',196).
mot_txt('added',197).
mot_txt('garbage',198).
mot_txt('collection',199).
mot_txt('thereby',200).
mot_txt('simplifying',201).
mot_txt('task',202).
mot_txt('making',203).
mot_txt('somewhat',204).
mot_txt('complicated',205).
mot_txt('common',206).
mot_txt('source',207).
mot_txt('complexity',208).
mot_txt('manyand',209).
mot_txt('applications',210).
mot_txt('storage',211).
mot_txt('management',212).
mot_txt('allocation',213).
mot_txt('freeing',214).
mot_txt('memory',215).
mot_txt('virtue',216).
mot_txt('having',217).
mot_txt('periodic',218).
mot_txt('being',219).
mot_txt('referenced',220).
mot_txt('makes',221).
mot_txt('easier',222).
mot_txt('it',223).
mot_txt('dramatically',224).
mot_txt('cuts',225).
mot_txt('down',226).
mot_txt('bugs',227).
mot_txt('folks',228).
mot_txt('at',229).
mot_txt('spend',230).
mot_txt('thinking',231).
mot_txt('instead',232).
mot_txt('spent',233).
mot_txt('mundane',234).
mot_txt('tasks',235).
mot_txt('central',236).
mot_txt('expertise',237).
mot_txt('teaching',238).
mot_txt('figuring',239).
mot_txt('out',240).
mot_txt('where',241).
mot_txt('wasted',242).
mot_txt('across',243).
mot_txt('20k',244).
mot_txt('lines',245).
mot_txt('code',246).
mot_txt('another',247).
mot_txt('aspect',248).
mot_txt('goals',249).
mot_txt('enable',250).
mot_txt('construction',251).
mot_txt('run',252).
mot_txt('stand',253).
mot_txt('alone',254).
mot_txt('machines',255).
mot_txt('interpreter',256).
mot_txt('libraries',257).
mot_txt('footprint',258).
mot_txt('size',259).
mot_txt('important',260).
mot_txt('downloaded',261).
mot_txt('over',262).
mot_txt('net',263).
mot_txt('unfortunately',264).
mot_txt('overused',265).
mot_txt('industry',266).
mot_txt('design',267).
mot_txt('very',268).
mot_txt('powerful',269).
mot_txt('because',270).
mot_txt('facilitates',271).
mot_txt('clean',272).
mot_txt('definition',273).
mot_txt('interfaces',274).
mot_txt('reusable',275).
mot_txt('ics',276).
mot_txt('simply',277).
mot_txt('stated',278).
mot_txt('technique',279).
mot_txt('focuses',280).
mot_txt('data',281).
mot_txt('objects',282).
mot_txt('analogy',283).
mot_txt('carpentry',284).
mot_txt('carpenter',285).
mot_txt('would',286).
mot_txt('mostly',287).
mot_txt('concerned',288).
mot_txt('chair',289).
mot_txt('he',290).
mot_txt('building',291).
mot_txt('secondarily',292).
mot_txt('tools',293).
mot_txt('non',294).
mot_txt('think',295).
mot_txt('his',296).
mot_txt('mechanism',297).
mot_txt('defining',298).
mot_txt('modules',299).
mot_txt('plug',300).
mot_txt('play',301).
mot_txt('facilities',302).
mot_txt('essentially',303).
mot_txt('extensions',304).
mot_txt('from',305).
mot_txt('objectivefor',306).
mot_txt('resolution',307).
mot_txt('had',308).
mot_txt('lots',309).
mot_txt('things',310).
mot_txt('simulation',311).
mot_txt('among',312).
mot_txt('ropes',313).
mot_txt('elastic',314).
mot_txt('bands',315).
mot_txt('initialversion',316).
mot_txt('product',317).
mot_txt('ended',318).
mot_txt('pretty',319).
mot_txt('big',320).
mot_txt('write',321).
mot_txt('separate',322).
mot_txt('versus',323).
mot_txt('rewrote',324).
mot_txt('application',325).
mot_txt('style',326).
mot_txt('define',327).
mot_txt('represented',328).
mot_txt('aspects',329).
mot_txt('then',330).
mot_txt('defined',331).
mot_txt('variations',332).
mot_txt('subclasses',333).
mot_txt('type',334).
mot_txt('came',335).
mot_txt('add',336).
mot_txt('chains',337).
mot_txt('snap',338).
mot_txt('been',339).
mot_txt('written',340).
mot_txt('before',341).
mot_txt('rather',342).
mot_txt('writing',343).
mot_txt('whole',344).
mot_txt('new',345).
mot_txt('has',346).
mot_txt('library',347).
mot_txt('routines',348).
mot_txt('coping',349).
mot_txt('tcp',350).
mot_txt('ip',351).
mot_txt('protocols',352).
mot_txt('like',353).
mot_txt('http',354).
mot_txt('ftp',355).
mot_txt('creating',356).
mot_txt('connections',357).
mot_txt('much',358).
mot_txt('inor',359).
mot_txt('open',360).
mot_txt('access',361).
mot_txt('thenet',362).
mot_txt('via',363).
mot_txt('urls',364).
mot_txt('same',365).
mot_txt('ease',366).
mot_txt('accessing',367).
mot_txt('local',368).
mot_txt('file',369).
mot_txt('built',370).
mot_txt('stuff',371).
mot_txt('cd',372).
mot_txt('rom',373).
mot_txt('some',374).
mot_txt('ideas',375).
mot_txt('interactive',376).
mot_txt('learning',377).
mot_txt('games',378).
mot_txt('try',379).
mot_txt('next',380).
mot_txt('example',381).
mot_txt('allow',382).
mot_txt('different',383).
mot_txt('computers',384).
mot_txt('cooperate',385).
mot_txt('machine',386).
mot_txt('all',387).
mot_txt('networking',388).
mot_txt('seen',389).
mot_txt('required',390).
mot_txt('specialists',391).
mot_txt('gave',392).
mot_txt('programs',393).
mot_txt('must',394).
mot_txt('variety',395).
mot_txt('ways',396).
mot_txt('puts',397).
mot_txt('emphasis',398).
mot_txt('early',399).
mot_txt('checking',400).
mot_txt('later',401).
mot_txt('runtime',402).
mot_txt('eliminating',403).
mot_txt('situations',404).
mot_txt('error',405).
mot_txt('prone',406).
mot_txt('advantages',407).
mot_txt('strongly',408).
mot_txt('typed',409).
mot_txt('compile',410).
mot_txt('inherits',411).
mot_txt('loopholes',412).
mot_txt('relatively',413).
mot_txt('lax',414).
mot_txt('particularly',415).
mot_txt('procedure',416).
mot_txt('declarations',417).
mot_txt('require',418).
mot_txt('do',419).
mot_txt('supportstyle',420).
mot_txt('implicit',421).
mot_txt('linker',422).
mot_txt('understands',423).
mot_txt('repeats',424).
mot_txt('checks',425).
mot_txt('done',426).
mot_txt('guard',427).
mot_txt('against',428).
mot_txt('version',429).
mot_txt('mismatch',430).
mot_txt('single',431).
mot_txt('biggest',432).
mot_txt('difference',433).
mot_txt('between',434).
mot_txt('pointer',435).
mot_txt('model',436).
mot_txt('eliminates',437).
mot_txt('possibility',438).
mot_txt('overwriting',439).
mot_txt('corrupting',440).
mot_txt('arithmetic',441).
mot_txt('true',442).
mot_txt('arrays',443).
mot_txt('subscript',444).
mot_txt('performed',445).
mot_txt('addition',446).
mot_txt('turn',447).
mot_txt('arbitrary',448).
mot_txt('integer',449).
mot_txt('into',450).
mot_txt('casting',451).
mot_txt('basically',452).
mot_txt('inpretty',453).
mot_txt('quickly',454).
mot_txt('schedule',455).
mot_txt('kept',456).
mot_txt('slipping',457).
mot_txt('through',458).
mot_txt('trouble',459).
mot_txt('corruptionversions',460).
mot_txt('sync',461).
mot_txt('interface',462).
mot_txt('mismatches',463).
mot_txt('gained',464).
mot_txt('let',465).
mot_txt('pull',466).
mot_txt('strange',467).
mot_txt('tricks',468).
mot_txt('theircode',469).
mot_txt('paid',470).
mot_txt('quality',471).
mot_txt('assurance',472).
mot_txt('reissue',473).
mot_txt('after',474).
mot_txt('first',475).
mot_txt('release',476).
mot_txt('slipped',477).
mot_txt('while',478).
mot_txt('doesn',479).
mot_txt('qa',480).
mot_txt('problem',481).
mot_txt('go',482).
mot_txt('away',483).
mot_txt('significantly',484).
mot_txt('languages',485).
mot_txt('lisp',486).
mot_txt('tcl',487).
mot_txt('smalltalk',488).
mot_txt('often',489).
mot_txt('prototyping',490).
mot_txt('reasons',491).
mot_txt('success',492).
mot_txt('you',493).
mot_txt('don',494).
mot_txt('worry',495).
mot_txt('or',496).
mot_txt('fearless',497).
mot_txt('dealing',498).
mot_txt('getting',499).
mot_txt('corrupted',500).
mot_txt('there',501).
mot_txt('no',502).
mot_txt('pointers',503).
mot_txt('accidentally',504).
mot_txt('overwrite',505).
mot_txt('end',506).
mot_txt('buffer',507).
mot_txt('cannot',508).
mot_txt('gain',509).
mot_txt('unauthorized',510).
mot_txt('happen',511).
mot_txt('reason',512).
mot_txt('good',513).
mot_txt('pin',514).
mot_txt('decisionstoo',515).
mot_txt('uses',516).
mot_txt('approach',517).
mot_txt('dilemma',518).
mot_txt('forces',519).
mot_txt('choices',520).
mot_txt('explicitly',521).
mot_txt('becauseit',522).
mot_txt('static',523).
mot_txt('typing',524).
mot_txt('enforces',525).
mot_txt('along',526).
mot_txt('comes',527).
mot_txt('assistance',528).
mot_txt('invocations',529).
mot_txt('if',530).
mot_txt('get',531).
mot_txt('something',532).
mot_txt('wrong',533).
mot_txt('informed',534).
mot_txt('havetoworry',535).
mot_txt('invocation',536).
mot_txt('networked',537).
mot_txt('environments',538).
mot_txt('toward',539).
mot_txt('placed',540).
mot_txt('security',541).
mot_txt('enables',542).
mot_txt('virus',543).
mot_txt('free',544).
mot_txt('tamper',545).
mot_txt('authentication',546).
mot_txt('techniques',547).
mot_txt('based',548).
mot_txt('public',549).
mot_txt('key',550).
mot_txt('encryption',551).
mot_txt('strong',552).
mot_txt('interplay',553).
mot_txt('changes',554).
mot_txt('semantics',555).
mot_txt('makeit',556).
mot_txt('impossible',557).
mot_txt('forge',558).
mot_txt('structures',559).
mot_txt('private',560).
mot_txt('closes',561).
mot_txt('door',562).
mot_txt('activities',563).
mot_txt('viruses',564).
mot_txt('someone',565).
mot_txt('wrote',566).
mot_txt('interesting',567).
mot_txt('patch',568).
mot_txt('pc',569).
mot_txt('posted',570).
mot_txt('major',571).
mot_txt('bulletin',572).
mot_txt('boards',573).
mot_txt('since',574).
mot_txt('available',575).
mot_txt('people',576).
mot_txt('hadn',577).
mot_txt('checked',578).
mot_txt('seemed',579).
mot_txt('work',580).
mot_txt('until',581).
mot_txt('april',582).
mot_txt('1st',583).
mot_txt('thousands',584).
mot_txt('discovered',585).
mot_txt('rude',586).
mot_txt('pictures',587).
mot_txt('popping',588).
mot_txt('children',589).
mot_txt('lessons',590).
mot_txt('needless',591).
mot_txt('say',592).
mot_txt('responsible',593).
mot_txt('incident',594).
mot_txt('still',595).
mot_txt('damage',596).
mot_txt('control',597).
mot_txt('support',598).
mot_txt('networks',599).
mot_txt('general',600).
mot_txt('composed',601).
mot_txt('cpu',602).
mot_txt('operating',603).
mot_txt('architectures',604).
mot_txt('execute',605).
mot_txt('anywhere',606).
mot_txt('generates',607).
mot_txt('format',608).
mot_txt('compiled',609).
mot_txt('executable',610).
mot_txt('processors',611).
mot_txt('given',612).
mot_txt('presence',613).
mot_txt('useful',614).
mot_txt('distribution',615).
mot_txt('present',616).
mot_txt('personal',617).
mot_txt('computer',618).
mot_txt('market',619).
mot_txt('writers',620).
mot_txt('produce',621).
mot_txt('versions',622).
mot_txt('compatible',623).
mot_txt('ibm',624).
mot_txt('apple',625).
mot_txt('macintosh',626).
mot_txt('windows',627).
mot_txt('nt',628).
mot_txt('diversifying',629).
mot_txt('moving',630).
mot_txt('off',631).
mot_txt('powerpc',632).
mot_txt('production',633).
mot_txt('runs',634).
mot_txt('platforms',635).
mot_txt('becomes',636).
mot_txt('nearly',637).
mot_txt('generating',638).
mot_txt('bytecode',639).
mot_txt('instructions',640).
mot_txt('nothing',641).
mot_txt('particular',642).
mot_txt('both',643).
mot_txt('easy',644).
mot_txt('interpret',645).
mot_txt('any',646).
mot_txt('translated',647).
mot_txt('native',648).
mot_txt('fly',649).
mot_txt('producing',650).
mot_txt('largest',651).
mot_txt('large',652).
mot_txt('enough',653).
mot_txt('afford',654).
mot_txt('port',655).
mot_txt('effort',656).
mot_txt('didn',657).
mot_txt('really',658).
mot_txt('pay',659).
mot_txt('couldn',660).
mot_txt('mips',661).
mot_txt('catch',662).
mot_txt('wave',663).
mot_txt('happening',664).
mot_txt('competitor',665).
mot_txt('jumped',666).
mot_txt('chunk',667).
mot_txt('unlikeand',668).
mot_txt('implementation',669).
mot_txt('dependent',670).
mot_txt('specification',671).
mot_txt('sizes',672).
mot_txt('primitive',673).
mot_txt('types',674).
mot_txt('specified',675).
mot_txt('int',676).
mot_txt('always',677).
mot_txt('means',678).
mot_txt('signed',679).
mot_txt('two',680).
mot_txt('complement',681).
mot_txt('32',682).
mot_txt('bit',683).
mot_txt('float',684).
mot_txt('ieee',685).
mot_txt('7',686).
mot_txt('4',687).
mot_txt('floating',688).
mot_txt('point',689).
mot_txt('feasible',690).
mot_txt('day',691).
mot_txt('age',692).
mot_txt('cpus',693).
mot_txt('share',694).
mot_txt('characteristics',695).
mot_txt('abstract',696).
mot_txt('window',697).
mot_txt('class',698).
mot_txt('implementations',699).
mot_txt('unix',700).
mot_txt('itself',701).
mot_txt('quite',702).
mot_txt('ansi',703).
mot_txt('portability',704).
mot_txt('boundary',705).
mot_txt('posix',706).
mot_txt('subset',707).
mot_txt('bytecodes',708).
mot_txt('stored',709).
mot_txt('linking',710).
mot_txt('incremental',711).
mot_txt('lightweight',712).
mot_txt('process',713).
mot_txt('development',714).
mot_txt('rapid',715).
mot_txt('exploratory',716).
mot_txt('stream',717).
mot_txt('information',718).
mot_txt('carried',719).
mot_txt('iswhat',720).
mot_txt('amenable',721).
mot_txt('debugging',722).
mot_txt('waiting',723).
mot_txt('link',724).
mot_txt('tracking',725).
mot_txt('senseless',726).
mot_txt('changed',727).
mot_txt('files',728).
mot_txt('despite',729).
mot_txt('using',730).
mot_txt('fancy',731).
mot_txt('facility',732).
mot_txt('caused',733).
mot_txt('track',734).
mot_txt('procedures',735).
mot_txt('declared',736).
mot_txt('inconsistently',737).
mot_txt('various',738).
mot_txt('parts',739).
mot_txt('couple',740).
mot_txt('months',741).
mot_txt('lost',742).
mot_txt('usually',743).
mot_txt('adequate',744).
mot_txt('higher',745).
mot_txt('running',746).
mot_txt('accustomed',747).
mot_txt('normal',748).
mot_txt('loader',749).
mot_txt('putting',750).
mot_txt('final',751).
mot_txt('generator',752).
mot_txt('codes',753).
mot_txt('mind',754).
mot_txt('actual',755).
mot_txt('generally',756).
mot_txt('efficient',757).
mot_txt('produced',758).
mot_txt('register',759).
mot_txt('andsome',760).
mot_txt('optimization',761).
mot_txt('calls',762).
mot_txt('per',763).
mot_txt('second',764).
mot_txt('sun',765).
mot_txt('microsystems',766).
mot_txt('sparcstation',767).
mot_txt('converted',768).
mot_txt('almost',769).
mot_txt('indistinguishable',770).
mot_txt('nativeor',771).
mot_txt('guide',772).
mot_txt('starting',773).
mot_txt('did',774).
mot_txt('prototype',775).
mot_txt('impressed',776).
mot_txt('investors',777).
mot_txt('theygot',778).
mot_txt('funded',779).
mot_txt('help',780).
mot_txt('simulations',781).
mot_txt('fast',782).
mot_txt('rewritten',783).
mot_txt('going',784).
mot_txt('world',785).
mot_txt('around',786).
mot_txt('us',787).
mot_txt('multithreading',788).
mot_txt('threads',789).
mot_txt('deal',790).
mot_txt('once',791).
mot_txt('difficult',792).
mot_txt('conventional',793).
mot_txt('threadedand',794).
mot_txt('sophisticated',795).
mot_txt('synchronization',796).
mot_txt('primitives',797).
mot_txt('widely',798).
mot_txt('monitor',799).
mot_txt('condition',800).
mot_txt('variable',801).
mot_txt('paradigm',802).
mot_txt('introduced',803).
mot_txt('bya',804).
mot_txt('hoare',805).
mot_txt('integrating',806).
mot_txt('classes',807).
mot_txt('become',808).
mot_txt('integration',809).
mot_txt('xerox',810).
mot_txt('cedar',811).
mot_txt('mesa',812).
mot_txt('other',813).
mot_txt('benefits',814).
mot_txt('better',815).
mot_txt('responsiveness',816).
mot_txt('limitedhowever',817).
mot_txt('underlying',818).
mot_txt('platform',819).
mot_txt('runningon',820).
mot_txt('top',821).
mot_txt('limits',822).
mot_txt('pulled',823).
mot_txt('wheels',824).
mot_txt('turning',825).
mot_txt('rocking',826).
mot_txt('input',827).
mot_txt('tracked',828).
mot_txt('threaded',829).
mot_txt('form',830).
mot_txt('each',831).
mot_txt('bemanually',832).
mot_txt('intermixed',833).
mot_txt('event',834).
mot_txt('loop',835).
mot_txt('made',836).
mot_txt('little',837).
mot_txt('cleaner',838).
mot_txt('mess',839).
mot_txt('became',840).
mot_txt('fragile',841).
mot_txt('hard',842).
mot_txt('understand',843).
mot_txt('pulling',844).
mot_txt('originally',845).
mot_txt('serialized',846).
mot_txt('communication',847).
mot_txt('slow',848).
mot_txt('multithreadedstyle',849).
mot_txt('trivial',850).
mot_txt('overlap',851).
mot_txt('thanor',852).
mot_txt('adapt',853).
mot_txt('evolving',854).
mot_txt('side',855).
mot_txt('effect',856).
mot_txt('implemented',857).
mot_txt('a',858).
mot_txt('components',859).
mot_txt('buys',860).
mot_txt('intheir',861).
mot_txt('its',862).
mot_txt('distributes',863).
mot_txt('will',864).
mot_txt('certainly',865).
mot_txt('recompileand',866).
mot_txt('redistribute',867).
mot_txt('own',868).
mot_txt('gets',869).
mot_txt('independently',870).
mot_txt('os',871).
mot_txt('vendor',872).
mot_txt('result',873).
mot_txt('upgrade',874).
mot_txt('break',875).
mot_txt('avoid',876).
mot_txt('extraordinarily',877).
mot_txt('effectively',878).
mot_txt('oo',879).
mot_txt('directly',880).
mot_txt('graphics',881).
mot_txt('released',882).
mot_txt('several',883).
mot_txt('manufacturers',884).
mot_txt('bundled',885).
mot_txt('customers',886).
mot_txt('bought',887).
mot_txt('dismay',888).
mot_txt('old',889).
mot_txt('longer',890).
mot_txt('worked',891).
mot_txt('life',892).
mot_txt('backwards',893).
mot_txt('compatibility',894).
mot_txt('isn',895).
mot_txt('priority',896).
mot_txt('never',897).
mot_txt('such',898).
mot_txt('ability',899).
mot_txt('change',900).
mot_txt('severely',901).
mot_txt('hindered',902).
mot_txt('expect',903).
mot_txt('recompile',904).
mot_txt('interconnections',905).
mot_txt('completely',906).
mot_txt('avoids',907).
mot_txt('straightforward',908).
mot_txt('freely',909).
mot_txt('methods',910).
mot_txt('instancevariables',911).
mot_txt('clients',912).
mot_txt('specifies',913).
mot_txt('perform',914).
mot_txt('leaves',915).
mot_txt('should',916).
mot_txt('implement',917).
mot_txt('implements',918).
mot_txt('implementing',919).
mot_txt('contained',920).
mot_txt('incontrast',921).
mot_txt('subclassing',922).
mot_txt('passes',923).
mot_txt('superclass',924).
mot_txt('subclass',925).
mot_txt('inherit',926).
mot_txt('promote',927).
mot_txt('flexibility',928).
mot_txt('reusability',929).
mot_txt('connecting',930).
mot_txt('terms',931).
mot_txt('howthey',932).
mot_txt('representation',933).
mot_txt('named',934).
mot_txt('instances',935).
mot_txt('contain',936).
mot_txt('definitions',937).
mot_txt('aor',938).
mot_txt('program',939).
mot_txt('know',940).
mot_txt('noway',941).
mot_txt('find',942).
mot_txt('however',943).
mot_txt('finding',944).
mot_txt('becausecasts',945).
mot_txt('trust',946).
mot_txt('cast',947).
mot_txt('hand',948).
mot_txt('inand',949).
mot_txt('trusts',950).
mot_txt('re',951).
mot_txt('right',952).
mot_txt('thing',953).
mot_txt('look',954).
mot_txt('string',955).
mot_txt('containing',956).
mot_txt('name',957).
mot_txt('compute',958).
mot_txt('dynamically',959).
mot_txt('linked',960).
mot_txt('expand',961).
mot_txt('revenue',962).
mot_txt('architect',963).
mot_txt('aftermarket',964).
mot_txt('extend',965).
mot_txt('barely',966).
mot_txt('hire',967).
mot_txt('newprogrammers',968).
/* (968 entrees) */


/**
 * Table txt_doc
 */

txt_doc(625,1,1).
txt_doc(512,1,1).
txt_doc(801,1,1).
txt_doc(286,1,1).
txt_doc(389,1,1).
txt_doc(751,1,1).
txt_doc(137,1,1).
txt_doc(601,1,1).
txt_doc(864,1,1).
txt_doc(252,1,1).
txt_doc(446,1,1).
txt_doc(831,1,1).
txt_doc(336,1,1).
txt_doc(133,1,1).
txt_doc(187,1,1).
txt_doc(74,1,1).
txt_doc(99,1,1).
txt_doc(722,1,1).
txt_doc(379,1,1).
txt_doc(30,1,1).
txt_doc(665,1,1).
txt_doc(56,1,1).
txt_doc(896,1,1).
txt_doc(781,1,1).
txt_doc(862,1,1).
txt_doc(238,1,1).
txt_doc(826,1,1).
txt_doc(469,1,1).
txt_doc(415,1,1).
txt_doc(793,1,1).
txt_doc(48,1,1).
txt_doc(666,1,1).
txt_doc(686,1,1).
txt_doc(295,1,1).
txt_doc(16,1,1).
txt_doc(42,1,2).
txt_doc(365,1,1).
txt_doc(9,1,1).
txt_doc(34,1,1).
txt_doc(411,1,2).
txt_doc(943,1,1).
txt_doc(120,1,1).
txt_doc(179,1,1).
txt_doc(753,1,1).
txt_doc(719,1,1).
txt_doc(456,1,1).
txt_doc(142,1,1).
txt_doc(763,1,1).
txt_doc(43,1,1).
txt_doc(97,1,1).
txt_doc(184,1,1).
txt_doc(618,1,1).
txt_doc(632,1,1).
txt_doc(556,1,1).
txt_doc(115,1,1).
txt_doc(348,1,1).
txt_doc(676,1,1).
txt_doc(88,1,1).
txt_doc(38,1,1).
txt_doc(803,1,1).
txt_doc(458,1,1).
txt_doc(160,1,1).
txt_doc(942,1,1).
txt_doc(673,1,1).
txt_doc(541,1,1).
txt_doc(387,1,1).
txt_doc(598,1,1).
txt_doc(679,1,1).
txt_doc(466,1,1).
txt_doc(707,1,1).
txt_doc(775,1,1).
txt_doc(602,1,2).
txt_doc(368,1,1).
txt_doc(351,1,1).
txt_doc(652,1,1).
txt_doc(915,1,1).
txt_doc(714,1,1).
txt_doc(815,1,1).
txt_doc(870,1,1).
txt_doc(455,1,1).
txt_doc(250,1,1).
txt_doc(811,1,1).
txt_doc(333,1,1).
txt_doc(925,1,1).
txt_doc(350,1,1).
txt_doc(209,1,1).
txt_doc(262,1,1).
txt_doc(112,1,1).
txt_doc(260,1,1).
txt_doc(210,1,1).
txt_doc(531,1,1).
txt_doc(814,1,1).
txt_doc(617,1,1).
txt_doc(661,1,1).
txt_doc(486,1,1).
txt_doc(481,1,2).
txt_doc(375,1,1).
txt_doc(345,1,1).
txt_doc(8,1,1).
txt_doc(504,1,1).
txt_doc(678,1,2).
txt_doc(895,1,1).
txt_doc(905,1,1).
txt_doc(79,1,1).
txt_doc(271,1,1).
txt_doc(588,1,1).
txt_doc(396,1,1).
txt_doc(175,1,1).
txt_doc(889,1,1).
txt_doc(537,1,2).
txt_doc(12,1,1).
txt_doc(44,1,1).
txt_doc(107,1,1).
txt_doc(143,1,1).
txt_doc(178,1,1).
txt_doc(106,1,1).
txt_doc(450,1,1).
txt_doc(888,1,1).
txt_doc(824,1,1).
txt_doc(37,1,1).
txt_doc(154,1,1).
txt_doc(494,1,1).
txt_doc(60,1,2).
txt_doc(309,1,1).
txt_doc(25,1,1).
txt_doc(373,1,1).
txt_doc(342,1,1).
txt_doc(259,1,2).
txt_doc(496,1,2).
txt_doc(23,1,1).
txt_doc(542,1,1).
txt_doc(258,1,1).
txt_doc(662,1,1).
txt_doc(715,1,3).
txt_doc(424,1,1).
txt_doc(18,1,1).
txt_doc(452,1,1).
txt_doc(695,1,1).
txt_doc(735,1,1).
txt_doc(239,1,1).
txt_doc(790,1,1).
txt_doc(119,1,1).
txt_doc(700,1,1).
txt_doc(920,1,1).
txt_doc(642,1,1).
txt_doc(310,1,1).
txt_doc(189,1,1).
txt_doc(427,1,1).
txt_doc(170,1,1).
txt_doc(804,1,2).
txt_doc(520,1,1).
txt_doc(5,1,1).
txt_doc(505,1,2).
txt_doc(2,1,1).
txt_doc(860,1,2).
txt_doc(497,1,1).
txt_doc(620,1,1).
txt_doc(283,1,1).
txt_doc(548,1,1).
txt_doc(321,1,1).
txt_doc(856,1,1).
txt_doc(39,1,1).
txt_doc(265,1,1).
txt_doc(117,1,1).
txt_doc(138,1,1).
txt_doc(234,1,1).
txt_doc(485,1,1).
txt_doc(938,1,1).
txt_doc(770,1,1).
txt_doc(341,1,1).
txt_doc(849,1,1).
txt_doc(313,1,1).
txt_doc(45,1,1).
txt_doc(569,1,1).
txt_doc(869,1,1).
txt_doc(702,1,1).
txt_doc(377,1,3).
txt_doc(593,1,1).
txt_doc(606,1,1).
txt_doc(261,1,1).
txt_doc(712,1,1).
txt_doc(550,1,1).
txt_doc(713,1,1).
txt_doc(693,1,1).
txt_doc(68,2,1).
txt_doc(861,2,1).
txt_doc(754,2,1).
txt_doc(797,2,1).
txt_doc(268,2,1).
txt_doc(30,2,1).
txt_doc(811,2,1).
txt_doc(899,2,1).
txt_doc(622,2,1).
txt_doc(537,2,1).
txt_doc(908,2,1).
txt_doc(43,2,1).
txt_doc(111,2,1).
txt_doc(601,2,1).
txt_doc(293,2,1).
txt_doc(916,2,1).
txt_doc(256,2,1).
txt_doc(303,2,1).
txt_doc(764,2,1).
txt_doc(744,2,1).
txt_doc(823,2,1).
txt_doc(593,2,1).
txt_doc(694,2,1).
txt_doc(477,2,1).
txt_doc(564,2,1).
txt_doc(812,2,1).
txt_doc(946,2,1).
txt_doc(743,2,1).
txt_doc(656,2,1).
txt_doc(150,2,1).
txt_doc(321,2,1).
txt_doc(540,2,1).
txt_doc(252,2,1).
txt_doc(737,2,1).
txt_doc(151,2,1).
txt_doc(721,2,1).
txt_doc(772,2,1).
txt_doc(214,2,1).
txt_doc(630,2,1).
txt_doc(219,2,1).
txt_doc(896,2,1).
txt_doc(456,2,1).
txt_doc(789,2,1).
txt_doc(181,2,2).
txt_doc(773,2,1).
txt_doc(122,2,1).
txt_doc(603,2,1).
txt_doc(620,2,1).
txt_doc(467,2,1).
txt_doc(28,2,1).
txt_doc(813,2,1).
txt_doc(749,2,1).
txt_doc(61,2,1).
txt_doc(641,2,1).
txt_doc(551,2,1).
txt_doc(404,2,2).
txt_doc(711,2,1).
txt_doc(571,2,1).
txt_doc(115,2,1).
txt_doc(134,2,1).
txt_doc(704,2,1).
txt_doc(123,2,1).
txt_doc(279,2,1).
txt_doc(258,2,1).
txt_doc(326,2,1).
txt_doc(292,2,1).
txt_doc(904,2,1).
txt_doc(631,2,1).
txt_doc(825,2,1).
txt_doc(42,2,1).
txt_doc(444,2,1).
txt_doc(164,2,1).
txt_doc(373,2,1).
txt_doc(114,2,1).
txt_doc(92,2,1).
txt_doc(263,2,1).
txt_doc(409,2,1).
txt_doc(274,2,1).
txt_doc(582,2,1).
txt_doc(516,2,1).
txt_doc(322,2,1).
txt_doc(714,2,1).
txt_doc(478,2,1).
txt_doc(665,2,2).
txt_doc(526,2,2).
txt_doc(175,2,1).
txt_doc(901,2,1).
txt_doc(728,2,1).
txt_doc(938,2,1).
txt_doc(428,2,1).
txt_doc(508,2,1).
txt_doc(621,2,1).
txt_doc(23,2,1).
txt_doc(917,2,1).
txt_doc(759,2,1).
txt_doc(382,2,3).
txt_doc(922,2,1).
txt_doc(465,2,1).
txt_doc(831,2,1).
txt_doc(378,2,1).
txt_doc(741,2,1).
txt_doc(227,2,1).
txt_doc(88,2,1).
txt_doc(189,2,1).
txt_doc(73,2,1).
txt_doc(793,2,1).
txt_doc(89,2,2).
txt_doc(523,2,1).
txt_doc(740,2,1).
txt_doc(557,2,1).
txt_doc(719,2,1).
txt_doc(503,2,1).
txt_doc(278,2,1).
txt_doc(573,2,1).
txt_doc(716,2,1).
txt_doc(82,2,1).
txt_doc(561,2,1).
txt_doc(900,2,1).
txt_doc(767,2,1).
txt_doc(949,2,1).
txt_doc(203,2,1).
txt_doc(833,2,1).
txt_doc(821,2,1).
txt_doc(361,2,1).
txt_doc(889,2,1).
txt_doc(221,2,1).
txt_doc(213,2,1).
txt_doc(939,2,1).
txt_doc(153,2,1).
txt_doc(233,2,2).
txt_doc(247,2,1).
txt_doc(314,2,1).
txt_doc(717,2,1).
txt_doc(869,2,2).
txt_doc(496,2,1).
txt_doc(229,2,1).
txt_doc(110,2,1).
txt_doc(237,2,1).
txt_doc(5,2,2).
txt_doc(902,2,1).
txt_doc(568,2,1).
txt_doc(679,2,1).
txt_doc(757,2,1).
txt_doc(65,2,2).
txt_doc(381,2,1).
txt_doc(330,2,1).
txt_doc(128,2,1).
txt_doc(471,2,2).
txt_doc(678,2,2).
txt_doc(926,2,1).
txt_doc(407,2,1).
txt_doc(610,2,1).
txt_doc(423,2,1).
txt_doc(335,2,1).
txt_doc(52,2,1).
txt_doc(947,2,1).
txt_doc(21,2,1).
txt_doc(71,2,1).
txt_doc(324,2,1).
txt_doc(308,2,1).
txt_doc(699,2,1).
txt_doc(712,2,1).
txt_doc(644,2,2).
txt_doc(885,2,2).
txt_doc(271,2,1).
txt_doc(655,2,1).
txt_doc(615,2,1).
txt_doc(196,2,1).
txt_doc(934,2,1).
txt_doc(295,2,1).
txt_doc(856,2,2).
txt_doc(507,2,1).
txt_doc(129,2,3).
txt_doc(967,2,1).
txt_doc(592,2,2).
txt_doc(393,2,1).
txt_doc(940,2,2).
txt_doc(649,2,1).
txt_doc(609,2,2).
txt_doc(64,3,1).
txt_doc(604,3,1).
txt_doc(454,3,1).
txt_doc(840,3,1).
txt_doc(174,3,1).
txt_doc(353,3,1).
txt_doc(290,3,1).
txt_doc(680,3,1).
txt_doc(273,3,1).
txt_doc(165,3,1).
txt_doc(308,3,1).
txt_doc(18,3,1).
txt_doc(41,3,1).
txt_doc(460,3,1).
txt_doc(661,3,1).
txt_doc(582,3,1).
txt_doc(766,3,1).
txt_doc(213,3,1).
txt_doc(133,3,1).
txt_doc(107,3,1).
txt_doc(648,3,1).
txt_doc(258,3,1).
txt_doc(189,3,1).
txt_doc(175,3,1).
txt_doc(202,3,1).
txt_doc(100,3,1).
txt_doc(448,3,1).
txt_doc(801,3,1).
txt_doc(877,3,1).
txt_doc(565,3,1).
txt_doc(890,3,1).
txt_doc(898,3,1).
txt_doc(426,3,1).
txt_doc(437,3,1).
txt_doc(425,3,1).
txt_doc(687,3,1).
txt_doc(447,3,1).
txt_doc(262,3,1).
txt_doc(747,3,1).
txt_doc(361,3,1).
txt_doc(161,3,1).
txt_doc(371,3,1).
txt_doc(413,3,1).
txt_doc(848,3,1).
txt_doc(673,3,1).
txt_doc(435,3,1).
txt_doc(772,3,1).
txt_doc(312,3,1).
txt_doc(483,3,1).
txt_doc(113,3,1).
txt_doc(820,3,1).
txt_doc(6,3,1).
txt_doc(719,3,1).
txt_doc(827,3,1).
txt_doc(159,3,1).
txt_doc(520,3,1).
txt_doc(84,3,1).
txt_doc(764,3,1).
txt_doc(47,3,1).
txt_doc(780,3,2).
txt_doc(179,3,1).
txt_doc(264,3,1).
txt_doc(146,3,1).
txt_doc(638,3,1).
txt_doc(2,3,1).
txt_doc(768,3,1).
txt_doc(316,3,1).
txt_doc(591,3,1).
txt_doc(527,3,3).
txt_doc(432,3,1).
txt_doc(411,3,1).
txt_doc(625,3,1).
txt_doc(947,3,1).
txt_doc(621,3,1).
txt_doc(366,3,1).
txt_doc(699,3,1).
txt_doc(805,3,1).
txt_doc(479,3,1).
txt_doc(596,3,1).
txt_doc(789,3,2).
txt_doc(549,3,1).
txt_doc(658,3,1).
txt_doc(398,3,1).
txt_doc(11,3,1).
txt_doc(576,3,1).
txt_doc(509,3,2).
txt_doc(650,3,1).
txt_doc(852,3,1).
txt_doc(196,3,1).
txt_doc(93,3,1).
txt_doc(911,3,1).
txt_doc(12,3,1).
txt_doc(455,3,1).
txt_doc(404,3,1).
txt_doc(294,3,1).
txt_doc(660,3,1).
txt_doc(753,3,1).
txt_doc(63,3,1).
txt_doc(613,3,2).
txt_doc(778,3,1).
txt_doc(577,3,1).
txt_doc(379,3,1).
txt_doc(696,3,1).
txt_doc(417,3,1).
txt_doc(779,3,1).
txt_doc(962,3,3).
txt_doc(551,3,1).
txt_doc(232,3,1).
txt_doc(704,3,1).
txt_doc(595,3,1).
txt_doc(822,3,1).
txt_doc(938,3,1).
txt_doc(249,3,1).
txt_doc(485,3,1).
txt_doc(659,3,1).
txt_doc(513,3,1).
txt_doc(439,3,1).
txt_doc(901,3,1).
txt_doc(337,3,1).
txt_doc(506,3,1).
txt_doc(617,3,2).
txt_doc(75,3,1).
txt_doc(908,3,1).
txt_doc(384,3,1).
txt_doc(713,3,2).
txt_doc(740,3,1).
txt_doc(769,3,1).
txt_doc(341,3,1).
txt_doc(293,3,1).
txt_doc(31,3,1).
txt_doc(651,3,1).
txt_doc(686,3,1).
txt_doc(275,3,1).
txt_doc(284,3,1).
txt_doc(468,3,1).
txt_doc(472,3,1).
txt_doc(419,3,1).
txt_doc(842,3,1).
txt_doc(254,3,1).
txt_doc(176,3,1).
txt_doc(502,3,1).
txt_doc(248,3,1).
txt_doc(600,3,2).
txt_doc(52,3,1).
txt_doc(954,3,1).
txt_doc(225,3,1).
txt_doc(512,3,1).
txt_doc(667,3,1).
txt_doc(951,3,2).
txt_doc(755,3,1).
txt_doc(142,3,1).
txt_doc(685,3,1).
txt_doc(14,3,1).
txt_doc(28,3,1).
txt_doc(554,3,1).
txt_doc(325,3,2).
txt_doc(441,3,1).
txt_doc(579,3,1).
txt_doc(210,3,1).
txt_doc(317,3,1).
txt_doc(929,3,1).
txt_doc(523,3,2).
txt_doc(959,3,1).
txt_doc(239,3,1).
txt_doc(123,3,1).
txt_doc(10,3,1).
txt_doc(466,3,1).
txt_doc(88,3,1).
txt_doc(102,3,2).
txt_doc(914,3,1).
txt_doc(712,3,1).
txt_doc(171,3,1).
txt_doc(421,3,1).
txt_doc(569,3,1).
txt_doc(681,3,1).
txt_doc(268,3,1).
txt_doc(543,3,1).
txt_doc(80,3,1).
txt_doc(812,3,1).
txt_doc(800,3,1).
txt_doc(939,3,1).
txt_doc(668,3,1).
txt_doc(380,3,1).
txt_doc(960,3,1).
txt_doc(860,3,1).
txt_doc(392,4,1).
txt_doc(628,4,1).
txt_doc(860,4,1).
txt_doc(209,4,1).
txt_doc(751,4,1).
txt_doc(914,4,1).
txt_doc(652,4,1).
txt_doc(861,4,1).
txt_doc(432,4,1).
txt_doc(130,4,1).
txt_doc(948,4,1).
txt_doc(445,4,1).
txt_doc(521,4,1).
txt_doc(363,4,1).
txt_doc(328,4,1).
txt_doc(100,4,1).
txt_doc(870,4,1).
txt_doc(953,4,1).
txt_doc(369,4,1).
txt_doc(472,4,1).
txt_doc(511,4,1).
txt_doc(827,4,1).
txt_doc(637,4,1).
txt_doc(515,4,1).
txt_doc(929,4,1).
txt_doc(112,4,1).
txt_doc(723,4,1).
txt_doc(354,4,1).
txt_doc(340,4,1).
txt_doc(671,4,1).
txt_doc(795,4,1).
txt_doc(845,4,1).
txt_doc(52,4,2).
txt_doc(912,4,1).
txt_doc(808,4,1).
txt_doc(662,4,1).
txt_doc(530,4,2).
txt_doc(146,4,1).
txt_doc(924,4,1).
txt_doc(841,4,2).
txt_doc(960,4,1).
txt_doc(844,4,1).
txt_doc(334,4,1).
txt_doc(95,4,1).
txt_doc(958,4,2).
txt_doc(50,4,1).
txt_doc(611,4,1).
txt_doc(779,4,1).
txt_doc(685,4,1).
txt_doc(748,4,1).
txt_doc(951,4,1).
txt_doc(692,4,1).
txt_doc(730,4,1).
txt_doc(402,4,1).
txt_doc(265,4,1).
txt_doc(665,4,1).
txt_doc(791,4,1).
txt_doc(129,4,2).
txt_doc(735,4,1).
txt_doc(28,4,1).
txt_doc(643,4,1).
txt_doc(428,4,1).
txt_doc(781,4,1).
txt_doc(448,4,1).
txt_doc(26,4,1).
txt_doc(661,4,1).
txt_doc(599,4,1).
txt_doc(355,4,1).
txt_doc(179,4,1).
txt_doc(335,4,1).
txt_doc(31,4,1).
txt_doc(85,4,1).
txt_doc(923,4,2).
txt_doc(352,4,2).
txt_doc(873,4,1).
txt_doc(716,4,1).
txt_doc(13,4,1).
txt_doc(206,4,1).
txt_doc(656,4,1).
txt_doc(784,4,1).
txt_doc(743,4,1).
txt_doc(390,4,2).
txt_doc(799,4,1).
txt_doc(373,4,1).
txt_doc(377,4,1).
txt_doc(633,4,1).
txt_doc(650,4,1).
txt_doc(451,4,1).
txt_doc(314,4,1).
txt_doc(792,4,2).
txt_doc(295,4,2).
txt_doc(19,4,1).
txt_doc(171,4,2).
txt_doc(341,4,1).
txt_doc(490,4,1).
txt_doc(955,4,1).
txt_doc(842,4,1).
txt_doc(857,4,2).
txt_doc(556,4,1).
txt_doc(846,4,1).
txt_doc(218,4,1).
txt_doc(715,4,2).
txt_doc(670,4,1).
txt_doc(642,4,2).
txt_doc(625,4,1).
txt_doc(554,4,2).
txt_doc(182,4,1).
txt_doc(881,4,2).
txt_doc(720,4,2).
txt_doc(463,4,1).
txt_doc(698,4,1).
txt_doc(805,4,1).
txt_doc(368,4,1).
txt_doc(649,4,1).
txt_doc(627,4,1).
txt_doc(934,4,1).
txt_doc(264,4,1).
txt_doc(574,4,2).
txt_doc(686,4,1).
txt_doc(154,4,1).
txt_doc(534,4,2).
txt_doc(824,4,1).
txt_doc(253,4,1).
txt_doc(339,4,2).
txt_doc(461,4,1).
txt_doc(561,4,1).
txt_doc(5,4,1).
txt_doc(604,4,1).
txt_doc(843,4,1).
txt_doc(210,4,1).
txt_doc(888,4,1).
txt_doc(380,4,1).
txt_doc(127,4,1).
txt_doc(9,4,1).
txt_doc(812,4,1).
txt_doc(164,4,2).
txt_doc(382,4,1).
txt_doc(871,4,1).
txt_doc(49,4,1).
txt_doc(205,4,1).
txt_doc(117,4,1).
txt_doc(565,4,2).
txt_doc(619,4,1).
txt_doc(403,4,1).
txt_doc(273,4,1).
txt_doc(174,4,2).
txt_doc(626,4,1).
txt_doc(58,4,1).
txt_doc(696,4,1).
txt_doc(502,4,1).
txt_doc(427,4,1).
txt_doc(965,4,1).
txt_doc(531,4,2).
txt_doc(125,4,1).
txt_doc(114,4,1).
txt_doc(737,4,1).
txt_doc(687,4,1).
txt_doc(430,4,2).
txt_doc(233,4,1).
txt_doc(695,4,2).
txt_doc(655,4,1).
txt_doc(481,4,1).
txt_doc(664,4,2).
txt_doc(51,4,2).
txt_doc(370,4,1).
txt_doc(399,4,1).
txt_doc(959,4,1).
txt_doc(484,4,1).
txt_doc(61,4,1).
txt_doc(711,4,1).
txt_doc(819,4,1).
txt_doc(203,4,1).
txt_doc(378,5,1).
txt_doc(130,5,1).
txt_doc(602,5,1).
txt_doc(681,5,1).
txt_doc(242,5,1).
txt_doc(115,5,1).
txt_doc(684,5,1).
txt_doc(904,5,1).
txt_doc(367,5,1).
txt_doc(141,5,1).
txt_doc(598,5,1).
txt_doc(150,5,1).
txt_doc(819,5,1).
txt_doc(376,5,1).
txt_doc(420,5,1).
txt_doc(32,5,1).
txt_doc(580,5,1).
txt_doc(772,5,1).
txt_doc(728,5,1).
txt_doc(248,5,1).
txt_doc(483,5,1).
txt_doc(780,5,1).
txt_doc(494,5,1).
txt_doc(3,5,1).
txt_doc(838,5,1).
txt_doc(751,5,1).
txt_doc(245,5,1).
txt_doc(309,5,1).
txt_doc(678,5,1).
txt_doc(764,5,1).
txt_doc(176,5,1).
txt_doc(496,5,1).
txt_doc(958,5,1).
txt_doc(885,5,1).
txt_doc(232,5,1).
txt_doc(334,5,1).
txt_doc(913,5,2).
txt_doc(340,5,1).
txt_doc(626,5,1).
txt_doc(854,5,1).
txt_doc(412,5,1).
txt_doc(591,5,1).
txt_doc(413,5,1).
txt_doc(180,5,1).
txt_doc(70,5,1).
txt_doc(461,5,1).
txt_doc(668,5,1).
txt_doc(578,5,1).
txt_doc(351,5,2).
txt_doc(816,5,1).
txt_doc(920,5,1).
txt_doc(791,5,1).
txt_doc(383,5,1).
txt_doc(42,5,1).
txt_doc(437,5,1).
txt_doc(643,5,1).
txt_doc(786,5,1).
txt_doc(94,5,1).
txt_doc(189,5,1).
txt_doc(610,5,1).
txt_doc(65,5,1).
txt_doc(170,5,1).
txt_doc(860,5,1).
txt_doc(670,5,1).
txt_doc(605,5,1).
txt_doc(659,5,2).
txt_doc(217,5,1).
txt_doc(908,5,1).
txt_doc(726,5,1).
txt_doc(286,5,2).
txt_doc(476,5,1).
txt_doc(635,5,1).
txt_doc(942,5,1).
txt_doc(907,5,1).
txt_doc(128,5,1).
txt_doc(565,5,1).
txt_doc(18,5,1).
txt_doc(601,5,1).
txt_doc(72,5,1).
txt_doc(654,5,1).
txt_doc(704,5,1).
txt_doc(750,5,1).
txt_doc(954,5,1).
txt_doc(89,5,1).
txt_doc(799,5,1).
txt_doc(555,5,2).
txt_doc(424,5,1).
txt_doc(447,5,2).
txt_doc(325,5,1).
txt_doc(202,5,1).
txt_doc(549,5,1).
txt_doc(432,5,1).
txt_doc(944,5,1).
txt_doc(251,5,1).
txt_doc(31,5,1).
txt_doc(522,5,1).
txt_doc(906,5,2).
txt_doc(450,5,1).
txt_doc(465,5,1).
txt_doc(746,5,1).
txt_doc(152,5,1).
txt_doc(327,5,1).
txt_doc(947,5,1).
txt_doc(162,5,1).
txt_doc(845,5,2).
txt_doc(489,5,1).
txt_doc(14,5,2).
txt_doc(656,5,1).
txt_doc(27,5,1).
txt_doc(743,5,1).
txt_doc(298,5,1).
txt_doc(852,5,1).
txt_doc(267,5,1).
txt_doc(623,5,1).
txt_doc(788,5,1).
txt_doc(216,5,1).
txt_doc(57,5,1).
txt_doc(497,5,1).
txt_doc(703,5,1).
txt_doc(928,5,2).
txt_doc(902,5,1).
txt_doc(472,5,1).
txt_doc(559,5,1).
txt_doc(614,5,1).
txt_doc(451,5,1).
txt_doc(86,5,1).
txt_doc(246,5,1).
txt_doc(148,5,1).
txt_doc(632,5,1).
txt_doc(540,5,1).
txt_doc(385,5,1).
txt_doc(319,5,2).
txt_doc(26,5,1).
txt_doc(946,5,2).
txt_doc(951,5,1).
txt_doc(793,5,1).
txt_doc(721,5,1).
txt_doc(171,5,1).
txt_doc(399,5,2).
txt_doc(132,5,1).
txt_doc(397,5,1).
txt_doc(331,5,1).
txt_doc(137,5,1).
txt_doc(768,5,1).
txt_doc(769,5,1).
txt_doc(342,5,1).
txt_doc(901,5,1).
txt_doc(444,5,1).
txt_doc(792,5,1).
txt_doc(820,5,1).
txt_doc(671,5,1).
txt_doc(317,5,1).
txt_doc(240,5,2).
txt_doc(929,5,1).
txt_doc(884,5,1).
txt_doc(117,5,2).
txt_doc(664,5,1).
txt_doc(50,5,1).
txt_doc(952,5,2).
txt_doc(753,5,1).
txt_doc(126,5,1).
txt_doc(389,5,1).
txt_doc(478,5,1).
txt_doc(477,5,1).
txt_doc(47,5,1).
txt_doc(950,5,2).
txt_doc(409,5,2).
txt_doc(411,5,1).
txt_doc(806,5,1).
txt_doc(777,5,1).
txt_doc(855,5,1).
txt_doc(809,5,1).
txt_doc(955,5,1).
txt_doc(488,5,1).
txt_doc(811,5,1).
txt_doc(430,5,1).
txt_doc(297,5,1).
txt_doc(485,5,1).
txt_doc(392,5,2).
txt_doc(151,5,1).
txt_doc(949,5,1).
txt_doc(133,6,1).
txt_doc(508,6,1).
txt_doc(602,6,1).
txt_doc(283,6,1).
txt_doc(419,6,1).
txt_doc(938,6,1).
txt_doc(164,6,1).
txt_doc(965,6,1).
txt_doc(549,6,1).
txt_doc(818,6,1).
txt_doc(692,6,1).
txt_doc(51,6,1).
txt_doc(720,6,1).
txt_doc(700,6,1).
txt_doc(456,6,1).
txt_doc(60,6,1).
txt_doc(212,6,1).
txt_doc(368,6,1).
txt_doc(263,6,1).
txt_doc(447,6,1).
txt_doc(538,6,1).
txt_doc(715,6,1).
txt_doc(600,6,1).
txt_doc(129,6,1).
txt_doc(648,6,1).
txt_doc(59,6,1).
txt_doc(327,6,1).
txt_doc(384,6,1).
txt_doc(506,6,1).
txt_doc(655,6,1).
txt_doc(70,6,1).
txt_doc(537,6,1).
txt_doc(678,6,1).
txt_doc(110,6,1).
txt_doc(201,6,1).
txt_doc(883,6,1).
txt_doc(631,6,1).
txt_doc(615,6,1).
txt_doc(869,6,1).
txt_doc(58,6,1).
txt_doc(234,6,1).
txt_doc(197,6,1).
txt_doc(194,6,1).
txt_doc(333,6,1).
txt_doc(425,6,1).
txt_doc(454,6,1).
txt_doc(459,6,1).
txt_doc(710,6,1).
txt_doc(511,6,1).
txt_doc(116,6,1).
txt_doc(809,6,1).
txt_doc(83,6,1).
txt_doc(249,6,1).
txt_doc(945,6,1).
txt_doc(939,6,1).
txt_doc(751,6,1).
txt_doc(31,6,1).
txt_doc(619,6,1).
txt_doc(295,6,1).
txt_doc(475,6,1).
txt_doc(20,6,1).
txt_doc(743,6,1).
txt_doc(688,6,1).
txt_doc(196,6,1).
txt_doc(529,6,1).
txt_doc(402,6,1).
txt_doc(69,6,1).
txt_doc(867,6,1).
txt_doc(381,6,1).
txt_doc(659,6,1).
txt_doc(440,6,1).
txt_doc(420,6,1).
txt_doc(242,6,1).
txt_doc(344,6,1).
txt_doc(608,6,1).
txt_doc(921,6,1).
txt_doc(503,6,1).
txt_doc(290,6,1).
txt_doc(39,6,1).
txt_doc(178,6,1).
txt_doc(131,6,1).
txt_doc(843,6,1).
txt_doc(163,6,1).
txt_doc(785,6,1).
txt_doc(666,6,1).
txt_doc(72,6,1).
txt_doc(312,6,2).
txt_doc(680,6,1).
txt_doc(193,6,1).
txt_doc(345,6,1).
txt_doc(846,6,1).
txt_doc(288,6,1).
txt_doc(553,6,2).
txt_doc(811,6,1).
txt_doc(23,6,1).
txt_doc(213,6,1).
txt_doc(104,6,1).
txt_doc(609,6,1).
txt_doc(825,6,1).
txt_doc(940,6,1).
txt_doc(61,6,1).
txt_doc(253,6,1).
txt_doc(585,6,1).
txt_doc(671,6,1).
txt_doc(707,6,1).
txt_doc(78,6,1).
txt_doc(62,6,1).
txt_doc(86,6,1).
txt_doc(467,6,1).
txt_doc(620,6,1).
txt_doc(579,6,1).
txt_doc(93,6,1).
txt_doc(286,6,1).
txt_doc(349,6,1).
txt_doc(568,6,1).
txt_doc(583,6,1).
txt_doc(571,6,1).
txt_doc(857,6,1).
txt_doc(835,6,1).
txt_doc(943,6,1).
txt_doc(662,6,1).
txt_doc(612,6,1).
txt_doc(160,6,1).
txt_doc(833,6,1).
txt_doc(683,6,1).
txt_doc(355,6,2).
txt_doc(722,6,1).
txt_doc(762,6,1).
txt_doc(101,6,1).
txt_doc(299,6,1).
txt_doc(884,6,1).
txt_doc(885,6,1).
txt_doc(307,6,1).
txt_doc(192,6,1).
txt_doc(896,6,1).
txt_doc(713,6,1).
txt_doc(653,6,1).
txt_doc(665,6,2).
txt_doc(7,6,1).
txt_doc(407,6,2).
txt_doc(961,6,1).
txt_doc(493,6,1).
txt_doc(491,6,2).
txt_doc(770,6,1).
txt_doc(113,6,1).
txt_doc(930,6,1).
txt_doc(179,6,2).
txt_doc(947,6,1).
txt_doc(369,6,1).
txt_doc(380,6,2).
txt_doc(478,6,1).
txt_doc(904,6,1).
txt_doc(169,6,1).
txt_doc(353,6,1).
txt_doc(373,6,1).
txt_doc(13,6,1).
txt_doc(763,6,1).
txt_doc(399,6,1).
txt_doc(819,6,1).
txt_doc(823,6,1).
txt_doc(837,6,1).
txt_doc(487,6,1).
txt_doc(800,6,1).
txt_doc(332,6,1).
txt_doc(200,6,1).
txt_doc(302,6,1).
txt_doc(32,6,2).
txt_doc(531,6,1).
txt_doc(324,6,1).
txt_doc(723,6,2).
txt_doc(593,6,1).
txt_doc(687,6,1).
txt_doc(343,6,1).
txt_doc(214,6,1).
txt_doc(19,6,1).
txt_doc(767,6,1).
txt_doc(601,6,1).
txt_doc(203,6,1).
txt_doc(499,6,1).
txt_doc(484,6,1).
txt_doc(841,6,1).
txt_doc(582,6,1).
txt_doc(932,6,1).
txt_doc(826,6,1).
txt_doc(629,6,1).
txt_doc(374,6,1).
txt_doc(851,6,2).
txt_doc(898,6,1).
txt_doc(570,6,1).
txt_doc(550,7,1).
txt_doc(406,7,1).
txt_doc(35,7,1).
txt_doc(776,7,1).
txt_doc(193,7,1).
txt_doc(483,7,1).
txt_doc(69,7,1).
txt_doc(870,7,1).
txt_doc(391,7,1).
txt_doc(214,7,1).
txt_doc(553,7,1).
txt_doc(561,7,1).
txt_doc(712,7,1).
txt_doc(540,7,1).
txt_doc(234,7,1).
txt_doc(869,7,1).
txt_doc(735,7,1).
txt_doc(83,7,1).
txt_doc(775,7,1).
txt_doc(405,7,1).
txt_doc(76,7,1).
txt_doc(328,7,1).
txt_doc(475,7,1).
txt_doc(401,7,1).
txt_doc(814,7,1).
txt_doc(954,7,1).
txt_doc(473,7,1).
txt_doc(795,7,1).
txt_doc(503,7,1).
txt_doc(10,7,1).
txt_doc(908,7,1).
txt_doc(508,7,1).
txt_doc(785,7,1).
txt_doc(341,7,1).
txt_doc(210,7,1).
txt_doc(910,7,1).
txt_doc(538,7,1).
txt_doc(509,7,1).
txt_doc(180,7,1).
txt_doc(236,7,1).
txt_doc(936,7,1).
txt_doc(669,7,1).
txt_doc(128,7,1).
txt_doc(900,7,1).
txt_doc(823,7,1).
txt_doc(951,7,1).
txt_doc(353,7,1).
txt_doc(276,7,1).
txt_doc(423,7,2).
txt_doc(325,7,1).
txt_doc(655,7,1).
txt_doc(488,7,2).
txt_doc(705,7,1).
txt_doc(564,7,1).
txt_doc(525,7,1).
txt_doc(227,7,1).
txt_doc(851,7,1).
txt_doc(695,7,1).
txt_doc(168,7,1).
txt_doc(337,7,1).
txt_doc(318,7,1).
txt_doc(183,7,1).
txt_doc(850,7,1).
txt_doc(547,7,1).
txt_doc(745,7,1).
txt_doc(958,7,1).
txt_doc(693,7,1).
txt_doc(923,7,1).
txt_doc(39,7,1).
txt_doc(649,7,1).
txt_doc(630,7,1).
txt_doc(733,7,1).
txt_doc(161,7,1).
txt_doc(226,7,1).
txt_doc(786,7,1).
txt_doc(638,7,1).
txt_doc(309,7,1).
txt_doc(842,7,1).
txt_doc(539,7,2).
txt_doc(687,7,1).
txt_doc(696,7,1).
txt_doc(753,7,2).
txt_doc(613,7,1).
txt_doc(822,7,1).
txt_doc(759,7,1).
txt_doc(783,7,2).
txt_doc(162,7,2).
txt_doc(563,7,1).
txt_doc(266,7,1).
txt_doc(28,7,1).
txt_doc(467,7,1).
txt_doc(159,7,1).
txt_doc(333,7,1).
txt_doc(63,7,1).
txt_doc(546,7,1).
txt_doc(897,7,1).
txt_doc(944,7,1).
txt_doc(46,7,2).
txt_doc(852,7,1).
txt_doc(247,7,1).
txt_doc(489,7,1).
txt_doc(298,7,1).
txt_doc(313,7,1).
txt_doc(383,7,1).
txt_doc(184,7,2).
txt_doc(757,7,1).
txt_doc(198,7,2).
txt_doc(582,7,1).
txt_doc(361,7,2).
txt_doc(784,7,1).
txt_doc(19,7,1).
txt_doc(766,7,1).
txt_doc(369,7,2).
txt_doc(88,7,1).
txt_doc(194,7,1).
txt_doc(635,7,1).
txt_doc(771,7,1).
txt_doc(931,7,1).
txt_doc(594,7,1).
txt_doc(811,7,2).
txt_doc(719,7,1).
txt_doc(585,7,1).
txt_doc(928,7,2).
txt_doc(674,7,2).
txt_doc(479,7,1).
txt_doc(262,7,1).
txt_doc(947,7,1).
txt_doc(454,7,1).
txt_doc(175,7,1).
txt_doc(628,7,1).
txt_doc(165,7,1).
txt_doc(640,7,1).
txt_doc(314,7,1).
txt_doc(462,7,1).
txt_doc(220,7,1).
txt_doc(445,7,1).
txt_doc(229,7,2).
txt_doc(436,7,1).
txt_doc(886,7,1).
txt_doc(342,7,2).
txt_doc(486,7,1).
txt_doc(302,7,1).
txt_doc(20,7,1).
txt_doc(697,7,1).
txt_doc(64,7,1).
txt_doc(464,7,1).
txt_doc(862,7,2).
txt_doc(79,7,1).
txt_doc(505,7,2).
txt_doc(153,7,1).
txt_doc(744,7,1).
txt_doc(8,7,2).
txt_doc(58,7,1).
txt_doc(711,7,1).
txt_doc(686,7,1).
txt_doc(578,7,1).
txt_doc(477,7,1).
txt_doc(468,7,1).
txt_doc(87,7,1).
txt_doc(940,7,1).
txt_doc(389,7,2).
txt_doc(75,7,1).
txt_doc(84,7,1).
txt_doc(860,7,2).
txt_doc(191,7,1).
txt_doc(833,7,2).
txt_doc(414,7,1).
txt_doc(672,7,1).
txt_doc(48,7,1).
txt_doc(209,7,2).
txt_doc(438,7,1).
txt_doc(295,7,1).
txt_doc(187,7,1).
txt_doc(810,7,1).
txt_doc(67,7,1).
txt_doc(597,7,1).
txt_doc(166,7,1).
txt_doc(592,8,1).
txt_doc(597,8,1).
txt_doc(909,8,1).
txt_doc(48,8,1).
txt_doc(550,8,1).
txt_doc(594,8,1).
txt_doc(438,8,1).
txt_doc(344,8,1).
txt_doc(839,8,1).
txt_doc(179,8,1).
txt_doc(290,8,1).
txt_doc(812,8,1).
txt_doc(773,8,1).
txt_doc(873,8,1).
txt_doc(947,8,1).
txt_doc(500,8,1).
txt_doc(177,8,1).
txt_doc(725,8,1).
txt_doc(626,8,1).
txt_doc(455,8,1).
txt_doc(314,8,1).
txt_doc(273,8,1).
txt_doc(854,8,1).
txt_doc(871,8,1).
txt_doc(225,8,1).
txt_doc(25,8,1).
txt_doc(547,8,1).
txt_doc(343,8,1).
txt_doc(901,8,1).
txt_doc(716,8,1).
txt_doc(68,8,1).
txt_doc(517,8,1).
txt_doc(87,8,1).
txt_doc(919,8,1).
txt_doc(637,8,1).
txt_doc(180,8,1).
txt_doc(306,8,1).
txt_doc(833,8,1).
txt_doc(869,8,1).
txt_doc(954,8,1).
txt_doc(8,8,1).
txt_doc(936,8,1).
txt_doc(635,8,1).
txt_doc(644,8,1).
txt_doc(474,8,1).
txt_doc(642,8,1).
txt_doc(422,8,1).
txt_doc(816,8,1).
txt_doc(836,8,1).
txt_doc(921,8,1).
txt_doc(263,8,1).
txt_doc(1,8,1).
txt_doc(430,8,1).
txt_doc(515,8,1).
txt_doc(186,8,1).
txt_doc(88,8,1).
txt_doc(966,8,1).
txt_doc(315,8,1).
txt_doc(927,8,1).
txt_doc(879,8,1).
txt_doc(881,8,2).
txt_doc(891,8,1).
txt_doc(691,8,1).
txt_doc(390,8,1).
txt_doc(577,8,1).
txt_doc(153,8,1).
txt_doc(18,8,1).
txt_doc(805,8,1).
txt_doc(224,8,1).
txt_doc(696,8,1).
txt_doc(697,8,1).
txt_doc(6,8,1).
txt_doc(239,8,1).
txt_doc(461,8,2).
txt_doc(938,8,1).
txt_doc(650,8,1).
txt_doc(575,8,1).
txt_doc(529,8,1).
txt_doc(243,8,1).
txt_doc(70,8,1).
txt_doc(408,8,1).
txt_doc(192,8,1).
txt_doc(252,8,1).
txt_doc(937,8,1).
txt_doc(118,8,1).
txt_doc(646,8,1).
txt_doc(155,8,1).
txt_doc(363,8,1).
txt_doc(4,8,1).
txt_doc(457,8,1).
txt_doc(191,8,1).
txt_doc(745,8,1).
txt_doc(478,8,2).
txt_doc(328,8,1).
txt_doc(284,8,2).
txt_doc(253,8,1).
txt_doc(120,8,2).
txt_doc(916,8,1).
txt_doc(358,8,1).
txt_doc(949,8,1).
txt_doc(537,8,2).
txt_doc(832,8,1).
txt_doc(801,8,1).
txt_doc(167,8,1).
txt_doc(778,8,1).
txt_doc(331,8,2).
txt_doc(442,8,1).
txt_doc(606,8,1).
txt_doc(793,8,2).
txt_doc(684,8,1).
txt_doc(790,8,2).
txt_doc(535,8,2).
txt_doc(503,8,1).
txt_doc(913,8,1).
txt_doc(71,8,2).
txt_doc(700,8,2).
txt_doc(426,8,1).
txt_doc(893,8,1).
txt_doc(29,8,1).
txt_doc(163,8,1).
txt_doc(726,8,2).
txt_doc(673,8,1).
txt_doc(701,8,1).
txt_doc(183,8,1).
txt_doc(477,8,1).
txt_doc(233,8,2).
txt_doc(383,8,1).
txt_doc(862,8,2).
txt_doc(580,8,1).
txt_doc(92,8,1).
txt_doc(524,8,1).
txt_doc(299,8,1).
txt_doc(116,8,1).
txt_doc(853,8,1).
txt_doc(755,8,1).
txt_doc(666,8,1).
txt_doc(453,8,1).
txt_doc(193,8,1).
txt_doc(235,8,1).
txt_doc(136,8,1).
txt_doc(859,8,2).
txt_doc(439,8,2).
txt_doc(356,8,1).
txt_doc(654,8,2).
txt_doc(262,8,1).
txt_doc(797,8,1).
txt_doc(539,8,2).
txt_doc(456,8,1).
txt_doc(600,8,1).
txt_doc(17,8,2).
txt_doc(960,8,1).
txt_doc(381,8,2).
txt_doc(361,8,1).
txt_doc(200,8,1).
txt_doc(266,8,2).
txt_doc(27,8,1).
txt_doc(630,8,2).
txt_doc(643,8,1).
txt_doc(733,8,1).
txt_doc(12,8,1).
txt_doc(526,8,1).
txt_doc(57,8,1).
txt_doc(714,8,1).
txt_doc(307,8,1).
txt_doc(603,8,2).
txt_doc(143,8,1).
txt_doc(509,8,1).
txt_doc(402,8,1).
txt_doc(260,8,1).
txt_doc(382,8,1).
txt_doc(320,8,1).
txt_doc(877,8,2).
txt_doc(352,8,1).
txt_doc(429,8,1).
txt_doc(119,8,1).
txt_doc(492,9,1).
txt_doc(769,9,1).
txt_doc(134,9,1).
txt_doc(16,9,1).
txt_doc(217,9,1).
txt_doc(921,9,1).
txt_doc(461,9,1).
txt_doc(234,9,1).
txt_doc(491,9,1).
txt_doc(715,9,1).
txt_doc(966,9,1).
txt_doc(937,9,1).
txt_doc(104,9,1).
txt_doc(412,9,1).
txt_doc(113,9,1).
txt_doc(17,9,1).
txt_doc(508,9,1).
txt_doc(813,9,1).
txt_doc(792,9,1).
txt_doc(71,9,1).
txt_doc(584,9,1).
txt_doc(443,9,1).
txt_doc(444,9,1).
txt_doc(820,9,1).
txt_doc(349,9,1).
txt_doc(598,9,1).
txt_doc(771,9,1).
txt_doc(834,9,1).
txt_doc(91,9,1).
txt_doc(278,9,1).
txt_doc(542,9,1).
txt_doc(240,9,1).
txt_doc(332,9,1).
txt_doc(254,9,1).
txt_doc(124,9,1).
txt_doc(645,9,1).
txt_doc(838,9,2).
txt_doc(879,9,1).
txt_doc(718,9,1).
txt_doc(2,9,1).
txt_doc(539,9,1).
txt_doc(49,9,1).
txt_doc(137,9,1).
txt_doc(744,9,1).
txt_doc(886,9,1).
txt_doc(472,9,1).
txt_doc(949,9,1).
txt_doc(915,9,1).
txt_doc(805,9,1).
txt_doc(828,9,1).
txt_doc(562,9,1).
txt_doc(835,9,1).
txt_doc(323,9,1).
txt_doc(823,9,1).
txt_doc(63,9,1).
txt_doc(84,9,1).
txt_doc(678,9,1).
txt_doc(335,9,1).
txt_doc(527,9,1).
txt_doc(660,9,1).
txt_doc(305,9,1).
txt_doc(141,9,1).
txt_doc(609,9,1).
txt_doc(123,9,1).
txt_doc(285,9,1).
txt_doc(667,9,1).
txt_doc(544,9,1).
txt_doc(103,9,1).
txt_doc(721,9,1).
txt_doc(635,9,1).
txt_doc(680,9,1).
txt_doc(578,9,1).
txt_doc(442,9,1).
txt_doc(789,9,1).
txt_doc(765,9,1).
txt_doc(831,9,1).
txt_doc(467,9,1).
txt_doc(700,9,1).
txt_doc(523,9,1).
txt_doc(650,9,1).
txt_doc(475,9,1).
txt_doc(962,9,1).
txt_doc(556,9,1).
txt_doc(945,9,1).
txt_doc(62,9,1).
txt_doc(593,9,3).
txt_doc(767,9,1).
txt_doc(119,9,1).
txt_doc(761,9,1).
txt_doc(371,9,1).
txt_doc(614,9,1).
txt_doc(759,9,1).
txt_doc(331,9,1).
txt_doc(807,9,1).
txt_doc(845,9,1).
txt_doc(422,9,1).
txt_doc(822,9,1).
txt_doc(179,9,1).
txt_doc(344,9,1).
txt_doc(389,9,1).
txt_doc(814,9,2).
txt_doc(871,9,1).
txt_doc(165,9,1).
txt_doc(766,9,1).
txt_doc(659,9,1).
txt_doc(516,9,1).
txt_doc(697,9,1).
txt_doc(11,9,1).
txt_doc(589,9,1).
txt_doc(319,9,1).
txt_doc(526,9,1).
txt_doc(128,9,1).
txt_doc(317,9,1).
txt_doc(870,9,1).
txt_doc(568,9,2).
txt_doc(174,9,1).
txt_doc(528,9,1).
txt_doc(843,9,1).
txt_doc(751,9,1).
txt_doc(603,9,1).
txt_doc(190,9,1).
txt_doc(877,9,1).
txt_doc(935,9,1).
txt_doc(488,9,1).
txt_doc(564,9,1).
txt_doc(171,9,1).
txt_doc(686,9,1).
txt_doc(720,9,1).
txt_doc(356,9,1).
txt_doc(847,9,1).
txt_doc(334,9,1).
txt_doc(926,9,1).
txt_doc(502,9,1).
txt_doc(702,9,1).
txt_doc(83,9,1).
txt_doc(146,9,1).
txt_doc(250,9,1).
txt_doc(683,9,1).
txt_doc(909,9,1).
txt_doc(677,9,1).
txt_doc(216,9,1).
txt_doc(261,9,1).
txt_doc(365,9,1).
txt_doc(850,9,1).
txt_doc(531,9,1).
txt_doc(129,9,1).
txt_doc(232,9,1).
txt_doc(27,9,1).
txt_doc(939,9,1).
txt_doc(694,9,1).
txt_doc(634,9,1).
txt_doc(518,9,1).
txt_doc(837,9,1).
txt_doc(264,9,1).
txt_doc(204,9,1).
txt_doc(370,9,2).
txt_doc(101,9,1).
txt_doc(514,9,1).
txt_doc(608,9,1).
txt_doc(301,9,1).
txt_doc(682,9,1).
txt_doc(29,9,1).
txt_doc(89,9,1).
txt_doc(760,9,2).
txt_doc(168,9,1).
txt_doc(420,9,1).
txt_doc(749,9,1).
txt_doc(873,9,2).
txt_doc(898,9,1).
txt_doc(628,9,1).
txt_doc(78,9,1).
txt_doc(465,9,1).
txt_doc(644,9,1).
txt_doc(175,9,1).
txt_doc(671,9,1).
txt_doc(387,9,1).
txt_doc(470,9,1).
txt_doc(320,9,1).
txt_doc(376,9,1).
txt_doc(228,9,1).
txt_doc(448,9,1).
txt_doc(673,9,1).
txt_doc(166,9,2).
txt_doc(757,9,1).
txt_doc(941,9,1).
txt_doc(353,9,2).
txt_doc(32,9,2).
txt_doc(521,9,1).
txt_doc(196,9,1).
txt_doc(642,10,1).
txt_doc(200,10,1).
txt_doc(849,10,1).
txt_doc(35,10,1).
txt_doc(664,10,1).
txt_doc(744,10,1).
txt_doc(7,10,1).
txt_doc(578,10,1).
txt_doc(570,10,1).
txt_doc(76,10,1).
txt_doc(402,10,1).
txt_doc(256,10,1).
txt_doc(308,10,1).
txt_doc(540,10,1).
txt_doc(968,10,1).
txt_doc(923,10,1).
txt_doc(234,10,1).
txt_doc(71,10,1).
txt_doc(158,10,1).
txt_doc(544,10,1).
txt_doc(948,10,2).
txt_doc(487,10,1).
txt_doc(224,10,1).
txt_doc(50,10,1).
txt_doc(498,10,1).
txt_doc(681,10,1).
txt_doc(827,10,1).
txt_doc(170,10,1).
txt_doc(418,10,1).
txt_doc(465,10,1).
txt_doc(299,10,1).
txt_doc(400,10,1).
txt_doc(144,10,1).
txt_doc(745,10,1).
txt_doc(497,10,1).
txt_doc(942,10,1).
txt_doc(385,10,1).
txt_doc(852,10,2).
txt_doc(201,10,1).
txt_doc(553,10,1).
txt_doc(34,10,1).
txt_doc(383,10,1).
txt_doc(823,10,1).
txt_doc(183,10,1).
txt_doc(133,10,1).
txt_doc(764,10,1).
txt_doc(543,10,2).
txt_doc(949,10,1).
txt_doc(349,10,1).
txt_doc(787,10,1).
txt_doc(837,10,1).
txt_doc(870,10,1).
txt_doc(894,10,1).
txt_doc(275,10,1).
txt_doc(240,10,1).
txt_doc(441,10,1).
txt_doc(759,10,1).
txt_doc(792,10,1).
txt_doc(138,10,1).
txt_doc(656,10,2).
txt_doc(29,10,1).
txt_doc(318,10,1).
txt_doc(237,10,1).
txt_doc(692,10,2).
txt_doc(668,10,1).
txt_doc(6,10,1).
txt_doc(241,10,1).
txt_doc(528,10,1).
txt_doc(716,10,1).
txt_doc(763,10,1).
txt_doc(867,10,1).
txt_doc(938,10,1).
txt_doc(917,10,1).
txt_doc(770,10,2).
txt_doc(311,10,1).
txt_doc(280,10,1).
txt_doc(620,10,1).
txt_doc(175,10,1).
txt_doc(875,10,1).
txt_doc(590,10,1).
txt_doc(891,10,1).
txt_doc(679,10,1).
txt_doc(897,10,1).
txt_doc(61,10,1).
txt_doc(63,10,1).
txt_doc(724,10,1).
txt_doc(678,10,1).
txt_doc(5,10,1).
txt_doc(285,10,2).
txt_doc(239,10,2).
txt_doc(47,10,1).
txt_doc(612,10,1).
txt_doc(905,10,1).
txt_doc(261,10,1).
txt_doc(845,10,1).
txt_doc(178,10,1).
txt_doc(530,10,2).
txt_doc(463,10,1).
txt_doc(435,10,1).
txt_doc(276,10,1).
txt_doc(786,10,1).
txt_doc(278,10,1).
txt_doc(532,10,1).
txt_doc(569,10,1).
txt_doc(79,10,1).
txt_doc(721,10,1).
txt_doc(370,10,1).
txt_doc(912,10,1).
txt_doc(929,10,2).
txt_doc(75,10,1).
txt_doc(861,10,1).
txt_doc(470,10,1).
txt_doc(182,10,2).
txt_doc(736,10,1).
txt_doc(760,10,1).
txt_doc(265,10,1).
txt_doc(44,10,2).
txt_doc(397,10,1).
txt_doc(749,10,2).
txt_doc(287,10,1).
txt_doc(101,10,1).
txt_doc(148,10,1).
txt_doc(281,10,2).
txt_doc(702,10,1).
txt_doc(259,10,1).
txt_doc(580,10,1).
txt_doc(469,10,1).
txt_doc(772,10,1).
txt_doc(113,10,1).
txt_doc(284,10,1).
txt_doc(406,10,1).
txt_doc(598,10,1).
txt_doc(693,10,1).
txt_doc(806,10,2).
txt_doc(458,10,2).
txt_doc(139,10,2).
txt_doc(910,10,1).
txt_doc(292,10,1).
txt_doc(731,10,2).
txt_doc(560,10,1).
txt_doc(631,10,2).
txt_doc(266,10,1).
txt_doc(501,10,1).
txt_doc(667,10,1).
txt_doc(691,10,1).
txt_doc(665,10,1).
txt_doc(866,10,1).
txt_doc(769,10,1).
txt_doc(630,10,2).
txt_doc(636,10,1).
txt_doc(46,10,1).
txt_doc(512,10,1).
txt_doc(958,10,1).
txt_doc(611,10,1).
txt_doc(453,10,1).
txt_doc(427,10,1).
txt_doc(152,10,1).
txt_doc(510,10,2).
txt_doc(783,10,2).
txt_doc(538,10,1).
txt_doc(720,10,1).
txt_doc(181,10,1).
txt_doc(217,10,1).
txt_doc(735,10,1).
txt_doc(382,10,1).
txt_doc(835,10,1).
txt_doc(508,10,1).
txt_doc(885,10,1).
txt_doc(135,10,1).
txt_doc(466,10,1).
txt_doc(753,10,2).
txt_doc(584,10,2).
txt_doc(110,10,2).
txt_doc(677,10,1).
txt_doc(301,10,1).
txt_doc(210,11,1).
txt_doc(233,11,1).
txt_doc(162,11,1).
txt_doc(902,11,1).
txt_doc(254,11,1).
txt_doc(533,11,1).
txt_doc(305,11,1).
txt_doc(245,11,1).
txt_doc(142,11,1).
txt_doc(849,11,1).
txt_doc(95,11,1).
txt_doc(713,11,1).
txt_doc(224,11,1).
txt_doc(636,11,1).
txt_doc(596,11,1).
txt_doc(752,11,1).
txt_doc(221,11,1).
txt_doc(398,11,1).
txt_doc(688,11,1).
txt_doc(278,11,1).
txt_doc(338,11,1).
txt_doc(119,11,1).
txt_doc(729,11,1).
txt_doc(592,11,1).
txt_doc(62,11,1).
txt_doc(4,11,1).
txt_doc(888,11,1).
txt_doc(181,11,1).
txt_doc(63,11,1).
txt_doc(479,11,2).
txt_doc(439,11,1).
txt_doc(229,11,1).
txt_doc(235,11,1).
txt_doc(720,11,1).
txt_doc(264,11,1).
txt_doc(855,11,1).
txt_doc(966,11,1).
txt_doc(554,11,1).
txt_doc(626,11,1).
txt_doc(478,11,1).
txt_doc(649,11,1).
txt_doc(274,11,1).
txt_doc(933,11,1).
txt_doc(792,11,1).
txt_doc(251,11,1).
txt_doc(524,11,1).
txt_doc(428,11,1).
txt_doc(347,11,1).
txt_doc(23,11,1).
txt_doc(850,11,1).
txt_doc(225,11,1).
txt_doc(140,11,1).
txt_doc(282,11,1).
txt_doc(247,11,1).
txt_doc(354,11,1).
txt_doc(318,11,1).
txt_doc(369,11,1).
txt_doc(856,11,1).
txt_doc(31,11,1).
txt_doc(822,11,1).
txt_doc(804,11,1).
txt_doc(594,11,1).
txt_doc(499,11,1).
txt_doc(419,11,1).
txt_doc(652,11,1).
txt_doc(150,11,1).
txt_doc(90,11,1).
txt_doc(132,11,1).
txt_doc(790,11,1).
txt_doc(407,11,1).
txt_doc(577,11,1).
txt_doc(654,11,1).
txt_doc(682,11,1).
txt_doc(410,11,1).
txt_doc(346,11,2).
txt_doc(896,11,1).
txt_doc(474,11,2).
txt_doc(757,11,1).
txt_doc(639,11,2).
txt_doc(460,11,1).
txt_doc(384,11,1).
txt_doc(879,11,1).
txt_doc(965,11,2).
txt_doc(947,11,1).
txt_doc(372,11,1).
txt_doc(281,11,1).
txt_doc(310,11,2).
txt_doc(65,11,1).
txt_doc(101,11,1).
txt_doc(504,11,2).
txt_doc(547,11,1).
txt_doc(775,11,1).
txt_doc(93,11,1).
txt_doc(810,11,1).
txt_doc(433,11,1).
txt_doc(425,11,1).
txt_doc(660,11,1).
txt_doc(925,11,1).
txt_doc(51,11,1).
txt_doc(868,11,2).
txt_doc(216,11,1).
txt_doc(878,11,2).
txt_doc(593,11,1).
txt_doc(138,11,1).
txt_doc(846,11,1).
txt_doc(584,11,1).
txt_doc(730,11,1).
txt_doc(397,11,1).
txt_doc(276,11,1).
txt_doc(669,11,1).
txt_doc(467,11,1).
txt_doc(837,11,1).
txt_doc(611,11,1).
txt_doc(825,11,1).
txt_doc(827,11,1).
txt_doc(212,11,1).
txt_doc(839,11,1).
txt_doc(723,11,1).
txt_doc(601,11,1).
txt_doc(21,11,1).
txt_doc(267,11,1).
txt_doc(368,11,1).
txt_doc(948,11,1).
txt_doc(155,11,1).
txt_doc(326,11,1).
txt_doc(171,11,1).
txt_doc(316,11,1).
txt_doc(441,11,1).
txt_doc(964,11,1).
txt_doc(631,11,1).
txt_doc(14,11,1).
txt_doc(258,11,1).
txt_doc(111,11,1).
txt_doc(39,11,1).
txt_doc(415,11,1).
txt_doc(309,11,1).
txt_doc(246,11,1).
txt_doc(423,11,1).
txt_doc(828,11,1).
txt_doc(700,11,1).
txt_doc(861,11,1).
txt_doc(22,11,1).
txt_doc(27,11,1).
txt_doc(402,11,2).
txt_doc(664,11,1).
txt_doc(464,11,1).
txt_doc(734,11,1).
txt_doc(333,11,1).
txt_doc(785,11,1).
txt_doc(795,11,2).
txt_doc(103,11,1).
txt_doc(367,11,1).
txt_doc(41,11,1).
txt_doc(79,11,1).
txt_doc(847,11,1).
txt_doc(642,11,1).
txt_doc(148,11,1).
txt_doc(736,11,2).
txt_doc(280,11,1).
txt_doc(475,11,1).
txt_doc(543,11,1).
txt_doc(514,11,1).
txt_doc(920,11,1).
txt_doc(359,11,1).
txt_doc(134,11,1).
txt_doc(516,11,2).
txt_doc(858,11,1).
txt_doc(56,11,3).
txt_doc(756,11,1).
txt_doc(438,11,2).
txt_doc(133,11,2).
txt_doc(937,11,1).
txt_doc(12,11,2).
txt_doc(733,11,1).
txt_doc(710,11,1).
txt_doc(738,11,2).
txt_doc(813,11,3).
txt_doc(306,11,1).
txt_doc(61,11,1).
txt_doc(335,12,1).
txt_doc(473,12,1).
txt_doc(820,12,1).
txt_doc(751,12,1).
txt_doc(442,12,1).
txt_doc(310,12,1).
txt_doc(635,12,1).
txt_doc(842,12,1).
txt_doc(213,12,1).
txt_doc(161,12,1).
txt_doc(15,12,1).
txt_doc(266,12,1).
txt_doc(245,12,1).
txt_doc(928,12,1).
txt_doc(844,12,1).
txt_doc(346,12,1).
txt_doc(596,12,1).
txt_doc(615,12,1).
txt_doc(361,12,1).
txt_doc(306,12,1).
txt_doc(443,12,1).
txt_doc(405,12,1).
txt_doc(431,12,1).
txt_doc(228,12,1).
txt_doc(460,12,1).
txt_doc(871,12,1).
txt_doc(407,12,1).
txt_doc(197,12,1).
txt_doc(409,12,1).
txt_doc(881,12,1).
txt_doc(428,12,1).
txt_doc(384,12,1).
txt_doc(385,12,1).
txt_doc(835,12,1).
txt_doc(792,12,1).
txt_doc(424,12,1).
txt_doc(693,12,1).
txt_doc(811,12,1).
txt_doc(362,12,1).
txt_doc(550,12,1).
txt_doc(140,12,1).
txt_doc(99,12,1).
txt_doc(886,12,1).
txt_doc(927,12,1).
txt_doc(162,12,1).
txt_doc(885,12,1).
txt_doc(952,12,2).
txt_doc(675,12,1).
txt_doc(265,12,1).
txt_doc(823,12,1).
txt_doc(902,12,1).
txt_doc(93,12,1).
txt_doc(620,12,1).
txt_doc(482,12,1).
txt_doc(738,12,1).
txt_doc(658,12,1).
txt_doc(626,12,1).
txt_doc(218,12,1).
txt_doc(890,12,1).
txt_doc(692,12,1).
txt_doc(795,12,1).
txt_doc(350,12,1).
txt_doc(554,12,1).
txt_doc(593,12,1).
txt_doc(659,12,1).
txt_doc(540,12,1).
txt_doc(926,12,1).
txt_doc(913,12,1).
txt_doc(770,12,1).
txt_doc(544,12,1).
txt_doc(651,12,1).
txt_doc(441,12,1).
txt_doc(177,12,2).
txt_doc(548,12,1).
txt_doc(778,12,1).
txt_doc(855,12,1).
txt_doc(906,12,1).
txt_doc(538,12,1).
txt_doc(52,12,1).
txt_doc(88,12,1).
txt_doc(965,12,1).
txt_doc(227,12,1).
txt_doc(536,12,1).
txt_doc(281,12,1).
txt_doc(152,12,1).
txt_doc(682,12,1).
txt_doc(611,12,3).
txt_doc(33,12,1).
txt_doc(119,12,1).
txt_doc(874,12,1).
txt_doc(481,12,1).
txt_doc(426,12,1).
txt_doc(315,12,1).
txt_doc(469,12,1).
txt_doc(400,12,1).
txt_doc(891,12,1).
txt_doc(314,12,2).
txt_doc(628,12,1).
txt_doc(551,12,1).
txt_doc(359,12,1).
txt_doc(252,12,1).
txt_doc(94,12,1).
txt_doc(171,12,1).
txt_doc(680,12,1).
txt_doc(645,12,1).
txt_doc(150,12,1).
txt_doc(910,12,2).
txt_doc(286,12,2).
txt_doc(12,12,1).
txt_doc(366,12,2).
txt_doc(230,12,2).
txt_doc(209,12,1).
txt_doc(523,12,1).
txt_doc(869,12,2).
txt_doc(240,12,1).
txt_doc(776,12,1).
txt_doc(916,12,2).
txt_doc(352,12,1).
txt_doc(321,12,2).
txt_doc(595,12,1).
txt_doc(736,12,1).
txt_doc(507,12,1).
txt_doc(16,12,1).
txt_doc(858,12,1).
txt_doc(261,12,1).
txt_doc(396,12,1).
txt_doc(483,12,1).
txt_doc(81,12,2).
txt_doc(941,12,1).
txt_doc(918,12,1).
txt_doc(494,12,1).
txt_doc(568,12,1).
txt_doc(63,12,3).
txt_doc(77,12,1).
txt_doc(601,12,1).
txt_doc(934,12,2).
txt_doc(60,12,2).
txt_doc(782,12,1).
txt_doc(163,12,1).
txt_doc(956,12,1).
txt_doc(297,12,1).
txt_doc(249,12,1).
txt_doc(145,12,1).
txt_doc(31,12,1).
txt_doc(236,12,1).
txt_doc(257,12,1).
txt_doc(716,12,1).
txt_doc(753,12,1).
txt_doc(813,12,1).
txt_doc(728,12,1).
txt_doc(23,12,1).
txt_doc(410,12,1).
txt_doc(415,12,1).
txt_doc(744,12,1).
txt_doc(26,12,1).
txt_doc(69,12,1).
txt_doc(485,12,1).
txt_doc(386,12,1).
txt_doc(803,12,1).
txt_doc(528,12,1).
txt_doc(30,12,1).
txt_doc(117,12,1).
txt_doc(276,12,1).
txt_doc(698,12,1).
txt_doc(848,12,1).
txt_doc(141,12,1).
txt_doc(450,12,1).
txt_doc(369,12,1).
txt_doc(549,12,1).
txt_doc(828,12,1).
txt_doc(103,12,2).
txt_doc(689,12,1).
txt_doc(539,12,1).
txt_doc(517,12,1).
txt_doc(123,12,1).
txt_doc(454,12,2).
txt_doc(148,12,1).
txt_doc(372,12,1).
txt_doc(129,12,1).
txt_doc(702,12,1).
txt_doc(115,12,1).
txt_doc(342,13,1).
txt_doc(197,13,1).
txt_doc(271,13,1).
txt_doc(774,13,1).
txt_doc(250,13,1).
txt_doc(625,13,1).
txt_doc(468,13,1).
txt_doc(920,13,1).
txt_doc(716,13,1).
txt_doc(798,13,1).
txt_doc(963,13,1).
txt_doc(526,13,1).
txt_doc(71,13,1).
txt_doc(579,13,1).
txt_doc(731,13,1).
txt_doc(810,13,1).
txt_doc(122,13,1).
txt_doc(192,13,1).
txt_doc(77,13,1).
txt_doc(771,13,1).
txt_doc(463,13,1).
txt_doc(600,13,1).
txt_doc(103,13,1).
txt_doc(203,13,1).
txt_doc(878,13,1).
txt_doc(780,13,1).
txt_doc(614,13,1).
txt_doc(866,13,1).
txt_doc(599,13,1).
txt_doc(20,13,1).
txt_doc(330,13,1).
txt_doc(772,13,1).
txt_doc(961,13,1).
txt_doc(79,13,1).
txt_doc(404,13,1).
txt_doc(721,13,1).
txt_doc(868,13,1).
txt_doc(722,13,1).
txt_doc(317,13,1).
txt_doc(399,13,1).
txt_doc(152,13,1).
txt_doc(932,13,1).
txt_doc(943,13,1).
txt_doc(667,13,1).
txt_doc(412,13,1).
txt_doc(876,13,1).
txt_doc(416,13,1).
txt_doc(672,13,1).
txt_doc(255,13,1).
txt_doc(613,13,2).
txt_doc(839,13,1).
txt_doc(728,13,1).
txt_doc(65,13,1).
txt_doc(119,13,1).
txt_doc(953,13,2).
txt_doc(153,13,1).
txt_doc(843,13,1).
txt_doc(760,13,1).
txt_doc(165,13,1).
txt_doc(189,13,2).
txt_doc(887,13,2).
txt_doc(755,13,1).
txt_doc(682,13,1).
txt_doc(316,13,1).
txt_doc(757,13,1).
txt_doc(43,13,1).
txt_doc(295,13,1).
txt_doc(383,13,1).
txt_doc(139,13,1).
txt_doc(784,13,1).
txt_doc(501,13,1).
txt_doc(217,13,1).
txt_doc(170,13,1).
txt_doc(183,13,2).
txt_doc(61,13,1).
txt_doc(597,13,1).
txt_doc(556,13,1).
txt_doc(577,13,1).
txt_doc(389,13,1).
txt_doc(497,13,1).
txt_doc(209,13,1).
txt_doc(671,13,2).
txt_doc(90,13,1).
txt_doc(946,13,1).
txt_doc(286,13,1).
txt_doc(235,13,1).
txt_doc(944,13,1).
txt_doc(368,13,2).
txt_doc(586,13,1).
txt_doc(391,13,1).
txt_doc(926,13,1).
txt_doc(915,13,1).
txt_doc(437,13,1).
txt_doc(696,13,1).
txt_doc(500,13,1).
txt_doc(327,13,1).
txt_doc(660,13,1).
txt_doc(598,13,1).
txt_doc(329,13,1).
txt_doc(458,13,1).
txt_doc(256,13,1).
txt_doc(142,13,1).
txt_doc(245,13,1).
txt_doc(841,13,1).
txt_doc(792,13,1).
txt_doc(199,13,1).
txt_doc(325,13,1).
txt_doc(525,13,1).
txt_doc(958,13,1).
txt_doc(326,13,1).
txt_doc(22,13,1).
txt_doc(659,13,1).
txt_doc(492,13,1).
txt_doc(726,13,1).
txt_doc(480,13,3).
txt_doc(606,13,1).
txt_doc(924,13,1).
txt_doc(104,13,2).
txt_doc(380,13,1).
txt_doc(113,13,2).
txt_doc(732,13,2).
txt_doc(27,13,1).
txt_doc(336,13,1).
txt_doc(781,13,1).
txt_doc(460,13,1).
txt_doc(32,13,1).
txt_doc(553,13,1).
txt_doc(47,13,1).
txt_doc(570,13,1).
txt_doc(155,13,1).
txt_doc(636,13,1).
txt_doc(775,13,1).
txt_doc(510,13,2).
txt_doc(578,13,1).
txt_doc(952,13,1).
txt_doc(238,13,1).
txt_doc(688,13,1).
txt_doc(703,13,1).
txt_doc(369,13,1).
txt_doc(366,13,1).
txt_doc(560,13,1).
txt_doc(185,13,1).
txt_doc(296,13,1).
txt_doc(52,13,1).
txt_doc(705,13,2).
txt_doc(261,13,2).
txt_doc(265,13,1).
txt_doc(127,13,1).
txt_doc(99,13,1).
txt_doc(551,13,2).
txt_doc(289,13,1).
txt_doc(471,13,1).
txt_doc(761,13,1).
txt_doc(44,13,1).
txt_doc(867,13,1).
txt_doc(479,13,1).
txt_doc(349,13,1).
txt_doc(591,13,1).
txt_doc(698,13,1).
txt_doc(718,13,1).
txt_doc(442,13,1).
txt_doc(232,13,1).
txt_doc(206,13,1).
txt_doc(38,13,1).
txt_doc(131,13,1).
txt_doc(902,13,1).
txt_doc(892,13,1).
txt_doc(673,13,1).
txt_doc(178,13,3).
txt_doc(112,13,1).
txt_doc(290,13,1).
txt_doc(293,13,1).
txt_doc(233,13,2).
txt_doc(957,13,1).
txt_doc(193,13,1).
txt_doc(151,13,1).
txt_doc(603,13,1).
txt_doc(710,13,1).
txt_doc(669,13,1).
txt_doc(845,13,1).
txt_doc(770,13,1).
txt_doc(727,14,1).
txt_doc(513,14,1).
txt_doc(741,14,1).
txt_doc(895,14,1).
txt_doc(780,14,1).
txt_doc(934,14,1).
txt_doc(961,14,1).
txt_doc(279,14,1).
txt_doc(71,14,1).
txt_doc(197,14,1).
txt_doc(437,14,1).
txt_doc(505,14,1).
txt_doc(896,14,1).
txt_doc(468,14,1).
txt_doc(646,14,1).
txt_doc(94,14,1).
txt_doc(545,14,1).
txt_doc(157,14,1).
txt_doc(835,14,1).
txt_doc(665,14,1).
txt_doc(103,14,1).
txt_doc(817,14,1).
txt_doc(326,14,1).
txt_doc(534,14,1).
txt_doc(735,14,1).
txt_doc(445,14,1).
txt_doc(40,14,1).
txt_doc(26,14,1).
txt_doc(160,14,1).
txt_doc(32,14,1).
txt_doc(953,14,1).
txt_doc(235,14,1).
txt_doc(112,14,1).
txt_doc(642,14,1).
txt_doc(460,14,1).
txt_doc(852,14,1).
txt_doc(346,14,1).
txt_doc(159,14,1).
txt_doc(338,14,1).
txt_doc(410,14,1).
txt_doc(703,14,1).
txt_doc(688,14,1).
txt_doc(461,14,1).
txt_doc(871,14,2).
txt_doc(109,14,1).
txt_doc(91,14,1).
txt_doc(694,14,1).
txt_doc(739,14,1).
txt_doc(902,14,1).
txt_doc(881,14,1).
txt_doc(420,14,1).
txt_doc(278,14,1).
txt_doc(324,14,1).
txt_doc(162,14,1).
txt_doc(883,14,1).
txt_doc(333,14,1).
txt_doc(318,14,1).
txt_doc(270,14,1).
txt_doc(280,14,1).
txt_doc(853,14,1).
txt_doc(792,14,1).
txt_doc(876,14,1).
txt_doc(731,14,1).
txt_doc(949,14,1).
txt_doc(350,14,1).
txt_doc(810,14,2).
txt_doc(747,14,1).
txt_doc(207,14,2).
txt_doc(288,14,1).
txt_doc(849,14,1).
txt_doc(631,14,1).
txt_doc(50,14,1).
txt_doc(698,14,1).
txt_doc(304,14,1).
txt_doc(59,14,1).
txt_doc(734,14,1).
txt_doc(947,14,1).
txt_doc(667,14,2).
txt_doc(35,14,1).
txt_doc(564,14,1).
txt_doc(256,14,1).
txt_doc(245,14,1).
txt_doc(914,14,2).
txt_doc(41,14,1).
txt_doc(706,14,1).
txt_doc(6,14,1).
txt_doc(960,14,1).
txt_doc(126,14,1).
txt_doc(784,14,1).
txt_doc(728,14,1).
txt_doc(521,14,1).
txt_doc(370,14,1).
txt_doc(862,14,1).
txt_doc(764,14,1).
txt_doc(141,14,1).
txt_doc(431,14,1).
txt_doc(785,14,1).
txt_doc(670,14,1).
txt_doc(791,14,1).
txt_doc(773,14,1).
txt_doc(907,14,1).
txt_doc(832,14,1).
txt_doc(426,14,1).
txt_doc(562,14,1).
txt_doc(415,14,1).
txt_doc(926,14,1).
txt_doc(367,14,1).
txt_doc(487,14,1).
txt_doc(858,14,1).
txt_doc(340,14,1).
txt_doc(276,14,1).
txt_doc(153,14,1).
txt_doc(185,14,1).
txt_doc(546,14,1).
txt_doc(119,14,1).
txt_doc(560,14,1).
txt_doc(720,14,1).
txt_doc(782,14,1).
txt_doc(578,14,1).
txt_doc(842,14,1).
txt_doc(193,14,1).
txt_doc(668,14,1).
txt_doc(514,14,2).
txt_doc(879,14,1).
txt_doc(593,14,1).
txt_doc(527,14,1).
txt_doc(78,14,1).
txt_doc(603,14,1).
txt_doc(258,14,2).
txt_doc(432,14,1).
txt_doc(213,14,1).
txt_doc(3,14,1).
txt_doc(559,14,1).
txt_doc(627,14,2).
txt_doc(915,14,1).
txt_doc(898,14,1).
txt_doc(345,14,1).
txt_doc(689,14,2).
txt_doc(830,14,1).
txt_doc(165,14,1).
txt_doc(413,14,1).
txt_doc(692,14,1).
txt_doc(27,14,1).
txt_doc(384,14,1).
txt_doc(449,14,1).
txt_doc(820,14,1).
txt_doc(746,14,1).
txt_doc(657,14,1).
txt_doc(774,14,1).
txt_doc(793,14,1).
txt_doc(243,14,1).
txt_doc(783,14,1).
txt_doc(821,14,1).
txt_doc(831,14,1).
txt_doc(814,14,1).
txt_doc(341,14,1).
txt_doc(869,14,2).
txt_doc(529,14,1).
txt_doc(137,14,2).
txt_doc(425,14,1).
txt_doc(294,14,1).
txt_doc(457,14,1).
txt_doc(412,14,1).
txt_doc(857,14,1).
txt_doc(632,14,1).
txt_doc(361,14,1).
txt_doc(597,14,1).
txt_doc(798,14,1).
txt_doc(736,14,1).
txt_doc(539,14,1).
txt_doc(816,14,1).
txt_doc(198,14,1).
txt_doc(429,14,2).
txt_doc(891,14,3).
txt_doc(312,14,2).
txt_doc(493,14,1).
txt_doc(212,14,2).
txt_doc(874,14,1).
txt_doc(194,14,1).
txt_doc(392,14,1).
txt_doc(693,14,1).
txt_doc(649,14,1).
txt_doc(888,14,2).
txt_doc(870,15,1).
txt_doc(145,15,1).
txt_doc(700,15,1).
txt_doc(459,15,1).
txt_doc(202,15,1).
txt_doc(780,15,1).
txt_doc(647,15,1).
txt_doc(504,15,1).
txt_doc(360,15,1).
txt_doc(110,15,1).
txt_doc(610,15,1).
txt_doc(937,15,1).
txt_doc(178,15,1).
txt_doc(917,15,1).
txt_doc(243,15,1).
txt_doc(122,15,1).
txt_doc(731,15,1).
txt_doc(86,15,1).
txt_doc(661,15,1).
txt_doc(852,15,1).
txt_doc(507,15,1).
txt_doc(430,15,1).
txt_doc(505,15,1).
txt_doc(319,15,1).
txt_doc(651,15,1).
txt_doc(939,15,1).
txt_doc(770,15,1).
txt_doc(772,15,1).
txt_doc(295,15,1).
txt_doc(269,15,1).
txt_doc(50,15,1).
txt_doc(810,15,1).
txt_doc(915,15,1).
txt_doc(620,15,1).
txt_doc(429,15,1).
txt_doc(212,15,1).
txt_doc(112,15,1).
txt_doc(776,15,1).
txt_doc(478,15,1).
txt_doc(663,15,1).
txt_doc(345,15,1).
txt_doc(843,15,1).
txt_doc(681,15,1).
txt_doc(955,15,2).
txt_doc(3,15,1).
txt_doc(323,15,1).
txt_doc(277,15,1).
txt_doc(948,15,1).
txt_doc(738,15,1).
txt_doc(474,15,1).
txt_doc(91,15,1).
txt_doc(327,15,1).
txt_doc(65,15,1).
txt_doc(533,15,1).
txt_doc(119,15,1).
txt_doc(339,15,2).
txt_doc(768,15,1).
txt_doc(510,15,1).
txt_doc(775,15,1).
txt_doc(650,15,1).
txt_doc(619,15,1).
txt_doc(807,15,1).
txt_doc(779,15,1).
txt_doc(528,15,1).
txt_doc(167,15,1).
txt_doc(332,15,1).
txt_doc(687,15,1).
txt_doc(162,15,1).
txt_doc(300,15,1).
txt_doc(875,15,1).
txt_doc(884,15,1).
txt_doc(517,15,1).
txt_doc(444,15,2).
txt_doc(114,15,1).
txt_doc(69,15,1).
txt_doc(863,15,1).
txt_doc(894,15,1).
txt_doc(17,15,2).
txt_doc(728,15,1).
txt_doc(611,15,1).
txt_doc(154,15,1).
txt_doc(854,15,1).
txt_doc(550,15,1).
txt_doc(2,15,1).
txt_doc(40,15,1).
txt_doc(911,15,1).
txt_doc(14,15,2).
txt_doc(753,15,1).
txt_doc(347,15,1).
txt_doc(536,15,1).
txt_doc(383,15,1).
txt_doc(297,15,1).
txt_doc(173,15,1).
txt_doc(164,15,2).
txt_doc(545,15,1).
txt_doc(367,15,2).
txt_doc(871,15,1).
txt_doc(170,15,1).
txt_doc(318,15,1).
txt_doc(633,15,1).
txt_doc(802,15,1).
txt_doc(518,15,1).
txt_doc(33,15,1).
txt_doc(882,15,1).
txt_doc(314,15,1).
txt_doc(362,15,1).
txt_doc(563,15,1).
txt_doc(208,15,1).
txt_doc(648,15,1).
txt_doc(739,15,1).
txt_doc(534,15,1).
txt_doc(515,15,1).
txt_doc(726,15,1).
txt_doc(439,15,1).
txt_doc(699,15,1).
txt_doc(306,15,1).
txt_doc(765,15,1).
txt_doc(723,15,1).
txt_doc(689,15,1).
txt_doc(26,15,1).
txt_doc(78,15,1).
txt_doc(138,15,2).
txt_doc(64,15,1).
txt_doc(836,15,1).
txt_doc(171,15,1).
txt_doc(680,15,1).
txt_doc(157,15,1).
txt_doc(581,15,1).
txt_doc(443,15,2).
txt_doc(706,15,2).
txt_doc(372,15,1).
txt_doc(280,15,1).
txt_doc(839,15,1).
txt_doc(486,15,1).
txt_doc(688,15,1).
txt_doc(744,15,1).
txt_doc(740,15,1).
txt_doc(201,15,1).
txt_doc(36,15,1).
txt_doc(529,15,1).
txt_doc(927,15,1).
txt_doc(217,15,1).
txt_doc(933,15,1).
txt_doc(442,15,1).
txt_doc(613,15,1).
txt_doc(481,15,2).
txt_doc(910,15,1).
txt_doc(822,15,1).
txt_doc(879,15,2).
txt_doc(124,15,1).
txt_doc(118,15,1).
txt_doc(664,15,1).
txt_doc(256,15,1).
txt_doc(471,15,1).
txt_doc(176,15,1).
txt_doc(716,15,1).
txt_doc(597,15,1).
txt_doc(886,15,1).
txt_doc(254,15,1).
txt_doc(641,15,1).
txt_doc(698,15,2).
txt_doc(850,15,1).
txt_doc(585,15,1).
txt_doc(887,15,1).
txt_doc(562,15,1).
txt_doc(316,15,1).
txt_doc(416,15,1).
txt_doc(521,15,1).
txt_doc(968,15,1).
txt_doc(721,15,1).
txt_doc(246,15,1).
txt_doc(736,15,1).
txt_doc(714,15,1).
txt_doc(364,15,1).
txt_doc(223,15,1).
txt_doc(542,15,1).
txt_doc(496,15,1).
txt_doc(155,15,1).
txt_doc(475,15,1).
txt_doc(656,15,1).
txt_doc(288,15,1).
txt_doc(127,15,1).
txt_doc(62,15,2).
txt_doc(406,15,1).
txt_doc(809,15,1).
txt_doc(951,15,1).
txt_doc(241,16,1).
txt_doc(504,16,1).
txt_doc(55,16,1).
txt_doc(524,16,1).
txt_doc(522,16,1).
txt_doc(948,16,1).
txt_doc(618,16,1).
txt_doc(249,16,2).
txt_doc(385,16,1).
txt_doc(575,16,1).
txt_doc(879,16,1).
txt_doc(712,16,1).
txt_doc(931,16,1).
txt_doc(45,16,1).
txt_doc(797,16,1).
txt_doc(124,16,1).
txt_doc(183,16,1).
txt_doc(503,16,1).
txt_doc(555,16,1).
txt_doc(414,16,1).
txt_doc(752,16,1).
txt_doc(251,16,1).
txt_doc(439,16,1).
txt_doc(956,16,1).
txt_doc(58,16,1).
txt_doc(373,16,1).
txt_doc(578,16,1).
txt_doc(198,16,1).
txt_doc(647,16,1).
txt_doc(925,16,1).
txt_doc(633,16,1).
txt_doc(396,16,1).
txt_doc(766,16,1).
txt_doc(254,16,1).
txt_doc(362,16,1).
txt_doc(722,16,1).
txt_doc(819,16,1).
txt_doc(520,16,1).
txt_doc(205,16,1).
txt_doc(14,16,1).
txt_doc(451,16,1).
txt_doc(891,16,1).
txt_doc(759,16,1).
txt_doc(309,16,1).
txt_doc(878,16,1).
txt_doc(41,16,1).
txt_doc(874,16,1).
txt_doc(936,16,1).
txt_doc(512,16,2).
txt_doc(861,16,1).
txt_doc(515,16,1).
txt_doc(913,16,1).
txt_doc(206,16,1).
txt_doc(823,16,1).
txt_doc(535,16,1).
txt_doc(352,16,1).
txt_doc(150,16,1).
txt_doc(423,16,1).
txt_doc(388,16,1).
txt_doc(680,16,1).
txt_doc(360,16,1).
txt_doc(18,16,1).
txt_doc(708,16,1).
txt_doc(157,16,1).
txt_doc(659,16,2).
txt_doc(643,16,1).
txt_doc(286,16,1).
txt_doc(166,16,1).
txt_doc(222,16,1).
txt_doc(276,16,1).
txt_doc(304,16,1).
txt_doc(121,16,1).
txt_doc(116,16,1).
txt_doc(394,16,1).
txt_doc(95,16,1).
txt_doc(462,16,1).
txt_doc(551,16,1).
txt_doc(541,16,1).
txt_doc(428,16,1).
txt_doc(344,16,1).
txt_doc(914,16,1).
txt_doc(949,16,1).
txt_doc(748,16,1).
txt_doc(6,16,1).
txt_doc(399,16,1).
txt_doc(179,16,1).
txt_doc(426,16,1).
txt_doc(966,16,1).
txt_doc(97,16,1).
txt_doc(125,16,1).
txt_doc(78,16,3).
txt_doc(910,16,1).
txt_doc(777,16,1).
txt_doc(143,16,1).
txt_doc(67,16,1).
txt_doc(732,16,1).
txt_doc(10,16,1).
txt_doc(474,16,2).
txt_doc(280,16,1).
txt_doc(815,16,1).
txt_doc(773,16,1).
txt_doc(587,16,1).
txt_doc(341,16,1).
txt_doc(685,16,1).
txt_doc(57,16,1).
txt_doc(698,16,1).
txt_doc(244,16,1).
txt_doc(279,16,1).
txt_doc(534,16,1).
txt_doc(620,16,1).
txt_doc(483,16,1).
txt_doc(932,16,1).
txt_doc(730,16,1).
txt_doc(724,16,1).
txt_doc(912,16,1).
txt_doc(813,16,1).
txt_doc(47,16,2).
txt_doc(4,16,1).
txt_doc(480,16,2).
txt_doc(805,16,1).
txt_doc(476,16,1).
txt_doc(735,16,3).
txt_doc(237,16,2).
txt_doc(468,16,1).
txt_doc(177,16,2).
txt_doc(71,16,1).
txt_doc(376,16,1).
txt_doc(576,16,1).
txt_doc(433,16,1).
txt_doc(80,16,1).
txt_doc(598,16,1).
txt_doc(8,16,1).
txt_doc(885,16,1).
txt_doc(622,16,1).
txt_doc(42,16,2).
txt_doc(397,16,1).
txt_doc(907,16,1).
txt_doc(90,16,1).
txt_doc(788,16,1).
txt_doc(849,16,2).
txt_doc(516,16,1).
txt_doc(260,16,1).
txt_doc(875,16,1).
txt_doc(454,16,1).
txt_doc(645,16,1).
txt_doc(294,16,1).
txt_doc(665,16,1).
txt_doc(315,16,1).
txt_doc(884,16,1).
txt_doc(623,16,1).
txt_doc(393,16,1).
txt_doc(24,16,1).
txt_doc(514,16,2).
txt_doc(60,16,1).
txt_doc(737,16,1).
txt_doc(408,16,1).
txt_doc(958,16,1).
txt_doc(923,16,1).
txt_doc(168,16,2).
txt_doc(678,16,2).
txt_doc(862,16,2).
txt_doc(673,16,1).
txt_doc(683,16,1).
txt_doc(754,16,1).
txt_doc(263,16,1).
txt_doc(156,16,1).
txt_doc(812,16,1).
txt_doc(328,16,1).
txt_doc(552,16,1).
txt_doc(789,16,2).
txt_doc(584,16,2).
txt_doc(16,16,1).
txt_doc(372,16,1).
txt_doc(700,16,2).
txt_doc(880,16,1).
txt_doc(11,16,2).
txt_doc(340,16,1).
txt_doc(682,16,1).
txt_doc(842,17,1).
txt_doc(341,17,1).
txt_doc(863,17,1).
txt_doc(684,17,1).
txt_doc(435,17,1).
txt_doc(313,17,1).
txt_doc(217,17,1).
txt_doc(709,17,1).
txt_doc(219,17,1).
txt_doc(522,17,1).
txt_doc(816,17,1).
txt_doc(83,17,1).
txt_doc(560,17,1).
txt_doc(544,17,1).
txt_doc(317,17,1).
txt_doc(392,17,1).
txt_doc(545,17,1).
txt_doc(661,17,1).
txt_doc(340,17,1).
txt_doc(880,17,1).
txt_doc(363,17,1).
txt_doc(192,17,1).
txt_doc(21,17,1).
txt_doc(785,17,1).
txt_doc(721,17,1).
txt_doc(32,17,1).
txt_doc(670,17,1).
txt_doc(685,17,1).
txt_doc(558,17,1).
txt_doc(499,17,1).
txt_doc(746,17,1).
txt_doc(293,17,1).
txt_doc(855,17,1).
txt_doc(656,17,1).
txt_doc(25,17,1).
txt_doc(148,17,1).
txt_doc(787,17,1).
txt_doc(494,17,1).
txt_doc(732,17,1).
txt_doc(625,17,2).
txt_doc(958,17,1).
txt_doc(750,17,1).
txt_doc(819,17,1).
txt_doc(786,17,1).
txt_doc(420,17,1).
txt_doc(267,17,1).
txt_doc(705,17,1).
txt_doc(595,17,1).
txt_doc(809,17,1).
txt_doc(719,17,1).
txt_doc(562,17,2).
txt_doc(414,17,2).
txt_doc(189,17,1).
txt_doc(243,17,1).
txt_doc(675,17,1).
txt_doc(492,17,1).
txt_doc(367,17,1).
txt_doc(760,17,1).
txt_doc(116,17,2).
txt_doc(130,17,1).
txt_doc(960,17,1).
txt_doc(307,17,1).
txt_doc(242,17,2).
txt_doc(111,17,1).
txt_doc(951,17,1).
txt_doc(424,17,1).
txt_doc(26,17,1).
txt_doc(751,17,1).
txt_doc(777,17,1).
txt_doc(594,17,1).
txt_doc(315,17,1).
txt_doc(381,17,1).
txt_doc(431,17,1).
txt_doc(361,17,1).
txt_doc(660,17,1).
txt_doc(278,17,1).
txt_doc(608,17,1).
txt_doc(965,17,1).
txt_doc(39,17,1).
txt_doc(956,17,1).
txt_doc(879,17,1).
txt_doc(781,17,1).
txt_doc(303,17,1).
txt_doc(165,17,1).
txt_doc(60,17,1).
txt_doc(92,17,1).
txt_doc(179,17,1).
txt_doc(543,17,1).
txt_doc(423,17,1).
txt_doc(469,17,1).
txt_doc(202,17,1).
txt_doc(354,17,1).
txt_doc(394,17,1).
txt_doc(873,17,1).
txt_doc(264,17,2).
txt_doc(147,17,2).
txt_doc(643,17,1).
txt_doc(799,17,1).
txt_doc(387,17,1).
txt_doc(553,17,1).
txt_doc(452,17,1).
txt_doc(570,17,1).
txt_doc(212,17,2).
txt_doc(820,17,1).
txt_doc(724,17,1).
txt_doc(740,17,1).
txt_doc(621,17,1).
txt_doc(581,17,1).
txt_doc(926,17,1).
txt_doc(206,17,1).
txt_doc(389,17,2).
txt_doc(5,17,1).
txt_doc(894,17,1).
txt_doc(339,17,1).
txt_doc(699,17,2).
txt_doc(700,17,1).
txt_doc(567,17,1).
txt_doc(441,17,1).
txt_doc(941,17,1).
txt_doc(758,17,1).
txt_doc(338,17,1).
txt_doc(540,17,2).
txt_doc(934,17,2).
txt_doc(382,17,1).
txt_doc(810,17,1).
txt_doc(54,17,1).
txt_doc(807,17,1).
txt_doc(704,17,2).
txt_doc(375,17,1).
txt_doc(416,17,1).
txt_doc(542,17,3).
txt_doc(662,17,1).
txt_doc(308,17,1).
txt_doc(386,17,1).
txt_doc(888,17,1).
txt_doc(351,17,2).
txt_doc(922,17,1).
txt_doc(2,17,1).
txt_doc(587,17,1).
txt_doc(273,17,1).
txt_doc(358,17,2).
txt_doc(672,17,1).
txt_doc(580,17,1).
txt_doc(356,17,2).
txt_doc(726,17,2).
txt_doc(297,17,1).
txt_doc(689,17,1).
txt_doc(77,17,1).
txt_doc(80,17,1).
txt_doc(906,17,1).
txt_doc(749,17,1).
txt_doc(576,17,1).
txt_doc(123,17,2).
txt_doc(836,17,1).
txt_doc(784,17,2).
txt_doc(105,17,1).
txt_doc(9,17,2).
txt_doc(98,17,1).
txt_doc(837,17,1).
txt_doc(355,17,1).
txt_doc(920,17,1).
txt_doc(615,17,1).
txt_doc(316,17,1).
txt_doc(115,17,1).
txt_doc(475,17,2).
txt_doc(48,17,1).
txt_doc(244,17,1).
txt_doc(783,17,2).
txt_doc(527,17,1).
txt_doc(556,17,1).
txt_doc(940,17,1).
txt_doc(22,17,1).
txt_doc(883,17,2).
txt_doc(663,17,2).
txt_doc(722,18,1).
txt_doc(343,18,1).
txt_doc(76,18,1).
txt_doc(864,18,1).
txt_doc(239,18,1).
txt_doc(210,18,1).
txt_doc(178,18,1).
txt_doc(125,18,1).
txt_doc(369,18,1).
txt_doc(840,18,1).
txt_doc(494,18,1).
txt_doc(188,18,1).
txt_doc(915,18,1).
txt_doc(665,18,1).
txt_doc(422,18,1).
txt_doc(861,18,1).
txt_doc(590,18,1).
txt_doc(592,18,1).
txt_doc(562,18,1).
txt_doc(624,18,1).
txt_doc(283,18,1).
txt_doc(462,18,1).
txt_doc(636,18,1).
txt_doc(92,18,1).
txt_doc(128,18,1).
txt_doc(880,18,1).
txt_doc(100,18,1).
txt_doc(661,18,1).
txt_doc(685,18,1).
txt_doc(212,18,1).
txt_doc(384,18,1).
txt_doc(848,18,1).
txt_doc(678,18,1).
txt_doc(470,18,1).
txt_doc(3,18,1).
txt_doc(884,18,1).
txt_doc(946,18,1).
txt_doc(673,18,1).
txt_doc(263,18,1).
txt_doc(14,18,1).
txt_doc(756,18,1).
txt_doc(806,18,1).
txt_doc(102,18,1).
txt_doc(596,18,1).
txt_doc(498,18,1).
txt_doc(347,18,1).
txt_doc(368,18,1).
txt_doc(845,18,1).
txt_doc(336,18,1).
txt_doc(450,18,1).
txt_doc(912,18,1).
txt_doc(250,18,1).
txt_doc(960,18,1).
txt_doc(825,18,1).
txt_doc(208,18,1).
txt_doc(154,18,1).
txt_doc(737,18,2).
txt_doc(488,18,1).
txt_doc(234,18,1).
txt_doc(817,18,1).
txt_doc(600,18,1).
txt_doc(296,18,1).
txt_doc(197,18,1).
txt_doc(601,18,1).
txt_doc(206,18,1).
txt_doc(541,18,1).
txt_doc(360,18,1).
txt_doc(477,18,1).
txt_doc(526,18,1).
txt_doc(290,18,1).
txt_doc(655,18,1).
txt_doc(788,18,1).
txt_doc(461,18,1).
txt_doc(889,18,1).
txt_doc(269,18,1).
txt_doc(834,18,1).
txt_doc(721,18,1).
txt_doc(968,18,1).
txt_doc(919,18,1).
txt_doc(646,18,1).
txt_doc(427,18,1).
txt_doc(687,18,2).
txt_doc(690,18,2).
txt_doc(710,18,1).
txt_doc(398,18,1).
txt_doc(774,18,1).
txt_doc(285,18,1).
txt_doc(447,18,1).
txt_doc(223,18,1).
txt_doc(929,18,1).
txt_doc(570,18,1).
txt_doc(958,18,1).
txt_doc(853,18,1).
txt_doc(344,18,1).
txt_doc(892,18,1).
txt_doc(34,18,1).
txt_doc(649,18,2).
txt_doc(160,18,1).
txt_doc(117,18,1).
txt_doc(524,18,1).
txt_doc(803,18,1).
txt_doc(703,18,1).
txt_doc(954,18,1).
txt_doc(823,18,1).
txt_doc(745,18,1).
txt_doc(169,18,2).
txt_doc(293,18,1).
txt_doc(94,18,1).
txt_doc(375,18,1).
txt_doc(571,18,1).
txt_doc(900,18,2).
txt_doc(778,18,1).
txt_doc(504,18,1).
txt_doc(559,18,1).
txt_doc(335,18,1).
txt_doc(113,18,1).
txt_doc(430,18,1).
txt_doc(805,18,1).
txt_doc(642,18,1).
txt_doc(760,18,1).
txt_doc(870,18,1).
txt_doc(733,18,1).
txt_doc(275,18,2).
txt_doc(22,18,2).
txt_doc(217,18,1).
txt_doc(584,18,1).
txt_doc(48,18,1).
txt_doc(715,18,1).
txt_doc(758,18,1).
txt_doc(777,18,1).
txt_doc(657,18,1).
txt_doc(939,18,2).
txt_doc(45,18,1).
txt_doc(182,18,1).
txt_doc(166,18,1).
txt_doc(226,18,1).
txt_doc(82,18,1).
txt_doc(214,18,1).
txt_doc(252,18,1).
txt_doc(43,18,1).
txt_doc(963,18,3).
txt_doc(783,18,2).
txt_doc(952,18,1).
txt_doc(444,18,1).
txt_doc(439,18,1).
txt_doc(728,18,2).
txt_doc(755,18,4).
txt_doc(727,18,2).
txt_doc(643,18,1).
txt_doc(337,18,1).
txt_doc(620,18,1).
txt_doc(515,18,2).
txt_doc(552,18,1).
txt_doc(413,18,1).
txt_doc(682,18,1).
txt_doc(648,18,1).
txt_doc(287,18,1).
txt_doc(839,18,1).
txt_doc(150,18,1).
txt_doc(622,18,1).
txt_doc(219,18,1).
txt_doc(638,18,1).
txt_doc(301,18,1).
txt_doc(621,18,1).
txt_doc(44,18,1).
txt_doc(377,18,1).
txt_doc(906,18,1).
txt_doc(51,18,2).
txt_doc(288,18,2).
txt_doc(938,18,1).
txt_doc(831,18,1).
txt_doc(692,18,2).
txt_doc(797,18,1).
txt_doc(672,18,1).
txt_doc(187,18,1).
txt_doc(459,18,1).
txt_doc(433,18,1).
txt_doc(811,18,1).
txt_doc(501,18,1).
txt_doc(301,19,1).
txt_doc(140,19,1).
txt_doc(696,19,1).
txt_doc(146,19,1).
txt_doc(877,19,1).
txt_doc(352,19,1).
txt_doc(858,19,1).
txt_doc(786,19,1).
txt_doc(644,19,1).
txt_doc(764,19,1).
txt_doc(493,19,1).
txt_doc(192,19,1).
txt_doc(40,19,1).
txt_doc(763,19,1).
txt_doc(60,19,1).
txt_doc(598,19,1).
txt_doc(98,19,1).
txt_doc(702,19,1).
txt_doc(899,19,1).
txt_doc(214,19,1).
txt_doc(609,19,1).
txt_doc(567,19,1).
txt_doc(438,19,1).
txt_doc(370,19,1).
txt_doc(913,19,2).
txt_doc(68,19,1).
txt_doc(252,19,1).
txt_doc(453,19,1).
txt_doc(178,19,1).
txt_doc(613,19,1).
txt_doc(2,19,1).
txt_doc(268,19,1).
txt_doc(851,19,1).
txt_doc(367,19,1).
txt_doc(319,19,1).
txt_doc(713,19,1).
txt_doc(261,19,1).
txt_doc(132,19,1).
txt_doc(559,19,1).
txt_doc(233,19,1).
txt_doc(353,19,1).
txt_doc(412,19,1).
txt_doc(921,19,1).
txt_doc(577,19,1).
txt_doc(293,19,1).
txt_doc(749,19,1).
txt_doc(936,19,1).
txt_doc(403,19,2).
txt_doc(683,19,1).
txt_doc(385,19,1).
txt_doc(200,19,1).
txt_doc(255,19,1).
txt_doc(147,19,1).
txt_doc(134,19,1).
txt_doc(859,19,1).
txt_doc(778,19,1).
txt_doc(430,19,1).
txt_doc(669,19,1).
txt_doc(498,19,1).
txt_doc(797,19,1).
txt_doc(540,19,1).
txt_doc(945,19,1).
txt_doc(791,19,1).
txt_doc(377,19,1).
txt_doc(822,19,1).
txt_doc(618,19,1).
txt_doc(825,19,1).
txt_doc(870,19,1).
txt_doc(37,19,1).
txt_doc(418,19,1).
txt_doc(614,19,1).
txt_doc(460,19,1).
txt_doc(424,19,1).
txt_doc(101,19,1).
txt_doc(408,19,1).
txt_doc(807,19,1).
txt_doc(529,19,1).
txt_doc(772,19,1).
txt_doc(112,19,1).
txt_doc(834,19,1).
txt_doc(31,19,1).
txt_doc(685,19,1).
txt_doc(724,19,1).
txt_doc(432,19,1).
txt_doc(589,19,1).
txt_doc(116,19,1).
txt_doc(10,19,1).
txt_doc(401,19,1).
txt_doc(787,19,1).
txt_doc(145,19,1).
txt_doc(56,19,1).
txt_doc(141,19,1).
txt_doc(653,19,2).
txt_doc(6,19,1).
txt_doc(623,19,2).
txt_doc(689,19,1).
txt_doc(50,19,1).
txt_doc(477,19,1).
txt_doc(113,19,1).
txt_doc(266,19,1).
txt_doc(581,19,1).
txt_doc(881,19,1).
txt_doc(382,19,1).
txt_doc(237,19,1).
txt_doc(27,19,1).
txt_doc(372,19,1).
txt_doc(3,19,1).
txt_doc(235,19,1).
txt_doc(89,19,1).
txt_doc(479,19,1).
txt_doc(360,19,1).
txt_doc(478,19,1).
txt_doc(219,19,2).
txt_doc(532,19,1).
txt_doc(281,19,1).
txt_doc(649,19,1).
txt_doc(421,19,1).
txt_doc(880,19,1).
txt_doc(601,19,2).
txt_doc(125,19,1).
txt_doc(621,19,1).
txt_doc(313,19,1).
txt_doc(678,19,1).
txt_doc(892,19,1).
txt_doc(840,19,1).
txt_doc(426,19,1).
txt_doc(508,19,1).
txt_doc(677,19,1).
txt_doc(19,19,1).
txt_doc(227,19,1).
txt_doc(760,19,1).
txt_doc(815,19,1).
txt_doc(152,19,1).
txt_doc(777,19,1).
txt_doc(728,19,1).
txt_doc(867,19,1).
txt_doc(602,19,1).
txt_doc(429,19,1).
txt_doc(24,19,1).
txt_doc(406,19,2).
txt_doc(114,19,1).
txt_doc(648,19,1).
txt_doc(811,19,1).
txt_doc(844,19,2).
txt_doc(693,19,2).
txt_doc(444,19,1).
txt_doc(300,19,1).
txt_doc(303,19,1).
txt_doc(651,19,1).
txt_doc(862,19,1).
txt_doc(726,19,2).
txt_doc(334,19,1).
txt_doc(823,19,1).
txt_doc(587,19,2).
txt_doc(257,19,1).
txt_doc(536,19,1).
txt_doc(624,19,2).
txt_doc(476,19,1).
txt_doc(890,19,1).
txt_doc(212,19,1).
txt_doc(149,19,1).
txt_doc(668,19,1).
txt_doc(817,19,1).
txt_doc(384,19,1).
txt_doc(67,19,1).
txt_doc(434,19,1).
txt_doc(754,19,1).
txt_doc(241,19,1).
txt_doc(922,19,1).
txt_doc(509,19,1).
txt_doc(184,19,1).
txt_doc(783,19,1).
txt_doc(546,19,1).
txt_doc(954,19,1).
txt_doc(77,19,1).
txt_doc(739,19,1).
txt_doc(291,19,1).
txt_doc(955,19,2).
txt_doc(761,19,1).
txt_doc(963,19,1).
txt_doc(247,19,1).
txt_doc(260,19,1).
txt_doc(399,19,1).
txt_doc(4,19,2).
txt_doc(842,19,2).
txt_doc(740,20,1).
txt_doc(273,20,1).
txt_doc(946,20,1).
txt_doc(119,20,1).
txt_doc(433,20,1).
txt_doc(1,20,1).
txt_doc(156,20,1).
txt_doc(963,20,1).
txt_doc(494,20,1).
txt_doc(912,20,1).
txt_doc(473,20,1).
txt_doc(361,20,1).
txt_doc(45,20,1).
txt_doc(306,20,1).
txt_doc(878,20,1).
txt_doc(115,20,1).
txt_doc(248,20,1).
txt_doc(533,20,1).
txt_doc(154,20,1).
txt_doc(867,20,1).
txt_doc(325,20,1).
txt_doc(958,20,2).
txt_doc(936,20,1).
txt_doc(948,20,1).
txt_doc(955,20,1).
txt_doc(62,20,1).
txt_doc(604,20,1).
txt_doc(451,20,1).
txt_doc(565,20,1).
txt_doc(167,20,1).
txt_doc(859,20,1).
txt_doc(791,20,1).
txt_doc(746,20,1).
txt_doc(635,20,1).
txt_doc(166,20,1).
txt_doc(51,20,1).
txt_doc(404,20,2).
txt_doc(636,20,2).
txt_doc(134,20,1).
txt_doc(28,20,1).
txt_doc(482,20,1).
txt_doc(129,20,1).
txt_doc(15,20,1).
txt_doc(495,20,2).
txt_doc(498,20,1).
txt_doc(80,20,1).
txt_doc(778,20,1).
txt_doc(509,20,1).
txt_doc(516,20,1).
txt_doc(915,20,1).
txt_doc(337,20,1).
txt_doc(828,20,1).
txt_doc(493,20,1).
txt_doc(228,20,1).
txt_doc(510,20,1).
txt_doc(777,20,1).
txt_doc(886,20,1).
txt_doc(260,20,1).
txt_doc(393,20,1).
txt_doc(233,20,1).
txt_doc(363,20,1).
txt_doc(19,20,1).
txt_doc(198,20,2).
txt_doc(434,20,1).
txt_doc(699,20,1).
txt_doc(517,20,1).
txt_doc(935,20,1).
txt_doc(903,20,1).
txt_doc(313,20,1).
txt_doc(208,20,1).
txt_doc(760,20,1).
txt_doc(102,20,1).
txt_doc(437,20,2).
txt_doc(22,20,1).
txt_doc(854,20,1).
txt_doc(586,20,3).
txt_doc(488,20,1).
txt_doc(99,20,1).
txt_doc(378,20,1).
txt_doc(797,20,1).
txt_doc(486,20,1).
txt_doc(530,20,1).
txt_doc(427,20,1).
txt_doc(357,20,1).
txt_doc(538,20,1).
txt_doc(927,20,1).
txt_doc(841,20,1).
txt_doc(432,20,1).
txt_doc(683,20,1).
txt_doc(318,20,1).
txt_doc(888,20,1).
txt_doc(689,20,1).
txt_doc(785,20,1).
txt_doc(216,20,1).
txt_doc(798,20,1).
txt_doc(246,20,1).
txt_doc(630,20,1).
txt_doc(708,20,2).
txt_doc(944,20,1).
txt_doc(943,20,2).
txt_doc(653,20,1).
txt_doc(410,20,1).
txt_doc(403,20,1).
txt_doc(676,20,2).
txt_doc(743,20,2).
txt_doc(536,20,1).
txt_doc(640,20,1).
txt_doc(160,20,1).
txt_doc(38,20,1).
txt_doc(85,20,1).
txt_doc(383,20,1).
txt_doc(669,20,1).
txt_doc(250,20,1).
txt_doc(637,20,1).
txt_doc(658,20,1).
txt_doc(298,20,2).
txt_doc(96,20,2).
txt_doc(77,20,1).
txt_doc(396,20,1).
txt_doc(576,20,1).
txt_doc(219,20,1).
txt_doc(414,20,1).
txt_doc(308,20,1).
txt_doc(485,20,1).
txt_doc(567,20,1).
txt_doc(686,20,1).
txt_doc(800,20,1).
txt_doc(827,20,1).
txt_doc(619,20,1).
txt_doc(74,20,1).
txt_doc(18,20,1).
txt_doc(391,20,1).
txt_doc(926,20,1).
txt_doc(611,20,1).
txt_doc(752,20,1).
txt_doc(302,20,1).
txt_doc(571,20,2).
txt_doc(870,20,1).
txt_doc(284,20,1).
txt_doc(150,20,1).
txt_doc(774,20,1).
txt_doc(340,20,1).
txt_doc(21,20,1).
txt_doc(387,20,1).
txt_doc(370,20,2).
txt_doc(424,20,1).
txt_doc(285,20,1).
txt_doc(401,20,1).
txt_doc(738,20,1).
txt_doc(716,20,1).
txt_doc(194,20,1).
txt_doc(83,20,1).
txt_doc(128,20,1).
txt_doc(262,20,1).
txt_doc(548,20,1).
txt_doc(118,20,1).
txt_doc(203,20,1).
txt_doc(857,20,1).
txt_doc(181,20,2).
txt_doc(191,20,1).
txt_doc(202,20,1).
txt_doc(799,20,1).
txt_doc(436,20,1).
txt_doc(131,20,1).
txt_doc(463,20,1).
txt_doc(147,20,1).
txt_doc(763,20,1).
txt_doc(563,20,1).
txt_doc(240,20,1).
txt_doc(95,20,1).
txt_doc(459,20,2).
txt_doc(825,20,1).
txt_doc(241,20,1).
txt_doc(901,20,1).
txt_doc(165,20,1).
txt_doc(348,20,1).
txt_doc(535,20,1).
txt_doc(120,20,1).
txt_doc(770,20,1).
txt_doc(739,20,1).
txt_doc(278,20,1).
txt_doc(196,20,1).
txt_doc(531,21,1).
txt_doc(353,21,1).
txt_doc(575,21,1).
txt_doc(433,21,1).
txt_doc(389,21,1).
txt_doc(700,21,1).
txt_doc(679,21,1).
txt_doc(317,21,1).
txt_doc(56,21,1).
txt_doc(668,21,2).
txt_doc(498,21,1).
txt_doc(138,21,1).
txt_doc(385,21,1).
txt_doc(669,21,1).
txt_doc(213,21,1).
txt_doc(234,21,1).
txt_doc(408,21,1).
txt_doc(841,21,1).
txt_doc(13,21,1).
txt_doc(955,21,1).
txt_doc(50,21,1).
txt_doc(393,21,1).
txt_doc(945,21,1).
txt_doc(69,21,1).
txt_doc(489,21,1).
txt_doc(493,21,1).
txt_doc(792,21,1).
txt_doc(620,21,1).
txt_doc(672,21,1).
txt_doc(437,21,1).
txt_doc(603,21,1).
txt_doc(502,21,1).
txt_doc(612,21,1).
txt_doc(359,21,1).
txt_doc(789,21,1).
txt_doc(845,21,1).
txt_doc(934,21,2).
txt_doc(390,21,1).
txt_doc(499,21,1).
txt_doc(510,21,1).
txt_doc(657,21,1).
txt_doc(847,21,1).
txt_doc(192,21,1).
txt_doc(855,21,1).
txt_doc(484,21,1).
txt_doc(469,21,1).
txt_doc(866,21,2).
txt_doc(822,21,1).
txt_doc(506,21,1).
txt_doc(487,21,1).
txt_doc(236,21,1).
txt_doc(547,21,2).
txt_doc(431,21,1).
txt_doc(637,21,2).
txt_doc(247,21,1).
txt_doc(894,21,1).
txt_doc(18,21,1).
txt_doc(587,21,1).
txt_doc(181,21,2).
txt_doc(20,21,1).
txt_doc(564,21,1).
txt_doc(186,21,2).
txt_doc(748,21,1).
txt_doc(747,21,1).
txt_doc(99,21,1).
txt_doc(935,21,1).
txt_doc(198,21,1).
txt_doc(117,21,1).
txt_doc(712,21,1).
txt_doc(878,21,1).
txt_doc(743,21,2).
txt_doc(9,21,2).
txt_doc(905,21,1).
txt_doc(43,21,2).
txt_doc(886,21,1).
txt_doc(15,21,1).
txt_doc(932,21,1).
txt_doc(46,21,1).
txt_doc(429,21,1).
txt_doc(266,21,1).
txt_doc(908,21,1).
txt_doc(826,21,1).
txt_doc(403,21,1).
txt_doc(304,21,1).
txt_doc(66,21,1).
txt_doc(897,21,2).
txt_doc(261,21,1).
txt_doc(658,21,1).
txt_doc(681,21,1).
txt_doc(778,21,1).
txt_doc(147,21,1).
txt_doc(32,21,1).
txt_doc(61,21,1).
txt_doc(756,21,1).
txt_doc(205,21,2).
txt_doc(800,21,1).
txt_doc(290,21,1).
txt_doc(495,21,1).
txt_doc(54,21,2).
txt_doc(566,21,1).
txt_doc(394,21,1).
txt_doc(605,21,1).
txt_doc(505,21,1).
txt_doc(818,21,1).
txt_doc(656,21,1).
txt_doc(691,21,1).
txt_doc(532,21,1).
txt_doc(142,21,2).
txt_doc(197,21,1).
txt_doc(386,21,1).
txt_doc(171,21,1).
txt_doc(678,21,1).
txt_doc(308,21,1).
txt_doc(159,21,1).
txt_doc(492,21,1).
txt_doc(88,21,1).
txt_doc(311,21,1).
txt_doc(86,21,1).
txt_doc(116,21,1).
txt_doc(863,21,1).
txt_doc(708,21,1).
txt_doc(409,21,2).
txt_doc(752,21,1).
txt_doc(89,21,1).
txt_doc(951,21,1).
txt_doc(222,21,1).
txt_doc(260,21,1).
txt_doc(707,21,1).
txt_doc(175,21,2).
txt_doc(577,21,1).
txt_doc(347,21,1).
txt_doc(96,21,1).
txt_doc(706,21,4).
txt_doc(447,21,1).
txt_doc(795,21,1).
txt_doc(307,21,1).
txt_doc(210,21,1).
txt_doc(869,21,1).
txt_doc(67,21,1).
txt_doc(288,21,1).
txt_doc(425,21,1).
txt_doc(931,21,2).
txt_doc(751,21,1).
txt_doc(862,21,1).
txt_doc(799,21,1).
txt_doc(844,21,1).
txt_doc(214,21,1).
txt_doc(383,21,4).
txt_doc(398,21,1).
txt_doc(257,21,1).
txt_doc(180,21,2).
txt_doc(160,21,1).
txt_doc(949,21,2).
txt_doc(821,21,1).
txt_doc(741,21,2).
txt_doc(379,21,1).
txt_doc(595,21,1).
txt_doc(850,21,1).
txt_doc(146,21,1).
txt_doc(105,21,1).
txt_doc(249,21,2).
txt_doc(610,21,1).
txt_doc(823,21,1).
txt_doc(805,21,1).
txt_doc(64,21,1).
txt_doc(313,21,1).
txt_doc(594,21,1).
txt_doc(203,21,1).
txt_doc(966,21,1).
txt_doc(124,21,1).
txt_doc(230,21,1).
txt_doc(830,21,1).
txt_doc(597,21,1).
txt_doc(453,22,1).
txt_doc(618,22,1).
txt_doc(36,22,1).
txt_doc(480,22,1).
txt_doc(573,22,1).
txt_doc(582,22,1).
txt_doc(786,22,1).
txt_doc(272,22,1).
txt_doc(231,22,1).
txt_doc(78,22,2).
txt_doc(433,22,1).
txt_doc(753,22,1).
txt_doc(222,22,1).
txt_doc(517,22,1).
txt_doc(653,22,1).
txt_doc(374,22,1).
txt_doc(769,22,1).
txt_doc(153,22,1).
txt_doc(238,22,1).
txt_doc(889,22,1).
txt_doc(646,22,1).
txt_doc(926,22,1).
txt_doc(299,22,1).
txt_doc(505,22,1).
txt_doc(925,22,1).
txt_doc(884,22,1).
txt_doc(738,22,1).
txt_doc(351,22,1).
txt_doc(910,22,1).
txt_doc(496,22,1).
txt_doc(263,22,2).
txt_doc(542,22,1).
txt_doc(628,22,1).
txt_doc(426,22,1).
txt_doc(891,22,1).
txt_doc(850,22,1).
txt_doc(481,22,1).
txt_doc(69,22,1).
txt_doc(178,22,1).
txt_doc(37,22,1).
txt_doc(952,22,1).
txt_doc(875,22,1).
txt_doc(709,22,1).
txt_doc(598,22,1).
txt_doc(157,22,1).
txt_doc(962,22,1).
txt_doc(751,22,1).
txt_doc(873,22,1).
txt_doc(115,22,1).
txt_doc(615,22,1).
txt_doc(330,22,1).
txt_doc(811,22,1).
txt_doc(942,22,1).
txt_doc(956,22,2).
txt_doc(616,22,1).
txt_doc(411,22,1).
txt_doc(212,22,1).
txt_doc(624,22,1).
txt_doc(818,22,1).
txt_doc(544,22,1).
txt_doc(304,22,1).
txt_doc(820,22,1).
txt_doc(288,22,1).
txt_doc(402,22,1).
txt_doc(398,22,1).
txt_doc(716,22,1).
txt_doc(46,22,1).
txt_doc(724,22,1).
txt_doc(808,22,1).
txt_doc(856,22,1).
txt_doc(447,22,1).
txt_doc(960,22,2).
txt_doc(205,22,1).
txt_doc(900,22,1).
txt_doc(522,22,1).
txt_doc(455,22,1).
txt_doc(879,22,1).
txt_doc(387,22,1).
txt_doc(285,22,1).
txt_doc(934,22,1).
txt_doc(555,22,1).
txt_doc(278,22,1).
txt_doc(497,22,1).
txt_doc(501,22,1).
txt_doc(785,22,1).
txt_doc(504,22,1).
txt_doc(29,22,1).
txt_doc(587,22,1).
txt_doc(89,22,1).
txt_doc(511,22,2).
txt_doc(10,22,1).
txt_doc(164,22,1).
txt_doc(554,22,1).
txt_doc(245,22,3).
txt_doc(232,22,1).
txt_doc(707,22,1).
txt_doc(663,22,1).
txt_doc(940,22,1).
txt_doc(644,22,1).
txt_doc(797,22,1).
txt_doc(581,22,2).
txt_doc(275,22,2).
txt_doc(764,22,1).
txt_doc(196,22,1).
txt_doc(51,22,1).
txt_doc(183,22,1).
txt_doc(35,22,1).
txt_doc(393,22,1).
txt_doc(672,22,2).
txt_doc(226,22,1).
txt_doc(282,22,1).
txt_doc(527,22,1).
txt_doc(508,22,1).
txt_doc(896,22,1).
txt_doc(23,22,1).
txt_doc(274,22,1).
txt_doc(471,22,1).
txt_doc(313,22,1).
txt_doc(853,22,1).
txt_doc(547,22,1).
txt_doc(964,22,1).
txt_doc(309,22,1).
txt_doc(649,22,1).
txt_doc(857,22,1).
txt_doc(247,22,1).
txt_doc(930,22,2).
txt_doc(927,22,1).
txt_doc(704,22,1).
txt_doc(840,22,1).
txt_doc(635,22,1).
txt_doc(104,22,1).
txt_doc(412,22,1).
txt_doc(215,22,1).
txt_doc(99,22,1).
txt_doc(362,22,1).
txt_doc(259,22,1).
txt_doc(56,22,1).
txt_doc(592,22,1).
txt_doc(476,22,1).
txt_doc(227,22,1).
txt_doc(434,22,1).
txt_doc(631,22,1).
txt_doc(605,22,2).
txt_doc(920,22,1).
txt_doc(637,22,1).
txt_doc(267,22,2).
txt_doc(449,22,1).
txt_doc(461,22,1).
txt_doc(967,22,1).
txt_doc(601,22,1).
txt_doc(775,22,1).
txt_doc(290,22,1).
txt_doc(156,22,2).
txt_doc(430,22,1).
txt_doc(965,22,1).
txt_doc(427,22,1).
txt_doc(45,22,1).
txt_doc(462,22,1).
txt_doc(759,22,1).
txt_doc(47,22,1).
txt_doc(86,22,1).
txt_doc(651,22,2).
txt_doc(489,22,1).
txt_doc(249,22,1).
txt_doc(777,22,1).
txt_doc(144,22,1).
txt_doc(325,22,1).
txt_doc(173,22,1).
txt_doc(766,22,1).
txt_doc(291,22,1).
txt_doc(317,22,1).
txt_doc(87,22,1).
txt_doc(16,22,1).
txt_doc(961,22,1).
txt_doc(543,22,1).
txt_doc(783,22,1).
txt_doc(6,22,1).
txt_doc(594,22,2).
txt_doc(33,22,1).
txt_doc(403,22,1).
txt_doc(640,22,2).
txt_doc(425,22,1).
txt_doc(912,22,1).
txt_doc(350,23,1).
txt_doc(423,23,1).
txt_doc(911,23,1).
txt_doc(429,23,1).
txt_doc(907,23,1).
txt_doc(678,23,1).
txt_doc(100,23,1).
txt_doc(271,23,1).
txt_doc(910,23,1).
txt_doc(484,23,1).
txt_doc(26,23,1).
txt_doc(414,23,1).
txt_doc(612,23,2).
txt_doc(473,23,1).
txt_doc(46,23,1).
txt_doc(936,23,1).
txt_doc(864,23,1).
txt_doc(808,23,1).
txt_doc(47,23,2).
txt_doc(862,23,1).
txt_doc(483,23,1).
txt_doc(260,23,1).
txt_doc(524,23,1).
txt_doc(736,23,1).
txt_doc(828,23,1).
txt_doc(840,23,1).
txt_doc(667,23,1).
txt_doc(273,23,1).
txt_doc(606,23,2).
txt_doc(229,23,1).
txt_doc(283,23,2).
txt_doc(63,23,1).
txt_doc(926,23,2).
txt_doc(717,23,1).
txt_doc(835,23,1).
txt_doc(860,23,1).
txt_doc(322,23,1).
txt_doc(479,23,1).
txt_doc(14,23,1).
txt_doc(345,23,1).
txt_doc(440,23,1).
txt_doc(877,23,1).
txt_doc(186,23,1).
txt_doc(427,23,1).
txt_doc(536,23,1).
txt_doc(879,23,1).
txt_doc(848,23,1).
txt_doc(197,23,1).
txt_doc(838,23,1).
txt_doc(52,23,1).
txt_doc(520,23,1).
txt_doc(767,23,1).
txt_doc(106,23,1).
txt_doc(174,23,1).
txt_doc(541,23,1).
txt_doc(347,23,1).
txt_doc(676,23,1).
txt_doc(552,23,1).
txt_doc(576,23,1).
txt_doc(141,23,1).
txt_doc(694,23,1).
txt_doc(686,23,1).
txt_doc(950,23,1).
txt_doc(661,23,1).
txt_doc(954,23,1).
txt_doc(598,23,2).
txt_doc(373,23,1).
txt_doc(568,23,1).
txt_doc(89,23,2).
txt_doc(850,23,1).
txt_doc(605,23,1).
txt_doc(544,23,1).
txt_doc(553,23,1).
txt_doc(548,23,1).
txt_doc(624,23,1).
txt_doc(250,23,1).
txt_doc(650,23,1).
txt_doc(195,23,2).
txt_doc(240,23,1).
txt_doc(142,23,1).
txt_doc(540,23,1).
txt_doc(556,23,1).
txt_doc(733,23,1).
txt_doc(54,23,3).
txt_doc(196,23,1).
txt_doc(967,23,3).
txt_doc(662,23,1).
txt_doc(563,23,1).
txt_doc(183,23,1).
txt_doc(656,23,1).
txt_doc(951,23,1).
txt_doc(527,23,1).
txt_doc(812,23,1).
txt_doc(139,23,1).
txt_doc(231,23,1).
txt_doc(120,23,1).
txt_doc(44,23,1).
txt_doc(311,23,1).
txt_doc(438,23,1).
txt_doc(704,23,1).
txt_doc(37,23,1).
txt_doc(762,23,1).
txt_doc(146,23,2).
txt_doc(287,23,1).
txt_doc(654,23,1).
txt_doc(23,23,1).
txt_doc(968,23,1).
txt_doc(153,23,2).
txt_doc(500,23,2).
txt_doc(435,23,1).
txt_doc(546,23,1).
txt_doc(406,23,1).
txt_doc(338,23,1).
txt_doc(303,23,1).
txt_doc(731,23,1).
txt_doc(683,23,1).
txt_doc(337,23,1).
txt_doc(327,23,1).
txt_doc(167,23,1).
txt_doc(103,23,1).
txt_doc(709,23,2).
txt_doc(149,23,1).
txt_doc(932,23,1).
txt_doc(158,23,1).
txt_doc(220,23,1).
txt_doc(387,23,1).
txt_doc(329,23,1).
txt_doc(904,23,2).
txt_doc(76,23,1).
txt_doc(623,23,1).
txt_doc(494,23,1).
txt_doc(669,23,1).
txt_doc(937,23,1).
txt_doc(324,23,1).
txt_doc(737,23,1).
txt_doc(773,23,1).
txt_doc(604,23,1).
txt_doc(876,23,1).
txt_doc(778,23,1).
txt_doc(741,23,1).
txt_doc(728,23,1).
txt_doc(888,23,1).
txt_doc(49,23,1).
txt_doc(274,23,1).
txt_doc(379,23,1).
txt_doc(647,23,1).
txt_doc(176,23,1).
txt_doc(275,23,1).
txt_doc(627,23,1).
txt_doc(955,23,1).
txt_doc(652,23,1).
txt_doc(11,23,1).
txt_doc(870,23,2).
txt_doc(942,23,3).
txt_doc(184,23,2).
txt_doc(161,23,1).
txt_doc(380,23,1).
txt_doc(940,23,1).
txt_doc(464,23,1).
txt_doc(897,23,1).
txt_doc(570,23,1).
txt_doc(461,23,1).
txt_doc(868,23,1).
txt_doc(735,23,1).
txt_doc(318,23,1).
txt_doc(241,23,1).
txt_doc(190,23,1).
txt_doc(820,23,1).
txt_doc(644,23,1).
txt_doc(616,23,1).
txt_doc(310,23,1).
txt_doc(549,23,1).
txt_doc(668,23,1).
txt_doc(189,23,1).
txt_doc(65,23,1).
txt_doc(424,23,1).
txt_doc(891,23,1).
txt_doc(306,23,1).
txt_doc(614,23,1).
txt_doc(882,24,1).
txt_doc(777,24,1).
txt_doc(330,24,1).
txt_doc(14,24,1).
txt_doc(839,24,1).
txt_doc(437,24,1).
txt_doc(38,24,1).
txt_doc(85,24,1).
txt_doc(754,24,1).
txt_doc(962,24,1).
txt_doc(197,24,1).
txt_doc(74,24,1).
txt_doc(877,24,1).
txt_doc(397,24,1).
txt_doc(132,24,1).
txt_doc(642,24,1).
txt_doc(953,24,1).
txt_doc(270,24,1).
txt_doc(180,24,1).
txt_doc(209,24,1).
txt_doc(530,24,1).
txt_doc(341,24,1).
txt_doc(373,24,1).
txt_doc(408,24,1).
txt_doc(503,24,1).
txt_doc(924,24,1).
txt_doc(637,24,1).
txt_doc(811,24,1).
txt_doc(208,24,1).
txt_doc(376,24,1).
txt_doc(114,24,1).
txt_doc(414,24,1).
txt_doc(357,24,1).
txt_doc(396,24,1).
txt_doc(488,24,1).
txt_doc(120,24,1).
txt_doc(548,24,1).
txt_doc(888,24,1).
txt_doc(836,24,1).
txt_doc(300,24,1).
txt_doc(866,24,1).
txt_doc(935,24,1).
txt_doc(510,24,1).
txt_doc(746,24,1).
txt_doc(321,24,1).
txt_doc(659,24,1).
txt_doc(106,24,1).
txt_doc(492,24,1).
txt_doc(675,24,1).
txt_doc(819,24,2).
txt_doc(202,24,1).
txt_doc(176,24,1).
txt_doc(963,24,1).
txt_doc(504,24,1).
txt_doc(762,24,1).
txt_doc(482,24,1).
txt_doc(61,24,1).
txt_doc(790,24,2).
txt_doc(818,24,1).
txt_doc(604,24,1).
txt_doc(375,24,1).
txt_doc(113,24,1).
txt_doc(785,24,2).
txt_doc(171,24,1).
txt_doc(890,24,1).
txt_doc(90,24,1).
txt_doc(98,24,1).
txt_doc(57,24,1).
txt_doc(168,24,1).
txt_doc(173,24,1).
txt_doc(657,24,1).
txt_doc(206,24,1).
txt_doc(873,24,1).
txt_doc(440,24,1).
txt_doc(485,24,1).
txt_doc(847,24,1).
txt_doc(743,24,1).
txt_doc(656,24,1).
txt_doc(34,24,1).
txt_doc(411,24,2).
txt_doc(636,24,2).
txt_doc(549,24,1).
txt_doc(910,24,1).
txt_doc(187,24,1).
txt_doc(256,24,1).
txt_doc(207,24,1).
txt_doc(268,24,1).
txt_doc(327,24,1).
txt_doc(619,24,1).
txt_doc(605,24,1).
txt_doc(566,24,1).
txt_doc(302,24,1).
txt_doc(9,24,1).
txt_doc(576,24,2).
txt_doc(248,24,1).
txt_doc(447,24,1).
txt_doc(876,24,1).
txt_doc(476,24,1).
txt_doc(897,24,1).
txt_doc(925,24,1).
txt_doc(781,24,1).
txt_doc(237,24,1).
txt_doc(705,24,1).
txt_doc(467,24,2).
txt_doc(31,24,1).
txt_doc(399,24,1).
txt_doc(77,24,1).
txt_doc(652,24,1).
txt_doc(45,24,1).
txt_doc(298,24,1).
txt_doc(219,24,1).
txt_doc(346,24,1).
txt_doc(644,24,1).
txt_doc(783,24,1).
txt_doc(462,24,1).
txt_doc(824,24,2).
txt_doc(515,24,1).
txt_doc(374,24,1).
txt_doc(522,24,1).
txt_doc(872,24,1).
txt_doc(714,24,1).
txt_doc(102,24,1).
txt_doc(477,24,1).
txt_doc(853,24,1).
txt_doc(831,24,1).
txt_doc(287,24,1).
txt_doc(941,24,1).
txt_doc(253,24,2).
txt_doc(441,24,1).
txt_doc(80,24,1).
txt_doc(796,24,1).
txt_doc(630,24,1).
txt_doc(766,24,1).
txt_doc(481,24,1).
txt_doc(683,24,1).
txt_doc(560,24,1).
txt_doc(511,24,1).
txt_doc(68,24,1).
txt_doc(681,24,1).
txt_doc(428,24,1).
txt_doc(884,24,1).
txt_doc(893,24,1).
txt_doc(123,24,1).
txt_doc(151,24,1).
txt_doc(965,24,1).
txt_doc(823,24,2).
txt_doc(239,24,1).
txt_doc(12,24,1).
txt_doc(689,24,1).
txt_doc(347,24,1).
txt_doc(651,24,1).
txt_doc(862,24,1).
txt_doc(573,24,2).
txt_doc(544,24,1).
txt_doc(71,24,1).
txt_doc(160,24,1).
txt_doc(557,24,1).
txt_doc(133,24,1).
txt_doc(127,24,1).
txt_doc(806,24,1).
txt_doc(79,24,1).
txt_doc(371,24,1).
txt_doc(767,24,1).
txt_doc(841,24,2).
txt_doc(11,24,1).
txt_doc(319,24,1).
txt_doc(665,24,1).
txt_doc(392,24,1).
txt_doc(564,24,1).
txt_doc(21,24,1).
txt_doc(892,24,1).
txt_doc(456,24,2).
txt_doc(837,24,1).
txt_doc(286,24,2).
txt_doc(854,24,1).
txt_doc(900,24,1).
txt_doc(810,24,1).
txt_doc(165,24,1).
txt_doc(789,24,1).
txt_doc(551,24,1).
txt_doc(109,24,1).
txt_doc(48,24,2).
txt_doc(310,24,1).
txt_doc(426,24,1).
txt_doc(194,24,1).
txt_doc(256,25,1).
txt_doc(955,25,1).
txt_doc(834,25,1).
txt_doc(847,25,1).
txt_doc(542,25,1).
txt_doc(239,25,1).
txt_doc(934,25,1).
txt_doc(222,25,1).
txt_doc(81,25,1).
txt_doc(44,25,1).
txt_doc(823,25,1).
txt_doc(152,25,1).
txt_doc(237,25,1).
txt_doc(348,25,1).
txt_doc(427,25,1).
txt_doc(683,25,1).
txt_doc(826,25,1).
txt_doc(284,25,1).
txt_doc(907,25,1).
txt_doc(836,25,1).
txt_doc(648,25,1).
txt_doc(161,25,1).
txt_doc(816,25,1).
txt_doc(371,25,1).
txt_doc(650,25,1).
txt_doc(711,25,1).
txt_doc(827,25,1).
txt_doc(718,25,1).
txt_doc(742,25,1).
txt_doc(773,25,1).
txt_doc(838,25,2).
txt_doc(459,25,1).
txt_doc(314,25,1).
txt_doc(617,25,1).
txt_doc(150,25,1).
txt_doc(963,25,1).
txt_doc(87,25,1).
txt_doc(480,25,1).
txt_doc(163,25,1).
txt_doc(479,25,1).
txt_doc(956,25,1).
txt_doc(369,25,1).
txt_doc(486,25,1).
txt_doc(771,25,1).
txt_doc(267,25,1).
txt_doc(780,25,1).
txt_doc(928,25,1).
txt_doc(503,25,1).
txt_doc(3,25,1).
txt_doc(556,25,1).
txt_doc(515,25,1).
txt_doc(272,25,1).
txt_doc(332,25,1).
txt_doc(966,25,1).
txt_doc(305,25,1).
txt_doc(801,25,1).
txt_doc(762,25,1).
txt_doc(909,25,1).
txt_doc(726,25,1).
txt_doc(395,25,1).
txt_doc(497,25,2).
txt_doc(63,25,1).
txt_doc(28,25,2).
txt_doc(942,25,1).
txt_doc(511,25,1).
txt_doc(388,25,1).
txt_doc(814,25,1).
txt_doc(761,25,1).
txt_doc(915,25,1).
txt_doc(401,25,1).
txt_doc(77,25,1).
txt_doc(311,25,1).
txt_doc(16,25,1).
txt_doc(673,25,1).
txt_doc(699,25,1).
txt_doc(317,25,1).
txt_doc(338,25,1).
txt_doc(96,25,1).
txt_doc(241,25,1).
txt_doc(848,25,1).
txt_doc(695,25,1).
txt_doc(281,25,1).
txt_doc(656,25,1).
txt_doc(793,25,1).
txt_doc(694,25,2).
txt_doc(40,25,1).
txt_doc(829,25,1).
txt_doc(483,25,1).
txt_doc(98,25,1).
txt_doc(331,25,1).
txt_doc(873,25,1).
txt_doc(193,25,1).
txt_doc(471,25,1).
txt_doc(248,25,1).
txt_doc(137,25,1).
txt_doc(784,25,1).
txt_doc(625,25,1).
txt_doc(282,25,1).
txt_doc(153,25,2).
txt_doc(240,25,1).
txt_doc(562,25,1).
txt_doc(661,25,1).
txt_doc(265,25,1).
txt_doc(817,25,1).
txt_doc(835,25,1).
txt_doc(710,25,1).
txt_doc(517,25,1).
txt_doc(722,25,1).
txt_doc(321,25,1).
txt_doc(921,25,1).
txt_doc(704,25,1).
txt_doc(233,25,1).
txt_doc(560,25,1).
txt_doc(967,25,1).
txt_doc(228,25,1).
txt_doc(468,25,2).
txt_doc(362,25,1).
txt_doc(952,25,1).
txt_doc(507,25,1).
txt_doc(143,25,1).
txt_doc(303,25,1).
txt_doc(895,25,1).
txt_doc(788,25,1).
txt_doc(756,25,1).
txt_doc(293,25,2).
txt_doc(566,25,1).
txt_doc(905,25,1).
txt_doc(785,25,1).
txt_doc(600,25,1).
txt_doc(488,25,2).
txt_doc(472,25,1).
txt_doc(175,25,3).
txt_doc(114,25,1).
txt_doc(712,25,1).
txt_doc(261,25,1).
txt_doc(302,25,1).
txt_doc(445,25,1).
txt_doc(525,25,1).
txt_doc(403,25,1).
txt_doc(528,25,1).
txt_doc(957,25,1).
txt_doc(737,25,1).
txt_doc(24,25,1).
txt_doc(162,25,1).
txt_doc(92,25,1).
txt_doc(270,25,1).
txt_doc(58,25,3).
txt_doc(408,25,2).
txt_doc(938,25,2).
txt_doc(769,25,1).
txt_doc(6,25,1).
txt_doc(60,25,1).
txt_doc(223,25,2).
txt_doc(195,25,2).
txt_doc(296,25,1).
txt_doc(764,25,1).
txt_doc(295,25,2).
txt_doc(126,25,1).
txt_doc(398,25,2).
txt_doc(475,25,1).
txt_doc(394,25,1).
txt_doc(372,25,1).
txt_doc(751,25,1).
txt_doc(414,25,1).
txt_doc(937,25,1).
txt_doc(790,25,1).
txt_doc(758,25,1).
txt_doc(82,25,1).
txt_doc(71,25,1).
txt_doc(7,25,2).
txt_doc(210,25,1).
txt_doc(319,25,1).
txt_doc(147,25,2).
txt_doc(642,25,1).
txt_doc(51,25,1).
txt_doc(274,25,1).
txt_doc(447,25,1).
txt_doc(846,25,3).
txt_doc(710,26,1).
txt_doc(664,26,1).
txt_doc(563,26,1).
txt_doc(557,26,1).
txt_doc(115,26,1).
txt_doc(393,26,1).
txt_doc(184,26,1).
txt_doc(414,26,1).
txt_doc(559,26,1).
txt_doc(789,26,1).
txt_doc(368,26,1).
txt_doc(640,26,1).
txt_doc(600,26,1).
txt_doc(904,26,1).
txt_doc(463,26,1).
txt_doc(907,26,2).
txt_doc(208,26,1).
txt_doc(788,26,1).
txt_doc(345,26,1).
txt_doc(249,26,1).
txt_doc(289,26,1).
txt_doc(94,26,1).
txt_doc(62,26,1).
txt_doc(657,26,1).
txt_doc(575,26,1).
txt_doc(453,26,1).
txt_doc(180,26,1).
txt_doc(164,26,1).
txt_doc(643,26,1).
txt_doc(425,26,1).
txt_doc(745,26,1).
txt_doc(878,26,1).
txt_doc(852,26,1).
txt_doc(633,26,1).
txt_doc(925,26,1).
txt_doc(498,26,1).
txt_doc(207,26,1).
txt_doc(293,26,1).
txt_doc(121,26,1).
txt_doc(131,26,1).
txt_doc(692,26,1).
txt_doc(327,26,1).
txt_doc(364,26,1).
txt_doc(684,26,1).
txt_doc(594,26,1).
txt_doc(709,26,1).
txt_doc(61,26,1).
txt_doc(42,26,1).
txt_doc(256,26,1).
txt_doc(268,26,1).
txt_doc(820,26,1).
txt_doc(424,26,1).
txt_doc(635,26,2).
txt_doc(292,26,1).
txt_doc(966,26,1).
txt_doc(147,26,2).
txt_doc(870,26,1).
txt_doc(31,26,2).
txt_doc(66,26,1).
txt_doc(480,26,1).
txt_doc(929,26,1).
txt_doc(258,26,1).
txt_doc(687,26,1).
txt_doc(628,26,1).
txt_doc(130,26,1).
txt_doc(236,26,2).
txt_doc(519,26,2).
txt_doc(810,26,1).
txt_doc(637,26,1).
txt_doc(168,26,1).
txt_doc(429,26,1).
txt_doc(797,26,2).
txt_doc(329,26,1).
txt_doc(889,26,1).
txt_doc(397,26,1).
txt_doc(3,26,1).
txt_doc(618,26,1).
txt_doc(809,26,1).
txt_doc(316,26,1).
txt_doc(917,26,2).
txt_doc(567,26,1).
txt_doc(503,26,1).
txt_doc(344,26,1).
txt_doc(5,26,1).
txt_doc(585,26,2).
txt_doc(795,26,1).
txt_doc(695,26,1).
txt_doc(858,26,1).
txt_doc(655,26,1).
txt_doc(719,26,1).
txt_doc(456,26,1).
txt_doc(171,26,1).
txt_doc(796,26,1).
txt_doc(840,26,1).
txt_doc(851,26,1).
txt_doc(355,26,1).
txt_doc(252,26,2).
txt_doc(8,26,1).
txt_doc(779,26,2).
txt_doc(814,26,1).
txt_doc(376,26,1).
txt_doc(71,26,1).
txt_doc(808,26,1).
txt_doc(233,26,1).
txt_doc(70,26,1).
txt_doc(288,26,1).
txt_doc(816,26,2).
txt_doc(595,26,2).
txt_doc(65,26,1).
txt_doc(758,26,1).
txt_doc(873,26,1).
txt_doc(307,26,1).
txt_doc(213,26,1).
txt_doc(419,26,1).
txt_doc(51,26,1).
txt_doc(276,26,1).
txt_doc(714,26,1).
txt_doc(257,26,1).
txt_doc(793,26,1).
txt_doc(751,26,1).
txt_doc(909,26,1).
txt_doc(47,26,1).
txt_doc(264,26,1).
txt_doc(417,26,1).
txt_doc(787,26,2).
txt_doc(58,26,1).
txt_doc(630,26,1).
txt_doc(433,26,2).
txt_doc(153,26,1).
txt_doc(241,26,1).
txt_doc(504,26,1).
txt_doc(170,26,1).
txt_doc(72,26,1).
txt_doc(822,26,2).
txt_doc(724,26,2).
txt_doc(328,26,1).
txt_doc(764,26,1).
txt_doc(868,26,1).
txt_doc(672,26,2).
txt_doc(960,26,1).
txt_doc(349,26,1).
txt_doc(783,26,1).
txt_doc(555,26,2).
txt_doc(548,26,1).
txt_doc(752,26,1).
txt_doc(360,26,1).
txt_doc(942,26,1).
txt_doc(418,26,1).
txt_doc(601,26,2).
txt_doc(193,26,2).
txt_doc(118,26,2).
txt_doc(895,26,1).
txt_doc(524,26,1).
txt_doc(916,26,1).
txt_doc(666,26,2).
txt_doc(670,26,1).
txt_doc(240,26,1).
txt_doc(7,26,1).
txt_doc(801,26,1).
txt_doc(690,26,1).
txt_doc(573,26,1).
txt_doc(198,26,2).
txt_doc(689,26,1).
txt_doc(18,26,1).
txt_doc(777,26,1).
txt_doc(372,26,1).
txt_doc(665,26,2).
txt_doc(454,26,2).
txt_doc(888,26,1).
txt_doc(354,26,2).
txt_doc(285,26,1).
txt_doc(908,26,1).
txt_doc(16,26,1).
txt_doc(514,27,1).
txt_doc(583,27,1).
txt_doc(705,27,1).
txt_doc(893,27,1).
txt_doc(906,27,1).
txt_doc(187,27,1).
txt_doc(722,27,1).
txt_doc(857,27,1).
txt_doc(231,27,1).
txt_doc(653,27,1).
txt_doc(374,27,1).
txt_doc(264,27,1).
txt_doc(829,27,1).
txt_doc(122,27,1).
txt_doc(832,27,1).
txt_doc(795,27,1).
txt_doc(883,27,1).
txt_doc(915,27,1).
txt_doc(68,27,1).
txt_doc(914,27,1).
txt_doc(798,27,1).
txt_doc(120,27,1).
txt_doc(555,27,1).
txt_doc(195,27,1).
txt_doc(101,27,1).
txt_doc(929,27,1).
txt_doc(925,27,1).
txt_doc(843,27,1).
txt_doc(501,27,1).
txt_doc(317,27,1).
txt_doc(14,27,1).
txt_doc(866,27,1).
txt_doc(157,27,1).
txt_doc(429,27,1).
txt_doc(837,27,1).
txt_doc(534,27,1).
txt_doc(362,27,1).
txt_doc(649,27,1).
txt_doc(283,27,1).
txt_doc(592,27,1).
txt_doc(18,27,1).
txt_doc(693,27,1).
txt_doc(348,27,1).
txt_doc(840,27,1).
txt_doc(458,27,1).
txt_doc(224,27,1).
txt_doc(845,27,1).
txt_doc(199,27,1).
txt_doc(402,27,1).
txt_doc(485,27,1).
txt_doc(35,27,1).
txt_doc(519,27,1).
txt_doc(42,27,1).
txt_doc(436,27,1).
txt_doc(153,27,1).
txt_doc(349,27,1).
txt_doc(178,27,1).
txt_doc(84,27,1).
txt_doc(757,27,1).
txt_doc(89,27,1).
txt_doc(613,27,1).
txt_doc(373,27,1).
txt_doc(303,27,1).
txt_doc(695,27,1).
txt_doc(818,27,1).
txt_doc(476,27,1).
txt_doc(523,27,1).
txt_doc(928,27,1).
txt_doc(158,27,1).
txt_doc(279,27,1).
txt_doc(730,27,1).
txt_doc(140,27,1).
txt_doc(188,27,1).
txt_doc(381,27,1).
txt_doc(606,27,1).
txt_doc(654,27,1).
txt_doc(105,27,1).
txt_doc(411,27,1).
txt_doc(943,27,1).
txt_doc(209,27,1).
txt_doc(79,27,1).
txt_doc(400,27,1).
txt_doc(636,27,1).
txt_doc(713,27,1).
txt_doc(909,27,1).
txt_doc(672,27,1).
txt_doc(935,27,1).
txt_doc(902,27,1).
txt_doc(251,27,1).
txt_doc(528,27,1).
txt_doc(968,27,1).
txt_doc(900,27,1).
txt_doc(564,27,1).
txt_doc(694,27,1).
txt_doc(728,27,1).
txt_doc(673,27,1).
txt_doc(1,27,1).
txt_doc(927,27,2).
txt_doc(604,27,2).
txt_doc(858,27,1).
txt_doc(792,27,1).
txt_doc(217,27,1).
txt_doc(78,27,1).
txt_doc(807,27,1).
txt_doc(110,27,1).
txt_doc(956,27,1).
txt_doc(443,27,1).
txt_doc(285,27,1).
txt_doc(237,27,2).
txt_doc(665,27,1).
txt_doc(723,27,1).
txt_doc(88,27,1).
txt_doc(689,27,1).
txt_doc(16,27,1).
txt_doc(631,27,1).
txt_doc(816,27,1).
txt_doc(496,27,1).
txt_doc(319,27,2).
txt_doc(544,27,1).
txt_doc(24,27,1).
txt_doc(182,27,1).
txt_doc(320,27,2).
txt_doc(467,27,1).
txt_doc(778,27,1).
txt_doc(785,27,2).
txt_doc(920,27,2).
txt_doc(186,27,1).
txt_doc(690,27,1).
txt_doc(173,27,1).
txt_doc(946,27,1).
txt_doc(174,27,1).
txt_doc(729,27,1).
txt_doc(47,27,1).
txt_doc(67,27,1).
txt_doc(235,27,1).
txt_doc(568,27,1).
txt_doc(797,27,2).
txt_doc(11,27,1).
txt_doc(292,27,1).
txt_doc(456,27,1).
txt_doc(139,27,1).
txt_doc(447,27,1).
txt_doc(38,27,1).
txt_doc(307,27,1).
txt_doc(576,27,1).
txt_doc(680,27,1).
txt_doc(95,27,1).
txt_doc(921,27,1).
txt_doc(879,27,1).
txt_doc(821,27,1).
txt_doc(430,27,1).
txt_doc(940,27,1).
txt_doc(90,27,1).
txt_doc(486,27,2).
txt_doc(744,27,2).
txt_doc(409,27,2).
txt_doc(753,27,1).
txt_doc(905,27,1).
txt_doc(952,27,1).
txt_doc(254,27,1).
txt_doc(870,27,1).
txt_doc(404,27,1).
txt_doc(626,27,1).
txt_doc(911,27,1).
txt_doc(19,27,1).
txt_doc(385,27,1).
txt_doc(205,27,1).
txt_doc(741,27,2).
txt_doc(802,27,1).
txt_doc(453,27,1).
txt_doc(799,27,1).
txt_doc(183,27,2).
txt_doc(321,27,1).
txt_doc(218,27,2).
txt_doc(27,27,1).
txt_doc(681,27,1).
txt_doc(281,27,2).
txt_doc(420,27,1).
txt_doc(761,27,1).
txt_doc(691,27,1).
txt_doc(462,27,1).
txt_doc(408,27,1).
txt_doc(382,27,1).
txt_doc(190,27,1).
txt_doc(262,27,1).
txt_doc(10,28,1).
txt_doc(924,28,1).
txt_doc(188,28,1).
txt_doc(275,28,1).
txt_doc(481,28,2).
txt_doc(513,28,1).
txt_doc(172,28,1).
txt_doc(450,28,1).
txt_doc(945,28,1).
txt_doc(955,28,1).
txt_doc(937,28,1).
txt_doc(278,28,1).
txt_doc(332,28,1).
txt_doc(453,28,1).
txt_doc(434,28,1).
txt_doc(962,28,1).
txt_doc(240,28,1).
txt_doc(657,28,1).
txt_doc(627,28,1).
txt_doc(432,28,1).
txt_doc(218,28,1).
txt_doc(699,28,1).
txt_doc(11,28,1).
txt_doc(354,28,1).
txt_doc(245,28,1).
txt_doc(492,28,1).
txt_doc(297,28,1).
txt_doc(205,28,1).
txt_doc(966,28,1).
txt_doc(204,28,1).
txt_doc(718,28,1).
txt_doc(742,28,1).
txt_doc(372,28,1).
txt_doc(228,28,1).
txt_doc(719,28,1).
txt_doc(958,28,1).
txt_doc(364,28,1).
txt_doc(112,28,1).
txt_doc(89,28,1).
txt_doc(891,28,1).
txt_doc(340,28,1).
txt_doc(809,28,1).
txt_doc(193,28,1).
txt_doc(944,28,1).
txt_doc(191,28,1).
txt_doc(209,28,1).
txt_doc(216,28,1).
txt_doc(523,28,1).
txt_doc(633,28,1).
txt_doc(315,28,1).
txt_doc(664,28,1).
txt_doc(685,28,1).
txt_doc(310,28,1).
txt_doc(474,28,1).
txt_doc(194,28,1).
txt_doc(764,28,1).
txt_doc(715,28,1).
txt_doc(208,28,1).
txt_doc(255,28,1).
txt_doc(106,28,1).
txt_doc(578,28,1).
txt_doc(367,28,1).
txt_doc(170,28,1).
txt_doc(151,28,1).
txt_doc(817,28,1).
txt_doc(324,28,1).
txt_doc(505,28,1).
txt_doc(621,28,1).
txt_doc(380,28,1).
txt_doc(816,28,1).
txt_doc(196,28,1).
txt_doc(146,28,1).
txt_doc(2,28,1).
txt_doc(274,28,1).
txt_doc(838,28,1).
txt_doc(757,28,1).
txt_doc(14,28,1).
txt_doc(694,28,1).
txt_doc(855,28,1).
txt_doc(545,28,1).
txt_doc(635,28,1).
txt_doc(466,28,1).
txt_doc(357,28,1).
txt_doc(684,28,1).
txt_doc(739,28,1).
txt_doc(273,28,1).
txt_doc(222,28,1).
txt_doc(507,28,1).
txt_doc(533,28,1).
txt_doc(921,28,1).
txt_doc(659,28,1).
txt_doc(821,28,2).
txt_doc(781,28,1).
txt_doc(500,28,1).
txt_doc(391,28,1).
txt_doc(799,28,1).
txt_doc(27,28,1).
txt_doc(646,28,1).
txt_doc(13,28,1).
txt_doc(728,28,1).
txt_doc(431,28,1).
txt_doc(253,28,1).
txt_doc(345,28,1).
txt_doc(814,28,1).
txt_doc(611,28,1).
txt_doc(307,28,1).
txt_doc(251,28,1).
txt_doc(326,28,2).
txt_doc(733,28,2).
txt_doc(644,28,1).
txt_doc(798,28,1).
txt_doc(203,28,1).
txt_doc(270,28,1).
txt_doc(497,28,1).
txt_doc(530,28,1).
txt_doc(32,28,1).
txt_doc(615,28,1).
txt_doc(743,28,1).
txt_doc(287,28,1).
txt_doc(184,28,1).
txt_doc(154,28,1).
txt_doc(723,28,1).
txt_doc(265,28,2).
txt_doc(158,28,1).
txt_doc(263,28,1).
txt_doc(261,28,1).
txt_doc(306,28,1).
txt_doc(336,28,1).
txt_doc(508,28,1).
txt_doc(252,28,1).
txt_doc(941,28,1).
txt_doc(691,28,1).
txt_doc(825,28,2).
txt_doc(895,28,1).
txt_doc(835,28,1).
txt_doc(549,28,1).
txt_doc(51,28,1).
txt_doc(296,28,1).
txt_doc(215,28,1).
txt_doc(546,28,1).
txt_doc(227,28,2).
txt_doc(444,28,1).
txt_doc(235,28,1).
txt_doc(951,28,1).
txt_doc(737,28,1).
txt_doc(724,28,1).
txt_doc(563,28,1).
txt_doc(119,28,1).
txt_doc(493,28,1).
txt_doc(560,28,2).
txt_doc(260,28,2).
txt_doc(165,28,1).
txt_doc(632,28,1).
txt_doc(385,28,2).
txt_doc(3,28,1).
txt_doc(392,28,1).
txt_doc(801,28,1).
txt_doc(477,28,1).
txt_doc(696,28,1).
txt_doc(847,28,1).
txt_doc(556,28,1).
txt_doc(277,28,2).
txt_doc(518,28,1).
txt_doc(712,28,1).
txt_doc(879,28,1).
txt_doc(501,28,1).
txt_doc(705,28,1).
txt_doc(173,28,1).
txt_doc(281,28,1).
txt_doc(903,28,2).
txt_doc(397,28,1).
txt_doc(779,28,1).
txt_doc(225,28,2).
txt_doc(454,28,1).
txt_doc(183,28,1).
txt_doc(437,28,1).
txt_doc(759,28,2).
txt_doc(565,28,1).
txt_doc(420,28,1).
txt_doc(609,28,1).
txt_doc(697,28,1).
txt_doc(409,28,1).
txt_doc(509,28,1).
txt_doc(589,28,2).
txt_doc(748,28,1).
txt_doc(751,29,1).
txt_doc(459,29,1).
txt_doc(341,29,1).
txt_doc(162,29,1).
txt_doc(78,29,1).
txt_doc(855,29,1).
txt_doc(581,29,1).
txt_doc(843,29,1).
txt_doc(762,29,1).
txt_doc(869,29,1).
txt_doc(110,29,2).
txt_doc(437,29,1).
txt_doc(369,29,1).
txt_doc(158,29,1).
txt_doc(726,29,1).
txt_doc(196,29,1).
txt_doc(326,29,1).
txt_doc(793,29,1).
txt_doc(207,29,1).
txt_doc(619,29,1).
txt_doc(771,29,1).
txt_doc(7,29,1).
txt_doc(499,29,1).
txt_doc(83,29,1).
txt_doc(392,29,1).
txt_doc(440,29,1).
txt_doc(321,29,1).
txt_doc(463,29,1).
txt_doc(507,29,1).
txt_doc(80,29,1).
txt_doc(339,29,1).
txt_doc(447,29,2).
txt_doc(590,29,1).
txt_doc(59,29,1).
txt_doc(163,29,1).
txt_doc(378,29,1).
txt_doc(381,29,1).
txt_doc(479,29,1).
txt_doc(545,29,3).
txt_doc(404,29,2).
txt_doc(562,29,1).
txt_doc(876,29,1).
txt_doc(528,29,1).
txt_doc(296,29,1).
txt_doc(161,29,1).
txt_doc(251,29,1).
txt_doc(403,29,1).
txt_doc(240,29,1).
txt_doc(267,29,3).
txt_doc(236,29,1).
txt_doc(589,29,1).
txt_doc(820,29,1).
txt_doc(789,29,1).
txt_doc(218,29,1).
txt_doc(914,29,1).
txt_doc(142,29,1).
txt_doc(710,29,1).
txt_doc(478,29,1).
txt_doc(379,29,1).
txt_doc(672,29,1).
txt_doc(921,29,1).
txt_doc(637,29,1).
txt_doc(797,29,1).
txt_doc(570,29,1).
txt_doc(118,29,1).
txt_doc(308,29,1).
txt_doc(474,29,1).
txt_doc(475,29,1).
txt_doc(823,29,1).
txt_doc(615,29,1).
txt_doc(657,29,1).
txt_doc(723,29,1).
txt_doc(70,29,1).
txt_doc(469,29,1).
txt_doc(898,29,1).
txt_doc(224,29,1).
txt_doc(794,29,1).
txt_doc(537,29,2).
txt_doc(735,29,1).
txt_doc(148,29,1).
txt_doc(826,29,1).
txt_doc(805,29,1).
txt_doc(542,29,1).
txt_doc(960,29,2).
txt_doc(310,29,1).
txt_doc(8,29,1).
txt_doc(316,29,1).
txt_doc(108,29,1).
txt_doc(269,29,1).
txt_doc(364,29,1).
txt_doc(355,29,2).
txt_doc(667,29,1).
txt_doc(525,29,2).
txt_doc(349,29,1).
txt_doc(174,29,1).
txt_doc(73,29,1).
txt_doc(271,29,1).
txt_doc(18,29,1).
txt_doc(209,29,1).
txt_doc(333,29,1).
txt_doc(307,29,1).
txt_doc(241,29,1).
txt_doc(143,29,1).
txt_doc(452,29,1).
txt_doc(93,29,1).
txt_doc(957,29,1).
txt_doc(761,29,1).
txt_doc(125,29,1).
txt_doc(524,29,1).
txt_doc(155,29,1).
txt_doc(946,29,1).
txt_doc(454,29,1).
txt_doc(748,29,1).
txt_doc(306,29,1).
txt_doc(446,29,1).
txt_doc(664,29,1).
txt_doc(491,29,1).
txt_doc(550,29,1).
txt_doc(830,29,1).
txt_doc(498,29,1).
txt_doc(188,29,1).
txt_doc(816,29,1).
txt_doc(967,29,1).
txt_doc(230,29,3).
txt_doc(741,29,1).
txt_doc(92,29,1).
txt_doc(573,29,1).
txt_doc(916,29,2).
txt_doc(39,29,1).
txt_doc(356,29,1).
txt_doc(669,29,1).
txt_doc(554,29,1).
txt_doc(20,29,1).
txt_doc(284,29,1).
txt_doc(450,29,1).
txt_doc(824,29,2).
txt_doc(658,29,1).
txt_doc(686,29,1).
txt_doc(46,29,1).
txt_doc(968,29,2).
txt_doc(150,29,1).
txt_doc(948,29,1).
txt_doc(68,29,1).
txt_doc(595,29,2).
txt_doc(239,29,2).
txt_doc(765,29,1).
txt_doc(276,29,2).
txt_doc(935,29,1).
txt_doc(737,29,1).
txt_doc(945,29,2).
txt_doc(146,29,1).
txt_doc(839,29,2).
txt_doc(925,29,1).
txt_doc(958,29,2).
txt_doc(250,29,1).
txt_doc(24,29,1).
txt_doc(468,29,2).
txt_doc(653,29,1).
txt_doc(242,29,1).
txt_doc(662,29,1).
txt_doc(822,29,1).
txt_doc(850,29,1).
txt_doc(220,29,1).
txt_doc(277,29,1).
txt_doc(467,29,1).
txt_doc(443,29,1).
txt_doc(109,29,1).
txt_doc(4,29,2).
txt_doc(886,29,1).
txt_doc(566,29,1).
txt_doc(646,29,1).
txt_doc(34,29,1).
txt_doc(391,29,1).
txt_doc(495,29,1).
txt_doc(71,29,1).
txt_doc(877,29,1).
txt_doc(573,30,1).
txt_doc(680,30,1).
txt_doc(541,30,1).
txt_doc(550,30,1).
txt_doc(752,30,1).
txt_doc(156,30,1).
txt_doc(940,30,1).
txt_doc(105,30,1).
txt_doc(204,30,1).
txt_doc(329,30,1).
txt_doc(471,30,1).
txt_doc(7,30,1).
txt_doc(865,30,1).
txt_doc(51,30,1).
txt_doc(622,30,1).
txt_doc(222,30,1).
txt_doc(920,30,1).
txt_doc(114,30,1).
txt_doc(53,30,1).
txt_doc(235,30,1).
txt_doc(216,30,1).
txt_doc(453,30,1).
txt_doc(787,30,1).
txt_doc(30,30,1).
txt_doc(134,30,1).
txt_doc(804,30,1).
txt_doc(345,30,1).
txt_doc(175,30,1).
txt_doc(667,30,1).
txt_doc(168,30,1).
txt_doc(337,30,1).
txt_doc(556,30,1).
txt_doc(927,30,1).
txt_doc(685,30,1).
txt_doc(334,30,1).
txt_doc(777,30,1).
txt_doc(612,30,1).
txt_doc(919,30,1).
txt_doc(283,30,1).
txt_doc(143,30,1).
txt_doc(88,30,1).
txt_doc(254,30,1).
txt_doc(709,30,1).
txt_doc(544,30,1).
txt_doc(36,30,1).
txt_doc(240,30,1).
txt_doc(320,30,1).
txt_doc(196,30,1).
txt_doc(285,30,1).
txt_doc(638,30,1).
txt_doc(703,30,1).
txt_doc(294,30,1).
txt_doc(882,30,2).
txt_doc(657,30,1).
txt_doc(202,30,1).
txt_doc(440,30,1).
txt_doc(344,30,1).
txt_doc(366,30,1).
txt_doc(526,30,1).
txt_doc(720,30,1).
txt_doc(450,30,1).
txt_doc(198,30,1).
txt_doc(751,30,1).
txt_doc(530,30,1).
txt_doc(173,30,1).
txt_doc(847,30,1).
txt_doc(477,30,1).
txt_doc(589,30,1).
txt_doc(65,30,1).
txt_doc(153,30,1).
txt_doc(253,30,1).
txt_doc(539,30,1).
txt_doc(287,30,1).
txt_doc(637,30,1).
txt_doc(83,30,1).
txt_doc(640,30,1).
txt_doc(548,30,1).
txt_doc(500,30,1).
txt_doc(872,30,1).
txt_doc(56,30,1).
txt_doc(518,30,1).
txt_doc(690,30,2).
txt_doc(194,30,1).
txt_doc(444,30,2).
txt_doc(217,30,1).
txt_doc(860,30,1).
txt_doc(472,30,1).
txt_doc(576,30,1).
txt_doc(184,30,1).
txt_doc(213,30,1).
txt_doc(385,30,1).
txt_doc(912,30,1).
txt_doc(520,30,1).
txt_doc(843,30,1).
txt_doc(21,30,1).
txt_doc(489,30,1).
txt_doc(77,30,1).
txt_doc(431,30,1).
txt_doc(187,30,2).
txt_doc(78,30,1).
txt_doc(243,30,1).
txt_doc(536,30,1).
txt_doc(359,30,1).
txt_doc(238,30,1).
txt_doc(16,30,1).
txt_doc(412,30,1).
txt_doc(434,30,1).
txt_doc(891,30,1).
txt_doc(938,30,1).
txt_doc(503,30,2).
txt_doc(505,30,1).
txt_doc(333,30,1).
txt_doc(376,30,2).
txt_doc(99,30,1).
txt_doc(897,30,1).
txt_doc(402,30,1).
txt_doc(201,30,2).
txt_doc(729,30,1).
txt_doc(15,30,1).
txt_doc(470,30,1).
txt_doc(716,30,2).
txt_doc(82,30,1).
txt_doc(837,30,1).
txt_doc(394,30,1).
txt_doc(620,30,1).
txt_doc(215,30,1).
txt_doc(342,30,1).
txt_doc(684,30,1).
txt_doc(538,30,1).
txt_doc(513,30,1).
txt_doc(252,30,1).
txt_doc(179,30,1).
txt_doc(122,30,1).
txt_doc(524,30,1).
txt_doc(542,30,1).
txt_doc(148,30,1).
txt_doc(894,30,1).
txt_doc(611,30,1).
txt_doc(775,30,1).
txt_doc(809,30,1).
txt_doc(692,30,1).
txt_doc(878,30,1).
txt_doc(390,30,1).
txt_doc(61,30,1).
txt_doc(631,30,1).
txt_doc(43,30,1).
txt_doc(832,30,1).
txt_doc(947,30,1).
txt_doc(322,30,1).
txt_doc(688,30,1).
txt_doc(705,30,1).
txt_doc(666,30,1).
txt_doc(674,30,1).
txt_doc(881,30,1).
txt_doc(678,30,1).
txt_doc(166,30,3).
txt_doc(765,30,1).
txt_doc(55,30,1).
txt_doc(97,30,1).
txt_doc(764,30,2).
txt_doc(958,30,1).
txt_doc(115,30,1).
txt_doc(339,30,1).
txt_doc(17,30,2).
txt_doc(255,30,1).
txt_doc(494,30,1).
txt_doc(522,30,1).
txt_doc(284,30,1).
txt_doc(174,30,1).
txt_doc(310,30,1).
txt_doc(89,30,1).
txt_doc(521,30,1).
txt_doc(338,30,1).
txt_doc(350,30,1).
txt_doc(365,30,1).
txt_doc(655,30,1).
txt_doc(381,30,1).
txt_doc(903,30,1).
txt_doc(896,30,1).
txt_doc(634,30,1).
txt_doc(698,30,1).
txt_doc(324,30,2).
txt_doc(101,30,1).
txt_doc(483,30,1).
txt_doc(535,30,1).
txt_doc(848,30,2).
txt_doc(454,31,1).
txt_doc(671,31,1).
txt_doc(115,31,1).
txt_doc(903,31,1).
txt_doc(730,31,1).
txt_doc(456,31,1).
txt_doc(953,31,1).
txt_doc(594,31,1).
txt_doc(809,31,1).
txt_doc(960,31,1).
txt_doc(759,31,1).
txt_doc(52,31,1).
txt_doc(496,31,1).
txt_doc(916,31,1).
txt_doc(164,31,1).
txt_doc(915,31,1).
txt_doc(197,31,1).
txt_doc(416,31,1).
txt_doc(945,31,1).
txt_doc(497,31,1).
txt_doc(315,31,1).
txt_doc(105,31,1).
txt_doc(663,31,1).
txt_doc(32,31,1).
txt_doc(500,31,1).
txt_doc(947,31,1).
txt_doc(776,31,1).
txt_doc(832,31,1).
txt_doc(488,31,1).
txt_doc(318,31,1).
txt_doc(184,31,1).
txt_doc(50,31,1).
txt_doc(241,31,1).
txt_doc(165,31,1).
txt_doc(352,31,1).
txt_doc(408,31,1).
txt_doc(354,31,1).
txt_doc(116,31,1).
txt_doc(644,31,1).
txt_doc(665,31,1).
txt_doc(949,31,1).
txt_doc(790,31,1).
txt_doc(309,31,1).
txt_doc(147,31,1).
txt_doc(564,31,1).
txt_doc(688,31,1).
txt_doc(542,31,1).
txt_doc(252,31,1).
txt_doc(273,31,1).
txt_doc(358,31,1).
txt_doc(967,31,1).
txt_doc(645,31,1).
txt_doc(966,31,1).
txt_doc(745,31,1).
txt_doc(857,31,1).
txt_doc(297,31,1).
txt_doc(224,31,1).
txt_doc(894,31,3).
txt_doc(340,31,1).
txt_doc(558,31,1).
txt_doc(101,31,2).
txt_doc(868,31,1).
txt_doc(754,31,1).
txt_doc(431,31,1).
txt_doc(259,31,1).
txt_doc(704,31,1).
txt_doc(4,31,2).
txt_doc(578,31,1).
txt_doc(138,31,1).
txt_doc(942,31,1).
txt_doc(494,31,1).
txt_doc(718,31,1).
txt_doc(445,31,1).
txt_doc(664,31,1).
txt_doc(614,31,1).
txt_doc(478,31,1).
txt_doc(136,31,1).
txt_doc(937,31,1).
txt_doc(257,31,1).
txt_doc(604,31,2).
txt_doc(394,31,2).
txt_doc(176,31,1).
txt_doc(134,31,1).
txt_doc(778,31,1).
txt_doc(950,31,1).
txt_doc(386,31,1).
txt_doc(34,31,1).
txt_doc(921,31,1).
txt_doc(145,31,1).
txt_doc(434,31,1).
txt_doc(72,31,1).
txt_doc(398,31,1).
txt_doc(314,31,1).
txt_doc(820,31,1).
txt_doc(637,31,2).
txt_doc(781,31,1).
txt_doc(565,31,1).
txt_doc(290,31,1).
txt_doc(701,31,1).
txt_doc(760,31,1).
txt_doc(182,31,1).
txt_doc(140,31,1).
txt_doc(823,31,1).
txt_doc(185,31,1).
txt_doc(525,31,1).
txt_doc(289,31,1).
txt_doc(756,31,2).
txt_doc(723,31,1).
txt_doc(166,31,2).
txt_doc(24,31,1).
txt_doc(206,31,2).
txt_doc(543,31,1).
txt_doc(511,31,1).
txt_doc(897,31,1).
txt_doc(583,31,1).
txt_doc(35,31,1).
txt_doc(687,31,1).
txt_doc(651,31,1).
txt_doc(199,31,1).
txt_doc(516,31,1).
txt_doc(646,31,1).
txt_doc(605,31,1).
txt_doc(518,31,1).
txt_doc(424,31,1).
txt_doc(261,31,1).
txt_doc(383,31,1).
txt_doc(468,31,1).
txt_doc(77,31,1).
txt_doc(526,31,1).
txt_doc(464,31,1).
txt_doc(936,31,1).
txt_doc(965,31,1).
txt_doc(128,31,1).
txt_doc(171,31,1).
txt_doc(260,31,2).
txt_doc(99,31,1).
txt_doc(227,31,1).
txt_doc(245,31,1).
txt_doc(528,31,1).
txt_doc(642,31,1).
txt_doc(554,31,2).
txt_doc(304,31,1).
txt_doc(53,31,1).
txt_doc(244,31,1).
txt_doc(640,31,1).
txt_doc(441,31,2).
txt_doc(487,31,1).
txt_doc(324,31,1).
txt_doc(323,31,1).
txt_doc(650,31,1).
txt_doc(369,31,1).
txt_doc(674,31,1).
txt_doc(2,31,1).
txt_doc(287,31,1).
txt_doc(711,31,1).
txt_doc(10,31,2).
txt_doc(149,31,1).
txt_doc(632,31,2).
txt_doc(806,31,1).
txt_doc(817,31,1).
txt_doc(559,31,1).
txt_doc(944,31,1).
txt_doc(118,31,1).
txt_doc(619,31,2).
txt_doc(90,31,1).
txt_doc(721,31,1).
txt_doc(95,31,1).
txt_doc(639,31,2).
txt_doc(689,31,1).
txt_doc(765,31,1).
txt_doc(907,31,1).
txt_doc(555,31,1).
txt_doc(714,31,1).
txt_doc(872,31,1).
txt_doc(231,31,1).
txt_doc(381,31,2).
txt_doc(766,31,2).
txt_doc(243,31,1).
txt_doc(242,31,1).
txt_doc(412,31,1).
txt_doc(119,31,1).
txt_doc(850,32,1).
txt_doc(389,32,1).
txt_doc(484,32,1).
txt_doc(321,32,1).
txt_doc(386,32,1).
txt_doc(116,32,1).
txt_doc(807,32,1).
txt_doc(596,32,1).
txt_doc(17,32,1).
txt_doc(361,32,1).
txt_doc(163,32,1).
txt_doc(669,32,1).
txt_doc(507,32,1).
txt_doc(878,32,1).
txt_doc(706,32,1).
txt_doc(162,32,1).
txt_doc(448,32,1).
txt_doc(366,32,1).
txt_doc(275,32,1).
txt_doc(893,32,1).
txt_doc(747,32,1).
txt_doc(930,32,1).
txt_doc(856,32,1).
txt_doc(334,32,1).
txt_doc(738,32,1).
txt_doc(434,32,1).
txt_doc(216,32,1).
txt_doc(751,32,2).
txt_doc(339,32,1).
txt_doc(627,32,1).
txt_doc(64,32,1).
txt_doc(442,32,1).
txt_doc(342,32,1).
txt_doc(710,32,1).
txt_doc(319,32,1).
txt_doc(221,32,1).
txt_doc(127,32,1).
txt_doc(583,32,1).
txt_doc(866,32,1).
txt_doc(965,32,1).
txt_doc(800,32,1).
txt_doc(845,32,1).
txt_doc(352,32,1).
txt_doc(184,32,1).
txt_doc(946,32,1).
txt_doc(921,32,1).
txt_doc(810,32,1).
txt_doc(485,32,1).
txt_doc(784,32,1).
txt_doc(740,32,1).
txt_doc(443,32,1).
txt_doc(626,32,1).
txt_doc(3,32,1).
txt_doc(755,32,1).
txt_doc(828,32,1).
txt_doc(167,32,1).
txt_doc(289,32,1).
txt_doc(852,32,1).
txt_doc(249,32,1).
txt_doc(460,32,1).
txt_doc(791,32,1).
txt_doc(417,32,1).
txt_doc(486,32,1).
txt_doc(530,32,1).
txt_doc(528,32,1).
txt_doc(870,32,1).
txt_doc(70,32,1).
txt_doc(430,32,1).
txt_doc(181,32,1).
txt_doc(539,32,1).
txt_doc(309,32,1).
txt_doc(461,32,1).
txt_doc(325,32,1).
txt_doc(6,32,1).
txt_doc(186,32,1).
txt_doc(711,32,2).
txt_doc(579,32,1).
txt_doc(332,32,1).
txt_doc(956,32,2).
txt_doc(882,32,1).
txt_doc(824,32,1).
txt_doc(536,32,2).
txt_doc(542,32,1).
txt_doc(355,32,1).
txt_doc(297,32,1).
txt_doc(408,32,1).
txt_doc(349,32,2).
txt_doc(115,32,1).
txt_doc(809,32,1).
txt_doc(639,32,1).
txt_doc(877,32,1).
txt_doc(805,32,1).
txt_doc(405,32,1).
txt_doc(412,32,1).
txt_doc(913,32,1).
txt_doc(267,32,2).
txt_doc(585,32,1).
txt_doc(868,32,1).
txt_doc(518,32,1).
txt_doc(860,32,2).
txt_doc(880,32,1).
txt_doc(211,32,1).
txt_doc(666,32,1).
txt_doc(159,32,1).
txt_doc(801,32,1).
txt_doc(215,32,1).
txt_doc(464,32,1).
txt_doc(825,32,3).
txt_doc(1,32,1).
txt_doc(59,32,1).
txt_doc(109,32,1).
txt_doc(618,32,1).
txt_doc(557,32,1).
txt_doc(429,32,1).
txt_doc(256,32,1).
txt_doc(378,32,1).
txt_doc(502,32,1).
txt_doc(281,32,1).
txt_doc(73,32,1).
txt_doc(649,32,3).
txt_doc(646,32,1).
txt_doc(756,32,1).
txt_doc(466,32,1).
txt_doc(887,32,1).
txt_doc(648,32,1).
txt_doc(26,32,2).
txt_doc(225,32,1).
txt_doc(265,32,1).
txt_doc(258,32,1).
txt_doc(897,32,1).
txt_doc(322,32,1).
txt_doc(584,32,1).
txt_doc(629,32,1).
txt_doc(827,32,1).
txt_doc(203,32,1).
txt_doc(244,32,1).
txt_doc(928,32,2).
txt_doc(739,32,1).
txt_doc(313,32,1).
txt_doc(686,32,1).
txt_doc(268,32,1).
txt_doc(559,32,1).
txt_doc(223,32,1).
txt_doc(551,32,1).
txt_doc(493,32,1).
txt_doc(254,32,1).
txt_doc(565,32,1).
txt_doc(214,32,1).
txt_doc(889,32,1).
txt_doc(182,32,2).
txt_doc(665,32,1).
txt_doc(172,32,1).
txt_doc(452,32,1).
txt_doc(158,32,2).
txt_doc(450,32,1).
txt_doc(311,32,1).
txt_doc(894,32,1).
txt_doc(479,32,1).
txt_doc(114,32,1).
txt_doc(491,32,1).
txt_doc(855,32,1).
txt_doc(761,32,2).
txt_doc(929,32,1).
txt_doc(391,32,2).
txt_doc(145,32,1).
txt_doc(142,32,2).
txt_doc(512,32,1).
txt_doc(231,32,2).
txt_doc(233,32,1).
txt_doc(610,32,1).
txt_doc(476,32,1).
txt_doc(492,32,1).
txt_doc(537,32,2).
txt_doc(606,32,1).
txt_doc(910,32,1).
txt_doc(525,32,1).
txt_doc(222,32,1).
txt_doc(745,32,1).
txt_doc(219,32,1).
txt_doc(967,32,1).
txt_doc(510,33,1).
txt_doc(705,33,1).
txt_doc(797,33,1).
txt_doc(516,33,1).
txt_doc(738,33,1).
txt_doc(6,33,1).
txt_doc(941,33,1).
txt_doc(100,33,2).
txt_doc(203,33,1).
txt_doc(598,33,1).
txt_doc(325,33,1).
txt_doc(94,33,1).
txt_doc(225,33,1).
txt_doc(210,33,1).
txt_doc(629,33,1).
txt_doc(936,33,1).
txt_doc(360,33,1).
txt_doc(687,33,1).
txt_doc(923,33,1).
txt_doc(97,33,1).
txt_doc(915,33,1).
txt_doc(514,33,1).
txt_doc(822,33,1).
txt_doc(481,33,1).
txt_doc(830,33,1).
txt_doc(657,33,1).
txt_doc(430,33,1).
txt_doc(131,33,1).
txt_doc(75,33,1).
txt_doc(944,33,1).
txt_doc(812,33,1).
txt_doc(718,33,1).
txt_doc(868,33,1).
txt_doc(179,33,1).
txt_doc(732,33,1).
txt_doc(540,33,1).
txt_doc(856,33,1).
txt_doc(865,33,1).
txt_doc(490,33,1).
txt_doc(665,33,1).
txt_doc(743,33,1).
txt_doc(621,33,1).
txt_doc(780,33,1).
txt_doc(925,33,1).
txt_doc(61,33,1).
txt_doc(853,33,1).
txt_doc(721,33,1).
txt_doc(59,33,1).
txt_doc(901,33,1).
txt_doc(482,33,1).
txt_doc(844,33,1).
txt_doc(689,33,1).
txt_doc(445,33,1).
txt_doc(56,33,1).
txt_doc(79,33,1).
txt_doc(322,33,1).
txt_doc(728,33,1).
txt_doc(862,33,1).
txt_doc(391,33,1).
txt_doc(734,33,1).
txt_doc(526,33,1).
txt_doc(130,33,1).
txt_doc(212,33,1).
txt_doc(753,33,1).
txt_doc(367,33,1).
txt_doc(236,33,1).
txt_doc(511,33,1).
txt_doc(617,33,1).
txt_doc(433,33,1).
txt_doc(449,33,1).
txt_doc(200,33,1).
txt_doc(948,33,1).
txt_doc(581,33,1).
txt_doc(301,33,1).
txt_doc(673,33,2).
txt_doc(649,33,1).
txt_doc(467,33,1).
txt_doc(846,33,1).
txt_doc(737,33,1).
txt_doc(120,33,1).
txt_doc(585,33,1).
txt_doc(461,33,1).
txt_doc(65,33,1).
txt_doc(328,33,1).
txt_doc(197,33,1).
txt_doc(173,33,1).
txt_doc(881,33,1).
txt_doc(694,33,1).
txt_doc(696,33,1).
txt_doc(370,33,1).
txt_doc(77,33,1).
txt_doc(206,33,1).
txt_doc(766,33,1).
txt_doc(501,33,1).
txt_doc(759,33,1).
txt_doc(88,33,1).
txt_doc(656,33,1).
txt_doc(448,33,1).
txt_doc(52,33,1).
txt_doc(855,33,1).
txt_doc(268,33,1).
txt_doc(497,33,1).
txt_doc(237,33,1).
txt_doc(484,33,1).
txt_doc(307,33,1).
txt_doc(503,33,1).
txt_doc(330,33,1).
txt_doc(413,33,1).
txt_doc(90,33,1).
txt_doc(25,33,1).
txt_doc(784,33,1).
txt_doc(586,33,1).
txt_doc(42,33,1).
txt_doc(538,33,1).
txt_doc(44,33,1).
txt_doc(927,33,1).
txt_doc(399,33,1).
txt_doc(34,33,1).
txt_doc(454,33,2).
txt_doc(389,33,1).
txt_doc(425,33,1).
txt_doc(248,33,2).
txt_doc(945,33,1).
txt_doc(798,33,1).
txt_doc(892,33,2).
txt_doc(431,33,1).
txt_doc(863,33,1).
txt_doc(265,33,1).
txt_doc(185,33,1).
txt_doc(831,33,1).
txt_doc(800,33,1).
txt_doc(412,33,3).
txt_doc(767,33,2).
txt_doc(457,33,1).
txt_doc(919,33,1).
txt_doc(792,33,1).
txt_doc(725,33,1).
txt_doc(238,33,2).
txt_doc(181,33,1).
txt_doc(345,33,1).
txt_doc(334,33,1).
txt_doc(295,33,1).
txt_doc(772,33,1).
txt_doc(91,33,1).
txt_doc(219,33,1).
txt_doc(631,33,1).
txt_doc(493,33,1).
txt_doc(436,33,1).
txt_doc(246,33,1).
txt_doc(745,33,1).
txt_doc(128,33,1).
txt_doc(590,33,1).
txt_doc(78,33,1).
txt_doc(377,33,1).
txt_doc(563,33,1).
txt_doc(735,33,1).
txt_doc(796,33,2).
txt_doc(170,33,1).
txt_doc(487,33,1).
txt_doc(943,33,1).
txt_doc(525,33,1).
txt_doc(272,33,1).
txt_doc(794,33,1).
txt_doc(639,33,1).
txt_doc(384,33,1).
txt_doc(280,33,1).
txt_doc(264,33,1).
txt_doc(469,33,1).
txt_doc(320,33,1).
txt_doc(2,33,1).
txt_doc(405,33,1).
txt_doc(900,33,1).
txt_doc(14,33,1).
txt_doc(486,33,1).
txt_doc(410,33,1).
txt_doc(3,33,1).
txt_doc(957,33,1).
txt_doc(747,33,1).
txt_doc(230,33,2).
txt_doc(648,33,2).
txt_doc(502,33,2).
txt_doc(202,33,1).
txt_doc(55,33,2).
txt_doc(349,33,1).
txt_doc(84,33,1).
txt_doc(357,33,1).
txt_doc(964,34,1).
txt_doc(889,34,1).
txt_doc(160,34,1).
txt_doc(414,34,1).
txt_doc(333,34,1).
txt_doc(529,34,1).
txt_doc(502,34,1).
txt_doc(134,34,1).
txt_doc(619,34,1).
txt_doc(925,34,1).
txt_doc(819,34,1).
txt_doc(924,34,1).
txt_doc(882,34,1).
txt_doc(586,34,1).
txt_doc(163,34,1).
txt_doc(797,34,1).
txt_doc(719,34,1).
txt_doc(378,34,1).
txt_doc(453,34,1).
txt_doc(602,34,1).
txt_doc(731,34,1).
txt_doc(495,34,1).
txt_doc(898,34,1).
txt_doc(872,34,1).
txt_doc(923,34,1).
txt_doc(876,34,1).
txt_doc(214,34,1).
txt_doc(196,34,1).
txt_doc(875,34,1).
txt_doc(347,34,1).
txt_doc(209,34,1).
txt_doc(338,34,1).
txt_doc(21,34,1).
txt_doc(812,34,1).
txt_doc(416,34,1).
txt_doc(857,34,1).
txt_doc(481,34,1).
txt_doc(457,34,1).
txt_doc(194,34,1).
txt_doc(512,34,1).
txt_doc(671,34,1).
txt_doc(676,34,1).
txt_doc(516,34,2).
txt_doc(571,34,1).
txt_doc(795,34,1).
txt_doc(438,34,1).
txt_doc(128,34,1).
txt_doc(231,34,1).
txt_doc(664,34,1).
txt_doc(26,34,1).
txt_doc(10,34,1).
txt_doc(590,34,1).
txt_doc(496,34,1).
txt_doc(167,34,1).
txt_doc(677,34,1).
txt_doc(113,34,1).
txt_doc(204,34,1).
txt_doc(358,34,1).
txt_doc(750,34,1).
txt_doc(440,34,1).
txt_doc(272,34,1).
txt_doc(807,34,1).
txt_doc(324,34,1).
txt_doc(911,34,1).
txt_doc(256,34,1).
txt_doc(308,34,1).
txt_doc(669,34,1).
txt_doc(50,34,1).
txt_doc(666,34,1).
txt_doc(733,34,1).
txt_doc(948,34,1).
txt_doc(367,34,1).
txt_doc(489,34,1).
txt_doc(667,34,1).
txt_doc(44,34,1).
txt_doc(721,34,1).
txt_doc(608,34,2).
txt_doc(574,34,1).
txt_doc(530,34,1).
txt_doc(643,34,1).
txt_doc(628,34,1).
txt_doc(879,34,1).
txt_doc(702,34,1).
txt_doc(177,34,1).
txt_doc(883,34,2).
txt_doc(250,34,1).
txt_doc(775,34,1).
txt_doc(723,34,1).
txt_doc(23,34,1).
txt_doc(754,34,1).
txt_doc(570,34,1).
txt_doc(662,34,1).
txt_doc(447,34,1).
txt_doc(914,34,1).
txt_doc(568,34,1).
txt_doc(849,34,1).
txt_doc(690,34,1).
txt_doc(126,34,2).
txt_doc(859,34,1).
txt_doc(510,34,1).
txt_doc(900,34,1).
txt_doc(111,34,1).
txt_doc(865,34,1).
txt_doc(228,34,1).
txt_doc(224,34,1).
txt_doc(888,34,1).
txt_doc(142,34,1).
txt_doc(760,34,1).
txt_doc(408,34,2).
txt_doc(207,34,1).
txt_doc(905,34,1).
txt_doc(382,34,1).
txt_doc(463,34,1).
txt_doc(19,34,1).
txt_doc(311,34,1).
txt_doc(210,34,1).
txt_doc(144,34,1).
txt_doc(143,34,1).
txt_doc(817,34,1).
txt_doc(655,34,1).
txt_doc(289,34,1).
txt_doc(594,34,1).
txt_doc(839,34,1).
txt_doc(278,34,1).
txt_doc(117,34,1).
txt_doc(211,34,1).
txt_doc(576,34,1).
txt_doc(686,34,1).
txt_doc(132,34,1).
txt_doc(370,34,1).
txt_doc(543,34,1).
txt_doc(673,34,1).
txt_doc(928,34,1).
txt_doc(115,34,1).
txt_doc(514,34,1).
txt_doc(63,34,1).
txt_doc(139,34,1).
txt_doc(471,34,1).
txt_doc(365,34,1).
txt_doc(771,34,1).
txt_doc(940,34,2).
txt_doc(17,34,1).
txt_doc(296,34,1).
txt_doc(83,34,1).
txt_doc(478,34,1).
txt_doc(189,34,1).
txt_doc(625,34,1).
txt_doc(234,34,1).
txt_doc(509,34,1).
txt_doc(579,34,1).
txt_doc(798,34,1).
txt_doc(874,34,2).
txt_doc(618,34,1).
txt_doc(724,34,1).
txt_doc(101,34,1).
txt_doc(855,34,1).
txt_doc(313,34,2).
txt_doc(778,34,1).
txt_doc(73,34,1).
txt_doc(707,34,1).
txt_doc(490,34,1).
txt_doc(192,34,1).
txt_doc(622,34,1).
txt_doc(454,34,1).
txt_doc(355,34,1).
txt_doc(949,34,1).
txt_doc(332,34,1).
txt_doc(331,34,1).
txt_doc(958,34,1).
txt_doc(897,34,1).
txt_doc(383,34,1).
txt_doc(138,34,1).
txt_doc(697,34,1).
txt_doc(480,34,1).
txt_doc(420,34,1).
txt_doc(479,34,1).
txt_doc(57,34,1).
txt_doc(630,34,1).
txt_doc(344,34,1).
txt_doc(513,34,1).
txt_doc(468,34,1).
txt_doc(687,34,1).
txt_doc(88,34,2).
txt_doc(613,34,2).
txt_doc(312,34,1).
txt_doc(417,34,1).
txt_doc(208,34,1).
txt_doc(381,34,1).
txt_doc(813,34,1).
txt_doc(774,34,1).
txt_doc(530,35,1).
txt_doc(226,35,1).
txt_doc(603,35,1).
txt_doc(927,35,1).
txt_doc(494,35,1).
txt_doc(308,35,1).
txt_doc(528,35,1).
txt_doc(198,35,1).
txt_doc(734,35,1).
txt_doc(498,35,1).
txt_doc(832,35,1).
txt_doc(688,35,1).
txt_doc(518,35,1).
txt_doc(923,35,1).
txt_doc(846,35,1).
txt_doc(611,35,2).
txt_doc(761,35,1).
txt_doc(418,35,1).
txt_doc(422,35,1).
txt_doc(191,35,1).
txt_doc(879,35,1).
txt_doc(421,35,1).
txt_doc(168,35,1).
txt_doc(141,35,1).
txt_doc(257,35,1).
txt_doc(81,35,1).
txt_doc(60,35,1).
txt_doc(850,35,1).
txt_doc(478,35,1).
txt_doc(284,35,1).
txt_doc(695,35,1).
txt_doc(500,35,1).
txt_doc(556,35,1).
txt_doc(273,35,1).
txt_doc(190,35,1).
txt_doc(340,35,1).
txt_doc(775,35,1).
txt_doc(230,35,1).
txt_doc(410,35,1).
txt_doc(643,35,1).
txt_doc(69,35,1).
txt_doc(656,35,1).
txt_doc(542,35,1).
txt_doc(211,35,1).
txt_doc(904,35,1).
txt_doc(218,35,1).
txt_doc(509,35,1).
txt_doc(507,35,1).
txt_doc(401,35,1).
txt_doc(165,35,1).
txt_doc(805,35,1).
txt_doc(92,35,1).
txt_doc(93,35,1).
txt_doc(18,35,1).
txt_doc(499,35,1).
txt_doc(915,35,1).
txt_doc(909,35,1).
txt_doc(2,35,1).
txt_doc(100,35,1).
txt_doc(872,35,1).
txt_doc(538,35,1).
txt_doc(853,35,1).
txt_doc(181,35,1).
txt_doc(951,35,1).
txt_doc(113,35,1).
txt_doc(571,35,1).
txt_doc(175,35,1).
txt_doc(847,35,1).
txt_doc(925,35,1).
txt_doc(289,35,1).
txt_doc(427,35,1).
txt_doc(202,35,1).
txt_doc(874,35,1).
txt_doc(798,35,1).
txt_doc(183,35,1).
txt_doc(87,35,1).
txt_doc(717,35,1).
txt_doc(704,35,2).
txt_doc(784,35,1).
txt_doc(956,35,1).
txt_doc(347,35,1).
txt_doc(564,35,1).
txt_doc(664,35,1).
txt_doc(358,35,1).
txt_doc(435,35,1).
txt_doc(61,35,1).
txt_doc(94,35,1).
txt_doc(960,35,1).
txt_doc(350,35,1).
txt_doc(663,35,2).
txt_doc(662,35,1).
txt_doc(800,35,1).
txt_doc(91,35,1).
txt_doc(186,35,1).
txt_doc(506,35,1).
txt_doc(786,35,2).
txt_doc(855,35,1).
txt_doc(533,35,1).
txt_doc(829,35,1).
txt_doc(892,35,1).
txt_doc(199,35,1).
txt_doc(823,35,2).
txt_doc(386,35,1).
txt_doc(387,35,1).
txt_doc(778,35,1).
txt_doc(893,35,1).
txt_doc(348,35,2).
txt_doc(313,35,2).
txt_doc(405,35,1).
txt_doc(540,35,1).
txt_doc(650,35,1).
txt_doc(14,35,1).
txt_doc(277,35,1).
txt_doc(723,35,1).
txt_doc(106,35,1).
txt_doc(229,35,1).
txt_doc(411,35,2).
txt_doc(814,35,1).
txt_doc(129,35,1).
txt_doc(167,35,1).
txt_doc(777,35,1).
txt_doc(210,35,1).
txt_doc(520,35,1).
txt_doc(640,35,1).
txt_doc(232,35,1).
txt_doc(622,35,1).
txt_doc(178,35,2).
txt_doc(77,35,1).
txt_doc(334,35,1).
txt_doc(466,35,1).
txt_doc(89,35,1).
txt_doc(231,35,1).
txt_doc(819,35,1).
txt_doc(958,35,1).
txt_doc(245,35,1).
txt_doc(457,35,1).
txt_doc(487,35,1).
txt_doc(673,35,1).
txt_doc(860,35,1).
txt_doc(488,35,1).
txt_doc(424,35,1).
txt_doc(444,35,1).
txt_doc(595,35,1).
txt_doc(933,35,1).
txt_doc(365,35,1).
txt_doc(773,35,2).
txt_doc(818,35,2).
txt_doc(25,35,1).
txt_doc(490,35,1).
txt_doc(928,35,1).
txt_doc(592,35,1).
txt_doc(393,35,1).
txt_doc(59,35,1).
txt_doc(841,35,1).
txt_doc(913,35,1).
txt_doc(628,35,1).
txt_doc(760,35,1).
txt_doc(535,35,1).
txt_doc(142,35,1).
txt_doc(145,35,2).
txt_doc(330,35,1).
txt_doc(555,35,1).
txt_doc(363,35,1).
txt_doc(460,35,1).
txt_doc(636,35,2).
txt_doc(543,35,1).
txt_doc(204,35,1).
txt_doc(268,35,1).
txt_doc(473,35,1).
txt_doc(90,35,1).
txt_doc(637,35,1).
txt_doc(143,35,1).
txt_doc(749,35,2).
txt_doc(763,35,1).
txt_doc(432,35,1).
txt_doc(743,35,1).
txt_doc(755,35,1).
txt_doc(770,35,1).
txt_doc(719,35,1).
txt_doc(856,35,1).
txt_doc(497,35,1).
txt_doc(264,35,1).
txt_doc(174,35,1).
txt_doc(710,35,1).
txt_doc(251,35,1).
txt_doc(795,35,1).
txt_doc(936,36,1).
txt_doc(282,36,1).
txt_doc(792,36,1).
txt_doc(851,36,1).
txt_doc(721,36,1).
txt_doc(722,36,1).
txt_doc(762,36,1).
txt_doc(176,36,1).
txt_doc(329,36,1).
txt_doc(132,36,1).
txt_doc(220,36,1).
txt_doc(70,36,1).
txt_doc(751,36,1).
txt_doc(733,36,1).
txt_doc(98,36,1).
txt_doc(961,36,1).
txt_doc(917,36,1).
txt_doc(21,36,1).
txt_doc(565,36,1).
txt_doc(696,36,1).
txt_doc(312,36,1).
txt_doc(726,36,1).
txt_doc(398,36,2).
txt_doc(765,36,1).
txt_doc(889,36,1).
txt_doc(136,36,1).
txt_doc(69,36,1).
txt_doc(737,36,1).
txt_doc(734,36,1).
txt_doc(28,36,1).
txt_doc(685,36,1).
txt_doc(15,36,1).
txt_doc(439,36,1).
txt_doc(38,36,1).
txt_doc(934,36,1).
txt_doc(757,36,1).
txt_doc(644,36,1).
txt_doc(203,36,1).
txt_doc(173,36,1).
txt_doc(588,36,1).
txt_doc(619,36,1).
txt_doc(723,36,1).
txt_doc(509,36,1).
txt_doc(725,36,1).
txt_doc(943,36,2).
txt_doc(680,36,1).
txt_doc(518,36,1).
txt_doc(469,36,1).
txt_doc(915,36,1).
txt_doc(447,36,1).
txt_doc(585,36,1).
txt_doc(80,36,1).
txt_doc(781,36,2).
txt_doc(330,36,1).
txt_doc(669,36,1).
txt_doc(672,36,1).
txt_doc(215,36,1).
txt_doc(806,36,1).
txt_doc(526,36,1).
txt_doc(729,36,1).
txt_doc(665,36,1).
txt_doc(143,36,1).
txt_doc(640,36,1).
txt_doc(298,36,1).
txt_doc(797,36,1).
txt_doc(745,36,1).
txt_doc(909,36,1).
txt_doc(477,36,2).
txt_doc(362,36,1).
txt_doc(756,36,1).
txt_doc(328,36,1).
txt_doc(399,36,1).
txt_doc(410,36,1).
txt_doc(530,36,1).
txt_doc(749,36,1).
txt_doc(834,36,1).
txt_doc(631,36,1).
txt_doc(430,36,1).
txt_doc(237,36,1).
txt_doc(424,36,1).
txt_doc(549,36,1).
txt_doc(4,36,1).
txt_doc(820,36,1).
txt_doc(646,36,1).
txt_doc(758,36,1).
txt_doc(332,36,1).
txt_doc(683,36,1).
txt_doc(638,36,1).
txt_doc(568,36,1).
txt_doc(907,36,2).
txt_doc(48,36,1).
txt_doc(874,36,1).
txt_doc(373,36,1).
txt_doc(12,36,1).
txt_doc(504,36,2).
txt_doc(461,36,2).
txt_doc(437,36,1).
txt_doc(938,36,1).
txt_doc(99,36,1).
txt_doc(645,36,1).
txt_doc(394,36,1).
txt_doc(595,36,1).
txt_doc(537,36,1).
txt_doc(309,36,1).
txt_doc(801,36,1).
txt_doc(727,36,1).
txt_doc(523,36,2).
txt_doc(73,36,1).
txt_doc(610,36,1).
txt_doc(579,36,1).
txt_doc(27,36,1).
txt_doc(188,36,1).
txt_doc(352,36,1).
txt_doc(404,36,1).
txt_doc(498,36,1).
txt_doc(418,36,1).
txt_doc(307,36,1).
txt_doc(572,36,1).
txt_doc(423,36,1).
txt_doc(670,36,1).
txt_doc(679,36,2).
txt_doc(475,36,1).
txt_doc(845,36,1).
txt_doc(266,36,1).
txt_doc(214,36,2).
txt_doc(571,36,1).
txt_doc(605,36,1).
txt_doc(322,36,1).
txt_doc(911,36,1).
txt_doc(962,36,1).
txt_doc(199,36,1).
txt_doc(922,36,1).
txt_doc(552,36,1).
txt_doc(455,36,1).
txt_doc(786,36,1).
txt_doc(825,36,1).
txt_doc(612,36,1).
txt_doc(952,36,1).
txt_doc(662,36,1).
txt_doc(152,36,1).
txt_doc(539,36,1).
txt_doc(512,36,1).
txt_doc(884,36,1).
txt_doc(557,36,1).
txt_doc(829,36,1).
txt_doc(230,36,1).
txt_doc(673,36,2).
txt_doc(473,36,1).
txt_doc(68,36,1).
txt_doc(574,36,1).
txt_doc(413,36,1).
txt_doc(365,36,1).
txt_doc(717,36,1).
txt_doc(360,36,1).
txt_doc(932,36,1).
txt_doc(51,36,1).
txt_doc(110,36,1).
txt_doc(881,36,1).
txt_doc(887,36,2).
txt_doc(828,36,1).
txt_doc(542,36,1).
txt_doc(403,36,1).
txt_doc(327,36,1).
txt_doc(865,36,2).
txt_doc(163,36,1).
txt_doc(697,36,1).
txt_doc(563,36,1).
txt_doc(338,36,1).
txt_doc(728,36,1).
txt_doc(883,36,2).
txt_doc(125,36,1).
txt_doc(165,36,1).
txt_doc(252,36,1).
txt_doc(34,36,2).
txt_doc(951,36,1).
txt_doc(401,36,1).
txt_doc(56,36,1).
txt_doc(452,36,1).
txt_doc(438,36,1).
txt_doc(600,36,1).
txt_doc(273,36,1).
txt_doc(431,36,1).
txt_doc(957,36,1).
txt_doc(390,36,1).
txt_doc(254,36,1).
txt_doc(109,37,1).
txt_doc(46,37,1).
txt_doc(545,37,1).
txt_doc(930,37,1).
txt_doc(264,37,1).
txt_doc(885,37,1).
txt_doc(171,37,1).
txt_doc(708,37,1).
txt_doc(138,37,1).
txt_doc(685,37,1).
txt_doc(309,37,1).
txt_doc(157,37,1).
txt_doc(81,37,1).
txt_doc(500,37,1).
txt_doc(86,37,1).
txt_doc(605,37,1).
txt_doc(210,37,1).
txt_doc(654,37,1).
txt_doc(628,37,1).
txt_doc(713,37,1).
txt_doc(842,37,1).
txt_doc(117,37,1).
txt_doc(700,37,1).
txt_doc(579,37,1).
txt_doc(738,37,1).
txt_doc(851,37,1).
txt_doc(716,37,1).
txt_doc(111,37,1).
txt_doc(554,37,1).
txt_doc(312,37,1).
txt_doc(866,37,1).
txt_doc(867,37,1).
txt_doc(514,37,1).
txt_doc(645,37,1).
txt_doc(217,37,1).
txt_doc(695,37,1).
txt_doc(167,37,1).
txt_doc(873,37,1).
txt_doc(60,37,1).
txt_doc(856,37,1).
txt_doc(345,37,2).
txt_doc(881,37,1).
txt_doc(749,37,2).
txt_doc(609,37,1).
txt_doc(144,37,1).
txt_doc(848,37,1).
txt_doc(246,37,1).
txt_doc(951,37,1).
txt_doc(765,37,1).
txt_doc(932,37,1).
txt_doc(466,37,1).
txt_doc(771,37,2).
txt_doc(380,37,1).
txt_doc(11,37,1).
txt_doc(725,37,2).
txt_doc(228,37,1).
txt_doc(250,37,1).
txt_doc(937,37,1).
txt_doc(454,37,1).
txt_doc(17,37,1).
txt_doc(372,37,1).
txt_doc(632,37,1).
txt_doc(397,37,1).
txt_doc(578,37,1).
txt_doc(913,37,1).
txt_doc(854,37,1).
txt_doc(682,37,1).
txt_doc(911,37,1).
txt_doc(517,37,1).
txt_doc(385,37,1).
txt_doc(22,37,1).
txt_doc(159,37,1).
txt_doc(805,37,1).
txt_doc(918,37,1).
txt_doc(288,37,1).
txt_doc(914,37,1).
txt_doc(525,37,1).
txt_doc(421,37,1).
txt_doc(68,37,1).
txt_doc(967,37,1).
txt_doc(746,37,1).
txt_doc(843,37,1).
txt_doc(468,37,1).
txt_doc(450,37,1).
txt_doc(301,37,1).
txt_doc(793,37,1).
txt_doc(747,37,1).
txt_doc(241,37,1).
txt_doc(965,37,1).
txt_doc(19,37,1).
txt_doc(75,37,1).
txt_doc(505,37,1).
txt_doc(927,37,1).
txt_doc(104,37,1).
txt_doc(263,37,2).
txt_doc(416,37,1).
txt_doc(114,37,1).
txt_doc(602,37,1).
txt_doc(660,37,1).
txt_doc(899,37,1).
txt_doc(102,37,2).
txt_doc(953,37,1).
txt_doc(955,37,1).
txt_doc(390,37,1).
txt_doc(426,37,1).
txt_doc(939,37,1).
txt_doc(163,37,1).
txt_doc(349,37,1).
txt_doc(13,37,1).
txt_doc(348,37,2).
txt_doc(386,37,1).
txt_doc(501,37,1).
txt_doc(38,37,1).
txt_doc(388,37,1).
txt_doc(283,37,1).
txt_doc(120,37,1).
txt_doc(658,37,1).
txt_doc(5,37,2).
txt_doc(412,37,1).
txt_doc(574,37,1).
txt_doc(239,37,1).
txt_doc(703,37,1).
txt_doc(297,37,1).
txt_doc(825,37,1).
txt_doc(57,37,1).
txt_doc(42,37,3).
txt_doc(174,37,1).
txt_doc(531,37,1).
txt_doc(522,37,1).
txt_doc(803,37,1).
txt_doc(538,37,1).
txt_doc(335,37,1).
txt_doc(540,37,3).
txt_doc(295,37,1).
txt_doc(888,37,1).
txt_doc(617,37,1).
txt_doc(405,37,1).
txt_doc(281,37,1).
txt_doc(593,37,1).
txt_doc(440,37,1).
txt_doc(863,37,1).
txt_doc(886,37,1).
txt_doc(586,37,1).
txt_doc(493,37,1).
txt_doc(341,37,1).
txt_doc(100,37,1).
txt_doc(569,37,1).
txt_doc(799,37,1).
txt_doc(258,37,1).
txt_doc(465,37,1).
txt_doc(3,37,1).
txt_doc(139,37,1).
txt_doc(580,37,1).
txt_doc(135,37,1).
txt_doc(496,37,2).
txt_doc(720,37,1).
txt_doc(601,37,1).
txt_doc(556,37,1).
txt_doc(641,37,1).
txt_doc(772,37,1).
txt_doc(853,37,1).
txt_doc(356,37,1).
txt_doc(940,37,1).
txt_doc(400,37,2).
txt_doc(847,37,1).
txt_doc(219,37,1).
txt_doc(826,37,1).
txt_doc(379,37,2).
txt_doc(782,37,1).
txt_doc(822,37,1).
txt_doc(557,37,2).
txt_doc(82,37,1).
txt_doc(774,37,1).
txt_doc(638,37,2).
txt_doc(530,37,1).
txt_doc(285,37,1).
txt_doc(714,37,1).
txt_doc(215,37,1).
txt_doc(195,37,1).
txt_doc(391,37,1).
txt_doc(511,37,1).
txt_doc(325,37,1).
txt_doc(65,37,1).
txt_doc(44,38,1).
txt_doc(835,38,1).
txt_doc(441,38,1).
txt_doc(614,38,1).
txt_doc(312,38,1).
txt_doc(567,38,1).
txt_doc(191,38,1).
txt_doc(149,38,1).
txt_doc(880,38,1).
txt_doc(145,38,1).
txt_doc(20,38,1).
txt_doc(564,38,1).
txt_doc(537,38,1).
txt_doc(780,38,1).
txt_doc(516,38,1).
txt_doc(931,38,1).
txt_doc(854,38,1).
txt_doc(952,38,1).
txt_doc(239,38,1).
txt_doc(830,38,1).
txt_doc(75,38,1).
txt_doc(329,38,1).
txt_doc(450,38,1).
txt_doc(810,38,1).
txt_doc(658,38,1).
txt_doc(102,38,1).
txt_doc(389,38,1).
txt_doc(306,38,1).
txt_doc(867,38,1).
txt_doc(228,38,1).
txt_doc(531,38,1).
txt_doc(813,38,2).
txt_doc(430,38,1).
txt_doc(649,38,1).
txt_doc(905,38,1).
txt_doc(412,38,1).
txt_doc(53,38,1).
txt_doc(868,38,1).
txt_doc(901,38,1).
txt_doc(316,38,1).
txt_doc(690,38,1).
txt_doc(755,38,1).
txt_doc(859,38,1).
txt_doc(322,38,1).
txt_doc(589,38,1).
txt_doc(909,38,1).
txt_doc(8,38,1).
txt_doc(781,38,1).
txt_doc(423,38,1).
txt_doc(28,38,1).
txt_doc(544,38,1).
txt_doc(253,38,1).
txt_doc(246,38,1).
txt_doc(679,38,1).
txt_doc(606,38,1).
txt_doc(37,38,1).
txt_doc(833,38,1).
txt_doc(178,38,1).
txt_doc(358,38,1).
txt_doc(272,38,2).
txt_doc(936,38,1).
txt_doc(827,38,1).
txt_doc(634,38,1).
txt_doc(406,38,1).
txt_doc(705,38,1).
txt_doc(754,38,1).
txt_doc(527,38,1).
txt_doc(968,38,2).
txt_doc(851,38,1).
txt_doc(761,38,1).
txt_doc(309,38,1).
txt_doc(183,38,1).
txt_doc(158,38,1).
txt_doc(738,38,2).
txt_doc(114,38,1).
txt_doc(482,38,1).
txt_doc(194,38,1).
txt_doc(758,38,1).
txt_doc(339,38,1).
txt_doc(650,38,1).
txt_doc(222,38,1).
txt_doc(501,38,1).
txt_doc(933,38,1).
txt_doc(784,38,1).
txt_doc(786,38,1).
txt_doc(174,38,1).
txt_doc(27,38,1).
txt_doc(395,38,1).
txt_doc(966,38,1).
txt_doc(13,38,1).
txt_doc(140,38,1).
txt_doc(491,38,1).
txt_doc(259,38,1).
txt_doc(230,38,1).
txt_doc(917,38,1).
txt_doc(301,38,1).
txt_doc(357,38,1).
txt_doc(951,38,1).
txt_doc(1,38,1).
txt_doc(313,38,1).
txt_doc(515,38,1).
txt_doc(453,38,1).
txt_doc(672,38,1).
txt_doc(156,38,1).
txt_doc(413,38,1).
txt_doc(850,38,1).
txt_doc(9,38,1).
txt_doc(492,38,1).
txt_doc(602,38,2).
txt_doc(708,38,3).
txt_doc(710,38,1).
txt_doc(420,38,1).
txt_doc(338,38,1).
txt_doc(663,38,2).
txt_doc(502,38,1).
txt_doc(184,38,1).
txt_doc(315,38,1).
txt_doc(226,38,1).
txt_doc(462,38,1).
txt_doc(471,38,2).
txt_doc(18,38,1).
txt_doc(865,38,1).
txt_doc(247,38,1).
txt_doc(853,38,1).
txt_doc(386,38,1).
txt_doc(457,38,1).
txt_doc(824,38,2).
txt_doc(373,38,1).
txt_doc(725,38,1).
txt_doc(846,38,1).
txt_doc(186,38,1).
txt_doc(348,38,1).
txt_doc(209,38,2).
txt_doc(295,38,1).
txt_doc(418,38,1).
txt_doc(604,38,1).
txt_doc(893,38,2).
txt_doc(697,38,1).
txt_doc(914,38,1).
txt_doc(245,38,2).
txt_doc(152,38,1).
txt_doc(288,38,1).
txt_doc(122,38,1).
txt_doc(764,38,1).
txt_doc(201,38,1).
txt_doc(444,38,1).
txt_doc(47,38,2).
txt_doc(631,38,1).
txt_doc(741,38,1).
txt_doc(713,38,1).
txt_doc(584,38,1).
txt_doc(193,38,2).
txt_doc(414,38,1).
txt_doc(942,38,1).
txt_doc(443,38,1).
txt_doc(258,38,1).
txt_doc(617,38,1).
txt_doc(319,38,1).
txt_doc(553,38,2).
txt_doc(547,38,1).
txt_doc(192,38,1).
txt_doc(621,38,1).
txt_doc(106,38,2).
txt_doc(298,38,1).
txt_doc(651,38,1).
txt_doc(954,38,1).
txt_doc(202,38,1).
txt_doc(719,38,1).
txt_doc(456,38,1).
txt_doc(164,38,1).
txt_doc(493,38,1).
txt_doc(820,38,2).
txt_doc(297,38,1).
txt_doc(629,38,1).
txt_doc(198,38,1).
txt_doc(882,38,1).
txt_doc(291,38,1).
txt_doc(802,38,1).
txt_doc(717,38,1).
txt_doc(435,38,1).
txt_doc(799,38,1).
txt_doc(95,38,1).
txt_doc(116,39,1).
txt_doc(932,39,1).
txt_doc(778,39,1).
txt_doc(863,39,1).
txt_doc(941,39,1).
txt_doc(118,39,1).
txt_doc(940,39,1).
txt_doc(854,39,1).
txt_doc(39,39,1).
txt_doc(516,39,1).
txt_doc(946,39,1).
txt_doc(530,39,1).
txt_doc(733,39,1).
txt_doc(840,39,1).
txt_doc(916,39,1).
txt_doc(657,39,1).
txt_doc(511,39,1).
txt_doc(837,39,1).
txt_doc(179,39,1).
txt_doc(285,39,1).
txt_doc(211,39,2).
txt_doc(238,39,1).
txt_doc(608,39,1).
txt_doc(877,39,1).
txt_doc(41,39,1).
txt_doc(801,39,1).
txt_doc(441,39,1).
txt_doc(884,39,1).
txt_doc(76,39,1).
txt_doc(754,39,1).
txt_doc(578,39,1).
txt_doc(415,39,1).
txt_doc(846,39,1).
txt_doc(638,39,1).
txt_doc(153,39,2).
txt_doc(955,39,2).
txt_doc(436,39,1).
txt_doc(494,39,1).
txt_doc(898,39,1).
txt_doc(710,39,1).
txt_doc(392,39,1).
txt_doc(862,39,1).
txt_doc(721,39,1).
txt_doc(440,39,1).
txt_doc(848,39,1).
txt_doc(42,39,1).
txt_doc(569,39,1).
txt_doc(167,39,1).
txt_doc(413,39,1).
txt_doc(716,39,2).
txt_doc(776,39,1).
txt_doc(206,39,1).
txt_doc(851,39,1).
txt_doc(202,39,1).
txt_doc(439,39,1).
txt_doc(655,39,1).
txt_doc(624,39,1).
txt_doc(135,39,1).
txt_doc(172,39,1).
txt_doc(570,39,1).
txt_doc(128,39,1).
txt_doc(295,39,1).
txt_doc(154,39,1).
txt_doc(72,39,1).
txt_doc(647,39,1).
txt_doc(497,39,1).
txt_doc(574,39,1).
txt_doc(500,39,1).
txt_doc(140,39,2).
txt_doc(501,39,1).
txt_doc(712,39,1).
txt_doc(509,39,1).
txt_doc(317,39,1).
txt_doc(152,39,1).
txt_doc(271,39,1).
txt_doc(783,39,1).
txt_doc(242,39,1).
txt_doc(370,39,2).
txt_doc(856,39,1).
txt_doc(333,39,1).
txt_doc(464,39,1).
txt_doc(67,39,1).
txt_doc(843,39,1).
txt_doc(666,39,1).
txt_doc(291,39,1).
txt_doc(394,39,2).
txt_doc(696,39,1).
txt_doc(585,39,1).
txt_doc(729,39,1).
txt_doc(61,39,1).
txt_doc(831,39,1).
txt_doc(693,39,1).
txt_doc(163,39,1).
txt_doc(788,39,1).
txt_doc(715,39,1).
txt_doc(219,39,1).
txt_doc(80,39,1).
txt_doc(178,39,2).
txt_doc(338,39,1).
txt_doc(663,39,1).
txt_doc(146,39,2).
txt_doc(23,39,1).
txt_doc(45,39,1).
txt_doc(397,39,1).
txt_doc(416,39,1).
txt_doc(842,39,1).
txt_doc(514,39,1).
txt_doc(901,39,1).
txt_doc(723,39,1).
txt_doc(918,39,1).
txt_doc(669,39,1).
txt_doc(220,39,1).
txt_doc(15,39,1).
txt_doc(150,39,1).
txt_doc(454,39,1).
txt_doc(539,39,1).
txt_doc(784,39,1).
txt_doc(180,39,1).
txt_doc(890,39,1).
txt_doc(764,39,1).
txt_doc(866,39,1).
txt_doc(517,39,1).
txt_doc(194,39,1).
txt_doc(883,39,1).
txt_doc(292,39,1).
txt_doc(630,39,1).
txt_doc(832,39,1).
txt_doc(168,39,2).
txt_doc(513,39,1).
txt_doc(896,39,2).
txt_doc(52,39,1).
txt_doc(632,39,1).
txt_doc(480,39,1).
txt_doc(187,39,1).
txt_doc(66,39,1).
txt_doc(820,39,1).
txt_doc(387,39,1).
txt_doc(425,39,1).
txt_doc(16,39,1).
txt_doc(47,39,1).
txt_doc(747,39,1).
txt_doc(640,39,1).
txt_doc(122,39,1).
txt_doc(13,39,2).
txt_doc(443,39,1).
txt_doc(524,39,1).
txt_doc(561,39,2).
txt_doc(253,39,1).
txt_doc(159,39,2).
txt_doc(705,39,2).
txt_doc(963,39,1).
txt_doc(470,39,1).
txt_doc(105,39,1).
txt_doc(24,39,2).
txt_doc(284,39,2).
txt_doc(327,39,1).
txt_doc(314,39,1).
txt_doc(477,39,1).
txt_doc(592,39,1).
txt_doc(813,39,1).
txt_doc(659,39,3).
txt_doc(137,39,1).
txt_doc(589,39,1).
txt_doc(809,39,1).
txt_doc(621,39,2).
txt_doc(175,39,1).
txt_doc(713,39,1).
txt_doc(384,39,1).
txt_doc(756,39,1).
txt_doc(430,39,1).
txt_doc(773,39,1).
txt_doc(365,39,1).
txt_doc(815,39,1).
txt_doc(556,39,1).
txt_doc(272,39,1).
txt_doc(844,39,1).
txt_doc(518,39,2).
txt_doc(689,39,1).
txt_doc(358,39,1).
txt_doc(933,40,1).
txt_doc(21,40,1).
txt_doc(182,40,1).
txt_doc(298,40,1).
txt_doc(832,40,1).
txt_doc(412,40,1).
txt_doc(574,40,1).
txt_doc(649,40,1).
txt_doc(692,40,1).
txt_doc(404,40,1).
txt_doc(452,40,1).
txt_doc(4,40,1).
txt_doc(187,40,1).
txt_doc(352,40,1).
txt_doc(688,40,1).
txt_doc(573,40,1).
txt_doc(561,40,1).
txt_doc(293,40,1).
txt_doc(764,40,1).
txt_doc(967,40,1).
txt_doc(374,40,1).
txt_doc(456,40,1).
txt_doc(258,40,1).
txt_doc(405,40,1).
txt_doc(466,40,1).
txt_doc(650,40,1).
txt_doc(894,40,1).
txt_doc(338,40,1).
txt_doc(460,40,1).
txt_doc(176,40,1).
txt_doc(112,40,1).
txt_doc(462,40,1).
txt_doc(390,40,1).
txt_doc(469,40,1).
txt_doc(529,40,1).
txt_doc(720,40,1).
txt_doc(428,40,1).
txt_doc(715,40,1).
txt_doc(204,40,1).
txt_doc(799,40,2).
txt_doc(470,40,1).
txt_doc(721,40,1).
txt_doc(250,40,1).
txt_doc(464,40,1).
txt_doc(269,40,1).
txt_doc(392,40,1).
txt_doc(395,40,1).
txt_doc(811,40,3).
txt_doc(431,40,1).
txt_doc(908,40,1).
txt_doc(443,40,1).
txt_doc(23,40,1).
txt_doc(238,40,1).
txt_doc(47,40,1).
txt_doc(22,40,2).
txt_doc(966,40,1).
txt_doc(442,40,1).
txt_doc(121,40,1).
txt_doc(636,40,1).
txt_doc(723,40,1).
txt_doc(255,40,1).
txt_doc(394,40,1).
txt_doc(704,40,1).
txt_doc(225,40,1).
txt_doc(853,40,1).
txt_doc(65,40,1).
txt_doc(198,40,1).
txt_doc(873,40,1).
txt_doc(211,40,1).
txt_doc(586,40,2).
txt_doc(403,40,1).
txt_doc(123,40,2).
txt_doc(672,40,1).
txt_doc(158,40,1).
txt_doc(792,40,1).
txt_doc(865,40,1).
txt_doc(189,40,1).
txt_doc(146,40,1).
txt_doc(590,40,2).
txt_doc(420,40,1).
txt_doc(495,40,1).
txt_doc(870,40,1).
txt_doc(1,40,1).
txt_doc(584,40,1).
txt_doc(90,40,1).
txt_doc(203,40,1).
txt_doc(116,40,1).
txt_doc(844,40,1).
txt_doc(12,40,1).
txt_doc(687,40,1).
txt_doc(758,40,1).
txt_doc(548,40,1).
txt_doc(432,40,1).
txt_doc(740,40,1).
txt_doc(699,40,1).
txt_doc(30,40,1).
txt_doc(154,40,1).
txt_doc(254,40,1).
txt_doc(856,40,1).
txt_doc(738,40,1).
txt_doc(950,40,1).
txt_doc(77,40,1).
txt_doc(781,40,1).
txt_doc(819,40,1).
txt_doc(480,40,1).
txt_doc(410,40,1).
txt_doc(138,40,2).
txt_doc(845,40,2).
txt_doc(909,40,1).
txt_doc(217,40,1).
txt_doc(447,40,1).
txt_doc(797,40,1).
txt_doc(793,40,1).
txt_doc(571,40,1).
txt_doc(810,40,1).
txt_doc(117,40,1).
txt_doc(783,40,1).
txt_doc(733,40,1).
txt_doc(893,40,1).
txt_doc(904,40,1).
txt_doc(755,40,1).
txt_doc(192,40,1).
txt_doc(739,40,1).
txt_doc(304,40,1).
txt_doc(361,40,1).
txt_doc(903,40,1).
txt_doc(965,40,1).
txt_doc(477,40,2).
txt_doc(327,40,1).
txt_doc(20,40,1).
txt_doc(789,40,1).
txt_doc(864,40,1).
txt_doc(345,40,1).
txt_doc(633,40,1).
txt_doc(215,40,1).
txt_doc(408,40,2).
txt_doc(716,40,1).
txt_doc(754,40,1).
txt_doc(476,40,2).
txt_doc(876,40,2).
txt_doc(342,40,1).
txt_doc(697,40,1).
txt_doc(524,40,1).
txt_doc(383,40,1).
txt_doc(222,40,1).
txt_doc(52,40,1).
txt_doc(209,40,1).
txt_doc(170,40,1).
txt_doc(120,40,1).
txt_doc(669,40,1).
txt_doc(691,40,1).
txt_doc(665,40,1).
txt_doc(148,40,1).
txt_doc(936,40,1).
txt_doc(896,40,2).
txt_doc(591,40,1).
txt_doc(386,40,1).
txt_doc(417,40,1).
txt_doc(232,40,1).
txt_doc(968,40,2).
txt_doc(101,40,2).
txt_doc(68,40,2).
txt_doc(558,40,1).
txt_doc(455,40,2).
txt_doc(600,40,1).
txt_doc(51,40,2).
txt_doc(662,40,1).
txt_doc(667,40,1).
txt_doc(512,40,1).
txt_doc(291,40,1).
txt_doc(761,40,1).
txt_doc(453,40,1).
txt_doc(702,40,1).
txt_doc(563,40,1).
txt_doc(960,40,1).
txt_doc(545,40,1).
txt_doc(539,40,3).
txt_doc(751,40,1).
txt_doc(806,40,1).
txt_doc(581,41,1).
txt_doc(50,41,1).
txt_doc(86,41,1).
txt_doc(550,41,1).
txt_doc(710,41,1).
txt_doc(343,41,1).
txt_doc(433,41,1).
txt_doc(572,41,2).
txt_doc(731,41,1).
txt_doc(495,41,1).
txt_doc(360,41,1).
txt_doc(744,41,1).
txt_doc(44,41,1).
txt_doc(763,41,1).
txt_doc(515,41,1).
txt_doc(668,41,1).
txt_doc(181,41,2).
txt_doc(292,41,1).
txt_doc(255,41,1).
txt_doc(443,41,1).
txt_doc(85,41,1).
txt_doc(543,41,1).
txt_doc(740,41,1).
txt_doc(281,41,1).
txt_doc(308,41,1).
txt_doc(405,41,1).
txt_doc(383,41,1).
txt_doc(966,41,1).
txt_doc(659,41,1).
txt_doc(577,41,1).
txt_doc(440,41,1).
txt_doc(438,41,1).
txt_doc(821,41,1).
txt_doc(691,41,1).
txt_doc(371,41,1).
txt_doc(412,41,1).
txt_doc(835,41,1).
txt_doc(112,41,1).
txt_doc(451,41,1).
txt_doc(866,41,1).
txt_doc(352,41,1).
txt_doc(311,41,1).
txt_doc(200,41,1).
txt_doc(468,41,1).
txt_doc(90,41,1).
txt_doc(904,41,1).
txt_doc(573,41,1).
txt_doc(798,41,1).
txt_doc(728,41,1).
txt_doc(204,41,1).
txt_doc(883,41,1).
txt_doc(504,41,1).
txt_doc(722,41,1).
txt_doc(58,41,1).
txt_doc(372,41,2).
txt_doc(173,41,1).
txt_doc(72,41,1).
txt_doc(150,41,2).
txt_doc(113,41,1).
txt_doc(228,41,1).
txt_doc(452,41,1).
txt_doc(110,41,1).
txt_doc(425,41,1).
txt_doc(431,41,1).
txt_doc(783,41,1).
txt_doc(836,41,1).
txt_doc(56,41,1).
txt_doc(791,41,1).
txt_doc(754,41,1).
txt_doc(796,41,1).
txt_doc(69,41,1).
txt_doc(532,41,1).
txt_doc(179,41,1).
txt_doc(388,41,1).
txt_doc(492,41,1).
txt_doc(600,41,1).
txt_doc(756,41,1).
txt_doc(704,41,1).
txt_doc(875,41,1).
txt_doc(813,41,1).
txt_doc(479,41,1).
txt_doc(167,41,1).
txt_doc(512,41,1).
txt_doc(583,41,1).
txt_doc(345,41,1).
txt_doc(539,41,1).
txt_doc(665,41,1).
txt_doc(263,41,1).
txt_doc(881,41,1).
txt_doc(364,41,1).
txt_doc(892,41,2).
txt_doc(144,41,1).
txt_doc(411,41,1).
txt_doc(351,41,1).
txt_doc(286,41,1).
txt_doc(389,41,1).
txt_doc(645,41,1).
txt_doc(948,41,1).
txt_doc(23,41,1).
txt_doc(814,41,1).
txt_doc(248,41,1).
txt_doc(338,41,1).
txt_doc(38,41,1).
txt_doc(568,41,1).
txt_doc(954,41,1).
txt_doc(145,41,1).
txt_doc(886,41,1).
txt_doc(745,41,1).
txt_doc(514,41,1).
txt_doc(340,41,1).
txt_doc(486,41,2).
txt_doc(598,41,1).
txt_doc(926,41,1).
txt_doc(363,41,1).
txt_doc(535,41,1).
txt_doc(305,41,1).
txt_doc(905,41,1).
txt_doc(293,41,1).
txt_doc(945,41,1).
txt_doc(450,41,1).
txt_doc(337,41,2).
txt_doc(873,41,1).
txt_doc(140,41,1).
txt_doc(643,41,1).
txt_doc(244,41,1).
txt_doc(218,41,1).
txt_doc(224,41,1).
txt_doc(735,41,1).
txt_doc(605,41,1).
txt_doc(264,41,1).
txt_doc(932,41,1).
txt_doc(256,41,1).
txt_doc(833,41,2).
txt_doc(520,41,1).
txt_doc(817,41,1).
txt_doc(241,41,1).
txt_doc(950,41,1).
txt_doc(693,41,1).
txt_doc(870,41,1).
txt_doc(284,41,1).
txt_doc(721,41,1).
txt_doc(231,41,2).
txt_doc(548,41,1).
txt_doc(258,41,1).
txt_doc(210,41,1).
txt_doc(930,41,1).
txt_doc(671,41,1).
txt_doc(547,41,1).
txt_doc(222,41,1).
txt_doc(481,41,1).
txt_doc(599,41,1).
txt_doc(563,41,1).
txt_doc(829,41,1).
txt_doc(2,41,1).
txt_doc(507,41,1).
txt_doc(70,41,1).
txt_doc(952,41,1).
txt_doc(769,41,1).
txt_doc(753,41,1).
txt_doc(661,41,1).
txt_doc(876,41,1).
txt_doc(146,41,1).
txt_doc(651,41,3).
txt_doc(718,41,1).
txt_doc(488,41,1).
txt_doc(587,41,1).
txt_doc(120,41,2).
txt_doc(261,41,1).
txt_doc(193,41,2).
txt_doc(584,41,2).
txt_doc(689,41,1).
txt_doc(825,41,2).
txt_doc(322,41,2).
txt_doc(25,41,1).
txt_doc(908,41,1).
txt_doc(919,41,1).
txt_doc(84,41,3).
txt_doc(45,41,1).
txt_doc(156,41,1).
txt_doc(874,41,2).
txt_doc(310,41,1).
txt_doc(641,42,1).
txt_doc(955,42,1).
txt_doc(504,42,1).
txt_doc(178,42,1).
txt_doc(413,42,1).
txt_doc(189,42,1).
txt_doc(914,42,1).
txt_doc(691,42,1).
txt_doc(839,42,1).
txt_doc(729,42,1).
txt_doc(567,42,1).
txt_doc(149,42,1).
txt_doc(30,42,1).
txt_doc(302,42,1).
txt_doc(282,42,1).
txt_doc(867,42,1).
txt_doc(903,42,1).
txt_doc(797,42,1).
txt_doc(655,42,1).
txt_doc(358,42,1).
txt_doc(871,42,1).
txt_doc(312,42,1).
txt_doc(260,42,1).
txt_doc(347,42,1).
txt_doc(454,42,1).
txt_doc(952,42,1).
txt_doc(620,42,1).
txt_doc(111,42,1).
txt_doc(649,42,1).
txt_doc(823,42,1).
txt_doc(466,42,1).
txt_doc(240,42,1).
txt_doc(557,42,1).
txt_doc(115,42,1).
txt_doc(157,42,1).
txt_doc(181,42,1).
txt_doc(962,42,1).
txt_doc(389,42,1).
txt_doc(405,42,1).
txt_doc(121,42,1).
txt_doc(844,42,1).
txt_doc(229,42,1).
txt_doc(831,42,1).
txt_doc(832,42,1).
txt_doc(528,42,1).
txt_doc(203,42,1).
txt_doc(679,42,1).
txt_doc(924,42,1).
txt_doc(907,42,1).
txt_doc(589,42,1).
txt_doc(50,42,1).
txt_doc(776,42,1).
txt_doc(214,42,1).
txt_doc(415,42,1).
txt_doc(911,42,1).
txt_doc(370,42,1).
txt_doc(4,42,1).
txt_doc(529,42,2).
txt_doc(851,42,1).
txt_doc(335,42,1).
txt_doc(85,42,1).
txt_doc(539,42,1).
txt_doc(496,42,1).
txt_doc(503,42,1).
txt_doc(720,42,1).
txt_doc(585,42,1).
txt_doc(743,42,1).
txt_doc(781,42,1).
txt_doc(922,42,1).
txt_doc(543,42,1).
txt_doc(535,42,2).
txt_doc(426,42,1).
txt_doc(224,42,1).
txt_doc(572,42,1).
txt_doc(734,42,1).
txt_doc(747,42,1).
txt_doc(610,42,1).
txt_doc(109,42,2).
txt_doc(868,42,1).
txt_doc(12,42,1).
txt_doc(698,42,1).
txt_doc(195,42,1).
txt_doc(103,42,1).
txt_doc(272,42,1).
txt_doc(334,42,1).
txt_doc(656,42,1).
txt_doc(556,42,1).
txt_doc(818,42,1).
txt_doc(60,42,1).
txt_doc(448,42,2).
txt_doc(268,42,1).
txt_doc(435,42,1).
txt_doc(654,42,1).
txt_doc(19,42,1).
txt_doc(58,42,1).
txt_doc(623,42,2).
txt_doc(309,42,1).
txt_doc(179,42,1).
txt_doc(28,42,1).
txt_doc(762,42,1).
txt_doc(741,42,1).
txt_doc(650,42,1).
txt_doc(660,42,1).
txt_doc(455,42,1).
txt_doc(269,42,1).
txt_doc(25,42,1).
txt_doc(444,42,1).
txt_doc(804,42,2).
txt_doc(758,42,1).
txt_doc(830,42,1).
txt_doc(468,42,1).
txt_doc(502,42,1).
txt_doc(467,42,1).
txt_doc(390,42,1).
txt_doc(731,42,1).
txt_doc(38,42,2).
txt_doc(775,42,1).
txt_doc(211,42,1).
txt_doc(36,42,1).
txt_doc(813,42,1).
txt_doc(117,42,1).
txt_doc(200,42,1).
txt_doc(644,42,2).
txt_doc(278,42,1).
txt_doc(913,42,1).
txt_doc(615,42,1).
txt_doc(303,42,1).
txt_doc(803,42,1).
txt_doc(110,42,2).
txt_doc(123,42,1).
txt_doc(625,42,1).
txt_doc(286,42,1).
txt_doc(852,42,1).
txt_doc(420,42,1).
txt_doc(352,42,1).
txt_doc(44,42,1).
txt_doc(323,42,1).
txt_doc(552,42,1).
txt_doc(418,42,2).
txt_doc(139,42,1).
txt_doc(75,42,1).
txt_doc(876,42,1).
txt_doc(532,42,1).
txt_doc(939,42,1).
txt_doc(697,42,1).
txt_doc(896,42,1).
txt_doc(404,42,1).
txt_doc(288,42,1).
txt_doc(77,42,1).
txt_doc(126,42,1).
txt_doc(383,42,1).
txt_doc(524,42,1).
txt_doc(930,42,1).
txt_doc(915,42,1).
txt_doc(313,42,1).
txt_doc(231,42,1).
txt_doc(61,42,1).
txt_doc(950,42,1).
txt_doc(274,42,1).
txt_doc(768,42,1).
txt_doc(213,42,2).
txt_doc(236,42,1).
txt_doc(482,42,1).
txt_doc(353,42,1).
txt_doc(812,42,1).
txt_doc(872,42,1).
txt_doc(624,42,1).
txt_doc(947,42,1).
txt_doc(689,42,1).
txt_doc(461,42,1).
txt_doc(142,42,1).
txt_doc(948,42,1).
txt_doc(15,42,1).
txt_doc(255,42,1).
txt_doc(465,42,1).
txt_doc(76,42,1).
txt_doc(789,42,1).
txt_doc(565,42,2).
txt_doc(760,42,1).
txt_doc(437,42,1).
txt_doc(46,42,1).
txt_doc(554,42,2).
txt_doc(422,42,1).
txt_doc(840,42,1).
txt_doc(967,42,1).
txt_doc(186,42,1).
txt_doc(506,42,1).
txt_doc(649,43,1).
txt_doc(959,43,1).
txt_doc(955,43,1).
txt_doc(682,43,1).
txt_doc(326,43,1).
txt_doc(903,43,1).
txt_doc(295,43,1).
txt_doc(355,43,1).
txt_doc(469,43,1).
txt_doc(736,43,1).
txt_doc(843,43,1).
txt_doc(503,43,1).
txt_doc(234,43,1).
txt_doc(864,43,1).
txt_doc(259,43,1).
txt_doc(8,43,1).
txt_doc(313,43,1).
txt_doc(1,43,1).
txt_doc(557,43,1).
txt_doc(159,43,1).
txt_doc(633,43,1).
txt_doc(407,43,1).
txt_doc(524,43,1).
txt_doc(575,43,1).
txt_doc(953,43,1).
txt_doc(308,43,1).
txt_doc(816,43,1).
txt_doc(342,43,1).
txt_doc(696,43,1).
txt_doc(740,43,1).
txt_doc(566,43,1).
txt_doc(156,43,1).
txt_doc(22,43,1).
txt_doc(599,43,1).
txt_doc(431,43,1).
txt_doc(68,43,1).
txt_doc(184,43,1).
txt_doc(731,43,1).
txt_doc(203,43,1).
txt_doc(314,43,1).
txt_doc(438,43,1).
txt_doc(966,43,1).
txt_doc(862,43,1).
txt_doc(448,43,1).
txt_doc(498,43,1).
txt_doc(200,43,1).
txt_doc(520,43,1).
txt_doc(489,43,1).
txt_doc(494,43,1).
txt_doc(309,43,1).
txt_doc(623,43,1).
txt_doc(227,43,1).
txt_doc(768,43,1).
txt_doc(160,43,1).
txt_doc(683,43,1).
txt_doc(618,43,2).
txt_doc(870,43,1).
txt_doc(117,43,1).
txt_doc(37,43,1).
txt_doc(75,43,1).
txt_doc(501,43,1).
txt_doc(928,43,1).
txt_doc(244,43,1).
txt_doc(253,43,1).
txt_doc(706,43,1).
txt_doc(170,43,1).
txt_doc(418,43,1).
txt_doc(678,43,1).
txt_doc(450,43,1).
txt_doc(472,43,1).
txt_doc(511,43,1).
txt_doc(299,43,1).
txt_doc(770,43,1).
txt_doc(658,43,1).
txt_doc(388,43,1).
txt_doc(631,43,1).
txt_doc(945,43,2).
txt_doc(934,43,1).
txt_doc(829,43,1).
txt_doc(83,43,1).
txt_doc(647,43,1).
txt_doc(428,43,1).
txt_doc(215,43,1).
txt_doc(794,43,1).
txt_doc(101,43,1).
txt_doc(324,43,1).
txt_doc(325,43,1).
txt_doc(400,43,1).
txt_doc(224,43,1).
txt_doc(440,43,1).
txt_doc(255,43,1).
txt_doc(761,43,1).
txt_doc(104,43,1).
txt_doc(213,43,1).
txt_doc(370,43,1).
txt_doc(758,43,1).
txt_doc(195,43,1).
txt_doc(13,43,1).
txt_doc(107,43,1).
txt_doc(167,43,1).
txt_doc(725,43,1).
txt_doc(65,43,1).
txt_doc(395,43,1).
txt_doc(152,43,1).
txt_doc(265,43,1).
txt_doc(447,43,1).
txt_doc(901,43,2).
txt_doc(672,43,1).
txt_doc(548,43,1).
txt_doc(114,43,1).
txt_doc(526,43,1).
txt_doc(612,43,1).
txt_doc(514,43,1).
txt_doc(964,43,1).
txt_doc(354,43,1).
txt_doc(47,43,1).
txt_doc(642,43,1).
txt_doc(775,43,1).
txt_doc(150,43,1).
txt_doc(748,43,1).
txt_doc(841,43,1).
txt_doc(804,43,1).
txt_doc(610,43,1).
txt_doc(890,43,1).
txt_doc(304,43,2).
txt_doc(707,43,1).
txt_doc(59,43,1).
txt_doc(165,43,1).
txt_doc(242,43,1).
txt_doc(55,43,2).
txt_doc(628,43,1).
txt_doc(732,43,2).
txt_doc(965,43,1).
txt_doc(21,43,1).
txt_doc(559,43,1).
txt_doc(345,43,1).
txt_doc(483,43,1).
txt_doc(779,43,1).
txt_doc(4,43,1).
txt_doc(789,43,2).
txt_doc(127,43,1).
txt_doc(689,43,1).
txt_doc(138,43,1).
txt_doc(709,43,2).
txt_doc(445,43,1).
txt_doc(778,43,1).
txt_doc(209,43,1).
txt_doc(232,43,2).
txt_doc(251,43,1).
txt_doc(292,43,1).
txt_doc(188,43,1).
txt_doc(471,43,2).
txt_doc(756,43,1).
txt_doc(275,43,1).
txt_doc(105,43,1).
txt_doc(842,43,1).
txt_doc(12,43,1).
txt_doc(852,43,1).
txt_doc(908,43,1).
txt_doc(310,43,1).
txt_doc(522,43,1).
txt_doc(374,43,1).
txt_doc(783,43,1).
txt_doc(487,43,1).
txt_doc(634,43,1).
txt_doc(225,43,1).
txt_doc(763,43,2).
txt_doc(563,43,1).
txt_doc(484,43,1).
txt_doc(162,43,1).
txt_doc(924,43,1).
txt_doc(88,43,2).
txt_doc(434,43,2).
txt_doc(968,43,1).
txt_doc(38,43,1).
txt_doc(131,43,1).
txt_doc(439,43,1).
txt_doc(817,43,2).
txt_doc(369,43,1).
txt_doc(293,43,1).
txt_doc(913,43,1).
txt_doc(512,43,1).
txt_doc(218,43,1).
txt_doc(648,43,1).
txt_doc(497,43,1).
txt_doc(869,43,1).
txt_doc(343,44,1).
txt_doc(314,44,1).
txt_doc(857,44,1).
txt_doc(106,44,1).
txt_doc(549,44,1).
txt_doc(907,44,1).
txt_doc(923,44,1).
txt_doc(309,44,1).
txt_doc(612,44,1).
txt_doc(523,44,1).
txt_doc(910,44,1).
txt_doc(939,44,1).
txt_doc(905,44,1).
txt_doc(710,44,1).
txt_doc(507,44,1).
txt_doc(324,44,1).
txt_doc(561,44,1).
txt_doc(914,44,1).
txt_doc(680,44,1).
txt_doc(945,44,1).
txt_doc(89,44,1).
txt_doc(434,44,1).
txt_doc(43,44,1).
txt_doc(714,44,1).
txt_doc(236,44,1).
txt_doc(518,44,1).
txt_doc(743,44,1).
txt_doc(15,44,1).
txt_doc(884,44,1).
txt_doc(171,44,1).
txt_doc(463,44,1).
txt_doc(796,44,1).
txt_doc(739,44,1).
txt_doc(12,44,1).
txt_doc(832,44,1).
txt_doc(893,44,1).
txt_doc(544,44,1).
txt_doc(512,44,1).
txt_doc(811,44,1).
txt_doc(722,44,1).
txt_doc(58,44,1).
txt_doc(570,44,1).
txt_doc(584,44,1).
txt_doc(607,44,1).
txt_doc(752,44,2).
txt_doc(823,44,1).
txt_doc(65,44,1).
txt_doc(83,44,1).
txt_doc(413,44,1).
txt_doc(925,44,2).
txt_doc(305,44,1).
txt_doc(955,44,1).
txt_doc(920,44,1).
txt_doc(131,44,1).
txt_doc(165,44,1).
txt_doc(490,44,1).
txt_doc(836,44,1).
txt_doc(246,44,1).
txt_doc(807,44,1).
txt_doc(958,44,1).
txt_doc(798,44,1).
txt_doc(934,44,1).
txt_doc(431,44,1).
txt_doc(231,44,1).
txt_doc(616,44,1).
txt_doc(560,44,1).
txt_doc(197,44,1).
txt_doc(334,44,1).
txt_doc(9,44,1).
txt_doc(70,44,1).
txt_doc(597,44,1).
txt_doc(589,44,1).
txt_doc(967,44,1).
txt_doc(224,44,1).
txt_doc(261,44,1).
txt_doc(219,44,1).
txt_doc(685,44,1).
txt_doc(129,44,1).
txt_doc(373,44,1).
txt_doc(278,44,1).
txt_doc(202,44,1).
txt_doc(906,44,1).
txt_doc(528,44,1).
txt_doc(904,44,1).
txt_doc(107,44,1).
txt_doc(17,44,1).
txt_doc(548,44,1).
txt_doc(645,44,1).
txt_doc(372,44,1).
txt_doc(269,44,1).
txt_doc(843,44,1).
txt_doc(346,44,2).
txt_doc(851,44,1).
txt_doc(531,44,1).
txt_doc(679,44,1).
txt_doc(913,44,1).
txt_doc(613,44,1).
txt_doc(91,44,1).
txt_doc(255,44,1).
txt_doc(226,44,2).
txt_doc(331,44,1).
txt_doc(422,44,1).
txt_doc(356,44,1).
txt_doc(301,44,1).
txt_doc(508,44,1).
txt_doc(855,44,1).
txt_doc(744,44,1).
txt_doc(687,44,1).
txt_doc(479,44,1).
txt_doc(283,44,1).
txt_doc(517,44,1).
txt_doc(498,44,1).
txt_doc(936,44,1).
txt_doc(558,44,1).
txt_doc(235,44,1).
txt_doc(765,44,2).
txt_doc(688,44,1).
txt_doc(34,44,1).
txt_doc(759,44,1).
txt_doc(730,44,1).
txt_doc(443,44,1).
txt_doc(458,44,1).
txt_doc(631,44,1).
txt_doc(861,44,1).
txt_doc(127,44,1).
txt_doc(946,44,1).
txt_doc(822,44,2).
txt_doc(146,44,1).
txt_doc(357,44,1).
txt_doc(414,44,2).
txt_doc(877,44,1).
txt_doc(128,44,1).
txt_doc(634,44,1).
txt_doc(288,44,1).
txt_doc(40,44,1).
txt_doc(894,44,1).
txt_doc(84,44,1).
txt_doc(163,44,1).
txt_doc(285,44,1).
txt_doc(41,44,1).
txt_doc(641,44,2).
txt_doc(49,44,1).
txt_doc(397,44,1).
txt_doc(689,44,2).
txt_doc(527,44,1).
txt_doc(830,44,1).
txt_doc(625,44,1).
txt_doc(601,44,2).
txt_doc(467,44,1).
txt_doc(118,44,1).
txt_doc(166,44,1).
txt_doc(53,44,2).
txt_doc(804,44,1).
txt_doc(580,44,1).
txt_doc(643,44,1).
txt_doc(782,44,1).
txt_doc(1,44,1).
txt_doc(340,44,1).
txt_doc(481,44,1).
txt_doc(32,44,1).
txt_doc(825,44,1).
txt_doc(280,44,2).
txt_doc(514,44,1).
txt_doc(159,44,1).
txt_doc(665,44,1).
txt_doc(270,44,1).
txt_doc(553,44,1).
txt_doc(682,44,1).
txt_doc(718,44,1).
txt_doc(336,44,2).
txt_doc(874,44,1).
txt_doc(656,44,1).
txt_doc(332,44,1).
txt_doc(858,44,1).
txt_doc(632,44,1).
txt_doc(966,44,2).
txt_doc(935,44,1).
txt_doc(417,44,1).
txt_doc(882,44,1).
txt_doc(535,44,1).
txt_doc(704,44,1).
txt_doc(529,44,1).
txt_doc(633,44,1).
txt_doc(229,44,3).
txt_doc(344,45,1).
txt_doc(501,45,1).
txt_doc(656,45,1).
txt_doc(699,45,1).
txt_doc(920,45,1).
txt_doc(111,45,1).
txt_doc(738,45,1).
txt_doc(638,45,1).
txt_doc(135,45,1).
txt_doc(550,45,1).
txt_doc(544,45,1).
txt_doc(445,45,1).
txt_doc(790,45,1).
txt_doc(140,45,1).
txt_doc(960,45,1).
txt_doc(926,45,1).
txt_doc(305,45,1).
txt_doc(202,45,1).
txt_doc(565,45,1).
txt_doc(537,45,1).
txt_doc(560,45,1).
txt_doc(447,45,1).
txt_doc(265,45,1).
txt_doc(179,45,1).
txt_doc(591,45,1).
txt_doc(187,45,1).
txt_doc(479,45,1).
txt_doc(128,45,1).
txt_doc(829,45,1).
txt_doc(899,45,1).
txt_doc(103,45,1).
txt_doc(76,45,1).
txt_doc(967,45,1).
txt_doc(123,45,1).
txt_doc(458,45,1).
txt_doc(543,45,1).
txt_doc(392,45,1).
txt_doc(480,45,1).
txt_doc(642,45,1).
txt_doc(314,45,1).
txt_doc(620,45,1).
txt_doc(410,45,1).
txt_doc(716,45,1).
txt_doc(222,45,1).
txt_doc(244,45,1).
txt_doc(614,45,1).
txt_doc(396,45,1).
txt_doc(936,45,2).
txt_doc(559,45,1).
txt_doc(8,45,1).
txt_doc(922,45,1).
txt_doc(720,45,1).
txt_doc(382,45,1).
txt_doc(939,45,1).
txt_doc(772,45,1).
txt_doc(149,45,1).
txt_doc(33,45,2).
txt_doc(47,45,1).
txt_doc(947,45,1).
txt_doc(824,45,1).
txt_doc(454,45,1).
txt_doc(315,45,1).
txt_doc(634,45,1).
txt_doc(138,45,2).
txt_doc(90,45,1).
txt_doc(30,45,1).
txt_doc(139,45,1).
txt_doc(449,45,1).
txt_doc(787,45,1).
txt_doc(57,45,1).
txt_doc(203,45,1).
txt_doc(242,45,1).
txt_doc(69,45,1).
txt_doc(841,45,1).
txt_doc(301,45,1).
txt_doc(953,45,1).
txt_doc(220,45,1).
txt_doc(156,45,1).
txt_doc(754,45,1).
txt_doc(142,45,1).
txt_doc(227,45,1).
txt_doc(548,45,1).
txt_doc(426,45,1).
txt_doc(415,45,1).
txt_doc(234,45,1).
txt_doc(852,45,1).
txt_doc(25,45,1).
txt_doc(63,45,1).
txt_doc(45,45,1).
txt_doc(297,45,1).
txt_doc(605,45,1).
txt_doc(931,45,1).
txt_doc(166,45,1).
txt_doc(628,45,1).
txt_doc(771,45,1).
txt_doc(580,45,1).
txt_doc(291,45,1).
txt_doc(857,45,1).
txt_doc(528,45,1).
txt_doc(621,45,1).
txt_doc(54,45,2).
txt_doc(583,45,2).
txt_doc(946,45,1).
txt_doc(246,45,1).
txt_doc(567,45,1).
txt_doc(365,45,1).
txt_doc(722,45,1).
txt_doc(384,45,1).
txt_doc(643,45,1).
txt_doc(428,45,2).
txt_doc(350,45,1).
txt_doc(741,45,2).
txt_doc(132,45,1).
txt_doc(676,45,1).
txt_doc(48,45,2).
txt_doc(494,45,1).
txt_doc(475,45,1).
txt_doc(508,45,1).
txt_doc(780,45,1).
txt_doc(683,45,1).
txt_doc(794,45,1).
txt_doc(68,45,1).
txt_doc(478,45,1).
txt_doc(842,45,1).
txt_doc(878,45,2).
txt_doc(331,45,1).
txt_doc(892,45,2).
txt_doc(487,45,1).
txt_doc(653,45,1).
txt_doc(943,45,1).
txt_doc(704,45,1).
txt_doc(800,45,1).
txt_doc(687,45,1).
txt_doc(385,45,1).
txt_doc(717,45,1).
txt_doc(200,45,2).
txt_doc(249,45,2).
txt_doc(162,45,1).
txt_doc(180,45,1).
txt_doc(866,45,1).
txt_doc(879,45,1).
txt_doc(515,45,1).
txt_doc(398,45,1).
txt_doc(49,45,1).
txt_doc(245,45,1).
txt_doc(574,45,1).
txt_doc(670,45,1).
txt_doc(556,45,2).
txt_doc(657,45,1).
txt_doc(387,45,1).
txt_doc(219,45,1).
txt_doc(367,45,1).
txt_doc(20,45,2).
txt_doc(522,45,1).
txt_doc(114,45,1).
txt_doc(321,45,1).
txt_doc(32,45,2).
txt_doc(318,45,1).
txt_doc(476,45,1).
txt_doc(348,45,1).
txt_doc(16,45,1).
txt_doc(817,45,1).
txt_doc(378,45,1).
txt_doc(803,45,1).
txt_doc(65,45,1).
txt_doc(659,45,2).
txt_doc(358,45,2).
txt_doc(466,45,1).
txt_doc(671,45,1).
txt_doc(596,45,1).
txt_doc(489,45,1).
txt_doc(213,45,1).
txt_doc(188,45,1).
txt_doc(874,45,1).
txt_doc(306,45,1).
txt_doc(962,45,1).
txt_doc(517,45,1).
txt_doc(462,45,1).
txt_doc(372,45,1).
txt_doc(523,45,1).
txt_doc(827,45,2).
txt_doc(221,45,1).
txt_doc(953,46,1).
txt_doc(141,46,1).
txt_doc(32,46,1).
txt_doc(95,46,1).
txt_doc(150,46,1).
txt_doc(211,46,1).
txt_doc(261,46,1).
txt_doc(288,46,1).
txt_doc(303,46,1).
txt_doc(62,46,1).
txt_doc(61,46,1).
txt_doc(461,46,1).
txt_doc(854,46,1).
txt_doc(351,46,2).
txt_doc(451,46,1).
txt_doc(218,46,1).
txt_doc(273,46,1).
txt_doc(778,46,1).
txt_doc(711,46,1).
txt_doc(312,46,1).
txt_doc(103,46,1).
txt_doc(161,46,1).
txt_doc(849,46,1).
txt_doc(877,46,1).
txt_doc(361,46,1).
txt_doc(842,46,1).
txt_doc(471,46,1).
txt_doc(862,46,1).
txt_doc(523,46,1).
txt_doc(164,46,1).
txt_doc(728,46,1).
txt_doc(513,46,1).
txt_doc(251,46,1).
txt_doc(400,46,1).
txt_doc(672,46,1).
txt_doc(191,46,1).
txt_doc(508,46,1).
txt_doc(764,46,1).
txt_doc(676,46,1).
txt_doc(209,46,1).
txt_doc(435,46,1).
txt_doc(83,46,1).
txt_doc(79,46,1).
txt_doc(6,46,1).
txt_doc(717,46,1).
txt_doc(599,46,1).
txt_doc(174,46,2).
txt_doc(470,46,1).
txt_doc(337,46,1).
txt_doc(636,46,1).
txt_doc(208,46,1).
txt_doc(220,46,1).
txt_doc(20,46,1).
txt_doc(204,46,1).
txt_doc(784,46,1).
txt_doc(699,46,1).
txt_doc(213,46,1).
txt_doc(365,46,1).
txt_doc(98,46,2).
txt_doc(656,46,1).
txt_doc(326,46,1).
txt_doc(215,46,1).
txt_doc(962,46,1).
txt_doc(119,46,1).
txt_doc(684,46,1).
txt_doc(136,46,1).
txt_doc(325,46,1).
txt_doc(87,46,1).
txt_doc(789,46,1).
txt_doc(321,46,1).
txt_doc(232,46,1).
txt_doc(934,46,1).
txt_doc(867,46,1).
txt_doc(195,46,1).
txt_doc(906,46,1).
txt_doc(567,46,1).
txt_doc(693,46,1).
txt_doc(297,46,1).
txt_doc(878,46,1).
txt_doc(682,46,1).
txt_doc(267,46,1).
txt_doc(257,46,1).
txt_doc(135,46,1).
txt_doc(743,46,1).
txt_doc(535,46,1).
txt_doc(484,46,1).
txt_doc(715,46,1).
txt_doc(739,46,1).
txt_doc(274,46,1).
txt_doc(724,46,1).
txt_doc(113,46,1).
txt_doc(607,46,1).
txt_doc(714,46,1).
txt_doc(898,46,1).
txt_doc(766,46,1).
txt_doc(917,46,1).
txt_doc(446,46,1).
txt_doc(686,46,4).
txt_doc(755,46,1).
txt_doc(465,46,2).
txt_doc(94,46,1).
txt_doc(646,46,1).
txt_doc(275,46,1).
txt_doc(909,46,1).
txt_doc(859,46,2).
txt_doc(774,46,1).
txt_doc(398,46,1).
txt_doc(964,46,2).
txt_doc(225,46,2).
txt_doc(891,46,1).
txt_doc(478,46,1).
txt_doc(162,46,1).
txt_doc(617,46,1).
txt_doc(122,46,1).
txt_doc(228,46,1).
txt_doc(536,46,1).
txt_doc(201,46,1).
txt_doc(71,46,1).
txt_doc(39,46,1).
txt_doc(586,46,1).
txt_doc(720,46,1).
txt_doc(740,46,1).
txt_doc(645,46,1).
txt_doc(777,46,1).
txt_doc(167,46,1).
txt_doc(495,46,1).
txt_doc(663,46,2).
txt_doc(362,46,1).
txt_doc(649,46,1).
txt_doc(476,46,1).
txt_doc(841,46,2).
txt_doc(582,46,1).
txt_doc(323,46,1).
txt_doc(300,46,1).
txt_doc(44,46,1).
txt_doc(604,46,1).
txt_doc(266,46,1).
txt_doc(104,46,1).
txt_doc(595,46,1).
txt_doc(576,46,2).
txt_doc(825,46,1).
txt_doc(280,46,1).
txt_doc(895,46,1).
txt_doc(224,46,1).
txt_doc(583,46,1).
txt_doc(602,46,1).
txt_doc(248,46,1).
txt_doc(894,46,1).
txt_doc(343,46,1).
txt_doc(614,46,1).
txt_doc(834,46,1).
txt_doc(568,46,1).
txt_doc(865,46,1).
txt_doc(529,46,1).
txt_doc(945,46,1).
txt_doc(665,46,2).
txt_doc(876,46,1).
txt_doc(181,46,1).
txt_doc(457,46,1).
txt_doc(330,46,1).
txt_doc(82,46,1).
txt_doc(615,46,1).
txt_doc(86,46,1).
txt_doc(719,46,1).
txt_doc(792,46,1).
txt_doc(756,46,1).
txt_doc(146,46,1).
txt_doc(479,46,1).
txt_doc(394,46,1).
txt_doc(521,46,1).
txt_doc(308,46,1).
txt_doc(360,46,1).
txt_doc(695,46,1).
txt_doc(317,46,1).
txt_doc(272,46,2).
txt_doc(128,46,1).
txt_doc(489,46,2).
txt_doc(689,46,3).
txt_doc(130,46,1).
txt_doc(373,46,1).
txt_doc(11,46,1).
txt_doc(697,46,1).
txt_doc(889,47,1).
txt_doc(742,47,1).
txt_doc(363,47,1).
txt_doc(49,47,1).
txt_doc(752,47,1).
txt_doc(927,47,1).
txt_doc(719,47,1).
txt_doc(904,47,1).
txt_doc(30,47,1).
txt_doc(99,47,1).
txt_doc(682,47,1).
txt_doc(175,47,1).
txt_doc(377,47,1).
txt_doc(817,47,1).
txt_doc(912,47,1).
txt_doc(854,47,1).
txt_doc(395,47,1).
txt_doc(275,47,1).
txt_doc(361,47,1).
txt_doc(882,47,1).
txt_doc(425,47,1).
txt_doc(845,47,1).
txt_doc(123,47,1).
txt_doc(113,47,1).
txt_doc(78,47,1).
txt_doc(284,47,1).
txt_doc(891,47,1).
txt_doc(754,47,1).
txt_doc(554,47,2).
txt_doc(90,47,1).
txt_doc(545,47,1).
txt_doc(433,47,1).
txt_doc(243,47,1).
txt_doc(512,47,1).
txt_doc(525,47,1).
txt_doc(806,47,1).
txt_doc(835,47,1).
txt_doc(274,47,1).
txt_doc(60,47,1).
txt_doc(8,47,1).
txt_doc(387,47,1).
txt_doc(506,47,1).
txt_doc(415,47,1).
txt_doc(358,47,1).
txt_doc(560,47,1).
txt_doc(229,47,1).
txt_doc(82,47,1).
txt_doc(144,47,1).
txt_doc(933,47,2).
txt_doc(513,47,1).
txt_doc(273,47,1).
txt_doc(593,47,1).
txt_doc(500,47,1).
txt_doc(134,47,1).
txt_doc(384,47,1).
txt_doc(66,47,2).
txt_doc(709,47,1).
txt_doc(871,47,1).
txt_doc(648,47,1).
txt_doc(315,47,1).
txt_doc(346,47,1).
txt_doc(396,47,1).
txt_doc(502,47,1).
txt_doc(55,47,2).
txt_doc(867,47,1).
txt_doc(165,47,1).
txt_doc(102,47,1).
txt_doc(873,47,1).
txt_doc(747,47,1).
txt_doc(829,47,1).
txt_doc(941,47,1).
txt_doc(846,47,1).
txt_doc(130,47,1).
txt_doc(137,47,1).
txt_doc(796,47,1).
txt_doc(269,47,1).
txt_doc(209,47,1).
txt_doc(836,47,1).
txt_doc(486,47,1).
txt_doc(677,47,1).
txt_doc(379,47,2).
txt_doc(477,47,1).
txt_doc(922,47,1).
txt_doc(654,47,1).
txt_doc(347,47,1).
txt_doc(656,47,1).
txt_doc(902,47,1).
txt_doc(930,47,1).
txt_doc(91,47,1).
txt_doc(658,47,1).
txt_doc(501,47,2).
txt_doc(872,47,1).
txt_doc(107,47,1).
txt_doc(83,47,1).
txt_doc(73,47,1).
txt_doc(287,47,1).
txt_doc(157,47,1).
txt_doc(383,47,1).
txt_doc(529,47,1).
txt_doc(257,47,1).
txt_doc(722,47,1).
txt_doc(606,47,1).
txt_doc(935,47,1).
txt_doc(708,47,1).
txt_doc(758,47,1).
txt_doc(892,47,1).
txt_doc(371,47,1).
txt_doc(141,47,1).
txt_doc(673,47,1).
txt_doc(655,47,1).
txt_doc(5,47,1).
txt_doc(763,47,1).
txt_doc(247,47,1).
txt_doc(409,47,2).
txt_doc(635,47,1).
txt_doc(600,47,1).
txt_doc(115,47,2).
txt_doc(488,47,2).
txt_doc(413,47,1).
txt_doc(158,47,1).
txt_doc(946,47,1).
txt_doc(919,47,2).
txt_doc(201,47,1).
txt_doc(496,47,1).
txt_doc(205,47,1).
txt_doc(668,47,1).
txt_doc(728,47,1).
txt_doc(235,47,1).
txt_doc(965,47,1).
txt_doc(237,47,1).
txt_doc(693,47,1).
txt_doc(856,47,1).
txt_doc(678,47,1).
txt_doc(878,47,1).
txt_doc(307,47,1).
txt_doc(681,47,1).
txt_doc(441,47,2).
txt_doc(587,47,1).
txt_doc(75,47,1).
txt_doc(242,47,1).
txt_doc(485,47,1).
txt_doc(458,47,1).
txt_doc(558,47,1).
txt_doc(532,47,1).
txt_doc(818,47,2).
txt_doc(748,47,1).
txt_doc(745,47,1).
txt_doc(160,47,1).
txt_doc(620,47,2).
txt_doc(868,47,1).
txt_doc(281,47,1).
txt_doc(240,47,1).
txt_doc(296,47,1).
txt_doc(52,47,1).
txt_doc(39,47,2).
txt_doc(731,47,1).
txt_doc(234,47,2).
txt_doc(22,47,1).
txt_doc(264,47,1).
txt_doc(92,47,1).
txt_doc(810,47,2).
txt_doc(641,47,1).
txt_doc(87,47,1).
txt_doc(42,47,1).
txt_doc(245,47,1).
txt_doc(724,47,2).
txt_doc(480,47,1).
txt_doc(510,47,2).
txt_doc(920,47,1).
txt_doc(760,47,2).
txt_doc(179,47,1).
txt_doc(382,47,1).
txt_doc(807,47,1).
txt_doc(18,47,1).
txt_doc(828,47,1).
txt_doc(887,47,2).
txt_doc(340,47,1).
txt_doc(95,47,1).
txt_doc(251,47,1).
txt_doc(555,47,1).
txt_doc(421,48,1).
txt_doc(163,48,2).
txt_doc(736,48,1).
txt_doc(860,48,1).
txt_doc(143,48,1).
txt_doc(681,48,1).
txt_doc(473,48,1).
txt_doc(481,48,1).
txt_doc(310,48,1).
txt_doc(803,48,1).
txt_doc(216,48,1).
txt_doc(556,48,1).
txt_doc(776,48,1).
txt_doc(911,48,1).
txt_doc(596,48,1).
txt_doc(814,48,1).
txt_doc(87,48,1).
txt_doc(145,48,1).
txt_doc(672,48,1).
txt_doc(800,48,1).
txt_doc(15,48,1).
txt_doc(340,48,1).
txt_doc(228,48,1).
txt_doc(938,48,1).
txt_doc(821,48,1).
txt_doc(39,48,2).
txt_doc(597,48,1).
txt_doc(947,48,1).
txt_doc(317,48,1).
txt_doc(348,48,1).
txt_doc(35,48,1).
txt_doc(80,48,1).
txt_doc(769,48,1).
txt_doc(907,48,1).
txt_doc(948,48,1).
txt_doc(866,48,1).
txt_doc(405,48,1).
txt_doc(160,48,1).
txt_doc(832,48,1).
txt_doc(162,48,1).
txt_doc(551,48,1).
txt_doc(335,48,1).
txt_doc(899,48,1).
txt_doc(967,48,1).
txt_doc(808,48,1).
txt_doc(825,48,1).
txt_doc(390,48,1).
txt_doc(136,48,1).
txt_doc(42,48,1).
txt_doc(294,48,1).
txt_doc(675,48,1).
txt_doc(615,48,1).
txt_doc(514,48,1).
txt_doc(230,48,1).
txt_doc(198,48,1).
txt_doc(183,48,1).
txt_doc(608,48,1).
txt_doc(650,48,1).
txt_doc(717,48,1).
txt_doc(433,48,1).
txt_doc(127,48,2).
txt_doc(79,48,1).
txt_doc(231,48,1).
txt_doc(750,48,1).
txt_doc(111,48,1).
txt_doc(663,48,1).
txt_doc(893,48,1).
txt_doc(72,48,1).
txt_doc(235,48,1).
txt_doc(497,48,1).
txt_doc(705,48,2).
txt_doc(504,48,1).
txt_doc(835,48,1).
txt_doc(337,48,1).
txt_doc(795,48,1).
txt_doc(444,48,2).
txt_doc(541,48,1).
txt_doc(641,48,2).
txt_doc(374,48,1).
txt_doc(921,48,1).
txt_doc(889,48,1).
txt_doc(376,48,1).
txt_doc(525,48,1).
txt_doc(384,48,1).
txt_doc(375,48,1).
txt_doc(93,48,1).
txt_doc(827,48,1).
txt_doc(887,48,1).
txt_doc(952,48,1).
txt_doc(3,48,1).
txt_doc(467,48,1).
txt_doc(361,48,1).
txt_doc(68,48,1).
txt_doc(552,48,1).
txt_doc(459,48,1).
txt_doc(22,48,1).
txt_doc(305,48,1).
txt_doc(864,48,1).
txt_doc(265,48,1).
txt_doc(639,48,1).
txt_doc(416,48,1).
txt_doc(437,48,1).
txt_doc(285,48,1).
txt_doc(321,48,2).
txt_doc(638,48,1).
txt_doc(453,48,1).
txt_doc(60,48,1).
txt_doc(543,48,1).
txt_doc(886,48,1).
txt_doc(21,48,2).
txt_doc(618,48,1).
txt_doc(727,48,2).
txt_doc(118,48,1).
txt_doc(168,48,1).
txt_doc(188,48,2).
txt_doc(63,48,1).
txt_doc(761,48,1).
txt_doc(227,48,1).
txt_doc(811,48,2).
txt_doc(49,48,1).
txt_doc(562,48,1).
txt_doc(271,48,1).
txt_doc(500,48,1).
txt_doc(478,48,1).
txt_doc(184,48,1).
txt_doc(571,48,2).
txt_doc(471,48,1).
txt_doc(590,48,1).
txt_doc(771,48,1).
txt_doc(19,48,1).
txt_doc(18,48,1).
txt_doc(428,48,1).
txt_doc(326,48,1).
txt_doc(57,48,1).
txt_doc(568,48,1).
txt_doc(931,48,1).
txt_doc(796,48,1).
txt_doc(805,48,1).
txt_doc(703,48,1).
txt_doc(926,48,1).
txt_doc(78,48,1).
txt_doc(766,48,2).
txt_doc(290,48,2).
txt_doc(730,48,1).
txt_doc(859,48,1).
txt_doc(632,48,1).
txt_doc(959,48,1).
txt_doc(191,48,1).
txt_doc(212,48,1).
txt_doc(278,48,1).
txt_doc(640,48,1).
txt_doc(555,48,1).
txt_doc(458,48,1).
txt_doc(11,48,1).
txt_doc(297,48,1).
txt_doc(325,48,1).
txt_doc(928,48,1).
txt_doc(155,48,1).
txt_doc(94,48,1).
txt_doc(895,48,1).
txt_doc(377,48,1).
txt_doc(689,48,1).
txt_doc(403,48,1).
txt_doc(40,48,1).
txt_doc(643,48,1).
txt_doc(511,48,1).
txt_doc(548,48,1).
txt_doc(829,48,2).
txt_doc(644,48,1).
txt_doc(175,48,1).
txt_doc(544,48,1).
txt_doc(448,48,1).
txt_doc(576,48,1).
txt_doc(269,48,2).
txt_doc(740,48,1).
txt_doc(549,48,1).
txt_doc(935,48,1).
txt_doc(223,48,1).
txt_doc(844,48,2).
txt_doc(610,48,1).
txt_doc(847,48,1).
txt_doc(936,48,1).
txt_doc(884,48,1).
txt_doc(599,49,1).
txt_doc(173,49,1).
txt_doc(59,49,1).
txt_doc(857,49,1).
txt_doc(125,49,1).
txt_doc(792,49,1).
txt_doc(246,49,1).
txt_doc(906,49,1).
txt_doc(418,49,1).
txt_doc(855,49,1).
txt_doc(426,49,1).
txt_doc(555,49,1).
txt_doc(548,49,1).
txt_doc(472,49,1).
txt_doc(834,49,1).
txt_doc(109,49,1).
txt_doc(392,49,1).
txt_doc(787,49,1).
txt_doc(558,49,1).
txt_doc(29,49,1).
txt_doc(742,49,1).
txt_doc(506,49,1).
txt_doc(816,49,1).
txt_doc(890,49,2).
txt_doc(286,49,1).
txt_doc(20,49,1).
txt_doc(68,49,1).
txt_doc(278,49,1).
txt_doc(678,49,1).
txt_doc(617,49,1).
txt_doc(773,49,1).
txt_doc(658,49,1).
txt_doc(520,49,1).
txt_doc(4,49,2).
txt_doc(229,49,1).
txt_doc(932,49,1).
txt_doc(560,49,1).
txt_doc(26,49,1).
txt_doc(7,49,1).
txt_doc(338,49,1).
txt_doc(35,49,1).
txt_doc(713,49,1).
txt_doc(804,49,2).
txt_doc(801,49,1).
txt_doc(1,49,1).
txt_doc(778,49,2).
txt_doc(829,49,1).
txt_doc(482,49,1).
txt_doc(737,49,1).
txt_doc(405,49,1).
txt_doc(843,49,1).
txt_doc(425,49,1).
txt_doc(94,49,1).
txt_doc(842,49,1).
txt_doc(404,49,1).
txt_doc(48,49,1).
txt_doc(99,49,1).
txt_doc(744,49,1).
txt_doc(123,49,1).
txt_doc(668,49,1).
txt_doc(183,49,1).
txt_doc(98,49,1).
txt_doc(534,49,1).
txt_doc(13,49,1).
txt_doc(157,49,1).
txt_doc(523,49,1).
txt_doc(545,49,1).
txt_doc(887,49,1).
txt_doc(701,49,1).
txt_doc(325,49,2).
txt_doc(41,49,1).
txt_doc(566,49,1).
txt_doc(962,49,1).
txt_doc(501,49,2).
txt_doc(940,49,1).
txt_doc(163,49,2).
txt_doc(725,49,1).
txt_doc(469,49,1).
txt_doc(199,49,1).
txt_doc(697,49,1).
txt_doc(556,49,1).
txt_doc(154,49,1).
txt_doc(81,49,1).
txt_doc(147,49,1).
txt_doc(551,49,1).
txt_doc(839,49,1).
txt_doc(403,49,1).
txt_doc(623,49,1).
txt_doc(378,49,2).
txt_doc(108,49,1).
txt_doc(484,49,1).
txt_doc(503,49,1).
txt_doc(272,49,1).
txt_doc(727,49,1).
txt_doc(174,49,3).
txt_doc(419,49,1).
txt_doc(223,49,1).
txt_doc(636,49,1).
txt_doc(952,49,1).
txt_doc(645,49,1).
txt_doc(687,49,1).
txt_doc(496,49,1).
txt_doc(252,49,1).
txt_doc(80,49,1).
txt_doc(187,49,1).
txt_doc(437,49,1).
txt_doc(232,49,1).
txt_doc(54,49,1).
txt_doc(302,49,1).
txt_doc(896,49,1).
txt_doc(176,49,2).
txt_doc(771,49,2).
txt_doc(111,49,2).
txt_doc(788,49,1).
txt_doc(688,49,1).
txt_doc(165,49,1).
txt_doc(540,49,1).
txt_doc(148,49,1).
txt_doc(581,49,2).
txt_doc(753,49,1).
txt_doc(620,49,1).
txt_doc(310,49,1).
txt_doc(106,49,1).
txt_doc(197,49,1).
txt_doc(275,49,1).
txt_doc(869,49,1).
txt_doc(516,49,1).
txt_doc(872,49,1).
txt_doc(38,49,1).
txt_doc(731,49,1).
txt_doc(359,49,1).
txt_doc(399,49,1).
txt_doc(254,49,1).
txt_doc(465,49,2).
txt_doc(946,49,1).
txt_doc(569,49,1).
txt_doc(240,49,1).
txt_doc(184,49,1).
txt_doc(767,49,1).
txt_doc(680,49,1).
txt_doc(313,49,1).
txt_doc(457,49,1).
txt_doc(480,49,1).
txt_doc(619,49,1).
txt_doc(606,49,1).
txt_doc(488,49,1).
txt_doc(954,49,1).
txt_doc(427,49,1).
txt_doc(859,49,1).
txt_doc(142,49,1).
txt_doc(964,49,1).
txt_doc(292,49,1).
txt_doc(5,49,1).
txt_doc(646,49,1).
txt_doc(622,49,1).
txt_doc(87,49,1).
txt_doc(818,49,1).
txt_doc(487,49,1).
txt_doc(476,49,1).
txt_doc(413,49,2).
txt_doc(703,49,1).
txt_doc(722,49,1).
txt_doc(44,49,1).
txt_doc(192,49,2).
txt_doc(604,49,1).
txt_doc(938,49,1).
txt_doc(460,49,1).
txt_doc(498,49,1).
txt_doc(393,49,2).
txt_doc(112,49,1).
txt_doc(190,49,2).
txt_doc(698,49,1).
txt_doc(101,49,1).
txt_doc(830,49,1).
txt_doc(583,49,1).
txt_doc(746,49,2).
txt_doc(398,49,1).
txt_doc(337,49,3).
txt_doc(43,50,1).
txt_doc(475,50,1).
txt_doc(328,50,1).
txt_doc(493,50,1).
txt_doc(678,50,1).
txt_doc(244,50,1).
txt_doc(204,50,1).
txt_doc(929,50,1).
txt_doc(353,50,1).
txt_doc(763,50,1).
txt_doc(705,50,1).
txt_doc(537,50,1).
txt_doc(593,50,1).
txt_doc(454,50,1).
txt_doc(942,50,1).
txt_doc(637,50,1).
txt_doc(952,50,1).
txt_doc(795,50,1).
txt_doc(367,50,1).
txt_doc(677,50,1).
txt_doc(772,50,1).
txt_doc(679,50,1).
txt_doc(416,50,1).
txt_doc(514,50,1).
txt_doc(129,50,1).
txt_doc(933,50,1).
txt_doc(745,50,1).
txt_doc(511,50,1).
txt_doc(652,50,1).
txt_doc(408,50,1).
txt_doc(920,50,1).
txt_doc(395,50,1).
txt_doc(251,50,1).
txt_doc(310,50,1).
txt_doc(92,50,1).
txt_doc(322,50,1).
txt_doc(183,50,1).
txt_doc(699,50,1).
txt_doc(421,50,1).
txt_doc(241,50,1).
txt_doc(636,50,1).
txt_doc(485,50,1).
txt_doc(571,50,1).
txt_doc(103,50,1).
txt_doc(331,50,1).
txt_doc(314,50,1).
txt_doc(363,50,1).
txt_doc(609,50,1).
txt_doc(346,50,1).
txt_doc(349,50,1).
txt_doc(323,50,1).
txt_doc(903,50,1).
txt_doc(161,50,1).
txt_doc(370,50,1).
txt_doc(1,50,1).
txt_doc(95,50,1).
txt_doc(90,50,2).
txt_doc(779,50,1).
txt_doc(410,50,1).
txt_doc(313,50,1).
txt_doc(556,50,1).
txt_doc(396,50,1).
txt_doc(918,50,1).
txt_doc(452,50,1).
txt_doc(703,50,1).
txt_doc(451,50,1).
txt_doc(57,50,1).
txt_doc(138,50,1).
txt_doc(426,50,1).
txt_doc(840,50,1).
txt_doc(692,50,1).
txt_doc(521,50,1).
txt_doc(199,50,1).
txt_doc(809,50,1).
txt_doc(604,50,1).
txt_doc(814,50,1).
txt_doc(302,50,1).
txt_doc(572,50,2).
txt_doc(119,50,2).
txt_doc(216,50,1).
txt_doc(905,50,1).
txt_doc(901,50,1).
txt_doc(782,50,1).
txt_doc(909,50,1).
txt_doc(601,50,1).
txt_doc(193,50,1).
txt_doc(924,50,1).
txt_doc(724,50,1).
txt_doc(124,50,1).
txt_doc(799,50,1).
txt_doc(847,50,1).
txt_doc(384,50,1).
txt_doc(115,50,2).
txt_doc(518,50,1).
txt_doc(600,50,1).
txt_doc(769,50,1).
txt_doc(233,50,1).
txt_doc(18,50,2).
txt_doc(320,50,1).
txt_doc(12,50,1).
txt_doc(881,50,1).
txt_doc(317,50,2).
txt_doc(160,50,1).
txt_doc(885,50,1).
txt_doc(281,50,1).
txt_doc(394,50,1).
txt_doc(529,50,1).
txt_doc(455,50,1).
txt_doc(133,50,1).
txt_doc(961,50,1).
txt_doc(359,50,1).
txt_doc(785,50,1).
txt_doc(658,50,1).
txt_doc(147,50,1).
txt_doc(598,50,1).
txt_doc(456,50,2).
txt_doc(756,50,1).
txt_doc(527,50,1).
txt_doc(661,50,1).
txt_doc(813,50,1).
txt_doc(887,50,1).
txt_doc(166,50,2).
txt_doc(209,50,1).
txt_doc(486,50,1).
txt_doc(704,50,2).
txt_doc(375,50,1).
txt_doc(654,50,1).
txt_doc(256,50,2).
txt_doc(687,50,2).
txt_doc(46,50,1).
txt_doc(957,50,2).
txt_doc(382,50,1).
txt_doc(347,50,1).
txt_doc(431,50,1).
txt_doc(239,50,1).
txt_doc(212,50,1).
txt_doc(146,50,2).
txt_doc(869,50,2).
txt_doc(890,50,2).
txt_doc(230,50,1).
txt_doc(701,50,1).
txt_doc(611,50,1).
txt_doc(425,50,2).
txt_doc(685,50,1).
txt_doc(64,50,1).
txt_doc(489,50,1).
txt_doc(32,50,2).
txt_doc(245,50,1).
txt_doc(576,50,1).
txt_doc(65,50,1).
txt_doc(253,50,1).
txt_doc(93,50,1).
txt_doc(917,50,1).
txt_doc(379,50,1).
txt_doc(484,50,1).
txt_doc(201,50,1).
txt_doc(880,50,2).
txt_doc(58,50,1).
txt_doc(26,50,1).
txt_doc(98,50,1).
txt_doc(189,50,1).
txt_doc(368,50,1).
txt_doc(904,50,1).
txt_doc(548,50,1).
txt_doc(937,50,1).
txt_doc(791,50,1).
txt_doc(853,50,1).
txt_doc(36,50,1).
txt_doc(354,50,1).
txt_doc(196,50,2).
txt_doc(727,50,1).
txt_doc(503,50,1).
txt_doc(121,50,2).
txt_doc(688,50,1).
txt_doc(888,50,1).
txt_doc(624,50,1).
txt_doc(215,50,1).
txt_doc(513,50,1).
txt_doc(205,50,2).
txt_doc(233,51,1).
txt_doc(249,51,1).
txt_doc(271,51,1).
txt_doc(488,51,1).
txt_doc(310,51,1).
txt_doc(634,51,1).
txt_doc(5,51,1).
txt_doc(942,51,1).
txt_doc(800,51,1).
txt_doc(193,51,1).
txt_doc(740,51,1).
txt_doc(470,51,1).
txt_doc(702,51,1).
txt_doc(781,51,1).
txt_doc(480,51,1).
txt_doc(520,51,1).
txt_doc(94,51,1).
txt_doc(200,51,1).
txt_doc(115,51,1).
txt_doc(529,51,1).
txt_doc(227,51,1).
txt_doc(122,51,1).
txt_doc(951,51,1).
txt_doc(852,51,1).
txt_doc(626,51,1).
txt_doc(232,51,1).
txt_doc(259,51,1).
txt_doc(16,51,1).
txt_doc(479,51,1).
txt_doc(366,51,1).
txt_doc(406,51,1).
txt_doc(856,51,1).
txt_doc(738,51,1).
txt_doc(652,51,1).
txt_doc(591,51,1).
txt_doc(191,51,1).
txt_doc(866,51,1).
txt_doc(263,51,1).
txt_doc(475,51,2).
txt_doc(162,51,1).
txt_doc(846,51,1).
txt_doc(148,51,1).
txt_doc(947,51,1).
txt_doc(14,51,1).
txt_doc(278,51,1).
txt_doc(45,51,1).
txt_doc(481,51,1).
txt_doc(235,51,1).
txt_doc(172,51,1).
txt_doc(507,51,1).
txt_doc(485,51,1).
txt_doc(97,51,1).
txt_doc(258,51,1).
txt_doc(256,51,2).
txt_doc(210,51,1).
txt_doc(446,51,1).
txt_doc(610,51,1).
txt_doc(655,51,1).
txt_doc(409,51,1).
txt_doc(294,51,1).
txt_doc(526,51,1).
txt_doc(405,51,1).
txt_doc(451,51,1).
txt_doc(564,51,2).
txt_doc(346,51,1).
txt_doc(383,51,1).
txt_doc(958,51,1).
txt_doc(938,51,1).
txt_doc(877,51,1).
txt_doc(859,51,1).
txt_doc(120,51,1).
txt_doc(848,51,1).
txt_doc(612,51,1).
txt_doc(495,51,1).
txt_doc(19,51,1).
txt_doc(730,51,2).
txt_doc(562,51,1).
txt_doc(307,51,1).
txt_doc(220,51,1).
txt_doc(456,51,1).
txt_doc(153,51,1).
txt_doc(962,51,2).
txt_doc(88,51,1).
txt_doc(7,51,1).
txt_doc(219,51,1).
txt_doc(419,51,1).
txt_doc(638,51,1).
txt_doc(515,51,1).
txt_doc(819,51,1).
txt_doc(135,51,1).
txt_doc(281,51,1).
txt_doc(359,51,1).
txt_doc(579,51,1).
txt_doc(264,51,1).
txt_doc(742,51,1).
txt_doc(317,51,1).
txt_doc(28,51,2).
txt_doc(812,51,1).
txt_doc(514,51,1).
txt_doc(749,51,1).
txt_doc(437,51,1).
txt_doc(716,51,1).
txt_doc(361,51,1).
txt_doc(952,51,1).
txt_doc(78,51,1).
txt_doc(808,51,1).
txt_doc(340,51,1).
txt_doc(303,51,1).
txt_doc(498,51,1).
txt_doc(679,51,2).
txt_doc(794,51,1).
txt_doc(509,51,2).
txt_doc(721,51,1).
txt_doc(707,51,1).
txt_doc(768,51,1).
txt_doc(69,51,1).
txt_doc(175,51,1).
txt_doc(517,51,1).
txt_doc(149,51,1).
txt_doc(801,51,1).
txt_doc(483,51,1).
txt_doc(277,51,1).
txt_doc(823,51,1).
txt_doc(354,51,2).
txt_doc(739,51,1).
txt_doc(344,51,1).
txt_doc(574,51,1).
txt_doc(864,51,2).
txt_doc(766,51,1).
txt_doc(642,51,1).
txt_doc(203,51,1).
txt_doc(790,51,1).
txt_doc(76,51,1).
txt_doc(607,51,1).
txt_doc(266,51,1).
txt_doc(911,51,1).
txt_doc(926,51,1).
txt_doc(21,51,1).
txt_doc(617,51,1).
txt_doc(941,51,1).
txt_doc(494,51,1).
txt_doc(74,51,1).
txt_doc(540,51,1).
txt_doc(568,51,1).
txt_doc(236,51,1).
txt_doc(226,51,1).
txt_doc(893,51,2).
txt_doc(657,51,1).
txt_doc(234,51,1).
txt_doc(95,51,1).
txt_doc(774,51,1).
txt_doc(306,51,1).
txt_doc(496,51,2).
txt_doc(254,51,1).
txt_doc(46,51,1).
txt_doc(339,51,1).
txt_doc(455,51,1).
txt_doc(231,51,2).
txt_doc(839,51,1).
txt_doc(729,51,1).
txt_doc(80,51,1).
txt_doc(553,51,1).
txt_doc(425,51,1).
txt_doc(922,51,1).
txt_doc(855,51,1).
txt_doc(832,51,1).
txt_doc(759,51,1).
txt_doc(463,51,1).
txt_doc(41,51,1).
txt_doc(837,51,1).
txt_doc(689,51,1).
txt_doc(183,51,1).
txt_doc(957,51,1).
txt_doc(385,51,2).
txt_doc(438,51,1).
txt_doc(847,51,1).
txt_doc(275,51,1).
txt_doc(392,51,1).
txt_doc(966,51,1).
txt_doc(946,51,1).
txt_doc(279,51,1).
txt_doc(468,51,1).
txt_doc(180,51,1).
txt_doc(593,51,1).
txt_doc(207,51,1).
txt_doc(168,51,1).
txt_doc(274,52,1).
txt_doc(498,52,1).
txt_doc(836,52,1).
txt_doc(964,52,1).
txt_doc(227,52,1).
txt_doc(957,52,1).
txt_doc(585,52,1).
txt_doc(127,52,1).
txt_doc(226,52,1).
txt_doc(128,52,1).
txt_doc(80,52,1).
txt_doc(83,52,1).
txt_doc(359,52,1).
txt_doc(190,52,1).
txt_doc(146,52,1).
txt_doc(640,52,1).
txt_doc(727,52,1).
txt_doc(30,52,1).
txt_doc(586,52,1).
txt_doc(351,52,1).
txt_doc(341,52,1).
txt_doc(911,52,1).
txt_doc(382,52,1).
txt_doc(733,52,1).
txt_doc(669,52,1).
txt_doc(738,52,1).
txt_doc(25,52,1).
txt_doc(690,52,1).
txt_doc(24,52,1).
txt_doc(634,52,2).
txt_doc(220,52,1).
txt_doc(472,52,1).
txt_doc(641,52,1).
txt_doc(954,52,1).
txt_doc(878,52,1).
txt_doc(789,52,1).
txt_doc(921,52,1).
txt_doc(292,52,1).
txt_doc(694,52,1).
txt_doc(711,52,1).
txt_doc(58,52,1).
txt_doc(18,52,1).
txt_doc(269,52,1).
txt_doc(642,52,1).
txt_doc(743,52,1).
txt_doc(655,52,1).
txt_doc(679,52,1).
txt_doc(583,52,1).
txt_doc(275,52,1).
txt_doc(841,52,1).
txt_doc(531,52,1).
txt_doc(327,52,1).
txt_doc(800,52,1).
txt_doc(701,52,1).
txt_doc(497,52,1).
txt_doc(759,52,1).
txt_doc(902,52,1).
txt_doc(512,52,1).
txt_doc(778,52,1).
txt_doc(494,52,1).
txt_doc(42,52,2).
txt_doc(671,52,1).
txt_doc(726,52,1).
txt_doc(304,52,1).
txt_doc(392,52,1).
txt_doc(314,52,1).
txt_doc(41,52,1).
txt_doc(436,52,1).
txt_doc(682,52,1).
txt_doc(940,52,1).
txt_doc(76,52,2).
txt_doc(415,52,1).
txt_doc(535,52,1).
txt_doc(109,52,1).
txt_doc(473,52,1).
txt_doc(762,52,1).
txt_doc(215,52,1).
txt_doc(747,52,1).
txt_doc(876,52,1).
txt_doc(491,52,1).
txt_doc(945,52,2).
txt_doc(231,52,1).
txt_doc(334,52,1).
txt_doc(479,52,1).
txt_doc(636,52,1).
txt_doc(151,52,1).
txt_doc(823,52,1).
txt_doc(111,52,1).
txt_doc(454,52,1).
txt_doc(956,52,1).
txt_doc(364,52,1).
txt_doc(37,52,1).
txt_doc(917,52,1).
txt_doc(71,52,1).
txt_doc(838,52,1).
txt_doc(187,52,1).
txt_doc(175,52,1).
txt_doc(904,52,1).
txt_doc(610,52,1).
txt_doc(523,52,1).
txt_doc(391,52,1).
txt_doc(909,52,1).
txt_doc(736,52,1).
txt_doc(530,52,1).
txt_doc(438,52,1).
txt_doc(942,52,1).
txt_doc(890,52,1).
txt_doc(706,52,1).
txt_doc(279,52,2).
txt_doc(611,52,1).
txt_doc(946,52,1).
txt_doc(863,52,1).
txt_doc(558,52,1).
txt_doc(57,52,1).
txt_doc(907,52,1).
txt_doc(696,52,1).
txt_doc(688,52,1).
txt_doc(646,52,1).
txt_doc(103,52,1).
txt_doc(901,52,1).
txt_doc(595,52,1).
txt_doc(266,52,1).
txt_doc(79,52,3).
txt_doc(433,52,1).
txt_doc(349,52,1).
txt_doc(657,52,1).
txt_doc(381,52,1).
txt_doc(357,52,1).
txt_doc(325,52,1).
txt_doc(221,52,1).
txt_doc(316,52,1).
txt_doc(407,52,1).
txt_doc(98,52,2).
txt_doc(775,52,1).
txt_doc(112,52,1).
txt_doc(785,52,1).
txt_doc(874,52,1).
txt_doc(138,52,1).
txt_doc(229,52,1).
txt_doc(10,52,2).
txt_doc(686,52,2).
txt_doc(469,52,1).
txt_doc(944,52,1).
txt_doc(960,52,1).
txt_doc(174,52,2).
txt_doc(790,52,1).
txt_doc(74,52,2).
txt_doc(833,52,1).
txt_doc(419,52,1).
txt_doc(537,52,1).
txt_doc(85,52,1).
txt_doc(704,52,1).
txt_doc(797,52,2).
txt_doc(271,52,1).
txt_doc(897,52,1).
txt_doc(249,52,1).
txt_doc(118,52,1).
txt_doc(534,52,2).
txt_doc(886,52,1).
txt_doc(92,52,2).
txt_doc(437,52,2).
txt_doc(654,52,1).
txt_doc(338,52,1).
txt_doc(306,52,2).
txt_doc(56,52,1).
txt_doc(211,52,1).
txt_doc(13,52,1).
txt_doc(198,52,1).
txt_doc(506,52,1).
txt_doc(263,52,1).
txt_doc(502,52,1).
txt_doc(934,52,1).
txt_doc(810,52,2).
txt_doc(375,52,1).
txt_doc(952,52,1).
txt_doc(526,52,1).
txt_doc(288,52,1).
txt_doc(668,52,2).
txt_doc(799,52,1).
txt_doc(495,52,1).
txt_doc(630,52,1).
txt_doc(476,53,1).
txt_doc(85,53,1).
txt_doc(261,53,1).
txt_doc(32,53,1).
txt_doc(138,53,1).
txt_doc(737,53,1).
txt_doc(822,53,1).
txt_doc(110,53,1).
txt_doc(464,53,1).
txt_doc(911,53,1).
txt_doc(589,53,1).
txt_doc(299,53,1).
txt_doc(199,53,1).
txt_doc(581,53,1).
txt_doc(240,53,1).
txt_doc(121,53,1).
txt_doc(417,53,1).
txt_doc(604,53,1).
txt_doc(566,53,1).
txt_doc(266,53,1).
txt_doc(478,53,1).
txt_doc(906,53,1).
txt_doc(569,53,1).
txt_doc(748,53,1).
txt_doc(119,53,1).
txt_doc(400,53,1).
txt_doc(466,53,1).
txt_doc(228,53,2).
txt_doc(330,53,1).
txt_doc(550,53,1).
txt_doc(535,53,1).
txt_doc(568,53,1).
txt_doc(84,53,1).
txt_doc(708,53,1).
txt_doc(300,53,1).
txt_doc(615,53,1).
txt_doc(166,53,1).
txt_doc(625,53,1).
txt_doc(533,53,1).
txt_doc(706,53,1).
txt_doc(526,53,1).
txt_doc(130,53,1).
txt_doc(319,53,1).
txt_doc(457,53,1).
txt_doc(630,53,1).
txt_doc(101,53,1).
txt_doc(954,53,1).
txt_doc(675,53,1).
txt_doc(320,53,1).
txt_doc(490,53,1).
txt_doc(382,53,1).
txt_doc(555,53,1).
txt_doc(714,53,2).
txt_doc(514,53,1).
txt_doc(770,53,1).
txt_doc(49,53,1).
txt_doc(766,53,1).
txt_doc(295,53,1).
txt_doc(725,53,1).
txt_doc(851,53,1).
txt_doc(599,53,1).
txt_doc(116,53,1).
txt_doc(919,53,1).
txt_doc(103,53,1).
txt_doc(81,53,2).
txt_doc(592,53,1).
txt_doc(810,53,1).
txt_doc(473,53,1).
txt_doc(229,53,1).
txt_doc(60,53,1).
txt_doc(51,53,1).
txt_doc(349,53,1).
txt_doc(791,53,1).
txt_doc(276,53,1).
txt_doc(440,53,1).
txt_doc(96,53,1).
txt_doc(83,53,1).
txt_doc(785,53,1).
txt_doc(270,53,1).
txt_doc(710,53,1).
txt_doc(505,53,1).
txt_doc(429,53,1).
txt_doc(414,53,1).
txt_doc(4,53,1).
txt_doc(255,53,1).
txt_doc(536,53,1).
txt_doc(728,53,1).
txt_doc(92,53,1).
txt_doc(551,53,1).
txt_doc(87,53,1).
txt_doc(45,53,1).
txt_doc(52,53,1).
txt_doc(749,53,1).
txt_doc(291,53,2).
txt_doc(420,53,1).
txt_doc(374,53,1).
txt_doc(274,53,1).
txt_doc(698,53,1).
txt_doc(157,53,1).
txt_doc(180,53,1).
txt_doc(55,53,1).
txt_doc(359,53,1).
txt_doc(322,53,2).
txt_doc(720,53,1).
txt_doc(284,53,1).
txt_doc(189,53,1).
txt_doc(11,53,1).
txt_doc(396,53,1).
txt_doc(214,53,1).
txt_doc(113,53,1).
txt_doc(215,53,1).
txt_doc(390,53,1).
txt_doc(719,53,1).
txt_doc(564,53,2).
txt_doc(909,53,1).
txt_doc(76,53,1).
txt_doc(670,53,1).
txt_doc(136,53,1).
txt_doc(247,53,1).
txt_doc(129,53,1).
txt_doc(815,53,1).
txt_doc(279,53,1).
txt_doc(612,53,1).
txt_doc(645,53,1).
txt_doc(786,53,1).
txt_doc(267,53,2).
txt_doc(231,53,1).
txt_doc(886,53,1).
txt_doc(115,53,1).
txt_doc(777,53,2).
txt_doc(358,53,1).
txt_doc(256,53,1).
txt_doc(913,53,1).
txt_doc(375,53,1).
txt_doc(140,53,3).
txt_doc(432,53,2).
txt_doc(861,53,2).
txt_doc(20,53,1).
txt_doc(754,53,1).
txt_doc(743,53,2).
txt_doc(552,53,1).
txt_doc(441,53,1).
txt_doc(293,53,1).
txt_doc(482,53,1).
txt_doc(297,53,1).
txt_doc(914,53,1).
txt_doc(679,53,2).
txt_doc(72,53,2).
txt_doc(294,53,1).
txt_doc(779,53,2).
txt_doc(207,53,1).
txt_doc(182,53,1).
txt_doc(262,53,1).
txt_doc(578,53,1).
txt_doc(753,53,1).
txt_doc(912,53,2).
txt_doc(933,53,1).
txt_doc(515,53,2).
txt_doc(165,53,1).
txt_doc(149,53,1).
txt_doc(366,53,1).
txt_doc(884,53,1).
txt_doc(137,53,1).
txt_doc(721,53,1).
txt_doc(22,53,1).
txt_doc(398,53,1).
txt_doc(316,53,1).
txt_doc(439,53,1).
txt_doc(100,53,1).
txt_doc(855,53,1).
txt_doc(280,53,1).
txt_doc(966,53,2).
txt_doc(605,53,1).
txt_doc(959,53,1).
txt_doc(793,53,1).
txt_doc(249,53,1).
txt_doc(528,53,1).
txt_doc(947,53,1).
txt_doc(27,53,1).
txt_doc(278,53,2).
txt_doc(639,54,1).
txt_doc(528,54,1).
txt_doc(759,54,1).
txt_doc(525,54,1).
txt_doc(917,54,1).
txt_doc(579,54,1).
txt_doc(113,54,1).
txt_doc(689,54,1).
txt_doc(538,54,1).
txt_doc(857,54,1).
txt_doc(274,54,1).
txt_doc(405,54,1).
txt_doc(173,54,1).
txt_doc(167,54,1).
txt_doc(932,54,1).
txt_doc(17,54,1).
txt_doc(416,54,1).
txt_doc(524,54,1).
txt_doc(402,54,1).
txt_doc(242,54,1).
txt_doc(511,54,1).
txt_doc(475,54,1).
txt_doc(251,54,1).
txt_doc(197,54,1).
txt_doc(226,54,1).
txt_doc(558,54,1).
txt_doc(160,54,1).
txt_doc(73,54,1).
txt_doc(161,54,1).
txt_doc(18,54,1).
txt_doc(275,54,1).
txt_doc(907,54,1).
txt_doc(294,54,1).
txt_doc(441,54,1).
txt_doc(451,54,1).
txt_doc(507,54,1).
txt_doc(965,54,1).
txt_doc(820,54,1).
txt_doc(687,54,1).
txt_doc(667,54,1).
txt_doc(457,54,1).
txt_doc(509,54,1).
txt_doc(777,54,1).
txt_doc(144,54,1).
txt_doc(278,54,1).
txt_doc(708,54,1).
txt_doc(590,54,1).
txt_doc(169,54,2).
txt_doc(819,54,1).
txt_doc(46,54,1).
txt_doc(904,54,2).
txt_doc(439,54,1).
txt_doc(498,54,1).
txt_doc(296,54,1).
txt_doc(513,54,1).
txt_doc(731,54,1).
txt_doc(230,54,1).
txt_doc(310,54,1).
txt_doc(21,54,1).
txt_doc(724,54,1).
txt_doc(604,54,1).
txt_doc(769,54,1).
txt_doc(342,54,1).
txt_doc(620,54,1).
txt_doc(192,54,1).
txt_doc(345,54,1).
txt_doc(96,54,1).
txt_doc(754,54,1).
txt_doc(844,54,1).
txt_doc(62,54,1).
txt_doc(923,54,1).
txt_doc(231,54,1).
txt_doc(968,54,1).
txt_doc(435,54,1).
txt_doc(952,54,1).
txt_doc(873,54,1).
txt_doc(219,54,1).
txt_doc(868,54,1).
txt_doc(837,54,1).
txt_doc(304,54,1).
txt_doc(142,54,1).
txt_doc(139,54,1).
txt_doc(835,54,1).
txt_doc(657,54,1).
txt_doc(562,54,1).
txt_doc(702,54,1).
txt_doc(417,54,1).
txt_doc(358,54,1).
txt_doc(67,54,1).
txt_doc(931,54,2).
txt_doc(81,54,1).
txt_doc(884,54,1).
txt_doc(442,54,1).
txt_doc(440,54,1).
txt_doc(136,54,2).
txt_doc(121,54,1).
txt_doc(787,54,1).
txt_doc(207,54,1).
txt_doc(569,54,1).
txt_doc(762,54,1).
txt_doc(6,54,1).
txt_doc(172,54,1).
txt_doc(576,54,1).
txt_doc(821,54,1).
txt_doc(61,54,1).
txt_doc(263,54,1).
txt_doc(582,54,1).
txt_doc(16,54,1).
txt_doc(291,54,2).
txt_doc(831,54,1).
txt_doc(59,54,1).
txt_doc(531,54,1).
txt_doc(170,54,1).
txt_doc(298,54,1).
txt_doc(7,54,1).
txt_doc(64,54,1).
txt_doc(632,54,1).
txt_doc(800,54,1).
txt_doc(307,54,1).
txt_doc(934,54,2).
txt_doc(940,54,1).
txt_doc(228,54,1).
txt_doc(705,54,1).
txt_doc(715,54,1).
txt_doc(629,54,1).
txt_doc(848,54,1).
txt_doc(878,54,2).
txt_doc(196,54,2).
txt_doc(712,54,1).
txt_doc(176,54,1).
txt_doc(814,54,1).
txt_doc(499,54,1).
txt_doc(34,54,1).
txt_doc(798,54,1).
txt_doc(713,54,2).
txt_doc(547,54,1).
txt_doc(823,54,2).
txt_doc(865,54,1).
txt_doc(764,54,1).
txt_doc(706,54,1).
txt_doc(852,54,1).
txt_doc(832,54,1).
txt_doc(455,54,1).
txt_doc(386,54,1).
txt_doc(926,54,1).
txt_doc(394,54,2).
txt_doc(387,54,2).
txt_doc(195,54,1).
txt_doc(122,54,1).
txt_doc(810,54,1).
txt_doc(503,54,2).
txt_doc(213,54,1).
txt_doc(661,54,1).
txt_doc(760,54,1).
txt_doc(721,54,2).
txt_doc(348,54,1).
txt_doc(508,54,2).
txt_doc(40,54,1).
txt_doc(372,54,1).
txt_doc(464,54,1).
txt_doc(444,54,1).
txt_doc(429,54,1).
txt_doc(905,54,1).
txt_doc(517,54,1).
txt_doc(736,54,1).
txt_doc(146,54,1).
txt_doc(641,54,1).
txt_doc(29,54,1).
txt_doc(198,54,2).
txt_doc(922,54,2).
txt_doc(750,54,2).
txt_doc(665,54,1).
txt_doc(722,54,1).
txt_doc(555,54,1).
txt_doc(805,54,1).
txt_doc(794,54,3).
txt_doc(293,54,2).
txt_doc(489,54,1).
txt_doc(401,54,1).
txt_doc(386,55,1).
txt_doc(682,55,1).
txt_doc(545,55,1).
txt_doc(569,55,1).
txt_doc(698,55,1).
txt_doc(36,55,1).
txt_doc(512,55,1).
txt_doc(754,55,1).
txt_doc(275,55,1).
txt_doc(732,55,1).
txt_doc(769,55,1).
txt_doc(746,55,1).
txt_doc(475,55,1).
txt_doc(159,55,1).
txt_doc(850,55,1).
txt_doc(119,55,1).
txt_doc(408,55,1).
txt_doc(823,55,1).
txt_doc(35,55,1).
txt_doc(313,55,2).
txt_doc(24,55,1).
txt_doc(182,55,1).
txt_doc(854,55,1).
txt_doc(54,55,1).
txt_doc(373,55,1).
txt_doc(105,55,1).
txt_doc(842,55,1).
txt_doc(347,55,1).
txt_doc(859,55,1).
txt_doc(780,55,2).
txt_doc(397,55,1).
txt_doc(949,55,1).
txt_doc(551,55,1).
txt_doc(690,55,1).
txt_doc(53,55,1).
txt_doc(937,55,1).
txt_doc(78,55,1).
txt_doc(307,55,1).
txt_doc(84,55,1).
txt_doc(525,55,1).
txt_doc(709,55,1).
txt_doc(485,55,1).
txt_doc(282,55,1).
txt_doc(795,55,1).
txt_doc(786,55,1).
txt_doc(70,55,1).
txt_doc(424,55,1).
txt_doc(775,55,1).
txt_doc(687,55,1).
txt_doc(904,55,1).
txt_doc(652,55,1).
txt_doc(558,55,1).
txt_doc(58,55,1).
txt_doc(403,55,1).
txt_doc(587,55,1).
txt_doc(392,55,1).
txt_doc(742,55,1).
txt_doc(955,55,1).
txt_doc(420,55,1).
txt_doc(295,55,1).
txt_doc(894,55,1).
txt_doc(335,55,1).
txt_doc(141,55,1).
txt_doc(694,55,1).
txt_doc(108,55,1).
txt_doc(497,55,1).
txt_doc(341,55,1).
txt_doc(731,55,1).
txt_doc(776,55,1).
txt_doc(378,55,1).
txt_doc(13,55,1).
txt_doc(429,55,1).
txt_doc(296,55,1).
txt_doc(6,55,1).
txt_doc(689,55,1).
txt_doc(164,55,1).
txt_doc(578,55,1).
txt_doc(729,55,1).
txt_doc(594,55,1).
txt_doc(298,55,1).
txt_doc(10,55,1).
txt_doc(343,55,1).
txt_doc(655,55,1).
txt_doc(751,55,1).
txt_doc(671,55,2).
txt_doc(641,55,1).
txt_doc(674,55,1).
txt_doc(493,55,1).
txt_doc(733,55,1).
txt_doc(661,55,1).
txt_doc(932,55,1).
txt_doc(901,55,1).
txt_doc(518,55,1).
txt_doc(800,55,1).
txt_doc(411,55,1).
txt_doc(15,55,1).
txt_doc(705,55,1).
txt_doc(870,55,1).
txt_doc(860,55,1).
txt_doc(293,55,1).
txt_doc(318,55,1).
txt_doc(835,55,1).
txt_doc(703,55,1).
txt_doc(494,55,1).
txt_doc(658,55,1).
txt_doc(526,55,1).
txt_doc(599,55,1).
txt_doc(660,55,1).
txt_doc(724,55,1).
txt_doc(873,55,1).
txt_doc(580,55,1).
txt_doc(774,55,1).
txt_doc(864,55,1).
txt_doc(367,55,1).
txt_doc(685,55,1).
txt_doc(718,55,1).
txt_doc(27,55,1).
txt_doc(888,55,1).
txt_doc(505,55,1).
txt_doc(632,55,1).
txt_doc(585,55,1).
txt_doc(428,55,1).
txt_doc(902,55,1).
txt_doc(288,55,1).
txt_doc(45,55,1).
txt_doc(695,55,1).
txt_doc(255,55,1).
txt_doc(441,55,1).
txt_doc(879,55,1).
txt_doc(107,55,1).
txt_doc(900,55,1).
txt_doc(592,55,1).
txt_doc(929,55,1).
txt_doc(158,55,1).
txt_doc(287,55,1).
txt_doc(235,55,1).
txt_doc(115,55,2).
txt_doc(804,55,1).
txt_doc(154,55,1).
txt_doc(697,55,1).
txt_doc(647,55,1).
txt_doc(380,55,1).
txt_doc(945,55,1).
txt_doc(490,55,1).
txt_doc(898,55,1).
txt_doc(536,55,1).
txt_doc(165,55,1).
txt_doc(40,55,1).
txt_doc(59,55,1).
txt_doc(284,55,1).
txt_doc(150,55,1).
txt_doc(883,55,1).
txt_doc(148,55,1).
txt_doc(472,55,1).
txt_doc(557,55,1).
txt_doc(919,55,1).
txt_doc(144,55,2).
txt_doc(112,55,1).
txt_doc(711,55,1).
txt_doc(943,55,1).
txt_doc(342,55,1).
txt_doc(51,55,2).
txt_doc(796,55,1).
txt_doc(89,55,1).
txt_doc(68,55,2).
txt_doc(664,55,1).
txt_doc(444,55,1).
txt_doc(353,55,1).
txt_doc(730,55,1).
txt_doc(328,55,1).
txt_doc(840,55,1).
txt_doc(160,55,1).
txt_doc(612,55,1).
txt_doc(492,55,1).
txt_doc(691,55,1).
txt_doc(605,55,1).
txt_doc(905,55,1).
txt_doc(964,55,3).
txt_doc(479,55,1).
txt_doc(802,55,2).
txt_doc(749,55,1).
txt_doc(186,55,2).
txt_doc(911,55,2).
txt_doc(254,55,1).
txt_doc(451,55,1).
txt_doc(292,55,1).
txt_doc(789,55,2).
txt_doc(634,56,1).
txt_doc(814,56,1).
txt_doc(23,56,1).
txt_doc(747,56,1).
txt_doc(661,56,1).
txt_doc(767,56,1).
txt_doc(733,56,1).
txt_doc(621,56,1).
txt_doc(328,56,1).
txt_doc(912,56,1).
txt_doc(541,56,1).
txt_doc(66,56,1).
txt_doc(498,56,1).
txt_doc(172,56,1).
txt_doc(466,56,1).
txt_doc(445,56,1).
txt_doc(253,56,1).
txt_doc(868,56,1).
txt_doc(844,56,1).
txt_doc(79,56,1).
txt_doc(433,56,1).
txt_doc(367,56,1).
txt_doc(586,56,1).
txt_doc(777,56,1).
txt_doc(55,56,1).
txt_doc(254,56,1).
txt_doc(825,56,1).
txt_doc(830,56,1).
txt_doc(394,56,2).
txt_doc(336,56,1).
txt_doc(183,56,1).
txt_doc(62,56,1).
txt_doc(215,56,1).
txt_doc(47,56,1).
txt_doc(572,56,1).
txt_doc(464,56,1).
txt_doc(762,56,1).
txt_doc(451,56,1).
txt_doc(811,56,2).
txt_doc(570,56,1).
txt_doc(304,56,1).
txt_doc(880,56,1).
txt_doc(715,56,1).
txt_doc(836,56,1).
txt_doc(526,56,1).
txt_doc(178,56,1).
txt_doc(568,56,1).
txt_doc(346,56,1).
txt_doc(638,56,1).
txt_doc(136,56,1).
txt_doc(932,56,1).
txt_doc(228,56,1).
txt_doc(59,56,1).
txt_doc(496,56,1).
txt_doc(878,56,2).
txt_doc(860,56,1).
txt_doc(375,56,1).
txt_doc(186,56,3).
txt_doc(185,56,1).
txt_doc(494,56,1).
txt_doc(961,56,1).
txt_doc(790,56,1).
txt_doc(476,56,1).
txt_doc(913,56,1).
txt_doc(903,56,1).
txt_doc(702,56,1).
txt_doc(352,56,1).
txt_doc(49,56,2).
txt_doc(160,56,1).
txt_doc(641,56,1).
txt_doc(46,56,1).
txt_doc(616,56,1).
txt_doc(41,56,1).
txt_doc(208,56,1).
txt_doc(773,56,1).
txt_doc(141,56,1).
txt_doc(75,56,1).
txt_doc(436,56,1).
txt_doc(139,56,1).
txt_doc(959,56,1).
txt_doc(314,56,2).
txt_doc(12,56,2).
txt_doc(205,56,1).
txt_doc(578,56,1).
txt_doc(615,56,3).
txt_doc(301,56,1).
txt_doc(826,56,1).
txt_doc(411,56,1).
txt_doc(143,56,1).
txt_doc(340,56,1).
txt_doc(133,56,1).
txt_doc(266,56,2).
txt_doc(426,56,1).
txt_doc(724,56,1).
txt_doc(364,56,1).
txt_doc(154,56,1).
txt_doc(296,56,1).
txt_doc(666,56,1).
txt_doc(678,56,1).
txt_doc(156,56,1).
txt_doc(390,56,2).
txt_doc(739,56,1).
txt_doc(422,56,1).
txt_doc(577,56,1).
txt_doc(418,56,1).
txt_doc(756,56,1).
txt_doc(934,56,1).
txt_doc(853,56,2).
txt_doc(552,56,1).
txt_doc(130,56,1).
txt_doc(123,56,1).
txt_doc(940,56,1).
txt_doc(378,56,1).
txt_doc(442,56,1).
txt_doc(687,56,1).
txt_doc(110,56,1).
txt_doc(456,56,1).
txt_doc(656,56,1).
txt_doc(945,56,1).
txt_doc(332,56,1).
txt_doc(339,56,1).
txt_doc(967,56,1).
txt_doc(276,56,1).
txt_doc(153,56,1).
txt_doc(430,56,1).
txt_doc(284,56,1).
txt_doc(206,56,1).
txt_doc(175,56,2).
txt_doc(905,56,1).
txt_doc(414,56,1).
txt_doc(44,56,1).
txt_doc(412,56,2).
txt_doc(226,56,1).
txt_doc(22,56,1).
txt_doc(493,56,2).
txt_doc(856,56,1).
txt_doc(512,56,2).
txt_doc(50,56,1).
txt_doc(753,56,1).
txt_doc(460,56,1).
txt_doc(798,56,1).
txt_doc(819,56,1).
txt_doc(386,56,1).
txt_doc(393,56,1).
txt_doc(473,56,1).
txt_doc(234,56,1).
txt_doc(859,56,1).
txt_doc(640,56,1).
txt_doc(698,56,1).
txt_doc(114,56,1).
txt_doc(57,56,1).
txt_doc(930,56,1).
txt_doc(429,56,1).
txt_doc(536,56,1).
txt_doc(723,56,1).
txt_doc(155,56,1).
txt_doc(288,56,2).
txt_doc(88,56,1).
txt_doc(272,56,1).
txt_doc(106,56,1).
txt_doc(38,56,1).
txt_doc(643,56,1).
txt_doc(159,56,1).
txt_doc(794,56,2).
txt_doc(815,56,1).
txt_doc(406,56,1).
txt_doc(770,56,1).
txt_doc(644,56,2).
txt_doc(565,56,1).
txt_doc(31,56,1).
txt_doc(374,56,1).
txt_doc(290,56,2).
txt_doc(81,56,1).
txt_doc(742,56,2).
txt_doc(950,56,1).
txt_doc(434,56,1).
txt_doc(2,56,1).
txt_doc(307,56,1).
txt_doc(582,57,1).
txt_doc(840,57,1).
txt_doc(668,57,1).
txt_doc(100,57,1).
txt_doc(520,57,1).
txt_doc(719,57,1).
txt_doc(277,57,1).
txt_doc(647,57,1).
txt_doc(134,57,1).
txt_doc(534,57,1).
txt_doc(615,57,1).
txt_doc(511,57,1).
txt_doc(52,57,1).
txt_doc(215,57,1).
txt_doc(132,57,1).
txt_doc(671,57,1).
txt_doc(127,57,1).
txt_doc(60,57,1).
txt_doc(816,57,1).
txt_doc(193,57,1).
txt_doc(200,57,1).
txt_doc(102,57,1).
txt_doc(433,57,1).
txt_doc(903,57,1).
txt_doc(771,57,1).
txt_doc(217,57,1).
txt_doc(197,57,1).
txt_doc(517,57,1).
txt_doc(192,57,1).
txt_doc(907,57,1).
txt_doc(46,57,2).
txt_doc(785,57,1).
txt_doc(502,57,1).
txt_doc(347,57,1).
txt_doc(229,57,1).
txt_doc(565,57,1).
txt_doc(629,57,1).
txt_doc(297,57,1).
txt_doc(537,57,2).
txt_doc(288,57,1).
txt_doc(662,57,1).
txt_doc(15,57,1).
txt_doc(918,57,1).
txt_doc(651,57,1).
txt_doc(614,57,1).
txt_doc(81,57,1).
txt_doc(656,57,1).
txt_doc(968,57,1).
txt_doc(826,57,1).
txt_doc(734,57,2).
txt_doc(552,57,1).
txt_doc(145,57,1).
txt_doc(805,57,1).
txt_doc(283,57,1).
txt_doc(794,57,1).
txt_doc(779,57,1).
txt_doc(632,57,1).
txt_doc(695,57,1).
txt_doc(30,57,1).
txt_doc(705,57,1).
txt_doc(906,57,1).
txt_doc(235,57,1).
txt_doc(17,57,1).
txt_doc(940,57,1).
txt_doc(796,57,1).
txt_doc(485,57,1).
txt_doc(675,57,1).
txt_doc(105,57,1).
txt_doc(710,57,1).
txt_doc(76,57,1).
txt_doc(455,57,1).
txt_doc(366,57,1).
txt_doc(426,57,1).
txt_doc(663,57,1).
txt_doc(780,57,1).
txt_doc(103,57,1).
txt_doc(928,57,1).
txt_doc(166,57,1).
txt_doc(467,57,2).
txt_doc(879,57,1).
txt_doc(273,57,1).
txt_doc(868,57,1).
txt_doc(319,57,1).
txt_doc(633,57,1).
txt_doc(306,57,1).
txt_doc(566,57,1).
txt_doc(403,57,1).
txt_doc(561,57,2).
txt_doc(369,57,1).
txt_doc(337,57,1).
txt_doc(844,57,1).
txt_doc(790,57,2).
txt_doc(296,57,1).
txt_doc(333,57,1).
txt_doc(47,57,1).
txt_doc(112,57,1).
txt_doc(185,57,1).
txt_doc(385,57,1).
txt_doc(589,57,1).
txt_doc(838,57,1).
txt_doc(712,57,2).
txt_doc(75,57,1).
txt_doc(497,57,1).
txt_doc(123,57,1).
txt_doc(55,57,1).
txt_doc(108,57,1).
txt_doc(831,57,1).
txt_doc(446,57,1).
txt_doc(5,57,1).
txt_doc(415,57,1).
txt_doc(33,57,1).
txt_doc(107,57,1).
txt_doc(885,57,1).
txt_doc(331,57,1).
txt_doc(383,57,2).
txt_doc(101,57,1).
txt_doc(682,57,1).
txt_doc(375,57,1).
txt_doc(77,57,1).
txt_doc(966,57,1).
txt_doc(361,57,1).
txt_doc(23,57,1).
txt_doc(681,57,1).
txt_doc(444,57,1).
txt_doc(152,57,1).
txt_doc(880,57,1).
txt_doc(744,57,1).
txt_doc(324,57,2).
txt_doc(724,57,1).
txt_doc(652,57,1).
txt_doc(211,57,1).
txt_doc(749,57,2).
txt_doc(307,57,1).
txt_doc(784,57,1).
txt_doc(449,57,1).
txt_doc(612,57,1).
txt_doc(206,57,1).
txt_doc(125,57,1).
txt_doc(122,57,1).
txt_doc(199,57,1).
txt_doc(36,57,1).
txt_doc(954,57,1).
txt_doc(402,57,1).
txt_doc(322,57,1).
txt_doc(778,57,2).
txt_doc(463,57,1).
txt_doc(559,57,1).
txt_doc(315,57,2).
txt_doc(300,57,1).
txt_doc(348,57,1).
txt_doc(254,57,1).
txt_doc(435,57,1).
txt_doc(243,57,1).
txt_doc(422,57,1).
txt_doc(564,57,2).
txt_doc(834,57,1).
txt_doc(162,57,1).
txt_doc(309,57,1).
txt_doc(290,57,1).
txt_doc(949,57,2).
txt_doc(736,57,1).
txt_doc(26,57,1).
txt_doc(947,57,1).
txt_doc(63,57,1).
txt_doc(458,57,1).
txt_doc(886,57,1).
txt_doc(397,57,1).
txt_doc(732,57,1).
txt_doc(504,57,1).
txt_doc(810,57,1).
txt_doc(644,57,1).
txt_doc(233,57,2).
txt_doc(470,57,1).
txt_doc(265,57,1).
txt_doc(113,57,2).
txt_doc(339,57,1).
txt_doc(607,57,1).
txt_doc(213,57,2).
txt_doc(759,57,1).
txt_doc(961,57,1).
txt_doc(494,57,1).
txt_doc(887,57,1).
txt_doc(679,57,1).
txt_doc(29,58,1).
txt_doc(298,58,1).
txt_doc(344,58,1).
txt_doc(683,58,1).
txt_doc(189,58,1).
txt_doc(188,58,1).
txt_doc(910,58,1).
txt_doc(493,58,1).
txt_doc(830,58,1).
txt_doc(98,58,1).
txt_doc(583,58,1).
txt_doc(964,58,1).
txt_doc(561,58,1).
txt_doc(565,58,1).
txt_doc(407,58,1).
txt_doc(386,58,1).
txt_doc(938,58,1).
txt_doc(479,58,1).
txt_doc(143,58,1).
txt_doc(372,58,1).
txt_doc(365,58,1).
txt_doc(37,58,1).
txt_doc(663,58,1).
txt_doc(579,58,1).
txt_doc(547,58,1).
txt_doc(2,58,1).
txt_doc(18,58,1).
txt_doc(341,58,1).
txt_doc(723,58,1).
txt_doc(457,58,1).
txt_doc(304,58,1).
txt_doc(810,58,1).
txt_doc(546,58,1).
txt_doc(349,58,1).
txt_doc(150,58,1).
txt_doc(685,58,1).
txt_doc(72,58,1).
txt_doc(924,58,1).
txt_doc(102,58,1).
txt_doc(227,58,1).
txt_doc(540,58,2).
txt_doc(647,58,2).
txt_doc(202,58,1).
txt_doc(307,58,1).
txt_doc(428,58,1).
txt_doc(321,58,1).
txt_doc(633,58,1).
txt_doc(950,58,1).
txt_doc(551,58,1).
txt_doc(664,58,1).
txt_doc(112,58,1).
txt_doc(824,58,1).
txt_doc(711,58,1).
txt_doc(45,58,1).
txt_doc(591,58,1).
txt_doc(943,58,1).
txt_doc(951,58,1).
txt_doc(826,58,1).
txt_doc(251,58,1).
txt_doc(84,58,1).
txt_doc(802,58,1).
txt_doc(62,58,1).
txt_doc(745,58,1).
txt_doc(145,58,1).
txt_doc(903,58,1).
txt_doc(166,58,1).
txt_doc(222,58,1).
txt_doc(960,58,1).
txt_doc(500,58,1).
txt_doc(418,58,2).
txt_doc(146,58,1).
txt_doc(562,58,2).
txt_doc(765,58,1).
txt_doc(199,58,1).
txt_doc(697,58,1).
txt_doc(692,58,1).
txt_doc(460,58,1).
txt_doc(568,58,1).
txt_doc(959,58,1).
txt_doc(364,58,1).
txt_doc(67,58,1).
txt_doc(211,58,1).
txt_doc(563,58,1).
txt_doc(266,58,1).
txt_doc(206,58,1).
txt_doc(653,58,1).
txt_doc(324,58,1).
txt_doc(577,58,1).
txt_doc(606,58,2).
txt_doc(172,58,1).
txt_doc(621,58,1).
txt_doc(729,58,1).
txt_doc(133,58,1).
txt_doc(662,58,1).
txt_doc(246,58,1).
txt_doc(516,58,1).
txt_doc(429,58,1).
txt_doc(173,58,1).
txt_doc(854,58,1).
txt_doc(535,58,1).
txt_doc(629,58,1).
txt_doc(68,58,1).
txt_doc(394,58,1).
txt_doc(350,58,1).
txt_doc(38,58,2).
txt_doc(1,58,1).
txt_doc(73,58,1).
txt_doc(762,58,1).
txt_doc(878,58,2).
txt_doc(775,58,1).
txt_doc(912,58,1).
txt_doc(58,58,1).
txt_doc(615,58,1).
txt_doc(263,58,1).
txt_doc(326,58,1).
txt_doc(946,58,1).
txt_doc(967,58,1).
txt_doc(559,58,2).
txt_doc(862,58,1).
txt_doc(501,58,1).
txt_doc(355,58,1).
txt_doc(373,58,1).
txt_doc(5,58,3).
txt_doc(911,58,1).
txt_doc(539,58,1).
txt_doc(175,58,1).
txt_doc(475,58,1).
txt_doc(638,58,2).
txt_doc(713,58,1).
txt_doc(41,58,2).
txt_doc(453,58,2).
txt_doc(828,58,1).
txt_doc(4,58,1).
txt_doc(15,58,1).
txt_doc(410,58,1).
txt_doc(674,58,1).
txt_doc(277,58,1).
txt_doc(798,58,1).
txt_doc(544,58,2).
txt_doc(849,58,3).
txt_doc(468,58,1).
txt_doc(578,58,1).
txt_doc(712,58,1).
txt_doc(799,58,1).
txt_doc(746,58,3).
txt_doc(155,58,1).
txt_doc(16,58,1).
txt_doc(767,58,1).
txt_doc(351,58,1).
txt_doc(176,58,1).
txt_doc(728,58,1).
txt_doc(891,58,1).
txt_doc(719,58,1).
txt_doc(178,58,1).
txt_doc(303,58,1).
txt_doc(758,58,1).
txt_doc(895,58,1).
txt_doc(269,58,1).
txt_doc(706,58,1).
txt_doc(626,58,1).
txt_doc(573,58,1).
txt_doc(96,58,2).
txt_doc(530,58,1).
txt_doc(113,58,1).
txt_doc(782,58,1).
txt_doc(812,58,1).
txt_doc(804,58,1).
txt_doc(838,58,1).
txt_doc(872,58,1).
txt_doc(507,58,1).
txt_doc(57,58,1).
txt_doc(739,58,2).
txt_doc(23,58,1).
txt_doc(436,58,1).
txt_doc(27,58,1).
txt_doc(431,58,1).
txt_doc(151,58,1).
txt_doc(604,58,1).
txt_doc(576,58,1).
txt_doc(192,58,1).
txt_doc(216,59,1).
txt_doc(266,59,1).
txt_doc(928,59,1).
txt_doc(372,59,1).
txt_doc(963,59,1).
txt_doc(228,59,1).
txt_doc(339,59,1).
txt_doc(647,59,1).
txt_doc(302,59,1).
txt_doc(817,59,1).
txt_doc(492,59,1).
txt_doc(814,59,1).
txt_doc(709,59,1).
txt_doc(765,59,1).
txt_doc(107,59,1).
txt_doc(764,59,1).
txt_doc(151,59,1).
txt_doc(159,59,1).
txt_doc(498,59,1).
txt_doc(59,59,1).
txt_doc(628,59,1).
txt_doc(637,59,1).
txt_doc(322,59,1).
txt_doc(428,59,1).
txt_doc(526,59,1).
txt_doc(175,59,2).
txt_doc(854,59,1).
txt_doc(173,59,1).
txt_doc(569,59,1).
txt_doc(680,59,1).
txt_doc(377,59,1).
txt_doc(128,59,1).
txt_doc(347,59,1).
txt_doc(289,59,1).
txt_doc(453,59,1).
txt_doc(265,59,1).
txt_doc(897,59,1).
txt_doc(318,59,1).
txt_doc(474,59,1).
txt_doc(154,59,1).
txt_doc(880,59,1).
txt_doc(192,59,1).
txt_doc(371,59,1).
txt_doc(258,59,1).
txt_doc(427,59,1).
txt_doc(845,59,1).
txt_doc(506,59,1).
txt_doc(454,59,1).
txt_doc(839,59,1).
txt_doc(165,59,1).
txt_doc(554,59,1).
txt_doc(527,59,1).
txt_doc(522,59,1).
txt_doc(729,59,1).
txt_doc(426,59,1).
txt_doc(161,59,1).
txt_doc(170,59,1).
txt_doc(178,59,1).
txt_doc(585,59,2).
txt_doc(357,59,1).
txt_doc(962,59,1).
txt_doc(719,59,1).
txt_doc(251,59,1).
txt_doc(593,59,1).
txt_doc(514,59,1).
txt_doc(515,59,1).
txt_doc(162,59,1).
txt_doc(894,59,1).
txt_doc(910,59,1).
txt_doc(138,59,1).
txt_doc(226,59,2).
txt_doc(416,59,1).
txt_doc(881,59,1).
txt_doc(330,59,1).
txt_doc(707,59,1).
txt_doc(925,59,1).
txt_doc(4,59,1).
txt_doc(179,59,1).
txt_doc(381,59,1).
txt_doc(409,59,1).
txt_doc(195,59,1).
txt_doc(955,59,1).
txt_doc(937,59,1).
txt_doc(77,59,1).
txt_doc(458,59,1).
txt_doc(617,59,1).
txt_doc(5,59,1).
txt_doc(636,59,1).
txt_doc(255,59,1).
txt_doc(673,59,1).
txt_doc(328,59,1).
txt_doc(447,59,1).
txt_doc(37,59,1).
txt_doc(856,59,1).
txt_doc(249,59,1).
txt_doc(768,59,1).
txt_doc(462,59,1).
txt_doc(695,59,1).
txt_doc(773,59,1).
txt_doc(46,59,1).
txt_doc(766,59,1).
txt_doc(549,59,1).
txt_doc(493,59,1).
txt_doc(479,59,1).
txt_doc(201,59,1).
txt_doc(781,59,1).
txt_doc(387,59,1).
txt_doc(392,59,1).
txt_doc(90,59,1).
txt_doc(43,59,1).
txt_doc(198,59,1).
txt_doc(270,59,2).
txt_doc(136,59,1).
txt_doc(763,59,1).
txt_doc(327,59,1).
txt_doc(320,59,1).
txt_doc(365,59,1).
txt_doc(594,59,1).
txt_doc(250,59,1).
txt_doc(208,59,1).
txt_doc(656,59,1).
txt_doc(601,59,1).
txt_doc(758,59,1).
txt_doc(938,59,1).
txt_doc(241,59,1).
txt_doc(595,59,1).
txt_doc(63,59,3).
txt_doc(687,59,1).
txt_doc(613,59,1).
txt_doc(615,59,1).
txt_doc(273,59,1).
txt_doc(245,59,2).
txt_doc(19,59,1).
txt_doc(802,59,1).
txt_doc(930,59,1).
txt_doc(420,59,1).
txt_doc(683,59,1).
txt_doc(681,59,1).
txt_doc(949,59,1).
txt_doc(304,59,1).
txt_doc(696,59,1).
txt_doc(100,59,1).
txt_doc(18,59,1).
txt_doc(164,59,1).
txt_doc(73,59,1).
txt_doc(316,59,1).
txt_doc(139,59,1).
txt_doc(819,59,1).
txt_doc(182,59,1).
txt_doc(25,59,1).
txt_doc(645,59,1).
txt_doc(889,59,1).
txt_doc(666,59,1).
txt_doc(39,59,1).
txt_doc(475,59,1).
txt_doc(891,59,1).
txt_doc(400,59,1).
txt_doc(117,59,1).
txt_doc(391,59,1).
txt_doc(468,59,1).
txt_doc(325,59,1).
txt_doc(385,59,1).
txt_doc(776,59,1).
txt_doc(174,59,1).
txt_doc(714,59,1).
txt_doc(863,59,2).
txt_doc(644,59,1).
txt_doc(885,59,2).
txt_doc(456,59,1).
txt_doc(735,59,1).
txt_doc(70,59,1).
txt_doc(918,59,1).
txt_doc(654,59,2).
txt_doc(831,59,2).
txt_doc(857,59,2).
txt_doc(563,59,1).
txt_doc(523,59,1).
txt_doc(629,59,2).
txt_doc(790,59,1).
txt_doc(163,59,1).
txt_doc(571,59,1).
txt_doc(305,59,1).
txt_doc(105,59,1).
txt_doc(256,59,2).
txt_doc(755,59,1).
txt_doc(335,59,1).
txt_doc(52,60,1).
txt_doc(920,60,1).
txt_doc(206,60,1).
txt_doc(903,60,1).
txt_doc(625,60,1).
txt_doc(791,60,1).
txt_doc(640,60,1).
txt_doc(572,60,1).
txt_doc(601,60,1).
txt_doc(957,60,1).
txt_doc(117,60,1).
txt_doc(722,60,1).
txt_doc(497,60,1).
txt_doc(433,60,1).
txt_doc(161,60,1).
txt_doc(617,60,1).
txt_doc(717,60,1).
txt_doc(793,60,1).
txt_doc(60,60,1).
txt_doc(825,60,1).
txt_doc(796,60,1).
txt_doc(74,60,1).
txt_doc(562,60,1).
txt_doc(763,60,1).
txt_doc(578,60,1).
txt_doc(648,60,1).
txt_doc(830,60,1).
txt_doc(426,60,1).
txt_doc(323,60,1).
txt_doc(365,60,1).
txt_doc(250,60,1).
txt_doc(646,60,1).
txt_doc(576,60,1).
txt_doc(715,60,1).
txt_doc(802,60,1).
txt_doc(229,60,1).
txt_doc(878,60,1).
txt_doc(627,60,1).
txt_doc(192,60,1).
txt_doc(919,60,1).
txt_doc(706,60,1).
txt_doc(958,60,1).
txt_doc(326,60,1).
txt_doc(345,60,1).
txt_doc(170,60,2).
txt_doc(20,60,2).
txt_doc(767,60,1).
txt_doc(396,60,1).
txt_doc(857,60,1).
txt_doc(480,60,1).
txt_doc(268,60,2).
txt_doc(143,60,1).
txt_doc(539,60,1).
txt_doc(266,60,1).
txt_doc(423,60,1).
txt_doc(739,60,1).
txt_doc(868,60,1).
txt_doc(65,60,1).
txt_doc(577,60,2).
txt_doc(823,60,2).
txt_doc(615,60,1).
txt_doc(727,60,1).
txt_doc(865,60,1).
txt_doc(283,60,1).
txt_doc(179,60,1).
txt_doc(815,60,1).
txt_doc(930,60,1).
txt_doc(549,60,1).
txt_doc(31,60,1).
txt_doc(427,60,1).
txt_doc(698,60,1).
txt_doc(305,60,1).
txt_doc(209,60,1).
txt_doc(287,60,1).
txt_doc(769,60,1).
txt_doc(197,60,1).
txt_doc(343,60,1).
txt_doc(894,60,1).
txt_doc(724,60,1).
txt_doc(398,60,1).
txt_doc(180,60,1).
txt_doc(102,60,1).
txt_doc(954,60,1).
txt_doc(76,60,1).
txt_doc(879,60,1).
txt_doc(190,60,1).
txt_doc(898,60,1).
txt_doc(841,60,1).
txt_doc(281,60,1).
txt_doc(701,60,1).
txt_doc(710,60,1).
txt_doc(552,60,1).
txt_doc(766,60,1).
txt_doc(484,60,1).
txt_doc(520,60,2).
txt_doc(95,60,1).
txt_doc(906,60,2).
txt_doc(354,60,1).
txt_doc(346,60,1).
txt_doc(51,60,1).
txt_doc(94,60,1).
txt_doc(307,60,1).
txt_doc(746,60,1).
txt_doc(278,60,1).
txt_doc(545,60,1).
txt_doc(392,60,1).
txt_doc(443,60,1).
txt_doc(519,60,1).
txt_doc(410,60,1).
txt_doc(432,60,2).
txt_doc(406,60,1).
txt_doc(686,60,1).
txt_doc(235,60,1).
txt_doc(598,60,1).
txt_doc(932,60,1).
txt_doc(705,60,1).
txt_doc(269,60,1).
txt_doc(751,60,1).
txt_doc(47,60,1).
txt_doc(935,60,1).
txt_doc(431,60,1).
txt_doc(333,60,1).
txt_doc(205,60,1).
txt_doc(642,60,1).
txt_doc(348,60,1).
txt_doc(955,60,1).
txt_doc(267,60,1).
txt_doc(707,60,1).
txt_doc(639,60,1).
txt_doc(310,60,1).
txt_doc(214,60,2).
txt_doc(786,60,1).
txt_doc(201,60,1).
txt_doc(773,60,1).
txt_doc(224,60,1).
txt_doc(188,60,2).
txt_doc(474,60,1).
txt_doc(679,60,1).
txt_doc(325,60,2).
txt_doc(304,60,2).
txt_doc(666,60,1).
txt_doc(34,60,1).
txt_doc(458,60,1).
txt_doc(853,60,1).
txt_doc(456,60,1).
txt_doc(785,60,1).
txt_doc(711,60,1).
txt_doc(687,60,1).
txt_doc(557,60,3).
txt_doc(837,60,1).
txt_doc(236,60,1).
txt_doc(755,60,1).
txt_doc(673,60,1).
txt_doc(441,60,1).
txt_doc(685,60,2).
txt_doc(542,60,1).
txt_doc(517,60,1).
txt_doc(499,60,2).
txt_doc(248,60,1).
txt_doc(184,60,1).
txt_doc(468,60,2).
txt_doc(684,60,1).
txt_doc(40,60,1).
txt_doc(719,60,1).
txt_doc(285,60,2).
txt_doc(444,60,1).
txt_doc(836,60,2).
txt_doc(558,60,1).
txt_doc(744,60,2).
txt_doc(18,60,1).
txt_doc(703,60,1).
txt_doc(488,60,1).
txt_doc(618,60,1).
txt_doc(163,60,1).
txt_doc(436,60,2).
txt_doc(332,60,2).
txt_doc(408,60,1).
txt_doc(349,60,1).
txt_doc(967,61,1).
txt_doc(4,61,1).
txt_doc(25,61,1).
txt_doc(190,61,1).
txt_doc(514,61,1).
txt_doc(532,61,2).
txt_doc(382,61,1).
txt_doc(790,61,1).
txt_doc(109,61,1).
txt_doc(352,61,1).
txt_doc(400,61,1).
txt_doc(951,61,1).
txt_doc(785,61,1).
txt_doc(228,61,1).
txt_doc(866,61,1).
txt_doc(172,61,1).
txt_doc(99,61,1).
txt_doc(254,61,1).
txt_doc(406,61,1).
txt_doc(137,61,1).
txt_doc(128,61,1).
txt_doc(105,61,1).
txt_doc(253,61,1).
txt_doc(173,61,1).
txt_doc(100,61,1).
txt_doc(168,61,2).
txt_doc(519,61,1).
txt_doc(438,61,2).
txt_doc(865,61,1).
txt_doc(76,61,1).
txt_doc(538,61,1).
txt_doc(528,61,1).
txt_doc(655,61,1).
txt_doc(629,61,1).
txt_doc(925,61,1).
txt_doc(271,61,1).
txt_doc(146,61,1).
txt_doc(662,61,1).
txt_doc(235,61,1).
txt_doc(448,61,1).
txt_doc(74,61,1).
txt_doc(614,61,1).
txt_doc(780,61,1).
txt_doc(640,61,1).
txt_doc(378,61,1).
txt_doc(545,61,1).
txt_doc(539,61,1).
txt_doc(95,61,1).
txt_doc(656,61,1).
txt_doc(393,61,1).
txt_doc(246,61,1).
txt_doc(887,61,1).
txt_doc(512,61,1).
txt_doc(699,61,1).
txt_doc(871,61,1).
txt_doc(302,61,2).
txt_doc(680,61,1).
txt_doc(525,61,1).
txt_doc(54,61,1).
txt_doc(879,61,1).
txt_doc(331,61,2).
txt_doc(359,61,2).
txt_doc(938,61,1).
txt_doc(387,61,1).
txt_doc(19,61,1).
txt_doc(854,61,1).
txt_doc(666,61,1).
txt_doc(596,61,1).
txt_doc(509,61,1).
txt_doc(644,61,1).
txt_doc(239,61,1).
txt_doc(231,61,1).
txt_doc(622,61,1).
txt_doc(358,61,1).
txt_doc(282,61,1).
txt_doc(123,61,1).
txt_doc(716,61,1).
txt_doc(90,61,1).
txt_doc(501,61,2).
txt_doc(793,61,1).
txt_doc(284,61,1).
txt_doc(898,61,1).
txt_doc(920,61,1).
txt_doc(205,61,1).
txt_doc(272,61,1).
txt_doc(136,61,1).
txt_doc(334,61,1).
txt_doc(497,61,1).
txt_doc(349,61,1).
txt_doc(117,61,1).
txt_doc(299,61,2).
txt_doc(328,61,1).
txt_doc(965,61,1).
txt_doc(937,61,1).
txt_doc(459,61,1).
txt_doc(694,61,2).
txt_doc(851,61,1).
txt_doc(1,61,1).
txt_doc(340,61,1).
txt_doc(867,61,1).
txt_doc(728,61,1).
txt_doc(827,61,1).
txt_doc(485,61,1).
txt_doc(784,61,1).
txt_doc(429,61,1).
txt_doc(768,61,2).
txt_doc(581,61,1).
txt_doc(428,61,1).
txt_doc(480,61,2).
txt_doc(837,61,1).
txt_doc(184,61,1).
txt_doc(81,61,1).
txt_doc(859,61,1).
txt_doc(9,61,1).
txt_doc(611,61,1).
txt_doc(775,61,1).
txt_doc(812,61,1).
txt_doc(737,61,1).
txt_doc(670,61,1).
txt_doc(600,61,1).
txt_doc(577,61,1).
txt_doc(912,61,1).
txt_doc(203,61,1).
txt_doc(430,61,1).
txt_doc(451,61,1).
txt_doc(783,61,1).
txt_doc(490,61,1).
txt_doc(583,61,1).
txt_doc(371,61,1).
txt_doc(325,61,1).
txt_doc(561,61,1).
txt_doc(759,61,1).
txt_doc(645,61,2).
txt_doc(409,61,1).
txt_doc(888,61,1).
txt_doc(442,61,2).
txt_doc(919,61,1).
txt_doc(767,61,1).
txt_doc(902,61,1).
txt_doc(441,61,1).
txt_doc(251,61,1).
txt_doc(363,61,1).
txt_doc(70,61,1).
txt_doc(493,61,1).
txt_doc(404,61,1).
txt_doc(543,61,1).
txt_doc(474,61,1).
txt_doc(667,61,1).
txt_doc(649,61,1).
txt_doc(610,61,1).
txt_doc(852,61,2).
txt_doc(578,61,1).
txt_doc(724,61,1).
txt_doc(59,61,1).
txt_doc(499,61,1).
txt_doc(313,61,1).
txt_doc(248,61,2).
txt_doc(219,61,3).
txt_doc(750,61,1).
txt_doc(949,61,1).
txt_doc(506,61,1).
txt_doc(31,61,1).
txt_doc(167,61,1).
txt_doc(846,61,1).
txt_doc(252,61,1).
txt_doc(460,61,1).
txt_doc(346,61,2).
txt_doc(296,61,1).
txt_doc(390,61,2).
txt_doc(177,61,1).
txt_doc(243,61,1).
txt_doc(717,61,2).
txt_doc(347,61,1).
txt_doc(730,61,1).
txt_doc(675,61,1).
txt_doc(478,61,1).
txt_doc(797,61,1).
txt_doc(616,61,1).
txt_doc(468,61,1).
txt_doc(197,61,1).
txt_doc(519,62,1).
txt_doc(943,62,1).
txt_doc(864,62,1).
txt_doc(396,62,1).
txt_doc(571,62,1).
txt_doc(666,62,1).
txt_doc(473,62,1).
txt_doc(669,62,1).
txt_doc(525,62,1).
txt_doc(114,62,1).
txt_doc(152,62,1).
txt_doc(150,62,1).
txt_doc(174,62,1).
txt_doc(965,62,1).
txt_doc(583,62,1).
txt_doc(178,62,1).
txt_doc(343,62,1).
txt_doc(604,62,1).
txt_doc(646,62,1).
txt_doc(347,62,1).
txt_doc(868,62,1).
txt_doc(915,62,1).
txt_doc(269,62,1).
txt_doc(673,62,1).
txt_doc(700,62,1).
txt_doc(386,62,1).
txt_doc(22,62,1).
txt_doc(774,62,1).
txt_doc(564,62,1).
txt_doc(752,62,1).
txt_doc(278,62,1).
txt_doc(53,62,1).
txt_doc(804,62,2).
txt_doc(43,62,1).
txt_doc(816,62,1).
txt_doc(218,62,1).
txt_doc(740,62,1).
txt_doc(917,62,1).
txt_doc(443,62,1).
txt_doc(100,62,1).
txt_doc(304,62,1).
txt_doc(139,62,1).
txt_doc(478,62,1).
txt_doc(75,62,1).
txt_doc(827,62,1).
txt_doc(853,62,1).
txt_doc(857,62,1).
txt_doc(664,62,1).
txt_doc(335,62,1).
txt_doc(179,62,1).
txt_doc(27,62,1).
txt_doc(101,62,1).
txt_doc(554,62,1).
txt_doc(551,62,1).
txt_doc(648,62,1).
txt_doc(670,62,1).
txt_doc(25,62,1).
txt_doc(718,62,1).
txt_doc(589,62,1).
txt_doc(846,62,1).
txt_doc(298,62,1).
txt_doc(354,62,1).
txt_doc(686,62,1).
txt_doc(668,62,1).
txt_doc(469,62,1).
txt_doc(845,62,1).
txt_doc(161,62,1).
txt_doc(264,62,1).
txt_doc(184,62,1).
txt_doc(96,62,1).
txt_doc(316,62,1).
txt_doc(902,62,2).
txt_doc(575,62,1).
txt_doc(211,62,1).
txt_doc(40,62,1).
txt_doc(682,62,1).
txt_doc(775,62,1).
txt_doc(538,62,1).
txt_doc(20,62,1).
txt_doc(893,62,1).
txt_doc(701,62,1).
txt_doc(509,62,1).
txt_doc(368,62,1).
txt_doc(168,62,2).
txt_doc(463,62,1).
txt_doc(451,62,1).
txt_doc(960,62,1).
txt_doc(601,62,1).
txt_doc(607,62,1).
txt_doc(502,62,1).
txt_doc(107,62,2).
txt_doc(940,62,1).
txt_doc(464,62,1).
txt_doc(484,62,1).
txt_doc(714,62,1).
txt_doc(889,62,1).
txt_doc(959,62,1).
txt_doc(64,62,1).
txt_doc(89,62,1).
txt_doc(58,62,2).
txt_doc(894,62,1).
txt_doc(431,62,1).
txt_doc(143,62,2).
txt_doc(350,62,1).
txt_doc(784,62,1).
txt_doc(117,62,1).
txt_doc(407,62,1).
txt_doc(320,62,1).
txt_doc(738,62,1).
txt_doc(159,62,1).
txt_doc(493,62,1).
txt_doc(672,62,2).
txt_doc(863,62,1).
txt_doc(569,62,1).
txt_doc(585,62,1).
txt_doc(529,62,1).
txt_doc(340,62,1).
txt_doc(321,62,1).
txt_doc(123,62,1).
txt_doc(852,62,2).
txt_doc(353,62,1).
txt_doc(638,62,1).
txt_doc(118,62,1).
txt_doc(506,62,1).
txt_doc(155,62,1).
txt_doc(208,62,1).
txt_doc(181,62,1).
txt_doc(426,62,1).
txt_doc(704,62,1).
txt_doc(667,62,1).
txt_doc(840,62,2).
txt_doc(271,62,2).
txt_doc(539,62,2).
txt_doc(757,62,1).
txt_doc(78,62,2).
txt_doc(797,62,1).
txt_doc(325,62,1).
txt_doc(405,62,1).
txt_doc(615,62,1).
txt_doc(567,62,1).
txt_doc(428,62,1).
txt_doc(233,62,1).
txt_doc(629,62,1).
txt_doc(388,62,1).
txt_doc(512,62,1).
txt_doc(224,62,1).
txt_doc(359,62,2).
txt_doc(7,62,1).
txt_doc(626,62,1).
txt_doc(487,62,1).
txt_doc(688,62,1).
txt_doc(292,62,1).
txt_doc(391,62,1).
txt_doc(173,62,1).
txt_doc(599,62,1).
txt_doc(751,62,1).
txt_doc(13,62,1).
txt_doc(228,62,1).
txt_doc(528,62,1).
txt_doc(272,62,1).
txt_doc(45,62,2).
txt_doc(743,62,1).
txt_doc(241,62,1).
txt_doc(658,62,1).
txt_doc(17,62,1).
txt_doc(745,62,1).
txt_doc(961,62,1).
txt_doc(662,62,2).
txt_doc(223,62,1).
txt_doc(362,62,1).
txt_doc(769,62,1).
txt_doc(333,62,1).
txt_doc(32,62,1).
txt_doc(698,62,1).
txt_doc(923,62,1).
txt_doc(63,62,1).
txt_doc(116,62,2).
txt_doc(488,62,2).
txt_doc(42,62,1).
txt_doc(283,62,2).
txt_doc(461,62,2).
txt_doc(303,63,1).
txt_doc(598,63,1).
txt_doc(873,63,1).
txt_doc(762,63,1).
txt_doc(768,63,1).
txt_doc(3,63,1).
txt_doc(4,63,1).
txt_doc(584,63,1).
txt_doc(15,63,1).
txt_doc(280,63,1).
txt_doc(89,63,1).
txt_doc(625,63,1).
txt_doc(886,63,1).
txt_doc(556,63,1).
txt_doc(867,63,1).
txt_doc(742,63,1).
txt_doc(790,63,1).
txt_doc(149,63,1).
txt_doc(725,63,1).
txt_doc(577,63,1).
txt_doc(828,63,1).
txt_doc(589,63,1).
txt_doc(697,63,1).
txt_doc(298,63,1).
txt_doc(667,63,1).
txt_doc(384,63,1).
txt_doc(631,63,1).
txt_doc(914,63,1).
txt_doc(167,63,1).
txt_doc(691,63,1).
txt_doc(789,63,1).
txt_doc(396,63,1).
txt_doc(947,63,1).
txt_doc(608,63,1).
txt_doc(693,63,1).
txt_doc(825,63,1).
txt_doc(478,63,1).
txt_doc(950,63,2).
txt_doc(516,63,2).
txt_doc(432,63,1).
txt_doc(708,63,1).
txt_doc(330,63,1).
txt_doc(629,63,1).
txt_doc(549,63,1).
txt_doc(701,63,1).
txt_doc(26,63,1).
txt_doc(467,63,1).
txt_doc(489,63,1).
txt_doc(591,63,1).
txt_doc(733,63,1).
txt_doc(123,63,1).
txt_doc(434,63,1).
txt_doc(495,63,1).
txt_doc(479,63,1).
txt_doc(588,63,1).
txt_doc(547,63,1).
txt_doc(951,63,1).
txt_doc(752,63,1).
txt_doc(772,63,1).
txt_doc(490,63,1).
txt_doc(275,63,1).
txt_doc(822,63,1).
txt_doc(442,63,1).
txt_doc(207,63,1).
txt_doc(18,63,1).
txt_doc(362,63,1).
txt_doc(79,63,1).
txt_doc(660,63,1).
txt_doc(336,63,1).
txt_doc(447,63,1).
txt_doc(456,63,1).
txt_doc(366,63,1).
txt_doc(664,63,1).
txt_doc(705,63,1).
txt_doc(538,63,1).
txt_doc(917,63,1).
txt_doc(256,63,1).
txt_doc(400,63,1).
txt_doc(946,63,1).
txt_doc(493,63,1).
txt_doc(880,63,1).
txt_doc(754,63,1).
txt_doc(375,63,1).
txt_doc(191,63,2).
txt_doc(820,63,1).
txt_doc(296,63,1).
txt_doc(203,63,1).
txt_doc(290,63,1).
txt_doc(743,63,1).
txt_doc(700,63,1).
txt_doc(746,63,1).
txt_doc(364,63,1).
txt_doc(208,63,2).
txt_doc(801,63,1).
txt_doc(953,63,1).
txt_doc(109,63,1).
txt_doc(945,63,2).
txt_doc(252,63,1).
txt_doc(579,63,1).
txt_doc(740,63,1).
txt_doc(955,63,1).
txt_doc(741,63,1).
txt_doc(502,63,1).
txt_doc(575,63,1).
txt_doc(537,63,1).
txt_doc(650,63,1).
txt_doc(839,63,1).
txt_doc(624,63,1).
txt_doc(509,63,1).
txt_doc(891,63,1).
txt_doc(389,63,1).
txt_doc(832,63,1).
txt_doc(905,63,1).
txt_doc(498,63,2).
txt_doc(627,63,1).
txt_doc(565,63,1).
txt_doc(359,63,1).
txt_doc(692,63,1).
txt_doc(714,63,1).
txt_doc(218,63,1).
txt_doc(440,63,3).
txt_doc(851,63,1).
txt_doc(174,63,1).
txt_doc(809,63,1).
txt_doc(35,63,1).
txt_doc(243,63,1).
txt_doc(391,63,1).
txt_doc(739,63,1).
txt_doc(232,63,1).
txt_doc(329,63,1).
txt_doc(141,63,1).
txt_doc(647,63,1).
txt_doc(77,63,1).
txt_doc(626,63,1).
txt_doc(678,63,2).
txt_doc(380,63,1).
txt_doc(562,63,1).
txt_doc(799,63,1).
txt_doc(793,63,1).
txt_doc(151,63,1).
txt_doc(818,63,2).
txt_doc(470,63,1).
txt_doc(304,63,1).
txt_doc(944,63,1).
txt_doc(91,63,1).
txt_doc(281,63,1).
txt_doc(31,63,1).
txt_doc(378,63,1).
txt_doc(371,63,1).
txt_doc(563,63,1).
txt_doc(103,63,1).
txt_doc(340,63,1).
txt_doc(276,63,1).
txt_doc(887,63,1).
txt_doc(557,63,1).
txt_doc(431,63,1).
txt_doc(405,63,3).
txt_doc(645,63,2).
txt_doc(513,63,2).
txt_doc(748,63,1).
txt_doc(707,63,1).
txt_doc(677,63,2).
txt_doc(189,63,1).
txt_doc(606,63,1).
txt_doc(36,63,1).
txt_doc(202,63,1).
txt_doc(305,63,2).
txt_doc(374,63,2).
txt_doc(327,63,1).
txt_doc(466,63,1).
txt_doc(655,63,1).
txt_doc(85,63,1).
txt_doc(44,63,1).
txt_doc(930,63,1).
txt_doc(640,63,2).
txt_doc(326,63,1).
txt_doc(32,63,1).
txt_doc(170,63,1).
txt_doc(781,63,1).
txt_doc(592,63,1).
txt_doc(68,63,1).
txt_doc(875,63,1).
txt_doc(193,64,1).
txt_doc(530,64,1).
txt_doc(183,64,1).
txt_doc(44,64,1).
txt_doc(122,64,1).
txt_doc(25,64,1).
txt_doc(23,64,1).
txt_doc(259,64,1).
txt_doc(532,64,1).
txt_doc(661,64,1).
txt_doc(59,64,1).
txt_doc(576,64,1).
txt_doc(939,64,1).
txt_doc(821,64,1).
txt_doc(856,64,1).
txt_doc(9,64,1).
txt_doc(508,64,1).
txt_doc(136,64,1).
txt_doc(961,64,1).
txt_doc(315,64,1).
txt_doc(823,64,2).
txt_doc(737,64,1).
txt_doc(615,64,1).
txt_doc(604,64,1).
txt_doc(968,64,1).
txt_doc(326,64,1).
txt_doc(797,64,1).
txt_doc(196,64,1).
txt_doc(478,64,1).
txt_doc(746,64,2).
txt_doc(943,64,1).
txt_doc(401,64,1).
txt_doc(323,64,1).
txt_doc(497,64,1).
txt_doc(372,64,2).
txt_doc(127,64,1).
txt_doc(56,64,1).
txt_doc(692,64,1).
txt_doc(816,64,1).
txt_doc(721,64,1).
txt_doc(901,64,1).
txt_doc(250,64,1).
txt_doc(38,64,1).
txt_doc(667,64,1).
txt_doc(161,64,1).
txt_doc(703,64,1).
txt_doc(593,64,2).
txt_doc(701,64,1).
txt_doc(793,64,1).
txt_doc(117,64,1).
txt_doc(356,64,1).
txt_doc(777,64,1).
txt_doc(148,64,1).
txt_doc(585,64,1).
txt_doc(160,64,1).
txt_doc(311,64,1).
txt_doc(509,64,1).
txt_doc(255,64,1).
txt_doc(433,64,1).
txt_doc(173,64,1).
txt_doc(98,64,1).
txt_doc(798,64,1).
txt_doc(28,64,1).
txt_doc(298,64,2).
txt_doc(903,64,1).
txt_doc(500,64,1).
txt_doc(61,64,1).
txt_doc(620,64,1).
txt_doc(641,64,1).
txt_doc(628,64,1).
txt_doc(562,64,1).
txt_doc(55,64,1).
txt_doc(392,64,1).
txt_doc(554,64,1).
txt_doc(290,64,1).
txt_doc(416,64,1).
txt_doc(462,64,1).
txt_doc(689,64,1).
txt_doc(329,64,1).
txt_doc(340,64,1).
txt_doc(732,64,1).
txt_doc(959,64,1).
txt_doc(480,64,2).
txt_doc(857,64,1).
txt_doc(405,64,1).
txt_doc(381,64,1).
txt_doc(453,64,1).
txt_doc(756,64,1).
txt_doc(632,64,1).
txt_doc(172,64,1).
txt_doc(645,64,2).
txt_doc(459,64,1).
txt_doc(867,64,1).
txt_doc(634,64,1).
txt_doc(956,64,1).
txt_doc(400,64,1).
txt_doc(633,64,1).
txt_doc(714,64,1).
txt_doc(74,64,1).
txt_doc(21,64,1).
txt_doc(369,64,1).
txt_doc(655,64,1).
txt_doc(870,64,1).
txt_doc(132,64,1).
txt_doc(126,64,1).
txt_doc(221,64,1).
txt_doc(559,64,1).
txt_doc(275,64,1).
txt_doc(674,64,1).
txt_doc(611,64,1).
txt_doc(490,64,1).
txt_doc(944,64,1).
txt_doc(428,64,1).
txt_doc(304,64,1).
txt_doc(638,64,1).
txt_doc(207,64,1).
txt_doc(567,64,1).
txt_doc(300,64,1).
txt_doc(936,64,2).
txt_doc(488,64,1).
txt_doc(219,64,1).
txt_doc(274,64,1).
txt_doc(143,64,1).
txt_doc(432,64,1).
txt_doc(121,64,1).
txt_doc(144,64,1).
txt_doc(608,64,1).
txt_doc(60,64,1).
txt_doc(945,64,1).
txt_doc(237,64,1).
txt_doc(680,64,1).
txt_doc(105,64,1).
txt_doc(330,64,1).
txt_doc(826,64,1).
txt_doc(294,64,1).
txt_doc(30,64,1).
txt_doc(673,64,2).
txt_doc(892,64,1).
txt_doc(630,64,1).
txt_doc(787,64,1).
txt_doc(723,64,1).
txt_doc(898,64,1).
txt_doc(558,64,1).
txt_doc(618,64,1).
txt_doc(97,64,1).
txt_doc(85,64,1).
txt_doc(162,64,3).
txt_doc(888,64,1).
txt_doc(288,64,1).
txt_doc(3,64,1).
txt_doc(357,64,1).
txt_doc(553,64,1).
txt_doc(779,64,1).
txt_doc(444,64,1).
txt_doc(873,64,2).
txt_doc(902,64,1).
txt_doc(15,64,1).
txt_doc(423,64,1).
txt_doc(751,64,1).
txt_doc(77,64,1).
txt_doc(91,64,1).
txt_doc(213,64,1).
txt_doc(327,64,1).
txt_doc(52,64,1).
txt_doc(442,64,1).
txt_doc(806,64,1).
txt_doc(13,64,1).
txt_doc(398,64,1).
txt_doc(584,64,1).
txt_doc(427,64,2).
txt_doc(170,64,1).
txt_doc(382,64,4).
txt_doc(215,64,2).
txt_doc(262,64,2).
txt_doc(491,64,1).
txt_doc(523,64,1).
txt_doc(836,64,1).
txt_doc(415,64,1).
txt_doc(862,64,1).
txt_doc(305,64,1).
txt_doc(163,64,1).
txt_doc(597,64,1).
txt_doc(226,65,1).
txt_doc(272,65,1).
txt_doc(354,65,1).
txt_doc(775,65,1).
txt_doc(151,65,1).
txt_doc(717,65,1).
txt_doc(681,65,1).
txt_doc(947,65,1).
txt_doc(699,65,1).
txt_doc(936,65,1).
txt_doc(549,65,1).
txt_doc(832,65,1).
txt_doc(319,65,1).
txt_doc(468,65,1).
txt_doc(788,65,1).
txt_doc(173,65,1).
txt_doc(635,65,1).
txt_doc(205,65,1).
txt_doc(489,65,2).
txt_doc(655,65,1).
txt_doc(86,65,1).
txt_doc(288,65,1).
txt_doc(391,65,1).
txt_doc(459,65,1).
txt_doc(844,65,1).
txt_doc(250,65,2).
txt_doc(195,65,1).
txt_doc(128,65,1).
txt_doc(660,65,1).
txt_doc(816,65,1).
txt_doc(864,65,1).
txt_doc(196,65,1).
txt_doc(966,65,1).
txt_doc(598,65,1).
txt_doc(296,65,2).
txt_doc(234,65,1).
txt_doc(318,65,1).
txt_doc(352,65,1).
txt_doc(14,65,1).
txt_doc(776,65,1).
txt_doc(138,65,1).
txt_doc(944,65,1).
txt_doc(846,65,1).
txt_doc(951,65,2).
txt_doc(227,65,1).
txt_doc(639,65,1).
txt_doc(847,65,1).
txt_doc(597,65,1).
txt_doc(532,65,1).
txt_doc(646,65,1).
txt_doc(102,65,1).
txt_doc(222,65,1).
txt_doc(867,65,1).
txt_doc(527,65,1).
txt_doc(218,65,1).
txt_doc(586,65,1).
txt_doc(78,65,1).
txt_doc(666,65,1).
txt_doc(811,65,1).
txt_doc(431,65,1).
txt_doc(803,65,1).
txt_doc(562,65,1).
txt_doc(286,65,1).
txt_doc(13,65,1).
txt_doc(967,65,1).
txt_doc(869,65,1).
txt_doc(247,65,1).
txt_doc(493,65,1).
txt_doc(498,65,1).
txt_doc(805,65,3).
txt_doc(385,65,1).
txt_doc(366,65,1).
txt_doc(740,65,1).
txt_doc(21,65,1).
txt_doc(858,65,1).
txt_doc(187,65,1).
txt_doc(898,65,1).
txt_doc(117,65,1).
txt_doc(265,65,1).
txt_doc(206,65,1).
txt_doc(760,65,1).
txt_doc(264,65,1).
txt_doc(878,65,1).
txt_doc(741,65,1).
txt_doc(853,65,1).
txt_doc(61,65,1).
txt_doc(323,65,1).
txt_doc(33,65,1).
txt_doc(211,65,1).
txt_doc(497,65,1).
txt_doc(70,65,1).
txt_doc(548,65,1).
txt_doc(417,65,1).
txt_doc(190,65,1).
txt_doc(667,65,1).
txt_doc(426,65,1).
txt_doc(960,65,1).
txt_doc(738,65,1).
txt_doc(201,65,1).
txt_doc(739,65,1).
txt_doc(749,65,1).
txt_doc(114,65,2).
txt_doc(106,65,1).
txt_doc(875,65,1).
txt_doc(587,65,1).
txt_doc(386,65,1).
txt_doc(96,65,1).
txt_doc(115,65,1).
txt_doc(100,65,1).
txt_doc(163,65,1).
txt_doc(623,65,1).
txt_doc(897,65,1).
txt_doc(156,65,1).
txt_doc(647,65,1).
txt_doc(76,65,1).
txt_doc(141,65,1).
txt_doc(781,65,1).
txt_doc(525,65,1).
txt_doc(331,65,1).
txt_doc(684,65,1).
txt_doc(478,65,1).
txt_doc(469,65,1).
txt_doc(71,65,1).
txt_doc(202,65,2).
txt_doc(486,65,1).
txt_doc(38,65,1).
txt_doc(940,65,1).
txt_doc(777,65,1).
txt_doc(259,65,1).
txt_doc(842,65,1).
txt_doc(523,65,1).
txt_doc(267,65,1).
txt_doc(485,65,1).
txt_doc(663,65,1).
txt_doc(397,65,2).
txt_doc(335,65,1).
txt_doc(906,65,1).
txt_doc(737,65,1).
txt_doc(580,65,1).
txt_doc(56,65,1).
txt_doc(901,65,1).
txt_doc(42,65,1).
txt_doc(589,65,1).
txt_doc(152,65,1).
txt_doc(827,65,1).
txt_doc(40,65,1).
txt_doc(479,65,1).
txt_doc(673,65,1).
txt_doc(524,65,1).
txt_doc(125,65,1).
txt_doc(600,65,2).
txt_doc(733,65,1).
txt_doc(433,65,1).
txt_doc(482,65,1).
txt_doc(723,65,1).
txt_doc(964,65,1).
txt_doc(606,65,1).
txt_doc(595,65,1).
txt_doc(648,65,1).
txt_doc(671,65,1).
txt_doc(108,65,1).
txt_doc(863,65,1).
txt_doc(866,65,1).
txt_doc(922,65,1).
txt_doc(395,65,1).
txt_doc(105,65,1).
txt_doc(59,65,1).
txt_doc(638,65,1).
txt_doc(157,65,1).
txt_doc(210,65,1).
txt_doc(280,65,1).
txt_doc(732,65,1).
txt_doc(181,65,1).
txt_doc(461,65,1).
txt_doc(446,65,1).
txt_doc(10,65,1).
txt_doc(348,65,1).
txt_doc(714,65,1).
txt_doc(221,65,1).
txt_doc(274,65,1).
txt_doc(389,65,2).
txt_doc(79,65,1).
txt_doc(552,65,1).
txt_doc(109,65,1).
txt_doc(560,65,2).
txt_doc(248,65,1).
txt_doc(192,65,1).
txt_doc(659,65,1).
txt_doc(664,66,1).
txt_doc(861,66,1).
txt_doc(61,66,1).
txt_doc(517,66,1).
txt_doc(372,66,1).
txt_doc(901,66,1).
txt_doc(646,66,1).
txt_doc(670,66,1).
txt_doc(757,66,1).
txt_doc(474,66,1).
txt_doc(695,66,1).
txt_doc(149,66,1).
txt_doc(189,66,1).
txt_doc(33,66,1).
txt_doc(171,66,1).
txt_doc(405,66,1).
txt_doc(880,66,1).
txt_doc(579,66,1).
txt_doc(649,66,1).
txt_doc(201,66,1).
txt_doc(555,66,1).
txt_doc(105,66,1).
txt_doc(84,66,1).
txt_doc(938,66,1).
txt_doc(932,66,1).
txt_doc(885,66,1).
txt_doc(277,66,1).
txt_doc(839,66,1).
txt_doc(581,66,1).
txt_doc(351,66,1).
txt_doc(716,66,1).
txt_doc(369,66,1).
txt_doc(475,66,1).
txt_doc(597,66,1).
txt_doc(827,66,1).
txt_doc(190,66,1).
txt_doc(700,66,1).
txt_doc(45,66,1).
txt_doc(843,66,1).
txt_doc(28,66,1).
txt_doc(403,66,1).
txt_doc(769,66,1).
txt_doc(572,66,1).
txt_doc(808,66,1).
txt_doc(943,66,1).
txt_doc(15,66,2).
txt_doc(161,66,1).
txt_doc(394,66,1).
txt_doc(203,66,1).
txt_doc(321,66,1).
txt_doc(153,66,1).
txt_doc(111,66,1).
txt_doc(275,66,1).
txt_doc(721,66,1).
txt_doc(511,66,1).
txt_doc(777,66,1).
txt_doc(90,66,1).
txt_doc(506,66,1).
txt_doc(241,66,2).
txt_doc(375,66,1).
txt_doc(658,66,1).
txt_doc(122,66,1).
txt_doc(946,66,1).
txt_doc(4,66,1).
txt_doc(820,66,1).
txt_doc(489,66,1).
txt_doc(802,66,1).
txt_doc(846,66,1).
txt_doc(638,66,1).
txt_doc(782,66,1).
txt_doc(301,66,1).
txt_doc(849,66,1).
txt_doc(100,66,1).
txt_doc(960,66,1).
txt_doc(432,66,1).
txt_doc(593,66,1).
txt_doc(648,66,1).
txt_doc(50,66,1).
txt_doc(709,66,1).
txt_doc(345,66,1).
txt_doc(283,66,1).
txt_doc(964,66,1).
txt_doc(601,66,1).
txt_doc(266,66,1).
txt_doc(69,66,1).
txt_doc(608,66,1).
txt_doc(892,66,1).
txt_doc(813,66,1).
txt_doc(710,66,1).
txt_doc(893,66,1).
txt_doc(253,66,1).
txt_doc(470,66,1).
txt_doc(568,66,1).
txt_doc(484,66,1).
txt_doc(544,66,2).
txt_doc(590,66,1).
txt_doc(248,66,1).
txt_doc(469,66,1).
txt_doc(830,66,1).
txt_doc(968,66,1).
txt_doc(94,66,1).
txt_doc(633,66,1).
txt_doc(912,66,1).
txt_doc(108,66,1).
txt_doc(295,66,1).
txt_doc(258,66,1).
txt_doc(528,66,1).
txt_doc(477,66,1).
txt_doc(671,66,1).
txt_doc(395,66,1).
txt_doc(962,66,2).
txt_doc(473,66,1).
txt_doc(503,66,2).
txt_doc(380,66,1).
txt_doc(586,66,1).
txt_doc(368,66,1).
txt_doc(324,66,1).
txt_doc(763,66,1).
txt_doc(407,66,1).
txt_doc(483,66,1).
txt_doc(594,66,1).
txt_doc(460,66,1).
txt_doc(438,66,1).
txt_doc(722,66,1).
txt_doc(738,66,1).
txt_doc(560,66,1).
txt_doc(386,66,1).
txt_doc(101,66,1).
txt_doc(736,66,1).
txt_doc(547,66,1).
txt_doc(107,66,1).
txt_doc(34,66,2).
txt_doc(494,66,1).
txt_doc(857,66,2).
txt_doc(86,66,1).
txt_doc(911,66,1).
txt_doc(39,66,1).
txt_doc(354,66,1).
txt_doc(724,66,2).
txt_doc(117,66,1).
txt_doc(408,66,1).
txt_doc(584,66,1).
txt_doc(656,66,1).
txt_doc(628,66,1).
txt_doc(132,66,1).
txt_doc(104,66,1).
txt_doc(543,66,1).
txt_doc(879,66,1).
txt_doc(618,66,1).
txt_doc(910,66,3).
txt_doc(571,66,1).
txt_doc(577,66,1).
txt_doc(245,66,1).
txt_doc(481,66,1).
txt_doc(817,66,1).
txt_doc(553,66,2).
txt_doc(193,66,1).
txt_doc(746,66,1).
txt_doc(765,66,4).
txt_doc(688,66,1).
txt_doc(338,66,1).
txt_doc(358,66,2).
txt_doc(582,66,1).
txt_doc(786,66,1).
txt_doc(924,66,1).
txt_doc(891,66,1).
txt_doc(52,66,1).
txt_doc(592,66,1).
txt_doc(95,66,1).
txt_doc(552,66,1).
txt_doc(238,66,1).
txt_doc(655,66,1).
txt_doc(486,66,1).
txt_doc(236,66,1).
txt_doc(121,66,1).
txt_doc(554,66,1).
txt_doc(833,66,1).
txt_doc(803,66,1).
txt_doc(467,66,1).
txt_doc(708,66,1).
txt_doc(362,66,1).
txt_doc(65,66,1).
txt_doc(887,66,2).
txt_doc(315,66,1).
txt_doc(366,67,1).
txt_doc(715,67,1).
txt_doc(409,67,1).
txt_doc(908,67,1).
txt_doc(387,67,1).
txt_doc(713,67,1).
txt_doc(66,67,1).
txt_doc(677,67,1).
txt_doc(405,67,1).
txt_doc(301,67,1).
txt_doc(470,67,1).
txt_doc(99,67,1).
txt_doc(796,67,1).
txt_doc(170,67,1).
txt_doc(185,67,1).
txt_doc(605,67,1).
txt_doc(367,67,1).
txt_doc(119,67,1).
txt_doc(757,67,1).
txt_doc(208,67,1).
txt_doc(472,67,1).
txt_doc(56,67,1).
txt_doc(238,67,1).
txt_doc(520,67,1).
txt_doc(127,67,1).
txt_doc(483,67,1).
txt_doc(221,67,1).
txt_doc(84,67,1).
txt_doc(923,67,1).
txt_doc(490,67,1).
txt_doc(747,67,1).
txt_doc(5,67,1).
txt_doc(759,67,1).
txt_doc(445,67,1).
txt_doc(756,67,1).
txt_doc(721,67,1).
txt_doc(237,67,1).
txt_doc(28,67,1).
txt_doc(688,67,1).
txt_doc(599,67,1).
txt_doc(706,67,1).
txt_doc(262,67,1).
txt_doc(620,67,1).
txt_doc(809,67,1).
txt_doc(40,67,1).
txt_doc(709,67,1).
txt_doc(294,67,2).
txt_doc(924,67,1).
txt_doc(746,67,1).
txt_doc(314,67,1).
txt_doc(386,67,1).
txt_doc(510,67,1).
txt_doc(655,67,1).
txt_doc(46,67,1).
txt_doc(853,67,1).
txt_doc(285,67,1).
txt_doc(8,67,1).
txt_doc(406,67,1).
txt_doc(174,67,1).
txt_doc(892,67,1).
txt_doc(810,67,1).
txt_doc(929,67,1).
txt_doc(135,67,1).
txt_doc(306,67,2).
txt_doc(594,67,1).
txt_doc(200,67,1).
txt_doc(233,67,1).
txt_doc(166,67,1).
txt_doc(779,67,1).
txt_doc(926,67,1).
txt_doc(765,67,1).
txt_doc(356,67,1).
txt_doc(351,67,1).
txt_doc(24,67,1).
txt_doc(298,67,1).
txt_doc(326,67,2).
txt_doc(774,67,1).
txt_doc(434,67,1).
txt_doc(537,67,1).
txt_doc(108,67,1).
txt_doc(695,67,1).
txt_doc(368,67,1).
txt_doc(88,67,1).
txt_doc(965,67,1).
txt_doc(425,67,1).
txt_doc(752,67,1).
txt_doc(83,67,1).
txt_doc(570,67,1).
txt_doc(962,67,1).
txt_doc(716,67,1).
txt_doc(109,67,1).
txt_doc(350,67,1).
txt_doc(162,67,1).
txt_doc(235,67,1).
txt_doc(286,67,1).
txt_doc(321,67,1).
txt_doc(177,67,1).
txt_doc(630,67,1).
txt_doc(346,67,1).
txt_doc(113,67,2).
txt_doc(640,67,2).
txt_doc(51,67,1).
txt_doc(667,67,1).
txt_doc(408,67,1).
txt_doc(930,67,1).
txt_doc(872,67,1).
txt_doc(110,67,1).
txt_doc(199,67,1).
txt_doc(80,67,2).
txt_doc(593,67,1).
txt_doc(172,67,1).
txt_doc(740,67,1).
txt_doc(293,67,1).
txt_doc(487,67,1).
txt_doc(880,67,2).
txt_doc(623,67,1).
txt_doc(718,67,2).
txt_doc(90,67,1).
txt_doc(764,67,1).
txt_doc(689,67,2).
txt_doc(805,67,2).
txt_doc(488,67,1).
txt_doc(35,67,1).
txt_doc(663,67,1).
txt_doc(844,67,1).
txt_doc(261,67,1).
txt_doc(949,67,1).
txt_doc(650,67,1).
txt_doc(561,67,1).
txt_doc(911,67,1).
txt_doc(867,67,1).
txt_doc(376,67,1).
txt_doc(760,67,1).
txt_doc(256,67,1).
txt_doc(11,67,1).
txt_doc(404,67,2).
txt_doc(313,67,1).
txt_doc(707,67,1).
txt_doc(22,67,1).
txt_doc(329,67,1).
txt_doc(950,67,1).
txt_doc(257,67,1).
txt_doc(146,67,1).
txt_doc(903,67,1).
txt_doc(78,67,1).
txt_doc(187,67,1).
txt_doc(243,67,1).
txt_doc(816,67,1).
txt_doc(812,67,1).
txt_doc(30,67,1).
txt_doc(907,67,1).
txt_doc(894,67,1).
txt_doc(851,67,1).
txt_doc(296,67,1).
txt_doc(428,67,1).
txt_doc(535,67,1).
txt_doc(652,67,2).
txt_doc(920,67,1).
txt_doc(10,67,1).
txt_doc(497,67,1).
txt_doc(206,67,1).
txt_doc(466,67,1).
txt_doc(474,67,2).
txt_doc(780,67,1).
txt_doc(714,67,2).
txt_doc(538,67,1).
txt_doc(89,67,1).
txt_doc(937,67,1).
txt_doc(550,67,1).
txt_doc(7,67,1).
txt_doc(414,67,2).
txt_doc(391,67,1).
txt_doc(67,67,1).
txt_doc(373,67,1).
txt_doc(508,67,1).
txt_doc(871,67,1).
txt_doc(272,67,1).
txt_doc(308,67,1).
txt_doc(899,67,1).
txt_doc(944,67,1).
txt_doc(178,67,1).
txt_doc(968,67,1).
txt_doc(273,67,2).
txt_doc(114,67,1).
txt_doc(201,68,1).
txt_doc(847,68,1).
txt_doc(775,68,1).
txt_doc(320,68,1).
txt_doc(749,68,1).
txt_doc(334,68,1).
txt_doc(520,68,1).
txt_doc(846,68,1).
txt_doc(70,68,1).
txt_doc(619,68,1).
txt_doc(238,68,1).
txt_doc(950,68,1).
txt_doc(578,68,1).
txt_doc(72,68,1).
txt_doc(427,68,1).
txt_doc(825,68,1).
txt_doc(966,68,1).
txt_doc(682,68,1).
txt_doc(877,68,1).
txt_doc(747,68,1).
txt_doc(800,68,1).
txt_doc(611,68,1).
txt_doc(622,68,1).
txt_doc(874,68,1).
txt_doc(319,68,1).
txt_doc(488,68,1).
txt_doc(98,68,1).
txt_doc(310,68,1).
txt_doc(168,68,1).
txt_doc(150,68,1).
txt_doc(835,68,2).
txt_doc(489,68,1).
txt_doc(905,68,1).
txt_doc(230,68,1).
txt_doc(655,68,1).
txt_doc(540,68,1).
txt_doc(956,68,1).
txt_doc(42,68,1).
txt_doc(205,68,1).
txt_doc(713,68,1).
txt_doc(693,68,1).
txt_doc(826,68,1).
txt_doc(739,68,1).
txt_doc(225,68,1).
txt_doc(606,68,1).
txt_doc(384,68,1).
txt_doc(389,68,2).
txt_doc(595,68,1).
txt_doc(430,68,2).
txt_doc(838,68,2).
txt_doc(175,68,2).
txt_doc(51,68,1).
txt_doc(122,68,1).
txt_doc(963,68,1).
txt_doc(255,68,1).
txt_doc(623,68,1).
txt_doc(852,68,1).
txt_doc(131,68,2).
txt_doc(934,68,1).
txt_doc(528,68,1).
txt_doc(636,68,1).
txt_doc(722,68,1).
txt_doc(63,68,1).
txt_doc(709,68,2).
txt_doc(767,68,1).
txt_doc(78,68,1).
txt_doc(432,68,1).
txt_doc(450,68,1).
txt_doc(827,68,1).
txt_doc(471,68,1).
txt_doc(89,68,1).
txt_doc(414,68,1).
txt_doc(127,68,1).
txt_doc(118,68,1).
txt_doc(300,68,1).
txt_doc(915,68,1).
txt_doc(820,68,1).
txt_doc(409,68,1).
txt_doc(730,68,1).
txt_doc(823,68,1).
txt_doc(513,68,1).
txt_doc(485,68,1).
txt_doc(433,68,1).
txt_doc(405,68,1).
txt_doc(620,68,1).
txt_doc(246,68,1).
txt_doc(125,68,1).
txt_doc(314,68,1).
txt_doc(535,68,1).
txt_doc(340,68,1).
txt_doc(551,68,1).
txt_doc(579,68,1).
txt_doc(449,68,2).
txt_doc(593,68,1).
txt_doc(833,68,1).
txt_doc(196,68,1).
txt_doc(609,68,1).
txt_doc(269,68,1).
txt_doc(696,68,1).
txt_doc(555,68,1).
txt_doc(316,68,1).
txt_doc(317,68,1).
txt_doc(876,68,1).
txt_doc(275,68,1).
txt_doc(151,68,1).
txt_doc(921,68,1).
txt_doc(306,68,1).
txt_doc(574,68,1).
txt_doc(268,68,1).
txt_doc(937,68,1).
txt_doc(226,68,1).
txt_doc(745,68,1).
txt_doc(547,68,1).
txt_doc(203,68,1).
txt_doc(519,68,1).
txt_doc(407,68,1).
txt_doc(117,68,2).
txt_doc(96,68,1).
txt_doc(479,68,1).
txt_doc(266,68,1).
txt_doc(895,68,1).
txt_doc(356,68,1).
txt_doc(12,68,1).
txt_doc(717,68,1).
txt_doc(818,68,1).
txt_doc(85,68,1).
txt_doc(560,68,1).
txt_doc(209,68,1).
txt_doc(638,68,1).
txt_doc(354,68,2).
txt_doc(442,68,1).
txt_doc(798,68,1).
txt_doc(896,68,1).
txt_doc(769,68,1).
txt_doc(464,68,1).
txt_doc(784,68,1).
txt_doc(392,68,1).
txt_doc(428,68,1).
txt_doc(110,68,1).
txt_doc(590,68,1).
txt_doc(771,68,1).
txt_doc(323,68,1).
txt_doc(415,68,1).
txt_doc(781,68,1).
txt_doc(735,68,2).
txt_doc(176,68,2).
txt_doc(311,68,1).
txt_doc(11,68,1).
txt_doc(758,68,1).
txt_doc(164,68,1).
txt_doc(137,68,1).
txt_doc(375,68,1).
txt_doc(542,68,1).
txt_doc(526,68,1).
txt_doc(725,68,1).
txt_doc(516,68,1).
txt_doc(105,68,2).
txt_doc(884,68,1).
txt_doc(104,68,1).
txt_doc(559,68,3).
txt_doc(138,68,1).
txt_doc(727,68,1).
txt_doc(52,68,1).
txt_doc(614,68,2).
txt_doc(254,68,1).
txt_doc(490,68,1).
txt_doc(245,68,1).
txt_doc(844,68,1).
txt_doc(318,68,1).
txt_doc(474,68,1).
txt_doc(598,68,1).
txt_doc(413,68,2).
txt_doc(495,68,2).
txt_doc(952,68,2).
txt_doc(88,68,1).
txt_doc(799,68,1).
txt_doc(187,68,1).
txt_doc(401,68,1).
txt_doc(499,68,1).
txt_doc(850,68,2).
txt_doc(162,69,1).
txt_doc(281,69,1).
txt_doc(122,69,1).
txt_doc(145,69,1).
txt_doc(318,69,1).
txt_doc(649,69,1).
txt_doc(66,69,1).
txt_doc(3,69,1).
txt_doc(496,69,1).
txt_doc(779,69,1).
txt_doc(901,69,1).
txt_doc(186,69,1).
txt_doc(812,69,1).
txt_doc(517,69,1).
txt_doc(878,69,1).
txt_doc(47,69,1).
txt_doc(880,69,1).
txt_doc(332,69,1).
txt_doc(163,69,1).
txt_doc(367,69,1).
txt_doc(99,69,1).
txt_doc(375,69,1).
txt_doc(560,69,1).
txt_doc(125,69,1).
txt_doc(200,69,1).
txt_doc(389,69,1).
txt_doc(617,69,1).
txt_doc(140,69,1).
txt_doc(481,69,1).
txt_doc(180,69,1).
txt_doc(52,69,1).
txt_doc(59,69,1).
txt_doc(203,69,1).
txt_doc(621,69,1).
txt_doc(460,69,1).
txt_doc(176,69,1).
txt_doc(802,69,1).
txt_doc(864,69,1).
txt_doc(516,69,1).
txt_doc(96,69,1).
txt_doc(684,69,1).
txt_doc(839,69,1).
txt_doc(101,69,1).
txt_doc(155,69,1).
txt_doc(269,69,1).
txt_doc(691,69,1).
txt_doc(166,69,1).
txt_doc(144,69,1).
txt_doc(230,69,1).
txt_doc(23,69,1).
txt_doc(314,69,1).
txt_doc(320,69,1).
txt_doc(780,69,1).
txt_doc(955,69,1).
txt_doc(925,69,1).
txt_doc(705,69,1).
txt_doc(491,69,1).
txt_doc(298,69,1).
txt_doc(843,69,1).
txt_doc(438,69,1).
txt_doc(730,69,1).
txt_doc(674,69,1).
txt_doc(816,69,1).
txt_doc(729,69,1).
txt_doc(199,69,1).
txt_doc(39,69,1).
txt_doc(485,69,1).
txt_doc(459,69,2).
txt_doc(747,69,1).
txt_doc(341,69,1).
txt_doc(749,69,1).
txt_doc(137,69,2).
txt_doc(348,69,1).
txt_doc(576,69,1).
txt_doc(762,69,1).
txt_doc(845,69,1).
txt_doc(150,69,1).
txt_doc(43,69,1).
txt_doc(819,69,1).
txt_doc(34,69,1).
txt_doc(924,69,1).
txt_doc(495,69,1).
txt_doc(832,69,1).
txt_doc(639,69,1).
txt_doc(221,69,1).
txt_doc(897,69,1).
txt_doc(274,69,1).
txt_doc(933,69,1).
txt_doc(612,69,1).
txt_doc(721,69,1).
txt_doc(593,69,1).
txt_doc(899,69,1).
txt_doc(247,69,1).
txt_doc(256,69,1).
txt_doc(211,69,1).
txt_doc(187,69,1).
txt_doc(97,69,2).
txt_doc(903,69,1).
txt_doc(896,69,1).
txt_doc(879,69,1).
txt_doc(411,69,1).
txt_doc(628,69,1).
txt_doc(436,69,2).
txt_doc(889,69,1).
txt_doc(19,69,1).
txt_doc(469,69,2).
txt_doc(941,69,1).
txt_doc(416,69,1).
txt_doc(184,69,1).
txt_doc(395,69,1).
txt_doc(220,69,1).
txt_doc(657,69,1).
txt_doc(966,69,1).
txt_doc(116,69,2).
txt_doc(88,69,3).
txt_doc(580,69,2).
txt_doc(470,69,1).
txt_doc(154,69,2).
txt_doc(556,69,1).
txt_doc(584,69,1).
txt_doc(249,69,1).
txt_doc(58,69,1).
txt_doc(945,69,1).
txt_doc(473,69,1).
txt_doc(569,69,1).
txt_doc(11,69,1).
txt_doc(596,69,1).
txt_doc(898,69,1).
txt_doc(751,69,1).
txt_doc(557,69,1).
txt_doc(310,69,1).
txt_doc(253,69,1).
txt_doc(744,69,1).
txt_doc(882,69,1).
txt_doc(648,69,1).
txt_doc(153,69,1).
txt_doc(73,69,1).
txt_doc(526,69,1).
txt_doc(407,69,1).
txt_doc(530,69,1).
txt_doc(232,69,1).
txt_doc(746,69,1).
txt_doc(29,69,1).
txt_doc(108,69,1).
txt_doc(306,69,2).
txt_doc(342,69,1).
txt_doc(728,69,1).
txt_doc(26,69,1).
txt_doc(731,69,1).
txt_doc(854,69,2).
txt_doc(894,69,1).
txt_doc(16,69,1).
txt_doc(736,69,1).
txt_doc(827,69,1).
txt_doc(550,69,1).
txt_doc(600,69,1).
txt_doc(228,69,1).
txt_doc(464,69,1).
txt_doc(659,69,1).
txt_doc(339,69,1).
txt_doc(511,69,1).
txt_doc(297,69,2).
txt_doc(366,69,1).
txt_doc(488,69,1).
txt_doc(207,69,1).
txt_doc(604,69,2).
txt_doc(208,69,1).
txt_doc(542,69,1).
txt_doc(836,69,2).
txt_doc(421,69,2).
txt_doc(217,69,1).
txt_doc(385,69,1).
txt_doc(252,69,1).
txt_doc(512,69,1).
txt_doc(105,69,2).
txt_doc(607,69,1).
txt_doc(267,69,2).
txt_doc(102,69,1).
txt_doc(892,69,1).
txt_doc(107,69,1).
txt_doc(369,69,1).
txt_doc(513,69,1).
/* (12531 entrees) */


/**
 * Table txt_page
 */

txt_page(650,1,1).
txt_page(242,1,1).
txt_page(578,1,1).
txt_page(359,1,1).
txt_page(75,1,1).
txt_page(277,1,1).
txt_page(187,1,1).
txt_page(841,1,1).
txt_page(362,1,1).
txt_page(238,1,1).
txt_page(916,1,1).
txt_page(827,1,1).
txt_page(226,1,1).
txt_page(736,1,1).
txt_page(866,1,1).
txt_page(268,1,1).
txt_page(334,1,1).
txt_page(74,1,1).
txt_page(209,1,1).
txt_page(565,1,1).
txt_page(926,2,1).
txt_page(591,2,1).
txt_page(44,2,1).
txt_page(863,2,1).
txt_page(67,2,1).
txt_page(445,2,1).
txt_page(356,2,1).
txt_page(642,2,1).
txt_page(751,2,1).
txt_page(737,2,1).
txt_page(411,2,1).
txt_page(906,2,1).
txt_page(281,2,1).
txt_page(608,2,1).
txt_page(564,2,1).
txt_page(427,2,1).
txt_page(509,2,1).
txt_page(854,2,1).
txt_page(634,2,2).
txt_page(29,3,1).
txt_page(365,3,1).
txt_page(901,3,1).
txt_page(740,3,1).
txt_page(788,3,1).
txt_page(262,3,1).
txt_page(588,3,1).
txt_page(263,3,1).
txt_page(199,3,1).
txt_page(696,3,1).
txt_page(254,3,1).
txt_page(883,3,1).
txt_page(145,3,1).
txt_page(417,3,1).
txt_page(3,3,1).
txt_page(182,3,1).
txt_page(907,3,1).
txt_page(105,3,1).
txt_page(75,3,1).
txt_page(129,3,1).
txt_page(933,4,1).
txt_page(129,4,1).
txt_page(618,4,1).
txt_page(447,4,1).
txt_page(624,4,1).
txt_page(917,4,1).
txt_page(799,4,1).
txt_page(439,4,1).
txt_page(831,4,1).
txt_page(870,4,1).
txt_page(583,4,1).
txt_page(470,4,1).
txt_page(74,4,1).
txt_page(355,4,1).
txt_page(630,4,1).
txt_page(397,4,1).
txt_page(147,4,1).
txt_page(546,4,1).
txt_page(441,4,2).
txt_page(452,5,1).
txt_page(277,5,1).
txt_page(296,5,1).
txt_page(69,5,1).
txt_page(758,5,1).
txt_page(448,5,1).
txt_page(77,5,1).
txt_page(907,5,1).
txt_page(848,5,1).
txt_page(410,5,1).
txt_page(512,5,1).
txt_page(261,5,1).
txt_page(216,5,1).
txt_page(89,5,1).
txt_page(897,5,1).
txt_page(138,5,1).
txt_page(239,5,1).
txt_page(551,5,1).
txt_page(536,5,1).
txt_page(266,5,1).
txt_page(197,6,1).
txt_page(602,6,1).
txt_page(778,6,1).
txt_page(557,6,1).
txt_page(182,6,1).
txt_page(838,6,1).
txt_page(673,6,1).
txt_page(634,6,1).
txt_page(641,6,1).
txt_page(322,6,1).
txt_page(50,6,1).
txt_page(791,6,1).
txt_page(27,6,1).
txt_page(904,6,1).
txt_page(691,6,1).
txt_page(223,6,2).
txt_page(486,6,1).
txt_page(84,6,1).
txt_page(947,6,1).
txt_page(426,7,1).
txt_page(254,7,1).
txt_page(608,7,1).
txt_page(499,7,1).
txt_page(968,7,1).
txt_page(334,7,1).
txt_page(341,7,1).
txt_page(573,7,1).
txt_page(347,7,1).
txt_page(280,7,1).
txt_page(76,7,1).
txt_page(581,7,1).
txt_page(318,7,1).
txt_page(240,7,1).
txt_page(396,7,1).
txt_page(12,7,1).
txt_page(373,7,1).
txt_page(448,7,1).
txt_page(65,7,1).
txt_page(943,7,1).
txt_page(889,8,1).
txt_page(146,8,1).
txt_page(369,8,1).
txt_page(84,8,1).
txt_page(700,8,1).
txt_page(484,8,1).
txt_page(190,8,1).
txt_page(196,8,1).
txt_page(471,8,1).
txt_page(144,8,1).
txt_page(150,8,1).
txt_page(334,8,1).
txt_page(176,8,1).
txt_page(456,8,1).
txt_page(806,8,1).
txt_page(291,8,1).
txt_page(639,8,2).
txt_page(352,8,1).
txt_page(100,8,1).
txt_page(891,9,1).
txt_page(466,9,1).
txt_page(358,9,1).
txt_page(934,9,1).
txt_page(723,9,1).
txt_page(334,9,1).
txt_page(333,9,1).
txt_page(229,9,1).
txt_page(612,9,1).
txt_page(946,9,1).
txt_page(276,9,1).
txt_page(966,9,1).
txt_page(67,9,1).
txt_page(914,9,1).
txt_page(519,9,1).
txt_page(806,9,1).
txt_page(897,9,1).
txt_page(80,9,1).
txt_page(570,9,1).
txt_page(170,9,1).
txt_page(85,10,1).
txt_page(713,10,1).
txt_page(237,10,1).
txt_page(444,10,1).
txt_page(353,10,1).
txt_page(557,10,1).
txt_page(39,10,1).
txt_page(855,10,1).
txt_page(486,10,1).
txt_page(210,10,1).
txt_page(536,10,1).
txt_page(505,10,1).
txt_page(251,10,1).
txt_page(125,10,1).
txt_page(418,10,1).
txt_page(47,10,1).
txt_page(885,10,1).
txt_page(865,10,1).
txt_page(676,10,1).
txt_page(783,10,1).
txt_page(687,11,1).
txt_page(108,11,1).
txt_page(845,11,1).
txt_page(640,11,1).
txt_page(930,11,1).
txt_page(294,11,1).
txt_page(662,11,1).
txt_page(211,11,1).
txt_page(946,11,1).
txt_page(622,11,1).
txt_page(130,11,1).
txt_page(719,11,1).
txt_page(605,11,1).
txt_page(360,11,1).
txt_page(585,11,1).
txt_page(247,11,1).
txt_page(853,11,1).
txt_page(55,11,1).
txt_page(560,11,1).
txt_page(850,11,1).
txt_page(122,12,1).
txt_page(281,12,1).
txt_page(892,12,1).
txt_page(314,12,1).
txt_page(51,12,1).
txt_page(221,12,1).
txt_page(738,12,1).
txt_page(966,12,1).
txt_page(958,12,2).
txt_page(960,12,1).
txt_page(339,12,1).
txt_page(406,12,1).
txt_page(835,12,1).
txt_page(45,12,1).
txt_page(102,12,1).
txt_page(107,12,1).
txt_page(911,12,1).
txt_page(201,12,1).
txt_page(694,12,1).
txt_page(258,13,1).
txt_page(781,13,1).
txt_page(278,13,1).
txt_page(571,13,1).
txt_page(386,13,1).
txt_page(420,13,1).
txt_page(453,13,1).
txt_page(115,13,1).
txt_page(549,13,1).
txt_page(594,13,1).
txt_page(127,13,1).
txt_page(556,13,1).
txt_page(56,13,1).
txt_page(711,13,1).
txt_page(109,13,1).
txt_page(421,13,1).
txt_page(70,13,1).
txt_page(72,13,1).
txt_page(934,13,1).
txt_page(601,13,1).
txt_page(713,14,1).
txt_page(483,14,1).
txt_page(56,14,1).
txt_page(353,14,1).
txt_page(762,14,1).
txt_page(223,14,1).
txt_page(311,14,1).
txt_page(958,14,1).
txt_page(91,14,1).
txt_page(593,14,1).
txt_page(638,14,1).
txt_page(193,14,1).
txt_page(92,14,1).
txt_page(371,14,1).
txt_page(902,14,1).
txt_page(688,14,1).
txt_page(661,14,1).
txt_page(182,14,1).
txt_page(696,14,1).
txt_page(740,14,1).
txt_page(700,15,1).
txt_page(961,15,1).
txt_page(744,15,1).
txt_page(916,15,1).
txt_page(713,15,1).
txt_page(772,15,1).
txt_page(146,15,1).
txt_page(242,15,1).
txt_page(187,15,1).
txt_page(896,15,1).
txt_page(252,15,1).
txt_page(870,15,1).
txt_page(847,15,1).
txt_page(137,15,1).
txt_page(33,15,1).
txt_page(387,15,1).
txt_page(151,15,1).
txt_page(417,15,1).
txt_page(494,15,1).
txt_page(651,15,1).
txt_page(139,16,1).
txt_page(258,16,1).
txt_page(171,16,1).
txt_page(262,16,1).
txt_page(248,16,1).
txt_page(384,16,1).
txt_page(695,16,1).
txt_page(133,16,1).
txt_page(95,16,1).
txt_page(119,16,1).
txt_page(838,16,1).
txt_page(553,16,1).
txt_page(729,16,1).
txt_page(834,16,1).
txt_page(953,16,1).
txt_page(4,16,1).
txt_page(765,16,1).
txt_page(716,16,1).
txt_page(158,16,1).
txt_page(385,16,1).
txt_page(166,17,1).
txt_page(526,17,1).
txt_page(327,17,1).
txt_page(70,17,1).
txt_page(662,17,1).
txt_page(510,17,1).
txt_page(945,17,1).
txt_page(62,17,1).
txt_page(647,17,1).
txt_page(107,17,1).
txt_page(138,17,1).
txt_page(781,17,1).
txt_page(443,17,1).
txt_page(258,17,1).
txt_page(132,17,1).
txt_page(551,17,1).
txt_page(611,17,1).
txt_page(249,17,1).
txt_page(949,17,1).
txt_page(208,17,1).
txt_page(596,18,1).
txt_page(137,18,1).
txt_page(870,18,1).
txt_page(200,18,1).
txt_page(275,18,1).
txt_page(199,18,1).
txt_page(744,18,1).
txt_page(272,18,1).
txt_page(614,18,1).
txt_page(753,18,1).
txt_page(563,18,1).
txt_page(691,18,1).
txt_page(674,18,1).
txt_page(116,18,1).
txt_page(104,18,1).
txt_page(48,18,1).
txt_page(66,18,1).
txt_page(606,18,1).
txt_page(554,18,1).
txt_page(371,18,1).
txt_page(633,19,1).
txt_page(64,19,1).
txt_page(791,19,1).
txt_page(809,19,1).
txt_page(684,19,1).
txt_page(171,19,1).
txt_page(523,19,1).
txt_page(526,19,1).
txt_page(306,19,1).
txt_page(547,19,1).
txt_page(202,19,1).
txt_page(881,19,1).
txt_page(376,19,1).
txt_page(426,19,1).
txt_page(487,19,1).
txt_page(824,19,1).
txt_page(425,19,1).
txt_page(476,19,1).
txt_page(832,19,1).
txt_page(462,19,1).
txt_page(510,20,1).
txt_page(815,20,1).
txt_page(336,20,1).
txt_page(909,20,1).
txt_page(198,20,1).
txt_page(69,20,1).
txt_page(585,20,1).
txt_page(660,20,1).
txt_page(717,20,1).
txt_page(179,20,1).
txt_page(470,20,1).
txt_page(14,20,1).
txt_page(389,20,1).
txt_page(936,20,1).
txt_page(944,20,1).
txt_page(879,20,1).
txt_page(251,20,1).
txt_page(596,20,1).
txt_page(626,20,1).
txt_page(193,20,1).
txt_page(736,21,1).
txt_page(30,21,1).
txt_page(549,21,1).
txt_page(629,21,1).
txt_page(880,21,1).
txt_page(565,21,1).
txt_page(135,21,1).
txt_page(500,21,1).
txt_page(717,21,1).
txt_page(78,21,1).
txt_page(640,21,1).
txt_page(741,21,1).
txt_page(559,21,1).
txt_page(627,21,1).
txt_page(569,21,1).
txt_page(623,21,1).
txt_page(937,21,1).
txt_page(215,21,1).
txt_page(947,21,1).
txt_page(57,21,1).
txt_page(965,22,1).
txt_page(563,22,1).
txt_page(770,22,2).
txt_page(806,22,1).
txt_page(566,22,1).
txt_page(580,22,1).
txt_page(680,22,1).
txt_page(589,22,1).
txt_page(652,22,1).
txt_page(325,22,1).
txt_page(317,22,1).
txt_page(412,22,1).
txt_page(35,22,1).
txt_page(941,22,1).
txt_page(195,22,1).
txt_page(9,22,1).
txt_page(260,22,1).
txt_page(468,22,1).
txt_page(933,22,1).
txt_page(166,23,1).
txt_page(379,23,1).
txt_page(343,23,1).
txt_page(245,23,1).
txt_page(210,23,1).
txt_page(569,23,1).
txt_page(290,23,1).
txt_page(847,23,1).
txt_page(927,23,1).
txt_page(836,23,1).
txt_page(268,23,1).
txt_page(94,23,1).
txt_page(540,23,1).
txt_page(602,23,1).
txt_page(59,23,1).
txt_page(152,23,1).
txt_page(546,23,1).
txt_page(22,23,1).
txt_page(333,23,1).
txt_page(637,23,1).
txt_page(487,24,1).
txt_page(388,24,1).
txt_page(695,24,1).
txt_page(666,24,1).
txt_page(452,24,1).
txt_page(309,24,1).
txt_page(204,24,1).
txt_page(86,24,1).
txt_page(176,24,1).
txt_page(265,24,1).
txt_page(710,24,1).
txt_page(696,24,1).
txt_page(845,24,1).
txt_page(314,24,1).
txt_page(642,24,1).
txt_page(109,24,1).
txt_page(603,24,1).
txt_page(906,24,1).
txt_page(486,24,1).
txt_page(18,24,1).
txt_page(931,25,1).
txt_page(182,25,1).
txt_page(516,25,1).
txt_page(877,25,1).
txt_page(4,25,1).
txt_page(755,25,1).
txt_page(482,25,1).
txt_page(837,25,1).
txt_page(843,25,1).
txt_page(437,25,1).
txt_page(564,25,1).
txt_page(111,25,1).
txt_page(284,25,1).
txt_page(829,25,1).
txt_page(457,25,1).
txt_page(661,25,1).
txt_page(901,25,1).
txt_page(277,25,1).
txt_page(598,25,1).
txt_page(33,25,1).
txt_page(719,26,1).
txt_page(113,26,1).
txt_page(269,26,1).
txt_page(662,26,1).
txt_page(959,26,1).
txt_page(421,26,1).
txt_page(790,26,1).
txt_page(833,26,1).
txt_page(583,26,1).
txt_page(845,26,1).
txt_page(378,26,1).
txt_page(383,26,1).
txt_page(472,26,1).
txt_page(588,26,1).
txt_page(629,26,1).
txt_page(422,26,1).
txt_page(275,26,1).
txt_page(365,26,1).
txt_page(738,26,1).
txt_page(607,26,1).
txt_page(38,27,1).
txt_page(858,27,1).
txt_page(843,27,1).
txt_page(944,27,1).
txt_page(223,27,1).
txt_page(418,27,1).
txt_page(964,27,1).
txt_page(943,27,1).
txt_page(442,27,1).
txt_page(774,27,1).
txt_page(918,27,1).
txt_page(312,27,1).
txt_page(855,27,1).
txt_page(221,27,1).
txt_page(821,27,2).
txt_page(156,27,1).
txt_page(351,27,1).
txt_page(160,27,1).
txt_page(296,27,1).
txt_page(413,28,1).
txt_page(796,28,1).
txt_page(946,28,1).
txt_page(67,28,1).
txt_page(896,28,1).
txt_page(432,28,1).
txt_page(571,28,1).
txt_page(897,28,1).
txt_page(784,28,1).
txt_page(194,28,1).
txt_page(827,28,1).
txt_page(35,28,1).
txt_page(595,28,1).
txt_page(230,28,1).
txt_page(416,28,1).
txt_page(620,28,1).
txt_page(431,28,1).
txt_page(508,28,1).
txt_page(874,28,1).
txt_page(678,28,1).
txt_page(584,29,1).
txt_page(520,29,1).
txt_page(878,29,1).
txt_page(914,29,1).
txt_page(178,29,1).
txt_page(734,29,1).
txt_page(302,29,1).
txt_page(580,29,1).
txt_page(143,29,1).
txt_page(11,29,1).
txt_page(968,29,1).
txt_page(635,29,1).
txt_page(130,29,1).
txt_page(589,29,1).
txt_page(691,29,1).
txt_page(375,29,1).
txt_page(471,29,1).
txt_page(503,29,1).
txt_page(351,29,1).
txt_page(601,29,1).
txt_page(68,30,1).
txt_page(611,30,1).
txt_page(352,30,1).
txt_page(44,30,1).
txt_page(914,30,1).
txt_page(792,30,1).
txt_page(902,30,1).
txt_page(452,30,1).
txt_page(299,30,1).
txt_page(215,30,1).
txt_page(335,30,1).
txt_page(418,30,1).
txt_page(596,30,1).
txt_page(935,30,1).
txt_page(864,30,1).
txt_page(852,30,1).
txt_page(729,30,1).
txt_page(614,30,2).
txt_page(120,30,1).
txt_page(116,31,1).
txt_page(958,31,1).
txt_page(18,31,1).
txt_page(273,31,1).
txt_page(754,31,1).
txt_page(897,31,1).
txt_page(588,31,1).
txt_page(710,31,1).
txt_page(390,31,1).
txt_page(79,31,1).
txt_page(304,31,1).
txt_page(274,31,1).
txt_page(201,31,1).
txt_page(816,31,1).
txt_page(422,31,1).
txt_page(685,31,1).
txt_page(216,31,1).
txt_page(325,31,1).
txt_page(817,31,1).
txt_page(801,31,1).
txt_page(812,32,1).
txt_page(30,32,1).
txt_page(53,32,1).
txt_page(500,32,1).
txt_page(925,32,1).
txt_page(968,32,1).
txt_page(728,32,1).
txt_page(11,32,1).
txt_page(720,32,1).
txt_page(107,32,1).
txt_page(772,32,1).
txt_page(402,32,1).
txt_page(68,32,1).
txt_page(770,32,1).
txt_page(954,32,1).
txt_page(623,32,1).
txt_page(139,32,1).
txt_page(613,32,1).
txt_page(315,32,1).
txt_page(14,32,1).
txt_page(877,33,1).
txt_page(638,33,1).
txt_page(210,33,1).
txt_page(951,33,1).
txt_page(213,33,1).
txt_page(344,33,1).
txt_page(327,33,1).
txt_page(942,33,1).
txt_page(400,33,1).
txt_page(566,33,1).
txt_page(645,33,1).
txt_page(172,33,1).
txt_page(260,33,1).
txt_page(398,33,1).
txt_page(899,33,1).
txt_page(346,33,1).
txt_page(824,33,1).
txt_page(165,33,1).
txt_page(695,33,1).
txt_page(957,33,1).
txt_page(677,34,1).
txt_page(882,34,1).
txt_page(857,34,1).
txt_page(359,34,1).
txt_page(862,34,1).
txt_page(259,34,1).
txt_page(65,34,1).
txt_page(848,34,1).
txt_page(393,34,1).
txt_page(339,34,1).
txt_page(759,34,1).
txt_page(349,34,1).
txt_page(457,34,1).
txt_page(254,34,1).
txt_page(641,34,1).
txt_page(660,34,1).
txt_page(19,34,1).
txt_page(786,34,1).
txt_page(469,34,1).
txt_page(211,34,1).
txt_page(79,35,1).
txt_page(891,35,1).
txt_page(888,35,1).
txt_page(684,35,1).
txt_page(441,35,1).
txt_page(371,35,1).
txt_page(491,35,1).
txt_page(562,35,1).
txt_page(769,35,1).
txt_page(359,35,1).
txt_page(461,35,1).
txt_page(138,35,1).
txt_page(892,35,1).
txt_page(358,35,1).
txt_page(789,35,1).
txt_page(480,35,1).
txt_page(558,35,1).
txt_page(751,35,1).
txt_page(322,35,1).
txt_page(517,35,1).
txt_page(254,36,1).
txt_page(856,36,1).
txt_page(708,36,1).
txt_page(533,36,1).
txt_page(805,36,1).
txt_page(461,36,1).
txt_page(240,36,1).
txt_page(95,36,1).
txt_page(118,36,1).
txt_page(109,36,1).
txt_page(70,36,1).
txt_page(807,36,1).
txt_page(870,36,1).
txt_page(228,36,1).
txt_page(505,36,1).
txt_page(859,36,1).
txt_page(112,36,2).
txt_page(96,36,1).
txt_page(426,36,1).
txt_page(280,37,1).
txt_page(311,37,2).
txt_page(113,37,1).
txt_page(784,37,1).
txt_page(310,37,1).
txt_page(527,37,1).
txt_page(671,37,1).
txt_page(84,37,1).
txt_page(656,37,1).
txt_page(909,37,1).
txt_page(4,37,1).
txt_page(10,37,1).
txt_page(936,37,1).
txt_page(788,37,1).
txt_page(644,37,1).
txt_page(245,37,1).
txt_page(896,37,1).
txt_page(72,37,1).
txt_page(337,37,1).
txt_page(284,38,1).
txt_page(605,38,1).
txt_page(414,38,1).
txt_page(77,38,1).
txt_page(914,38,1).
txt_page(156,38,1).
txt_page(701,38,1).
txt_page(523,38,1).
txt_page(804,38,1).
txt_page(576,38,1).
txt_page(882,38,1).
txt_page(171,38,1).
txt_page(31,38,1).
txt_page(550,38,1).
txt_page(308,38,1).
txt_page(587,38,1).
txt_page(785,38,1).
txt_page(384,38,1).
txt_page(627,38,1).
txt_page(644,38,1).
txt_page(332,39,1).
txt_page(507,39,1).
txt_page(478,39,1).
txt_page(60,39,1).
txt_page(84,39,1).
txt_page(781,39,1).
txt_page(611,39,1).
txt_page(439,39,1).
txt_page(580,39,1).
txt_page(606,39,1).
txt_page(36,39,1).
txt_page(861,39,1).
txt_page(695,39,1).
txt_page(865,39,1).
txt_page(674,39,1).
txt_page(446,39,1).
txt_page(42,39,1).
txt_page(203,39,1).
txt_page(809,39,1).
txt_page(315,39,1).
txt_page(39,40,1).
txt_page(267,40,1).
txt_page(520,40,1).
txt_page(425,40,1).
txt_page(271,40,1).
txt_page(925,40,1).
txt_page(411,40,1).
txt_page(607,40,1).
txt_page(543,40,1).
txt_page(384,40,1).
txt_page(734,40,1).
txt_page(385,40,1).
txt_page(849,40,1).
txt_page(500,40,1).
txt_page(78,40,1).
txt_page(103,40,1).
txt_page(545,40,1).
txt_page(929,40,1).
txt_page(275,40,1).
txt_page(902,40,1).
txt_page(624,41,1).
txt_page(683,41,1).
txt_page(404,41,1).
txt_page(230,41,1).
txt_page(729,41,1).
txt_page(767,41,1).
txt_page(635,41,1).
txt_page(447,41,1).
txt_page(334,41,1).
txt_page(659,41,1).
txt_page(172,41,1).
txt_page(798,41,1).
txt_page(645,41,1).
txt_page(285,41,1).
txt_page(171,41,1).
txt_page(175,41,1).
txt_page(725,41,1).
txt_page(707,41,1).
txt_page(177,41,1).
txt_page(785,41,1).
txt_page(838,42,1).
txt_page(337,42,1).
txt_page(353,42,1).
txt_page(248,42,1).
txt_page(600,42,1).
txt_page(701,42,1).
txt_page(462,42,1).
txt_page(525,42,1).
txt_page(97,42,1).
txt_page(19,42,1).
txt_page(875,42,1).
txt_page(675,42,1).
txt_page(168,42,1).
txt_page(16,42,1).
txt_page(273,42,1).
txt_page(938,42,1).
txt_page(730,42,1).
txt_page(601,42,1).
txt_page(129,42,1).
txt_page(113,42,1).
txt_page(34,43,1).
txt_page(602,43,1).
txt_page(729,43,1).
txt_page(513,43,1).
txt_page(850,43,1).
txt_page(728,43,1).
txt_page(106,43,1).
txt_page(505,43,1).
txt_page(508,43,1).
txt_page(749,43,1).
txt_page(429,43,1).
txt_page(236,43,1).
txt_page(13,43,1).
txt_page(735,43,1).
txt_page(460,43,1).
txt_page(161,43,1).
txt_page(324,43,1).
txt_page(531,43,1).
txt_page(37,43,1).
txt_page(805,43,1).
txt_page(826,44,1).
txt_page(596,44,1).
txt_page(655,44,1).
txt_page(453,44,1).
txt_page(321,44,1).
txt_page(109,44,1).
txt_page(169,44,1).
txt_page(498,44,1).
txt_page(483,44,1).
txt_page(616,44,1).
txt_page(522,44,1).
txt_page(903,44,1).
txt_page(241,44,1).
txt_page(264,44,1).
txt_page(116,44,1).
txt_page(521,44,1).
txt_page(663,44,1).
txt_page(550,44,1).
txt_page(861,44,1).
txt_page(84,44,1).
txt_page(483,45,1).
txt_page(445,45,1).
txt_page(632,45,1).
txt_page(326,45,1).
txt_page(883,45,1).
txt_page(544,45,1).
txt_page(630,45,1).
txt_page(392,45,1).
txt_page(191,45,1).
txt_page(916,45,1).
txt_page(733,45,1).
txt_page(671,45,1).
txt_page(815,45,1).
txt_page(229,45,1).
txt_page(821,45,1).
txt_page(337,45,1).
txt_page(741,45,1).
txt_page(115,45,1).
txt_page(595,45,1).
txt_page(223,45,1).
txt_page(652,46,1).
txt_page(516,46,1).
txt_page(16,46,1).
txt_page(646,46,1).
txt_page(153,46,1).
txt_page(829,46,1).
txt_page(945,46,1).
txt_page(29,46,1).
txt_page(456,46,1).
txt_page(375,46,1).
txt_page(768,46,1).
txt_page(309,46,1).
txt_page(795,46,1).
txt_page(629,46,1).
txt_page(182,46,1).
txt_page(221,46,1).
txt_page(555,46,1).
txt_page(240,46,1).
txt_page(863,46,1).
txt_page(463,46,1).
txt_page(19,47,1).
txt_page(875,47,1).
txt_page(3,47,1).
txt_page(948,47,1).
txt_page(266,47,1).
txt_page(350,47,1).
txt_page(221,47,1).
txt_page(300,47,1).
txt_page(20,47,1).
txt_page(936,47,1).
txt_page(863,47,1).
txt_page(256,47,1).
txt_page(928,47,1).
txt_page(348,47,1).
txt_page(676,47,1).
txt_page(424,47,1).
txt_page(924,47,1).
txt_page(833,47,1).
txt_page(656,47,1).
txt_page(627,47,1).
txt_page(712,48,1).
txt_page(68,48,1).
txt_page(962,48,1).
txt_page(257,48,1).
txt_page(843,48,1).
txt_page(301,48,1).
txt_page(230,48,1).
txt_page(217,48,1).
txt_page(398,48,1).
txt_page(670,48,1).
txt_page(866,48,1).
txt_page(584,48,1).
txt_page(599,48,1).
txt_page(646,48,1).
txt_page(421,48,1).
txt_page(671,48,1).
txt_page(315,48,1).
txt_page(407,48,1).
txt_page(650,48,1).
txt_page(388,48,1).
txt_page(686,49,1).
txt_page(505,49,1).
txt_page(644,49,1).
txt_page(674,49,1).
txt_page(182,49,1).
txt_page(368,49,1).
txt_page(566,49,1).
txt_page(675,49,1).
txt_page(224,49,1).
txt_page(328,49,1).
txt_page(165,49,1).
txt_page(289,49,1).
txt_page(521,49,1).
txt_page(228,49,1).
txt_page(950,49,1).
txt_page(315,49,1).
txt_page(244,49,1).
txt_page(249,49,1).
txt_page(719,49,1).
txt_page(754,49,1).
txt_page(857,50,1).
txt_page(163,50,1).
txt_page(275,50,1).
txt_page(801,50,1).
txt_page(220,50,1).
txt_page(570,50,1).
txt_page(664,50,1).
txt_page(848,50,1).
txt_page(53,50,1).
txt_page(50,50,1).
txt_page(169,50,2).
txt_page(918,50,1).
txt_page(74,50,1).
txt_page(836,50,1).
txt_page(456,50,1).
txt_page(85,50,1).
txt_page(359,50,1).
txt_page(509,50,1).
txt_page(522,50,1).
txt_page(241,51,1).
txt_page(684,51,1).
txt_page(831,51,1).
txt_page(827,51,1).
txt_page(147,51,1).
txt_page(363,51,1).
txt_page(695,51,1).
txt_page(472,51,1).
txt_page(290,51,1).
txt_page(520,51,1).
txt_page(852,51,1).
txt_page(483,51,1).
txt_page(940,51,1).
txt_page(85,51,1).
txt_page(501,51,1).
txt_page(475,51,1).
txt_page(631,51,1).
txt_page(533,51,1).
txt_page(626,51,1).
txt_page(480,51,1).
txt_page(685,52,1).
txt_page(201,52,1).
txt_page(222,52,1).
txt_page(819,52,1).
txt_page(558,52,1).
txt_page(712,52,1).
txt_page(371,52,1).
txt_page(593,52,1).
txt_page(103,52,1).
txt_page(842,52,1).
txt_page(799,52,1).
txt_page(840,52,1).
txt_page(589,52,1).
txt_page(521,52,1).
txt_page(866,52,1).
txt_page(258,52,1).
txt_page(414,52,1).
txt_page(36,52,1).
txt_page(30,52,1).
txt_page(511,52,1).
txt_page(597,53,1).
txt_page(901,53,1).
txt_page(580,53,1).
txt_page(672,53,1).
txt_page(931,53,1).
txt_page(810,53,1).
txt_page(372,53,1).
txt_page(898,53,1).
txt_page(72,53,1).
txt_page(75,53,1).
txt_page(846,53,1).
txt_page(835,53,1).
txt_page(632,53,1).
txt_page(508,53,1).
txt_page(479,53,1).
txt_page(644,53,1).
txt_page(601,53,1).
txt_page(165,53,1).
txt_page(84,53,1).
txt_page(221,53,1).
txt_page(792,54,1).
txt_page(826,54,1).
txt_page(923,54,1).
txt_page(871,54,1).
txt_page(782,54,1).
txt_page(851,54,1).
txt_page(383,54,1).
txt_page(576,54,1).
txt_page(346,54,1).
txt_page(554,54,1).
txt_page(725,54,1).
txt_page(663,54,1).
txt_page(118,54,1).
txt_page(83,54,1).
txt_page(690,54,1).
txt_page(927,54,1).
txt_page(105,54,1).
txt_page(419,54,1).
txt_page(541,54,1).
txt_page(879,54,1).
txt_page(967,55,1).
txt_page(934,55,1).
txt_page(226,55,1).
txt_page(381,55,1).
txt_page(200,55,1).
txt_page(552,55,1).
txt_page(13,55,1).
txt_page(12,55,1).
txt_page(538,55,1).
txt_page(245,55,1).
txt_page(585,55,1).
txt_page(766,55,1).
txt_page(398,55,1).
txt_page(354,55,1).
txt_page(844,55,1).
txt_page(510,55,1).
txt_page(196,55,1).
txt_page(532,55,1).
txt_page(82,55,1).
txt_page(428,55,1).
txt_page(341,56,1).
txt_page(918,56,1).
txt_page(66,56,1).
txt_page(297,56,1).
txt_page(115,56,1).
txt_page(760,56,1).
txt_page(31,56,1).
txt_page(270,56,1).
txt_page(561,56,1).
txt_page(267,56,1).
txt_page(724,56,1).
txt_page(309,56,1).
txt_page(814,56,1).
txt_page(562,56,1).
txt_page(661,56,1).
txt_page(859,56,1).
txt_page(86,56,1).
txt_page(352,56,1).
txt_page(729,56,1).
txt_page(702,56,1).
txt_page(787,57,1).
txt_page(74,57,1).
txt_page(425,57,1).
txt_page(209,57,1).
txt_page(735,57,1).
txt_page(568,57,1).
txt_page(462,57,1).
txt_page(16,57,1).
txt_page(824,57,1).
txt_page(931,57,1).
txt_page(828,57,1).
txt_page(288,57,1).
txt_page(700,57,1).
txt_page(125,57,1).
txt_page(642,57,1).
txt_page(940,57,1).
txt_page(860,57,1).
txt_page(809,57,1).
txt_page(59,57,1).
txt_page(194,57,1).
txt_page(258,58,1).
txt_page(607,58,1).
txt_page(492,58,1).
txt_page(185,58,1).
txt_page(256,58,1).
txt_page(739,58,1).
txt_page(619,58,1).
txt_page(805,58,1).
txt_page(263,58,1).
txt_page(706,58,1).
txt_page(383,58,1).
txt_page(131,58,1).
txt_page(71,58,1).
txt_page(820,58,1).
txt_page(837,58,1).
txt_page(202,58,1).
txt_page(558,58,1).
txt_page(807,58,1).
txt_page(631,58,1).
txt_page(551,58,1).
txt_page(742,59,1).
txt_page(905,59,1).
txt_page(745,59,1).
txt_page(286,59,1).
txt_page(840,59,1).
txt_page(504,59,1).
txt_page(35,59,1).
txt_page(557,59,1).
txt_page(288,59,1).
txt_page(87,59,1).
txt_page(709,59,1).
txt_page(484,59,1).
txt_page(881,59,1).
txt_page(272,59,1).
txt_page(48,59,1).
txt_page(347,59,1).
txt_page(542,59,1).
txt_page(313,59,1).
txt_page(122,59,1).
txt_page(523,59,1).
txt_page(924,60,1).
txt_page(933,60,1).
txt_page(719,60,1).
txt_page(330,60,1).
txt_page(791,60,1).
txt_page(151,60,1).
txt_page(128,60,1).
txt_page(403,60,1).
txt_page(144,60,1).
txt_page(17,60,1).
txt_page(414,60,1).
txt_page(206,60,1).
txt_page(47,60,1).
txt_page(15,60,1).
txt_page(944,60,1).
txt_page(850,60,1).
txt_page(649,60,1).
txt_page(955,60,1).
txt_page(580,60,1).
txt_page(876,60,1).
txt_page(895,61,1).
txt_page(93,61,1).
txt_page(678,61,1).
txt_page(883,61,1).
txt_page(68,61,1).
txt_page(889,61,1).
txt_page(376,61,1).
txt_page(699,61,1).
txt_page(167,61,1).
txt_page(296,61,1).
txt_page(84,61,1).
txt_page(4,61,1).
txt_page(311,61,1).
txt_page(106,61,1).
txt_page(185,61,1).
txt_page(638,61,1).
txt_page(129,61,1).
txt_page(456,61,2).
txt_page(820,61,1).
txt_page(147,62,1).
txt_page(32,62,1).
txt_page(751,62,1).
txt_page(533,62,1).
txt_page(410,62,1).
txt_page(780,62,1).
txt_page(48,62,1).
txt_page(848,62,1).
txt_page(584,62,1).
txt_page(894,62,1).
txt_page(1,62,1).
txt_page(163,62,1).
txt_page(207,62,1).
txt_page(851,62,1).
txt_page(662,62,1).
txt_page(379,62,1).
txt_page(401,62,1).
txt_page(635,62,1).
txt_page(365,62,1).
txt_page(588,62,1).
txt_page(835,63,1).
txt_page(489,63,1).
txt_page(373,63,1).
txt_page(185,63,1).
txt_page(336,63,1).
txt_page(784,63,1).
txt_page(198,63,1).
txt_page(483,63,1).
txt_page(908,63,1).
txt_page(948,63,1).
txt_page(386,63,1).
txt_page(84,63,1).
txt_page(50,63,1).
txt_page(566,63,1).
txt_page(471,63,1).
txt_page(836,63,1).
txt_page(647,63,1).
txt_page(765,63,1).
txt_page(263,63,1).
txt_page(476,63,1).
txt_page(34,64,1).
txt_page(884,64,1).
txt_page(181,64,1).
txt_page(9,64,1).
txt_page(133,64,1).
txt_page(334,64,1).
txt_page(347,64,1).
txt_page(382,64,1).
txt_page(819,64,1).
txt_page(280,64,1).
txt_page(85,64,1).
txt_page(699,64,1).
txt_page(30,64,1).
txt_page(710,64,1).
txt_page(944,64,1).
txt_page(487,64,1).
txt_page(795,64,1).
txt_page(787,64,1).
txt_page(965,64,1).
txt_page(945,64,1).
txt_page(77,65,1).
txt_page(361,65,1).
txt_page(793,65,1).
txt_page(766,65,1).
txt_page(564,65,1).
txt_page(938,65,1).
txt_page(666,65,1).
txt_page(41,65,1).
txt_page(425,65,1).
txt_page(162,65,1).
txt_page(307,65,1).
txt_page(673,65,1).
txt_page(240,65,1).
txt_page(413,65,1).
txt_page(679,65,1).
txt_page(473,65,1).
txt_page(858,65,1).
txt_page(797,65,1).
txt_page(288,65,1).
txt_page(694,65,1).
txt_page(940,66,1).
txt_page(698,66,1).
txt_page(65,66,1).
txt_page(205,66,1).
txt_page(643,66,1).
txt_page(330,66,1).
txt_page(724,66,1).
txt_page(233,66,1).
txt_page(954,66,1).
txt_page(17,66,1).
txt_page(73,66,1).
txt_page(375,66,1).
txt_page(231,66,1).
txt_page(253,66,1).
txt_page(423,66,1).
txt_page(943,66,1).
txt_page(786,66,1).
txt_page(275,66,1).
txt_page(444,66,1).
txt_page(861,66,1).
txt_page(952,67,1).
txt_page(182,67,1).
txt_page(853,67,1).
txt_page(518,67,1).
txt_page(643,67,1).
txt_page(106,67,1).
txt_page(701,67,1).
txt_page(521,67,1).
txt_page(73,67,1).
txt_page(219,67,1).
txt_page(455,67,1).
txt_page(338,67,1).
txt_page(368,67,1).
txt_page(86,67,1).
txt_page(914,67,1).
txt_page(297,67,1).
txt_page(434,67,1).
txt_page(478,67,1).
txt_page(534,67,1).
txt_page(204,67,1).
txt_page(369,68,1).
txt_page(161,68,1).
txt_page(961,68,1).
txt_page(420,68,1).
txt_page(676,68,1).
txt_page(511,68,1).
txt_page(927,68,1).
txt_page(552,68,1).
txt_page(275,68,1).
txt_page(569,68,1).
txt_page(61,68,1).
txt_page(238,68,1).
txt_page(248,68,1).
txt_page(480,68,1).
txt_page(706,68,1).
txt_page(822,68,1).
txt_page(750,68,1).
txt_page(120,68,1).
txt_page(338,68,1).
txt_page(380,68,1).
txt_page(942,69,1).
txt_page(337,69,1).
txt_page(759,69,1).
txt_page(847,69,1).
txt_page(267,69,1).
txt_page(73,69,1).
txt_page(887,69,1).
txt_page(75,69,1).
txt_page(491,69,1).
txt_page(681,69,1).
txt_page(865,69,1).
txt_page(783,69,1).
txt_page(733,69,1).
txt_page(120,69,1).
txt_page(96,69,1).
txt_page(823,69,1).
txt_page(526,69,1).
txt_page(38,69,1).
txt_page(582,69,1).
txt_page(765,69,1).
txt_page(726,70,1).
txt_page(479,70,1).
txt_page(286,70,1).
txt_page(472,70,1).
txt_page(894,70,1).
txt_page(941,70,1).
txt_page(645,70,1).
txt_page(868,70,1).
txt_page(271,70,1).
txt_page(57,70,1).
txt_page(582,70,1).
txt_page(474,70,1).
txt_page(234,70,1).
txt_page(764,70,1).
txt_page(179,70,1).
txt_page(879,70,1).
txt_page(901,70,1).
txt_page(889,70,1).
txt_page(561,70,1).
txt_page(752,70,1).
txt_page(290,71,1).
txt_page(109,71,1).
txt_page(788,71,1).
txt_page(760,71,1).
txt_page(723,71,1).
txt_page(731,71,1).
txt_page(354,71,1).
txt_page(227,71,1).
txt_page(103,71,1).
txt_page(550,71,1).
txt_page(395,71,1).
txt_page(733,71,1).
txt_page(941,71,1).
txt_page(695,71,1).
txt_page(430,71,1).
txt_page(823,71,1).
txt_page(743,71,1).
txt_page(565,71,1).
txt_page(908,71,1).
txt_page(381,71,1).
txt_page(900,72,1).
txt_page(883,72,1).
txt_page(663,72,1).
txt_page(328,72,1).
txt_page(685,72,1).
txt_page(4,72,1).
txt_page(770,72,1).
txt_page(53,72,1).
txt_page(455,72,1).
txt_page(586,72,1).
txt_page(619,72,1).
txt_page(279,72,1).
txt_page(147,72,1).
txt_page(30,72,1).
txt_page(420,72,1).
txt_page(651,72,1).
txt_page(490,72,1).
txt_page(125,72,1).
txt_page(669,72,1).
txt_page(88,72,1).
txt_page(657,73,1).
txt_page(100,73,1).
txt_page(821,73,1).
txt_page(899,73,1).
txt_page(47,73,1).
txt_page(349,73,1).
txt_page(114,73,2).
txt_page(948,73,1).
txt_page(710,73,1).
txt_page(540,73,1).
txt_page(445,73,1).
txt_page(563,73,1).
txt_page(610,73,1).
txt_page(950,73,1).
txt_page(848,73,1).
txt_page(626,73,1).
txt_page(954,73,1).
txt_page(600,73,1).
txt_page(889,73,1).
txt_page(479,74,1).
txt_page(740,74,1).
txt_page(946,74,1).
txt_page(163,74,1).
txt_page(910,74,1).
txt_page(217,74,1).
txt_page(658,74,1).
txt_page(371,74,1).
txt_page(781,74,1).
txt_page(511,74,1).
txt_page(200,74,1).
txt_page(586,74,1).
txt_page(959,74,1).
txt_page(25,74,1).
txt_page(783,74,1).
txt_page(224,74,2).
txt_page(267,74,1).
txt_page(143,74,1).
txt_page(141,74,1).
txt_page(823,75,1).
txt_page(602,75,1).
txt_page(54,75,1).
txt_page(944,75,1).
txt_page(926,75,1).
txt_page(698,75,1).
txt_page(266,75,1).
txt_page(749,75,1).
txt_page(82,75,1).
txt_page(75,75,1).
txt_page(281,75,1).
txt_page(621,75,1).
txt_page(30,75,1).
txt_page(719,75,1).
txt_page(559,75,1).
txt_page(760,75,1).
txt_page(402,75,1).
txt_page(453,75,1).
txt_page(670,75,1).
txt_page(438,75,1).
txt_page(82,76,1).
txt_page(758,76,1).
txt_page(52,76,1).
txt_page(611,76,1).
txt_page(940,76,1).
txt_page(657,76,1).
txt_page(409,76,1).
txt_page(423,76,1).
txt_page(283,76,1).
txt_page(535,76,1).
txt_page(926,76,1).
txt_page(716,76,1).
txt_page(344,76,1).
txt_page(97,76,1).
txt_page(151,76,1).
txt_page(822,76,1).
txt_page(587,76,1).
txt_page(416,76,1).
txt_page(700,76,1).
txt_page(232,76,1).
txt_page(606,77,1).
txt_page(354,77,1).
txt_page(458,77,1).
txt_page(307,77,1).
txt_page(368,77,1).
txt_page(797,77,1).
txt_page(53,77,1).
txt_page(652,77,1).
txt_page(812,77,1).
txt_page(380,77,1).
txt_page(436,77,1).
txt_page(916,77,1).
txt_page(408,77,1).
txt_page(164,77,1).
txt_page(396,77,2).
txt_page(249,77,1).
txt_page(776,77,1).
txt_page(335,77,1).
txt_page(161,77,1).
txt_page(336,78,1).
txt_page(678,78,1).
txt_page(389,78,1).
txt_page(368,78,1).
txt_page(133,78,1).
txt_page(521,78,1).
txt_page(855,78,1).
txt_page(677,78,1).
txt_page(566,78,1).
txt_page(954,78,1).
txt_page(64,78,1).
txt_page(751,78,1).
txt_page(284,78,1).
txt_page(453,78,1).
txt_page(656,78,1).
txt_page(852,78,1).
txt_page(883,78,1).
txt_page(940,78,1).
txt_page(632,78,1).
txt_page(273,78,1).
txt_page(258,79,1).
txt_page(264,79,1).
txt_page(721,79,1).
txt_page(51,79,1).
txt_page(619,79,1).
txt_page(124,79,1).
txt_page(523,79,1).
txt_page(491,79,1).
txt_page(966,79,1).
txt_page(345,79,1).
txt_page(780,79,1).
txt_page(223,79,1).
txt_page(225,79,1).
txt_page(927,79,1).
txt_page(85,79,1).
txt_page(285,79,1).
txt_page(362,79,1).
txt_page(189,79,1).
txt_page(564,79,1).
txt_page(423,79,1).
txt_page(321,80,1).
txt_page(394,80,1).
txt_page(545,80,1).
txt_page(405,80,1).
txt_page(865,80,1).
txt_page(283,80,1).
txt_page(231,80,1).
txt_page(670,80,1).
txt_page(633,80,1).
txt_page(892,80,1).
txt_page(79,80,1).
txt_page(568,80,1).
txt_page(451,80,1).
txt_page(843,80,1).
txt_page(176,80,1).
txt_page(266,80,1).
txt_page(933,80,1).
txt_page(30,80,1).
txt_page(445,80,1).
txt_page(459,80,1).
txt_page(390,81,1).
txt_page(214,81,1).
txt_page(90,81,1).
txt_page(572,81,1).
txt_page(363,81,1).
txt_page(640,81,1).
txt_page(721,81,1).
txt_page(362,81,1).
txt_page(50,81,1).
txt_page(567,81,1).
txt_page(370,81,1).
txt_page(256,81,1).
txt_page(247,81,1).
txt_page(368,81,1).
txt_page(916,81,1).
txt_page(386,81,1).
txt_page(775,81,1).
txt_page(181,81,2).
txt_page(753,81,1).
txt_page(898,82,1).
txt_page(573,82,1).
txt_page(564,82,1).
txt_page(157,82,1).
txt_page(446,82,1).
txt_page(684,82,1).
txt_page(342,82,1).
txt_page(155,82,1).
txt_page(281,82,1).
txt_page(259,82,1).
txt_page(707,82,1).
txt_page(562,82,1).
txt_page(263,82,1).
txt_page(877,82,1).
txt_page(551,82,1).
txt_page(169,82,1).
txt_page(370,82,1).
txt_page(472,82,1).
txt_page(429,82,2).
txt_page(123,83,1).
txt_page(347,83,1).
txt_page(553,83,1).
txt_page(569,83,1).
txt_page(515,83,1).
txt_page(680,83,1).
txt_page(519,83,1).
txt_page(546,83,1).
txt_page(825,83,1).
txt_page(721,83,1).
txt_page(282,83,1).
txt_page(734,83,1).
txt_page(762,83,1).
txt_page(807,83,1).
txt_page(806,83,1).
txt_page(772,83,1).
txt_page(886,83,1).
txt_page(883,83,1).
txt_page(172,83,1).
txt_page(781,83,1).
txt_page(231,84,1).
txt_page(504,84,1).
txt_page(72,84,1).
txt_page(164,84,1).
txt_page(131,84,1).
txt_page(326,84,1).
txt_page(56,84,1).
txt_page(744,84,1).
txt_page(454,84,1).
txt_page(962,84,1).
txt_page(651,84,1).
txt_page(364,84,1).
txt_page(81,84,1).
txt_page(170,84,1).
txt_page(389,84,1).
txt_page(321,84,1).
txt_page(585,84,1).
txt_page(764,84,1).
txt_page(967,84,1).
txt_page(537,84,1).
txt_page(276,85,1).
txt_page(770,85,1).
txt_page(486,85,1).
txt_page(252,85,1).
txt_page(379,85,1).
txt_page(700,85,1).
txt_page(318,85,1).
txt_page(296,85,1).
txt_page(622,85,1).
txt_page(684,85,1).
txt_page(255,85,1).
txt_page(136,85,1).
txt_page(265,85,1).
txt_page(499,85,1).
txt_page(951,85,1).
txt_page(957,85,1).
txt_page(34,85,1).
txt_page(519,85,1).
txt_page(484,85,1).
txt_page(749,85,1).
txt_page(566,86,1).
txt_page(178,86,1).
txt_page(857,86,1).
txt_page(535,86,1).
txt_page(370,86,1).
txt_page(94,86,1).
txt_page(798,86,1).
txt_page(163,86,1).
txt_page(860,86,1).
txt_page(729,86,1).
txt_page(633,86,1).
txt_page(395,86,1).
txt_page(153,86,1).
txt_page(244,86,1).
txt_page(57,86,1).
txt_page(190,86,1).
txt_page(276,86,1).
txt_page(947,86,1).
txt_page(342,86,1).
txt_page(142,86,1).
txt_page(446,87,1).
txt_page(250,87,1).
txt_page(554,87,1).
txt_page(697,87,1).
txt_page(149,87,1).
txt_page(767,87,1).
txt_page(279,87,1).
txt_page(232,87,1).
txt_page(243,87,1).
txt_page(476,87,1).
txt_page(128,87,1).
txt_page(142,87,1).
txt_page(444,87,1).
txt_page(878,87,1).
txt_page(512,87,1).
txt_page(171,87,1).
txt_page(303,87,1).
txt_page(937,87,1).
txt_page(625,87,1).
txt_page(150,87,1).
txt_page(188,88,1).
txt_page(524,88,1).
txt_page(388,88,1).
txt_page(893,88,1).
txt_page(803,88,1).
txt_page(858,88,1).
txt_page(621,88,1).
txt_page(565,88,1).
txt_page(3,88,1).
txt_page(551,88,1).
txt_page(790,88,1).
txt_page(921,88,1).
txt_page(161,88,1).
txt_page(478,88,1).
txt_page(449,88,1).
txt_page(292,88,1).
txt_page(277,88,1).
txt_page(446,88,1).
txt_page(169,88,1).
txt_page(582,88,1).
txt_page(643,89,1).
txt_page(892,89,1).
txt_page(896,89,1).
txt_page(189,89,1).
txt_page(347,89,1).
txt_page(270,89,1).
txt_page(512,89,1).
txt_page(898,89,1).
txt_page(708,89,1).
txt_page(298,89,1).
txt_page(589,89,1).
txt_page(278,89,1).
txt_page(332,89,1).
txt_page(625,89,1).
txt_page(811,89,1).
txt_page(751,89,1).
txt_page(508,89,1).
txt_page(401,89,1).
txt_page(72,89,1).
txt_page(391,89,1).
txt_page(499,90,1).
txt_page(22,90,1).
txt_page(436,90,1).
txt_page(307,90,1).
txt_page(830,90,1).
txt_page(175,90,1).
txt_page(27,90,1).
txt_page(702,90,1).
txt_page(738,90,1).
txt_page(867,90,1).
txt_page(384,90,1).
txt_page(851,90,1).
txt_page(725,90,1).
txt_page(112,90,1).
txt_page(810,90,1).
txt_page(34,90,1).
txt_page(647,90,1).
txt_page(405,90,1).
txt_page(635,90,1).
txt_page(454,90,1).
txt_page(191,91,1).
txt_page(200,91,1).
txt_page(874,91,1).
txt_page(439,91,1).
txt_page(482,91,1).
txt_page(851,91,1).
txt_page(363,91,1).
txt_page(69,91,1).
txt_page(334,91,1).
txt_page(87,91,1).
txt_page(83,91,1).
txt_page(955,91,1).
txt_page(959,91,1).
txt_page(889,91,1).
txt_page(853,91,1).
txt_page(415,91,1).
txt_page(201,91,1).
txt_page(498,91,1).
txt_page(725,91,1).
txt_page(255,91,1).
txt_page(461,92,1).
txt_page(713,92,1).
txt_page(429,92,1).
txt_page(699,92,1).
txt_page(495,92,1).
txt_page(758,92,1).
txt_page(26,92,1).
txt_page(422,92,1).
txt_page(262,92,1).
txt_page(692,92,1).
txt_page(922,92,1).
txt_page(289,92,1).
txt_page(1,92,1).
txt_page(705,92,1).
txt_page(581,92,1).
txt_page(297,92,1).
txt_page(681,92,1).
txt_page(693,92,1).
txt_page(83,92,1).
txt_page(345,92,1).
txt_page(757,93,1).
txt_page(911,93,1).
txt_page(594,93,1).
txt_page(424,93,1).
txt_page(396,93,1).
txt_page(69,93,1).
txt_page(44,93,1).
txt_page(259,93,1).
txt_page(853,93,1).
txt_page(439,93,1).
txt_page(328,93,1).
txt_page(130,93,1).
txt_page(682,93,1).
txt_page(675,93,1).
txt_page(680,93,1).
txt_page(62,93,1).
txt_page(204,93,1).
txt_page(762,93,1).
txt_page(855,93,1).
txt_page(917,93,1).
txt_page(966,94,1).
txt_page(56,94,1).
txt_page(761,94,1).
txt_page(440,94,1).
txt_page(573,94,1).
txt_page(298,94,1).
txt_page(500,94,1).
txt_page(525,94,1).
txt_page(122,94,1).
txt_page(674,94,1).
txt_page(768,94,1).
txt_page(594,94,1).
txt_page(253,94,1).
txt_page(117,94,1).
txt_page(156,94,1).
txt_page(581,94,1).
txt_page(475,94,1).
txt_page(882,94,1).
txt_page(470,94,1).
txt_page(899,94,1).
txt_page(572,95,1).
txt_page(305,95,1).
txt_page(238,95,1).
txt_page(654,95,1).
txt_page(620,95,1).
txt_page(673,95,1).
txt_page(102,95,1).
txt_page(768,95,1).
txt_page(337,95,1).
txt_page(240,95,1).
txt_page(330,95,1).
txt_page(791,95,1).
txt_page(655,95,1).
txt_page(474,95,1).
txt_page(968,95,1).
txt_page(539,95,1).
txt_page(868,95,1).
txt_page(640,95,1).
txt_page(960,95,1).
txt_page(84,95,1).
txt_page(489,96,1).
txt_page(392,96,1).
txt_page(876,96,1).
txt_page(698,96,1).
txt_page(187,96,1).
txt_page(867,96,1).
txt_page(685,96,1).
txt_page(43,96,1).
txt_page(456,96,1).
txt_page(532,96,1).
txt_page(3,96,1).
txt_page(510,96,1).
txt_page(295,96,1).
txt_page(119,96,1).
txt_page(333,96,1).
txt_page(652,96,1).
txt_page(371,96,1).
txt_page(770,96,1).
txt_page(735,96,1).
txt_page(128,96,1).
txt_page(405,97,1).
txt_page(519,97,1).
txt_page(966,97,1).
txt_page(187,97,1).
txt_page(238,97,1).
txt_page(786,97,1).
txt_page(669,97,1).
txt_page(686,97,1).
txt_page(31,97,1).
txt_page(673,97,1).
txt_page(106,97,1).
txt_page(680,97,1).
txt_page(301,97,1).
txt_page(491,97,1).
txt_page(455,97,1).
txt_page(864,97,1).
txt_page(58,97,1).
txt_page(737,97,1).
txt_page(644,97,1).
txt_page(586,97,1).
txt_page(188,98,1).
txt_page(42,98,1).
txt_page(167,98,1).
txt_page(198,98,1).
txt_page(86,98,1).
txt_page(195,98,1).
txt_page(286,98,1).
txt_page(344,98,1).
txt_page(861,98,1).
txt_page(456,98,1).
txt_page(67,98,1).
txt_page(244,98,1).
txt_page(904,98,1).
txt_page(618,98,1).
txt_page(528,98,1).
txt_page(100,98,1).
txt_page(447,98,1).
txt_page(607,98,1).
txt_page(89,98,1).
txt_page(714,98,1).
txt_page(886,99,1).
txt_page(895,99,1).
txt_page(804,99,1).
txt_page(473,99,1).
txt_page(106,99,1).
txt_page(347,99,1).
txt_page(959,99,1).
txt_page(564,99,1).
txt_page(6,99,1).
txt_page(180,99,1).
txt_page(380,99,1).
txt_page(912,99,1).
txt_page(45,99,1).
txt_page(329,99,1).
txt_page(215,99,1).
txt_page(717,99,1).
txt_page(104,99,1).
txt_page(521,99,1).
txt_page(84,99,1).
txt_page(528,99,1).
txt_page(364,100,1).
txt_page(661,100,1).
txt_page(825,100,1).
txt_page(587,100,1).
txt_page(217,100,1).
txt_page(389,100,1).
txt_page(952,100,1).
txt_page(707,100,1).
txt_page(662,100,1).
txt_page(310,100,1).
txt_page(71,100,1).
txt_page(279,100,1).
txt_page(254,100,1).
txt_page(859,100,1).
txt_page(540,100,1).
txt_page(786,100,1).
txt_page(615,100,1).
txt_page(655,100,1).
txt_page(445,100,1).
txt_page(361,100,1).
txt_page(445,101,1).
txt_page(726,101,1).
txt_page(620,101,1).
txt_page(219,101,1).
txt_page(414,101,1).
txt_page(751,101,1).
txt_page(239,101,1).
txt_page(247,101,1).
txt_page(372,101,1).
txt_page(238,101,1).
txt_page(456,101,1).
txt_page(665,101,1).
txt_page(261,101,1).
txt_page(908,101,1).
txt_page(467,101,2).
txt_page(484,101,1).
txt_page(155,101,1).
txt_page(144,101,1).
txt_page(451,101,1).
txt_page(234,102,1).
txt_page(727,102,1).
txt_page(252,102,1).
txt_page(476,102,2).
txt_page(69,102,1).
txt_page(605,102,1).
txt_page(122,102,1).
txt_page(66,102,1).
txt_page(756,102,1).
txt_page(141,102,1).
txt_page(309,102,1).
txt_page(258,102,1).
txt_page(300,102,1).
txt_page(210,102,1).
txt_page(339,102,1).
txt_page(803,102,1).
txt_page(157,102,1).
txt_page(929,102,1).
txt_page(458,102,1).
txt_page(913,103,1).
txt_page(603,103,1).
txt_page(221,103,1).
txt_page(720,103,1).
txt_page(104,103,1).
txt_page(158,103,1).
txt_page(725,103,1).
txt_page(273,103,1).
txt_page(659,103,1).
txt_page(230,103,1).
txt_page(36,103,1).
txt_page(244,103,1).
txt_page(855,103,1).
txt_page(310,103,1).
txt_page(693,103,1).
txt_page(91,103,1).
txt_page(10,103,1).
txt_page(809,103,1).
txt_page(207,103,1).
txt_page(212,103,1).
txt_page(18,104,1).
txt_page(871,104,1).
txt_page(591,104,1).
txt_page(276,104,1).
txt_page(822,104,1).
txt_page(297,104,1).
txt_page(48,104,1).
txt_page(441,104,1).
txt_page(280,104,1).
txt_page(290,104,1).
txt_page(362,104,1).
txt_page(197,104,1).
txt_page(505,104,1).
txt_page(367,104,1).
txt_page(470,104,1).
txt_page(952,104,1).
txt_page(141,104,1).
txt_page(880,104,1).
txt_page(372,104,1).
txt_page(19,104,1).
txt_page(490,105,1).
txt_page(802,105,1).
txt_page(418,105,1).
txt_page(684,105,1).
txt_page(785,105,1).
txt_page(357,105,1).
txt_page(358,105,1).
txt_page(395,105,1).
txt_page(393,105,1).
txt_page(820,105,1).
txt_page(186,105,1).
txt_page(351,105,1).
txt_page(701,105,1).
txt_page(38,105,1).
txt_page(450,105,1).
txt_page(713,105,1).
txt_page(851,105,1).
txt_page(336,105,1).
txt_page(966,105,1).
txt_page(754,105,1).
txt_page(608,106,1).
txt_page(202,106,1).
txt_page(407,106,1).
txt_page(243,106,1).
txt_page(828,106,1).
txt_page(384,106,1).
txt_page(682,106,1).
txt_page(599,106,1).
txt_page(742,106,1).
txt_page(15,106,1).
txt_page(656,106,1).
txt_page(531,106,1).
txt_page(391,106,1).
txt_page(739,106,1).
txt_page(314,106,1).
txt_page(519,106,1).
txt_page(764,106,1).
txt_page(333,106,1).
txt_page(847,106,1).
txt_page(463,106,1).
txt_page(20,107,1).
txt_page(494,107,1).
txt_page(370,107,1).
txt_page(843,107,1).
txt_page(914,107,1).
txt_page(898,107,1).
txt_page(355,107,1).
txt_page(500,107,1).
txt_page(943,107,1).
txt_page(901,107,1).
txt_page(61,107,1).
txt_page(797,107,1).
txt_page(49,107,1).
txt_page(755,107,1).
txt_page(487,107,1).
txt_page(538,107,1).
txt_page(519,107,1).
txt_page(636,107,1).
txt_page(704,107,1).
txt_page(323,107,1).
txt_page(557,108,1).
txt_page(590,108,1).
txt_page(732,108,1).
txt_page(5,108,1).
txt_page(614,108,1).
txt_page(626,108,1).
txt_page(623,108,1).
txt_page(705,108,1).
txt_page(819,108,1).
txt_page(269,108,1).
txt_page(220,108,1).
txt_page(66,108,1).
txt_page(421,108,1).
txt_page(877,108,1).
txt_page(728,108,1).
txt_page(552,108,1).
txt_page(70,108,1).
txt_page(497,108,1).
txt_page(731,108,1).
txt_page(84,108,1).
txt_page(195,109,1).
txt_page(795,109,1).
txt_page(50,109,1).
txt_page(362,109,1).
txt_page(27,109,1).
txt_page(567,109,1).
txt_page(839,109,1).
txt_page(237,109,1).
txt_page(732,109,1).
txt_page(620,109,1).
txt_page(665,109,1).
txt_page(462,109,1).
txt_page(883,109,2).
txt_page(556,109,1).
txt_page(944,109,1).
txt_page(194,109,1).
txt_page(850,109,1).
txt_page(761,109,1).
txt_page(37,109,1).
txt_page(45,110,1).
txt_page(647,110,1).
txt_page(897,110,1).
txt_page(103,110,1).
txt_page(319,110,1).
txt_page(713,110,1).
txt_page(37,110,1).
txt_page(663,110,1).
txt_page(913,110,1).
txt_page(526,110,1).
txt_page(794,110,1).
txt_page(726,110,1).
txt_page(840,110,1).
txt_page(88,110,1).
txt_page(756,110,1).
txt_page(672,110,1).
txt_page(752,110,1).
txt_page(853,110,1).
txt_page(594,110,1).
txt_page(523,110,1).
txt_page(730,111,1).
txt_page(525,111,1).
txt_page(531,111,1).
txt_page(369,111,1).
txt_page(261,111,1).
txt_page(917,111,1).
txt_page(860,111,1).
txt_page(32,111,2).
txt_page(43,111,1).
txt_page(425,111,1).
txt_page(243,111,1).
txt_page(828,111,1).
txt_page(46,111,1).
txt_page(745,111,1).
txt_page(690,111,1).
txt_page(247,111,1).
txt_page(743,111,1).
txt_page(273,111,1).
txt_page(602,111,1).
txt_page(834,112,1).
txt_page(676,112,1).
txt_page(210,112,1).
txt_page(324,112,1).
txt_page(158,112,1).
txt_page(93,112,1).
txt_page(292,112,1).
txt_page(577,112,1).
txt_page(379,112,1).
txt_page(641,112,1).
txt_page(291,112,1).
txt_page(890,112,1).
txt_page(79,112,1).
txt_page(478,112,1).
txt_page(496,112,1).
txt_page(675,112,1).
txt_page(694,112,1).
txt_page(560,112,1).
txt_page(548,112,1).
txt_page(129,112,1).
txt_page(909,113,1).
txt_page(201,113,1).
txt_page(208,113,1).
txt_page(712,113,1).
txt_page(557,113,1).
txt_page(488,113,1).
txt_page(251,113,1).
txt_page(67,113,1).
txt_page(766,113,1).
txt_page(274,113,1).
txt_page(781,113,1).
txt_page(172,113,1).
txt_page(313,113,1).
txt_page(769,113,1).
txt_page(12,113,1).
txt_page(834,113,1).
txt_page(41,113,1).
txt_page(176,113,1).
txt_page(547,113,1).
txt_page(583,113,1).
txt_page(217,114,1).
txt_page(742,114,1).
txt_page(752,114,1).
txt_page(795,114,1).
txt_page(270,114,1).
txt_page(352,114,1).
txt_page(170,114,1).
txt_page(573,114,1).
txt_page(538,114,1).
txt_page(557,114,1).
txt_page(490,114,1).
txt_page(374,114,1).
txt_page(794,114,1).
txt_page(533,114,1).
txt_page(556,114,1).
txt_page(758,114,1).
txt_page(364,114,1).
txt_page(615,114,1).
txt_page(115,114,1).
txt_page(618,114,1).
txt_page(579,115,1).
txt_page(33,115,1).
txt_page(106,115,1).
txt_page(627,115,1).
txt_page(924,115,1).
txt_page(541,115,1).
txt_page(359,115,1).
txt_page(653,115,1).
txt_page(934,115,1).
txt_page(305,115,1).
txt_page(83,115,1).
txt_page(940,115,1).
txt_page(875,115,1).
txt_page(24,115,1).
txt_page(583,115,1).
txt_page(598,115,1).
txt_page(363,115,1).
txt_page(554,115,1).
txt_page(537,115,1).
txt_page(914,115,1).
txt_page(953,116,1).
txt_page(206,116,1).
txt_page(694,116,1).
txt_page(658,116,1).
txt_page(649,116,1).
txt_page(889,116,1).
txt_page(548,116,1).
txt_page(217,116,1).
txt_page(382,116,1).
txt_page(435,116,1).
txt_page(411,116,1).
txt_page(231,116,1).
txt_page(837,116,1).
txt_page(193,116,1).
txt_page(178,116,1).
txt_page(333,116,1).
txt_page(147,116,1).
txt_page(449,116,1).
txt_page(469,116,1).
txt_page(536,116,1).
txt_page(888,117,1).
txt_page(123,117,1).
txt_page(274,117,1).
txt_page(317,117,1).
txt_page(749,117,1).
txt_page(251,117,1).
txt_page(622,117,1).
txt_page(400,117,1).
txt_page(923,117,1).
txt_page(826,117,1).
txt_page(501,117,1).
txt_page(462,117,1).
txt_page(82,117,1).
txt_page(873,117,1).
txt_page(6,117,1).
txt_page(657,117,1).
txt_page(799,117,1).
txt_page(705,117,1).
txt_page(822,117,1).
txt_page(816,117,1).
txt_page(414,118,1).
txt_page(679,118,1).
txt_page(824,118,1).
txt_page(435,118,1).
txt_page(632,118,1).
txt_page(820,118,1).
txt_page(741,118,1).
txt_page(33,118,1).
txt_page(903,118,1).
txt_page(127,118,1).
txt_page(224,118,1).
txt_page(284,118,1).
txt_page(236,118,1).
txt_page(194,118,1).
txt_page(584,118,1).
txt_page(631,118,1).
txt_page(774,118,1).
txt_page(519,118,1).
txt_page(297,118,1).
txt_page(798,118,1).
txt_page(858,119,1).
txt_page(634,119,1).
txt_page(412,119,1).
txt_page(801,119,1).
txt_page(697,119,1).
txt_page(143,119,1).
txt_page(10,119,1).
txt_page(665,119,1).
txt_page(2,119,1).
txt_page(741,119,1).
txt_page(515,119,1).
txt_page(61,119,1).
txt_page(677,119,1).
txt_page(175,119,1).
txt_page(513,119,1).
txt_page(855,119,1).
txt_page(612,119,1).
txt_page(387,119,1).
txt_page(9,119,1).
txt_page(752,119,1).
txt_page(475,120,1).
txt_page(777,120,1).
txt_page(79,120,1).
txt_page(246,120,1).
txt_page(488,120,1).
txt_page(477,120,1).
txt_page(662,120,1).
txt_page(773,120,1).
txt_page(797,120,1).
txt_page(582,120,1).
txt_page(629,120,1).
txt_page(826,120,1).
txt_page(678,120,1).
txt_page(770,120,1).
txt_page(443,120,1).
txt_page(274,120,1).
txt_page(366,120,1).
txt_page(718,120,1).
txt_page(633,120,1).
txt_page(845,120,1).
txt_page(570,121,1).
txt_page(342,121,1).
txt_page(289,121,1).
txt_page(205,121,1).
txt_page(414,121,2).
txt_page(17,121,1).
txt_page(766,121,1).
txt_page(693,121,1).
txt_page(584,121,1).
txt_page(80,121,1).
txt_page(655,121,1).
txt_page(925,121,1).
txt_page(364,121,1).
txt_page(712,121,1).
txt_page(380,121,1).
txt_page(807,121,1).
txt_page(374,121,1).
txt_page(77,121,1).
txt_page(56,121,1).
txt_page(621,122,1).
txt_page(11,122,1).
txt_page(813,122,1).
txt_page(347,122,1).
txt_page(94,122,1).
txt_page(844,122,1).
txt_page(110,122,1).
txt_page(652,122,1).
txt_page(460,122,1).
txt_page(743,122,1).
txt_page(575,122,1).
txt_page(212,122,1).
txt_page(697,122,1).
txt_page(390,122,1).
txt_page(339,122,1).
txt_page(352,122,1).
txt_page(178,122,1).
txt_page(389,122,1).
txt_page(855,122,1).
txt_page(509,122,1).
txt_page(80,123,1).
txt_page(487,123,1).
txt_page(187,123,1).
txt_page(44,123,1).
txt_page(841,123,1).
txt_page(490,123,1).
txt_page(2,123,1).
txt_page(247,123,1).
txt_page(334,123,1).
txt_page(833,123,1).
txt_page(28,123,1).
txt_page(241,123,1).
txt_page(407,123,1).
txt_page(446,123,1).
txt_page(468,123,1).
txt_page(606,123,1).
txt_page(109,123,1).
txt_page(895,123,1).
txt_page(372,123,1).
txt_page(899,123,1).
txt_page(600,124,1).
txt_page(897,124,1).
txt_page(152,124,1).
txt_page(675,124,1).
txt_page(589,124,1).
txt_page(819,124,1).
txt_page(613,124,1).
txt_page(706,124,1).
txt_page(59,124,1).
txt_page(570,124,1).
txt_page(394,124,1).
txt_page(330,124,1).
txt_page(98,124,1).
txt_page(887,124,1).
txt_page(522,124,1).
txt_page(461,124,1).
txt_page(832,124,1).
txt_page(269,124,1).
txt_page(405,124,1).
txt_page(438,124,1).
txt_page(30,125,1).
txt_page(294,125,1).
txt_page(757,125,1).
txt_page(815,125,1).
txt_page(906,125,1).
txt_page(764,125,1).
txt_page(310,125,1).
txt_page(520,125,1).
txt_page(409,125,1).
txt_page(824,125,1).
txt_page(101,125,1).
txt_page(766,125,1).
txt_page(317,125,1).
txt_page(653,125,1).
txt_page(256,125,1).
txt_page(156,125,1).
txt_page(637,125,1).
txt_page(909,125,1).
txt_page(851,125,1).
txt_page(624,125,1).
txt_page(598,126,1).
txt_page(40,126,1).
txt_page(861,126,1).
txt_page(52,126,1).
txt_page(620,126,1).
txt_page(889,126,1).
txt_page(664,126,1).
txt_page(314,126,1).
txt_page(24,126,1).
txt_page(909,126,1).
txt_page(99,126,2).
txt_page(674,126,1).
txt_page(429,126,1).
txt_page(939,126,1).
txt_page(609,126,1).
txt_page(148,126,1).
txt_page(187,126,1).
txt_page(754,126,1).
txt_page(808,126,1).
txt_page(285,127,1).
txt_page(77,127,1).
txt_page(948,127,1).
txt_page(187,127,1).
txt_page(678,127,1).
txt_page(658,127,1).
txt_page(811,127,1).
txt_page(449,127,1).
txt_page(809,127,1).
txt_page(922,127,1).
txt_page(907,127,1).
txt_page(108,127,1).
txt_page(618,127,1).
txt_page(847,127,1).
txt_page(74,127,1).
txt_page(37,127,1).
txt_page(278,127,1).
txt_page(798,127,1).
txt_page(837,127,1).
txt_page(95,127,1).
txt_page(507,128,1).
txt_page(196,128,1).
txt_page(333,128,1).
txt_page(723,128,1).
txt_page(394,128,1).
txt_page(884,128,1).
txt_page(514,128,1).
txt_page(403,128,1).
txt_page(616,128,1).
txt_page(115,128,1).
txt_page(866,128,1).
txt_page(950,128,1).
txt_page(658,128,1).
txt_page(958,128,1).
txt_page(761,128,1).
txt_page(763,128,1).
txt_page(537,128,1).
txt_page(95,128,1).
txt_page(202,128,1).
txt_page(291,128,1).
txt_page(796,129,1).
txt_page(716,129,1).
txt_page(884,129,1).
txt_page(888,129,1).
txt_page(652,129,1).
txt_page(672,129,1).
txt_page(190,129,1).
txt_page(77,129,1).
txt_page(912,129,1).
txt_page(176,129,1).
txt_page(682,129,1).
txt_page(256,129,1).
txt_page(850,129,1).
txt_page(597,129,1).
txt_page(429,129,1).
txt_page(234,129,1).
txt_page(929,129,1).
txt_page(920,129,1).
txt_page(662,129,1).
txt_page(252,129,1).
txt_page(278,130,1).
txt_page(339,130,1).
txt_page(85,130,1).
txt_page(122,130,1).
txt_page(924,130,1).
txt_page(577,130,1).
txt_page(245,130,1).
txt_page(149,130,1).
txt_page(497,130,1).
txt_page(153,130,1).
txt_page(966,130,1).
txt_page(260,130,1).
txt_page(412,130,1).
txt_page(654,130,1).
txt_page(586,130,1).
txt_page(312,130,1).
txt_page(378,130,1).
txt_page(448,130,1).
txt_page(110,130,1).
txt_page(552,130,1).
txt_page(313,131,1).
txt_page(251,131,1).
txt_page(855,131,1).
txt_page(932,131,1).
txt_page(456,131,1).
txt_page(481,131,1).
txt_page(266,131,1).
txt_page(840,131,1).
txt_page(684,131,1).
txt_page(440,131,1).
txt_page(713,131,1).
txt_page(887,131,1).
txt_page(766,131,1).
txt_page(307,131,1).
txt_page(635,131,1).
txt_page(776,131,1).
txt_page(947,131,1).
txt_page(111,131,2).
txt_page(133,131,1).
txt_page(534,132,1).
txt_page(550,132,1).
txt_page(964,132,1).
txt_page(633,132,1).
txt_page(740,132,1).
txt_page(471,132,1).
txt_page(721,132,1).
txt_page(34,132,1).
txt_page(965,132,1).
txt_page(312,132,1).
txt_page(87,132,1).
txt_page(535,132,1).
txt_page(835,132,1).
txt_page(528,132,1).
txt_page(200,132,1).
txt_page(120,132,1).
txt_page(743,132,1).
txt_page(900,132,1).
txt_page(95,132,1).
txt_page(877,132,1).
txt_page(358,133,1).
txt_page(689,133,1).
txt_page(241,133,1).
txt_page(719,133,1).
txt_page(148,133,1).
txt_page(420,133,1).
txt_page(77,133,1).
txt_page(832,133,1).
txt_page(718,133,1).
txt_page(933,133,1).
txt_page(809,133,1).
txt_page(850,133,1).
txt_page(481,133,1).
txt_page(231,133,1).
txt_page(513,133,1).
txt_page(429,133,1).
txt_page(706,133,1).
txt_page(437,133,1).
txt_page(519,133,1).
txt_page(938,133,1).
txt_page(239,134,1).
txt_page(811,134,1).
txt_page(643,134,1).
txt_page(108,134,1).
txt_page(469,134,1).
txt_page(726,134,1).
txt_page(594,134,1).
txt_page(490,134,1).
txt_page(703,134,1).
txt_page(330,134,1).
txt_page(102,134,1).
txt_page(939,134,1).
txt_page(929,134,1).
txt_page(914,134,1).
txt_page(956,134,1).
txt_page(326,134,1).
txt_page(169,134,1).
txt_page(24,134,1).
txt_page(761,134,1).
txt_page(166,134,1).
txt_page(778,135,1).
txt_page(558,135,1).
txt_page(506,135,1).
txt_page(952,135,1).
txt_page(965,135,1).
txt_page(796,135,1).
txt_page(785,135,1).
txt_page(47,135,1).
txt_page(195,135,1).
txt_page(467,135,1).
txt_page(960,135,1).
txt_page(279,135,1).
txt_page(518,135,1).
txt_page(193,135,1).
txt_page(597,135,1).
txt_page(773,135,1).
txt_page(784,135,1).
txt_page(589,135,1).
txt_page(135,135,2).
txt_page(900,136,1).
txt_page(121,136,1).
txt_page(196,136,1).
txt_page(250,136,1).
txt_page(899,136,1).
txt_page(43,136,1).
txt_page(266,136,1).
txt_page(108,136,1).
txt_page(757,136,1).
txt_page(883,136,1).
txt_page(967,136,1).
txt_page(515,136,1).
txt_page(815,136,1).
txt_page(697,136,1).
txt_page(487,136,1).
txt_page(348,136,1).
txt_page(413,136,1).
txt_page(546,136,1).
txt_page(759,136,1).
txt_page(257,136,1).
txt_page(530,137,1).
txt_page(567,137,1).
txt_page(559,137,1).
txt_page(227,137,1).
txt_page(549,137,1).
txt_page(833,137,1).
txt_page(188,137,1).
txt_page(297,137,1).
txt_page(165,137,1).
txt_page(380,137,1).
txt_page(598,137,1).
txt_page(655,137,1).
txt_page(890,137,1).
txt_page(824,137,1).
txt_page(519,137,1).
txt_page(555,137,1).
txt_page(422,137,1).
txt_page(145,137,1).
txt_page(129,137,1).
txt_page(587,137,1).
txt_page(288,138,1).
txt_page(579,138,1).
txt_page(23,138,1).
txt_page(321,138,1).
txt_page(968,138,1).
txt_page(616,138,1).
txt_page(80,138,1).
txt_page(206,138,1).
txt_page(266,138,1).
txt_page(945,138,1).
txt_page(919,138,1).
txt_page(918,138,1).
txt_page(858,138,1).
txt_page(197,138,1).
txt_page(633,138,1).
txt_page(237,138,1).
txt_page(110,138,1).
txt_page(106,138,1).
txt_page(90,138,1).
txt_page(701,138,1).
txt_page(189,139,1).
txt_page(403,139,1).
txt_page(744,139,1).
txt_page(152,139,1).
txt_page(224,139,1).
txt_page(511,139,1).
txt_page(244,139,1).
txt_page(762,139,1).
txt_page(571,139,1).
txt_page(838,139,1).
txt_page(757,139,1).
txt_page(239,139,1).
txt_page(773,139,1).
txt_page(457,139,1).
txt_page(367,139,1).
txt_page(440,139,1).
txt_page(788,139,1).
txt_page(581,139,1).
txt_page(943,139,1).
txt_page(682,139,1).
txt_page(448,140,1).
txt_page(799,140,1).
txt_page(663,140,1).
txt_page(879,140,1).
txt_page(599,140,1).
txt_page(901,140,1).
txt_page(209,140,1).
txt_page(821,140,1).
txt_page(335,140,1).
txt_page(92,140,1).
txt_page(20,140,1).
txt_page(487,140,1).
txt_page(60,140,1).
txt_page(55,140,1).
txt_page(582,140,1).
txt_page(661,140,1).
txt_page(626,140,1).
txt_page(422,140,1).
txt_page(967,140,1).
txt_page(529,140,1).
txt_page(57,141,1).
txt_page(828,141,1).
txt_page(934,141,1).
txt_page(318,141,1).
txt_page(174,141,1).
txt_page(229,141,1).
txt_page(593,141,1).
txt_page(900,141,1).
txt_page(452,141,1).
txt_page(367,141,1).
txt_page(885,141,1).
txt_page(741,141,1).
txt_page(139,141,1).
txt_page(907,141,1).
txt_page(534,141,1).
txt_page(519,141,1).
txt_page(710,141,1).
txt_page(108,141,1).
txt_page(425,141,1).
txt_page(823,141,1).
txt_page(249,142,1).
txt_page(761,142,1).
txt_page(23,142,1).
txt_page(961,142,1).
txt_page(309,142,1).
txt_page(754,142,1).
txt_page(195,142,1).
txt_page(486,142,1).
txt_page(144,142,1).
txt_page(25,142,1).
txt_page(639,142,1).
txt_page(911,142,1).
txt_page(669,142,1).
txt_page(550,142,1).
txt_page(324,142,1).
txt_page(511,142,1).
txt_page(536,142,1).
txt_page(514,142,1).
txt_page(965,142,1).
txt_page(647,142,1).
txt_page(660,143,1).
txt_page(910,143,1).
txt_page(909,143,1).
txt_page(557,143,1).
txt_page(427,143,1).
txt_page(301,143,1).
txt_page(87,143,1).
txt_page(624,143,1).
txt_page(364,143,1).
txt_page(791,143,1).
txt_page(504,143,1).
txt_page(296,143,1).
txt_page(638,143,1).
txt_page(363,143,1).
txt_page(69,143,1).
txt_page(233,143,1).
txt_page(507,143,1).
txt_page(823,143,1).
txt_page(252,143,1).
txt_page(1,143,1).
txt_page(75,144,1).
txt_page(378,144,1).
txt_page(810,144,1).
txt_page(622,144,1).
txt_page(643,144,1).
txt_page(573,144,1).
txt_page(369,144,1).
txt_page(442,144,1).
txt_page(462,144,1).
txt_page(233,144,1).
txt_page(557,144,1).
txt_page(776,144,1).
txt_page(739,144,1).
txt_page(886,144,1).
txt_page(823,144,1).
txt_page(190,144,1).
txt_page(634,144,1).
txt_page(64,144,1).
txt_page(578,144,1).
txt_page(38,144,1).
txt_page(185,145,1).
txt_page(679,145,1).
txt_page(315,145,1).
txt_page(745,145,1).
txt_page(259,145,1).
txt_page(599,145,1).
txt_page(700,145,1).
txt_page(790,145,1).
txt_page(285,145,1).
txt_page(848,145,1).
txt_page(541,145,1).
txt_page(108,145,1).
txt_page(3,145,1).
txt_page(765,145,1).
txt_page(538,145,1).
txt_page(24,145,1).
txt_page(623,145,1).
txt_page(130,145,1).
txt_page(404,145,1).
txt_page(718,145,1).
txt_page(498,146,1).
txt_page(965,146,1).
txt_page(78,146,1).
txt_page(117,146,1).
txt_page(270,146,1).
txt_page(310,146,1).
txt_page(337,146,1).
txt_page(846,146,1).
txt_page(406,146,1).
txt_page(666,146,1).
txt_page(393,146,1).
txt_page(294,146,1).
txt_page(907,146,1).
txt_page(227,146,1).
txt_page(323,146,1).
txt_page(330,146,1).
txt_page(354,146,1).
txt_page(608,146,1).
txt_page(885,146,1).
txt_page(953,146,1).
txt_page(249,147,1).
txt_page(801,147,1).
txt_page(368,147,1).
txt_page(603,147,1).
txt_page(777,147,1).
txt_page(733,147,1).
txt_page(705,147,1).
txt_page(730,147,1).
txt_page(454,147,1).
txt_page(130,147,1).
txt_page(107,147,1).
txt_page(146,147,1).
txt_page(607,147,1).
txt_page(362,147,1).
txt_page(202,147,1).
txt_page(110,147,1).
txt_page(863,147,1).
txt_page(23,147,1).
txt_page(333,147,1).
txt_page(475,147,1).
txt_page(310,148,1).
txt_page(96,148,1).
txt_page(134,148,1).
txt_page(304,148,1).
txt_page(832,148,1).
txt_page(789,148,1).
txt_page(370,148,1).
txt_page(951,148,1).
txt_page(532,148,1).
txt_page(572,148,1).
txt_page(340,148,1).
txt_page(705,148,1).
txt_page(366,148,1).
txt_page(80,148,1).
txt_page(336,148,1).
txt_page(878,148,1).
txt_page(425,148,1).
txt_page(250,148,1).
txt_page(809,148,1).
txt_page(332,148,1).
txt_page(832,149,1).
txt_page(487,149,1).
txt_page(66,149,1).
txt_page(234,149,1).
txt_page(872,149,1).
txt_page(265,149,1).
txt_page(742,149,1).
txt_page(88,149,1).
txt_page(805,149,1).
txt_page(457,149,1).
txt_page(918,149,1).
txt_page(662,149,1).
txt_page(454,149,1).
txt_page(46,149,1).
txt_page(442,149,1).
txt_page(941,149,1).
txt_page(341,149,1).
txt_page(255,149,1).
txt_page(693,149,1).
txt_page(620,149,1).
txt_page(914,150,1).
txt_page(22,150,1).
txt_page(110,150,1).
txt_page(197,150,1).
txt_page(927,150,1).
txt_page(177,150,1).
txt_page(537,150,1).
txt_page(922,150,1).
txt_page(412,150,1).
txt_page(353,150,1).
txt_page(184,150,1).
txt_page(612,150,1).
txt_page(961,150,1).
txt_page(937,150,1).
txt_page(56,150,2).
txt_page(101,150,1).
txt_page(966,150,1).
txt_page(163,150,1).
txt_page(892,150,1).
txt_page(500,151,1).
txt_page(166,151,1).
txt_page(613,151,1).
txt_page(121,151,1).
txt_page(515,151,1).
txt_page(296,151,1).
txt_page(770,151,1).
txt_page(646,151,1).
txt_page(355,151,1).
txt_page(185,151,1).
txt_page(807,151,1).
txt_page(608,151,1).
txt_page(177,151,1).
txt_page(465,151,1).
txt_page(315,151,1).
txt_page(493,151,1).
txt_page(469,151,1).
txt_page(252,151,1).
txt_page(875,151,1).
txt_page(405,151,1).
txt_page(151,152,1).
txt_page(347,152,1).
txt_page(454,152,1).
txt_page(566,152,1).
txt_page(915,152,1).
txt_page(729,152,1).
txt_page(340,152,1).
txt_page(557,152,1).
txt_page(505,152,1).
txt_page(784,152,1).
txt_page(830,152,1).
txt_page(783,152,1).
txt_page(407,152,1).
txt_page(68,152,1).
txt_page(442,152,1).
txt_page(698,152,1).
txt_page(187,152,2).
txt_page(573,152,1).
txt_page(921,152,1).
txt_page(104,153,1).
txt_page(729,153,1).
txt_page(616,153,1).
txt_page(792,153,1).
txt_page(253,153,1).
txt_page(871,153,1).
txt_page(288,153,1).
txt_page(145,153,1).
txt_page(677,153,1).
txt_page(320,153,1).
txt_page(900,153,1).
txt_page(553,153,1).
txt_page(525,153,1).
txt_page(468,153,1).
txt_page(367,153,1).
txt_page(179,153,1).
txt_page(59,153,1).
txt_page(153,153,1).
txt_page(346,153,1).
txt_page(817,153,1).
txt_page(911,154,1).
txt_page(436,154,1).
txt_page(773,154,1).
txt_page(930,154,1).
txt_page(416,154,1).
txt_page(768,154,1).
txt_page(144,154,1).
txt_page(375,154,1).
txt_page(227,154,1).
txt_page(879,154,1).
txt_page(477,154,1).
txt_page(913,154,1).
txt_page(472,154,1).
txt_page(264,154,1).
txt_page(366,154,1).
txt_page(96,154,1).
txt_page(728,154,1).
txt_page(294,154,1).
txt_page(457,154,1).
txt_page(530,154,1).
txt_page(939,155,1).
txt_page(92,155,1).
txt_page(644,155,1).
txt_page(862,155,1).
txt_page(957,155,1).
txt_page(43,155,1).
txt_page(456,155,1).
txt_page(124,155,1).
txt_page(160,155,1).
txt_page(485,155,1).
txt_page(24,155,1).
txt_page(172,155,1).
txt_page(163,155,1).
txt_page(526,155,1).
txt_page(605,155,2).
txt_page(5,155,1).
txt_page(860,155,1).
txt_page(640,155,1).
txt_page(437,155,1).
txt_page(639,156,1).
txt_page(176,156,1).
txt_page(55,156,1).
txt_page(927,156,1).
txt_page(835,156,1).
txt_page(245,156,1).
txt_page(886,156,1).
txt_page(612,156,1).
txt_page(858,156,1).
txt_page(124,156,1).
txt_page(812,156,1).
txt_page(242,156,1).
txt_page(848,156,1).
txt_page(602,156,1).
txt_page(742,156,1).
txt_page(209,156,1).
txt_page(85,156,1).
txt_page(793,156,1).
txt_page(502,156,1).
txt_page(648,156,1).
txt_page(461,157,1).
txt_page(92,157,1).
txt_page(26,157,1).
txt_page(946,157,1).
txt_page(469,157,1).
txt_page(663,157,1).
txt_page(696,157,1).
txt_page(675,157,1).
txt_page(257,157,1).
txt_page(787,157,1).
txt_page(171,157,1).
txt_page(900,157,1).
txt_page(809,157,1).
txt_page(721,157,1).
txt_page(432,157,1).
txt_page(821,157,1).
txt_page(503,157,1).
txt_page(254,157,1).
txt_page(733,157,1).
txt_page(374,157,1).
txt_page(403,158,1).
txt_page(15,158,1).
txt_page(407,158,1).
txt_page(436,158,1).
txt_page(637,158,1).
txt_page(658,158,1).
txt_page(418,158,1).
txt_page(269,158,1).
txt_page(628,158,1).
txt_page(719,158,1).
txt_page(558,158,1).
txt_page(625,158,1).
txt_page(915,158,1).
txt_page(916,158,1).
txt_page(275,158,1).
txt_page(728,158,1).
txt_page(497,158,1).
txt_page(642,158,1).
txt_page(79,158,1).
txt_page(673,158,1).
txt_page(858,159,1).
txt_page(912,159,1).
txt_page(556,159,1).
txt_page(410,159,1).
txt_page(644,159,1).
txt_page(109,159,1).
txt_page(581,159,1).
txt_page(563,159,1).
txt_page(439,159,1).
txt_page(192,159,1).
txt_page(404,159,1).
txt_page(443,159,1).
txt_page(896,159,1).
txt_page(134,159,1).
txt_page(934,159,1).
txt_page(907,159,1).
txt_page(307,159,1).
txt_page(778,159,1).
txt_page(795,159,1).
txt_page(49,159,1).
txt_page(440,160,1).
txt_page(946,160,1).
txt_page(73,160,1).
txt_page(274,160,1).
txt_page(693,160,1).
txt_page(949,160,1).
txt_page(74,160,1).
txt_page(736,160,1).
txt_page(397,160,1).
txt_page(146,160,1).
txt_page(634,160,1).
txt_page(604,160,1).
txt_page(454,160,1).
txt_page(402,160,1).
txt_page(745,160,1).
txt_page(575,160,1).
txt_page(374,160,1).
txt_page(179,160,1).
txt_page(393,160,1).
txt_page(512,160,1).
txt_page(871,161,1).
txt_page(417,161,1).
txt_page(551,161,1).
txt_page(656,161,1).
txt_page(232,161,1).
txt_page(436,161,1).
txt_page(237,161,1).
txt_page(874,161,1).
txt_page(112,161,1).
txt_page(345,161,1).
txt_page(481,161,1).
txt_page(847,161,1).
txt_page(180,161,1).
txt_page(861,161,1).
txt_page(535,161,1).
txt_page(789,161,1).
txt_page(172,161,1).
txt_page(960,161,1).
txt_page(622,161,1).
txt_page(694,161,1).
txt_page(190,162,1).
txt_page(474,162,1).
txt_page(755,162,1).
txt_page(862,162,1).
txt_page(173,162,1).
txt_page(867,162,1).
txt_page(100,162,1).
txt_page(544,162,1).
txt_page(781,162,1).
txt_page(161,162,1).
txt_page(334,162,1).
txt_page(954,162,1).
txt_page(137,162,1).
txt_page(321,162,1).
txt_page(478,162,1).
txt_page(3,162,1).
txt_page(393,162,1).
txt_page(378,162,1).
txt_page(465,162,1).
txt_page(85,162,1).
txt_page(928,163,1).
txt_page(102,163,1).
txt_page(538,163,1).
txt_page(153,163,1).
txt_page(939,163,1).
txt_page(250,163,1).
txt_page(543,163,1).
txt_page(627,163,1).
txt_page(290,163,1).
txt_page(198,163,1).
txt_page(202,163,1).
txt_page(591,163,1).
txt_page(955,163,1).
txt_page(900,163,1).
txt_page(504,163,1).
txt_page(787,163,1).
txt_page(492,163,1).
txt_page(75,163,1).
txt_page(881,163,1).
txt_page(450,163,1).
txt_page(957,164,1).
txt_page(287,164,1).
txt_page(140,164,1).
txt_page(392,164,1).
txt_page(81,164,1).
txt_page(646,164,1).
txt_page(329,164,1).
txt_page(188,164,1).
txt_page(738,164,1).
txt_page(505,164,1).
txt_page(352,164,1).
txt_page(757,164,1).
txt_page(180,164,1).
txt_page(771,164,1).
txt_page(215,164,1).
txt_page(669,164,1).
txt_page(714,164,1).
txt_page(582,164,1).
txt_page(232,164,1).
txt_page(136,164,1).
txt_page(33,165,1).
txt_page(297,165,1).
txt_page(551,165,1).
txt_page(200,165,1).
txt_page(511,165,1).
txt_page(409,165,1).
txt_page(914,165,1).
txt_page(221,165,1).
txt_page(686,165,1).
txt_page(440,165,1).
txt_page(743,165,1).
txt_page(655,165,1).
txt_page(863,165,1).
txt_page(871,165,1).
txt_page(870,165,1).
txt_page(837,165,1).
txt_page(99,165,1).
txt_page(378,165,1).
txt_page(652,165,1).
txt_page(270,165,1).
txt_page(569,166,1).
txt_page(595,166,1).
txt_page(198,166,1).
txt_page(162,166,1).
txt_page(237,166,1).
txt_page(223,166,1).
txt_page(354,166,1).
txt_page(865,166,1).
txt_page(752,166,1).
txt_page(70,166,1).
txt_page(613,166,1).
txt_page(847,166,1).
txt_page(226,166,1).
txt_page(103,166,1).
txt_page(230,166,1).
txt_page(575,166,1).
txt_page(566,166,1).
txt_page(113,166,1).
txt_page(730,166,1).
txt_page(682,166,1).
txt_page(332,167,1).
txt_page(809,167,1).
txt_page(885,167,1).
txt_page(182,167,1).
txt_page(458,167,1).
txt_page(248,167,1).
txt_page(190,167,1).
txt_page(238,167,1).
txt_page(638,167,1).
txt_page(282,167,1).
txt_page(811,167,1).
txt_page(285,167,1).
txt_page(803,167,1).
txt_page(648,167,1).
txt_page(663,167,1).
txt_page(495,167,1).
txt_page(740,167,1).
txt_page(336,167,1).
txt_page(440,167,1).
txt_page(437,167,1).
txt_page(4,168,1).
txt_page(412,168,1).
txt_page(227,168,1).
txt_page(852,168,1).
txt_page(567,168,1).
txt_page(306,168,1).
txt_page(616,168,1).
txt_page(254,168,1).
txt_page(621,168,1).
txt_page(35,168,1).
txt_page(766,168,1).
txt_page(813,168,1).
txt_page(482,168,1).
txt_page(115,168,1).
txt_page(339,168,1).
txt_page(622,168,1).
txt_page(727,168,1).
txt_page(754,168,1).
txt_page(696,168,1).
txt_page(624,168,1).
txt_page(177,169,1).
txt_page(37,169,1).
txt_page(588,169,1).
txt_page(798,169,1).
txt_page(405,169,1).
txt_page(370,169,1).
txt_page(709,169,1).
txt_page(689,169,1).
txt_page(510,169,1).
txt_page(786,169,1).
txt_page(409,169,1).
txt_page(854,169,1).
txt_page(424,169,1).
txt_page(524,169,1).
txt_page(268,169,1).
txt_page(330,169,1).
txt_page(71,169,1).
txt_page(941,169,1).
txt_page(815,169,1).
txt_page(203,169,1).
txt_page(432,170,1).
txt_page(442,170,1).
txt_page(869,170,1).
txt_page(160,170,1).
txt_page(133,170,1).
txt_page(790,170,1).
txt_page(101,170,1).
txt_page(520,170,1).
txt_page(303,170,1).
txt_page(19,170,1).
txt_page(817,170,1).
txt_page(484,170,1).
txt_page(703,170,1).
txt_page(98,170,1).
txt_page(624,170,1).
txt_page(787,170,1).
txt_page(642,170,1).
txt_page(534,170,1).
txt_page(967,170,1).
txt_page(655,170,1).
txt_page(799,171,1).
txt_page(589,171,1).
txt_page(936,171,1).
txt_page(480,171,1).
txt_page(367,171,1).
txt_page(21,171,1).
txt_page(253,171,1).
txt_page(270,171,1).
txt_page(435,171,1).
txt_page(947,171,1).
txt_page(366,171,1).
txt_page(575,171,1).
txt_page(858,171,1).
txt_page(451,171,1).
txt_page(628,171,1).
txt_page(141,171,1).
txt_page(95,171,1).
txt_page(806,171,1).
txt_page(180,171,1).
txt_page(78,171,1).
txt_page(768,172,1).
txt_page(944,172,1).
txt_page(702,172,1).
txt_page(912,172,1).
txt_page(862,172,1).
txt_page(39,172,1).
txt_page(251,172,1).
txt_page(164,172,1).
txt_page(220,172,1).
txt_page(898,172,1).
txt_page(446,172,1).
txt_page(66,172,1).
txt_page(700,172,1).
txt_page(125,172,1).
txt_page(753,172,1).
txt_page(266,172,1).
txt_page(772,172,1).
txt_page(340,172,1).
txt_page(216,172,1).
txt_page(675,172,1).
txt_page(226,173,1).
txt_page(268,173,1).
txt_page(549,173,1).
txt_page(403,173,1).
txt_page(668,173,1).
txt_page(218,173,1).
txt_page(420,173,1).
txt_page(752,173,1).
txt_page(223,173,1).
txt_page(4,173,1).
txt_page(686,173,1).
txt_page(211,173,1).
txt_page(771,173,1).
txt_page(224,173,1).
txt_page(352,173,1).
txt_page(846,173,1).
txt_page(498,173,1).
txt_page(664,173,1).
txt_page(937,173,1).
txt_page(635,173,1).
txt_page(182,174,1).
txt_page(965,174,1).
txt_page(607,174,1).
txt_page(854,174,1).
txt_page(681,174,1).
txt_page(268,174,1).
txt_page(173,174,1).
txt_page(154,174,1).
txt_page(418,174,1).
txt_page(125,174,1).
txt_page(223,174,1).
txt_page(690,174,1).
txt_page(703,174,1).
txt_page(151,174,1).
txt_page(890,174,1).
txt_page(710,174,1).
txt_page(452,174,1).
txt_page(400,174,1).
txt_page(407,174,1).
txt_page(271,174,1).
txt_page(179,175,1).
txt_page(853,175,1).
txt_page(668,175,1).
txt_page(293,175,1).
txt_page(764,175,1).
txt_page(389,175,1).
txt_page(163,175,1).
txt_page(255,175,1).
txt_page(623,175,1).
txt_page(673,175,1).
txt_page(339,175,1).
txt_page(540,175,1).
txt_page(656,175,1).
txt_page(213,175,1).
txt_page(832,175,1).
txt_page(367,175,1).
txt_page(211,175,1).
txt_page(64,175,1).
txt_page(688,175,1).
txt_page(237,175,1).
txt_page(81,176,1).
txt_page(829,176,1).
txt_page(612,176,1).
txt_page(138,176,1).
txt_page(529,176,1).
txt_page(170,176,1).
txt_page(83,176,1).
txt_page(889,176,1).
txt_page(559,176,1).
txt_page(355,176,1).
txt_page(836,176,1).
txt_page(380,176,1).
txt_page(335,176,1).
txt_page(264,176,1).
txt_page(29,176,1).
txt_page(911,176,1).
txt_page(427,176,1).
txt_page(384,176,1).
txt_page(491,176,1).
txt_page(159,176,1).
txt_page(249,177,1).
txt_page(949,177,1).
txt_page(689,177,1).
txt_page(627,177,1).
txt_page(588,177,1).
txt_page(190,177,1).
txt_page(304,177,1).
txt_page(836,177,1).
txt_page(602,177,1).
txt_page(221,177,1).
txt_page(124,177,1).
txt_page(380,177,1).
txt_page(617,177,1).
txt_page(777,177,1).
txt_page(516,177,1).
txt_page(78,177,1).
txt_page(176,177,1).
txt_page(226,177,1).
txt_page(826,177,1).
txt_page(429,177,1).
txt_page(322,178,1).
txt_page(395,178,1).
txt_page(283,178,1).
txt_page(737,178,1).
txt_page(389,178,1).
txt_page(40,178,1).
txt_page(150,178,1).
txt_page(478,178,1).
txt_page(893,178,1).
txt_page(4,178,1).
txt_page(716,178,1).
txt_page(929,178,1).
txt_page(262,178,1).
txt_page(505,178,1).
txt_page(45,178,1).
txt_page(890,178,1).
txt_page(940,178,1).
txt_page(274,178,1).
txt_page(956,178,1).
txt_page(787,178,1).
txt_page(14,179,1).
txt_page(316,179,2).
txt_page(718,179,1).
txt_page(258,179,1).
txt_page(416,179,1).
txt_page(17,179,1).
txt_page(904,179,1).
txt_page(68,179,1).
txt_page(242,179,1).
txt_page(675,179,1).
txt_page(678,179,1).
txt_page(26,179,1).
txt_page(845,179,1).
txt_page(45,179,1).
txt_page(898,179,1).
txt_page(64,179,1).
txt_page(633,179,1).
txt_page(826,179,1).
txt_page(708,179,1).
txt_page(646,180,1).
txt_page(279,180,1).
txt_page(239,180,1).
txt_page(319,180,1).
txt_page(300,180,1).
txt_page(62,180,1).
txt_page(770,180,1).
txt_page(605,180,1).
txt_page(224,180,1).
txt_page(144,180,1).
txt_page(366,180,1).
txt_page(821,180,1).
txt_page(451,180,1).
txt_page(832,180,1).
txt_page(675,180,1).
txt_page(622,180,1).
txt_page(351,180,1).
txt_page(857,180,1).
txt_page(581,180,1).
txt_page(427,180,1).
txt_page(662,181,1).
txt_page(956,181,1).
txt_page(364,181,1).
txt_page(358,181,1).
txt_page(300,181,1).
txt_page(305,181,1).
txt_page(667,181,1).
txt_page(401,181,1).
txt_page(430,181,1).
txt_page(952,181,1).
txt_page(750,181,1).
txt_page(152,181,1).
txt_page(78,181,1).
txt_page(293,181,1).
txt_page(909,181,1).
txt_page(421,181,1).
txt_page(727,181,1).
txt_page(176,181,1).
txt_page(880,181,1).
txt_page(80,181,1).
txt_page(283,182,1).
txt_page(744,182,1).
txt_page(258,182,1).
txt_page(56,182,1).
txt_page(136,182,1).
txt_page(864,182,1).
txt_page(391,182,1).
txt_page(475,182,1).
txt_page(643,182,1).
txt_page(63,182,1).
txt_page(411,182,1).
txt_page(151,182,1).
txt_page(146,182,1).
txt_page(483,182,1).
txt_page(84,182,1).
txt_page(778,182,1).
txt_page(362,182,1).
txt_page(406,182,1).
txt_page(751,182,1).
txt_page(397,182,1).
txt_page(521,183,1).
txt_page(839,183,1).
txt_page(106,183,1).
txt_page(474,183,1).
txt_page(599,183,1).
txt_page(738,183,1).
txt_page(803,183,1).
txt_page(512,183,1).
txt_page(220,183,1).
txt_page(692,183,2).
txt_page(755,183,1).
txt_page(529,183,1).
txt_page(351,183,1).
txt_page(97,183,1).
txt_page(159,183,1).
txt_page(791,183,1).
txt_page(648,183,1).
txt_page(85,183,1).
txt_page(946,183,1).
txt_page(89,184,1).
txt_page(713,184,1).
txt_page(675,184,1).
txt_page(881,184,1).
txt_page(854,184,1).
txt_page(678,184,1).
txt_page(642,184,1).
txt_page(929,184,1).
txt_page(479,184,1).
txt_page(90,184,1).
txt_page(469,184,1).
txt_page(529,184,1).
txt_page(392,184,1).
txt_page(887,184,1).
txt_page(911,184,1).
txt_page(773,184,1).
txt_page(384,184,1).
txt_page(173,184,1).
txt_page(663,184,1).
txt_page(576,184,1).
txt_page(212,185,1).
txt_page(643,185,1).
txt_page(76,185,1).
txt_page(244,185,1).
txt_page(495,185,1).
txt_page(750,185,1).
txt_page(485,185,1).
txt_page(45,185,1).
txt_page(291,185,1).
txt_page(719,185,1).
txt_page(480,185,1).
txt_page(237,185,1).
txt_page(772,185,1).
txt_page(225,185,1).
txt_page(828,185,1).
txt_page(716,185,1).
txt_page(706,185,1).
txt_page(563,185,1).
txt_page(318,185,1).
txt_page(350,185,1).
txt_page(373,186,1).
txt_page(244,186,1).
txt_page(598,186,1).
txt_page(271,186,1).
txt_page(434,186,1).
txt_page(83,186,1).
txt_page(758,186,1).
txt_page(784,186,1).
txt_page(787,186,1).
txt_page(921,186,1).
txt_page(658,186,1).
txt_page(188,186,1).
txt_page(859,186,1).
txt_page(260,186,1).
txt_page(356,186,1).
txt_page(249,186,1).
txt_page(525,186,1).
txt_page(771,186,1).
txt_page(364,186,1).
txt_page(98,186,1).
txt_page(222,187,1).
txt_page(97,187,1).
txt_page(522,187,1).
txt_page(445,187,1).
txt_page(883,187,1).
txt_page(205,187,1).
txt_page(950,187,1).
txt_page(773,187,1).
txt_page(146,187,1).
txt_page(516,187,1).
txt_page(483,187,1).
txt_page(586,187,1).
txt_page(862,187,1).
txt_page(778,187,1).
txt_page(442,187,1).
txt_page(354,187,2).
txt_page(474,187,1).
txt_page(751,187,1).
txt_page(837,187,1).
txt_page(49,188,1).
txt_page(890,188,1).
txt_page(178,188,1).
txt_page(536,188,1).
txt_page(200,188,1).
txt_page(694,188,1).
txt_page(699,188,1).
txt_page(701,188,1).
txt_page(129,188,1).
txt_page(892,188,1).
txt_page(145,188,1).
txt_page(923,188,1).
txt_page(755,188,1).
txt_page(778,188,1).
txt_page(266,188,1).
txt_page(276,188,1).
txt_page(486,188,1).
txt_page(949,188,1).
txt_page(916,188,1).
txt_page(702,188,1).
txt_page(877,189,1).
txt_page(303,189,1).
txt_page(134,189,1).
txt_page(854,189,1).
txt_page(502,189,1).
txt_page(305,189,1).
txt_page(959,189,1).
txt_page(905,189,1).
txt_page(224,189,1).
txt_page(942,189,1).
txt_page(74,189,1).
txt_page(40,189,1).
txt_page(579,189,1).
txt_page(839,189,1).
txt_page(544,189,1).
txt_page(834,189,1).
txt_page(338,189,1).
txt_page(51,189,1).
txt_page(111,189,1).
txt_page(965,189,1).
txt_page(281,190,1).
txt_page(57,190,1).
txt_page(586,190,1).
txt_page(20,190,1).
txt_page(733,190,1).
txt_page(602,190,1).
txt_page(299,190,1).
txt_page(750,190,1).
txt_page(233,190,1).
txt_page(859,190,1).
txt_page(774,190,1).
txt_page(4,190,1).
txt_page(742,190,1).
txt_page(135,190,1).
txt_page(155,190,1).
txt_page(59,190,1).
txt_page(283,190,1).
txt_page(845,190,1).
txt_page(967,190,1).
txt_page(639,190,1).
txt_page(225,191,1).
txt_page(913,191,1).
txt_page(885,191,1).
txt_page(905,191,1).
txt_page(294,191,1).
txt_page(485,191,1).
txt_page(40,191,1).
txt_page(571,191,1).
txt_page(137,191,1).
txt_page(509,191,1).
txt_page(702,191,1).
txt_page(829,191,1).
txt_page(484,191,1).
txt_page(647,191,1).
txt_page(816,191,1).
txt_page(406,191,1).
txt_page(576,191,1).
txt_page(167,191,1).
txt_page(131,191,1).
txt_page(165,191,1).
txt_page(268,192,1).
txt_page(274,192,1).
txt_page(864,192,1).
txt_page(720,192,1).
txt_page(393,192,1).
txt_page(371,192,1).
txt_page(755,192,2).
txt_page(427,192,1).
txt_page(618,192,1).
txt_page(486,192,1).
txt_page(512,192,1).
txt_page(524,192,1).
txt_page(915,192,1).
txt_page(794,192,1).
txt_page(581,192,1).
txt_page(759,192,1).
txt_page(364,192,1).
txt_page(627,192,1).
txt_page(240,192,1).
txt_page(962,193,1).
txt_page(309,193,1).
txt_page(97,193,1).
txt_page(805,193,1).
txt_page(741,193,1).
txt_page(826,193,1).
txt_page(817,193,1).
txt_page(695,193,1).
txt_page(710,193,1).
txt_page(68,193,1).
txt_page(862,193,1).
txt_page(105,193,1).
txt_page(788,193,1).
txt_page(523,193,1).
txt_page(235,193,1).
txt_page(636,193,1).
txt_page(332,193,1).
txt_page(520,193,1).
txt_page(524,193,1).
txt_page(94,193,1).
txt_page(412,194,1).
txt_page(574,194,1).
txt_page(73,194,1).
txt_page(689,194,1).
txt_page(808,194,1).
txt_page(488,194,1).
txt_page(527,194,1).
txt_page(149,194,1).
txt_page(774,194,1).
txt_page(133,194,1).
txt_page(604,194,1).
txt_page(335,194,1).
txt_page(92,194,1).
txt_page(625,194,1).
txt_page(612,194,1).
txt_page(281,194,1).
txt_page(231,194,1).
txt_page(857,194,1).
txt_page(152,194,1).
txt_page(41,194,1).
txt_page(874,195,1).
txt_page(554,195,1).
txt_page(40,195,1).
txt_page(719,195,1).
txt_page(869,195,1).
txt_page(629,195,1).
txt_page(612,195,1).
txt_page(935,195,1).
txt_page(320,195,1).
txt_page(448,195,1).
txt_page(806,195,1).
txt_page(152,195,1).
txt_page(762,195,1).
txt_page(941,195,1).
txt_page(509,195,1).
txt_page(21,195,1).
txt_page(932,195,1).
txt_page(682,195,1).
txt_page(129,195,1).
txt_page(519,195,1).
txt_page(77,196,1).
txt_page(569,196,1).
txt_page(812,196,1).
txt_page(84,196,1).
txt_page(371,196,1).
txt_page(709,196,1).
txt_page(792,196,1).
txt_page(799,196,1).
txt_page(892,196,1).
txt_page(653,196,1).
txt_page(186,196,1).
txt_page(28,196,1).
txt_page(772,196,1).
txt_page(227,196,1).
txt_page(476,196,1).
txt_page(261,196,1).
txt_page(881,196,1).
txt_page(847,196,1).
txt_page(918,196,1).
txt_page(6,196,1).
txt_page(505,197,1).
txt_page(219,197,1).
txt_page(519,197,1).
txt_page(394,197,1).
txt_page(326,197,1).
txt_page(862,197,1).
txt_page(417,197,1).
txt_page(464,197,1).
txt_page(249,197,1).
txt_page(39,197,1).
txt_page(672,197,1).
txt_page(899,197,1).
txt_page(104,197,1).
txt_page(60,197,1).
txt_page(658,197,1).
txt_page(276,197,1).
txt_page(753,197,1).
txt_page(173,197,1).
txt_page(814,197,1).
txt_page(403,197,1).
txt_page(932,198,1).
txt_page(657,198,1).
txt_page(611,198,1).
txt_page(111,198,1).
txt_page(626,198,1).
txt_page(65,198,1).
txt_page(603,198,1).
txt_page(359,198,1).
txt_page(260,198,1).
txt_page(944,198,1).
txt_page(254,198,1).
txt_page(732,198,1).
txt_page(87,198,1).
txt_page(213,198,1).
txt_page(159,198,1).
txt_page(810,198,1).
txt_page(166,198,1).
txt_page(646,198,1).
txt_page(40,198,1).
txt_page(458,198,1).
txt_page(951,199,1).
txt_page(8,199,1).
txt_page(323,199,1).
txt_page(332,199,1).
txt_page(348,199,1).
txt_page(604,199,1).
txt_page(912,199,1).
txt_page(968,199,1).
txt_page(183,199,1).
txt_page(187,199,1).
txt_page(901,199,1).
txt_page(850,199,1).
txt_page(19,199,1).
txt_page(731,199,1).
txt_page(433,199,1).
txt_page(170,199,1).
txt_page(671,199,2).
txt_page(729,199,1).
txt_page(63,199,1).
txt_page(809,200,1).
txt_page(483,200,1).
txt_page(142,200,1).
txt_page(544,200,1).
txt_page(870,200,1).
txt_page(535,200,1).
txt_page(650,200,1).
txt_page(739,200,1).
txt_page(26,200,1).
txt_page(318,200,1).
txt_page(45,200,1).
txt_page(14,200,1).
txt_page(424,200,1).
txt_page(926,200,1).
txt_page(740,200,1).
txt_page(830,200,1).
txt_page(494,200,1).
txt_page(556,200,1).
txt_page(192,200,1).
txt_page(384,200,1).
txt_page(763,201,1).
txt_page(205,201,1).
txt_page(43,201,1).
txt_page(95,201,1).
txt_page(215,201,1).
txt_page(442,201,1).
txt_page(40,201,1).
txt_page(583,201,1).
txt_page(451,201,1).
txt_page(19,201,1).
txt_page(510,201,1).
txt_page(446,201,1).
txt_page(514,201,1).
txt_page(618,201,1).
txt_page(702,201,1).
txt_page(288,201,1).
txt_page(330,201,1).
txt_page(39,201,1).
txt_page(212,201,1).
txt_page(624,201,1).
txt_page(441,202,1).
txt_page(787,202,1).
txt_page(687,202,1).
txt_page(575,202,1).
txt_page(671,202,1).
txt_page(98,202,1).
txt_page(23,202,1).
txt_page(78,202,1).
txt_page(757,202,1).
txt_page(154,202,1).
txt_page(319,202,1).
txt_page(59,202,1).
txt_page(379,202,1).
txt_page(323,202,1).
txt_page(542,202,1).
txt_page(193,202,2).
txt_page(253,202,1).
txt_page(961,202,1).
txt_page(452,202,1).
txt_page(201,203,1).
txt_page(872,203,1).
txt_page(704,203,1).
txt_page(902,203,1).
txt_page(636,203,1).
txt_page(149,203,1).
txt_page(938,203,1).
txt_page(477,203,1).
txt_page(207,203,1).
txt_page(321,203,1).
txt_page(949,203,1).
txt_page(415,203,1).
txt_page(353,203,1).
txt_page(824,203,1).
txt_page(131,203,1).
txt_page(5,203,1).
txt_page(332,203,2).
txt_page(307,203,1).
txt_page(896,203,1).
txt_page(55,204,1).
txt_page(315,204,1).
txt_page(508,204,1).
txt_page(142,204,1).
txt_page(107,204,1).
txt_page(473,204,1).
txt_page(893,204,1).
txt_page(621,204,1).
txt_page(681,204,1).
txt_page(246,204,1).
txt_page(889,204,1).
txt_page(374,204,1).
txt_page(818,204,1).
txt_page(301,204,1).
txt_page(15,204,1).
txt_page(190,204,1).
txt_page(407,204,1).
txt_page(834,204,1).
txt_page(705,204,1).
txt_page(906,204,1).
txt_page(100,205,1).
txt_page(638,205,1).
txt_page(490,205,1).
txt_page(912,205,1).
txt_page(246,205,1).
txt_page(597,205,1).
txt_page(207,205,1).
txt_page(525,205,1).
txt_page(767,205,1).
txt_page(925,205,1).
txt_page(861,205,1).
txt_page(392,205,1).
txt_page(923,205,1).
txt_page(922,205,1).
txt_page(794,205,1).
txt_page(211,205,1).
txt_page(726,205,1).
txt_page(590,205,1).
txt_page(461,205,1).
txt_page(206,205,1).
txt_page(295,206,1).
txt_page(716,206,1).
txt_page(919,206,1).
txt_page(460,206,1).
txt_page(120,206,1).
txt_page(508,206,1).
txt_page(95,206,1).
txt_page(781,206,1).
txt_page(173,206,1).
txt_page(185,206,1).
txt_page(965,206,1).
txt_page(250,206,1).
txt_page(762,206,1).
txt_page(693,206,1).
txt_page(132,206,1).
txt_page(783,206,1).
txt_page(109,206,1).
txt_page(838,206,1).
txt_page(655,206,1).
txt_page(340,206,1).
txt_page(357,207,1).
txt_page(366,207,1).
txt_page(107,207,1).
txt_page(911,207,1).
txt_page(821,207,1).
txt_page(831,207,1).
txt_page(702,207,1).
txt_page(664,207,1).
txt_page(828,207,1).
txt_page(336,207,1).
txt_page(81,207,1).
txt_page(9,207,1).
txt_page(657,207,1).
txt_page(295,207,1).
txt_page(309,207,1).
txt_page(719,207,1).
txt_page(504,207,1).
txt_page(769,207,1).
txt_page(788,207,1).
txt_page(532,207,1).
txt_page(843,208,1).
txt_page(297,208,1).
txt_page(895,208,1).
txt_page(943,208,1).
txt_page(158,208,1).
txt_page(166,208,1).
txt_page(929,208,1).
txt_page(395,208,1).
txt_page(684,208,1).
txt_page(41,208,1).
txt_page(812,208,1).
txt_page(942,208,1).
txt_page(230,208,1).
txt_page(841,208,1).
txt_page(565,208,1).
txt_page(675,208,1).
txt_page(832,208,1).
txt_page(272,208,1).
txt_page(370,208,1).
txt_page(907,208,1).
txt_page(468,209,1).
txt_page(759,209,1).
txt_page(897,209,1).
txt_page(284,209,1).
txt_page(932,209,1).
txt_page(788,209,1).
txt_page(548,209,1).
txt_page(804,209,1).
txt_page(75,209,1).
txt_page(33,209,1).
txt_page(636,209,1).
txt_page(249,209,1).
txt_page(173,209,1).
txt_page(288,209,1).
txt_page(60,209,1).
txt_page(493,209,1).
txt_page(189,209,1).
txt_page(298,209,1).
txt_page(805,209,1).
txt_page(351,209,1).
txt_page(517,210,1).
txt_page(836,210,1).
txt_page(452,210,1).
txt_page(35,210,1).
txt_page(60,210,1).
txt_page(945,210,1).
txt_page(374,210,1).
txt_page(215,210,1).
txt_page(571,210,1).
txt_page(391,210,1).
txt_page(196,210,1).
txt_page(259,210,1).
txt_page(867,210,1).
txt_page(252,210,1).
txt_page(291,210,2).
txt_page(686,210,1).
txt_page(830,210,1).
txt_page(821,210,1).
txt_page(81,210,1).
txt_page(126,211,1).
txt_page(10,211,1).
txt_page(204,211,1).
txt_page(509,211,1).
txt_page(845,211,1).
txt_page(902,211,1).
txt_page(330,211,1).
txt_page(670,211,1).
txt_page(365,211,1).
txt_page(274,211,1).
txt_page(908,211,1).
txt_page(68,211,1).
txt_page(464,211,1).
txt_page(667,211,1).
txt_page(292,211,1).
txt_page(322,211,1).
txt_page(266,211,1).
txt_page(158,211,1).
txt_page(402,211,1).
txt_page(302,211,1).
txt_page(21,212,1).
txt_page(86,212,1).
txt_page(637,212,1).
txt_page(450,212,1).
txt_page(562,212,1).
txt_page(780,212,1).
txt_page(184,212,1).
txt_page(937,212,1).
txt_page(876,212,1).
txt_page(125,212,1).
txt_page(842,212,1).
txt_page(473,212,1).
txt_page(407,212,1).
txt_page(516,212,1).
txt_page(23,212,1).
txt_page(824,212,1).
txt_page(858,212,1).
txt_page(463,212,1).
txt_page(292,212,1).
txt_page(217,212,1).
txt_page(209,213,1).
txt_page(517,213,1).
txt_page(396,213,1).
txt_page(735,213,1).
txt_page(787,213,1).
txt_page(542,213,1).
txt_page(711,213,1).
txt_page(822,213,1).
txt_page(322,213,1).
txt_page(666,213,1).
txt_page(391,213,1).
txt_page(520,213,1).
txt_page(364,213,1).
txt_page(883,213,1).
txt_page(540,213,1).
txt_page(115,213,1).
txt_page(933,213,1).
txt_page(433,213,1).
txt_page(724,213,1).
txt_page(444,213,1).
txt_page(245,214,1).
txt_page(896,214,1).
txt_page(908,214,1).
txt_page(115,214,1).
txt_page(146,214,1).
txt_page(389,214,1).
txt_page(644,214,1).
txt_page(560,214,1).
txt_page(90,214,1).
txt_page(47,214,1).
txt_page(595,214,1).
txt_page(230,214,1).
txt_page(771,214,1).
txt_page(470,214,1).
txt_page(570,214,2).
txt_page(35,214,1).
txt_page(900,214,1).
txt_page(811,214,1).
txt_page(397,214,1).
txt_page(114,215,1).
txt_page(237,215,1).
txt_page(115,215,1).
txt_page(893,215,1).
txt_page(407,215,1).
txt_page(29,215,1).
txt_page(305,215,1).
txt_page(306,215,1).
txt_page(320,215,1).
txt_page(76,215,1).
txt_page(751,215,1).
txt_page(925,215,1).
txt_page(525,215,1).
txt_page(533,215,1).
txt_page(538,215,1).
txt_page(349,215,1).
txt_page(222,215,1).
txt_page(825,215,1).
txt_page(605,215,1).
txt_page(765,215,1).
txt_page(694,216,1).
txt_page(269,216,1).
txt_page(441,216,1).
txt_page(98,216,1).
txt_page(6,216,1).
txt_page(253,216,1).
txt_page(715,216,1).
txt_page(273,216,1).
txt_page(47,216,1).
txt_page(573,216,1).
txt_page(263,216,1).
txt_page(395,216,1).
txt_page(949,216,1).
txt_page(821,216,1).
txt_page(221,216,1).
txt_page(284,216,1).
txt_page(717,216,1).
txt_page(337,216,1).
txt_page(683,216,1).
txt_page(61,216,1).
txt_page(421,217,1).
txt_page(191,217,1).
txt_page(97,217,1).
txt_page(244,217,1).
txt_page(538,217,1).
txt_page(237,217,1).
txt_page(301,217,1).
txt_page(182,217,1).
txt_page(864,217,1).
txt_page(259,217,1).
txt_page(772,217,1).
txt_page(639,217,1).
txt_page(867,217,1).
txt_page(94,217,2).
txt_page(25,217,1).
txt_page(653,217,1).
txt_page(346,217,1).
txt_page(807,217,1).
txt_page(763,217,1).
txt_page(144,218,1).
txt_page(282,218,1).
txt_page(519,218,1).
txt_page(130,218,1).
txt_page(898,218,1).
txt_page(461,218,1).
txt_page(337,218,1).
txt_page(754,218,1).
txt_page(451,218,1).
txt_page(499,218,1).
txt_page(588,218,1).
txt_page(108,218,1).
txt_page(312,218,1).
txt_page(500,218,1).
txt_page(603,218,2).
txt_page(768,218,1).
txt_page(590,218,1).
txt_page(133,218,1).
txt_page(651,218,1).
txt_page(85,219,1).
txt_page(852,219,1).
txt_page(883,219,1).
txt_page(756,219,1).
txt_page(243,219,1).
txt_page(254,219,1).
txt_page(412,219,1).
txt_page(195,219,1).
txt_page(521,219,1).
txt_page(643,219,1).
txt_page(340,219,1).
txt_page(474,219,1).
txt_page(362,219,1).
txt_page(519,219,1).
txt_page(102,219,1).
txt_page(205,219,1).
txt_page(440,219,1).
txt_page(942,219,1).
txt_page(741,219,1).
txt_page(944,219,1).
txt_page(920,220,1).
txt_page(223,220,1).
txt_page(231,220,1).
txt_page(53,220,1).
txt_page(764,220,1).
txt_page(151,220,1).
txt_page(67,220,1).
txt_page(470,220,1).
txt_page(705,220,1).
txt_page(546,220,1).
txt_page(843,220,1).
txt_page(538,220,1).
txt_page(213,220,1).
txt_page(349,220,1).
txt_page(37,220,1).
txt_page(885,220,1).
txt_page(771,220,1).
txt_page(664,220,1).
txt_page(713,220,2).
txt_page(120,221,1).
txt_page(435,221,1).
txt_page(151,221,1).
txt_page(424,221,1).
txt_page(639,221,1).
txt_page(88,221,1).
txt_page(885,221,1).
txt_page(600,221,1).
txt_page(506,221,1).
txt_page(673,221,1).
txt_page(324,221,1).
txt_page(262,221,1).
txt_page(887,221,1).
txt_page(357,221,1).
txt_page(358,221,1).
txt_page(446,221,1).
txt_page(740,221,1).
txt_page(612,221,1).
txt_page(128,221,1).
txt_page(323,221,1).
txt_page(155,222,1).
txt_page(191,222,1).
txt_page(330,222,1).
txt_page(559,222,1).
txt_page(110,222,1).
txt_page(153,222,1).
txt_page(813,222,1).
txt_page(157,222,1).
txt_page(451,222,1).
txt_page(721,222,1).
txt_page(65,222,1).
txt_page(814,222,1).
txt_page(276,222,1).
txt_page(331,222,1).
txt_page(789,222,1).
txt_page(119,222,1).
txt_page(332,222,1).
txt_page(649,222,1).
txt_page(685,222,1).
txt_page(791,222,1).
txt_page(179,223,1).
txt_page(404,223,1).
txt_page(49,223,1).
txt_page(804,223,1).
txt_page(529,223,1).
txt_page(408,223,1).
txt_page(40,223,1).
txt_page(454,223,1).
txt_page(731,223,1).
txt_page(319,223,1).
txt_page(771,223,1).
txt_page(459,223,1).
txt_page(792,223,1).
txt_page(725,223,1).
txt_page(581,223,1).
txt_page(327,223,1).
txt_page(859,223,1).
txt_page(936,223,1).
txt_page(691,223,1).
txt_page(41,223,1).
txt_page(118,224,1).
txt_page(445,224,1).
txt_page(574,224,1).
txt_page(98,224,1).
txt_page(803,224,1).
txt_page(611,224,1).
txt_page(555,224,1).
txt_page(278,224,1).
txt_page(476,224,1).
txt_page(308,224,1).
txt_page(735,224,1).
txt_page(451,224,1).
txt_page(337,224,1).
txt_page(962,224,1).
txt_page(773,224,1).
txt_page(717,224,1).
txt_page(43,224,1).
txt_page(726,224,1).
txt_page(245,224,1).
txt_page(416,224,1).
txt_page(207,225,1).
txt_page(756,225,1).
txt_page(48,225,1).
txt_page(34,225,1).
txt_page(45,225,1).
txt_page(499,225,1).
txt_page(726,225,1).
txt_page(10,225,1).
txt_page(229,225,1).
txt_page(684,225,1).
txt_page(566,225,1).
txt_page(271,225,1).
txt_page(212,225,1).
txt_page(739,225,1).
txt_page(242,225,1).
txt_page(706,225,1).
txt_page(150,225,1).
txt_page(669,225,1).
txt_page(356,225,1).
txt_page(252,225,1).
txt_page(549,226,1).
txt_page(948,226,1).
txt_page(437,226,1).
txt_page(246,226,1).
txt_page(178,226,1).
txt_page(258,226,1).
txt_page(755,226,1).
txt_page(786,226,1).
txt_page(135,226,1).
txt_page(160,226,1).
txt_page(150,226,1).
txt_page(813,226,1).
txt_page(891,226,1).
txt_page(142,226,1).
txt_page(254,226,1).
txt_page(740,226,1).
txt_page(68,226,1).
txt_page(205,226,1).
txt_page(866,226,1).
txt_page(379,226,1).
txt_page(581,227,1).
txt_page(355,227,1).
txt_page(472,227,1).
txt_page(489,227,1).
txt_page(143,227,1).
txt_page(829,227,1).
txt_page(826,227,1).
txt_page(274,227,1).
txt_page(397,227,1).
txt_page(38,227,1).
txt_page(396,227,1).
txt_page(334,227,1).
txt_page(887,227,1).
txt_page(744,227,1).
txt_page(203,227,1).
txt_page(297,227,1).
txt_page(667,227,1).
txt_page(715,227,1).
txt_page(106,227,1).
txt_page(530,227,1).
txt_page(57,228,1).
txt_page(303,228,1).
txt_page(211,228,1).
txt_page(469,228,1).
txt_page(328,228,1).
txt_page(752,228,1).
txt_page(456,228,1).
txt_page(108,228,1).
txt_page(331,228,1).
txt_page(258,228,1).
txt_page(678,228,1).
txt_page(782,228,1).
txt_page(703,228,1).
txt_page(117,228,1).
txt_page(507,228,1).
txt_page(112,228,1).
txt_page(946,228,1).
txt_page(300,228,1).
txt_page(149,228,1).
txt_page(352,228,1).
txt_page(653,229,1).
txt_page(503,229,1).
txt_page(770,229,1).
txt_page(701,229,1).
txt_page(584,229,1).
txt_page(404,229,1).
txt_page(309,229,1).
txt_page(684,229,1).
txt_page(851,229,1).
txt_page(473,229,1).
txt_page(740,229,1).
txt_page(438,229,1).
txt_page(370,229,1).
txt_page(865,229,1).
txt_page(784,229,1).
txt_page(231,229,1).
txt_page(141,229,1).
txt_page(938,229,1).
txt_page(96,229,2).
txt_page(457,230,1).
txt_page(543,230,1).
txt_page(66,230,1).
txt_page(356,230,1).
txt_page(131,230,1).
txt_page(536,230,1).
txt_page(397,230,1).
txt_page(884,230,1).
txt_page(296,230,1).
txt_page(596,230,1).
txt_page(660,230,1).
txt_page(751,230,1).
txt_page(638,230,1).
txt_page(117,230,1).
txt_page(672,230,1).
txt_page(401,230,1).
txt_page(61,230,1).
txt_page(416,230,1).
txt_page(81,230,1).
txt_page(586,230,1).
txt_page(133,231,1).
txt_page(821,231,1).
txt_page(678,231,1).
txt_page(370,231,1).
txt_page(748,231,1).
txt_page(606,231,1).
txt_page(260,231,1).
txt_page(9,231,1).
txt_page(95,231,1).
txt_page(648,231,1).
txt_page(666,231,1).
txt_page(457,231,1).
txt_page(233,231,1).
txt_page(104,231,1).
txt_page(413,231,1).
txt_page(58,231,1).
txt_page(940,231,1).
txt_page(876,231,1).
txt_page(441,231,1).
txt_page(156,231,1).
txt_page(774,232,1).
txt_page(780,232,1).
txt_page(799,232,1).
txt_page(650,232,1).
txt_page(411,232,1).
txt_page(886,232,1).
txt_page(544,232,1).
txt_page(719,232,1).
txt_page(90,232,1).
txt_page(400,232,1).
txt_page(835,232,1).
txt_page(418,232,1).
txt_page(199,232,1).
txt_page(322,232,1).
txt_page(621,232,1).
txt_page(836,232,1).
txt_page(597,232,1).
txt_page(198,232,1).
txt_page(195,232,1).
txt_page(328,232,1).
txt_page(742,233,1).
txt_page(871,233,1).
txt_page(861,233,1).
txt_page(620,233,1).
txt_page(189,233,1).
txt_page(167,233,1).
txt_page(122,233,1).
txt_page(192,233,1).
txt_page(726,233,1).
txt_page(787,233,1).
txt_page(452,233,1).
txt_page(526,233,1).
txt_page(927,233,1).
txt_page(613,233,1).
txt_page(633,233,1).
txt_page(156,233,1).
txt_page(596,233,1).
txt_page(614,233,1).
txt_page(892,233,1).
txt_page(844,233,1).
txt_page(657,234,1).
txt_page(109,234,1).
txt_page(450,234,1).
txt_page(680,234,1).
txt_page(301,234,1).
txt_page(307,234,1).
txt_page(702,234,1).
txt_page(800,234,1).
txt_page(505,234,1).
txt_page(808,234,1).
txt_page(433,234,1).
txt_page(147,234,1).
txt_page(872,234,1).
txt_page(582,234,1).
txt_page(631,234,1).
txt_page(871,234,1).
txt_page(310,234,1).
txt_page(328,234,1).
txt_page(445,234,1).
txt_page(700,234,1).
txt_page(490,235,1).
txt_page(56,235,1).
txt_page(72,235,1).
txt_page(525,235,1).
txt_page(549,235,1).
txt_page(802,235,1).
txt_page(943,235,1).
txt_page(869,235,1).
txt_page(701,235,1).
txt_page(355,235,1).
txt_page(184,235,1).
txt_page(824,235,1).
txt_page(143,235,1).
txt_page(610,235,1).
txt_page(395,235,1).
txt_page(899,235,1).
txt_page(919,235,1).
txt_page(949,235,1).
txt_page(786,235,1).
txt_page(294,235,1).
txt_page(744,236,1).
txt_page(267,236,1).
txt_page(152,236,1).
txt_page(648,236,1).
txt_page(194,236,1).
txt_page(637,236,1).
txt_page(426,236,1).
txt_page(293,236,1).
txt_page(11,236,1).
txt_page(759,236,1).
txt_page(806,236,1).
txt_page(803,236,1).
txt_page(790,236,1).
txt_page(961,236,1).
txt_page(694,236,2).
txt_page(363,236,1).
txt_page(886,236,1).
txt_page(125,236,1).
txt_page(129,236,1).
txt_page(836,237,1).
txt_page(861,237,1).
txt_page(50,237,1).
txt_page(681,237,1).
txt_page(146,237,1).
txt_page(326,237,1).
txt_page(521,237,1).
txt_page(114,237,1).
txt_page(785,237,1).
txt_page(957,237,1).
txt_page(941,237,1).
txt_page(167,237,1).
txt_page(324,237,1).
txt_page(558,237,1).
txt_page(69,237,1).
txt_page(388,237,1).
txt_page(220,237,1).
txt_page(116,237,1).
txt_page(764,237,1).
txt_page(622,237,1).
txt_page(837,238,1).
txt_page(830,238,1).
txt_page(259,238,1).
txt_page(467,238,1).
txt_page(285,238,1).
txt_page(882,238,1).
txt_page(672,238,1).
txt_page(387,238,1).
txt_page(451,238,1).
txt_page(91,238,1).
txt_page(28,238,1).
txt_page(727,238,1).
txt_page(549,238,1).
txt_page(123,238,1).
txt_page(863,238,1).
txt_page(52,238,1).
txt_page(956,238,1).
txt_page(946,238,1).
txt_page(501,238,1).
txt_page(360,238,1).
txt_page(327,239,1).
txt_page(430,239,1).
txt_page(310,239,1).
txt_page(230,239,1).
txt_page(561,239,1).
txt_page(246,239,1).
txt_page(413,239,1).
txt_page(452,239,1).
txt_page(447,239,1).
txt_page(473,239,1).
txt_page(11,239,1).
txt_page(151,239,1).
txt_page(672,239,1).
txt_page(678,239,1).
txt_page(715,239,1).
txt_page(211,239,1).
txt_page(766,239,1).
txt_page(558,239,1).
txt_page(22,239,1).
txt_page(289,239,1).
txt_page(816,240,1).
txt_page(669,240,1).
txt_page(900,240,1).
txt_page(739,240,1).
txt_page(97,240,1).
txt_page(233,240,1).
txt_page(742,240,1).
txt_page(759,240,1).
txt_page(732,240,1).
txt_page(449,240,1).
txt_page(648,240,1).
txt_page(89,240,1).
txt_page(432,240,1).
txt_page(562,240,1).
txt_page(22,240,1).
txt_page(465,240,1).
txt_page(339,240,1).
txt_page(141,240,1).
txt_page(214,240,1).
txt_page(398,240,1).
txt_page(487,241,1).
txt_page(706,241,1).
txt_page(69,241,1).
txt_page(253,241,1).
txt_page(427,241,1).
txt_page(334,241,1).
txt_page(151,241,1).
txt_page(834,241,1).
txt_page(240,241,1).
txt_page(208,241,1).
txt_page(213,241,1).
txt_page(494,241,1).
txt_page(696,241,1).
txt_page(656,241,1).
txt_page(425,241,1).
txt_page(768,241,1).
txt_page(776,241,1).
txt_page(812,241,1).
txt_page(430,241,1).
txt_page(222,241,1).
txt_page(806,242,1).
txt_page(837,242,1).
txt_page(670,242,1).
txt_page(263,242,1).
txt_page(849,242,1).
txt_page(694,242,1).
txt_page(944,242,1).
txt_page(932,242,1).
txt_page(795,242,1).
txt_page(170,242,1).
txt_page(185,242,1).
txt_page(871,242,1).
txt_page(492,242,1).
txt_page(514,242,1).
txt_page(468,242,1).
txt_page(24,242,1).
txt_page(300,242,1).
txt_page(820,242,1).
txt_page(182,242,1).
txt_page(525,242,1).
txt_page(235,243,1).
txt_page(644,243,1).
txt_page(862,243,1).
txt_page(432,243,1).
txt_page(15,243,1).
txt_page(731,243,1).
txt_page(261,243,1).
txt_page(14,243,1).
txt_page(907,243,1).
txt_page(804,243,1).
txt_page(943,243,1).
txt_page(396,243,1).
txt_page(181,243,1).
txt_page(78,243,1).
txt_page(713,243,1).
txt_page(377,243,1).
txt_page(675,243,1).
txt_page(930,243,1).
txt_page(646,243,1).
txt_page(867,243,1).
txt_page(471,244,1).
txt_page(881,244,1).
txt_page(466,244,1).
txt_page(87,244,1).
txt_page(211,244,1).
txt_page(671,244,1).
txt_page(319,244,1).
txt_page(702,244,1).
txt_page(701,244,2).
txt_page(396,244,1).
txt_page(562,244,1).
txt_page(15,244,1).
txt_page(461,244,1).
txt_page(364,244,1).
txt_page(159,244,1).
txt_page(548,244,1).
txt_page(483,244,1).
txt_page(216,244,1).
txt_page(78,244,1).
txt_page(554,245,1).
txt_page(382,245,1).
txt_page(315,245,1).
txt_page(763,245,1).
txt_page(31,245,1).
txt_page(691,245,1).
txt_page(326,245,1).
txt_page(393,245,1).
txt_page(432,245,1).
txt_page(164,245,1).
txt_page(689,245,1).
txt_page(351,245,1).
txt_page(140,245,1).
txt_page(891,245,1).
txt_page(675,245,1).
txt_page(696,245,1).
txt_page(747,245,1).
txt_page(345,245,1).
txt_page(334,245,1).
txt_page(79,245,1).
txt_page(560,246,1).
txt_page(591,246,1).
txt_page(584,246,1).
txt_page(689,246,1).
txt_page(365,246,1).
txt_page(652,246,1).
txt_page(74,246,1).
txt_page(596,246,1).
txt_page(510,246,1).
txt_page(76,246,1).
txt_page(894,246,1).
txt_page(493,246,1).
txt_page(937,246,1).
txt_page(565,246,1).
txt_page(547,246,1).
txt_page(77,246,1).
txt_page(292,246,1).
txt_page(613,246,1).
txt_page(695,246,1).
txt_page(335,246,1).
txt_page(413,247,1).
txt_page(645,247,1).
txt_page(543,247,1).
txt_page(203,247,1).
txt_page(350,247,1).
txt_page(944,247,1).
txt_page(900,247,1).
txt_page(892,247,1).
txt_page(294,247,1).
txt_page(97,247,1).
txt_page(796,247,1).
txt_page(728,247,1).
txt_page(552,247,1).
txt_page(820,247,1).
txt_page(667,247,1).
txt_page(162,247,1).
txt_page(3,247,1).
txt_page(529,247,1).
txt_page(169,247,1).
txt_page(285,247,1).
txt_page(480,248,1).
txt_page(39,248,1).
txt_page(749,248,1).
txt_page(54,248,1).
txt_page(964,248,1).
txt_page(57,248,1).
txt_page(89,248,1).
txt_page(560,248,1).
txt_page(446,248,1).
txt_page(695,248,1).
txt_page(930,248,1).
txt_page(959,248,1).
txt_page(617,248,1).
txt_page(197,248,1).
txt_page(856,248,1).
txt_page(573,248,1).
txt_page(847,248,1).
txt_page(523,248,1).
txt_page(273,248,1).
txt_page(264,248,1).
txt_page(779,249,1).
txt_page(420,249,1).
txt_page(773,249,1).
txt_page(608,249,1).
txt_page(112,249,1).
txt_page(504,249,1).
txt_page(724,249,1).
txt_page(90,249,1).
txt_page(614,249,1).
txt_page(892,249,1).
txt_page(296,249,1).
txt_page(782,249,1).
txt_page(814,249,1).
txt_page(635,249,1).
txt_page(642,249,1).
txt_page(529,249,1).
txt_page(69,249,1).
txt_page(434,249,1).
txt_page(404,249,1).
txt_page(515,249,1).
txt_page(144,250,1).
txt_page(301,250,1).
txt_page(506,250,1).
txt_page(157,250,1).
txt_page(825,250,1).
txt_page(664,250,1).
txt_page(133,250,1).
txt_page(447,250,1).
txt_page(96,250,1).
txt_page(364,250,1).
txt_page(528,250,1).
txt_page(726,250,1).
txt_page(152,250,1).
txt_page(55,250,1).
txt_page(393,250,1).
txt_page(335,250,1).
txt_page(428,250,1).
txt_page(657,250,2).
txt_page(795,250,1).
txt_page(636,251,1).
txt_page(440,251,1).
txt_page(621,251,1).
txt_page(705,251,1).
txt_page(324,251,1).
txt_page(879,251,1).
txt_page(669,251,1).
txt_page(102,251,1).
txt_page(811,251,1).
txt_page(4,251,1).
txt_page(44,251,1).
txt_page(148,251,1).
txt_page(820,251,2).
txt_page(620,251,1).
txt_page(99,251,1).
txt_page(335,251,1).
txt_page(143,251,1).
txt_page(770,251,1).
txt_page(19,251,1).
txt_page(737,252,1).
txt_page(764,252,1).
txt_page(340,252,1).
txt_page(678,252,1).
txt_page(3,252,1).
txt_page(99,252,1).
txt_page(922,252,1).
txt_page(659,252,1).
txt_page(540,252,1).
txt_page(919,252,1).
txt_page(477,252,1).
txt_page(579,252,1).
txt_page(406,252,2).
txt_page(873,252,1).
txt_page(71,252,1).
txt_page(353,252,1).
txt_page(591,252,1).
txt_page(899,252,1).
txt_page(356,252,1).
txt_page(945,253,1).
txt_page(862,253,1).
txt_page(627,253,1).
txt_page(21,253,1).
txt_page(213,253,1).
txt_page(177,253,1).
txt_page(71,253,1).
txt_page(441,253,1).
txt_page(734,253,1).
txt_page(463,253,1).
txt_page(745,253,1).
txt_page(76,253,1).
txt_page(541,253,1).
txt_page(828,253,1).
txt_page(824,253,1).
txt_page(570,253,1).
txt_page(465,253,1).
txt_page(731,253,1).
txt_page(823,253,1).
txt_page(913,253,1).
txt_page(306,254,1).
txt_page(947,254,1).
txt_page(502,254,1).
txt_page(958,254,1).
txt_page(483,254,2).
txt_page(768,254,1).
txt_page(399,254,1).
txt_page(914,254,1).
txt_page(179,254,1).
txt_page(164,254,1).
txt_page(850,254,1).
txt_page(24,254,1).
txt_page(91,254,1).
txt_page(884,254,1).
txt_page(617,254,1).
txt_page(183,254,1).
txt_page(107,254,1).
txt_page(704,254,1).
txt_page(336,254,1).
txt_page(591,255,1).
txt_page(396,255,1).
txt_page(476,255,1).
txt_page(400,255,1).
txt_page(17,255,1).
txt_page(404,255,1).
txt_page(479,255,1).
txt_page(858,255,1).
txt_page(94,255,1).
txt_page(245,255,1).
txt_page(821,255,1).
txt_page(290,255,1).
txt_page(358,255,1).
txt_page(450,255,1).
txt_page(598,255,1).
txt_page(385,255,1).
txt_page(149,255,1).
txt_page(83,255,1).
txt_page(364,255,1).
txt_page(251,255,1).
txt_page(774,256,1).
txt_page(528,256,1).
txt_page(729,256,1).
txt_page(654,256,1).
txt_page(62,256,1).
txt_page(617,256,1).
txt_page(118,256,1).
txt_page(828,256,1).
txt_page(937,256,1).
txt_page(556,256,1).
txt_page(616,256,1).
txt_page(55,256,1).
txt_page(931,256,1).
txt_page(241,256,1).
txt_page(177,256,1).
txt_page(26,256,1).
txt_page(368,256,1).
txt_page(53,256,1).
txt_page(334,256,1).
txt_page(719,256,1).
txt_page(456,257,1).
txt_page(259,257,1).
txt_page(629,257,1).
txt_page(874,257,1).
txt_page(770,257,1).
txt_page(245,257,1).
txt_page(363,257,1).
txt_page(348,257,1).
txt_page(563,257,1).
txt_page(825,257,1).
txt_page(93,257,1).
txt_page(258,257,1).
txt_page(624,257,1).
txt_page(367,257,1).
txt_page(66,257,1).
txt_page(398,257,1).
txt_page(404,257,1).
txt_page(375,257,2).
txt_page(450,257,1).
txt_page(808,258,1).
txt_page(61,258,1).
txt_page(779,258,1).
txt_page(685,258,1).
txt_page(845,258,1).
txt_page(766,258,2).
txt_page(365,258,1).
txt_page(596,258,1).
txt_page(686,258,1).
txt_page(826,258,1).
txt_page(641,258,1).
txt_page(875,258,1).
txt_page(251,258,1).
txt_page(708,258,1).
txt_page(90,258,1).
txt_page(67,258,1).
txt_page(314,258,1).
txt_page(17,258,1).
txt_page(442,258,1).
txt_page(862,259,1).
txt_page(197,259,1).
txt_page(275,259,1).
txt_page(14,259,1).
txt_page(725,259,1).
txt_page(143,259,1).
txt_page(918,259,1).
txt_page(382,259,1).
txt_page(423,259,1).
txt_page(649,259,1).
txt_page(199,259,1).
txt_page(166,259,1).
txt_page(552,259,1).
txt_page(686,259,1).
txt_page(399,259,1).
txt_page(56,259,1).
txt_page(30,259,1).
txt_page(457,259,1).
txt_page(108,259,1).
txt_page(523,259,1).
txt_page(513,260,1).
txt_page(388,260,1).
txt_page(776,260,1).
txt_page(91,260,1).
txt_page(123,260,1).
txt_page(593,260,1).
txt_page(101,260,1).
txt_page(443,260,1).
txt_page(340,260,1).
txt_page(409,260,1).
txt_page(602,260,1).
txt_page(250,260,1).
txt_page(481,260,1).
txt_page(768,260,1).
txt_page(707,260,1).
txt_page(630,260,1).
txt_page(479,260,1).
txt_page(723,260,1).
txt_page(426,260,1).
txt_page(163,260,1).
txt_page(133,261,1).
txt_page(717,261,1).
txt_page(88,261,1).
txt_page(197,261,1).
txt_page(899,261,1).
txt_page(548,261,1).
txt_page(865,261,1).
txt_page(125,261,1).
txt_page(472,261,1).
txt_page(532,261,1).
txt_page(555,261,1).
txt_page(43,261,1).
txt_page(805,261,1).
txt_page(259,261,1).
txt_page(706,261,1).
txt_page(254,261,1).
txt_page(311,261,1).
txt_page(151,261,1).
txt_page(891,261,1).
txt_page(388,261,1).
txt_page(150,262,1).
txt_page(365,262,1).
txt_page(281,262,1).
txt_page(210,262,1).
txt_page(727,262,1).
txt_page(129,262,1).
txt_page(311,262,1).
txt_page(34,262,1).
txt_page(559,262,1).
txt_page(38,262,1).
txt_page(503,262,1).
txt_page(675,262,1).
txt_page(215,262,1).
txt_page(452,262,1).
txt_page(746,262,1).
txt_page(943,262,1).
txt_page(906,262,1).
txt_page(569,262,1).
txt_page(351,262,1).
txt_page(107,262,1).
txt_page(239,263,1).
txt_page(649,263,1).
txt_page(206,263,1).
txt_page(826,263,1).
txt_page(633,263,1).
txt_page(399,263,1).
txt_page(339,263,1).
txt_page(679,263,1).
txt_page(872,263,1).
txt_page(535,263,1).
txt_page(846,263,1).
txt_page(1,263,1).
txt_page(40,263,1).
txt_page(874,263,1).
txt_page(303,263,1).
txt_page(621,263,1).
txt_page(761,263,1).
txt_page(248,263,1).
txt_page(721,263,1).
txt_page(819,263,1).
txt_page(566,264,1).
txt_page(463,264,1).
txt_page(373,264,1).
txt_page(657,264,1).
txt_page(218,264,1).
txt_page(687,264,1).
txt_page(404,264,1).
txt_page(248,264,1).
txt_page(504,264,1).
txt_page(768,264,1).
txt_page(886,264,1).
txt_page(102,264,1).
txt_page(937,264,1).
txt_page(460,264,1).
txt_page(305,264,1).
txt_page(14,264,1).
txt_page(35,264,1).
txt_page(66,264,1).
txt_page(439,264,1).
txt_page(27,264,1).
txt_page(557,265,1).
txt_page(72,265,1).
txt_page(876,265,1).
txt_page(634,265,1).
txt_page(430,265,1).
txt_page(90,265,1).
txt_page(696,265,1).
txt_page(108,265,1).
txt_page(912,265,1).
txt_page(266,265,1).
txt_page(650,265,1).
txt_page(555,265,1).
txt_page(499,265,1).
txt_page(852,265,1).
txt_page(342,265,1).
txt_page(253,265,1).
txt_page(127,265,1).
txt_page(480,265,1).
txt_page(435,265,1).
txt_page(402,265,1).
txt_page(944,266,1).
txt_page(120,266,1).
txt_page(283,266,1).
txt_page(941,266,1).
txt_page(798,266,1).
txt_page(420,266,1).
txt_page(223,266,1).
txt_page(459,266,1).
txt_page(833,266,1).
txt_page(568,266,1).
txt_page(774,266,1).
txt_page(412,266,1).
txt_page(163,266,1).
txt_page(246,266,1).
txt_page(680,266,1).
txt_page(968,266,1).
txt_page(959,266,1).
txt_page(295,266,1).
txt_page(532,266,1).
txt_page(701,266,1).
txt_page(766,267,1).
txt_page(196,267,1).
txt_page(577,267,1).
txt_page(95,267,1).
txt_page(367,267,1).
txt_page(10,267,1).
txt_page(660,267,1).
txt_page(82,267,1).
txt_page(28,267,1).
txt_page(292,267,1).
txt_page(355,267,1).
txt_page(152,267,1).
txt_page(733,267,1).
txt_page(477,267,1).
txt_page(385,267,1).
txt_page(184,267,1).
txt_page(168,267,1).
txt_page(837,267,1).
txt_page(848,267,1).
txt_page(522,267,1).
txt_page(720,268,1).
txt_page(711,268,1).
txt_page(482,268,1).
txt_page(349,268,1).
txt_page(695,268,1).
txt_page(889,268,1).
txt_page(495,268,1).
txt_page(608,268,1).
txt_page(382,268,1).
txt_page(481,268,1).
txt_page(247,268,1).
txt_page(670,268,1).
txt_page(77,268,1).
txt_page(273,268,1).
txt_page(589,268,1).
txt_page(174,268,1).
txt_page(12,268,1).
txt_page(699,268,1).
txt_page(131,268,1).
txt_page(766,268,1).
txt_page(264,269,1).
txt_page(928,269,1).
txt_page(138,269,1).
txt_page(316,269,1).
txt_page(106,269,1).
txt_page(509,269,1).
txt_page(183,269,1).
txt_page(297,269,1).
txt_page(360,269,1).
txt_page(675,269,1).
txt_page(387,269,1).
txt_page(492,269,1).
txt_page(378,269,1).
txt_page(785,269,1).
txt_page(365,269,1).
txt_page(11,269,1).
txt_page(333,269,1).
txt_page(89,269,1).
txt_page(711,269,1).
txt_page(80,269,1).
txt_page(428,270,1).
txt_page(484,270,1).
txt_page(679,270,1).
txt_page(758,270,1).
txt_page(34,270,1).
txt_page(310,270,1).
txt_page(571,270,1).
txt_page(337,270,1).
txt_page(570,270,1).
txt_page(168,270,1).
txt_page(757,270,1).
txt_page(487,270,1).
txt_page(670,270,1).
txt_page(652,270,2).
txt_page(223,270,1).
txt_page(22,270,1).
txt_page(477,270,1).
txt_page(416,270,1).
txt_page(485,270,1).
txt_page(222,271,1).
txt_page(487,271,1).
txt_page(518,271,1).
txt_page(931,271,1).
txt_page(872,271,1).
txt_page(82,271,1).
txt_page(814,271,1).
txt_page(898,271,1).
txt_page(363,271,1).
txt_page(286,271,1).
txt_page(815,271,1).
txt_page(622,271,1).
txt_page(535,271,1).
txt_page(711,271,1).
txt_page(278,271,1).
txt_page(110,271,1).
txt_page(579,271,1).
txt_page(505,271,1).
txt_page(327,271,1).
txt_page(689,271,1).
txt_page(338,272,1).
txt_page(588,272,1).
txt_page(518,272,1).
txt_page(200,272,1).
txt_page(759,272,2).
txt_page(744,272,1).
txt_page(1,272,1).
txt_page(29,272,1).
txt_page(902,272,1).
txt_page(774,272,1).
txt_page(706,272,1).
txt_page(594,272,1).
txt_page(254,272,1).
txt_page(111,272,1).
txt_page(206,272,1).
txt_page(517,272,1).
txt_page(33,272,1).
txt_page(692,272,1).
txt_page(397,272,1).
txt_page(252,273,1).
txt_page(681,273,1).
txt_page(77,273,1).
txt_page(375,273,1).
txt_page(411,273,1).
txt_page(836,273,1).
txt_page(657,273,1).
txt_page(761,273,1).
txt_page(25,273,1).
txt_page(323,273,1).
txt_page(108,273,1).
txt_page(174,273,1).
txt_page(26,273,1).
txt_page(183,273,1).
txt_page(897,273,1).
txt_page(548,273,1).
txt_page(713,273,1).
txt_page(361,273,1).
txt_page(188,273,1).
txt_page(480,273,1).
txt_page(313,274,1).
txt_page(449,274,1).
txt_page(847,274,1).
txt_page(946,274,1).
txt_page(675,274,1).
txt_page(390,274,1).
txt_page(268,274,1).
txt_page(397,274,1).
txt_page(420,274,1).
txt_page(894,274,1).
txt_page(562,274,1).
txt_page(110,274,1).
txt_page(716,274,1).
txt_page(678,274,1).
txt_page(457,274,1).
txt_page(49,274,1).
txt_page(250,274,1).
txt_page(866,274,1).
txt_page(636,274,1).
txt_page(846,274,1).
txt_page(389,275,1).
txt_page(761,275,1).
txt_page(101,275,1).
txt_page(561,275,1).
txt_page(475,275,1).
txt_page(412,275,1).
txt_page(606,275,1).
txt_page(529,275,1).
txt_page(254,275,1).
txt_page(843,275,1).
txt_page(264,275,1).
txt_page(220,275,1).
txt_page(694,275,1).
txt_page(287,275,1).
txt_page(452,275,1).
txt_page(103,275,1).
txt_page(339,275,1).
txt_page(557,275,1).
txt_page(808,275,1).
txt_page(788,275,1).
txt_page(776,276,1).
txt_page(170,276,1).
txt_page(92,276,1).
txt_page(356,276,1).
txt_page(762,276,1).
txt_page(288,276,1).
txt_page(615,276,1).
txt_page(152,276,1).
txt_page(71,276,1).
txt_page(210,276,1).
txt_page(559,276,1).
txt_page(246,276,1).
txt_page(520,276,1).
txt_page(816,276,1).
txt_page(337,276,1).
txt_page(198,276,1).
txt_page(84,276,1).
txt_page(620,276,1).
txt_page(186,276,1).
txt_page(550,276,1).
txt_page(179,277,1).
txt_page(505,277,1).
txt_page(511,277,1).
txt_page(803,277,1).
txt_page(898,277,1).
txt_page(895,277,1).
txt_page(800,277,1).
txt_page(774,277,1).
txt_page(167,277,1).
txt_page(677,277,1).
txt_page(318,277,1).
txt_page(703,277,1).
txt_page(40,277,1).
txt_page(75,277,1).
txt_page(776,277,1).
txt_page(704,277,1).
txt_page(798,277,1).
txt_page(914,277,1).
txt_page(857,277,1).
txt_page(916,277,1).
txt_page(571,278,1).
txt_page(81,278,1).
txt_page(908,278,1).
txt_page(768,278,1).
txt_page(743,278,1).
txt_page(246,278,1).
txt_page(516,278,1).
txt_page(242,278,1).
txt_page(187,278,1).
txt_page(16,278,1).
txt_page(547,278,1).
txt_page(446,278,1).
txt_page(86,278,1).
txt_page(561,278,1).
txt_page(853,278,1).
txt_page(548,278,1).
txt_page(677,278,1).
txt_page(502,278,1).
txt_page(204,278,1).
txt_page(99,278,1).
txt_page(656,279,1).
txt_page(726,279,1).
txt_page(241,279,1).
txt_page(555,279,1).
txt_page(332,279,1).
txt_page(283,279,1).
txt_page(900,279,1).
txt_page(236,279,1).
txt_page(680,279,1).
txt_page(201,279,1).
txt_page(143,279,1).
txt_page(563,279,1).
txt_page(204,279,1).
txt_page(30,279,1).
txt_page(556,279,1).
txt_page(358,279,1).
txt_page(135,279,1).
txt_page(398,279,1).
txt_page(256,279,1).
txt_page(659,279,1).
txt_page(461,280,1).
txt_page(686,280,1).
txt_page(772,280,1).
txt_page(832,280,1).
txt_page(537,280,1).
txt_page(599,280,1).
txt_page(873,280,1).
txt_page(354,280,1).
txt_page(244,280,1).
txt_page(486,280,1).
txt_page(304,280,1).
txt_page(63,280,1).
txt_page(113,280,1).
txt_page(777,280,1).
txt_page(913,280,1).
txt_page(643,280,1).
txt_page(232,280,1).
txt_page(668,280,1).
txt_page(662,280,1).
txt_page(925,280,1).
txt_page(923,281,1).
txt_page(808,281,1).
txt_page(166,281,1).
txt_page(136,281,1).
txt_page(551,281,1).
txt_page(721,281,1).
txt_page(919,281,1).
txt_page(247,281,1).
txt_page(685,281,1).
txt_page(225,281,1).
txt_page(963,281,1).
txt_page(348,281,1).
txt_page(502,281,1).
txt_page(505,281,1).
txt_page(483,281,2).
txt_page(457,281,1).
txt_page(462,281,1).
txt_page(585,281,1).
txt_page(864,281,1).
txt_page(548,282,1).
txt_page(802,282,1).
txt_page(59,282,1).
txt_page(946,282,1).
txt_page(436,282,1).
txt_page(812,282,1).
txt_page(664,282,1).
txt_page(92,282,1).
txt_page(592,282,1).
txt_page(964,282,1).
txt_page(495,282,1).
txt_page(98,282,1).
txt_page(788,282,1).
txt_page(675,282,1).
txt_page(519,282,1).
txt_page(867,282,1).
txt_page(164,282,1).
txt_page(836,282,1).
txt_page(71,282,1).
txt_page(326,282,1).
txt_page(940,283,1).
txt_page(872,283,1).
txt_page(798,283,1).
txt_page(384,283,1).
txt_page(191,283,1).
txt_page(620,283,1).
txt_page(634,283,1).
txt_page(464,283,1).
txt_page(379,283,1).
txt_page(308,283,1).
txt_page(290,283,1).
txt_page(155,283,1).
txt_page(968,283,1).
txt_page(426,283,1).
txt_page(256,283,1).
txt_page(686,283,1).
txt_page(927,283,1).
txt_page(767,283,1).
txt_page(424,283,1).
txt_page(312,283,1).
txt_page(633,284,1).
txt_page(731,284,1).
txt_page(846,284,1).
txt_page(545,284,1).
txt_page(497,284,1).
txt_page(103,284,1).
txt_page(936,284,1).
txt_page(364,284,1).
txt_page(587,284,1).
txt_page(220,284,1).
txt_page(510,284,1).
txt_page(666,284,1).
txt_page(319,284,1).
txt_page(953,284,1).
txt_page(621,284,1).
txt_page(865,284,1).
txt_page(630,284,1).
txt_page(941,284,1).
txt_page(822,284,1).
txt_page(964,284,1).
txt_page(374,285,1).
txt_page(321,285,1).
txt_page(962,285,1).
txt_page(163,285,1).
txt_page(710,285,1).
txt_page(411,285,1).
txt_page(623,285,1).
txt_page(617,285,1).
txt_page(1,285,1).
txt_page(351,285,1).
txt_page(748,285,1).
txt_page(865,285,1).
txt_page(282,285,1).
txt_page(111,285,1).
txt_page(961,285,1).
txt_page(24,285,1).
txt_page(325,285,1).
txt_page(17,285,1).
txt_page(438,285,1).
txt_page(697,285,1).
txt_page(838,286,1).
txt_page(60,286,1).
txt_page(98,286,1).
txt_page(960,286,1).
txt_page(747,286,1).
txt_page(56,286,1).
txt_page(451,286,1).
txt_page(763,286,1).
txt_page(754,286,1).
txt_page(470,286,1).
txt_page(164,286,1).
txt_page(54,286,1).
txt_page(903,286,1).
txt_page(541,286,1).
txt_page(814,286,1).
txt_page(88,286,1).
txt_page(532,286,1).
txt_page(347,286,1).
txt_page(687,286,1).
txt_page(659,286,1).
txt_page(410,287,1).
txt_page(69,287,1).
txt_page(466,287,1).
txt_page(808,287,1).
txt_page(546,287,1).
txt_page(886,287,1).
txt_page(90,287,1).
txt_page(343,287,1).
txt_page(919,287,1).
txt_page(583,287,1).
txt_page(628,287,1).
txt_page(776,287,1).
txt_page(200,287,1).
txt_page(592,287,1).
txt_page(532,287,1).
txt_page(293,287,1).
txt_page(557,287,1).
txt_page(679,287,1).
txt_page(125,287,1).
txt_page(83,287,1).
txt_page(323,288,1).
txt_page(170,288,1).
txt_page(500,288,1).
txt_page(637,288,1).
txt_page(304,288,1).
txt_page(339,288,1).
txt_page(141,288,1).
txt_page(465,288,1).
txt_page(436,288,1).
txt_page(338,288,1).
txt_page(10,288,1).
txt_page(898,288,1).
txt_page(99,288,1).
txt_page(80,288,1).
txt_page(341,288,1).
txt_page(200,288,1).
txt_page(443,288,1).
txt_page(591,288,1).
txt_page(610,288,1).
txt_page(686,288,1).
txt_page(489,289,1).
txt_page(689,289,1).
txt_page(102,289,1).
txt_page(638,289,1).
txt_page(124,289,1).
txt_page(684,289,1).
txt_page(922,289,1).
txt_page(644,289,1).
txt_page(156,289,1).
txt_page(344,289,1).
txt_page(799,289,1).
txt_page(309,289,1).
txt_page(733,289,1).
txt_page(64,289,1).
txt_page(282,289,1).
txt_page(915,289,1).
txt_page(40,289,1).
txt_page(913,289,1).
txt_page(401,289,1).
txt_page(22,289,1).
txt_page(573,290,1).
txt_page(251,290,1).
txt_page(234,290,1).
txt_page(351,290,1).
txt_page(131,290,1).
txt_page(257,290,1).
txt_page(607,290,1).
txt_page(299,290,1).
txt_page(677,290,1).
txt_page(477,290,1).
txt_page(5,290,1).
txt_page(841,290,1).
txt_page(9,290,1).
txt_page(683,290,1).
txt_page(749,290,1).
txt_page(667,290,1).
txt_page(886,290,1).
txt_page(753,290,1).
txt_page(194,290,1).
txt_page(74,290,1).
txt_page(39,291,1).
txt_page(201,291,1).
txt_page(329,291,1).
txt_page(261,291,1).
txt_page(383,291,1).
txt_page(761,291,1).
txt_page(759,291,1).
txt_page(810,291,1).
txt_page(771,291,1).
txt_page(658,291,1).
txt_page(927,291,1).
txt_page(309,291,1).
txt_page(67,291,1).
txt_page(610,291,1).
txt_page(699,291,1).
txt_page(181,291,1).
txt_page(179,291,1).
txt_page(23,291,1).
txt_page(616,291,1).
txt_page(55,291,1).
txt_page(4,292,1).
txt_page(719,292,1).
txt_page(265,292,1).
txt_page(273,292,1).
txt_page(443,292,1).
txt_page(166,292,1).
txt_page(315,292,1).
txt_page(140,292,1).
txt_page(261,292,1).
txt_page(783,292,1).
txt_page(81,292,1).
txt_page(282,292,1).
txt_page(159,292,1).
txt_page(72,292,1).
txt_page(685,292,1).
txt_page(107,292,1).
txt_page(395,292,1).
txt_page(204,292,1).
txt_page(793,292,1).
txt_page(602,292,1).
txt_page(379,293,1).
txt_page(335,293,1).
txt_page(533,293,1).
txt_page(201,293,1).
txt_page(15,293,1).
txt_page(766,293,1).
txt_page(701,293,1).
txt_page(236,293,1).
txt_page(436,293,1).
txt_page(357,293,1).
txt_page(98,293,1).
txt_page(11,293,1).
txt_page(272,293,1).
txt_page(763,293,1).
txt_page(321,293,1).
txt_page(332,293,1).
txt_page(239,293,1).
txt_page(16,293,1).
txt_page(748,293,1).
txt_page(390,293,1).
txt_page(503,294,1).
txt_page(706,294,1).
txt_page(482,294,1).
txt_page(956,294,1).
txt_page(159,294,1).
txt_page(773,294,1).
txt_page(772,294,1).
txt_page(438,294,1).
txt_page(867,294,1).
txt_page(649,294,1).
txt_page(121,294,1).
txt_page(38,294,1).
txt_page(240,294,1).
txt_page(138,294,1).
txt_page(611,294,1).
txt_page(111,294,1).
txt_page(534,294,1).
txt_page(642,294,1).
txt_page(177,294,1).
txt_page(36,294,1).
txt_page(777,295,1).
txt_page(727,295,1).
txt_page(603,295,1).
txt_page(550,295,1).
txt_page(293,295,1).
txt_page(471,295,1).
txt_page(101,295,1).
txt_page(374,295,1).
txt_page(337,295,1).
txt_page(427,295,1).
txt_page(853,295,1).
txt_page(513,295,1).
txt_page(885,295,1).
txt_page(730,295,1).
txt_page(254,295,1).
txt_page(741,295,1).
txt_page(959,295,1).
txt_page(621,295,1).
txt_page(306,295,1).
txt_page(783,295,1).
txt_page(704,296,1).
txt_page(219,296,1).
txt_page(920,296,1).
txt_page(309,296,1).
txt_page(872,296,1).
txt_page(761,296,1).
txt_page(542,296,1).
txt_page(916,296,1).
txt_page(47,296,1).
txt_page(728,296,1).
txt_page(240,296,1).
txt_page(819,296,1).
txt_page(532,296,1).
txt_page(143,296,1).
txt_page(811,296,1).
txt_page(786,296,1).
txt_page(306,296,1).
txt_page(596,296,1).
txt_page(204,296,1).
txt_page(880,296,1).
txt_page(737,297,1).
txt_page(279,297,1).
txt_page(865,297,1).
txt_page(602,297,1).
txt_page(483,297,1).
txt_page(156,297,1).
txt_page(55,297,1).
txt_page(488,297,1).
txt_page(422,297,1).
txt_page(176,297,1).
txt_page(726,297,1).
txt_page(925,297,1).
txt_page(557,297,1).
txt_page(494,297,1).
txt_page(399,297,1).
txt_page(678,297,1).
txt_page(12,297,1).
txt_page(227,297,1).
txt_page(413,297,1).
txt_page(13,297,1).
txt_page(229,298,1).
txt_page(813,298,1).
txt_page(215,298,1).
txt_page(426,298,1).
txt_page(339,298,1).
txt_page(866,298,1).
txt_page(731,298,1).
txt_page(510,298,1).
txt_page(126,298,1).
txt_page(185,298,1).
txt_page(604,298,1).
txt_page(492,298,1).
txt_page(700,298,1).
txt_page(267,298,1).
txt_page(46,298,1).
txt_page(423,298,1).
txt_page(69,298,1).
txt_page(5,298,1).
txt_page(929,298,1).
txt_page(542,298,1).
txt_page(63,299,1).
txt_page(213,299,1).
txt_page(696,299,1).
txt_page(695,299,1).
txt_page(641,299,1).
txt_page(610,299,1).
txt_page(83,299,1).
txt_page(895,299,1).
txt_page(405,299,1).
txt_page(528,299,1).
txt_page(877,299,1).
txt_page(477,299,1).
txt_page(621,299,1).
txt_page(826,299,1).
txt_page(329,299,1).
txt_page(863,299,1).
txt_page(754,299,1).
txt_page(761,299,1).
txt_page(741,299,1).
txt_page(841,299,1).
txt_page(138,300,1).
txt_page(88,300,1).
txt_page(791,300,1).
txt_page(50,300,1).
txt_page(194,300,1).
txt_page(584,300,1).
txt_page(382,300,1).
txt_page(782,300,1).
txt_page(701,300,1).
txt_page(274,300,1).
txt_page(110,300,1).
txt_page(306,300,1).
txt_page(34,300,1).
txt_page(179,300,1).
txt_page(260,300,1).
txt_page(638,300,1).
txt_page(687,300,1).
txt_page(629,300,1).
txt_page(324,300,1).
txt_page(190,300,1).
txt_page(656,301,1).
txt_page(354,301,1).
txt_page(649,301,1).
txt_page(823,301,1).
txt_page(425,301,1).
txt_page(900,301,1).
txt_page(955,301,1).
txt_page(24,301,1).
txt_page(889,301,2).
txt_page(804,301,1).
txt_page(190,301,1).
txt_page(808,301,1).
txt_page(485,301,1).
txt_page(867,301,1).
txt_page(539,301,1).
txt_page(137,301,1).
txt_page(429,301,1).
txt_page(410,301,1).
txt_page(336,301,1).
txt_page(448,302,1).
txt_page(36,302,1).
txt_page(963,302,1).
txt_page(560,302,1).
txt_page(294,302,1).
txt_page(525,302,1).
txt_page(559,302,1).
txt_page(170,302,1).
txt_page(577,302,1).
txt_page(120,302,1).
txt_page(848,302,1).
txt_page(117,302,1).
txt_page(637,302,1).
txt_page(521,302,1).
txt_page(814,302,1).
txt_page(207,302,1).
txt_page(108,302,1).
txt_page(699,302,1).
txt_page(409,302,1).
txt_page(707,302,1).
txt_page(235,303,1).
txt_page(764,303,1).
txt_page(333,303,1).
txt_page(547,303,1).
txt_page(213,303,1).
txt_page(656,303,1).
txt_page(216,303,1).
txt_page(300,303,1).
txt_page(5,303,1).
txt_page(780,303,1).
txt_page(950,303,1).
txt_page(895,303,1).
txt_page(398,303,1).
txt_page(864,303,1).
txt_page(772,303,1).
txt_page(311,303,1).
txt_page(720,303,1).
txt_page(693,303,1).
txt_page(579,303,1).
txt_page(494,303,1).
txt_page(212,304,1).
txt_page(200,304,1).
txt_page(80,304,1).
txt_page(107,304,1).
txt_page(609,304,1).
txt_page(343,304,1).
txt_page(7,304,1).
txt_page(737,304,1).
txt_page(744,304,1).
txt_page(731,304,1).
txt_page(712,304,1).
txt_page(438,304,1).
txt_page(44,304,1).
txt_page(148,304,1).
txt_page(163,304,1).
txt_page(367,304,1).
txt_page(728,304,1).
txt_page(504,304,1).
txt_page(29,304,1).
txt_page(670,304,1).
txt_page(49,305,1).
txt_page(275,305,1).
txt_page(966,305,1).
txt_page(434,305,1).
txt_page(94,305,1).
txt_page(529,305,1).
txt_page(421,305,1).
txt_page(264,305,1).
txt_page(622,305,1).
txt_page(967,305,1).
txt_page(115,305,1).
txt_page(167,305,1).
txt_page(71,305,1).
txt_page(799,305,1).
txt_page(511,305,1).
txt_page(747,305,1).
txt_page(370,305,1).
txt_page(894,305,1).
txt_page(272,305,1).
txt_page(964,305,1).
txt_page(22,306,1).
txt_page(161,306,1).
txt_page(500,306,1).
txt_page(10,306,1).
txt_page(621,306,1).
txt_page(125,306,1).
txt_page(339,306,1).
txt_page(680,306,1).
txt_page(175,306,1).
txt_page(457,306,1).
txt_page(498,306,1).
txt_page(228,306,1).
txt_page(877,306,1).
txt_page(176,306,1).
txt_page(490,306,1).
txt_page(396,306,1).
txt_page(307,306,1).
txt_page(670,306,1).
txt_page(629,306,1).
txt_page(834,306,1).
txt_page(94,307,1).
txt_page(949,307,1).
txt_page(730,307,1).
txt_page(22,307,1).
txt_page(404,307,1).
txt_page(258,307,2).
txt_page(729,307,1).
txt_page(837,307,1).
txt_page(166,307,1).
txt_page(880,307,1).
txt_page(44,307,1).
txt_page(514,307,1).
txt_page(472,307,1).
txt_page(722,307,1).
txt_page(571,307,1).
txt_page(115,307,1).
txt_page(63,307,1).
txt_page(10,307,1).
txt_page(709,307,1).
txt_page(851,308,1).
txt_page(218,308,1).
txt_page(899,308,1).
txt_page(392,308,1).
txt_page(873,308,1).
txt_page(497,308,1).
txt_page(455,308,1).
txt_page(707,308,1).
txt_page(89,308,1).
txt_page(859,308,1).
txt_page(714,308,1).
txt_page(965,308,1).
txt_page(788,308,1).
txt_page(942,308,1).
txt_page(347,308,1).
txt_page(228,308,1).
txt_page(764,308,1).
txt_page(716,308,1).
txt_page(342,308,1).
txt_page(555,308,1).
txt_page(810,309,1).
txt_page(948,309,1).
txt_page(908,309,1).
txt_page(872,309,1).
txt_page(320,309,1).
txt_page(660,309,1).
txt_page(201,309,1).
txt_page(583,309,1).
txt_page(355,309,1).
txt_page(627,309,1).
txt_page(883,309,1).
txt_page(904,309,1).
txt_page(112,309,1).
txt_page(723,309,1).
txt_page(150,309,1).
txt_page(573,309,1).
txt_page(651,309,1).
txt_page(379,309,1).
txt_page(878,309,1).
txt_page(376,309,1).
txt_page(185,310,1).
txt_page(836,310,1).
txt_page(379,310,1).
txt_page(201,310,1).
txt_page(761,310,1).
txt_page(911,310,1).
txt_page(359,310,1).
txt_page(489,310,1).
txt_page(624,310,1).
txt_page(232,310,1).
txt_page(98,310,1).
txt_page(894,310,1).
txt_page(47,310,1).
txt_page(903,310,1).
txt_page(410,310,1).
txt_page(600,310,1).
txt_page(583,310,1).
txt_page(354,310,1).
txt_page(313,310,1).
txt_page(175,310,1).
txt_page(147,311,1).
txt_page(755,311,1).
txt_page(188,311,1).
txt_page(320,311,1).
txt_page(765,311,1).
txt_page(124,311,1).
txt_page(261,311,1).
txt_page(819,311,1).
txt_page(514,311,1).
txt_page(272,311,1).
txt_page(788,311,1).
txt_page(318,311,1).
txt_page(729,311,1).
txt_page(749,311,1).
txt_page(142,311,1).
txt_page(602,311,1).
txt_page(369,311,1).
txt_page(420,311,1).
txt_page(557,311,1).
txt_page(534,311,1).
txt_page(478,312,1).
txt_page(557,312,1).
txt_page(906,312,1).
txt_page(421,312,1).
txt_page(212,312,1).
txt_page(677,312,1).
txt_page(71,312,1).
txt_page(738,312,1).
txt_page(203,312,1).
txt_page(820,312,1).
txt_page(746,312,1).
txt_page(808,312,1).
txt_page(678,312,1).
txt_page(772,312,1).
txt_page(442,312,1).
txt_page(119,312,1).
txt_page(713,312,1).
txt_page(578,312,1).
txt_page(786,312,2).
txt_page(105,313,1).
txt_page(626,313,1).
txt_page(385,313,1).
txt_page(149,313,1).
txt_page(461,313,1).
txt_page(859,313,1).
txt_page(919,313,1).
txt_page(286,313,1).
txt_page(301,313,1).
txt_page(26,313,1).
txt_page(776,313,1).
txt_page(839,313,1).
txt_page(69,313,1).
txt_page(205,313,1).
txt_page(901,313,1).
txt_page(802,313,2).
txt_page(135,313,1).
txt_page(179,313,1).
txt_page(692,313,1).
txt_page(374,314,1).
txt_page(41,314,1).
txt_page(894,314,1).
txt_page(317,314,1).
txt_page(216,314,1).
txt_page(834,314,1).
txt_page(52,314,1).
txt_page(461,314,1).
txt_page(779,314,1).
txt_page(192,314,1).
txt_page(725,314,1).
txt_page(393,314,1).
txt_page(746,314,1).
txt_page(455,314,1).
txt_page(95,314,1).
txt_page(114,314,1).
txt_page(871,314,1).
txt_page(214,314,1).
txt_page(325,314,1).
txt_page(380,314,1).
txt_page(8,315,1).
txt_page(739,315,1).
txt_page(35,315,1).
txt_page(331,315,1).
txt_page(71,315,1).
txt_page(300,315,1).
txt_page(361,315,1).
txt_page(128,315,1).
txt_page(673,315,1).
txt_page(734,315,1).
txt_page(250,315,1).
txt_page(871,315,1).
txt_page(449,315,1).
txt_page(139,315,1).
txt_page(12,315,1).
txt_page(579,315,1).
txt_page(82,315,1).
txt_page(61,315,1).
txt_page(921,315,1).
txt_page(919,315,1).
txt_page(97,316,1).
txt_page(661,316,1).
txt_page(147,316,1).
txt_page(678,316,1).
txt_page(288,316,1).
txt_page(141,316,1).
txt_page(887,316,1).
txt_page(850,316,1).
txt_page(252,316,1).
txt_page(596,316,1).
txt_page(699,316,1).
txt_page(687,316,1).
txt_page(857,316,1).
txt_page(96,316,1).
txt_page(454,316,1).
txt_page(541,316,1).
txt_page(833,316,1).
txt_page(530,316,1).
txt_page(380,316,1).
txt_page(832,316,1).
txt_page(376,317,1).
txt_page(170,317,1).
txt_page(645,317,1).
txt_page(217,317,1).
txt_page(472,317,1).
txt_page(215,317,1).
txt_page(167,317,1).
txt_page(9,317,1).
txt_page(766,317,1).
txt_page(323,317,1).
txt_page(568,317,1).
txt_page(746,317,1).
txt_page(396,317,1).
txt_page(20,317,1).
txt_page(551,317,1).
txt_page(524,317,1).
txt_page(322,317,1).
txt_page(253,317,1).
txt_page(531,317,1).
txt_page(542,317,1).
txt_page(423,318,1).
txt_page(725,318,1).
txt_page(300,318,1).
txt_page(42,318,1).
txt_page(965,318,1).
txt_page(826,318,1).
txt_page(933,318,1).
txt_page(475,318,1).
txt_page(91,318,1).
txt_page(226,318,1).
txt_page(183,318,1).
txt_page(666,318,1).
txt_page(407,318,1).
txt_page(904,318,1).
txt_page(962,318,1).
txt_page(400,318,1).
txt_page(849,318,1).
txt_page(574,318,1).
txt_page(208,318,1).
txt_page(634,318,1).
txt_page(597,319,1).
txt_page(73,319,1).
txt_page(269,319,1).
txt_page(735,319,1).
txt_page(763,319,1).
txt_page(117,319,1).
txt_page(96,319,1).
txt_page(374,319,1).
txt_page(101,319,1).
txt_page(727,319,1).
txt_page(785,319,1).
txt_page(565,319,1).
txt_page(458,319,1).
txt_page(576,319,1).
txt_page(100,319,1).
txt_page(890,319,1).
txt_page(571,319,1).
txt_page(641,319,1).
txt_page(442,319,1).
txt_page(106,319,1).
txt_page(476,320,1).
txt_page(478,320,1).
txt_page(360,320,1).
txt_page(82,320,1).
txt_page(252,320,1).
txt_page(47,320,1).
txt_page(661,320,1).
txt_page(228,320,1).
txt_page(219,320,1).
txt_page(43,320,1).
txt_page(201,320,1).
txt_page(845,320,1).
txt_page(164,320,1).
txt_page(923,320,1).
txt_page(837,320,1).
txt_page(404,320,1).
txt_page(894,320,1).
txt_page(210,320,1).
txt_page(172,320,1).
txt_page(831,320,1).
txt_page(891,321,1).
txt_page(902,321,1).
txt_page(845,321,1).
txt_page(769,321,1).
txt_page(176,321,1).
txt_page(795,321,1).
txt_page(526,321,1).
txt_page(148,321,1).
txt_page(54,321,1).
txt_page(422,321,1).
txt_page(745,321,1).
txt_page(403,321,1).
txt_page(935,321,1).
txt_page(592,321,1).
txt_page(271,321,1).
txt_page(18,321,1).
txt_page(134,321,1).
txt_page(166,321,1).
txt_page(96,321,1).
txt_page(924,321,1).
txt_page(860,322,1).
txt_page(862,322,1).
txt_page(831,322,1).
txt_page(899,322,1).
txt_page(465,322,1).
txt_page(43,322,1).
txt_page(657,322,1).
txt_page(330,322,1).
txt_page(65,322,1).
txt_page(408,322,1).
txt_page(50,322,1).
txt_page(516,322,1).
txt_page(890,322,1).
txt_page(889,322,1).
txt_page(762,322,1).
txt_page(353,322,1).
txt_page(327,322,1).
txt_page(611,322,1).
txt_page(242,322,1).
txt_page(742,322,1).
txt_page(748,323,1).
txt_page(312,323,1).
txt_page(354,323,1).
txt_page(352,323,1).
txt_page(946,323,1).
txt_page(901,323,1).
txt_page(146,323,1).
txt_page(38,323,1).
txt_page(373,323,1).
txt_page(371,323,1).
txt_page(147,323,1).
txt_page(790,323,1).
txt_page(383,323,1).
txt_page(698,323,1).
txt_page(243,323,1).
txt_page(277,323,1).
txt_page(224,323,1).
txt_page(714,323,1).
txt_page(669,323,1).
txt_page(769,323,1).
txt_page(410,324,1).
txt_page(100,324,1).
txt_page(250,324,1).
txt_page(424,324,1).
txt_page(588,324,1).
txt_page(884,324,1).
txt_page(796,324,1).
txt_page(956,324,1).
txt_page(910,324,1).
txt_page(527,324,1).
txt_page(370,324,1).
txt_page(731,324,1).
txt_page(327,324,1).
txt_page(650,324,1).
txt_page(502,324,1).
txt_page(198,324,1).
txt_page(567,324,1).
txt_page(795,324,1).
txt_page(501,324,1).
txt_page(318,324,1).
txt_page(754,325,1).
txt_page(103,325,1).
txt_page(932,325,1).
txt_page(71,325,1).
txt_page(199,325,1).
txt_page(335,325,1).
txt_page(434,325,1).
txt_page(958,325,1).
txt_page(601,325,1).
txt_page(861,325,1).
txt_page(328,325,1).
txt_page(916,325,1).
txt_page(482,325,1).
txt_page(559,325,1).
txt_page(555,325,1).
txt_page(760,325,1).
txt_page(633,325,1).
txt_page(854,325,1).
txt_page(94,325,1).
txt_page(829,325,1).
txt_page(70,326,1).
txt_page(598,326,1).
txt_page(102,326,1).
txt_page(189,326,1).
txt_page(684,326,1).
txt_page(489,326,1).
txt_page(397,326,1).
txt_page(715,326,1).
txt_page(696,326,1).
txt_page(216,326,1).
txt_page(339,326,1).
txt_page(224,326,1).
txt_page(407,326,1).
txt_page(643,326,1).
txt_page(88,326,1).
txt_page(40,326,1).
txt_page(273,326,1).
txt_page(405,326,1).
txt_page(439,326,1).
txt_page(610,326,1).
txt_page(374,327,1).
txt_page(712,327,1).
txt_page(501,327,1).
txt_page(255,327,1).
txt_page(875,327,1).
txt_page(593,327,1).
txt_page(937,327,1).
txt_page(403,327,1).
txt_page(443,327,1).
txt_page(96,327,1).
txt_page(11,327,1).
txt_page(894,327,1).
txt_page(802,327,1).
txt_page(595,327,1).
txt_page(661,327,1).
txt_page(198,327,1).
txt_page(690,327,1).
txt_page(957,327,1).
txt_page(638,327,1).
txt_page(947,327,1).
txt_page(132,328,1).
txt_page(236,328,1).
txt_page(424,328,1).
txt_page(724,328,1).
txt_page(598,328,1).
txt_page(918,328,1).
txt_page(774,328,1).
txt_page(670,328,1).
txt_page(5,328,1).
txt_page(255,328,1).
txt_page(44,328,1).
txt_page(405,328,1).
txt_page(448,328,1).
txt_page(47,328,1).
txt_page(256,328,1).
txt_page(546,328,1).
txt_page(115,328,1).
txt_page(243,328,1).
txt_page(36,328,1).
txt_page(462,328,1).
txt_page(959,329,1).
txt_page(80,329,1).
txt_page(952,329,1).
txt_page(923,329,1).
txt_page(811,329,1).
txt_page(585,329,1).
txt_page(646,329,1).
txt_page(237,329,1).
txt_page(475,329,1).
txt_page(399,329,1).
txt_page(740,329,1).
txt_page(696,329,1).
txt_page(592,329,1).
txt_page(318,329,1).
txt_page(485,329,1).
txt_page(493,329,1).
txt_page(293,329,1).
txt_page(709,329,1).
txt_page(851,329,1).
txt_page(271,329,1).
txt_page(891,330,1).
txt_page(377,330,1).
txt_page(936,330,1).
txt_page(764,330,1).
txt_page(55,330,1).
txt_page(581,330,1).
txt_page(226,330,1).
txt_page(683,330,1).
txt_page(439,330,1).
txt_page(414,330,1).
txt_page(788,330,1).
txt_page(385,330,1).
txt_page(386,330,1).
txt_page(953,330,1).
txt_page(824,330,2).
txt_page(827,330,1).
txt_page(499,330,1).
txt_page(691,330,1).
txt_page(271,330,1).
txt_page(286,331,1).
txt_page(162,331,1).
txt_page(960,331,1).
txt_page(788,331,1).
txt_page(480,331,1).
txt_page(531,331,1).
txt_page(699,331,1).
txt_page(439,331,1).
txt_page(98,331,1).
txt_page(509,331,2).
txt_page(492,331,1).
txt_page(786,331,1).
txt_page(187,331,1).
txt_page(948,331,1).
txt_page(461,331,1).
txt_page(734,331,1).
txt_page(598,331,1).
txt_page(64,331,1).
txt_page(99,331,1).
txt_page(470,332,1).
txt_page(53,332,1).
txt_page(723,332,1).
txt_page(569,332,1).
txt_page(851,332,1).
txt_page(822,332,1).
txt_page(360,332,1).
txt_page(758,332,1).
txt_page(912,332,1).
txt_page(114,332,1).
txt_page(492,332,1).
txt_page(268,332,1).
txt_page(197,332,1).
txt_page(229,332,1).
txt_page(159,332,1).
txt_page(597,332,1).
txt_page(388,332,1).
txt_page(136,332,1).
txt_page(942,332,1).
txt_page(563,332,1).
txt_page(51,333,1).
txt_page(504,333,1).
txt_page(734,333,1).
txt_page(701,333,1).
txt_page(593,333,1).
txt_page(347,333,1).
txt_page(275,333,1).
txt_page(249,333,1).
txt_page(783,333,1).
txt_page(136,333,1).
txt_page(401,333,1).
txt_page(461,333,1).
txt_page(169,333,1).
txt_page(515,333,1).
txt_page(447,333,1).
txt_page(597,333,1).
txt_page(325,333,1).
txt_page(524,333,1).
txt_page(516,333,1).
txt_page(560,333,1).
txt_page(712,334,1).
txt_page(928,334,1).
txt_page(289,334,1).
txt_page(534,334,1).
txt_page(220,334,1).
txt_page(657,334,1).
txt_page(869,334,1).
txt_page(172,334,1).
txt_page(790,334,1).
txt_page(875,334,1).
txt_page(850,334,1).
txt_page(640,334,1).
txt_page(44,334,1).
txt_page(28,334,1).
txt_page(867,334,1).
txt_page(633,334,1).
txt_page(624,334,1).
txt_page(429,334,1).
txt_page(777,334,1).
txt_page(30,334,1).
txt_page(650,335,1).
txt_page(737,335,1).
txt_page(965,335,1).
txt_page(808,335,1).
txt_page(938,335,1).
txt_page(95,335,1).
txt_page(284,335,1).
txt_page(909,335,1).
txt_page(317,335,1).
txt_page(788,335,1).
txt_page(890,335,1).
txt_page(275,335,2).
txt_page(708,335,1).
txt_page(533,335,1).
txt_page(454,335,1).
txt_page(296,335,1).
txt_page(603,335,1).
txt_page(728,335,1).
txt_page(545,335,1).
txt_page(282,336,1).
txt_page(99,336,1).
txt_page(146,336,1).
txt_page(753,336,1).
txt_page(811,336,1).
txt_page(55,336,1).
txt_page(567,336,1).
txt_page(166,336,1).
txt_page(666,336,1).
txt_page(153,336,1).
txt_page(794,336,1).
txt_page(549,336,1).
txt_page(720,336,1).
txt_page(274,336,1).
txt_page(302,336,1).
txt_page(664,336,1).
txt_page(572,336,1).
txt_page(589,336,1).
txt_page(568,336,1).
txt_page(97,336,1).
txt_page(523,337,1).
txt_page(129,337,1).
txt_page(256,337,1).
txt_page(790,337,1).
txt_page(948,337,1).
txt_page(569,337,1).
txt_page(203,337,1).
txt_page(953,337,1).
txt_page(681,337,1).
txt_page(260,337,1).
txt_page(783,337,1).
txt_page(901,337,1).
txt_page(548,337,1).
txt_page(228,337,1).
txt_page(749,337,1).
txt_page(797,337,1).
txt_page(772,337,1).
txt_page(368,337,1).
txt_page(849,337,1).
txt_page(583,337,1).
txt_page(382,338,1).
txt_page(812,338,1).
txt_page(483,338,1).
txt_page(702,338,1).
txt_page(120,338,1).
txt_page(343,338,1).
txt_page(230,338,1).
txt_page(80,338,1).
txt_page(286,338,1).
txt_page(752,338,1).
txt_page(172,338,1).
txt_page(636,338,1).
txt_page(174,338,1).
txt_page(640,338,1).
txt_page(654,338,1).
txt_page(832,338,1).
txt_page(59,338,1).
txt_page(864,338,1).
txt_page(78,338,1).
txt_page(524,338,1).
txt_page(285,339,1).
txt_page(368,339,1).
txt_page(174,339,1).
txt_page(740,339,1).
txt_page(775,339,1).
txt_page(48,339,1).
txt_page(835,339,1).
txt_page(69,339,1).
txt_page(274,339,1).
txt_page(676,339,1).
txt_page(820,339,1).
txt_page(58,339,1).
txt_page(498,339,1).
txt_page(588,339,1).
txt_page(721,339,1).
txt_page(200,339,1).
txt_page(760,339,1).
txt_page(400,339,1).
txt_page(423,339,1).
txt_page(288,339,1).
txt_page(319,340,1).
txt_page(213,340,1).
txt_page(240,340,1).
txt_page(401,340,1).
txt_page(248,340,1).
txt_page(936,340,1).
txt_page(405,340,1).
txt_page(748,340,1).
txt_page(201,340,1).
txt_page(889,340,1).
txt_page(791,340,1).
txt_page(779,340,1).
txt_page(241,340,1).
txt_page(773,340,1).
txt_page(903,340,1).
txt_page(800,340,1).
txt_page(196,340,1).
txt_page(652,340,1).
txt_page(929,340,1).
txt_page(445,340,1).
txt_page(670,341,1).
txt_page(336,341,1).
txt_page(273,341,1).
txt_page(314,341,1).
txt_page(458,341,1).
txt_page(705,341,1).
txt_page(914,341,1).
txt_page(391,341,1).
txt_page(373,341,1).
txt_page(854,341,1).
txt_page(421,341,1).
txt_page(154,341,1).
txt_page(335,341,1).
txt_page(783,341,1).
txt_page(645,341,1).
txt_page(844,341,1).
txt_page(774,341,1).
txt_page(413,341,1).
txt_page(566,341,1).
txt_page(712,341,1).
txt_page(688,342,1).
txt_page(954,342,1).
txt_page(24,342,1).
txt_page(727,342,1).
txt_page(588,342,1).
txt_page(429,342,1).
txt_page(924,342,1).
txt_page(3,342,1).
txt_page(838,342,1).
txt_page(309,342,1).
txt_page(50,342,1).
txt_page(331,342,1).
txt_page(26,342,1).
txt_page(610,342,1).
txt_page(874,342,1).
txt_page(723,342,1).
txt_page(740,342,1).
txt_page(630,342,1).
txt_page(62,342,1).
txt_page(828,342,1).
txt_page(228,343,1).
txt_page(110,343,1).
txt_page(855,343,1).
txt_page(530,343,1).
txt_page(275,343,1).
txt_page(814,343,1).
txt_page(305,343,1).
txt_page(395,343,1).
txt_page(559,343,1).
txt_page(553,343,1).
txt_page(903,343,1).
txt_page(1,343,1).
txt_page(706,343,1).
txt_page(520,343,1).
txt_page(949,343,1).
txt_page(586,343,1).
txt_page(243,343,1).
txt_page(319,343,1).
txt_page(219,343,1).
txt_page(556,343,1).
txt_page(293,344,1).
txt_page(45,344,1).
txt_page(808,344,1).
txt_page(202,344,1).
txt_page(463,344,1).
txt_page(282,344,1).
txt_page(626,344,1).
txt_page(676,344,1).
txt_page(668,344,1).
txt_page(238,344,1).
txt_page(592,344,1).
txt_page(759,344,1).
txt_page(878,344,1).
txt_page(575,344,1).
txt_page(469,344,1).
txt_page(534,344,1).
txt_page(189,344,1).
txt_page(512,344,1).
txt_page(882,344,1).
txt_page(650,344,1).
txt_page(127,345,1).
txt_page(298,345,1).
txt_page(119,345,1).
txt_page(284,345,1).
txt_page(117,345,1).
txt_page(640,345,1).
txt_page(449,345,1).
txt_page(116,345,1).
txt_page(748,345,1).
txt_page(81,345,1).
txt_page(37,345,1).
txt_page(373,345,1).
txt_page(194,345,1).
txt_page(95,345,1).
txt_page(687,345,1).
txt_page(864,345,1).
txt_page(517,345,1).
txt_page(587,345,1).
txt_page(481,345,1).
txt_page(44,345,1).
txt_page(894,346,1).
txt_page(714,346,1).
txt_page(927,346,1).
txt_page(240,346,1).
txt_page(946,346,1).
txt_page(869,346,1).
txt_page(103,346,1).
txt_page(761,346,1).
txt_page(518,346,1).
txt_page(292,346,1).
txt_page(788,346,1).
txt_page(688,346,1).
txt_page(556,346,1).
txt_page(17,346,1).
txt_page(740,346,1).
txt_page(477,346,1).
txt_page(356,346,1).
txt_page(531,346,1).
txt_page(192,346,1).
txt_page(687,346,1).
txt_page(447,347,1).
txt_page(599,347,1).
txt_page(417,347,1).
txt_page(841,347,1).
txt_page(291,347,1).
txt_page(25,347,1).
txt_page(710,347,2).
txt_page(97,347,1).
txt_page(590,347,1).
txt_page(351,347,1).
txt_page(160,347,1).
txt_page(387,347,1).
txt_page(791,347,1).
txt_page(513,347,1).
txt_page(366,347,1).
txt_page(412,347,1).
txt_page(870,347,1).
txt_page(549,347,1).
txt_page(354,347,1).
txt_page(394,348,1).
txt_page(413,348,1).
txt_page(304,348,1).
txt_page(300,348,1).
txt_page(646,348,1).
txt_page(29,348,1).
txt_page(737,348,1).
txt_page(101,348,1).
txt_page(68,348,1).
txt_page(388,348,1).
txt_page(397,348,1).
txt_page(895,348,1).
txt_page(792,348,1).
txt_page(363,348,1).
txt_page(674,348,1).
txt_page(336,348,1).
txt_page(876,348,1).
txt_page(27,348,1).
txt_page(48,348,1).
txt_page(213,348,1).
txt_page(146,349,1).
txt_page(914,349,1).
txt_page(496,349,1).
txt_page(968,349,1).
txt_page(55,349,1).
txt_page(445,349,1).
txt_page(347,349,1).
txt_page(212,349,1).
txt_page(174,349,1).
txt_page(709,349,1).
txt_page(569,349,1).
txt_page(43,349,1).
txt_page(258,349,1).
txt_page(433,349,1).
txt_page(893,349,1).
txt_page(60,349,1).
txt_page(637,349,1).
txt_page(949,349,1).
txt_page(440,349,1).
txt_page(17,349,1).
txt_page(518,350,1).
txt_page(261,350,1).
txt_page(1,350,1).
txt_page(756,350,1).
txt_page(374,350,1).
txt_page(889,350,1).
txt_page(470,350,1).
txt_page(311,350,1).
txt_page(541,350,1).
txt_page(227,350,1).
txt_page(181,350,1).
txt_page(594,350,1).
txt_page(375,350,1).
txt_page(822,350,1).
txt_page(52,350,1).
txt_page(615,350,1).
txt_page(854,350,1).
txt_page(890,350,1).
txt_page(477,350,1).
txt_page(646,350,1).
txt_page(282,351,1).
txt_page(15,351,1).
txt_page(424,351,1).
txt_page(10,351,1).
txt_page(560,351,1).
txt_page(905,351,1).
txt_page(132,351,1).
txt_page(635,351,1).
txt_page(383,351,1).
txt_page(524,351,1).
txt_page(62,351,1).
txt_page(39,351,1).
txt_page(906,351,1).
txt_page(333,351,1).
txt_page(799,351,1).
txt_page(416,351,1).
txt_page(888,351,1).
txt_page(872,351,1).
txt_page(577,351,1).
txt_page(871,351,1).
txt_page(304,352,1).
txt_page(55,352,1).
txt_page(902,352,1).
txt_page(26,352,1).
txt_page(409,352,1).
txt_page(742,352,1).
txt_page(942,352,1).
txt_page(896,352,1).
txt_page(794,352,1).
txt_page(418,352,1).
txt_page(759,352,1).
txt_page(4,352,1).
txt_page(265,352,1).
txt_page(789,352,1).
txt_page(717,352,1).
txt_page(818,352,1).
txt_page(907,352,1).
txt_page(58,352,1).
txt_page(198,352,1).
txt_page(860,352,1).
txt_page(202,353,1).
txt_page(37,353,1).
txt_page(914,353,1).
txt_page(704,353,1).
txt_page(543,353,1).
txt_page(172,353,1).
txt_page(577,353,1).
txt_page(660,353,1).
txt_page(820,353,1).
txt_page(371,353,1).
txt_page(361,353,1).
txt_page(212,353,1).
txt_page(161,353,1).
txt_page(459,353,1).
txt_page(56,353,1).
txt_page(475,353,1).
txt_page(954,353,1).
txt_page(78,353,1).
txt_page(203,353,1).
txt_page(625,353,1).
txt_page(953,354,1).
txt_page(88,354,1).
txt_page(501,354,1).
txt_page(510,354,1).
txt_page(65,354,1).
txt_page(931,354,1).
txt_page(133,354,1).
txt_page(131,354,1).
txt_page(460,354,1).
txt_page(358,354,1).
txt_page(872,354,1).
txt_page(54,354,1).
txt_page(427,354,1).
txt_page(116,354,1).
txt_page(959,354,1).
txt_page(619,354,1).
txt_page(343,354,1).
txt_page(859,354,1).
txt_page(846,354,2).
txt_page(291,355,1).
txt_page(695,355,1).
txt_page(151,355,1).
txt_page(131,355,1).
txt_page(109,355,1).
txt_page(525,355,1).
txt_page(528,355,1).
txt_page(472,355,1).
txt_page(623,355,1).
txt_page(477,355,1).
txt_page(292,355,1).
txt_page(664,355,1).
txt_page(700,355,1).
txt_page(416,355,1).
txt_page(411,355,1).
txt_page(758,355,1).
txt_page(158,355,1).
txt_page(574,355,1).
txt_page(72,355,1).
txt_page(256,355,1).
txt_page(677,356,1).
txt_page(20,356,1).
txt_page(130,356,1).
txt_page(431,356,1).
txt_page(224,356,1).
txt_page(624,356,1).
txt_page(961,356,1).
txt_page(250,356,1).
txt_page(101,356,2).
txt_page(741,356,1).
txt_page(567,356,1).
txt_page(216,356,1).
txt_page(572,356,1).
txt_page(940,356,1).
txt_page(748,356,1).
txt_page(342,356,1).
txt_page(674,356,1).
txt_page(731,356,1).
txt_page(442,356,1).
txt_page(526,357,1).
txt_page(496,357,1).
txt_page(61,357,1).
txt_page(746,357,1).
txt_page(6,357,1).
txt_page(182,357,1).
txt_page(530,357,1).
txt_page(865,357,1).
txt_page(353,357,1).
txt_page(253,357,1).
txt_page(867,357,1).
txt_page(802,357,1).
txt_page(361,357,1).
txt_page(717,357,1).
txt_page(815,357,1).
txt_page(495,357,1).
txt_page(607,357,1).
txt_page(639,357,1).
txt_page(234,357,1).
txt_page(805,357,1).
txt_page(821,358,1).
txt_page(334,358,1).
txt_page(502,358,1).
txt_page(682,358,1).
txt_page(387,358,1).
txt_page(552,358,1).
txt_page(138,358,1).
txt_page(944,358,1).
txt_page(665,358,1).
txt_page(586,358,1).
txt_page(318,358,1).
txt_page(746,358,1).
txt_page(653,358,1).
txt_page(570,358,1).
txt_page(798,358,1).
txt_page(219,358,1).
txt_page(622,358,1).
txt_page(744,358,1).
txt_page(457,358,1).
txt_page(467,358,1).
txt_page(671,359,1).
txt_page(43,359,1).
txt_page(585,359,1).
txt_page(244,359,1).
txt_page(64,359,1).
txt_page(699,359,1).
txt_page(659,359,1).
txt_page(329,359,1).
txt_page(529,359,1).
txt_page(839,359,1).
txt_page(731,359,1).
txt_page(487,359,1).
txt_page(737,359,1).
txt_page(662,359,1).
txt_page(11,359,1).
txt_page(326,359,1).
txt_page(848,359,1).
txt_page(637,359,1).
txt_page(325,359,1).
txt_page(171,359,1).
txt_page(611,360,1).
txt_page(109,360,1).
txt_page(323,360,1).
txt_page(142,360,1).
txt_page(350,360,1).
txt_page(697,360,1).
txt_page(265,360,1).
txt_page(812,360,1).
txt_page(865,360,1).
txt_page(406,360,1).
txt_page(672,360,2).
txt_page(909,360,1).
txt_page(218,360,1).
txt_page(900,360,1).
txt_page(36,360,1).
txt_page(729,360,1).
txt_page(234,360,1).
txt_page(572,360,1).
txt_page(421,360,1).
txt_page(77,361,1).
txt_page(382,361,1).
txt_page(876,361,1).
txt_page(629,361,1).
txt_page(413,361,1).
txt_page(819,361,1).
txt_page(143,361,1).
txt_page(63,361,1).
txt_page(253,361,1).
txt_page(336,361,1).
txt_page(505,361,1).
txt_page(166,361,1).
txt_page(708,361,1).
txt_page(704,361,1).
txt_page(144,361,1).
txt_page(775,361,1).
txt_page(964,361,1).
txt_page(534,361,1).
txt_page(507,361,1).
txt_page(347,361,1).
txt_page(562,362,1).
txt_page(911,362,1).
txt_page(683,362,1).
txt_page(793,362,1).
txt_page(350,362,1).
txt_page(191,362,1).
txt_page(500,362,1).
txt_page(180,362,1).
txt_page(807,362,1).
txt_page(714,362,1).
txt_page(581,362,1).
txt_page(939,362,1).
txt_page(25,362,1).
txt_page(554,362,1).
txt_page(802,362,1).
txt_page(586,362,1).
txt_page(571,362,1).
txt_page(49,362,1).
txt_page(918,362,1).
txt_page(58,362,1).
txt_page(431,363,1).
txt_page(655,363,1).
txt_page(354,363,1).
txt_page(294,363,1).
txt_page(573,363,1).
txt_page(186,363,1).
txt_page(565,363,1).
txt_page(362,363,1).
txt_page(968,363,1).
txt_page(938,363,1).
txt_page(524,363,1).
txt_page(18,363,1).
txt_page(342,363,1).
txt_page(465,363,1).
txt_page(60,363,1).
txt_page(11,363,1).
txt_page(25,363,1).
txt_page(696,363,1).
txt_page(940,363,1).
txt_page(266,363,1).
txt_page(432,364,1).
txt_page(90,364,1).
txt_page(457,364,1).
txt_page(532,364,1).
txt_page(363,364,1).
txt_page(275,364,1).
txt_page(912,364,1).
txt_page(201,364,1).
txt_page(129,364,1).
txt_page(803,364,1).
txt_page(714,364,1).
txt_page(511,364,1).
txt_page(184,364,1).
txt_page(773,364,1).
txt_page(536,364,1).
txt_page(541,364,1).
txt_page(103,364,1).
txt_page(590,364,1).
txt_page(710,364,1).
txt_page(267,364,1).
txt_page(22,365,1).
txt_page(153,365,1).
txt_page(44,365,1).
txt_page(762,365,1).
txt_page(663,365,1).
txt_page(482,365,1).
txt_page(311,365,1).
txt_page(584,365,1).
txt_page(281,365,1).
txt_page(71,365,1).
txt_page(672,365,1).
txt_page(238,365,1).
txt_page(712,365,1).
txt_page(360,365,1).
txt_page(965,365,1).
txt_page(337,365,1).
txt_page(588,365,1).
txt_page(568,365,1).
txt_page(879,365,2).
txt_page(949,366,1).
txt_page(758,366,1).
txt_page(921,366,1).
txt_page(20,366,1).
txt_page(126,366,1).
txt_page(527,366,1).
txt_page(507,366,1).
txt_page(72,366,1).
txt_page(380,366,1).
txt_page(621,366,1).
txt_page(80,366,1).
txt_page(764,366,1).
txt_page(356,366,1).
txt_page(165,366,1).
txt_page(745,366,1).
txt_page(606,366,1).
txt_page(35,366,1).
txt_page(787,366,1).
txt_page(930,366,1).
txt_page(500,366,1).
txt_page(905,367,1).
txt_page(488,367,1).
txt_page(657,367,1).
txt_page(531,367,1).
txt_page(122,367,1).
txt_page(54,367,1).
txt_page(22,367,1).
txt_page(363,367,1).
txt_page(682,367,1).
txt_page(816,367,1).
txt_page(953,367,1).
txt_page(643,367,1).
txt_page(558,367,1).
txt_page(754,367,1).
txt_page(49,367,1).
txt_page(78,367,1).
txt_page(638,367,1).
txt_page(480,367,1).
txt_page(140,367,1).
txt_page(536,367,1).
txt_page(914,368,1).
txt_page(263,368,1).
txt_page(327,368,1).
txt_page(904,368,1).
txt_page(641,368,1).
txt_page(928,368,1).
txt_page(920,368,1).
txt_page(386,368,1).
txt_page(490,368,1).
txt_page(179,368,1).
txt_page(18,368,1).
txt_page(271,368,1).
txt_page(422,368,1).
txt_page(852,368,1).
txt_page(754,368,1).
txt_page(689,368,1).
txt_page(402,368,1).
txt_page(620,368,1).
txt_page(816,368,1).
txt_page(700,368,1).
txt_page(785,369,1).
txt_page(400,369,1).
txt_page(961,369,1).
txt_page(112,369,1).
txt_page(445,369,1).
txt_page(94,369,1).
txt_page(51,369,1).
txt_page(715,369,1).
txt_page(209,369,1).
txt_page(57,369,1).
txt_page(783,369,1).
txt_page(87,369,1).
txt_page(465,369,1).
txt_page(612,369,1).
txt_page(484,369,1).
txt_page(728,369,2).
txt_page(884,369,1).
txt_page(401,369,1).
txt_page(850,369,1).
txt_page(853,370,1).
txt_page(593,370,1).
txt_page(835,370,1).
txt_page(129,370,1).
txt_page(860,370,1).
txt_page(292,370,1).
txt_page(244,370,1).
txt_page(533,370,1).
txt_page(12,370,1).
txt_page(798,370,1).
txt_page(563,370,1).
txt_page(19,370,1).
txt_page(963,370,1).
txt_page(138,370,1).
txt_page(151,370,1).
txt_page(529,370,1).
txt_page(394,370,1).
txt_page(9,370,1).
txt_page(167,370,1).
txt_page(349,370,1).
txt_page(446,371,1).
txt_page(42,371,1).
txt_page(433,371,1).
txt_page(878,371,1).
txt_page(5,371,1).
txt_page(392,371,1).
txt_page(657,371,1).
txt_page(68,371,1).
txt_page(117,371,1).
txt_page(15,371,1).
txt_page(217,371,1).
txt_page(763,371,1).
txt_page(495,371,1).
txt_page(314,371,1).
txt_page(322,371,1).
txt_page(710,371,1).
txt_page(60,371,1).
txt_page(917,371,1).
txt_page(760,371,1).
txt_page(773,371,1).
txt_page(962,372,1).
txt_page(52,372,1).
txt_page(903,372,1).
txt_page(838,372,1).
txt_page(635,372,1).
txt_page(407,372,1).
txt_page(547,372,1).
txt_page(127,372,1).
txt_page(647,372,1).
txt_page(293,372,1).
txt_page(81,372,1).
txt_page(331,372,1).
txt_page(824,372,1).
txt_page(239,372,1).
txt_page(952,372,1).
txt_page(486,372,1).
txt_page(560,372,1).
txt_page(268,372,1).
txt_page(139,372,1).
txt_page(226,372,1).
txt_page(319,373,1).
txt_page(122,373,1).
txt_page(361,373,1).
txt_page(368,373,1).
txt_page(504,373,1).
txt_page(931,373,1).
txt_page(712,373,1).
txt_page(933,373,1).
txt_page(112,373,1).
txt_page(177,373,1).
txt_page(452,373,1).
txt_page(95,373,1).
txt_page(162,373,1).
txt_page(552,373,1).
txt_page(627,373,1).
txt_page(379,373,1).
txt_page(630,373,1).
txt_page(715,373,1).
txt_page(11,373,1).
txt_page(937,373,1).
txt_page(752,374,1).
txt_page(235,374,1).
txt_page(272,374,1).
txt_page(139,374,1).
txt_page(106,374,1).
txt_page(523,374,1).
txt_page(680,374,1).
txt_page(682,374,1).
txt_page(276,374,1).
txt_page(944,374,1).
txt_page(107,374,1).
txt_page(296,374,1).
txt_page(937,374,1).
txt_page(871,374,1).
txt_page(26,374,1).
txt_page(45,374,1).
txt_page(820,374,1).
txt_page(436,374,1).
txt_page(208,374,1).
txt_page(319,374,1).
txt_page(432,375,1).
txt_page(569,375,1).
txt_page(544,375,1).
txt_page(677,375,1).
txt_page(86,375,1).
txt_page(884,375,1).
txt_page(605,375,1).
txt_page(856,375,1).
txt_page(434,375,1).
txt_page(103,375,1).
txt_page(915,375,1).
txt_page(384,375,1).
txt_page(428,375,1).
txt_page(471,375,1).
txt_page(70,375,1).
txt_page(179,375,1).
txt_page(78,375,1).
txt_page(927,375,1).
txt_page(46,375,1).
txt_page(137,375,1).
txt_page(890,376,1).
txt_page(346,376,1).
txt_page(743,376,1).
txt_page(247,376,1).
txt_page(618,376,1).
txt_page(806,376,1).
txt_page(795,376,1).
txt_page(450,376,1).
txt_page(599,376,1).
txt_page(64,376,1).
txt_page(943,376,1).
txt_page(449,376,1).
txt_page(183,376,1).
txt_page(224,376,1).
txt_page(929,376,1).
txt_page(793,376,1).
txt_page(35,376,1).
txt_page(706,376,1).
txt_page(481,376,1).
txt_page(624,376,1).
txt_page(535,377,1).
txt_page(826,377,1).
txt_page(710,377,1).
txt_page(96,377,1).
txt_page(30,377,1).
txt_page(410,377,1).
txt_page(646,377,1).
txt_page(73,377,1).
txt_page(480,377,1).
txt_page(282,377,1).
txt_page(334,377,1).
txt_page(598,377,1).
txt_page(843,377,1).
txt_page(74,377,1).
txt_page(45,377,1).
txt_page(580,377,1).
txt_page(429,377,1).
txt_page(904,377,1).
txt_page(681,377,1).
txt_page(762,377,1).
txt_page(343,378,1).
txt_page(185,378,1).
txt_page(841,378,1).
txt_page(211,378,1).
txt_page(827,378,1).
txt_page(685,378,1).
txt_page(772,378,1).
txt_page(155,378,1).
txt_page(421,378,1).
txt_page(399,378,1).
txt_page(196,378,1).
txt_page(540,378,1).
txt_page(674,378,1).
txt_page(288,378,1).
txt_page(694,378,1).
txt_page(473,378,1).
txt_page(123,378,1).
txt_page(454,378,1).
txt_page(863,378,1).
txt_page(680,378,1).
txt_page(115,379,1).
txt_page(672,379,1).
txt_page(925,379,1).
txt_page(868,379,1).
txt_page(699,379,1).
txt_page(294,379,1).
txt_page(31,379,1).
txt_page(494,379,1).
txt_page(510,379,1).
txt_page(808,379,1).
txt_page(338,379,1).
txt_page(490,379,1).
txt_page(253,379,1).
txt_page(832,379,1).
txt_page(835,379,1).
txt_page(599,379,2).
txt_page(416,379,1).
txt_page(739,379,1).
txt_page(90,379,1).
txt_page(237,380,1).
txt_page(43,380,1).
txt_page(863,380,1).
txt_page(294,380,1).
txt_page(781,380,1).
txt_page(554,380,1).
txt_page(856,380,1).
txt_page(644,380,1).
txt_page(768,380,1).
txt_page(6,380,1).
txt_page(638,380,1).
txt_page(766,380,1).
txt_page(749,380,1).
txt_page(336,380,1).
txt_page(392,380,1).
txt_page(45,380,2).
txt_page(29,380,1).
txt_page(622,380,1).
txt_page(500,380,1).
txt_page(16,381,1).
txt_page(662,381,1).
txt_page(32,381,1).
txt_page(418,381,1).
txt_page(800,381,1).
txt_page(391,381,1).
txt_page(805,381,1).
txt_page(847,381,1).
txt_page(728,381,1).
txt_page(639,381,1).
txt_page(550,381,1).
txt_page(960,381,1).
txt_page(782,381,1).
txt_page(160,381,1).
txt_page(721,381,1).
txt_page(790,381,1).
txt_page(701,381,1).
txt_page(705,381,1).
txt_page(456,381,1).
txt_page(934,381,1).
txt_page(158,382,1).
txt_page(964,382,1).
txt_page(143,382,1).
txt_page(506,382,1).
txt_page(850,382,1).
txt_page(683,382,1).
txt_page(579,382,1).
txt_page(324,382,1).
txt_page(894,382,1).
txt_page(405,382,1).
txt_page(718,382,1).
txt_page(349,382,1).
txt_page(246,382,1).
txt_page(759,382,1).
txt_page(265,382,1).
txt_page(526,382,1).
txt_page(67,382,1).
txt_page(262,382,1).
txt_page(261,382,1).
txt_page(809,382,1).
txt_page(670,383,1).
txt_page(792,383,1).
txt_page(95,383,1).
txt_page(230,383,1).
txt_page(622,383,1).
txt_page(825,383,1).
txt_page(936,383,1).
txt_page(144,383,1).
txt_page(706,383,1).
txt_page(677,383,1).
txt_page(523,383,1).
txt_page(15,383,1).
txt_page(504,383,1).
txt_page(208,383,1).
txt_page(848,383,1).
txt_page(472,383,1).
txt_page(579,383,1).
txt_page(815,383,1).
txt_page(796,383,1).
txt_page(642,383,1).
txt_page(629,384,1).
txt_page(716,384,1).
txt_page(699,384,1).
txt_page(725,384,1).
txt_page(191,384,1).
txt_page(682,384,1).
txt_page(713,384,1).
txt_page(721,384,1).
txt_page(509,384,1).
txt_page(540,384,1).
txt_page(298,384,1).
txt_page(879,384,1).
txt_page(261,384,1).
txt_page(504,384,1).
txt_page(720,384,1).
txt_page(366,384,1).
txt_page(679,384,1).
txt_page(598,384,1).
txt_page(830,384,1).
txt_page(791,384,1).
txt_page(767,385,1).
txt_page(817,385,1).
txt_page(120,385,1).
txt_page(441,385,1).
txt_page(879,385,1).
txt_page(180,385,1).
txt_page(323,385,1).
txt_page(802,385,1).
txt_page(729,385,1).
txt_page(311,385,1).
txt_page(21,385,1).
txt_page(561,385,1).
txt_page(159,385,1).
txt_page(347,385,1).
txt_page(513,385,1).
txt_page(256,385,1).
txt_page(92,385,1).
txt_page(130,385,1).
txt_page(881,385,1).
txt_page(491,385,1).
txt_page(860,386,1).
txt_page(696,386,1).
txt_page(612,386,1).
txt_page(417,386,1).
txt_page(547,386,1).
txt_page(862,386,1).
txt_page(388,386,1).
txt_page(918,386,1).
txt_page(573,386,1).
txt_page(605,386,1).
txt_page(49,386,1).
txt_page(532,386,1).
txt_page(600,386,1).
txt_page(327,386,1).
txt_page(338,386,1).
txt_page(146,386,1).
txt_page(229,386,1).
txt_page(713,386,1).
txt_page(512,386,1).
txt_page(710,386,1).
txt_page(343,387,1).
txt_page(188,387,1).
txt_page(840,387,1).
txt_page(799,387,1).
txt_page(658,387,1).
txt_page(287,387,1).
txt_page(918,387,1).
txt_page(846,387,1).
txt_page(574,387,1).
txt_page(512,387,1).
txt_page(227,387,1).
txt_page(588,387,1).
txt_page(917,387,1).
txt_page(357,387,1).
txt_page(728,387,1).
txt_page(94,387,1).
txt_page(892,387,1).
txt_page(593,387,1).
txt_page(631,387,1).
txt_page(434,387,1).
txt_page(231,388,1).
txt_page(693,388,1).
txt_page(648,388,1).
txt_page(271,388,1).
txt_page(456,388,1).
txt_page(832,388,1).
txt_page(470,388,1).
txt_page(12,388,1).
txt_page(685,388,1).
txt_page(218,388,1).
txt_page(212,388,1).
txt_page(935,388,1).
txt_page(621,388,1).
txt_page(520,388,1).
txt_page(873,388,1).
txt_page(425,388,1).
txt_page(17,388,1).
txt_page(831,388,1).
txt_page(510,388,1).
txt_page(857,388,1).
txt_page(827,389,1).
txt_page(172,389,1).
txt_page(784,389,1).
txt_page(539,389,1).
txt_page(602,389,1).
txt_page(855,389,1).
txt_page(131,389,1).
txt_page(230,389,1).
txt_page(581,389,1).
txt_page(236,389,1).
txt_page(598,389,1).
txt_page(955,389,1).
txt_page(502,389,1).
txt_page(737,389,1).
txt_page(795,389,1).
txt_page(765,389,1).
txt_page(707,389,1).
txt_page(405,389,1).
txt_page(592,389,1).
txt_page(370,389,1).
txt_page(919,390,1).
txt_page(406,390,1).
txt_page(886,390,1).
txt_page(855,390,1).
txt_page(959,390,1).
txt_page(81,390,1).
txt_page(647,390,1).
txt_page(403,390,1).
txt_page(148,390,1).
txt_page(249,390,1).
txt_page(369,390,1).
txt_page(215,390,1).
txt_page(887,390,1).
txt_page(624,390,1).
txt_page(58,390,1).
txt_page(798,390,2).
txt_page(661,390,1).
txt_page(680,390,1).
txt_page(674,390,1).
txt_page(552,391,1).
txt_page(271,391,1).
txt_page(958,391,1).
txt_page(79,391,1).
txt_page(334,391,1).
txt_page(794,391,1).
txt_page(66,391,1).
txt_page(56,391,1).
txt_page(255,391,1).
txt_page(59,391,1).
txt_page(75,391,1).
txt_page(677,391,1).
txt_page(219,391,1).
txt_page(306,391,1).
txt_page(483,391,1).
txt_page(294,391,1).
txt_page(220,391,1).
txt_page(826,391,1).
txt_page(260,391,1).
txt_page(755,391,1).
txt_page(453,392,1).
txt_page(688,392,1).
txt_page(463,392,1).
txt_page(602,392,1).
txt_page(856,392,1).
txt_page(923,392,1).
txt_page(929,392,1).
txt_page(22,392,1).
txt_page(629,392,1).
txt_page(689,392,1).
txt_page(181,392,1).
txt_page(785,392,1).
txt_page(930,392,1).
txt_page(460,392,1).
txt_page(711,392,1).
txt_page(116,392,1).
txt_page(682,392,1).
txt_page(605,392,1).
txt_page(389,392,1).
txt_page(813,392,1).
txt_page(28,393,1).
txt_page(928,393,1).
txt_page(265,393,1).
txt_page(927,393,1).
txt_page(236,393,1).
txt_page(392,393,1).
txt_page(918,393,1).
txt_page(695,393,1).
txt_page(799,393,1).
txt_page(272,393,1).
txt_page(816,393,1).
txt_page(535,393,1).
txt_page(201,393,1).
txt_page(572,393,1).
txt_page(513,393,1).
txt_page(898,393,1).
txt_page(68,393,1).
txt_page(329,393,1).
txt_page(41,393,1).
txt_page(558,393,1).
txt_page(389,394,1).
txt_page(590,394,1).
txt_page(833,394,1).
txt_page(950,394,1).
txt_page(620,394,1).
txt_page(885,394,1).
txt_page(948,394,1).
txt_page(947,394,1).
txt_page(160,394,1).
txt_page(185,394,1).
txt_page(940,394,1).
txt_page(459,394,1).
txt_page(355,394,1).
txt_page(190,394,1).
txt_page(426,394,1).
txt_page(20,394,1).
txt_page(843,394,2).
txt_page(663,394,1).
txt_page(337,394,1).
txt_page(491,395,1).
txt_page(350,395,1).
txt_page(464,395,1).
txt_page(148,395,1).
txt_page(627,395,1).
txt_page(961,395,1).
txt_page(613,395,1).
txt_page(507,395,1).
txt_page(249,395,1).
txt_page(540,395,1).
txt_page(146,395,1).
txt_page(636,395,1).
txt_page(923,395,1).
txt_page(70,395,1).
txt_page(115,395,1).
txt_page(334,395,1).
txt_page(437,395,1).
txt_page(328,395,1).
txt_page(549,395,1).
txt_page(297,395,1).
txt_page(317,396,1).
txt_page(809,396,1).
txt_page(607,396,1).
txt_page(863,396,1).
txt_page(183,396,1).
txt_page(945,396,1).
txt_page(304,396,1).
txt_page(240,396,1).
txt_page(344,396,1).
txt_page(232,396,1).
txt_page(27,396,1).
txt_page(537,396,1).
txt_page(704,396,1).
txt_page(275,396,1).
txt_page(41,396,1).
txt_page(667,396,1).
txt_page(449,396,1).
txt_page(342,396,1).
txt_page(505,396,1).
txt_page(837,396,1).
txt_page(874,397,1).
txt_page(714,397,1).
txt_page(582,397,1).
txt_page(908,397,1).
txt_page(876,397,1).
txt_page(417,397,1).
txt_page(44,397,1).
txt_page(568,397,1).
txt_page(221,397,1).
txt_page(763,397,1).
txt_page(936,397,2).
txt_page(713,397,1).
txt_page(776,397,1).
txt_page(155,397,1).
txt_page(860,397,1).
txt_page(325,397,1).
txt_page(124,397,1).
txt_page(228,397,1).
txt_page(106,397,1).
txt_page(352,398,1).
txt_page(931,398,1).
txt_page(251,398,1).
txt_page(94,398,1).
txt_page(550,398,1).
txt_page(80,398,1).
txt_page(426,398,1).
txt_page(649,398,1).
txt_page(788,398,1).
txt_page(527,398,1).
txt_page(288,398,1).
txt_page(902,398,1).
txt_page(188,398,1).
txt_page(12,398,1).
txt_page(102,398,1).
txt_page(282,398,1).
txt_page(255,398,1).
txt_page(259,398,1).
txt_page(146,398,1).
txt_page(578,398,1).
txt_page(588,399,1).
txt_page(131,399,1).
txt_page(770,399,1).
txt_page(132,399,1).
txt_page(733,399,1).
txt_page(782,399,1).
txt_page(395,399,1).
txt_page(549,399,1).
txt_page(333,399,1).
txt_page(407,399,1).
txt_page(661,399,1).
txt_page(940,399,1).
txt_page(594,399,1).
txt_page(429,399,1).
txt_page(785,399,1).
txt_page(706,399,1).
txt_page(496,399,1).
txt_page(756,399,1).
txt_page(774,399,1).
txt_page(606,399,1).
txt_page(606,400,1).
txt_page(934,400,1).
txt_page(36,400,1).
txt_page(966,400,1).
txt_page(310,400,1).
txt_page(478,400,1).
txt_page(164,400,1).
txt_page(318,400,1).
txt_page(460,400,1).
txt_page(589,400,1).
txt_page(675,400,1).
txt_page(823,400,1).
txt_page(591,400,1).
txt_page(751,400,1).
txt_page(177,400,1).
txt_page(552,400,1).
txt_page(872,400,1).
txt_page(200,400,1).
txt_page(75,400,1).
txt_page(47,400,1).
txt_page(366,401,1).
txt_page(180,401,1).
txt_page(244,401,1).
txt_page(855,401,1).
txt_page(311,401,1).
txt_page(12,401,1).
txt_page(57,401,1).
txt_page(636,401,1).
txt_page(600,401,1).
txt_page(533,401,1).
txt_page(82,401,1).
txt_page(128,401,1).
txt_page(567,401,1).
txt_page(673,401,1).
txt_page(611,401,1).
txt_page(359,401,1).
txt_page(731,401,1).
txt_page(370,401,1).
txt_page(831,401,1).
txt_page(718,401,1).
txt_page(423,402,1).
txt_page(768,402,1).
txt_page(572,402,1).
txt_page(54,402,1).
txt_page(35,402,1).
txt_page(357,402,1).
txt_page(243,402,1).
txt_page(504,402,1).
txt_page(100,402,1).
txt_page(526,402,1).
txt_page(335,402,1).
txt_page(361,402,1).
txt_page(72,402,1).
txt_page(342,402,1).
txt_page(480,402,1).
txt_page(637,402,1).
txt_page(144,402,1).
txt_page(224,402,1).
txt_page(321,402,2).
txt_page(580,403,1).
txt_page(351,403,1).
txt_page(607,403,1).
txt_page(732,403,1).
txt_page(332,403,1).
txt_page(217,403,1).
txt_page(348,403,1).
txt_page(587,403,1).
txt_page(69,403,1).
txt_page(674,403,1).
txt_page(605,403,1).
txt_page(823,403,1).
txt_page(894,403,1).
txt_page(832,403,1).
txt_page(504,403,1).
txt_page(604,403,1).
txt_page(711,403,2).
txt_page(427,403,1).
txt_page(795,403,1).
txt_page(463,404,1).
txt_page(245,404,1).
txt_page(252,404,1).
txt_page(626,404,1).
txt_page(47,404,1).
txt_page(174,404,1).
txt_page(757,404,1).
txt_page(801,404,1).
txt_page(72,404,1).
txt_page(724,404,1).
txt_page(901,404,1).
txt_page(95,404,1).
txt_page(258,404,1).
txt_page(633,404,1).
txt_page(90,404,1).
txt_page(388,404,1).
txt_page(404,404,1).
txt_page(232,404,1).
txt_page(686,404,1).
txt_page(670,404,1).
txt_page(743,405,1).
txt_page(140,405,1).
txt_page(375,405,1).
txt_page(936,405,1).
txt_page(100,405,1).
txt_page(335,405,1).
txt_page(291,405,1).
txt_page(357,405,1).
txt_page(8,405,1).
txt_page(623,405,1).
txt_page(436,405,1).
txt_page(609,405,1).
txt_page(3,405,1).
txt_page(647,405,1).
txt_page(579,405,1).
txt_page(387,405,1).
txt_page(645,405,1).
txt_page(366,405,1).
txt_page(460,405,1).
txt_page(361,405,1).
txt_page(815,406,1).
txt_page(85,406,1).
txt_page(746,406,1).
txt_page(183,406,1).
txt_page(883,406,1).
txt_page(469,406,1).
txt_page(355,406,1).
txt_page(549,406,1).
txt_page(303,406,1).
txt_page(951,406,1).
txt_page(895,406,1).
txt_page(687,406,1).
txt_page(593,406,1).
txt_page(597,406,1).
txt_page(759,406,1).
txt_page(837,406,1).
txt_page(924,406,1).
txt_page(342,406,1).
txt_page(112,406,1).
txt_page(866,406,1).
txt_page(450,407,1).
txt_page(894,407,1).
txt_page(500,407,1).
txt_page(313,407,1).
txt_page(664,407,1).
txt_page(821,407,1).
txt_page(710,407,1).
txt_page(642,407,1).
txt_page(572,407,1).
txt_page(97,407,1).
txt_page(878,407,1).
txt_page(724,407,1).
txt_page(484,407,1).
txt_page(378,407,1).
txt_page(603,407,1).
txt_page(548,407,1).
txt_page(41,407,1).
txt_page(694,407,1).
txt_page(840,407,1).
txt_page(24,407,1).
txt_page(327,408,1).
txt_page(650,408,1).
txt_page(784,408,1).
txt_page(447,408,1).
txt_page(16,408,1).
txt_page(781,408,1).
txt_page(474,408,1).
txt_page(462,408,1).
txt_page(529,408,1).
txt_page(77,408,1).
txt_page(386,408,1).
txt_page(14,408,1).
txt_page(287,408,1).
txt_page(787,408,1).
txt_page(372,408,1).
txt_page(73,408,1).
txt_page(648,408,1).
txt_page(701,408,1).
txt_page(957,408,1).
txt_page(272,408,1).
txt_page(578,409,1).
txt_page(312,409,1).
txt_page(625,409,1).
txt_page(662,409,1).
txt_page(503,409,1).
txt_page(81,409,1).
txt_page(948,409,1).
txt_page(104,409,1).
txt_page(34,409,1).
txt_page(656,409,1).
txt_page(762,409,1).
txt_page(379,409,1).
txt_page(617,409,1).
txt_page(475,409,1).
txt_page(241,409,1).
txt_page(563,409,1).
txt_page(886,409,1).
txt_page(449,409,1).
txt_page(73,409,1).
txt_page(347,409,1).
txt_page(243,410,1).
txt_page(80,410,1).
txt_page(963,410,1).
txt_page(310,410,1).
txt_page(354,410,1).
txt_page(77,410,1).
txt_page(432,410,1).
txt_page(546,410,1).
txt_page(780,410,1).
txt_page(54,410,1).
txt_page(935,410,1).
txt_page(905,410,1).
txt_page(153,410,1).
txt_page(91,410,1).
txt_page(558,410,1).
txt_page(272,410,1).
txt_page(959,410,1).
txt_page(99,410,1).
txt_page(899,410,1).
txt_page(624,410,1).
txt_page(175,411,1).
txt_page(816,411,1).
txt_page(732,411,1).
txt_page(118,411,1).
txt_page(897,411,1).
txt_page(631,411,1).
txt_page(949,411,1).
txt_page(585,411,1).
txt_page(655,411,1).
txt_page(749,411,1).
txt_page(888,411,1).
txt_page(959,411,1).
txt_page(820,411,1).
txt_page(116,411,1).
txt_page(212,411,1).
txt_page(216,411,1).
txt_page(262,411,1).
txt_page(367,411,1).
txt_page(536,411,1).
txt_page(350,411,1).
txt_page(176,412,1).
txt_page(688,412,1).
txt_page(365,412,1).
txt_page(383,412,1).
txt_page(283,412,1).
txt_page(164,412,1).
txt_page(263,412,1).
txt_page(855,412,1).
txt_page(396,412,1).
txt_page(434,412,1).
txt_page(237,412,1).
txt_page(338,412,1).
txt_page(596,412,1).
txt_page(11,412,1).
txt_page(397,412,1).
txt_page(419,412,2).
txt_page(7,412,1).
txt_page(102,412,1).
txt_page(504,412,1).
txt_page(510,413,1).
txt_page(357,413,1).
txt_page(645,413,1).
txt_page(230,413,1).
txt_page(516,413,1).
txt_page(594,413,1).
txt_page(910,413,1).
txt_page(680,413,1).
txt_page(646,413,1).
txt_page(371,413,1).
txt_page(14,413,1).
txt_page(463,413,1).
txt_page(100,413,1).
txt_page(732,413,1).
txt_page(831,413,1).
txt_page(478,413,1).
txt_page(186,413,1).
txt_page(505,413,1).
txt_page(164,413,1).
txt_page(233,413,1).
txt_page(516,414,1).
txt_page(666,414,1).
txt_page(83,414,1).
txt_page(931,414,1).
txt_page(634,414,1).
txt_page(479,414,1).
txt_page(508,414,1).
txt_page(482,414,1).
txt_page(531,414,1).
txt_page(722,414,1).
txt_page(804,414,1).
txt_page(286,414,1).
txt_page(965,414,1).
txt_page(450,414,1).
txt_page(730,414,1).
txt_page(327,414,1).
txt_page(257,414,1).
txt_page(305,414,1).
txt_page(506,414,1).
txt_page(145,414,1).
txt_page(227,415,1).
txt_page(140,415,1).
txt_page(246,415,1).
txt_page(545,415,1).
txt_page(33,415,1).
txt_page(842,415,1).
txt_page(261,415,1).
txt_page(311,415,1).
txt_page(111,415,1).
txt_page(226,415,1).
txt_page(50,415,1).
txt_page(441,415,1).
txt_page(810,415,1).
txt_page(867,415,1).
txt_page(789,415,1).
txt_page(362,415,1).
txt_page(16,415,1).
txt_page(778,415,1).
txt_page(701,415,1).
txt_page(870,415,1).
txt_page(661,416,1).
txt_page(625,416,1).
txt_page(922,416,1).
txt_page(837,416,1).
txt_page(854,416,1).
txt_page(612,416,1).
txt_page(448,416,1).
txt_page(903,416,1).
txt_page(908,416,1).
txt_page(628,416,1).
txt_page(780,416,1).
txt_page(759,416,1).
txt_page(224,416,1).
txt_page(640,416,1).
txt_page(258,416,1).
txt_page(965,416,1).
txt_page(548,416,1).
txt_page(957,416,1).
txt_page(137,416,2).
txt_page(42,417,1).
txt_page(256,417,1).
txt_page(307,417,1).
txt_page(207,417,1).
txt_page(31,417,1).
txt_page(518,417,1).
txt_page(408,417,1).
txt_page(53,417,1).
txt_page(430,417,1).
txt_page(328,417,1).
txt_page(81,417,1).
txt_page(25,417,1).
txt_page(806,417,1).
txt_page(382,417,1).
txt_page(566,417,1).
txt_page(98,417,1).
txt_page(226,417,1).
txt_page(665,417,1).
txt_page(41,417,1).
txt_page(604,417,1).
txt_page(856,418,1).
txt_page(246,418,1).
txt_page(920,418,1).
txt_page(292,418,1).
txt_page(17,418,1).
txt_page(103,418,1).
txt_page(106,418,1).
txt_page(258,418,1).
txt_page(896,418,1).
txt_page(710,418,1).
txt_page(663,418,1).
txt_page(522,418,1).
txt_page(598,418,1).
txt_page(580,418,1).
txt_page(545,418,1).
txt_page(774,418,1).
txt_page(175,418,1).
txt_page(763,418,1).
txt_page(65,418,1).
txt_page(798,418,1).
txt_page(212,419,1).
txt_page(568,419,1).
txt_page(294,419,1).
txt_page(710,419,1).
txt_page(335,419,1).
txt_page(465,419,1).
txt_page(590,419,1).
txt_page(249,419,1).
txt_page(830,419,1).
txt_page(213,419,1).
txt_page(356,419,1).
txt_page(554,419,1).
txt_page(208,419,1).
txt_page(216,419,1).
txt_page(599,419,1).
txt_page(770,419,1).
txt_page(409,419,1).
txt_page(255,419,1).
txt_page(626,419,1).
txt_page(437,419,1).
txt_page(1,420,1).
txt_page(92,420,1).
txt_page(269,420,1).
txt_page(813,420,1).
txt_page(267,420,1).
txt_page(933,420,1).
txt_page(388,420,1).
txt_page(857,420,1).
txt_page(789,420,1).
txt_page(235,420,1).
txt_page(463,420,2).
txt_page(68,420,1).
txt_page(313,420,1).
txt_page(444,420,1).
txt_page(262,420,1).
txt_page(393,420,1).
txt_page(138,420,1).
txt_page(374,420,1).
txt_page(360,420,1).
txt_page(813,421,1).
txt_page(240,421,1).
txt_page(647,421,1).
txt_page(611,421,1).
txt_page(475,421,1).
txt_page(502,421,1).
txt_page(710,421,1).
txt_page(198,421,1).
txt_page(792,421,1).
txt_page(594,421,1).
txt_page(839,421,1).
txt_page(736,421,1).
txt_page(780,421,1).
txt_page(394,421,1).
txt_page(260,421,1).
txt_page(377,421,1).
txt_page(38,421,1).
txt_page(768,421,1).
txt_page(921,421,1).
txt_page(187,421,1).
txt_page(62,422,1).
txt_page(449,422,1).
txt_page(28,422,1).
txt_page(732,422,1).
txt_page(946,422,1).
txt_page(305,422,1).
txt_page(616,422,1).
txt_page(135,422,1).
txt_page(689,422,1).
txt_page(869,422,1).
txt_page(176,422,1).
txt_page(158,422,1).
txt_page(569,422,1).
txt_page(484,422,1).
txt_page(758,422,1).
txt_page(494,422,1).
txt_page(220,422,1).
txt_page(89,422,1).
txt_page(539,422,1).
txt_page(37,422,1).
txt_page(42,423,1).
txt_page(613,423,1).
txt_page(443,423,1).
txt_page(188,423,1).
txt_page(753,423,1).
txt_page(864,423,1).
txt_page(793,423,1).
txt_page(950,423,1).
txt_page(596,423,1).
txt_page(379,423,1).
txt_page(670,423,1).
txt_page(692,423,1).
txt_page(783,423,1).
txt_page(894,423,1).
txt_page(957,423,1).
txt_page(157,423,1).
txt_page(359,423,1).
txt_page(24,423,1).
txt_page(14,423,1).
txt_page(357,423,1).
txt_page(846,424,1).
txt_page(116,424,1).
txt_page(821,424,1).
txt_page(853,424,1).
txt_page(866,424,1).
txt_page(799,424,1).
txt_page(474,424,1).
txt_page(464,424,1).
txt_page(698,424,1).
txt_page(187,424,1).
txt_page(634,424,1).
txt_page(318,424,1).
txt_page(281,424,1).
txt_page(733,424,1).
txt_page(408,424,1).
txt_page(457,424,1).
txt_page(162,424,1).
txt_page(107,424,1).
txt_page(633,424,1).
txt_page(263,424,1).
txt_page(47,425,1).
txt_page(697,425,1).
txt_page(846,425,1).
txt_page(20,425,1).
txt_page(790,425,1).
txt_page(781,425,1).
txt_page(145,425,1).
txt_page(124,425,1).
txt_page(214,425,1).
txt_page(802,425,1).
txt_page(829,425,1).
txt_page(671,425,1).
txt_page(10,425,1).
txt_page(195,425,1).
txt_page(231,425,1).
txt_page(619,425,1).
txt_page(502,425,1).
txt_page(232,425,1).
txt_page(355,425,1).
txt_page(104,425,1).
txt_page(701,426,1).
txt_page(350,426,1).
txt_page(856,426,1).
txt_page(148,426,1).
txt_page(562,426,1).
txt_page(518,426,1).
txt_page(459,426,1).
txt_page(833,426,1).
txt_page(278,426,1).
txt_page(49,426,1).
txt_page(671,426,1).
txt_page(328,426,1).
txt_page(655,426,1).
txt_page(368,426,1).
txt_page(275,426,1).
txt_page(123,426,1).
txt_page(884,426,1).
txt_page(179,426,1).
txt_page(602,426,1).
txt_page(26,426,1).
txt_page(514,427,1).
txt_page(219,427,1).
txt_page(902,427,1).
txt_page(143,427,1).
txt_page(51,427,1).
txt_page(741,427,1).
txt_page(828,427,1).
txt_page(697,427,1).
txt_page(763,427,1).
txt_page(85,427,1).
txt_page(693,427,1).
txt_page(515,427,1).
txt_page(426,427,1).
txt_page(605,427,1).
txt_page(241,427,1).
txt_page(254,427,1).
txt_page(859,427,1).
txt_page(27,427,1).
txt_page(110,427,1).
txt_page(635,427,1).
txt_page(572,428,1).
txt_page(943,428,1).
txt_page(78,428,1).
txt_page(99,428,1).
txt_page(720,428,1).
txt_page(816,428,1).
txt_page(671,428,1).
txt_page(229,428,1).
txt_page(155,428,1).
txt_page(61,428,1).
txt_page(307,428,1).
txt_page(519,428,1).
txt_page(688,428,1).
txt_page(547,428,1).
txt_page(841,428,1).
txt_page(551,428,1).
txt_page(253,428,1).
txt_page(444,428,1).
txt_page(236,428,1).
txt_page(335,428,1).
txt_page(698,429,1).
txt_page(822,429,1).
txt_page(33,429,1).
txt_page(744,429,1).
txt_page(341,429,1).
txt_page(361,429,1).
txt_page(107,429,1).
txt_page(293,429,1).
txt_page(392,429,1).
txt_page(947,429,1).
txt_page(847,429,1).
txt_page(803,429,1).
txt_page(703,429,1).
txt_page(174,429,1).
txt_page(368,429,1).
txt_page(21,429,1).
txt_page(217,429,1).
txt_page(246,429,1).
txt_page(524,429,1).
txt_page(247,429,1).
txt_page(28,430,1).
txt_page(765,430,1).
txt_page(269,430,1).
txt_page(814,430,1).
txt_page(349,430,1).
txt_page(736,430,1).
txt_page(897,430,1).
txt_page(492,430,1).
txt_page(432,430,1).
txt_page(777,430,1).
txt_page(596,430,1).
txt_page(422,430,1).
txt_page(158,430,1).
txt_page(57,430,1).
txt_page(518,430,1).
txt_page(535,430,1).
txt_page(376,430,1).
txt_page(419,430,1).
txt_page(261,430,1).
txt_page(129,430,1).
txt_page(439,431,1).
txt_page(204,431,1).
txt_page(246,431,1).
txt_page(322,431,1).
txt_page(525,431,1).
txt_page(68,431,1).
txt_page(251,431,1).
txt_page(18,431,1).
txt_page(872,431,1).
txt_page(592,431,1).
txt_page(888,431,1).
txt_page(84,431,1).
txt_page(168,431,1).
txt_page(890,431,1).
txt_page(128,431,1).
txt_page(191,431,1).
txt_page(908,431,1).
txt_page(507,431,1).
txt_page(674,431,1).
txt_page(47,431,1).
txt_page(378,432,1).
txt_page(592,432,1).
txt_page(217,432,1).
txt_page(792,432,1).
txt_page(436,432,1).
txt_page(843,432,1).
txt_page(212,432,2).
txt_page(884,432,1).
txt_page(809,432,1).
txt_page(393,432,1).
txt_page(724,432,1).
txt_page(525,432,1).
txt_page(785,432,1).
txt_page(779,432,1).
txt_page(543,432,1).
txt_page(545,432,1).
txt_page(45,432,1).
txt_page(158,432,1).
txt_page(116,432,1).
txt_page(224,433,1).
txt_page(339,433,1).
txt_page(208,433,1).
txt_page(269,433,1).
txt_page(1,433,1).
txt_page(148,433,1).
txt_page(320,433,1).
txt_page(797,433,2).
txt_page(574,433,1).
txt_page(219,433,1).
txt_page(45,433,1).
txt_page(954,433,1).
txt_page(461,433,1).
txt_page(930,433,1).
txt_page(415,433,1).
txt_page(85,433,1).
txt_page(714,433,1).
txt_page(967,433,1).
txt_page(110,433,1).
txt_page(57,434,1).
txt_page(742,434,1).
txt_page(376,434,1).
txt_page(275,434,1).
txt_page(61,434,1).
txt_page(519,434,1).
txt_page(891,434,1).
txt_page(396,434,1).
txt_page(687,434,1).
txt_page(464,434,2).
txt_page(319,434,1).
txt_page(453,434,1).
txt_page(142,434,1).
txt_page(863,434,1).
txt_page(75,434,1).
txt_page(469,434,1).
txt_page(795,434,1).
txt_page(416,434,1).
txt_page(182,434,1).
txt_page(271,435,1).
txt_page(457,435,1).
txt_page(73,435,1).
txt_page(256,435,1).
txt_page(558,435,1).
txt_page(794,435,1).
txt_page(682,435,1).
txt_page(946,435,1).
txt_page(921,435,1).
txt_page(705,435,1).
txt_page(361,435,1).
txt_page(40,435,1).
txt_page(852,435,1).
txt_page(208,435,1).
txt_page(907,435,1).
txt_page(622,435,1).
txt_page(133,435,1).
txt_page(963,435,1).
txt_page(755,435,1).
txt_page(878,435,1).
txt_page(660,436,1).
txt_page(860,436,1).
txt_page(662,436,1).
txt_page(209,436,1).
txt_page(541,436,1).
txt_page(484,436,1).
txt_page(689,436,1).
txt_page(876,436,1).
txt_page(569,436,1).
txt_page(638,436,1).
txt_page(523,436,1).
txt_page(785,436,1).
txt_page(157,436,1).
txt_page(766,436,1).
txt_page(901,436,1).
txt_page(4,436,2).
txt_page(17,436,1).
txt_page(871,436,1).
txt_page(568,436,1).
txt_page(486,437,1).
txt_page(937,437,1).
txt_page(145,437,1).
txt_page(244,437,1).
txt_page(819,437,1).
txt_page(787,437,1).
txt_page(156,437,1).
txt_page(185,437,1).
txt_page(105,437,1).
txt_page(75,437,1).
txt_page(953,437,1).
txt_page(449,437,1).
txt_page(776,437,1).
txt_page(946,437,1).
txt_page(949,437,1).
txt_page(702,437,1).
txt_page(302,437,1).
txt_page(159,437,1).
txt_page(722,437,1).
txt_page(193,437,1).
txt_page(696,438,1).
txt_page(525,438,1).
txt_page(44,438,1).
txt_page(897,438,1).
txt_page(391,438,1).
txt_page(399,438,1).
txt_page(302,438,1).
txt_page(195,438,1).
txt_page(836,438,1).
txt_page(72,438,1).
txt_page(852,438,1).
txt_page(874,438,1).
txt_page(509,438,1).
txt_page(691,438,1).
txt_page(699,438,1).
txt_page(441,438,1).
txt_page(156,438,1).
txt_page(350,438,1).
txt_page(434,438,1).
txt_page(77,438,1).
txt_page(164,439,1).
txt_page(127,439,1).
txt_page(654,439,1).
txt_page(293,439,1).
txt_page(85,439,1).
txt_page(902,439,1).
txt_page(694,439,1).
txt_page(321,439,1).
txt_page(342,439,2).
txt_page(741,439,1).
txt_page(374,439,1).
txt_page(336,439,1).
txt_page(515,439,1).
txt_page(787,439,1).
txt_page(434,439,1).
txt_page(171,439,1).
txt_page(931,439,1).
txt_page(769,439,1).
txt_page(30,439,1).
txt_page(454,440,1).
txt_page(370,440,1).
txt_page(408,440,1).
txt_page(176,440,1).
txt_page(864,440,1).
txt_page(536,440,1).
txt_page(386,440,1).
txt_page(366,440,1).
txt_page(411,440,1).
txt_page(538,440,1).
txt_page(222,440,1).
txt_page(619,440,1).
txt_page(627,440,1).
txt_page(390,440,1).
txt_page(394,440,1).
txt_page(715,440,1).
txt_page(89,440,1).
txt_page(509,440,1).
txt_page(547,440,1).
txt_page(702,440,1).
txt_page(215,441,1).
txt_page(160,441,1).
txt_page(563,441,1).
txt_page(379,441,1).
txt_page(201,441,1).
txt_page(8,441,1).
txt_page(811,441,1).
txt_page(882,441,1).
txt_page(889,441,1).
txt_page(283,441,1).
txt_page(360,441,1).
txt_page(924,441,1).
txt_page(17,441,1).
txt_page(36,441,1).
txt_page(42,441,1).
txt_page(818,441,1).
txt_page(806,441,1).
txt_page(228,441,1).
txt_page(265,441,1).
txt_page(339,441,1).
txt_page(106,442,1).
txt_page(382,442,1).
txt_page(91,442,1).
txt_page(380,442,1).
txt_page(407,442,1).
txt_page(222,442,1).
txt_page(21,442,1).
txt_page(431,442,1).
txt_page(146,442,1).
txt_page(76,442,1).
txt_page(469,442,1).
txt_page(420,442,1).
txt_page(542,442,1).
txt_page(405,442,1).
txt_page(663,442,1).
txt_page(393,442,1).
txt_page(543,442,1).
txt_page(942,442,1).
txt_page(856,442,1).
txt_page(792,442,1).
txt_page(13,443,1).
txt_page(371,443,1).
txt_page(324,443,1).
txt_page(278,443,1).
txt_page(453,443,1).
txt_page(214,443,1).
txt_page(525,443,1).
txt_page(10,443,1).
txt_page(227,443,1).
txt_page(609,443,1).
txt_page(156,443,1).
txt_page(417,443,1).
txt_page(379,443,1).
txt_page(762,443,1).
txt_page(757,443,1).
txt_page(142,443,1).
txt_page(957,443,1).
txt_page(460,443,1).
txt_page(740,443,1).
txt_page(748,443,1).
txt_page(204,444,1).
txt_page(595,444,1).
txt_page(443,444,1).
txt_page(374,444,1).
txt_page(141,444,1).
txt_page(831,444,1).
txt_page(888,444,1).
txt_page(256,444,1).
txt_page(899,444,1).
txt_page(356,444,1).
txt_page(930,444,1).
txt_page(679,444,1).
txt_page(270,444,1).
txt_page(936,444,1).
txt_page(228,444,1).
txt_page(214,444,1).
txt_page(607,444,1).
txt_page(150,444,1).
txt_page(254,444,1).
txt_page(478,444,1).
txt_page(443,445,1).
txt_page(299,445,1).
txt_page(66,445,1).
txt_page(549,445,1).
txt_page(763,445,1).
txt_page(815,445,1).
txt_page(509,445,1).
txt_page(678,445,1).
txt_page(665,445,1).
txt_page(715,445,1).
txt_page(380,445,1).
txt_page(606,445,1).
txt_page(845,445,1).
txt_page(497,445,1).
txt_page(966,445,1).
txt_page(394,445,1).
txt_page(9,445,1).
txt_page(242,445,1).
txt_page(649,445,1).
txt_page(306,445,1).
txt_page(580,446,1).
txt_page(387,446,1).
txt_page(280,446,1).
txt_page(305,446,1).
txt_page(172,446,1).
txt_page(772,446,1).
txt_page(408,446,1).
txt_page(332,446,1).
txt_page(905,446,1).
txt_page(109,446,1).
txt_page(593,446,1).
txt_page(760,446,1).
txt_page(167,446,1).
txt_page(266,446,1).
txt_page(706,446,1).
txt_page(278,446,1).
txt_page(148,446,1).
txt_page(351,446,1).
txt_page(255,446,1).
txt_page(614,446,1).
txt_page(519,447,1).
txt_page(773,447,1).
txt_page(41,447,1).
txt_page(238,447,1).
txt_page(669,447,1).
txt_page(385,447,1).
txt_page(414,447,1).
txt_page(247,447,1).
txt_page(674,447,1).
txt_page(818,447,1).
txt_page(365,447,1).
txt_page(66,447,1).
txt_page(692,447,1).
txt_page(303,447,1).
txt_page(369,447,1).
txt_page(153,447,1).
txt_page(910,447,1).
txt_page(398,447,1).
txt_page(477,447,1).
txt_page(55,447,1).
txt_page(452,448,1).
txt_page(7,448,1).
txt_page(298,448,1).
txt_page(478,448,1).
txt_page(690,448,1).
txt_page(288,448,1).
txt_page(662,448,1).
txt_page(418,448,1).
txt_page(466,448,1).
txt_page(16,448,1).
txt_page(211,448,1).
txt_page(684,448,1).
txt_page(243,448,1).
txt_page(794,448,1).
txt_page(569,448,1).
txt_page(751,448,1).
txt_page(687,448,1).
txt_page(663,448,1).
txt_page(911,448,1).
txt_page(960,448,1).
txt_page(313,449,1).
txt_page(705,449,1).
txt_page(777,449,1).
txt_page(124,449,1).
txt_page(422,449,1).
txt_page(96,449,1).
txt_page(714,449,1).
txt_page(837,449,1).
txt_page(69,449,1).
txt_page(869,449,1).
txt_page(183,449,1).
txt_page(404,449,1).
txt_page(614,449,1).
txt_page(35,449,1).
txt_page(838,449,1).
txt_page(750,449,1).
txt_page(440,449,1).
txt_page(378,449,1).
txt_page(395,449,1).
txt_page(405,449,1).
txt_page(61,450,1).
txt_page(689,450,1).
txt_page(319,450,1).
txt_page(338,450,1).
txt_page(849,450,1).
txt_page(792,450,1).
txt_page(648,450,1).
txt_page(470,450,1).
txt_page(930,450,1).
txt_page(450,450,1).
txt_page(140,450,1).
txt_page(557,450,1).
txt_page(867,450,1).
txt_page(530,450,1).
txt_page(619,450,1).
txt_page(885,450,1).
txt_page(675,450,1).
txt_page(630,450,1).
txt_page(631,450,1).
txt_page(896,450,1).
txt_page(734,451,1).
txt_page(172,451,1).
txt_page(188,451,1).
txt_page(848,451,1).
txt_page(97,451,1).
txt_page(876,451,1).
txt_page(800,451,1).
txt_page(207,451,1).
txt_page(682,451,1).
txt_page(884,451,1).
txt_page(965,451,1).
txt_page(521,451,1).
txt_page(444,451,1).
txt_page(633,451,1).
txt_page(58,451,1).
txt_page(725,451,1).
txt_page(571,451,1).
txt_page(841,451,1).
txt_page(882,451,1).
txt_page(157,451,1).
txt_page(400,452,1).
txt_page(176,452,1).
txt_page(572,452,1).
txt_page(175,452,1).
txt_page(388,452,1).
txt_page(110,452,1).
txt_page(874,452,1).
txt_page(648,452,1).
txt_page(36,452,1).
txt_page(827,452,1).
txt_page(610,452,1).
txt_page(257,452,1).
txt_page(647,452,1).
txt_page(269,452,1).
txt_page(365,452,1).
txt_page(419,452,1).
txt_page(380,452,1).
txt_page(100,452,1).
txt_page(214,452,1).
txt_page(676,452,1).
txt_page(600,453,1).
txt_page(781,453,1).
txt_page(865,453,1).
txt_page(771,453,1).
txt_page(783,453,1).
txt_page(580,453,1).
txt_page(54,453,1).
txt_page(41,453,1).
txt_page(119,453,1).
txt_page(555,453,1).
txt_page(776,453,1).
txt_page(148,453,1).
txt_page(859,453,1).
txt_page(615,453,1).
txt_page(774,453,1).
txt_page(713,453,1).
txt_page(164,453,1).
txt_page(379,453,1).
txt_page(605,453,1).
txt_page(236,453,1).
txt_page(11,454,1).
txt_page(121,454,1).
txt_page(899,454,1).
txt_page(162,454,1).
txt_page(625,454,1).
txt_page(111,454,1).
txt_page(918,454,1).
txt_page(681,454,1).
txt_page(713,454,1).
txt_page(729,454,1).
txt_page(747,454,1).
txt_page(733,454,1).
txt_page(617,454,1).
txt_page(437,454,1).
txt_page(172,454,1).
txt_page(753,454,1).
txt_page(952,454,1).
txt_page(771,454,1).
txt_page(175,454,1).
txt_page(90,454,1).
txt_page(338,455,1).
txt_page(279,455,1).
txt_page(734,455,1).
txt_page(557,455,1).
txt_page(403,455,1).
txt_page(782,455,1).
txt_page(259,455,1).
txt_page(181,455,1).
txt_page(8,455,1).
txt_page(343,455,1).
txt_page(340,455,1).
txt_page(169,455,1).
txt_page(670,455,1).
txt_page(858,455,1).
txt_page(677,455,1).
txt_page(43,455,1).
txt_page(771,455,1).
txt_page(270,455,1).
txt_page(49,455,1).
txt_page(377,455,1).
txt_page(560,456,1).
txt_page(513,456,1).
txt_page(324,456,1).
txt_page(839,456,1).
txt_page(901,456,1).
txt_page(637,456,1).
txt_page(870,456,1).
txt_page(51,456,1).
txt_page(48,456,1).
txt_page(122,456,1).
txt_page(386,456,1).
txt_page(371,456,1).
txt_page(487,456,1).
txt_page(573,456,1).
txt_page(834,456,1).
txt_page(178,456,1).
txt_page(555,456,1).
txt_page(960,456,1).
txt_page(935,456,1).
txt_page(950,456,1).
txt_page(577,457,1).
txt_page(908,457,1).
txt_page(528,457,1).
txt_page(721,457,1).
txt_page(471,457,1).
txt_page(239,457,1).
txt_page(536,457,1).
txt_page(459,457,1).
txt_page(316,457,1).
txt_page(261,457,1).
txt_page(649,457,1).
txt_page(325,457,1).
txt_page(137,457,1).
txt_page(75,457,1).
txt_page(229,457,2).
txt_page(365,457,1).
txt_page(279,457,1).
txt_page(926,457,1).
txt_page(601,457,1).
txt_page(565,458,1).
txt_page(895,458,1).
txt_page(36,458,1).
txt_page(468,458,1).
txt_page(534,458,1).
txt_page(503,458,1).
txt_page(749,458,1).
txt_page(382,458,1).
txt_page(242,458,1).
txt_page(725,458,1).
txt_page(799,458,1).
txt_page(264,458,1).
txt_page(639,458,1).
txt_page(621,458,1).
txt_page(146,458,1).
txt_page(126,458,1).
txt_page(567,458,1).
txt_page(56,458,1).
txt_page(879,458,1).
txt_page(552,458,1).
txt_page(637,459,1).
txt_page(68,459,1).
txt_page(556,459,1).
txt_page(960,459,1).
txt_page(808,459,1).
txt_page(844,459,1).
txt_page(872,459,1).
txt_page(552,459,1).
txt_page(538,459,1).
txt_page(968,459,1).
txt_page(39,459,1).
txt_page(855,459,1).
txt_page(296,459,1).
txt_page(967,459,1).
txt_page(869,459,1).
txt_page(619,459,1).
txt_page(197,459,1).
txt_page(41,459,1).
txt_page(312,459,1).
txt_page(946,459,1).
txt_page(383,460,1).
txt_page(169,460,1).
txt_page(691,460,1).
txt_page(348,460,1).
txt_page(394,460,1).
txt_page(53,460,1).
txt_page(545,460,1).
txt_page(318,460,1).
txt_page(129,460,1).
txt_page(791,460,1).
txt_page(33,460,1).
txt_page(479,460,1).
txt_page(219,460,1).
txt_page(368,460,1).
txt_page(174,460,1).
txt_page(748,460,1).
txt_page(939,460,1).
txt_page(683,460,1).
txt_page(570,460,1).
txt_page(276,460,1).
txt_page(249,461,1).
txt_page(605,461,1).
txt_page(113,461,1).
txt_page(634,461,1).
txt_page(539,461,1).
txt_page(332,461,1).
txt_page(721,461,1).
txt_page(911,461,1).
txt_page(65,461,1).
txt_page(300,461,1).
txt_page(259,461,1).
txt_page(869,461,1).
txt_page(206,461,1).
txt_page(954,461,1).
txt_page(380,461,1).
txt_page(801,461,1).
txt_page(714,461,1).
txt_page(207,461,1).
txt_page(244,461,1).
txt_page(362,461,1).
txt_page(603,462,1).
txt_page(300,462,1).
txt_page(576,462,1).
txt_page(403,462,1).
txt_page(599,462,1).
txt_page(953,462,1).
txt_page(810,462,1).
txt_page(507,462,1).
txt_page(59,462,1).
txt_page(480,462,1).
txt_page(201,462,1).
txt_page(702,462,1).
txt_page(11,462,1).
txt_page(653,462,1).
txt_page(933,462,1).
txt_page(686,462,1).
txt_page(91,462,1).
txt_page(30,462,1).
txt_page(620,462,1).
txt_page(579,462,1).
txt_page(441,463,1).
txt_page(131,463,1).
txt_page(719,463,1).
txt_page(806,463,1).
txt_page(296,463,1).
txt_page(890,463,1).
txt_page(944,463,1).
txt_page(902,463,1).
txt_page(702,463,1).
txt_page(839,463,1).
txt_page(778,463,1).
txt_page(332,463,1).
txt_page(19,463,1).
txt_page(92,463,1).
txt_page(860,463,1).
txt_page(605,463,1).
txt_page(390,463,1).
txt_page(750,463,1).
txt_page(51,463,1).
txt_page(840,463,1).
txt_page(410,464,1).
txt_page(702,464,1).
txt_page(227,464,1).
txt_page(798,464,1).
txt_page(137,464,1).
txt_page(461,464,1).
txt_page(251,464,1).
txt_page(349,464,1).
txt_page(712,464,1).
txt_page(788,464,1).
txt_page(640,464,1).
txt_page(784,464,1).
txt_page(318,464,1).
txt_page(235,464,1).
txt_page(844,464,1).
txt_page(802,464,1).
txt_page(283,464,1).
txt_page(926,464,1).
txt_page(347,464,1).
txt_page(673,464,1).
txt_page(834,465,1).
txt_page(399,465,1).
txt_page(107,465,1).
txt_page(420,465,1).
txt_page(489,465,1).
txt_page(356,465,1).
txt_page(842,465,1).
txt_page(935,465,1).
txt_page(219,465,1).
txt_page(110,465,1).
txt_page(701,465,1).
txt_page(947,465,1).
txt_page(622,465,1).
txt_page(299,465,1).
txt_page(901,465,1).
txt_page(739,465,1).
txt_page(809,465,1).
txt_page(419,465,1).
txt_page(700,465,1).
txt_page(561,465,1).
txt_page(307,466,1).
txt_page(873,466,1).
txt_page(449,466,1).
txt_page(712,466,1).
txt_page(855,466,1).
txt_page(742,466,1).
txt_page(56,466,1).
txt_page(618,466,1).
txt_page(357,466,1).
txt_page(823,466,1).
txt_page(152,466,1).
txt_page(218,466,1).
txt_page(556,466,1).
txt_page(884,466,1).
txt_page(461,466,1).
txt_page(43,466,1).
txt_page(530,466,1).
txt_page(24,466,1).
txt_page(572,466,1).
txt_page(526,466,1).
txt_page(650,467,1).
txt_page(628,467,1).
txt_page(693,467,1).
txt_page(896,467,1).
txt_page(52,467,1).
txt_page(949,467,1).
txt_page(439,467,1).
txt_page(78,467,1).
txt_page(710,467,1).
txt_page(890,467,1).
txt_page(776,467,1).
txt_page(141,467,1).
txt_page(485,467,1).
txt_page(425,467,1).
txt_page(214,467,1).
txt_page(54,467,1).
txt_page(284,467,1).
txt_page(489,467,1).
txt_page(441,467,1).
txt_page(198,467,1).
txt_page(321,468,1).
txt_page(464,468,1).
txt_page(595,468,1).
txt_page(735,468,1).
txt_page(497,468,1).
txt_page(959,468,1).
txt_page(498,468,1).
txt_page(121,468,1).
txt_page(53,468,1).
txt_page(657,468,1).
txt_page(249,468,1).
txt_page(266,468,1).
txt_page(480,468,1).
txt_page(129,468,1).
txt_page(816,468,1).
txt_page(415,468,1).
txt_page(904,468,1).
txt_page(713,468,1).
txt_page(189,468,1).
txt_page(748,468,1).
txt_page(517,469,1).
txt_page(290,469,1).
txt_page(400,469,1).
txt_page(322,469,1).
txt_page(871,469,1).
txt_page(165,469,1).
txt_page(404,469,1).
txt_page(622,469,1).
txt_page(83,469,1).
txt_page(411,469,1).
txt_page(107,469,1).
txt_page(786,469,1).
txt_page(541,469,1).
txt_page(500,469,1).
txt_page(9,469,1).
txt_page(903,469,1).
txt_page(109,469,1).
txt_page(92,469,1).
txt_page(342,469,1).
txt_page(309,469,1).
txt_page(108,470,1).
txt_page(97,470,1).
txt_page(373,470,1).
txt_page(766,470,1).
txt_page(749,470,1).
txt_page(719,470,1).
txt_page(773,470,1).
txt_page(512,470,1).
txt_page(867,470,1).
txt_page(823,470,1).
txt_page(850,470,1).
txt_page(826,470,1).
txt_page(127,470,1).
txt_page(564,470,1).
txt_page(228,470,1).
txt_page(610,470,1).
txt_page(44,470,1).
txt_page(772,470,1).
txt_page(13,470,1).
txt_page(788,470,1).
txt_page(644,471,1).
txt_page(891,471,1).
txt_page(273,471,1).
txt_page(377,471,1).
txt_page(433,471,1).
txt_page(880,471,1).
txt_page(42,471,1).
txt_page(741,471,1).
txt_page(879,471,1).
txt_page(643,471,1).
txt_page(621,471,1).
txt_page(382,471,1).
txt_page(767,471,1).
txt_page(595,471,1).
txt_page(575,471,1).
txt_page(177,471,1).
txt_page(418,471,1).
txt_page(389,471,1).
txt_page(478,471,1).
txt_page(637,471,1).
txt_page(960,472,1).
txt_page(919,472,1).
txt_page(842,472,1).
txt_page(863,472,1).
txt_page(731,472,1).
txt_page(188,472,1).
txt_page(48,472,1).
txt_page(2,472,1).
txt_page(658,472,1).
txt_page(512,472,1).
txt_page(453,472,1).
txt_page(676,472,1).
txt_page(469,472,1).
txt_page(322,472,1).
txt_page(425,472,1).
txt_page(470,472,1).
txt_page(780,472,1).
txt_page(661,472,1).
txt_page(708,472,1).
txt_page(846,472,1).
txt_page(136,473,1).
txt_page(628,473,1).
txt_page(825,473,1).
txt_page(436,473,1).
txt_page(544,473,1).
txt_page(901,473,1).
txt_page(475,473,1).
txt_page(93,473,1).
txt_page(654,473,1).
txt_page(593,473,1).
txt_page(319,473,1).
txt_page(608,473,1).
txt_page(20,473,1).
txt_page(644,473,1).
txt_page(244,473,1).
txt_page(26,473,1).
txt_page(624,473,1).
txt_page(378,473,1).
txt_page(611,473,1).
txt_page(169,473,1).
txt_page(856,474,1).
txt_page(313,474,1).
txt_page(120,474,1).
txt_page(182,474,1).
txt_page(345,474,1).
txt_page(207,474,1).
txt_page(412,474,1).
txt_page(384,474,1).
txt_page(681,474,1).
txt_page(539,474,1).
txt_page(250,474,1).
txt_page(392,474,1).
txt_page(113,474,1).
txt_page(893,474,1).
txt_page(375,474,1).
txt_page(405,474,1).
txt_page(733,474,1).
txt_page(886,474,1).
txt_page(172,474,1).
txt_page(652,474,1).
txt_page(788,475,1).
txt_page(279,475,1).
txt_page(867,475,1).
txt_page(30,475,1).
txt_page(532,475,1).
txt_page(331,475,1).
txt_page(115,475,1).
txt_page(916,475,1).
txt_page(707,475,1).
txt_page(769,475,1).
txt_page(674,475,1).
txt_page(456,475,1).
txt_page(522,475,1).
txt_page(306,475,1).
txt_page(446,475,1).
txt_page(961,475,1).
txt_page(399,475,1).
txt_page(826,475,1).
txt_page(304,475,1).
txt_page(919,475,1).
txt_page(274,476,1).
txt_page(845,476,1).
txt_page(334,476,1).
txt_page(187,476,1).
txt_page(816,476,1).
txt_page(74,476,1).
txt_page(953,476,1).
txt_page(330,476,1).
txt_page(629,476,1).
txt_page(785,476,1).
txt_page(847,476,1).
txt_page(822,476,1).
txt_page(942,476,1).
txt_page(87,476,1).
txt_page(569,476,1).
txt_page(522,476,1).
txt_page(132,476,1).
txt_page(305,476,1).
txt_page(594,476,1).
txt_page(333,476,1).
txt_page(820,477,1).
txt_page(540,477,1).
txt_page(380,477,1).
txt_page(644,477,1).
txt_page(597,477,1).
txt_page(916,477,1).
txt_page(685,477,1).
txt_page(285,477,1).
txt_page(197,477,1).
txt_page(529,477,1).
txt_page(411,477,1).
txt_page(445,477,1).
txt_page(22,477,1).
txt_page(342,477,1).
txt_page(572,477,1).
txt_page(620,477,1).
txt_page(422,477,1).
txt_page(77,477,1).
txt_page(894,477,1).
txt_page(911,477,1).
txt_page(570,478,1).
txt_page(374,478,1).
txt_page(147,478,1).
txt_page(467,478,1).
txt_page(210,478,1).
txt_page(740,478,1).
txt_page(638,478,1).
txt_page(606,478,1).
txt_page(251,478,2).
txt_page(203,478,1).
txt_page(336,478,1).
txt_page(926,478,1).
txt_page(639,478,1).
txt_page(819,478,1).
txt_page(372,478,1).
txt_page(775,478,1).
txt_page(262,478,1).
txt_page(312,478,1).
txt_page(480,478,1).
txt_page(438,479,1).
txt_page(130,479,1).
txt_page(23,479,1).
txt_page(323,479,1).
txt_page(928,479,1).
txt_page(152,479,1).
txt_page(294,479,1).
txt_page(792,479,1).
txt_page(761,479,1).
txt_page(320,479,1).
txt_page(522,479,1).
txt_page(947,479,1).
txt_page(708,479,1).
txt_page(524,479,1).
txt_page(267,479,1).
txt_page(820,479,1).
txt_page(527,479,1).
txt_page(962,479,1).
txt_page(190,479,1).
txt_page(187,479,1).
txt_page(440,480,1).
txt_page(92,480,1).
txt_page(698,480,1).
txt_page(405,480,1).
txt_page(864,480,1).
txt_page(804,480,1).
txt_page(251,480,1).
txt_page(525,480,1).
txt_page(914,480,1).
txt_page(938,480,1).
txt_page(802,480,1).
txt_page(547,480,1).
txt_page(906,480,1).
txt_page(271,480,1).
txt_page(194,480,1).
txt_page(816,480,1).
txt_page(352,480,1).
txt_page(742,480,1).
txt_page(320,480,1).
txt_page(280,480,1).
txt_page(13,481,1).
txt_page(563,481,1).
txt_page(504,481,1).
txt_page(849,481,1).
txt_page(820,481,1).
txt_page(815,481,1).
txt_page(161,481,1).
txt_page(931,481,1).
txt_page(287,481,1).
txt_page(232,481,1).
txt_page(495,481,1).
txt_page(587,481,1).
txt_page(158,481,1).
txt_page(942,481,1).
txt_page(542,481,1).
txt_page(220,481,1).
txt_page(127,481,1).
txt_page(579,481,1).
txt_page(51,481,1).
txt_page(293,481,1).
txt_page(830,482,1).
txt_page(527,482,1).
txt_page(508,482,1).
txt_page(469,482,1).
txt_page(30,482,1).
txt_page(11,482,1).
txt_page(399,482,1).
txt_page(534,482,1).
txt_page(372,482,1).
txt_page(183,482,1).
txt_page(347,482,1).
txt_page(155,482,1).
txt_page(162,482,1).
txt_page(234,482,1).
txt_page(515,482,1).
txt_page(154,482,1).
txt_page(692,482,1).
txt_page(486,482,1).
txt_page(457,482,1).
txt_page(264,482,1).
txt_page(437,483,1).
txt_page(93,483,1).
txt_page(461,483,1).
txt_page(621,483,1).
txt_page(760,483,1).
txt_page(787,483,1).
txt_page(703,483,1).
txt_page(968,483,1).
txt_page(67,483,1).
txt_page(948,483,1).
txt_page(548,483,1).
txt_page(84,483,1).
txt_page(897,483,1).
txt_page(201,483,1).
txt_page(526,483,1).
txt_page(282,483,1).
txt_page(376,483,1).
txt_page(9,483,1).
txt_page(208,483,1).
txt_page(877,483,1).
txt_page(128,484,1).
txt_page(934,484,1).
txt_page(211,484,1).
txt_page(821,484,1).
txt_page(696,484,1).
txt_page(545,484,1).
txt_page(594,484,1).
txt_page(143,484,1).
txt_page(571,484,1).
txt_page(296,484,1).
txt_page(249,484,1).
txt_page(178,484,1).
txt_page(303,484,1).
txt_page(626,484,1).
txt_page(258,484,1).
txt_page(818,484,1).
txt_page(418,484,1).
txt_page(885,484,1).
txt_page(39,484,1).
txt_page(808,484,1).
txt_page(325,485,1).
txt_page(272,485,1).
txt_page(782,485,1).
txt_page(106,485,1).
txt_page(63,485,1).
txt_page(930,485,1).
txt_page(763,485,1).
txt_page(209,485,1).
txt_page(251,485,1).
txt_page(150,485,1).
txt_page(641,485,1).
txt_page(471,485,1).
txt_page(715,485,1).
txt_page(487,485,1).
txt_page(648,485,1).
txt_page(70,485,1).
txt_page(298,485,1).
txt_page(193,485,1).
txt_page(868,485,1).
txt_page(444,485,1).
txt_page(254,486,1).
txt_page(941,486,1).
txt_page(775,486,1).
txt_page(902,486,1).
txt_page(95,486,1).
txt_page(317,486,1).
txt_page(41,486,1).
txt_page(776,486,1).
txt_page(872,486,1).
txt_page(134,486,1).
txt_page(103,486,1).
txt_page(479,486,1).
txt_page(592,486,1).
txt_page(528,486,1).
txt_page(723,486,1).
txt_page(921,486,1).
txt_page(483,486,1).
txt_page(357,486,1).
txt_page(220,486,1).
txt_page(410,486,1).
txt_page(937,487,1).
txt_page(804,487,1).
txt_page(853,487,1).
txt_page(404,487,1).
txt_page(829,487,1).
txt_page(222,487,1).
txt_page(172,487,1).
txt_page(555,487,1).
txt_page(256,487,1).
txt_page(630,487,1).
txt_page(114,487,1).
txt_page(757,487,1).
txt_page(190,487,1).
txt_page(441,487,1).
txt_page(6,487,1).
txt_page(301,487,1).
txt_page(883,487,1).
txt_page(758,487,1).
txt_page(50,487,1).
txt_page(783,487,1).
txt_page(331,488,1).
txt_page(140,488,1).
txt_page(508,488,1).
txt_page(168,488,1).
txt_page(458,488,1).
txt_page(111,488,1).
txt_page(651,488,1).
txt_page(649,488,1).
txt_page(852,488,1).
txt_page(505,488,1).
txt_page(43,488,1).
txt_page(819,488,1).
txt_page(20,488,1).
txt_page(456,488,1).
txt_page(320,488,1).
txt_page(419,488,1).
txt_page(123,488,1).
txt_page(46,488,1).
txt_page(169,488,1).
txt_page(48,488,1).
txt_page(391,489,1).
txt_page(81,489,1).
txt_page(412,489,1).
txt_page(432,489,1).
txt_page(234,489,1).
txt_page(677,489,1).
txt_page(901,489,1).
txt_page(718,489,1).
txt_page(608,489,1).
txt_page(236,489,1).
txt_page(634,489,1).
txt_page(451,489,1).
txt_page(137,489,1).
txt_page(200,489,1).
txt_page(807,489,1).
txt_page(707,489,1).
txt_page(735,489,1).
txt_page(804,489,1).
txt_page(83,489,1).
txt_page(911,489,1).
txt_page(226,490,1).
txt_page(667,490,1).
txt_page(161,490,1).
txt_page(599,490,1).
txt_page(359,490,1).
txt_page(843,490,1).
txt_page(218,490,1).
txt_page(111,490,1).
txt_page(99,490,1).
txt_page(604,490,1).
txt_page(302,490,1).
txt_page(788,490,1).
txt_page(562,490,1).
txt_page(23,490,1).
txt_page(967,490,1).
txt_page(253,490,1).
txt_page(430,490,1).
txt_page(307,490,1).
txt_page(896,490,1).
txt_page(190,490,1).
txt_page(146,491,1).
txt_page(863,491,1).
txt_page(758,491,1).
txt_page(652,491,1).
txt_page(493,491,1).
txt_page(83,491,1).
txt_page(673,491,1).
txt_page(870,491,1).
txt_page(919,491,1).
txt_page(346,491,1).
txt_page(635,491,1).
txt_page(715,491,1).
txt_page(234,491,1).
txt_page(161,491,1).
txt_page(619,491,1).
txt_page(587,491,1).
txt_page(305,491,1).
txt_page(724,491,1).
txt_page(621,491,1).
txt_page(753,491,1).
txt_page(116,492,1).
txt_page(948,492,1).
txt_page(194,492,1).
txt_page(663,492,1).
txt_page(506,492,1).
txt_page(430,492,1).
txt_page(930,492,1).
txt_page(153,492,1).
txt_page(409,492,1).
txt_page(703,492,1).
txt_page(856,492,1).
txt_page(445,492,1).
txt_page(564,492,1).
txt_page(714,492,1).
txt_page(303,492,1).
txt_page(524,492,1).
txt_page(261,492,1).
txt_page(458,492,1).
txt_page(787,492,1).
txt_page(941,492,1).
txt_page(66,493,1).
txt_page(137,493,1).
txt_page(701,493,1).
txt_page(45,493,1).
txt_page(340,493,1).
txt_page(788,493,1).
txt_page(370,493,1).
txt_page(521,493,1).
txt_page(529,493,1).
txt_page(57,493,1).
txt_page(773,493,1).
txt_page(507,493,1).
txt_page(623,493,1).
txt_page(879,493,1).
txt_page(134,493,1).
txt_page(566,493,1).
txt_page(418,493,1).
txt_page(690,493,1).
txt_page(88,493,1).
txt_page(104,493,1).
txt_page(808,494,1).
txt_page(208,494,1).
txt_page(107,494,1).
txt_page(243,494,1).
txt_page(753,494,1).
txt_page(486,494,1).
txt_page(853,494,1).
txt_page(685,494,1).
txt_page(357,494,1).
txt_page(484,494,1).
txt_page(90,494,1).
txt_page(439,494,1).
txt_page(589,494,1).
txt_page(473,494,1).
txt_page(823,494,1).
txt_page(109,494,1).
txt_page(211,494,1).
txt_page(757,494,1).
txt_page(768,494,1).
txt_page(220,494,1).
txt_page(819,495,1).
txt_page(544,495,1).
txt_page(13,495,1).
txt_page(918,495,1).
txt_page(531,495,1).
txt_page(679,495,1).
txt_page(318,495,1).
txt_page(362,495,1).
txt_page(821,495,1).
txt_page(904,495,1).
txt_page(815,495,1).
txt_page(793,495,1).
txt_page(415,495,1).
txt_page(442,495,1).
txt_page(291,495,1).
txt_page(655,495,1).
txt_page(774,495,1).
txt_page(967,495,1).
txt_page(589,495,1).
txt_page(545,495,1).
txt_page(559,496,1).
txt_page(787,496,1).
txt_page(529,496,1).
txt_page(182,496,1).
txt_page(937,496,1).
txt_page(275,496,1).
txt_page(890,496,1).
txt_page(703,496,1).
txt_page(953,496,1).
txt_page(905,496,1).
txt_page(699,496,1).
txt_page(384,496,1).
txt_page(887,496,1).
txt_page(146,496,1).
txt_page(810,496,1).
txt_page(17,496,1).
txt_page(852,496,1).
txt_page(678,496,1).
txt_page(79,496,1).
txt_page(386,496,1).
txt_page(285,497,1).
txt_page(155,497,1).
txt_page(154,497,1).
txt_page(712,497,1).
txt_page(162,497,1).
txt_page(932,497,1).
txt_page(127,497,1).
txt_page(533,497,1).
txt_page(737,497,1).
txt_page(502,497,1).
txt_page(249,497,1).
txt_page(866,497,1).
txt_page(881,497,1).
txt_page(692,497,1).
txt_page(404,497,1).
txt_page(446,497,1).
txt_page(888,497,1).
txt_page(799,497,1).
txt_page(290,497,1).
txt_page(41,497,1).
txt_page(858,498,1).
txt_page(446,498,1).
txt_page(301,498,1).
txt_page(585,498,1).
txt_page(839,498,1).
txt_page(406,498,1).
txt_page(826,498,1).
txt_page(532,498,1).
txt_page(149,498,1).
txt_page(493,498,1).
txt_page(259,498,1).
txt_page(174,498,1).
txt_page(955,498,1).
txt_page(574,498,2).
txt_page(781,498,1).
txt_page(911,498,1).
txt_page(947,498,1).
txt_page(936,498,1).
txt_page(43,498,1).
txt_page(553,499,1).
txt_page(703,499,1).
txt_page(309,499,1).
txt_page(420,499,1).
txt_page(236,499,1).
txt_page(847,499,1).
txt_page(66,499,1).
txt_page(552,499,1).
txt_page(555,499,1).
txt_page(294,499,1).
txt_page(525,499,1).
txt_page(423,499,1).
txt_page(670,499,1).
txt_page(168,499,1).
txt_page(181,499,1).
txt_page(733,499,1).
txt_page(809,499,1).
txt_page(4,499,1).
txt_page(651,499,1).
txt_page(127,499,1).
txt_page(491,500,1).
txt_page(869,500,1).
txt_page(200,500,1).
txt_page(336,500,1).
txt_page(463,500,1).
txt_page(23,500,1).
txt_page(215,500,1).
txt_page(528,500,1).
txt_page(352,500,1).
txt_page(306,500,1).
txt_page(173,500,1).
txt_page(191,500,1).
txt_page(65,500,1).
txt_page(535,500,1).
txt_page(505,500,1).
txt_page(786,500,1).
txt_page(631,500,1).
txt_page(98,500,1).
txt_page(79,500,1).
txt_page(359,500,1).
txt_page(391,501,1).
txt_page(379,501,1).
txt_page(137,501,1).
txt_page(863,501,1).
txt_page(460,501,1).
txt_page(10,501,1).
txt_page(518,501,1).
txt_page(707,501,1).
txt_page(258,501,1).
txt_page(155,501,1).
txt_page(566,501,1).
txt_page(25,501,1).
txt_page(830,501,1).
txt_page(341,501,1).
txt_page(410,501,1).
txt_page(463,501,1).
txt_page(656,501,1).
txt_page(725,501,1).
txt_page(839,501,1).
txt_page(849,501,1).
txt_page(138,502,1).
txt_page(616,502,1).
txt_page(560,502,1).
txt_page(743,502,1).
txt_page(44,502,1).
txt_page(573,502,1).
txt_page(255,502,1).
txt_page(420,502,1).
txt_page(909,502,1).
txt_page(661,502,1).
txt_page(115,502,1).
txt_page(287,502,1).
txt_page(719,502,1).
txt_page(466,502,1).
txt_page(812,502,1).
txt_page(408,502,1).
txt_page(213,502,1).
txt_page(206,502,1).
txt_page(500,502,1).
txt_page(563,502,1).
txt_page(304,503,1).
txt_page(101,503,1).
txt_page(383,503,1).
txt_page(123,503,1).
txt_page(423,503,1).
txt_page(324,503,1).
txt_page(209,503,1).
txt_page(412,503,1).
txt_page(479,503,1).
txt_page(227,503,1).
txt_page(112,503,1).
txt_page(257,503,1).
txt_page(526,503,1).
txt_page(643,503,1).
txt_page(447,503,1).
txt_page(902,503,1).
txt_page(664,503,1).
txt_page(495,503,1).
txt_page(458,503,1).
txt_page(730,503,1).
txt_page(853,504,1).
txt_page(774,504,1).
txt_page(9,504,1).
txt_page(334,504,1).
txt_page(323,504,1).
txt_page(950,504,1).
txt_page(35,504,2).
txt_page(728,504,1).
txt_page(687,504,1).
txt_page(543,504,1).
txt_page(835,504,1).
txt_page(883,504,1).
txt_page(587,504,1).
txt_page(911,504,1).
txt_page(880,504,1).
txt_page(339,504,1).
txt_page(176,504,1).
txt_page(97,504,1).
txt_page(515,504,1).
txt_page(834,505,1).
txt_page(252,505,1).
txt_page(618,505,1).
txt_page(610,505,1).
txt_page(409,505,1).
txt_page(825,505,1).
txt_page(743,505,1).
txt_page(324,505,1).
txt_page(848,505,1).
txt_page(920,505,1).
txt_page(384,505,1).
txt_page(452,505,1).
txt_page(660,505,1).
txt_page(475,505,1).
txt_page(554,505,1).
txt_page(531,505,1).
txt_page(374,505,1).
txt_page(558,505,1).
txt_page(418,505,1).
txt_page(710,505,1).
txt_page(497,506,1).
txt_page(176,506,1).
txt_page(20,506,1).
txt_page(512,506,1).
txt_page(943,506,1).
txt_page(324,506,1).
txt_page(743,506,1).
txt_page(779,506,1).
txt_page(930,506,1).
txt_page(681,506,1).
txt_page(659,506,1).
txt_page(111,506,1).
txt_page(660,506,1).
txt_page(290,506,1).
txt_page(304,506,1).
txt_page(529,506,1).
txt_page(341,506,1).
txt_page(947,506,1).
txt_page(824,506,1).
txt_page(955,506,1).
txt_page(886,507,1).
txt_page(423,507,1).
txt_page(783,507,1).
txt_page(104,507,1).
txt_page(124,507,1).
txt_page(679,507,1).
txt_page(670,507,1).
txt_page(390,507,1).
txt_page(633,507,1).
txt_page(774,507,1).
txt_page(506,507,1).
txt_page(29,507,1).
txt_page(636,507,1).
txt_page(312,507,1).
txt_page(664,507,1).
txt_page(916,507,1).
txt_page(234,507,1).
txt_page(446,507,1).
txt_page(483,507,1).
txt_page(190,507,1).
txt_page(815,508,1).
txt_page(419,508,1).
txt_page(755,508,1).
txt_page(69,508,1).
txt_page(830,508,1).
txt_page(857,508,1).
txt_page(626,508,1).
txt_page(224,508,1).
txt_page(313,508,1).
txt_page(189,508,1).
txt_page(356,508,1).
txt_page(489,508,1).
txt_page(127,508,1).
txt_page(775,508,1).
txt_page(150,508,1).
txt_page(544,508,1).
txt_page(920,508,1).
txt_page(855,508,1).
txt_page(110,508,1).
txt_page(756,508,1).
txt_page(30,509,1).
txt_page(308,509,1).
txt_page(555,509,1).
txt_page(508,509,1).
txt_page(949,509,1).
txt_page(765,509,1).
txt_page(643,509,1).
txt_page(483,509,1).
txt_page(268,509,1).
txt_page(7,509,1).
txt_page(363,509,1).
txt_page(549,509,1).
txt_page(171,509,1).
txt_page(693,509,1).
txt_page(521,509,1).
txt_page(205,509,1).
txt_page(76,509,1).
txt_page(832,509,1).
txt_page(962,509,1).
txt_page(785,509,1).
txt_page(515,510,1).
txt_page(8,510,1).
txt_page(750,510,1).
txt_page(200,510,1).
txt_page(853,510,1).
txt_page(572,510,1).
txt_page(105,510,1).
txt_page(952,510,1).
txt_page(872,510,1).
txt_page(651,510,1).
txt_page(473,510,1).
txt_page(786,510,1).
txt_page(165,510,1).
txt_page(518,510,1).
txt_page(431,510,1).
txt_page(5,510,1).
txt_page(187,510,1).
txt_page(340,510,1).
txt_page(120,510,1).
txt_page(278,510,1).
txt_page(98,511,1).
txt_page(445,511,1).
txt_page(448,511,1).
txt_page(308,511,1).
txt_page(124,511,1).
txt_page(90,511,1).
txt_page(26,511,1).
txt_page(788,511,1).
txt_page(865,511,1).
txt_page(962,511,1).
txt_page(83,511,1).
txt_page(368,511,1).
txt_page(722,511,1).
txt_page(683,511,1).
txt_page(910,511,1).
txt_page(529,511,1).
txt_page(432,511,1).
txt_page(1,511,1).
txt_page(61,511,1).
txt_page(606,511,1).
txt_page(29,512,1).
txt_page(156,512,1).
txt_page(228,512,1).
txt_page(602,512,1).
txt_page(143,512,1).
txt_page(635,512,1).
txt_page(558,512,1).
txt_page(7,512,1).
txt_page(690,512,1).
txt_page(388,512,1).
txt_page(421,512,1).
txt_page(947,512,1).
txt_page(124,512,2).
txt_page(48,512,1).
txt_page(711,512,1).
txt_page(631,512,1).
txt_page(503,512,1).
txt_page(798,512,1).
txt_page(874,512,1).
txt_page(522,513,1).
txt_page(479,513,1).
txt_page(341,513,1).
txt_page(332,513,1).
txt_page(728,513,1).
txt_page(702,513,1).
txt_page(355,513,1).
txt_page(389,513,1).
txt_page(315,513,1).
txt_page(767,513,1).
txt_page(49,513,1).
txt_page(82,513,1).
txt_page(647,513,1).
txt_page(33,513,1).
txt_page(171,513,1).
txt_page(705,513,1).
txt_page(954,513,1).
txt_page(750,513,1).
txt_page(523,513,1).
txt_page(325,513,1).
txt_page(550,514,1).
txt_page(532,514,1).
txt_page(933,514,1).
txt_page(463,514,1).
txt_page(812,514,1).
txt_page(742,514,1).
txt_page(682,514,1).
txt_page(603,514,1).
txt_page(270,514,1).
txt_page(858,514,1).
txt_page(95,514,1).
txt_page(727,514,1).
txt_page(46,514,1).
txt_page(860,514,1).
txt_page(530,514,1).
txt_page(516,514,1).
txt_page(380,514,1).
txt_page(448,514,1).
txt_page(519,514,1).
txt_page(691,514,1).
txt_page(409,515,1).
txt_page(705,515,1).
txt_page(557,515,1).
txt_page(429,515,1).
txt_page(606,515,1).
txt_page(753,515,1).
txt_page(47,515,1).
txt_page(683,515,1).
txt_page(570,515,1).
txt_page(210,515,1).
txt_page(932,515,1).
txt_page(77,515,1).
txt_page(692,515,1).
txt_page(213,515,1).
txt_page(728,515,1).
txt_page(357,515,1).
txt_page(825,515,1).
txt_page(903,515,1).
txt_page(507,515,1).
txt_page(101,515,1).
txt_page(333,516,1).
txt_page(354,516,1).
txt_page(645,516,1).
txt_page(97,516,1).
txt_page(111,516,1).
txt_page(753,516,1).
txt_page(169,516,1).
txt_page(614,516,1).
txt_page(656,516,1).
txt_page(22,516,1).
txt_page(829,516,1).
txt_page(930,516,1).
txt_page(605,516,1).
txt_page(4,516,1).
txt_page(860,516,1).
txt_page(811,516,1).
txt_page(85,516,1).
txt_page(277,516,1).
txt_page(728,516,1).
txt_page(87,516,1).
txt_page(176,517,1).
txt_page(821,517,1).
txt_page(82,517,1).
txt_page(523,517,1).
txt_page(555,517,1).
txt_page(656,517,1).
txt_page(358,517,1).
txt_page(547,517,1).
txt_page(623,517,1).
txt_page(606,517,1).
txt_page(44,517,1).
txt_page(62,517,1).
txt_page(175,517,1).
txt_page(64,517,1).
txt_page(533,517,1).
txt_page(589,517,1).
txt_page(239,517,1).
txt_page(958,517,1).
txt_page(209,517,1).
txt_page(357,517,1).
txt_page(460,518,1).
txt_page(397,518,1).
txt_page(705,518,1).
txt_page(962,518,1).
txt_page(275,518,1).
txt_page(183,518,1).
txt_page(170,518,1).
txt_page(897,518,1).
txt_page(698,518,1).
txt_page(510,518,1).
txt_page(438,518,1).
txt_page(540,518,1).
txt_page(148,518,1).
txt_page(607,518,1).
txt_page(652,518,1).
txt_page(648,518,1).
txt_page(947,518,1).
txt_page(63,518,1).
txt_page(625,518,1).
txt_page(699,518,1).
txt_page(632,519,1).
txt_page(631,519,1).
txt_page(332,519,1).
txt_page(745,519,1).
txt_page(500,519,1).
txt_page(503,519,1).
txt_page(16,519,1).
txt_page(648,519,1).
txt_page(566,519,1).
txt_page(633,519,1).
txt_page(951,519,1).
txt_page(290,519,1).
txt_page(640,519,1).
txt_page(214,519,1).
txt_page(687,519,1).
txt_page(75,519,1).
txt_page(163,519,1).
txt_page(172,519,1).
txt_page(744,519,2).
txt_page(473,520,1).
txt_page(85,520,1).
txt_page(470,520,1).
txt_page(732,520,1).
txt_page(735,520,1).
txt_page(241,520,1).
txt_page(458,520,1).
txt_page(427,520,1).
txt_page(222,520,1).
txt_page(497,520,1).
txt_page(49,520,1).
txt_page(463,520,1).
txt_page(255,520,1).
txt_page(769,520,1).
txt_page(134,520,2).
txt_page(326,520,1).
txt_page(437,520,1).
txt_page(109,520,1).
txt_page(526,520,1).
txt_page(256,521,1).
txt_page(298,521,1).
txt_page(494,521,2).
txt_page(638,521,1).
txt_page(409,521,1).
txt_page(841,521,1).
txt_page(78,521,1).
txt_page(756,521,1).
txt_page(567,521,1).
txt_page(708,521,1).
txt_page(599,521,1).
txt_page(653,521,1).
txt_page(299,521,1).
txt_page(60,521,1).
txt_page(963,521,1).
txt_page(858,521,1).
txt_page(890,521,1).
txt_page(115,521,1).
txt_page(683,521,1).
txt_page(661,522,1).
txt_page(598,522,1).
txt_page(476,522,1).
txt_page(820,522,1).
txt_page(318,522,1).
txt_page(522,522,1).
txt_page(26,522,1).
txt_page(172,522,1).
txt_page(410,522,1).
txt_page(407,522,1).
txt_page(640,522,1).
txt_page(962,522,1).
txt_page(728,522,1).
txt_page(623,522,1).
txt_page(440,522,1).
txt_page(319,522,1).
txt_page(492,522,1).
txt_page(447,522,1).
txt_page(72,522,1).
txt_page(23,522,1).
txt_page(410,523,1).
txt_page(289,523,1).
txt_page(720,523,1).
txt_page(121,523,1).
txt_page(585,523,1).
txt_page(472,523,1).
txt_page(214,523,1).
txt_page(875,523,1).
txt_page(790,523,1).
txt_page(249,523,1).
txt_page(97,523,1).
txt_page(550,523,1).
txt_page(647,523,1).
txt_page(110,523,1).
txt_page(926,523,1).
txt_page(674,523,1).
txt_page(210,523,1).
txt_page(28,523,1).
txt_page(519,523,1).
txt_page(71,523,1).
txt_page(872,524,1).
txt_page(694,524,1).
txt_page(768,524,1).
txt_page(99,524,1).
txt_page(435,524,1).
txt_page(878,524,1).
txt_page(741,524,1).
txt_page(479,524,1).
txt_page(258,524,1).
txt_page(611,524,1).
txt_page(603,524,1).
txt_page(600,524,1).
txt_page(332,524,1).
txt_page(314,524,1).
txt_page(81,524,1).
txt_page(428,524,1).
txt_page(568,524,1).
txt_page(630,524,1).
txt_page(497,524,1).
txt_page(557,524,1).
txt_page(506,525,1).
txt_page(801,525,1).
txt_page(277,525,1).
txt_page(782,525,1).
txt_page(236,525,2).
txt_page(508,525,1).
txt_page(806,525,1).
txt_page(478,525,1).
txt_page(140,525,1).
txt_page(467,525,1).
txt_page(432,525,1).
txt_page(434,525,1).
txt_page(113,525,1).
txt_page(896,525,1).
txt_page(567,525,1).
txt_page(699,525,1).
txt_page(335,525,1).
txt_page(31,525,1).
txt_page(351,525,1).
txt_page(892,526,1).
txt_page(734,526,1).
txt_page(792,526,1).
txt_page(181,526,1).
txt_page(610,526,1).
txt_page(783,526,1).
txt_page(879,526,1).
txt_page(731,526,1).
txt_page(535,526,1).
txt_page(380,526,1).
txt_page(900,526,1).
txt_page(643,526,1).
txt_page(953,526,1).
txt_page(777,526,1).
txt_page(471,526,1).
txt_page(798,526,1).
txt_page(268,526,1).
txt_page(306,526,1).
txt_page(239,526,1).
txt_page(172,526,1).
txt_page(134,527,1).
txt_page(780,527,1).
txt_page(158,527,1).
txt_page(259,527,1).
txt_page(952,527,1).
txt_page(907,527,1).
txt_page(241,527,1).
txt_page(413,527,1).
txt_page(13,527,1).
txt_page(262,527,1).
txt_page(194,527,1).
txt_page(962,527,1).
txt_page(701,527,1).
txt_page(961,527,1).
txt_page(703,527,1).
txt_page(667,527,1).
txt_page(464,527,1).
txt_page(462,527,1).
txt_page(875,527,1).
txt_page(681,527,1).
txt_page(873,528,1).
txt_page(268,528,1).
txt_page(544,528,1).
txt_page(944,528,1).
txt_page(15,528,1).
txt_page(251,528,1).
txt_page(558,528,1).
txt_page(613,528,1).
txt_page(352,528,1).
txt_page(915,528,1).
txt_page(366,528,1).
txt_page(530,528,1).
txt_page(134,528,1).
txt_page(815,528,1).
txt_page(640,528,1).
txt_page(900,528,1).
txt_page(535,528,1).
txt_page(880,528,1).
txt_page(672,528,1).
txt_page(853,528,1).
txt_page(115,529,1).
txt_page(776,529,1).
txt_page(193,529,1).
txt_page(826,529,1).
txt_page(923,529,1).
txt_page(83,529,1).
txt_page(33,529,1).
txt_page(3,529,1).
txt_page(887,529,1).
txt_page(105,529,1).
txt_page(934,529,1).
txt_page(226,529,1).
txt_page(64,529,1).
txt_page(296,529,1).
txt_page(491,529,1).
txt_page(13,529,1).
txt_page(113,529,1).
txt_page(568,529,1).
txt_page(640,529,1).
txt_page(370,529,1).
txt_page(75,530,1).
txt_page(487,530,1).
txt_page(473,530,1).
txt_page(509,530,1).
txt_page(840,530,1).
txt_page(629,530,1).
txt_page(161,530,1).
txt_page(950,530,1).
txt_page(178,530,1).
txt_page(361,530,1).
txt_page(383,530,1).
txt_page(680,530,1).
txt_page(471,530,1).
txt_page(44,530,1).
txt_page(79,530,1).
txt_page(272,530,1).
txt_page(260,530,1).
txt_page(108,530,1).
txt_page(17,530,1).
txt_page(514,530,1).
txt_page(424,531,1).
txt_page(481,531,1).
txt_page(762,531,1).
txt_page(330,531,1).
txt_page(255,531,1).
txt_page(929,531,1).
txt_page(493,531,1).
txt_page(9,531,1).
txt_page(821,531,1).
txt_page(531,531,1).
txt_page(777,531,1).
txt_page(884,531,1).
txt_page(46,531,1).
txt_page(231,531,1).
txt_page(892,531,1).
txt_page(907,531,1).
txt_page(579,531,1).
txt_page(580,531,1).
txt_page(460,531,1).
txt_page(24,531,1).
txt_page(175,532,1).
txt_page(268,532,1).
txt_page(130,532,1).
txt_page(88,532,1).
txt_page(276,532,1).
txt_page(136,532,1).
txt_page(916,532,1).
txt_page(249,532,1).
txt_page(140,532,1).
txt_page(129,532,1).
txt_page(127,532,1).
txt_page(635,532,1).
txt_page(31,532,1).
txt_page(419,532,1).
txt_page(869,532,1).
txt_page(230,532,1).
txt_page(945,532,1).
txt_page(288,532,1).
txt_page(806,532,1).
txt_page(891,532,1).
txt_page(210,533,1).
txt_page(434,533,1).
txt_page(764,533,1).
txt_page(823,533,1).
txt_page(360,533,1).
txt_page(591,533,1).
txt_page(121,533,1).
txt_page(403,533,1).
txt_page(643,533,1).
txt_page(675,533,1).
txt_page(161,533,1).
txt_page(259,533,1).
txt_page(380,533,1).
txt_page(495,533,1).
txt_page(267,533,1).
txt_page(841,533,1).
txt_page(714,533,1).
txt_page(669,533,1).
txt_page(177,533,1).
txt_page(922,533,1).
txt_page(861,534,1).
txt_page(300,534,1).
txt_page(678,534,1).
txt_page(529,534,1).
txt_page(99,534,1).
txt_page(169,534,1).
txt_page(583,534,1).
txt_page(410,534,1).
txt_page(603,534,1).
txt_page(605,534,1).
txt_page(856,534,1).
txt_page(869,534,1).
txt_page(574,534,1).
txt_page(914,534,1).
txt_page(439,534,1).
txt_page(709,534,1).
txt_page(835,534,1).
txt_page(283,534,1).
txt_page(332,534,1).
txt_page(558,534,1).
txt_page(743,535,1).
txt_page(5,535,1).
txt_page(922,535,1).
txt_page(233,535,1).
txt_page(898,535,1).
txt_page(38,535,1).
txt_page(669,535,1).
txt_page(359,535,1).
txt_page(229,535,1).
txt_page(70,535,1).
txt_page(486,535,1).
txt_page(538,535,1).
txt_page(714,535,1).
txt_page(909,535,1).
txt_page(231,535,1).
txt_page(791,535,1).
txt_page(759,535,1).
txt_page(904,535,1).
txt_page(350,535,1).
txt_page(165,535,1).
txt_page(914,536,1).
txt_page(225,536,1).
txt_page(62,536,1).
txt_page(710,536,1).
txt_page(610,536,1).
txt_page(951,536,1).
txt_page(52,536,1).
txt_page(652,536,1).
txt_page(604,536,1).
txt_page(373,536,1).
txt_page(168,536,1).
txt_page(263,536,1).
txt_page(895,536,1).
txt_page(299,536,1).
txt_page(681,536,1).
txt_page(164,536,1).
txt_page(805,536,1).
txt_page(348,536,1).
txt_page(166,536,1).
txt_page(550,536,1).
txt_page(690,537,1).
txt_page(115,537,1).
txt_page(264,537,1).
txt_page(869,537,1).
txt_page(518,537,1).
txt_page(921,537,1).
txt_page(839,537,1).
txt_page(506,537,1).
txt_page(146,537,1).
txt_page(830,537,1).
txt_page(89,537,1).
txt_page(880,537,1).
txt_page(575,537,1).
txt_page(676,537,1).
txt_page(351,537,1).
txt_page(752,537,1).
txt_page(700,537,1).
txt_page(345,537,1).
txt_page(184,537,1).
txt_page(61,537,1).
txt_page(442,538,1).
txt_page(298,538,1).
txt_page(467,538,1).
txt_page(448,538,1).
txt_page(548,538,1).
txt_page(525,538,1).
txt_page(632,538,1).
txt_page(496,538,1).
txt_page(208,538,1).
txt_page(748,538,1).
txt_page(799,538,1).
txt_page(873,538,1).
txt_page(763,538,1).
txt_page(669,538,1).
txt_page(638,538,1).
txt_page(340,538,1).
txt_page(483,538,1).
txt_page(323,538,1).
txt_page(343,538,1).
txt_page(11,538,1).
txt_page(795,539,1).
txt_page(16,539,1).
txt_page(869,539,1).
txt_page(39,539,1).
txt_page(426,539,1).
txt_page(244,539,1).
txt_page(185,539,1).
txt_page(797,539,1).
txt_page(456,539,1).
txt_page(395,539,1).
txt_page(761,539,1).
txt_page(770,539,1).
txt_page(584,539,1).
txt_page(864,539,1).
txt_page(184,539,1).
txt_page(172,539,1).
txt_page(455,539,1).
txt_page(264,539,2).
txt_page(148,539,1).
txt_page(337,540,1).
txt_page(947,540,1).
txt_page(937,540,1).
txt_page(342,540,1).
txt_page(180,540,1).
txt_page(600,540,1).
txt_page(56,540,1).
txt_page(503,540,1).
txt_page(543,540,1).
txt_page(8,540,1).
txt_page(261,540,1).
txt_page(117,540,1).
txt_page(441,540,1).
txt_page(875,540,1).
txt_page(696,540,1).
txt_page(619,540,1).
txt_page(61,540,1).
txt_page(693,540,1).
txt_page(920,540,1).
txt_page(459,540,1).
txt_page(647,541,1).
txt_page(333,541,1).
txt_page(588,541,1).
txt_page(527,541,1).
txt_page(306,541,1).
txt_page(151,541,1).
txt_page(400,541,1).
txt_page(65,541,1).
txt_page(637,541,1).
txt_page(110,541,1).
txt_page(318,541,1).
txt_page(734,541,1).
txt_page(198,541,1).
txt_page(459,541,1).
txt_page(968,541,1).
txt_page(216,541,1).
txt_page(892,541,1).
txt_page(513,541,1).
txt_page(325,541,1).
txt_page(206,541,1).
txt_page(771,542,1).
txt_page(618,542,1).
txt_page(717,542,1).
txt_page(216,542,1).
txt_page(633,542,1).
txt_page(894,542,1).
txt_page(402,542,1).
txt_page(566,542,1).
txt_page(944,542,1).
txt_page(483,542,1).
txt_page(814,542,1).
txt_page(303,542,1).
txt_page(667,542,1).
txt_page(816,542,1).
txt_page(544,542,1).
txt_page(741,542,1).
txt_page(269,542,1).
txt_page(460,542,2).
txt_page(61,542,1).
txt_page(871,543,1).
txt_page(430,543,1).
txt_page(589,543,1).
txt_page(770,543,1).
txt_page(932,543,1).
txt_page(270,543,1).
txt_page(744,543,1).
txt_page(137,543,1).
txt_page(730,543,1).
txt_page(45,543,1).
txt_page(361,543,1).
txt_page(196,543,1).
txt_page(415,543,1).
txt_page(860,543,1).
txt_page(328,543,1).
txt_page(594,543,1).
txt_page(714,543,1).
txt_page(387,543,1).
txt_page(743,543,1).
txt_page(572,543,1).
txt_page(37,544,1).
txt_page(165,544,1).
txt_page(395,544,1).
txt_page(502,544,1).
txt_page(125,544,1).
txt_page(354,544,1).
txt_page(225,544,1).
txt_page(503,544,1).
txt_page(753,544,1).
txt_page(250,544,1).
txt_page(269,544,1).
txt_page(609,544,1).
txt_page(398,544,1).
txt_page(789,544,1).
txt_page(805,544,1).
txt_page(679,544,1).
txt_page(938,544,1).
txt_page(460,544,1).
txt_page(894,544,1).
txt_page(778,544,1).
txt_page(398,545,1).
txt_page(841,545,1).
txt_page(384,545,1).
txt_page(261,545,1).
txt_page(894,545,1).
txt_page(594,545,1).
txt_page(547,545,1).
txt_page(733,545,1).
txt_page(909,545,1).
txt_page(557,545,1).
txt_page(882,545,1).
txt_page(628,545,1).
txt_page(569,545,1).
txt_page(89,545,1).
txt_page(316,545,1).
txt_page(28,545,1).
txt_page(117,545,1).
txt_page(871,545,1).
txt_page(233,545,1).
txt_page(853,545,1).
txt_page(869,546,1).
txt_page(750,546,1).
txt_page(516,546,1).
txt_page(112,546,1).
txt_page(152,546,1).
txt_page(645,546,1).
txt_page(216,546,1).
txt_page(704,546,1).
txt_page(634,546,1).
txt_page(57,546,1).
txt_page(734,546,1).
txt_page(887,546,1).
txt_page(82,546,1).
txt_page(234,546,1).
txt_page(607,546,1).
txt_page(30,546,1).
txt_page(638,546,1).
txt_page(485,546,1).
txt_page(442,546,1).
txt_page(840,546,1).
txt_page(456,547,1).
txt_page(118,547,1).
txt_page(667,547,1).
txt_page(370,547,1).
txt_page(741,547,1).
txt_page(597,547,1).
txt_page(895,547,1).
txt_page(396,547,1).
txt_page(612,547,1).
txt_page(256,547,1).
txt_page(762,547,1).
txt_page(303,547,1).
txt_page(287,547,1).
txt_page(700,547,1).
txt_page(355,547,1).
txt_page(683,547,1).
txt_page(114,547,1).
txt_page(931,547,1).
txt_page(270,547,1).
txt_page(898,547,1).
txt_page(882,548,1).
txt_page(593,548,1).
txt_page(37,548,1).
txt_page(851,548,1).
txt_page(671,548,1).
txt_page(278,548,1).
txt_page(693,548,1).
txt_page(55,548,1).
txt_page(656,548,1).
txt_page(628,548,1).
txt_page(777,548,1).
txt_page(339,548,1).
txt_page(4,548,1).
txt_page(633,548,1).
txt_page(426,548,1).
txt_page(782,548,1).
txt_page(548,548,1).
txt_page(348,548,1).
txt_page(916,548,1).
txt_page(263,548,1).
txt_page(423,549,1).
txt_page(814,549,1).
txt_page(28,549,1).
txt_page(608,549,1).
txt_page(968,549,1).
txt_page(930,549,1).
txt_page(751,549,1).
txt_page(447,549,1).
txt_page(383,549,1).
txt_page(869,549,1).
txt_page(185,549,1).
txt_page(793,549,1).
txt_page(403,549,1).
txt_page(284,549,1).
txt_page(310,549,1).
txt_page(661,549,1).
txt_page(401,549,1).
txt_page(701,549,1).
txt_page(476,549,1).
txt_page(740,549,1).
txt_page(26,550,1).
txt_page(563,550,1).
txt_page(460,550,1).
txt_page(561,550,1).
txt_page(32,550,1).
txt_page(720,550,1).
txt_page(680,550,1).
txt_page(14,550,1).
txt_page(75,550,1).
txt_page(25,550,1).
txt_page(48,550,1).
txt_page(128,550,1).
txt_page(331,550,1).
txt_page(17,550,1).
txt_page(664,550,1).
txt_page(741,550,1).
txt_page(925,550,1).
txt_page(704,550,1).
txt_page(442,550,1).
txt_page(941,550,1).
txt_page(447,551,1).
txt_page(346,551,1).
txt_page(448,551,1).
txt_page(193,551,1).
txt_page(418,551,1).
txt_page(616,551,1).
txt_page(248,551,1).
txt_page(511,551,1).
txt_page(957,551,1).
txt_page(455,551,1).
txt_page(484,551,1).
txt_page(590,551,1).
txt_page(172,551,1).
txt_page(210,551,1).
txt_page(362,551,1).
txt_page(515,551,1).
txt_page(610,551,1).
txt_page(912,551,1).
txt_page(336,551,1).
txt_page(236,551,1).
txt_page(425,552,1).
txt_page(303,552,1).
txt_page(526,552,1).
txt_page(78,552,1).
txt_page(787,552,1).
txt_page(884,552,1).
txt_page(413,552,1).
txt_page(295,552,1).
txt_page(412,552,1).
txt_page(261,552,1).
txt_page(472,552,1).
txt_page(84,552,1).
txt_page(55,552,1).
txt_page(601,552,1).
txt_page(180,552,1).
txt_page(600,552,1).
txt_page(522,552,1).
txt_page(653,552,1).
txt_page(382,552,1).
txt_page(650,552,1).
txt_page(284,553,1).
txt_page(594,553,1).
txt_page(821,553,1).
txt_page(50,553,1).
txt_page(19,553,1).
txt_page(254,553,1).
txt_page(470,553,1).
txt_page(738,553,1).
txt_page(108,553,1).
txt_page(294,553,1).
txt_page(622,553,1).
txt_page(855,553,1).
txt_page(2,553,1).
txt_page(369,553,1).
txt_page(838,553,1).
txt_page(897,553,1).
txt_page(116,553,1).
txt_page(706,553,1).
txt_page(769,553,1).
txt_page(16,553,1).
txt_page(217,554,1).
txt_page(146,554,1).
txt_page(376,554,1).
txt_page(256,554,1).
txt_page(220,554,1).
txt_page(801,554,1).
txt_page(493,554,1).
txt_page(22,554,1).
txt_page(823,554,1).
txt_page(850,554,1).
txt_page(924,554,1).
txt_page(965,554,1).
txt_page(656,554,1).
txt_page(261,554,1).
txt_page(824,554,1).
txt_page(783,554,1).
txt_page(918,554,1).
txt_page(546,554,1).
txt_page(531,554,1).
txt_page(224,554,1).
txt_page(73,555,1).
txt_page(882,555,1).
txt_page(710,555,1).
txt_page(846,555,1).
txt_page(194,555,1).
txt_page(364,555,1).
txt_page(289,555,1).
txt_page(302,555,1).
txt_page(481,555,1).
txt_page(225,555,1).
txt_page(191,555,1).
txt_page(915,555,1).
txt_page(261,555,1).
txt_page(727,555,1).
txt_page(26,555,1).
txt_page(5,555,1).
txt_page(13,555,1).
txt_page(672,555,1).
txt_page(742,555,1).
txt_page(41,555,1).
txt_page(155,556,1).
txt_page(391,556,1).
txt_page(98,556,1).
txt_page(769,556,1).
txt_page(565,556,1).
txt_page(789,556,1).
txt_page(944,556,1).
txt_page(872,556,1).
txt_page(512,556,1).
txt_page(863,556,1).
txt_page(817,556,1).
txt_page(464,556,1).
txt_page(344,556,1).
txt_page(883,556,1).
txt_page(57,556,1).
txt_page(22,556,1).
txt_page(601,556,1).
txt_page(521,556,1).
txt_page(405,556,1).
txt_page(292,556,1).
txt_page(806,557,1).
txt_page(840,557,1).
txt_page(421,557,1).
txt_page(517,557,1).
txt_page(950,557,1).
txt_page(448,557,1).
txt_page(874,557,1).
txt_page(855,557,1).
txt_page(264,557,1).
txt_page(954,557,1).
txt_page(769,557,1).
txt_page(443,557,1).
txt_page(552,557,1).
txt_page(161,557,1).
txt_page(201,557,1).
txt_page(276,557,1).
txt_page(889,557,1).
txt_page(1,557,1).
txt_page(566,557,1).
txt_page(942,557,1).
txt_page(233,558,1).
txt_page(639,558,1).
txt_page(566,558,1).
txt_page(623,558,1).
txt_page(344,558,1).
txt_page(535,558,1).
txt_page(178,558,1).
txt_page(703,558,1).
txt_page(768,558,1).
txt_page(349,558,1).
txt_page(763,558,1).
txt_page(914,558,1).
txt_page(355,558,1).
txt_page(965,558,1).
txt_page(275,558,1).
txt_page(265,558,1).
txt_page(4,558,1).
txt_page(72,558,1).
txt_page(278,558,1).
txt_page(664,558,1).
txt_page(155,559,1).
txt_page(236,559,1).
txt_page(884,559,1).
txt_page(299,559,1).
txt_page(56,559,1).
txt_page(220,559,1).
txt_page(577,559,1).
txt_page(336,559,1).
txt_page(686,559,1).
txt_page(432,559,1).
txt_page(961,559,1).
txt_page(392,559,1).
txt_page(780,559,1).
txt_page(104,559,1).
txt_page(421,559,1).
txt_page(804,559,1).
txt_page(272,559,1).
txt_page(695,559,1).
txt_page(546,559,1).
txt_page(589,559,1).
txt_page(167,560,1).
txt_page(314,560,1).
txt_page(376,560,1).
txt_page(294,560,1).
txt_page(887,560,1).
txt_page(741,560,1).
txt_page(617,560,1).
txt_page(385,560,1).
txt_page(743,560,1).
txt_page(823,560,1).
txt_page(702,560,1).
txt_page(132,560,1).
txt_page(718,560,1).
txt_page(253,560,1).
txt_page(96,560,1).
txt_page(175,560,1).
txt_page(489,560,1).
txt_page(960,560,1).
txt_page(272,560,1).
txt_page(943,560,1).
txt_page(504,561,1).
txt_page(6,561,1).
txt_page(291,561,1).
txt_page(685,561,1).
txt_page(585,561,1).
txt_page(381,561,1).
txt_page(607,561,1).
txt_page(322,561,1).
txt_page(747,561,1).
txt_page(458,561,1).
txt_page(714,561,1).
txt_page(544,561,1).
txt_page(815,561,1).
txt_page(673,561,1).
txt_page(273,561,1).
txt_page(454,561,1).
txt_page(111,561,1).
txt_page(103,561,1).
txt_page(662,561,1).
txt_page(106,561,1).
txt_page(357,562,1).
txt_page(767,562,1).
txt_page(393,562,1).
txt_page(702,562,1).
txt_page(248,562,1).
txt_page(794,562,1).
txt_page(787,562,1).
txt_page(128,562,1).
txt_page(671,562,1).
txt_page(708,562,1).
txt_page(184,562,1).
txt_page(930,562,1).
txt_page(141,562,1).
txt_page(506,562,1).
txt_page(539,562,1).
txt_page(682,562,1).
txt_page(639,562,1).
txt_page(154,562,1).
txt_page(193,562,1).
txt_page(183,562,1).
txt_page(406,563,1).
txt_page(112,563,1).
txt_page(419,563,1).
txt_page(685,563,1).
txt_page(399,563,1).
txt_page(861,563,1).
txt_page(23,563,1).
txt_page(31,563,1).
txt_page(173,563,1).
txt_page(349,563,1).
txt_page(629,563,1).
txt_page(809,563,1).
txt_page(293,563,1).
txt_page(586,563,1).
txt_page(865,563,1).
txt_page(133,563,1).
txt_page(512,563,1).
txt_page(170,563,1).
txt_page(24,563,1).
txt_page(777,563,1).
txt_page(716,564,1).
txt_page(889,564,1).
txt_page(361,564,1).
txt_page(64,564,1).
txt_page(440,564,1).
txt_page(674,564,1).
txt_page(963,564,1).
txt_page(698,564,1).
txt_page(14,564,1).
txt_page(823,564,1).
txt_page(580,564,1).
txt_page(667,564,1).
txt_page(216,564,1).
txt_page(111,564,1).
txt_page(732,564,1).
txt_page(613,564,1).
txt_page(874,564,1).
txt_page(290,564,2).
txt_page(519,564,1).
txt_page(662,565,1).
txt_page(32,565,1).
txt_page(562,565,1).
txt_page(28,565,1).
txt_page(490,565,1).
txt_page(819,565,1).
txt_page(697,565,1).
txt_page(746,565,1).
txt_page(13,565,1).
txt_page(771,565,1).
txt_page(869,565,1).
txt_page(155,565,1).
txt_page(489,565,1).
txt_page(851,565,1).
txt_page(687,565,1).
txt_page(785,565,1).
txt_page(170,565,1).
txt_page(856,565,1).
txt_page(773,565,1).
txt_page(204,565,1).
txt_page(881,566,1).
txt_page(460,566,1).
txt_page(868,566,1).
txt_page(661,566,1).
txt_page(704,566,1).
txt_page(204,566,1).
txt_page(434,566,1).
txt_page(274,566,1).
txt_page(208,566,1).
txt_page(301,566,1).
txt_page(304,566,1).
txt_page(392,566,1).
txt_page(839,566,1).
txt_page(178,566,1).
txt_page(665,566,1).
txt_page(966,566,1).
txt_page(242,566,1).
txt_page(738,566,1).
txt_page(335,566,1).
txt_page(831,566,1).
txt_page(967,567,1).
txt_page(834,567,1).
txt_page(634,567,1).
txt_page(414,567,1).
txt_page(437,567,1).
txt_page(248,567,1).
txt_page(576,567,1).
txt_page(473,567,1).
txt_page(274,567,1).
txt_page(390,567,1).
txt_page(130,567,1).
txt_page(526,567,1).
txt_page(456,567,1).
txt_page(810,567,1).
txt_page(830,567,1).
txt_page(83,567,1).
txt_page(549,567,1).
txt_page(862,567,1).
txt_page(184,567,1).
txt_page(406,567,1).
txt_page(336,568,1).
txt_page(319,568,1).
txt_page(395,568,1).
txt_page(810,568,1).
txt_page(917,568,1).
txt_page(904,568,1).
txt_page(701,568,1).
txt_page(261,568,1).
txt_page(340,568,1).
txt_page(318,568,1).
txt_page(945,568,1).
txt_page(668,568,1).
txt_page(426,568,1).
txt_page(11,568,1).
txt_page(122,568,1).
txt_page(637,568,1).
txt_page(440,568,1).
txt_page(776,568,1).
txt_page(113,568,1).
txt_page(68,568,1).
txt_page(851,569,1).
txt_page(365,569,1).
txt_page(502,569,1).
txt_page(73,569,1).
txt_page(80,569,1).
txt_page(475,569,1).
txt_page(131,569,1).
txt_page(904,569,1).
txt_page(226,569,1).
txt_page(515,569,1).
txt_page(6,569,1).
txt_page(426,569,1).
txt_page(421,569,1).
txt_page(432,569,1).
txt_page(403,569,1).
txt_page(171,569,1).
txt_page(938,569,1).
txt_page(593,569,1).
txt_page(910,569,1).
txt_page(964,569,1).
txt_page(218,570,1).
txt_page(941,570,1).
txt_page(727,570,1).
txt_page(409,570,1).
txt_page(919,570,1).
txt_page(817,570,1).
txt_page(447,570,1).
txt_page(456,570,1).
txt_page(79,570,1).
txt_page(938,570,1).
txt_page(214,570,1).
txt_page(322,570,1).
txt_page(647,570,1).
txt_page(93,570,1).
txt_page(448,570,1).
txt_page(366,570,1).
txt_page(256,570,1).
txt_page(176,570,1).
txt_page(315,570,1).
txt_page(618,570,1).
txt_page(226,571,1).
txt_page(903,571,1).
txt_page(691,571,1).
txt_page(752,571,1).
txt_page(520,571,1).
txt_page(115,571,1).
txt_page(264,571,1).
txt_page(247,571,1).
txt_page(631,571,1).
txt_page(24,571,1).
txt_page(59,571,1).
txt_page(858,571,1).
txt_page(257,571,1).
txt_page(472,571,1).
txt_page(762,571,1).
txt_page(966,571,1).
txt_page(413,571,1).
txt_page(656,571,1).
txt_page(510,571,1).
txt_page(439,571,1).
txt_page(792,572,1).
txt_page(810,572,1).
txt_page(355,572,1).
txt_page(589,572,1).
txt_page(372,572,1).
txt_page(279,572,1).
txt_page(609,572,1).
txt_page(880,572,1).
txt_page(350,572,1).
txt_page(688,572,1).
txt_page(612,572,1).
txt_page(41,572,1).
txt_page(918,572,1).
txt_page(957,572,1).
txt_page(441,572,1).
txt_page(901,572,1).
txt_page(10,572,1).
txt_page(33,572,1).
txt_page(323,572,1).
txt_page(200,572,1).
txt_page(755,573,1).
txt_page(344,573,1).
txt_page(805,573,1).
txt_page(176,573,1).
txt_page(58,573,1).
txt_page(624,573,1).
txt_page(575,573,1).
txt_page(632,573,1).
txt_page(802,573,1).
txt_page(56,573,1).
txt_page(456,573,1).
txt_page(694,573,1).
txt_page(730,573,1).
txt_page(814,573,1).
txt_page(164,573,1).
txt_page(22,573,1).
txt_page(347,573,1).
txt_page(511,573,1).
txt_page(80,573,1).
txt_page(666,573,1).
txt_page(553,574,1).
txt_page(868,574,1).
txt_page(406,574,1).
txt_page(404,574,1).
txt_page(258,574,1).
txt_page(278,574,1).
txt_page(770,574,1).
txt_page(837,574,1).
txt_page(427,574,1).
txt_page(933,574,1).
txt_page(246,574,1).
txt_page(574,574,1).
txt_page(188,574,1).
txt_page(593,574,2).
txt_page(156,574,1).
txt_page(902,574,1).
txt_page(958,574,1).
txt_page(330,574,1).
txt_page(511,574,1).
txt_page(574,575,1).
txt_page(508,575,1).
txt_page(945,575,1).
txt_page(115,575,1).
txt_page(223,575,1).
txt_page(796,575,1).
txt_page(510,575,1).
txt_page(924,575,1).
txt_page(355,575,1).
txt_page(16,575,1).
txt_page(819,575,1).
txt_page(730,575,1).
txt_page(769,575,1).
txt_page(36,575,1).
txt_page(719,575,1).
txt_page(793,575,1).
txt_page(13,575,1).
txt_page(85,575,1).
txt_page(682,575,1).
txt_page(872,575,1).
txt_page(311,576,1).
txt_page(45,576,1).
txt_page(813,576,1).
txt_page(941,576,1).
txt_page(685,576,1).
txt_page(192,576,1).
txt_page(500,576,1).
txt_page(660,576,1).
txt_page(309,576,1).
txt_page(320,576,1).
txt_page(487,576,1).
txt_page(845,576,1).
txt_page(193,576,1).
txt_page(124,576,1).
txt_page(651,576,1).
txt_page(617,576,1).
txt_page(910,576,1).
txt_page(867,576,1).
txt_page(721,576,1).
txt_page(215,576,1).
txt_page(240,577,1).
txt_page(235,577,1).
txt_page(891,577,1).
txt_page(926,577,1).
txt_page(175,577,1).
txt_page(319,577,1).
txt_page(254,577,1).
txt_page(505,577,1).
txt_page(586,577,1).
txt_page(567,577,1).
txt_page(52,577,1).
txt_page(77,577,1).
txt_page(943,577,1).
txt_page(490,577,1).
txt_page(501,577,1).
txt_page(53,577,1).
txt_page(846,577,1).
txt_page(588,577,1).
txt_page(251,577,2).
txt_page(469,578,1).
txt_page(412,578,1).
txt_page(346,578,1).
txt_page(343,578,1).
txt_page(364,578,1).
txt_page(197,578,1).
txt_page(234,578,1).
txt_page(402,578,1).
txt_page(951,578,1).
txt_page(611,578,1).
txt_page(134,578,1).
txt_page(475,578,1).
txt_page(763,578,1).
txt_page(16,578,1).
txt_page(610,578,1).
txt_page(290,578,1).
txt_page(710,578,1).
txt_page(330,578,1).
txt_page(218,578,1).
txt_page(268,578,1).
txt_page(100,579,1).
txt_page(382,579,1).
txt_page(64,579,1).
txt_page(868,579,1).
txt_page(767,579,1).
txt_page(285,579,1).
txt_page(637,579,1).
txt_page(326,579,1).
txt_page(324,579,1).
txt_page(966,579,1).
txt_page(620,579,1).
txt_page(172,579,1).
txt_page(54,579,1).
txt_page(341,579,1).
txt_page(816,579,1).
txt_page(921,579,1).
txt_page(441,579,1).
txt_page(864,579,1).
txt_page(660,579,1).
txt_page(923,579,1).
txt_page(370,580,1).
txt_page(54,580,1).
txt_page(610,580,1).
txt_page(387,580,1).
txt_page(358,580,1).
txt_page(380,580,1).
txt_page(675,580,1).
txt_page(121,580,1).
txt_page(716,580,1).
txt_page(512,580,1).
txt_page(860,580,1).
txt_page(280,580,1).
txt_page(638,580,1).
txt_page(41,580,1).
txt_page(807,580,1).
txt_page(847,580,1).
txt_page(240,580,1).
txt_page(603,580,1).
txt_page(656,580,1).
txt_page(815,580,1).
txt_page(479,581,1).
txt_page(151,581,1).
txt_page(831,581,1).
txt_page(133,581,1).
txt_page(871,581,1).
txt_page(70,581,1).
txt_page(288,581,1).
txt_page(933,581,1).
txt_page(358,581,1).
txt_page(899,581,1).
txt_page(255,581,1).
txt_page(945,581,1).
txt_page(542,581,1).
txt_page(936,581,1).
txt_page(806,581,1).
txt_page(886,581,1).
txt_page(874,581,1).
txt_page(934,581,1).
txt_page(370,581,1).
txt_page(342,581,1).
txt_page(527,582,1).
txt_page(257,582,1).
txt_page(348,582,1).
txt_page(81,582,1).
txt_page(234,582,1).
txt_page(344,582,1).
txt_page(108,582,1).
txt_page(730,582,1).
txt_page(268,582,2).
txt_page(859,582,1).
txt_page(548,582,1).
txt_page(703,582,1).
txt_page(894,582,1).
txt_page(330,582,1).
txt_page(908,582,1).
txt_page(803,582,1).
txt_page(667,582,1).
txt_page(521,582,1).
txt_page(582,582,1).
txt_page(765,583,1).
txt_page(166,583,1).
txt_page(627,583,1).
txt_page(433,583,1).
txt_page(252,583,1).
txt_page(62,583,1).
txt_page(117,583,1).
txt_page(394,583,1).
txt_page(494,583,1).
txt_page(42,583,1).
txt_page(343,583,1).
txt_page(532,583,1).
txt_page(710,583,1).
txt_page(414,583,1).
txt_page(965,583,1).
txt_page(741,583,1).
txt_page(893,583,1).
txt_page(862,583,1).
txt_page(509,583,1).
txt_page(676,583,1).
txt_page(231,584,1).
txt_page(490,584,1).
txt_page(63,584,1).
txt_page(272,584,1).
txt_page(461,584,1).
txt_page(157,584,1).
txt_page(383,584,1).
txt_page(675,584,1).
txt_page(616,584,1).
txt_page(868,584,1).
txt_page(795,584,1).
txt_page(661,584,1).
txt_page(887,584,1).
txt_page(375,584,1).
txt_page(670,584,1).
txt_page(243,584,1).
txt_page(426,584,1).
txt_page(386,584,1).
txt_page(187,584,1).
txt_page(912,584,1).
txt_page(483,585,1).
txt_page(198,585,1).
txt_page(855,585,1).
txt_page(545,585,1).
txt_page(531,585,1).
txt_page(351,585,1).
txt_page(720,585,1).
txt_page(902,585,1).
txt_page(882,585,1).
txt_page(672,585,1).
txt_page(234,585,1).
txt_page(859,585,1).
txt_page(529,585,1).
txt_page(276,585,1).
txt_page(657,585,1).
txt_page(578,585,1).
txt_page(956,585,1).
txt_page(923,585,1).
txt_page(492,585,1).
txt_page(98,585,1).
txt_page(936,586,1).
txt_page(110,586,1).
txt_page(154,586,1).
txt_page(898,586,1).
txt_page(393,586,1).
txt_page(774,586,1).
txt_page(406,586,1).
txt_page(29,586,1).
txt_page(591,586,1).
txt_page(620,586,1).
txt_page(598,586,1).
txt_page(300,586,1).
txt_page(81,586,1).
txt_page(707,586,1).
txt_page(491,586,1).
txt_page(468,586,1).
txt_page(593,586,1).
txt_page(172,586,1).
txt_page(119,586,1).
txt_page(205,586,1).
txt_page(890,587,1).
txt_page(746,587,1).
txt_page(390,587,1).
txt_page(684,587,1).
txt_page(439,587,1).
txt_page(172,587,1).
txt_page(355,587,1).
txt_page(515,587,1).
txt_page(825,587,1).
txt_page(639,587,1).
txt_page(331,587,1).
txt_page(191,587,1).
txt_page(503,587,1).
txt_page(793,587,1).
txt_page(481,587,1).
txt_page(335,587,1).
txt_page(919,587,1).
txt_page(852,587,1).
txt_page(654,587,1).
txt_page(665,587,1).
txt_page(133,588,1).
txt_page(634,588,1).
txt_page(844,588,1).
txt_page(249,588,1).
txt_page(385,588,1).
txt_page(403,588,1).
txt_page(956,588,1).
txt_page(372,588,1).
txt_page(224,588,1).
txt_page(284,588,1).
txt_page(733,588,1).
txt_page(12,588,1).
txt_page(762,588,1).
txt_page(738,588,1).
txt_page(815,588,1).
txt_page(452,588,1).
txt_page(849,588,1).
txt_page(732,588,1).
txt_page(155,588,1).
txt_page(533,588,1).
txt_page(79,589,1).
txt_page(594,589,1).
txt_page(276,589,1).
txt_page(204,589,1).
txt_page(75,589,1).
txt_page(553,589,1).
txt_page(250,589,1).
txt_page(219,589,1).
txt_page(957,589,1).
txt_page(869,589,1).
txt_page(922,589,1).
txt_page(46,589,1).
txt_page(965,589,1).
txt_page(876,589,1).
txt_page(352,589,1).
txt_page(449,589,1).
txt_page(247,589,1).
txt_page(255,589,1).
txt_page(71,589,1).
txt_page(341,589,1).
txt_page(757,590,1).
txt_page(14,590,1).
txt_page(906,590,1).
txt_page(742,590,1).
txt_page(531,590,1).
txt_page(916,590,1).
txt_page(653,590,1).
txt_page(680,590,1).
txt_page(780,590,1).
txt_page(101,590,1).
txt_page(907,590,1).
txt_page(563,590,1).
txt_page(197,590,1).
txt_page(904,590,1).
txt_page(47,590,1).
txt_page(879,590,1).
txt_page(851,590,1).
txt_page(730,590,1).
txt_page(214,590,2).
txt_page(80,591,1).
txt_page(701,591,1).
txt_page(865,591,1).
txt_page(169,591,1).
txt_page(478,591,1).
txt_page(156,591,1).
txt_page(921,591,1).
txt_page(697,591,1).
txt_page(194,591,1).
txt_page(533,591,1).
txt_page(546,591,1).
txt_page(291,591,1).
txt_page(663,591,1).
txt_page(502,591,1).
txt_page(379,591,1).
txt_page(251,591,1).
txt_page(545,591,1).
txt_page(830,591,1).
txt_page(558,591,1).
txt_page(963,591,1).
txt_page(411,592,1).
txt_page(844,592,1).
txt_page(598,592,1).
txt_page(486,592,1).
txt_page(849,592,1).
txt_page(943,592,1).
txt_page(796,592,1).
txt_page(959,592,1).
txt_page(784,592,1).
txt_page(791,592,1).
txt_page(919,592,1).
txt_page(29,592,1).
txt_page(402,592,1).
txt_page(380,592,1).
txt_page(941,592,1).
txt_page(827,592,1).
txt_page(414,592,1).
txt_page(36,592,1).
txt_page(422,592,1).
txt_page(819,592,1).
txt_page(302,593,1).
txt_page(756,593,1).
txt_page(168,593,1).
txt_page(64,593,1).
txt_page(435,593,1).
txt_page(779,593,1).
txt_page(669,593,1).
txt_page(767,593,1).
txt_page(25,593,1).
txt_page(878,593,1).
txt_page(476,593,1).
txt_page(226,593,1).
txt_page(925,593,1).
txt_page(386,593,1).
txt_page(579,593,1).
txt_page(578,593,1).
txt_page(504,593,1).
txt_page(871,593,1).
txt_page(456,593,1).
txt_page(20,593,1).
txt_page(908,594,1).
txt_page(680,594,1).
txt_page(95,594,1).
txt_page(175,594,1).
txt_page(919,594,1).
txt_page(875,594,1).
txt_page(818,594,1).
txt_page(708,594,1).
txt_page(218,594,1).
txt_page(806,594,1).
txt_page(543,594,1).
txt_page(793,594,1).
txt_page(276,594,1).
txt_page(12,594,1).
txt_page(670,594,1).
txt_page(285,594,1).
txt_page(303,594,1).
txt_page(59,594,1).
txt_page(714,594,1).
txt_page(684,594,1).
txt_page(377,595,1).
txt_page(495,595,1).
txt_page(252,595,1).
txt_page(742,595,1).
txt_page(677,595,1).
txt_page(401,595,1).
txt_page(891,595,1).
txt_page(478,595,1).
txt_page(454,595,1).
txt_page(862,595,1).
txt_page(278,595,1).
txt_page(374,595,1).
txt_page(843,595,1).
txt_page(290,595,1).
txt_page(718,595,1).
txt_page(394,595,1).
txt_page(702,595,1).
txt_page(772,595,1).
txt_page(855,595,1).
txt_page(597,595,1).
txt_page(507,596,1).
txt_page(190,596,1).
txt_page(952,596,1).
txt_page(613,596,1).
txt_page(274,596,1).
txt_page(59,596,1).
txt_page(589,596,1).
txt_page(922,596,2).
txt_page(961,596,1).
txt_page(857,596,1).
txt_page(717,596,1).
txt_page(159,596,1).
txt_page(208,596,1).
txt_page(652,596,1).
txt_page(940,596,1).
txt_page(763,596,1).
txt_page(765,596,1).
txt_page(662,596,1).
txt_page(63,596,1).
txt_page(136,597,1).
txt_page(461,597,1).
txt_page(772,597,1).
txt_page(775,597,1).
txt_page(437,597,1).
txt_page(89,597,1).
txt_page(546,597,1).
txt_page(123,597,1).
txt_page(773,597,1).
txt_page(936,597,1).
txt_page(370,597,1).
txt_page(428,597,1).
txt_page(692,597,1).
txt_page(107,597,1).
txt_page(718,597,1).
txt_page(365,597,1).
txt_page(51,597,1).
txt_page(86,597,1).
txt_page(95,597,1).
txt_page(545,597,1).
txt_page(757,598,1).
txt_page(565,598,1).
txt_page(162,598,1).
txt_page(726,598,1).
txt_page(604,598,1).
txt_page(323,598,1).
txt_page(442,598,1).
txt_page(708,598,1).
txt_page(141,598,1).
txt_page(842,598,1).
txt_page(188,598,1).
txt_page(780,598,1).
txt_page(707,598,1).
txt_page(482,598,1).
txt_page(109,598,1).
txt_page(810,598,1).
txt_page(276,598,1).
txt_page(283,598,1).
txt_page(603,598,1).
txt_page(495,598,1).
txt_page(228,599,1).
txt_page(905,599,1).
txt_page(372,599,1).
txt_page(911,599,1).
txt_page(660,599,1).
txt_page(691,599,1).
txt_page(32,599,1).
txt_page(790,599,1).
txt_page(68,599,1).
txt_page(625,599,1).
txt_page(856,599,1).
txt_page(322,599,2).
txt_page(772,599,1).
txt_page(138,599,1).
txt_page(251,599,1).
txt_page(874,599,1).
txt_page(564,599,1).
txt_page(655,599,1).
txt_page(319,599,1).
txt_page(143,600,1).
txt_page(830,600,1).
txt_page(335,600,1).
txt_page(179,600,1).
txt_page(896,600,1).
txt_page(195,600,1).
txt_page(324,600,1).
txt_page(966,600,1).
txt_page(574,600,1).
txt_page(682,600,1).
txt_page(413,600,1).
txt_page(641,600,1).
txt_page(317,600,1).
txt_page(173,600,1).
txt_page(484,600,1).
txt_page(626,600,1).
txt_page(54,600,1).
txt_page(248,600,1).
txt_page(539,600,1).
txt_page(488,600,1).
txt_page(413,601,1).
txt_page(610,601,1).
txt_page(827,601,1).
txt_page(478,601,1).
txt_page(208,601,1).
txt_page(698,601,1).
txt_page(821,601,1).
txt_page(366,601,1).
txt_page(953,601,1).
txt_page(218,601,1).
txt_page(72,601,1).
txt_page(235,601,1).
txt_page(485,601,1).
txt_page(752,601,1).
txt_page(728,601,1).
txt_page(453,601,1).
txt_page(123,601,1).
txt_page(925,601,1).
txt_page(492,601,1).
txt_page(792,601,1).
txt_page(878,602,1).
txt_page(269,602,1).
txt_page(380,602,1).
txt_page(364,602,1).
txt_page(479,602,1).
txt_page(687,602,1).
txt_page(607,602,1).
txt_page(386,602,1).
txt_page(34,602,1).
txt_page(337,602,1).
txt_page(619,602,1).
txt_page(346,602,1).
txt_page(287,602,1).
txt_page(749,602,1).
txt_page(867,602,1).
txt_page(791,602,1).
txt_page(155,602,1).
txt_page(635,602,1).
txt_page(356,602,1).
txt_page(534,602,1).
txt_page(84,603,1).
txt_page(206,603,1).
txt_page(796,603,1).
txt_page(340,603,1).
txt_page(791,603,1).
txt_page(484,603,1).
txt_page(496,603,1).
txt_page(683,603,1).
txt_page(99,603,2).
txt_page(612,603,1).
txt_page(488,603,1).
txt_page(87,603,1).
txt_page(824,603,1).
txt_page(150,603,1).
txt_page(13,603,1).
txt_page(756,603,1).
txt_page(884,603,1).
txt_page(359,603,1).
txt_page(367,603,1).
txt_page(337,604,1).
txt_page(858,604,1).
txt_page(446,604,1).
txt_page(774,604,1).
txt_page(578,604,1).
txt_page(949,604,1).
txt_page(307,604,1).
txt_page(851,604,1).
txt_page(628,604,1).
txt_page(621,604,1).
txt_page(502,604,1).
txt_page(643,604,1).
txt_page(916,604,1).
txt_page(812,604,1).
txt_page(208,604,1).
txt_page(58,604,1).
txt_page(390,604,1).
txt_page(322,604,1).
txt_page(261,604,1).
txt_page(32,604,1).
txt_page(747,605,1).
txt_page(421,605,1).
txt_page(204,605,1).
txt_page(595,605,1).
txt_page(182,605,1).
txt_page(567,605,1).
txt_page(237,605,1).
txt_page(933,605,1).
txt_page(27,605,1).
txt_page(613,605,1).
txt_page(362,605,1).
txt_page(637,605,1).
txt_page(759,605,1).
txt_page(266,605,1).
txt_page(57,605,1).
txt_page(348,605,1).
txt_page(213,605,1).
txt_page(184,605,1).
txt_page(309,605,1).
txt_page(665,605,1).
txt_page(593,606,1).
txt_page(895,606,1).
txt_page(135,606,1).
txt_page(960,606,1).
txt_page(474,606,1).
txt_page(376,606,1).
txt_page(745,606,1).
txt_page(799,606,1).
txt_page(726,606,1).
txt_page(653,606,1).
txt_page(622,606,1).
txt_page(820,606,1).
txt_page(929,606,1).
txt_page(783,606,1).
txt_page(837,606,1).
txt_page(252,606,1).
txt_page(597,606,1).
txt_page(579,606,1).
txt_page(607,606,1).
txt_page(198,606,1).
txt_page(277,607,1).
txt_page(304,607,1).
txt_page(608,607,1).
txt_page(55,607,1).
txt_page(231,607,1).
txt_page(745,607,1).
txt_page(879,607,1).
txt_page(784,607,1).
txt_page(19,607,1).
txt_page(153,607,1).
txt_page(309,607,1).
txt_page(377,607,1).
txt_page(707,607,1).
txt_page(843,607,1).
txt_page(490,607,1).
txt_page(481,607,1).
txt_page(390,607,1).
txt_page(285,607,1).
txt_page(265,607,1).
txt_page(578,607,1).
txt_page(234,608,1).
txt_page(236,608,1).
txt_page(205,608,1).
txt_page(567,608,1).
txt_page(716,608,1).
txt_page(358,608,1).
txt_page(473,608,1).
txt_page(348,608,1).
txt_page(176,608,1).
txt_page(263,608,1).
txt_page(583,608,1).
txt_page(450,608,1).
txt_page(261,608,1).
txt_page(581,608,1).
txt_page(401,608,1).
txt_page(708,608,1).
txt_page(924,608,1).
txt_page(655,608,1).
txt_page(775,608,1).
txt_page(779,608,1).
txt_page(788,609,1).
txt_page(674,609,1).
txt_page(317,609,1).
txt_page(708,609,1).
txt_page(769,609,1).
txt_page(792,609,1).
txt_page(340,609,1).
txt_page(948,609,1).
txt_page(95,609,1).
txt_page(294,609,1).
txt_page(230,609,1).
txt_page(644,609,1).
txt_page(535,609,1).
txt_page(353,609,1).
txt_page(849,609,1).
txt_page(367,609,1).
txt_page(723,609,1).
txt_page(898,609,1).
txt_page(446,609,1).
txt_page(397,609,1).
txt_page(412,610,1).
txt_page(748,610,1).
txt_page(560,610,1).
txt_page(463,610,1).
txt_page(434,610,1).
txt_page(177,610,1).
txt_page(878,610,1).
txt_page(403,610,1).
txt_page(696,610,1).
txt_page(832,610,1).
txt_page(216,610,1).
txt_page(709,610,1).
txt_page(641,610,1).
txt_page(278,610,1).
txt_page(595,610,1).
txt_page(281,610,1).
txt_page(647,610,1).
txt_page(355,610,1).
txt_page(967,610,1).
txt_page(549,610,1).
txt_page(548,611,1).
txt_page(132,611,1).
txt_page(278,611,1).
txt_page(412,611,1).
txt_page(262,611,1).
txt_page(802,611,1).
txt_page(494,611,1).
txt_page(15,611,1).
txt_page(605,611,1).
txt_page(914,611,1).
txt_page(389,611,1).
txt_page(302,611,1).
txt_page(788,611,1).
txt_page(242,611,1).
txt_page(190,611,1).
txt_page(816,611,1).
txt_page(289,611,1).
txt_page(396,611,1).
txt_page(942,611,1).
txt_page(796,611,1).
txt_page(505,612,1).
txt_page(69,612,1).
txt_page(690,612,1).
txt_page(297,612,1).
txt_page(881,612,1).
txt_page(564,612,1).
txt_page(184,612,1).
txt_page(958,612,1).
txt_page(308,612,1).
txt_page(595,612,1).
txt_page(111,612,1).
txt_page(53,612,1).
txt_page(616,612,1).
txt_page(152,612,1).
txt_page(946,612,1).
txt_page(522,612,1).
txt_page(208,612,1).
txt_page(786,612,1).
txt_page(233,612,1).
txt_page(71,612,1).
txt_page(868,613,1).
txt_page(638,613,1).
txt_page(955,613,1).
txt_page(690,613,1).
txt_page(235,613,1).
txt_page(220,613,1).
txt_page(311,613,1).
txt_page(940,613,1).
txt_page(180,613,1).
txt_page(131,613,1).
txt_page(607,613,1).
txt_page(682,613,1).
txt_page(161,613,1).
txt_page(572,613,1).
txt_page(688,613,1).
txt_page(354,613,1).
txt_page(487,613,1).
txt_page(268,613,1).
txt_page(551,613,1).
txt_page(400,613,1).
txt_page(419,614,1).
txt_page(810,614,1).
txt_page(548,614,1).
txt_page(582,614,1).
txt_page(115,614,1).
txt_page(531,614,1).
txt_page(514,614,1).
txt_page(24,614,1).
txt_page(932,614,1).
txt_page(358,614,1).
txt_page(474,614,1).
txt_page(510,614,1).
txt_page(338,614,1).
txt_page(811,614,1).
txt_page(293,614,1).
txt_page(382,614,1).
txt_page(379,614,1).
txt_page(664,614,1).
txt_page(391,614,1).
txt_page(618,614,1).
txt_page(331,615,1).
txt_page(654,615,1).
txt_page(79,615,1).
txt_page(328,615,1).
txt_page(670,615,1).
txt_page(386,615,2).
txt_page(391,615,1).
txt_page(67,615,1).
txt_page(834,615,1).
txt_page(343,615,1).
txt_page(379,615,1).
txt_page(119,615,1).
txt_page(93,615,1).
txt_page(871,615,1).
txt_page(248,615,1).
txt_page(336,615,1).
txt_page(469,615,1).
txt_page(2,615,1).
txt_page(717,615,1).
txt_page(88,616,1).
txt_page(960,616,1).
txt_page(233,616,1).
txt_page(309,616,1).
txt_page(426,616,1).
txt_page(638,616,1).
txt_page(259,616,1).
txt_page(189,616,1).
txt_page(385,616,1).
txt_page(332,616,1).
txt_page(671,616,1).
txt_page(614,616,1).
txt_page(800,616,1).
txt_page(967,616,1).
txt_page(29,616,1).
txt_page(114,616,1).
txt_page(619,616,1).
txt_page(400,616,1).
txt_page(23,616,1).
txt_page(307,616,1).
txt_page(266,617,1).
txt_page(731,617,1).
txt_page(100,617,1).
txt_page(366,617,1).
txt_page(198,617,1).
txt_page(299,617,1).
txt_page(897,617,1).
txt_page(823,617,1).
txt_page(598,617,1).
txt_page(788,617,1).
txt_page(358,617,1).
txt_page(760,617,1).
txt_page(593,617,1).
txt_page(697,617,1).
txt_page(42,617,1).
txt_page(401,617,1).
txt_page(217,617,1).
txt_page(212,617,1).
txt_page(571,617,1).
txt_page(936,617,1).
txt_page(355,618,1).
txt_page(599,618,1).
txt_page(799,618,1).
txt_page(7,618,1).
txt_page(751,618,1).
txt_page(663,618,1).
txt_page(194,618,1).
txt_page(163,618,1).
txt_page(56,618,1).
txt_page(150,618,1).
txt_page(204,618,1).
txt_page(857,618,1).
txt_page(916,618,1).
txt_page(309,618,1).
txt_page(11,618,1).
txt_page(107,618,1).
txt_page(540,618,1).
txt_page(236,618,1).
txt_page(900,618,1).
txt_page(706,618,1).
txt_page(17,619,1).
txt_page(722,619,1).
txt_page(51,619,1).
txt_page(305,619,1).
txt_page(129,619,1).
txt_page(333,619,1).
txt_page(256,619,1).
txt_page(792,619,1).
txt_page(286,619,1).
txt_page(163,619,1).
txt_page(754,619,1).
txt_page(62,619,1).
txt_page(209,619,1).
txt_page(767,619,1).
txt_page(259,619,1).
txt_page(395,619,1).
txt_page(429,619,1).
txt_page(447,619,1).
txt_page(449,619,1).
txt_page(469,619,1).
txt_page(408,620,1).
txt_page(569,620,1).
txt_page(747,620,1).
txt_page(442,620,1).
txt_page(449,620,1).
txt_page(237,620,1).
txt_page(856,620,1).
txt_page(56,620,1).
txt_page(76,620,1).
txt_page(928,620,1).
txt_page(942,620,1).
txt_page(798,620,1).
txt_page(30,620,1).
txt_page(343,620,1).
txt_page(247,620,1).
txt_page(193,620,1).
txt_page(145,620,2).
txt_page(61,620,1).
txt_page(70,620,1).
txt_page(249,621,1).
txt_page(954,621,1).
txt_page(693,621,1).
txt_page(841,621,1).
txt_page(336,621,1).
txt_page(826,621,1).
txt_page(121,621,1).
txt_page(27,621,1).
txt_page(504,621,1).
txt_page(487,621,1).
txt_page(194,621,1).
txt_page(68,621,1).
txt_page(282,621,1).
txt_page(364,621,1).
txt_page(471,621,1).
txt_page(363,621,1).
txt_page(892,621,1).
txt_page(798,621,1).
txt_page(888,621,1).
txt_page(147,621,1).
txt_page(677,622,1).
txt_page(183,622,1).
txt_page(805,622,1).
txt_page(573,622,1).
txt_page(250,622,1).
txt_page(159,622,1).
txt_page(318,622,1).
txt_page(425,622,1).
txt_page(358,622,1).
txt_page(723,622,1).
txt_page(605,622,1).
txt_page(700,622,1).
txt_page(195,622,1).
txt_page(357,622,1).
txt_page(257,622,1).
txt_page(831,622,1).
txt_page(633,622,1).
txt_page(526,622,1).
txt_page(790,622,1).
txt_page(622,622,1).
txt_page(437,623,1).
txt_page(942,623,1).
txt_page(892,623,1).
txt_page(206,623,1).
txt_page(788,623,1).
txt_page(215,623,1).
txt_page(378,623,1).
txt_page(599,623,1).
txt_page(541,623,1).
txt_page(579,623,1).
txt_page(746,623,1).
txt_page(292,623,1).
txt_page(8,623,1).
txt_page(166,623,1).
txt_page(36,623,1).
txt_page(267,623,1).
txt_page(227,623,1).
txt_page(68,623,1).
txt_page(555,623,1).
txt_page(623,623,1).
txt_page(470,624,1).
txt_page(438,624,1).
txt_page(109,624,1).
txt_page(254,624,1).
txt_page(693,624,1).
txt_page(255,624,1).
txt_page(428,624,1).
txt_page(671,624,1).
txt_page(349,624,1).
txt_page(87,624,1).
txt_page(104,624,1).
txt_page(235,624,1).
txt_page(803,624,1).
txt_page(644,624,1).
txt_page(442,624,1).
txt_page(38,624,1).
txt_page(704,624,1).
txt_page(36,624,1).
txt_page(155,624,2).
txt_page(933,625,1).
txt_page(855,625,1).
txt_page(438,625,1).
txt_page(194,625,1).
txt_page(902,625,1).
txt_page(394,625,1).
txt_page(510,625,1).
txt_page(39,625,1).
txt_page(32,625,1).
txt_page(786,625,1).
txt_page(440,625,1).
txt_page(43,625,1).
txt_page(621,625,1).
txt_page(723,625,1).
txt_page(841,625,1).
txt_page(34,625,1).
txt_page(875,625,1).
txt_page(12,625,1).
txt_page(609,625,1).
txt_page(831,625,1).
txt_page(185,626,1).
txt_page(121,626,1).
txt_page(11,626,1).
txt_page(749,626,1).
txt_page(429,626,1).
txt_page(861,626,1).
txt_page(800,626,1).
txt_page(455,626,1).
txt_page(177,626,1).
txt_page(545,626,1).
txt_page(247,626,1).
txt_page(540,626,1).
txt_page(526,626,1).
txt_page(654,626,1).
txt_page(483,626,1).
txt_page(875,626,1).
txt_page(13,626,1).
txt_page(845,626,1).
txt_page(403,626,1).
txt_page(420,626,1).
txt_page(943,627,1).
txt_page(764,627,1).
txt_page(239,627,1).
txt_page(555,627,1).
txt_page(368,627,1).
txt_page(287,627,1).
txt_page(343,627,1).
txt_page(612,627,1).
txt_page(902,627,1).
txt_page(139,627,1).
txt_page(331,627,1).
txt_page(313,627,1).
txt_page(358,627,1).
txt_page(299,627,1).
txt_page(50,627,1).
txt_page(445,627,1).
txt_page(450,627,1).
txt_page(115,627,1).
txt_page(282,627,1).
txt_page(495,627,1).
txt_page(345,628,1).
txt_page(449,628,1).
txt_page(862,628,1).
txt_page(256,628,1).
txt_page(736,628,1).
txt_page(428,628,1).
txt_page(919,628,1).
txt_page(691,628,1).
txt_page(745,628,1).
txt_page(159,628,1).
txt_page(367,628,1).
txt_page(689,628,1).
txt_page(423,628,1).
txt_page(918,628,1).
txt_page(688,628,1).
txt_page(671,628,1).
txt_page(559,628,1).
txt_page(34,628,1).
txt_page(312,628,1).
txt_page(235,628,1).
txt_page(570,629,1).
txt_page(270,629,1).
txt_page(52,629,1).
txt_page(723,629,1).
txt_page(931,629,1).
txt_page(126,629,1).
txt_page(374,629,1).
txt_page(235,629,1).
txt_page(273,629,1).
txt_page(853,629,1).
txt_page(67,629,1).
txt_page(608,629,1).
txt_page(240,629,1).
txt_page(519,629,1).
txt_page(41,629,1).
txt_page(150,629,1).
txt_page(626,629,1).
txt_page(268,629,1).
txt_page(614,629,1).
txt_page(797,629,1).
txt_page(604,630,1).
txt_page(90,630,1).
txt_page(136,630,1).
txt_page(480,630,1).
txt_page(445,630,1).
txt_page(477,630,1).
txt_page(563,630,1).
txt_page(76,630,1).
txt_page(526,630,1).
txt_page(800,630,1).
txt_page(708,630,1).
txt_page(468,630,1).
txt_page(355,630,1).
txt_page(657,630,1).
txt_page(539,630,1).
txt_page(691,630,1).
txt_page(30,630,1).
txt_page(394,630,1).
txt_page(288,630,1).
txt_page(773,630,1).
txt_page(461,631,1).
txt_page(401,631,1).
txt_page(226,631,1).
txt_page(449,631,1).
txt_page(258,631,1).
txt_page(369,631,1).
txt_page(269,631,1).
txt_page(760,631,1).
txt_page(79,631,1).
txt_page(918,631,1).
txt_page(243,631,1).
txt_page(260,631,1).
txt_page(903,631,1).
txt_page(205,631,1).
txt_page(569,631,1).
txt_page(284,631,1).
txt_page(698,631,1).
txt_page(583,631,1).
txt_page(486,631,1).
txt_page(683,631,1).
txt_page(802,632,1).
txt_page(170,632,1).
txt_page(868,632,1).
txt_page(203,632,1).
txt_page(549,632,1).
txt_page(876,632,1).
txt_page(375,632,1).
txt_page(492,632,1).
txt_page(142,632,1).
txt_page(946,632,1).
txt_page(527,632,1).
txt_page(650,632,1).
txt_page(428,632,1).
txt_page(554,632,1).
txt_page(280,632,1).
txt_page(917,632,1).
txt_page(612,632,1).
txt_page(518,632,1).
txt_page(335,632,1).
txt_page(839,632,1).
txt_page(330,633,1).
txt_page(310,633,1).
txt_page(730,633,1).
txt_page(632,633,1).
txt_page(564,633,1).
txt_page(766,633,1).
txt_page(960,633,1).
txt_page(343,633,1).
txt_page(411,633,1).
txt_page(626,633,1).
txt_page(836,633,1).
txt_page(830,633,1).
txt_page(232,633,1).
txt_page(504,633,1).
txt_page(373,633,1).
txt_page(759,633,1).
txt_page(887,633,1).
txt_page(123,633,1).
txt_page(773,633,1).
txt_page(892,633,1).
txt_page(685,634,1).
txt_page(75,634,1).
txt_page(74,634,1).
txt_page(957,634,1).
txt_page(155,634,1).
txt_page(465,634,1).
txt_page(892,634,1).
txt_page(422,634,1).
txt_page(907,634,1).
txt_page(706,634,1).
txt_page(844,634,1).
txt_page(453,634,1).
txt_page(398,634,1).
txt_page(632,634,1).
txt_page(516,634,1).
txt_page(152,634,1).
txt_page(653,634,1).
txt_page(710,634,1).
txt_page(811,634,1).
txt_page(371,634,1).
txt_page(557,635,1).
txt_page(402,635,1).
txt_page(81,635,1).
txt_page(517,635,1).
txt_page(535,635,1).
txt_page(564,635,1).
txt_page(134,635,1).
txt_page(315,635,1).
txt_page(203,635,1).
txt_page(526,635,1).
txt_page(749,635,1).
txt_page(574,635,1).
txt_page(590,635,1).
txt_page(888,635,1).
txt_page(595,635,1).
txt_page(275,635,1).
txt_page(581,635,1).
txt_page(435,635,1).
txt_page(404,635,1).
txt_page(502,635,1).
txt_page(630,636,1).
txt_page(446,636,1).
txt_page(526,636,1).
txt_page(420,636,1).
txt_page(566,636,1).
txt_page(269,636,1).
txt_page(871,636,1).
txt_page(330,636,1).
txt_page(769,636,1).
txt_page(898,636,1).
txt_page(754,636,1).
txt_page(572,636,1).
txt_page(143,636,1).
txt_page(812,636,1).
txt_page(782,636,1).
txt_page(749,636,1).
txt_page(435,636,1).
txt_page(277,636,1).
txt_page(332,636,1).
txt_page(421,636,1).
txt_page(457,637,1).
txt_page(428,637,1).
txt_page(589,637,1).
txt_page(378,637,1).
txt_page(713,637,1).
txt_page(913,637,1).
txt_page(574,637,1).
txt_page(886,637,1).
txt_page(274,637,1).
txt_page(623,637,1).
txt_page(263,637,1).
txt_page(954,637,1).
txt_page(91,637,1).
txt_page(372,637,1).
txt_page(272,637,1).
txt_page(518,637,1).
txt_page(785,637,1).
txt_page(159,637,1).
txt_page(359,637,1).
txt_page(217,637,1).
txt_page(344,638,1).
txt_page(529,638,1).
txt_page(184,638,1).
txt_page(582,638,1).
txt_page(634,638,1).
txt_page(702,638,1).
txt_page(516,638,1).
txt_page(378,638,1).
txt_page(136,638,1).
txt_page(722,638,1).
txt_page(515,638,1).
txt_page(858,638,1).
txt_page(129,638,1).
txt_page(372,638,1).
txt_page(511,638,1).
txt_page(312,638,1).
txt_page(762,638,1).
txt_page(285,638,1).
txt_page(931,638,1).
txt_page(561,638,1).
txt_page(460,639,1).
txt_page(3,639,1).
txt_page(444,639,1).
txt_page(434,639,1).
txt_page(347,639,1).
txt_page(15,639,1).
txt_page(875,639,1).
txt_page(806,639,1).
txt_page(315,639,1).
txt_page(823,639,1).
txt_page(251,639,1).
txt_page(153,639,1).
txt_page(513,639,1).
txt_page(759,639,1).
txt_page(825,639,1).
txt_page(681,639,1).
txt_page(586,639,1).
txt_page(959,639,1).
txt_page(156,639,1).
txt_page(589,639,1).
txt_page(213,640,1).
txt_page(791,640,1).
txt_page(420,640,1).
txt_page(768,640,1).
txt_page(709,640,1).
txt_page(18,640,1).
txt_page(834,640,1).
txt_page(467,640,1).
txt_page(905,640,1).
txt_page(916,640,1).
txt_page(348,640,1).
txt_page(410,640,1).
txt_page(523,640,1).
txt_page(882,640,1).
txt_page(858,640,1).
txt_page(124,640,1).
txt_page(672,640,1).
txt_page(250,640,1).
txt_page(756,640,1).
txt_page(745,640,1).
txt_page(44,641,1).
txt_page(252,641,1).
txt_page(53,641,1).
txt_page(941,641,1).
txt_page(256,641,1).
txt_page(521,641,1).
txt_page(223,641,1).
txt_page(919,641,1).
txt_page(347,641,1).
txt_page(874,641,1).
txt_page(945,641,1).
txt_page(118,641,1).
txt_page(194,641,1).
txt_page(691,641,1).
txt_page(49,641,1).
txt_page(723,641,1).
txt_page(611,641,1).
txt_page(25,641,1).
txt_page(260,641,1).
txt_page(525,641,1).
txt_page(794,642,1).
txt_page(52,642,1).
txt_page(138,642,1).
txt_page(249,642,1).
txt_page(891,642,1).
txt_page(128,642,1).
txt_page(510,642,1).
txt_page(289,642,1).
txt_page(816,642,1).
txt_page(930,642,1).
txt_page(466,642,1).
txt_page(231,642,2).
txt_page(872,642,1).
txt_page(113,642,1).
txt_page(890,642,1).
txt_page(842,642,1).
txt_page(515,642,1).
txt_page(853,642,1).
txt_page(580,642,1).
txt_page(847,643,1).
txt_page(794,643,1).
txt_page(530,643,1).
txt_page(54,643,1).
txt_page(269,643,1).
txt_page(190,643,1).
txt_page(461,643,1).
txt_page(417,643,1).
txt_page(250,643,1).
txt_page(665,643,1).
txt_page(499,643,1).
txt_page(251,643,1).
txt_page(357,643,1).
txt_page(946,643,1).
txt_page(756,643,1).
txt_page(348,643,1).
txt_page(268,643,1).
txt_page(472,643,1).
txt_page(693,643,1).
txt_page(867,643,1).
txt_page(814,644,1).
txt_page(934,644,1).
txt_page(301,644,1).
txt_page(809,644,1).
txt_page(926,644,1).
txt_page(905,644,1).
txt_page(154,644,1).
txt_page(869,644,1).
txt_page(679,644,1).
txt_page(360,644,1).
txt_page(448,644,1).
txt_page(406,644,1).
txt_page(830,644,1).
txt_page(910,644,1).
txt_page(57,644,1).
txt_page(326,644,1).
txt_page(762,644,1).
txt_page(790,644,1).
txt_page(927,644,1).
txt_page(196,644,1).
txt_page(350,645,1).
txt_page(569,645,1).
txt_page(290,645,1).
txt_page(521,645,1).
txt_page(919,645,1).
txt_page(486,645,1).
txt_page(113,645,1).
txt_page(373,645,1).
txt_page(254,645,1).
txt_page(768,645,1).
txt_page(131,645,1).
txt_page(752,645,1).
txt_page(742,645,1).
txt_page(288,645,1).
txt_page(116,645,1).
txt_page(668,645,1).
txt_page(712,645,1).
txt_page(615,645,1).
txt_page(41,645,1).
txt_page(441,645,1).
txt_page(398,646,1).
txt_page(57,646,1).
txt_page(839,646,1).
txt_page(312,646,1).
txt_page(766,646,1).
txt_page(534,646,1).
txt_page(547,646,1).
txt_page(949,646,1).
txt_page(851,646,1).
txt_page(719,646,1).
txt_page(671,646,1).
txt_page(150,646,1).
txt_page(763,646,1).
txt_page(920,646,1).
txt_page(605,646,1).
txt_page(498,646,1).
txt_page(349,646,1).
txt_page(733,646,1).
txt_page(953,646,1).
txt_page(841,646,1).
txt_page(436,647,1).
txt_page(534,647,1).
txt_page(938,647,1).
txt_page(149,647,1).
txt_page(924,647,1).
txt_page(713,647,1).
txt_page(687,647,1).
txt_page(840,647,1).
txt_page(78,647,1).
txt_page(691,647,1).
txt_page(349,647,1).
txt_page(87,647,1).
txt_page(286,647,1).
txt_page(414,647,1).
txt_page(926,647,1).
txt_page(830,647,1).
txt_page(361,647,1).
txt_page(277,647,1).
txt_page(649,647,1).
txt_page(756,647,1).
txt_page(860,648,1).
txt_page(621,648,1).
txt_page(834,648,1).
txt_page(135,648,1).
txt_page(272,648,1).
txt_page(82,648,1).
txt_page(227,648,1).
txt_page(178,648,1).
txt_page(143,648,1).
txt_page(303,648,1).
txt_page(677,648,1).
txt_page(800,648,1).
txt_page(866,648,1).
txt_page(867,648,1).
txt_page(706,648,1).
txt_page(651,648,1).
txt_page(158,648,1).
txt_page(512,648,1).
txt_page(506,648,1).
txt_page(358,648,1).
txt_page(194,649,1).
txt_page(33,649,1).
txt_page(324,649,1).
txt_page(315,649,1).
txt_page(205,649,1).
txt_page(36,649,1).
txt_page(407,649,1).
txt_page(45,649,1).
txt_page(817,649,1).
txt_page(852,649,1).
txt_page(134,649,1).
txt_page(887,649,1).
txt_page(892,649,1).
txt_page(120,649,1).
txt_page(387,649,1).
txt_page(266,649,1).
txt_page(767,649,1).
txt_page(470,649,1).
txt_page(568,649,1).
txt_page(55,649,1).
txt_page(732,650,1).
txt_page(946,650,1).
txt_page(875,650,1).
txt_page(920,650,1).
txt_page(822,650,1).
txt_page(276,650,1).
txt_page(263,650,1).
txt_page(642,650,1).
txt_page(188,650,1).
txt_page(569,650,1).
txt_page(198,650,1).
txt_page(284,650,1).
txt_page(509,650,1).
txt_page(194,650,1).
txt_page(486,650,1).
txt_page(74,650,1).
txt_page(567,650,1).
txt_page(967,650,1).
txt_page(234,650,1).
txt_page(70,650,1).
txt_page(683,651,1).
txt_page(904,651,1).
txt_page(890,651,1).
txt_page(930,651,1).
txt_page(144,651,1).
txt_page(259,651,1).
txt_page(12,651,1).
txt_page(406,651,1).
txt_page(800,651,1).
txt_page(39,651,1).
txt_page(856,651,1).
txt_page(344,651,1).
txt_page(11,651,1).
txt_page(762,651,1).
txt_page(268,651,1).
txt_page(881,651,1).
txt_page(112,651,1).
txt_page(443,651,1).
txt_page(457,651,1).
txt_page(45,651,1).
txt_page(465,652,1).
txt_page(706,652,1).
txt_page(163,652,1).
txt_page(953,652,1).
txt_page(477,652,1).
txt_page(512,652,1).
txt_page(689,652,1).
txt_page(771,652,1).
txt_page(963,652,1).
txt_page(237,652,1).
txt_page(269,652,1).
txt_page(296,652,1).
txt_page(327,652,1).
txt_page(176,652,1).
txt_page(111,652,1).
txt_page(788,652,1).
txt_page(288,652,1).
txt_page(675,652,1).
txt_page(893,652,1).
txt_page(115,652,1).
txt_page(277,653,1).
txt_page(852,653,1).
txt_page(633,653,1).
txt_page(234,653,1).
txt_page(421,653,1).
txt_page(656,653,1).
txt_page(593,653,1).
txt_page(238,653,1).
txt_page(536,653,1).
txt_page(135,653,1).
txt_page(83,653,1).
txt_page(382,653,1).
txt_page(117,653,1).
txt_page(706,653,1).
txt_page(149,653,1).
txt_page(883,653,1).
txt_page(89,653,1).
txt_page(861,653,1).
txt_page(885,653,2).
txt_page(815,654,1).
txt_page(717,654,1).
txt_page(439,654,1).
txt_page(188,654,1).
txt_page(10,654,1).
txt_page(734,654,1).
txt_page(344,654,1).
txt_page(25,654,1).
txt_page(832,654,1).
txt_page(713,654,1).
txt_page(831,654,1).
txt_page(509,654,1).
txt_page(343,654,2).
txt_page(367,654,1).
txt_page(566,654,1).
txt_page(865,654,1).
txt_page(351,654,1).
txt_page(733,654,1).
txt_page(345,654,1).
txt_page(213,655,1).
txt_page(967,655,1).
txt_page(275,655,1).
txt_page(646,655,1).
txt_page(847,655,1).
txt_page(670,655,1).
txt_page(262,655,1).
txt_page(731,655,1).
txt_page(671,655,1).
txt_page(908,655,1).
txt_page(249,655,1).
txt_page(52,655,1).
txt_page(580,655,1).
txt_page(133,655,1).
txt_page(69,655,1).
txt_page(466,655,1).
txt_page(764,655,1).
txt_page(945,655,1).
txt_page(111,655,1).
txt_page(47,655,1).
txt_page(484,656,1).
txt_page(118,656,1).
txt_page(755,656,1).
txt_page(771,656,1).
txt_page(857,656,1).
txt_page(610,656,1).
txt_page(661,656,1).
txt_page(730,656,1).
txt_page(71,656,1).
txt_page(856,656,1).
txt_page(879,656,1).
txt_page(92,656,1).
txt_page(639,656,1).
txt_page(354,656,1).
txt_page(350,656,1).
txt_page(87,656,1).
txt_page(600,656,1).
txt_page(229,656,1).
txt_page(30,656,1).
txt_page(221,656,1).
txt_page(618,657,1).
txt_page(211,657,1).
txt_page(949,657,1).
txt_page(520,657,1).
txt_page(139,657,1).
txt_page(354,657,1).
txt_page(705,657,1).
txt_page(279,657,1).
txt_page(805,657,1).
txt_page(99,657,1).
txt_page(428,657,1).
txt_page(858,657,1).
txt_page(176,657,1).
txt_page(761,657,1).
txt_page(818,657,1).
txt_page(439,657,1).
txt_page(795,657,1).
txt_page(414,657,1).
txt_page(538,657,1).
txt_page(935,657,1).
txt_page(286,658,1).
txt_page(362,658,1).
txt_page(757,658,1).
txt_page(434,658,1).
txt_page(816,658,1).
txt_page(188,658,1).
txt_page(121,658,1).
txt_page(682,658,1).
txt_page(730,658,1).
txt_page(1,658,1).
txt_page(845,658,1).
txt_page(110,658,1).
txt_page(97,658,1).
txt_page(624,658,1).
txt_page(590,658,1).
txt_page(656,658,1).
txt_page(652,658,1).
txt_page(649,658,1).
txt_page(341,658,1).
txt_page(519,658,1).
txt_page(567,659,1).
txt_page(607,659,1).
txt_page(643,659,1).
txt_page(468,659,1).
txt_page(200,659,1).
txt_page(105,659,1).
txt_page(215,659,1).
txt_page(119,659,1).
txt_page(367,659,1).
txt_page(139,659,1).
txt_page(620,659,2).
txt_page(151,659,1).
txt_page(474,659,1).
txt_page(137,659,1).
txt_page(802,659,1).
txt_page(231,659,1).
txt_page(762,659,1).
txt_page(438,659,1).
txt_page(74,659,1).
txt_page(68,660,1).
txt_page(930,660,1).
txt_page(646,660,1).
txt_page(27,660,1).
txt_page(691,660,1).
txt_page(844,660,1).
txt_page(923,660,1).
txt_page(340,660,1).
txt_page(680,660,1).
txt_page(570,660,1).
txt_page(207,660,1).
txt_page(362,660,1).
txt_page(595,660,1).
txt_page(275,660,1).
txt_page(578,660,1).
txt_page(615,660,1).
txt_page(794,660,1).
txt_page(452,660,1).
txt_page(822,660,2).
txt_page(700,661,1).
txt_page(505,661,1).
txt_page(561,661,1).
txt_page(306,661,1).
txt_page(487,661,1).
txt_page(55,661,1).
txt_page(180,661,1).
txt_page(437,661,1).
txt_page(527,661,1).
txt_page(747,661,1).
txt_page(516,661,1).
txt_page(640,661,1).
txt_page(554,661,1).
txt_page(65,661,1).
txt_page(693,661,1).
txt_page(392,661,1).
txt_page(463,661,1).
txt_page(27,661,1).
txt_page(391,661,1).
txt_page(348,661,1).
txt_page(616,662,1).
txt_page(357,662,1).
txt_page(834,662,1).
txt_page(856,662,1).
txt_page(354,662,1).
txt_page(352,662,1).
txt_page(851,662,1).
txt_page(738,662,1).
txt_page(10,662,1).
txt_page(327,662,1).
txt_page(364,662,1).
txt_page(1,662,2).
txt_page(401,662,1).
txt_page(359,662,1).
txt_page(570,662,1).
txt_page(234,662,1).
txt_page(378,662,1).
txt_page(295,662,1).
txt_page(308,662,1).
txt_page(533,663,1).
txt_page(798,663,1).
txt_page(269,663,1).
txt_page(323,663,1).
txt_page(738,663,1).
txt_page(675,663,1).
txt_page(717,663,1).
txt_page(200,663,1).
txt_page(410,663,1).
txt_page(332,663,1).
txt_page(26,663,1).
txt_page(73,663,1).
txt_page(424,663,1).
txt_page(315,663,1).
txt_page(138,663,1).
txt_page(708,663,1).
txt_page(36,663,1).
txt_page(325,663,1).
txt_page(577,663,1).
txt_page(946,663,1).
txt_page(799,664,1).
txt_page(140,664,1).
txt_page(554,664,1).
txt_page(392,664,1).
txt_page(264,664,1).
txt_page(372,664,1).
txt_page(805,664,1).
txt_page(563,664,1).
txt_page(460,664,1).
txt_page(787,664,1).
txt_page(677,664,1).
txt_page(283,664,1).
txt_page(635,664,1).
txt_page(641,664,1).
txt_page(398,664,1).
txt_page(551,664,1).
txt_page(845,664,1).
txt_page(409,664,1).
txt_page(340,664,1).
txt_page(820,664,1).
txt_page(536,665,1).
txt_page(139,665,1).
txt_page(936,665,1).
txt_page(845,665,1).
txt_page(838,665,1).
txt_page(628,665,1).
txt_page(44,665,1).
txt_page(203,665,1).
txt_page(743,665,1).
txt_page(362,665,1).
txt_page(824,665,1).
txt_page(808,665,1).
txt_page(753,665,1).
txt_page(194,665,1).
txt_page(115,665,1).
txt_page(796,665,1).
txt_page(779,665,1).
txt_page(395,665,1).
txt_page(70,665,1).
txt_page(450,665,1).
txt_page(459,666,1).
txt_page(839,666,1).
txt_page(348,666,1).
txt_page(750,666,1).
txt_page(504,666,1).
txt_page(745,666,1).
txt_page(709,666,1).
txt_page(775,666,1).
txt_page(549,666,1).
txt_page(578,666,1).
txt_page(498,666,1).
txt_page(824,666,1).
txt_page(535,666,1).
txt_page(710,666,1).
txt_page(625,666,1).
txt_page(815,666,1).
txt_page(777,666,1).
txt_page(132,666,1).
txt_page(602,666,1).
txt_page(842,666,1).
txt_page(181,667,1).
txt_page(32,667,1).
txt_page(901,667,1).
txt_page(474,667,1).
txt_page(411,667,1).
txt_page(424,667,1).
txt_page(882,667,1).
txt_page(486,667,1).
txt_page(815,667,1).
txt_page(110,667,1).
txt_page(34,667,1).
txt_page(477,667,1).
txt_page(949,667,1).
txt_page(689,667,1).
txt_page(883,667,1).
txt_page(462,667,1).
txt_page(358,667,1).
txt_page(387,667,1).
txt_page(58,667,1).
txt_page(407,667,1).
txt_page(566,668,1).
txt_page(249,668,1).
txt_page(705,668,1).
txt_page(739,668,1).
txt_page(632,668,1).
txt_page(250,668,1).
txt_page(283,668,1).
txt_page(154,668,1).
txt_page(377,668,1).
txt_page(807,668,1).
txt_page(761,668,1).
txt_page(626,668,1).
txt_page(366,668,1).
txt_page(275,668,1).
txt_page(579,668,1).
txt_page(479,668,1).
txt_page(631,668,1).
txt_page(486,668,2).
txt_page(463,668,1).
txt_page(155,669,1).
txt_page(400,669,1).
txt_page(79,669,1).
txt_page(88,669,1).
txt_page(332,669,1).
txt_page(212,669,1).
txt_page(798,669,1).
txt_page(768,669,1).
txt_page(428,669,1).
txt_page(300,669,1).
txt_page(335,669,1).
txt_page(917,669,1).
txt_page(828,669,1).
txt_page(810,669,1).
txt_page(802,669,1).
txt_page(551,669,1).
txt_page(906,669,1).
txt_page(787,669,1).
txt_page(78,669,1).
txt_page(520,669,1).
txt_page(833,670,1).
txt_page(895,670,1).
txt_page(371,670,1).
txt_page(512,670,1).
txt_page(185,670,1).
txt_page(768,670,1).
txt_page(144,670,1).
txt_page(356,670,1).
txt_page(395,670,1).
txt_page(468,670,1).
txt_page(907,670,1).
txt_page(92,670,1).
txt_page(618,670,1).
txt_page(265,670,1).
txt_page(614,670,1).
txt_page(457,670,1).
txt_page(952,670,1).
txt_page(247,670,1).
txt_page(463,670,1).
txt_page(259,670,1).
txt_page(75,671,1).
txt_page(744,671,1).
txt_page(453,671,1).
txt_page(832,671,1).
txt_page(545,671,1).
txt_page(202,671,1).
txt_page(659,671,1).
txt_page(839,671,1).
txt_page(240,671,1).
txt_page(8,671,1).
txt_page(61,671,1).
txt_page(449,671,1).
txt_page(411,671,1).
txt_page(730,671,1).
txt_page(631,671,1).
txt_page(362,671,1).
txt_page(884,671,1).
txt_page(286,671,1).
txt_page(602,671,1).
txt_page(396,671,1).
txt_page(256,672,1).
txt_page(290,672,1).
txt_page(348,672,1).
txt_page(606,672,1).
txt_page(335,672,1).
txt_page(100,672,1).
txt_page(351,672,1).
txt_page(274,672,1).
txt_page(677,672,1).
txt_page(633,672,1).
txt_page(584,672,1).
txt_page(636,672,1).
txt_page(923,672,1).
txt_page(292,672,1).
txt_page(191,672,1).
txt_page(560,672,1).
txt_page(646,672,1).
txt_page(511,672,1).
txt_page(323,672,1).
txt_page(60,672,1).
txt_page(83,673,1).
txt_page(310,673,1).
txt_page(234,673,1).
txt_page(383,673,1).
txt_page(587,673,1).
txt_page(854,673,1).
txt_page(245,673,1).
txt_page(336,673,1).
txt_page(672,673,1).
txt_page(825,673,1).
txt_page(142,673,1).
txt_page(955,673,1).
txt_page(511,673,1).
txt_page(139,673,1).
txt_page(835,673,2).
txt_page(629,673,1).
txt_page(263,673,1).
txt_page(413,673,1).
txt_page(769,673,1).
txt_page(279,674,1).
txt_page(373,674,1).
txt_page(556,674,1).
txt_page(97,674,1).
txt_page(933,674,1).
txt_page(682,674,1).
txt_page(817,674,1).
txt_page(118,674,1).
txt_page(374,674,1).
txt_page(392,674,1).
txt_page(51,674,1).
txt_page(300,674,1).
txt_page(102,674,1).
txt_page(850,674,1).
txt_page(419,674,1).
txt_page(702,674,1).
txt_page(618,674,1).
txt_page(254,674,1).
txt_page(676,674,1).
txt_page(610,674,1).
txt_page(819,675,1).
txt_page(199,675,1).
txt_page(908,675,1).
txt_page(181,675,1).
txt_page(897,675,1).
txt_page(234,675,1).
txt_page(66,675,1).
txt_page(178,675,1).
txt_page(822,675,1).
txt_page(258,675,1).
txt_page(257,675,1).
txt_page(120,675,1).
txt_page(17,675,1).
txt_page(399,675,1).
txt_page(70,675,1).
txt_page(848,675,1).
txt_page(238,675,1).
txt_page(773,675,1).
txt_page(71,675,1).
txt_page(878,675,1).
txt_page(125,676,1).
txt_page(858,676,1).
txt_page(955,676,1).
txt_page(933,676,1).
txt_page(55,676,1).
txt_page(507,676,1).
txt_page(722,676,1).
txt_page(821,676,1).
txt_page(527,676,1).
txt_page(662,676,1).
txt_page(659,676,1).
txt_page(780,676,1).
txt_page(141,676,1).
txt_page(667,676,1).
txt_page(798,676,1).
txt_page(479,676,1).
txt_page(51,676,1).
txt_page(571,676,1).
txt_page(927,676,1).
txt_page(572,676,1).
txt_page(792,677,1).
txt_page(935,677,1).
txt_page(427,677,1).
txt_page(959,677,1).
txt_page(405,677,1).
txt_page(716,677,1).
txt_page(589,677,1).
txt_page(254,677,1).
txt_page(819,677,1).
txt_page(431,677,1).
txt_page(393,677,1).
txt_page(216,677,1).
txt_page(384,677,1).
txt_page(730,677,1).
txt_page(532,677,1).
txt_page(283,677,1).
txt_page(852,677,1).
txt_page(805,677,1).
txt_page(368,677,1).
txt_page(844,677,1).
txt_page(40,678,1).
txt_page(105,678,1).
txt_page(447,678,1).
txt_page(440,678,1).
txt_page(736,678,1).
txt_page(549,678,1).
txt_page(18,678,1).
txt_page(270,678,1).
txt_page(167,678,1).
txt_page(89,678,1).
txt_page(417,678,1).
txt_page(213,678,1).
txt_page(172,678,1).
txt_page(534,678,1).
txt_page(30,678,1).
txt_page(490,678,1).
txt_page(880,678,1).
txt_page(151,678,1).
txt_page(398,678,1).
txt_page(962,678,1).
txt_page(517,679,1).
txt_page(309,679,1).
txt_page(855,679,1).
txt_page(832,679,1).
txt_page(120,679,1).
txt_page(625,679,1).
txt_page(761,679,1).
txt_page(882,679,1).
txt_page(554,679,1).
txt_page(24,679,1).
txt_page(229,679,1).
txt_page(595,679,1).
txt_page(499,679,1).
txt_page(908,679,1).
txt_page(857,679,1).
txt_page(575,679,1).
txt_page(557,679,1).
txt_page(346,679,1).
txt_page(305,679,1).
txt_page(665,679,1).
txt_page(719,680,1).
txt_page(754,680,1).
txt_page(101,680,1).
txt_page(796,680,1).
txt_page(524,680,1).
txt_page(68,680,1).
txt_page(489,680,1).
txt_page(735,680,1).
txt_page(158,680,1).
txt_page(172,680,1).
txt_page(788,680,1).
txt_page(255,680,1).
txt_page(896,680,1).
txt_page(403,680,1).
txt_page(381,680,1).
txt_page(966,680,1).
txt_page(466,680,2).
txt_page(226,680,1).
txt_page(425,680,1).
txt_page(309,681,1).
txt_page(257,681,1).
txt_page(793,681,1).
txt_page(371,681,1).
txt_page(571,681,1).
txt_page(712,681,1).
txt_page(476,681,1).
txt_page(460,681,1).
txt_page(398,681,1).
txt_page(655,681,1).
txt_page(204,681,1).
txt_page(136,681,1).
txt_page(563,681,1).
txt_page(828,681,1).
txt_page(698,681,1).
txt_page(893,681,1).
txt_page(443,681,2).
txt_page(529,681,1).
txt_page(877,681,1).
txt_page(852,682,1).
txt_page(584,682,1).
txt_page(181,682,1).
txt_page(610,682,1).
txt_page(746,682,1).
txt_page(367,682,1).
txt_page(303,682,1).
txt_page(623,682,1).
txt_page(205,682,1).
txt_page(698,682,1).
txt_page(738,682,1).
txt_page(386,682,1).
txt_page(789,682,1).
txt_page(917,682,1).
txt_page(47,682,1).
txt_page(742,682,1).
txt_page(851,682,2).
txt_page(245,682,1).
txt_page(208,682,1).
txt_page(611,683,1).
txt_page(72,683,1).
txt_page(441,683,1).
txt_page(126,683,1).
txt_page(83,683,1).
txt_page(175,683,1).
txt_page(575,683,1).
txt_page(717,683,1).
txt_page(55,683,1).
txt_page(602,683,1).
txt_page(433,683,1).
txt_page(587,683,1).
txt_page(916,683,1).
txt_page(268,683,1).
txt_page(735,683,1).
txt_page(795,683,1).
txt_page(419,683,1).
txt_page(453,683,2).
txt_page(747,683,1).
txt_page(190,684,1).
txt_page(892,684,1).
txt_page(885,684,1).
txt_page(211,684,1).
txt_page(348,684,1).
txt_page(571,684,1).
txt_page(735,684,1).
txt_page(860,684,1).
txt_page(635,684,1).
txt_page(637,684,1).
txt_page(407,684,1).
txt_page(761,684,1).
txt_page(631,684,1).
txt_page(369,684,1).
txt_page(505,684,1).
txt_page(172,684,1).
txt_page(275,684,1).
txt_page(401,684,1).
txt_page(757,684,1).
txt_page(587,684,1).
txt_page(56,685,1).
txt_page(560,685,1).
txt_page(339,685,1).
txt_page(407,685,1).
txt_page(340,685,1).
txt_page(397,685,1).
txt_page(179,685,1).
txt_page(755,685,1).
txt_page(589,685,1).
txt_page(253,685,1).
txt_page(720,685,1).
txt_page(544,685,1).
txt_page(769,685,1).
txt_page(123,685,1).
txt_page(211,685,1).
txt_page(730,685,1).
txt_page(929,685,1).
txt_page(561,685,1).
txt_page(506,685,1).
txt_page(709,685,1).
txt_page(274,686,1).
txt_page(305,686,1).
txt_page(949,686,1).
txt_page(823,686,1).
txt_page(819,686,1).
txt_page(56,686,1).
txt_page(493,686,1).
txt_page(524,686,1).
txt_page(749,686,1).
txt_page(733,686,1).
txt_page(178,686,1).
txt_page(308,686,1).
txt_page(942,686,1).
txt_page(624,686,1).
txt_page(74,686,1).
txt_page(523,686,1).
txt_page(482,686,1).
txt_page(424,686,1).
txt_page(730,686,1).
txt_page(457,686,1).
txt_page(755,687,1).
txt_page(484,687,1).
txt_page(18,687,1).
txt_page(329,687,1).
txt_page(384,687,1).
txt_page(542,687,1).
txt_page(7,687,1).
txt_page(623,687,1).
txt_page(370,687,1).
txt_page(634,687,1).
txt_page(722,687,1).
txt_page(855,687,1).
txt_page(947,687,1).
txt_page(510,687,1).
txt_page(223,687,1).
txt_page(905,687,1).
txt_page(322,687,1).
txt_page(41,687,1).
txt_page(936,687,1).
txt_page(72,687,1).
txt_page(456,688,1).
txt_page(354,688,1).
txt_page(348,688,1).
txt_page(965,688,1).
txt_page(497,688,1).
txt_page(93,688,1).
txt_page(856,688,1).
txt_page(45,688,1).
txt_page(886,688,1).
txt_page(295,688,1).
txt_page(479,688,1).
txt_page(352,688,1).
txt_page(604,688,1).
txt_page(36,688,1).
txt_page(398,688,1).
txt_page(70,688,1).
txt_page(522,688,1).
txt_page(137,688,1).
txt_page(206,688,1).
txt_page(165,688,1).
txt_page(210,689,1).
txt_page(244,689,1).
txt_page(82,689,1).
txt_page(313,689,1).
txt_page(345,689,1).
txt_page(798,689,1).
txt_page(936,689,1).
txt_page(452,689,1).
txt_page(167,689,1).
txt_page(594,689,1).
txt_page(139,689,1).
txt_page(396,689,1).
txt_page(425,689,1).
txt_page(181,689,1).
txt_page(383,689,1).
txt_page(681,689,1).
txt_page(829,689,1).
txt_page(458,689,1).
txt_page(693,689,1).
txt_page(909,689,1).
txt_page(688,690,1).
txt_page(88,690,1).
txt_page(21,690,1).
txt_page(355,690,1).
txt_page(865,690,1).
txt_page(58,690,1).
txt_page(175,690,1).
txt_page(108,690,1).
txt_page(432,690,1).
txt_page(309,690,1).
txt_page(634,690,1).
txt_page(229,690,1).
txt_page(686,690,1).
txt_page(418,690,1).
txt_page(538,690,1).
txt_page(684,690,1).
txt_page(834,690,1).
txt_page(443,690,1).
txt_page(755,690,1).
txt_page(145,690,1).
txt_page(439,691,1).
txt_page(721,691,1).
txt_page(163,691,1).
txt_page(691,691,1).
txt_page(332,691,1).
txt_page(617,691,1).
txt_page(359,691,1).
txt_page(707,691,1).
txt_page(940,691,1).
txt_page(507,691,1).
txt_page(1,691,1).
txt_page(210,691,1).
txt_page(826,691,1).
txt_page(458,691,1).
txt_page(814,691,1).
txt_page(20,691,1).
txt_page(751,691,1).
txt_page(823,691,1).
txt_page(786,691,1).
txt_page(103,691,1).
txt_page(43,692,1).
txt_page(696,692,1).
txt_page(301,692,1).
txt_page(563,692,1).
txt_page(760,692,1).
txt_page(798,692,1).
txt_page(531,692,1).
txt_page(623,692,1).
txt_page(401,692,1).
txt_page(194,692,1).
txt_page(938,692,1).
txt_page(176,692,1).
txt_page(738,692,1).
txt_page(456,692,1).
txt_page(628,692,1).
txt_page(849,692,1).
txt_page(613,692,1).
txt_page(334,692,1).
txt_page(66,692,1).
txt_page(837,692,1).
txt_page(933,693,1).
txt_page(260,693,1).
txt_page(488,693,1).
txt_page(464,693,1).
txt_page(765,693,1).
txt_page(182,693,1).
txt_page(13,693,1).
txt_page(795,693,1).
txt_page(490,693,1).
txt_page(816,693,1).
txt_page(732,693,1).
txt_page(556,693,1).
txt_page(895,693,1).
txt_page(100,693,1).
txt_page(84,693,1).
txt_page(211,693,1).
txt_page(333,693,1).
txt_page(916,693,1).
txt_page(132,693,1).
txt_page(731,693,1).
txt_page(606,694,1).
txt_page(578,694,1).
txt_page(551,694,1).
txt_page(807,694,1).
txt_page(552,694,1).
txt_page(733,694,1).
txt_page(198,694,1).
txt_page(575,694,1).
txt_page(889,694,1).
txt_page(445,694,1).
txt_page(399,694,1).
txt_page(123,694,1).
txt_page(907,694,1).
txt_page(677,694,1).
txt_page(699,694,1).
txt_page(559,694,1).
txt_page(865,694,1).
txt_page(183,694,1).
txt_page(162,694,1).
txt_page(155,694,1).
txt_page(121,695,1).
txt_page(417,695,1).
txt_page(608,695,1).
txt_page(736,695,1).
txt_page(152,695,1).
txt_page(544,695,1).
txt_page(576,695,1).
txt_page(301,695,1).
txt_page(388,695,1).
txt_page(185,695,1).
txt_page(30,695,1).
txt_page(94,695,1).
txt_page(146,695,1).
txt_page(956,695,1).
txt_page(838,695,1).
txt_page(959,695,1).
txt_page(359,695,2).
txt_page(402,695,1).
txt_page(835,695,1).
txt_page(57,696,1).
txt_page(28,696,1).
txt_page(155,696,1).
txt_page(465,696,1).
txt_page(845,696,1).
txt_page(586,696,1).
txt_page(329,696,1).
txt_page(761,696,1).
txt_page(233,696,1).
txt_page(957,696,1).
txt_page(378,696,1).
txt_page(339,696,1).
txt_page(858,696,1).
txt_page(25,696,1).
txt_page(502,696,1).
txt_page(360,696,1).
txt_page(670,696,1).
txt_page(563,696,1).
txt_page(227,696,1).
txt_page(945,696,1).
txt_page(945,697,1).
txt_page(441,697,1).
txt_page(773,697,1).
txt_page(924,697,1).
txt_page(298,697,1).
txt_page(919,697,1).
txt_page(737,697,1).
txt_page(286,697,1).
txt_page(434,697,1).
txt_page(244,697,1).
txt_page(684,697,1).
txt_page(384,697,1).
txt_page(440,697,1).
txt_page(912,697,1).
txt_page(307,697,1).
txt_page(9,697,1).
txt_page(39,697,1).
txt_page(26,697,1).
txt_page(226,697,1).
txt_page(23,697,1).
txt_page(220,698,1).
txt_page(678,698,1).
txt_page(454,698,1).
txt_page(249,698,1).
txt_page(774,698,1).
txt_page(848,698,1).
txt_page(338,698,1).
txt_page(434,698,1).
txt_page(153,698,1).
txt_page(11,698,1).
txt_page(721,698,1).
txt_page(532,698,1).
txt_page(86,698,1).
txt_page(485,698,1).
txt_page(456,698,1).
txt_page(900,698,1).
txt_page(559,698,1).
txt_page(233,698,1).
txt_page(888,698,1).
txt_page(10,698,1).
txt_page(712,699,1).
txt_page(873,699,1).
txt_page(633,699,1).
txt_page(423,699,1).
txt_page(51,699,1).
txt_page(908,699,1).
txt_page(775,699,1).
txt_page(79,699,1).
txt_page(396,699,1).
txt_page(815,699,1).
txt_page(500,699,1).
txt_page(576,699,1).
txt_page(490,699,1).
txt_page(255,699,1).
txt_page(2,699,1).
txt_page(26,699,1).
txt_page(797,699,1).
txt_page(128,699,1).
txt_page(731,699,1).
txt_page(242,699,1).
txt_page(664,700,1).
txt_page(231,700,1).
txt_page(577,700,1).
txt_page(733,700,1).
txt_page(219,700,1).
txt_page(854,700,1).
txt_page(641,700,1).
txt_page(465,700,1).
txt_page(593,700,1).
txt_page(428,700,1).
txt_page(441,700,1).
txt_page(711,700,1).
txt_page(932,700,1).
txt_page(249,700,1).
txt_page(205,700,1).
txt_page(473,700,1).
txt_page(532,700,1).
txt_page(175,700,1).
txt_page(809,700,1).
txt_page(936,700,1).
txt_page(579,701,1).
txt_page(942,701,1).
txt_page(145,701,1).
txt_page(950,701,1).
txt_page(745,701,1).
txt_page(353,701,1).
txt_page(847,701,1).
txt_page(392,701,1).
txt_page(200,701,1).
txt_page(652,701,1).
txt_page(5,701,1).
txt_page(282,701,1).
txt_page(141,701,1).
txt_page(255,701,1).
txt_page(829,701,1).
txt_page(333,701,1).
txt_page(906,701,1).
txt_page(598,701,1).
txt_page(86,701,1).
txt_page(715,701,1).
txt_page(135,702,1).
txt_page(89,702,1).
txt_page(426,702,1).
txt_page(878,702,1).
txt_page(569,702,1).
txt_page(362,702,1).
txt_page(436,702,1).
txt_page(303,702,1).
txt_page(872,702,1).
txt_page(573,702,1).
txt_page(431,702,1).
txt_page(807,702,2).
txt_page(713,702,1).
txt_page(576,702,1).
txt_page(151,702,1).
txt_page(826,702,1).
txt_page(532,702,1).
txt_page(575,702,1).
txt_page(155,702,1).
txt_page(450,703,1).
txt_page(176,703,1).
txt_page(49,703,1).
txt_page(232,703,1).
txt_page(924,703,1).
txt_page(633,703,1).
txt_page(341,703,1).
txt_page(610,703,1).
txt_page(135,703,1).
txt_page(475,703,1).
txt_page(333,703,1).
txt_page(700,703,1).
txt_page(108,703,1).
txt_page(240,703,1).
txt_page(199,703,1).
txt_page(463,703,1).
txt_page(245,703,1).
txt_page(848,703,1).
txt_page(731,703,1).
txt_page(650,703,1).
txt_page(484,704,1).
txt_page(585,704,1).
txt_page(542,704,1).
txt_page(844,704,1).
txt_page(414,704,1).
txt_page(781,704,1).
txt_page(15,704,1).
txt_page(909,704,1).
txt_page(654,704,1).
txt_page(27,704,1).
txt_page(954,704,1).
txt_page(491,704,1).
txt_page(943,704,1).
txt_page(8,704,1).
txt_page(102,704,2).
txt_page(717,704,1).
txt_page(705,704,1).
txt_page(822,704,1).
txt_page(498,704,1).
txt_page(22,705,1).
txt_page(304,705,2).
txt_page(870,705,1).
txt_page(4,705,1).
txt_page(798,705,1).
txt_page(168,705,1).
txt_page(624,705,1).
txt_page(220,705,1).
txt_page(217,705,1).
txt_page(583,705,1).
txt_page(393,705,1).
txt_page(858,705,1).
txt_page(338,705,1).
txt_page(155,705,1).
txt_page(224,705,1).
txt_page(271,705,1).
txt_page(308,705,1).
txt_page(522,705,1).
txt_page(861,705,1).
txt_page(495,706,1).
txt_page(642,706,1).
txt_page(738,706,1).
txt_page(205,706,1).
txt_page(395,706,1).
txt_page(85,706,1).
txt_page(311,706,1).
txt_page(93,706,1).
txt_page(932,706,1).
txt_page(45,706,1).
txt_page(702,706,1).
txt_page(664,706,1).
txt_page(906,706,1).
txt_page(798,706,1).
txt_page(874,706,1).
txt_page(501,706,1).
txt_page(669,706,1).
txt_page(351,706,1).
txt_page(763,706,1).
txt_page(453,706,1).
txt_page(125,707,1).
txt_page(720,707,1).
txt_page(702,707,1).
txt_page(178,707,1).
txt_page(878,707,1).
txt_page(564,707,1).
txt_page(97,707,1).
txt_page(626,707,1).
txt_page(869,707,1).
txt_page(575,707,1).
txt_page(57,707,1).
txt_page(220,707,1).
txt_page(93,707,1).
txt_page(642,707,1).
txt_page(434,707,1).
txt_page(313,707,1).
txt_page(312,707,1).
txt_page(114,707,1).
txt_page(92,707,1).
txt_page(605,707,1).
txt_page(956,708,1).
txt_page(703,708,1).
txt_page(603,708,1).
txt_page(551,708,1).
txt_page(679,708,1).
txt_page(797,708,1).
txt_page(528,708,1).
txt_page(241,708,1).
txt_page(389,708,1).
txt_page(734,708,1).
txt_page(60,708,1).
txt_page(434,708,1).
txt_page(247,708,1).
txt_page(451,708,1).
txt_page(833,708,1).
txt_page(570,708,1).
txt_page(382,708,1).
txt_page(50,708,1).
txt_page(927,708,1).
txt_page(850,708,1).
txt_page(420,709,1).
txt_page(1,709,1).
txt_page(92,709,1).
txt_page(287,709,1).
txt_page(58,709,1).
txt_page(440,709,1).
txt_page(161,709,1).
txt_page(941,709,1).
txt_page(40,709,1).
txt_page(300,709,1).
txt_page(101,709,1).
txt_page(956,709,1).
txt_page(123,709,1).
txt_page(66,709,1).
txt_page(474,709,1).
txt_page(8,709,1).
txt_page(851,709,1).
txt_page(31,709,1).
txt_page(581,709,2).
txt_page(250,710,1).
txt_page(404,710,1).
txt_page(223,710,1).
txt_page(960,710,1).
txt_page(696,710,1).
txt_page(720,710,1).
txt_page(926,710,1).
txt_page(920,710,1).
txt_page(518,710,1).
txt_page(267,710,1).
txt_page(496,710,1).
txt_page(272,710,1).
txt_page(239,710,1).
txt_page(426,710,1).
txt_page(192,710,1).
txt_page(56,710,1).
txt_page(191,710,1).
txt_page(244,710,1).
txt_page(770,710,1).
txt_page(756,710,1).
txt_page(681,711,1).
txt_page(395,711,1).
txt_page(953,711,1).
txt_page(183,711,1).
txt_page(123,711,1).
txt_page(966,711,1).
txt_page(253,711,1).
txt_page(542,711,1).
txt_page(209,711,1).
txt_page(257,711,1).
txt_page(124,711,1).
txt_page(590,711,1).
txt_page(872,711,1).
txt_page(347,711,1).
txt_page(27,711,1).
txt_page(786,711,1).
txt_page(645,711,1).
txt_page(643,711,1).
txt_page(858,711,1).
txt_page(427,711,1).
txt_page(503,712,1).
txt_page(83,712,1).
txt_page(303,712,1).
txt_page(721,712,1).
txt_page(459,712,1).
txt_page(926,712,1).
txt_page(403,712,1).
txt_page(114,712,1).
txt_page(65,712,1).
txt_page(792,712,1).
txt_page(952,712,1).
txt_page(950,712,1).
txt_page(131,712,1).
txt_page(727,712,1).
txt_page(89,712,1).
txt_page(425,712,1).
txt_page(597,712,1).
txt_page(809,712,1).
txt_page(371,712,1).
txt_page(765,712,1).
txt_page(161,713,1).
txt_page(522,713,1).
txt_page(483,713,1).
txt_page(328,713,1).
txt_page(665,713,1).
txt_page(732,713,1).
txt_page(70,713,1).
txt_page(649,713,1).
txt_page(366,713,1).
txt_page(934,713,1).
txt_page(82,713,1).
txt_page(84,713,1).
txt_page(401,713,1).
txt_page(712,713,1).
txt_page(946,713,1).
txt_page(95,713,1).
txt_page(132,713,1).
txt_page(955,713,1).
txt_page(287,713,1).
txt_page(772,713,1).
txt_page(408,714,1).
txt_page(888,714,1).
txt_page(510,714,1).
txt_page(491,714,1).
txt_page(339,714,1).
txt_page(130,714,1).
txt_page(523,714,1).
txt_page(775,714,1).
txt_page(14,714,1).
txt_page(837,714,1).
txt_page(507,714,1).
txt_page(889,714,1).
txt_page(469,714,1).
txt_page(280,714,1).
txt_page(414,714,1).
txt_page(615,714,1).
txt_page(948,714,1).
txt_page(413,714,1).
txt_page(835,714,1).
txt_page(939,714,1).
txt_page(756,715,1).
txt_page(590,715,1).
txt_page(181,715,1).
txt_page(707,715,1).
txt_page(66,715,1).
txt_page(714,715,1).
txt_page(480,715,1).
txt_page(242,715,1).
txt_page(706,715,1).
txt_page(920,715,2).
txt_page(710,715,1).
txt_page(944,715,1).
txt_page(924,715,1).
txt_page(598,715,1).
txt_page(199,715,1).
txt_page(661,715,1).
txt_page(879,715,1).
txt_page(889,715,1).
txt_page(493,715,1).
txt_page(378,716,1).
txt_page(222,716,1).
txt_page(117,716,1).
txt_page(455,716,1).
txt_page(248,716,1).
txt_page(815,716,1).
txt_page(675,716,1).
txt_page(15,716,1).
txt_page(730,716,1).
txt_page(863,716,1).
txt_page(672,716,1).
txt_page(812,716,1).
txt_page(888,716,1).
txt_page(132,716,2).
txt_page(624,716,1).
txt_page(195,716,1).
txt_page(253,716,1).
txt_page(699,716,1).
txt_page(261,716,1).
txt_page(450,717,1).
txt_page(143,717,1).
txt_page(644,717,1).
txt_page(492,717,1).
txt_page(965,717,1).
txt_page(301,717,1).
txt_page(429,717,1).
txt_page(614,717,1).
txt_page(70,717,1).
txt_page(822,717,1).
txt_page(171,717,1).
txt_page(427,717,1).
txt_page(763,717,1).
txt_page(625,717,1).
txt_page(563,717,1).
txt_page(513,717,1).
txt_page(744,717,1).
txt_page(329,717,1).
txt_page(135,717,1).
txt_page(877,717,1).
txt_page(826,718,1).
txt_page(537,718,1).
txt_page(361,718,1).
txt_page(183,718,1).
txt_page(335,718,1).
txt_page(559,718,1).
txt_page(359,718,1).
txt_page(685,718,1).
txt_page(262,718,1).
txt_page(611,718,1).
txt_page(457,718,1).
txt_page(730,718,1).
txt_page(698,718,1).
txt_page(100,718,1).
txt_page(547,718,1).
txt_page(801,718,1).
txt_page(233,718,1).
txt_page(845,718,1).
txt_page(639,718,1).
txt_page(927,718,1).
txt_page(958,719,1).
txt_page(262,719,1).
txt_page(190,719,1).
txt_page(235,719,1).
txt_page(481,719,1).
txt_page(413,719,1).
txt_page(831,719,1).
txt_page(727,719,1).
txt_page(297,719,1).
txt_page(948,719,1).
txt_page(300,719,1).
txt_page(443,719,1).
txt_page(876,719,1).
txt_page(378,719,1).
txt_page(72,719,1).
txt_page(296,719,1).
txt_page(595,719,1).
txt_page(961,719,1).
txt_page(376,719,1).
txt_page(597,719,1).
txt_page(758,720,1).
txt_page(2,720,1).
txt_page(195,720,1).
txt_page(568,720,1).
txt_page(411,720,1).
txt_page(107,720,1).
txt_page(796,720,1).
txt_page(699,720,1).
txt_page(682,720,1).
txt_page(236,720,1).
txt_page(218,720,1).
txt_page(432,720,1).
txt_page(491,720,1).
txt_page(864,720,1).
txt_page(508,720,1).
txt_page(613,720,1).
txt_page(764,720,1).
txt_page(316,720,1).
txt_page(11,720,1).
txt_page(64,720,1).
txt_page(323,721,1).
txt_page(402,721,1).
txt_page(835,721,1).
txt_page(263,721,1).
txt_page(532,721,1).
txt_page(789,721,1).
txt_page(752,721,1).
txt_page(364,721,1).
txt_page(538,721,1).
txt_page(490,721,1).
txt_page(389,721,1).
txt_page(649,721,1).
txt_page(373,721,1).
txt_page(147,721,1).
txt_page(950,721,1).
txt_page(620,721,1).
txt_page(220,721,1).
txt_page(896,721,1).
txt_page(282,721,2).
txt_page(699,722,1).
txt_page(151,722,1).
txt_page(89,722,1).
txt_page(15,722,1).
txt_page(939,722,1).
txt_page(775,722,1).
txt_page(336,722,1).
txt_page(593,722,1).
txt_page(674,722,1).
txt_page(253,722,1).
txt_page(626,722,1).
txt_page(702,722,1).
txt_page(324,722,1).
txt_page(52,722,1).
txt_page(666,722,1).
txt_page(320,722,1).
txt_page(66,722,2).
txt_page(806,722,1).
txt_page(524,722,1).
txt_page(632,723,1).
txt_page(173,723,1).
txt_page(442,723,1).
txt_page(184,723,1).
txt_page(128,723,1).
txt_page(782,723,1).
txt_page(506,723,1).
txt_page(456,723,1).
txt_page(785,723,1).
txt_page(555,723,1).
txt_page(477,723,1).
txt_page(464,723,1).
txt_page(859,723,1).
txt_page(552,723,1).
txt_page(263,723,1).
txt_page(657,723,1).
txt_page(81,723,1).
txt_page(158,723,1).
txt_page(329,723,1).
txt_page(249,723,1).
txt_page(710,724,1).
txt_page(452,724,1).
txt_page(30,724,1).
txt_page(544,724,1).
txt_page(69,724,1).
txt_page(368,724,1).
txt_page(883,724,1).
txt_page(434,724,1).
txt_page(225,724,1).
txt_page(687,724,1).
txt_page(339,724,1).
txt_page(906,724,1).
txt_page(326,724,1).
txt_page(229,724,1).
txt_page(528,724,1).
txt_page(875,724,1).
txt_page(777,724,1).
txt_page(704,724,1).
txt_page(535,724,1).
txt_page(607,724,1).
txt_page(472,725,1).
txt_page(943,725,1).
txt_page(128,725,1).
txt_page(372,725,1).
txt_page(303,725,1).
txt_page(725,725,1).
txt_page(301,725,1).
txt_page(934,725,2).
txt_page(657,725,1).
txt_page(395,725,1).
txt_page(134,725,1).
txt_page(816,725,1).
txt_page(421,725,1).
txt_page(458,725,1).
txt_page(650,725,1).
txt_page(336,725,1).
txt_page(58,725,1).
txt_page(552,725,1).
txt_page(449,725,1).
txt_page(522,726,1).
txt_page(666,726,1).
txt_page(464,726,1).
txt_page(200,726,1).
txt_page(234,726,1).
txt_page(13,726,1).
txt_page(631,726,1).
txt_page(232,726,1).
txt_page(404,726,1).
txt_page(319,726,1).
txt_page(118,726,1).
txt_page(238,726,1).
txt_page(566,726,1).
txt_page(231,726,1).
txt_page(839,726,1).
txt_page(588,726,1).
txt_page(692,726,1).
txt_page(312,726,1).
txt_page(754,726,1).
txt_page(60,726,1).
txt_page(141,727,1).
txt_page(643,727,1).
txt_page(591,727,1).
txt_page(119,727,1).
txt_page(958,727,1).
txt_page(236,727,2).
txt_page(361,727,2).
txt_page(498,727,1).
txt_page(378,727,1).
txt_page(835,727,1).
txt_page(753,727,1).
txt_page(354,727,1).
txt_page(312,727,1).
txt_page(859,727,1).
txt_page(846,727,1).
txt_page(873,727,1).
txt_page(219,727,1).
txt_page(22,727,1).
txt_page(453,728,1).
txt_page(391,728,1).
txt_page(487,728,1).
txt_page(499,728,1).
txt_page(634,728,1).
txt_page(545,728,1).
txt_page(573,728,1).
txt_page(131,728,1).
txt_page(810,728,1).
txt_page(291,728,1).
txt_page(23,728,1).
txt_page(73,728,1).
txt_page(126,728,1).
txt_page(737,728,1).
txt_page(272,728,1).
txt_page(702,728,1).
txt_page(735,728,1).
txt_page(444,728,1).
txt_page(226,728,1).
txt_page(762,728,1).
txt_page(513,729,1).
txt_page(659,729,1).
txt_page(383,729,1).
txt_page(808,729,1).
txt_page(236,729,1).
txt_page(936,729,1).
txt_page(915,729,1).
txt_page(116,729,1).
txt_page(153,729,1).
txt_page(781,729,1).
txt_page(556,729,1).
txt_page(84,729,1).
txt_page(344,729,1).
txt_page(365,729,1).
txt_page(670,729,1).
txt_page(629,729,1).
txt_page(658,729,1).
txt_page(544,729,1).
txt_page(99,729,1).
txt_page(246,729,1).
txt_page(675,730,1).
txt_page(93,730,1).
txt_page(74,730,1).
txt_page(120,730,1).
txt_page(872,730,1).
txt_page(178,730,1).
txt_page(233,730,1).
txt_page(701,730,1).
txt_page(892,730,1).
txt_page(494,730,1).
txt_page(103,730,1).
txt_page(593,730,1).
txt_page(244,730,1).
txt_page(313,730,1).
txt_page(485,730,1).
txt_page(721,730,1).
txt_page(582,730,1).
txt_page(951,730,1).
txt_page(17,730,1).
txt_page(214,730,1).
txt_page(377,731,1).
txt_page(584,731,1).
txt_page(660,731,1).
txt_page(291,731,1).
txt_page(508,731,1).
txt_page(633,731,1).
txt_page(477,731,1).
txt_page(855,731,1).
txt_page(417,731,1).
txt_page(148,731,1).
txt_page(5,731,1).
txt_page(173,731,1).
txt_page(492,731,1).
txt_page(800,731,2).
txt_page(280,731,1).
txt_page(947,731,1).
txt_page(653,731,1).
txt_page(295,731,1).
txt_page(436,731,1).
txt_page(674,732,1).
txt_page(529,732,1).
txt_page(232,732,1).
txt_page(594,732,1).
txt_page(759,732,1).
txt_page(432,732,1).
txt_page(44,732,1).
txt_page(47,732,1).
txt_page(74,732,1).
txt_page(9,732,1).
txt_page(532,732,1).
txt_page(577,732,1).
txt_page(71,732,1).
txt_page(94,732,1).
txt_page(932,732,1).
txt_page(480,732,1).
txt_page(795,732,1).
txt_page(502,732,1).
txt_page(347,732,1).
txt_page(585,732,1).
txt_page(60,733,1).
txt_page(714,733,1).
txt_page(81,733,1).
txt_page(648,733,1).
txt_page(428,733,1).
txt_page(659,733,1).
txt_page(201,733,1).
txt_page(228,733,1).
txt_page(155,733,1).
txt_page(6,733,1).
txt_page(162,733,1).
txt_page(895,733,1).
txt_page(458,733,1).
txt_page(158,733,1).
txt_page(448,733,1).
txt_page(745,733,1).
txt_page(690,733,1).
txt_page(571,733,1).
txt_page(117,733,1).
txt_page(461,733,1).
txt_page(703,734,1).
txt_page(511,734,1).
txt_page(248,734,1).
txt_page(490,734,1).
txt_page(163,734,1).
txt_page(962,734,1).
txt_page(416,734,1).
txt_page(780,734,1).
txt_page(525,734,1).
txt_page(102,734,1).
txt_page(434,734,1).
txt_page(301,734,1).
txt_page(912,734,1).
txt_page(333,734,1).
txt_page(469,734,1).
txt_page(655,734,1).
txt_page(43,734,1).
txt_page(767,734,2).
txt_page(571,734,1).
txt_page(477,735,1).
txt_page(592,735,1).
txt_page(264,735,1).
txt_page(560,735,1).
txt_page(287,735,1).
txt_page(216,735,1).
txt_page(740,735,1).
txt_page(850,735,1).
txt_page(741,735,1).
txt_page(486,735,1).
txt_page(201,735,1).
txt_page(17,735,1).
txt_page(280,735,1).
txt_page(126,735,1).
txt_page(22,735,1).
txt_page(172,735,1).
txt_page(202,735,1).
txt_page(353,735,1).
txt_page(769,735,1).
txt_page(895,735,1).
txt_page(681,736,1).
txt_page(904,736,1).
txt_page(105,736,1).
txt_page(224,736,1).
txt_page(40,736,1).
txt_page(918,736,1).
txt_page(408,736,1).
txt_page(232,736,1).
txt_page(655,736,1).
txt_page(915,736,1).
txt_page(740,736,1).
txt_page(880,736,1).
txt_page(275,736,1).
txt_page(483,736,1).
txt_page(791,736,1).
txt_page(628,736,1).
txt_page(645,736,1).
txt_page(349,736,1).
txt_page(230,736,1).
txt_page(965,736,1).
txt_page(364,737,1).
txt_page(664,737,1).
txt_page(591,737,1).
txt_page(175,737,1).
txt_page(115,737,1).
txt_page(660,737,1).
txt_page(964,737,1).
txt_page(734,737,1).
txt_page(814,737,1).
txt_page(116,737,1).
txt_page(961,737,1).
txt_page(741,737,1).
txt_page(3,737,1).
txt_page(498,737,1).
txt_page(564,737,1).
txt_page(752,737,1).
txt_page(551,737,1).
txt_page(745,737,1).
txt_page(682,737,1).
txt_page(765,737,1).
txt_page(734,738,1).
txt_page(459,738,1).
txt_page(717,738,1).
txt_page(727,738,1).
txt_page(744,738,1).
txt_page(395,738,1).
txt_page(410,738,1).
txt_page(226,738,1).
txt_page(292,738,1).
txt_page(781,738,1).
txt_page(500,738,1).
txt_page(119,738,1).
txt_page(510,738,1).
txt_page(319,738,1).
txt_page(182,738,1).
txt_page(40,738,1).
txt_page(680,738,1).
txt_page(897,738,1).
txt_page(751,738,1).
txt_page(606,738,1).
txt_page(228,739,1).
txt_page(958,739,1).
txt_page(614,739,1).
txt_page(30,739,1).
txt_page(7,739,1).
txt_page(498,739,1).
txt_page(447,739,1).
txt_page(116,739,1).
txt_page(802,739,1).
txt_page(824,739,1).
txt_page(688,739,1).
txt_page(741,739,1).
txt_page(485,739,1).
txt_page(941,739,1).
txt_page(938,739,1).
txt_page(333,739,1).
txt_page(879,739,1).
txt_page(708,739,1).
txt_page(960,739,1).
txt_page(489,739,1).
txt_page(346,740,1).
txt_page(29,740,1).
txt_page(565,740,1).
txt_page(340,740,1).
txt_page(116,740,1).
txt_page(861,740,1).
txt_page(555,740,1).
txt_page(247,740,1).
txt_page(36,740,1).
txt_page(224,740,1).
txt_page(525,740,1).
txt_page(880,740,1).
txt_page(487,740,1).
txt_page(849,740,1).
txt_page(193,740,2).
txt_page(27,740,1).
txt_page(966,740,1).
txt_page(39,740,1).
txt_page(135,740,1).
txt_page(630,741,1).
txt_page(480,741,1).
txt_page(252,741,1).
txt_page(186,741,1).
txt_page(801,741,1).
txt_page(718,741,1).
txt_page(368,741,1).
txt_page(339,741,1).
txt_page(819,741,1).
txt_page(663,741,1).
txt_page(218,741,1).
txt_page(569,741,1).
txt_page(581,741,1).
txt_page(670,741,1).
txt_page(734,741,1).
txt_page(637,741,1).
txt_page(12,741,1).
txt_page(409,741,1).
txt_page(577,741,1).
txt_page(747,741,1).
txt_page(876,742,1).
txt_page(579,742,1).
txt_page(407,742,1).
txt_page(148,742,1).
txt_page(416,742,1).
txt_page(457,742,1).
txt_page(224,742,1).
txt_page(403,742,1).
txt_page(210,742,1).
txt_page(215,742,1).
txt_page(394,742,1).
txt_page(812,742,1).
txt_page(334,742,1).
txt_page(319,742,1).
txt_page(35,742,1).
txt_page(966,742,1).
txt_page(878,742,1).
txt_page(426,742,1).
txt_page(8,742,1).
txt_page(221,742,1).
txt_page(486,743,1).
txt_page(224,743,1).
txt_page(324,743,1).
txt_page(571,743,1).
txt_page(38,743,1).
txt_page(406,743,1).
txt_page(428,743,1).
txt_page(882,743,1).
txt_page(342,743,1).
txt_page(340,743,1).
txt_page(725,743,1).
txt_page(50,743,1).
txt_page(921,743,1).
txt_page(524,743,1).
txt_page(129,743,1).
txt_page(327,743,1).
txt_page(507,743,1).
txt_page(466,743,1).
txt_page(762,743,1).
txt_page(769,743,1).
txt_page(351,744,1).
txt_page(493,744,1).
txt_page(752,744,1).
txt_page(709,744,1).
txt_page(640,744,1).
txt_page(481,744,1).
txt_page(60,744,1).
txt_page(727,744,1).
txt_page(389,744,1).
txt_page(648,744,1).
txt_page(262,744,1).
txt_page(673,744,1).
txt_page(783,744,1).
txt_page(248,744,1).
txt_page(401,744,1).
txt_page(291,744,1).
txt_page(500,744,1).
txt_page(867,744,1).
txt_page(644,744,1).
txt_page(740,744,1).
txt_page(781,745,1).
txt_page(207,745,1).
txt_page(284,745,1).
txt_page(425,745,1).
txt_page(500,745,1).
txt_page(522,745,1).
txt_page(878,745,1).
txt_page(72,745,1).
txt_page(857,745,1).
txt_page(357,745,1).
txt_page(611,745,1).
txt_page(869,745,1).
txt_page(661,745,1).
txt_page(549,745,1).
txt_page(686,745,2).
txt_page(277,745,1).
txt_page(464,745,1).
txt_page(701,745,1).
txt_page(408,745,1).
txt_page(520,746,1).
txt_page(249,746,1).
txt_page(596,746,1).
txt_page(759,746,1).
txt_page(92,746,1).
txt_page(489,746,1).
txt_page(964,746,1).
txt_page(46,746,1).
txt_page(732,746,1).
txt_page(738,746,1).
txt_page(369,746,1).
txt_page(751,746,1).
txt_page(80,746,1).
txt_page(811,746,1).
txt_page(404,746,1).
txt_page(41,746,1).
txt_page(421,746,1).
txt_page(192,746,1).
txt_page(437,746,1).
txt_page(142,746,1).
txt_page(498,747,1).
txt_page(771,747,1).
txt_page(105,747,1).
txt_page(588,747,1).
txt_page(548,747,1).
txt_page(673,747,1).
txt_page(838,747,1).
txt_page(332,747,1).
txt_page(50,747,1).
txt_page(827,747,1).
txt_page(667,747,1).
txt_page(488,747,1).
txt_page(702,747,1).
txt_page(28,747,2).
txt_page(967,747,1).
txt_page(115,747,1).
txt_page(743,747,1).
txt_page(270,747,1).
txt_page(201,747,1).
txt_page(350,748,1).
txt_page(817,748,1).
txt_page(917,748,1).
txt_page(146,748,1).
txt_page(123,748,1).
txt_page(258,748,1).
txt_page(567,748,1).
txt_page(305,748,1).
txt_page(813,748,1).
txt_page(684,748,1).
txt_page(686,748,1).
txt_page(112,748,1).
txt_page(575,748,1).
txt_page(930,748,1).
txt_page(414,748,1).
txt_page(536,748,1).
txt_page(47,748,2).
txt_page(615,748,1).
txt_page(790,748,1).
txt_page(412,749,1).
txt_page(827,749,1).
txt_page(339,749,1).
txt_page(272,749,1).
txt_page(678,749,1).
txt_page(436,749,1).
txt_page(876,749,1).
txt_page(65,749,1).
txt_page(225,749,1).
txt_page(792,749,1).
txt_page(521,749,1).
txt_page(859,749,1).
txt_page(270,749,1).
txt_page(181,749,1).
txt_page(952,749,1).
txt_page(940,749,1).
txt_page(814,749,1).
txt_page(336,749,1).
txt_page(372,749,1).
txt_page(576,749,1).
txt_page(652,750,1).
txt_page(7,750,1).
txt_page(149,750,1).
txt_page(342,750,1).
txt_page(647,750,1).
txt_page(412,750,1).
txt_page(605,750,1).
txt_page(532,750,1).
txt_page(219,750,1).
txt_page(675,750,2).
txt_page(938,750,1).
txt_page(621,750,1).
txt_page(108,750,1).
txt_page(829,750,1).
txt_page(821,750,1).
txt_page(287,750,1).
txt_page(22,750,1).
txt_page(759,750,1).
txt_page(769,750,1).
txt_page(203,751,1).
txt_page(588,751,1).
txt_page(193,751,1).
txt_page(248,751,1).
txt_page(210,751,1).
txt_page(399,751,1).
txt_page(684,751,1).
txt_page(632,751,1).
txt_page(76,751,1).
txt_page(670,751,1).
txt_page(638,751,1).
txt_page(423,751,1).
txt_page(870,751,1).
txt_page(609,751,1).
txt_page(737,751,1).
txt_page(35,751,1).
txt_page(60,751,1).
txt_page(207,751,1).
txt_page(392,751,1).
txt_page(245,751,1).
txt_page(710,752,1).
txt_page(90,752,1).
txt_page(47,752,1).
txt_page(271,752,1).
txt_page(173,752,1).
txt_page(930,752,1).
txt_page(961,752,1).
txt_page(842,752,1).
txt_page(614,752,1).
txt_page(280,752,1).
txt_page(437,752,1).
txt_page(495,752,1).
txt_page(191,752,1).
txt_page(207,752,1).
txt_page(98,752,1).
txt_page(44,752,1).
txt_page(502,752,1).
txt_page(860,752,1).
txt_page(257,752,1).
txt_page(114,752,1).
txt_page(393,753,1).
txt_page(495,753,1).
txt_page(446,753,1).
txt_page(220,753,1).
txt_page(917,753,1).
txt_page(160,753,1).
txt_page(318,753,1).
txt_page(813,753,1).
txt_page(248,753,1).
txt_page(651,753,1).
txt_page(478,753,1).
txt_page(771,753,1).
txt_page(614,753,1).
txt_page(307,753,1).
txt_page(952,753,1).
txt_page(948,753,1).
txt_page(864,753,1).
txt_page(894,753,1).
txt_page(613,753,1).
txt_page(99,753,1).
txt_page(935,754,1).
txt_page(569,754,1).
txt_page(864,754,1).
txt_page(599,754,1).
txt_page(318,754,1).
txt_page(45,754,1).
txt_page(414,754,1).
txt_page(677,754,1).
txt_page(959,754,1).
txt_page(210,754,1).
txt_page(188,754,1).
txt_page(138,754,1).
txt_page(814,754,1).
txt_page(272,754,1).
txt_page(559,754,1).
txt_page(537,754,2).
txt_page(77,754,1).
txt_page(313,754,1).
txt_page(252,754,1).
txt_page(138,755,1).
txt_page(795,755,1).
txt_page(171,755,1).
txt_page(10,755,1).
txt_page(378,755,1).
txt_page(548,755,1).
txt_page(883,755,1).
txt_page(743,755,1).
txt_page(397,755,1).
txt_page(967,755,1).
txt_page(715,755,1).
txt_page(55,755,1).
txt_page(14,755,1).
txt_page(657,755,1).
txt_page(607,755,1).
txt_page(709,755,1).
txt_page(457,755,1).
txt_page(768,755,1).
txt_page(786,755,1).
txt_page(675,755,1).
txt_page(191,756,1).
txt_page(522,756,1).
txt_page(25,756,1).
txt_page(716,756,1).
txt_page(148,756,1).
txt_page(610,756,1).
txt_page(99,756,1).
txt_page(18,756,1).
txt_page(655,756,1).
txt_page(496,756,1).
txt_page(311,756,1).
txt_page(327,756,1).
txt_page(786,756,1).
txt_page(489,756,1).
txt_page(456,756,1).
txt_page(854,756,1).
txt_page(457,756,1).
txt_page(884,756,1).
txt_page(590,756,1).
txt_page(955,756,1).
txt_page(247,757,1).
txt_page(861,757,1).
txt_page(224,757,1).
txt_page(13,757,1).
txt_page(842,757,1).
txt_page(834,757,1).
txt_page(862,757,1).
txt_page(895,757,1).
txt_page(605,757,1).
txt_page(622,757,1).
txt_page(125,757,1).
txt_page(510,757,1).
txt_page(679,757,1).
txt_page(150,757,1).
txt_page(945,757,1).
txt_page(952,757,1).
txt_page(900,757,1).
txt_page(220,757,1).
txt_page(713,757,1).
txt_page(447,757,1).
txt_page(964,758,1).
txt_page(174,758,1).
txt_page(548,758,1).
txt_page(127,758,1).
txt_page(682,758,1).
txt_page(406,758,1).
txt_page(660,758,1).
txt_page(50,758,1).
txt_page(468,758,1).
txt_page(773,758,1).
txt_page(175,758,1).
txt_page(255,758,1).
txt_page(478,758,1).
txt_page(828,758,1).
txt_page(423,758,1).
txt_page(71,758,1).
txt_page(698,758,1).
txt_page(763,758,1).
txt_page(179,758,1).
txt_page(549,758,1).
txt_page(443,759,1).
txt_page(228,759,1).
txt_page(432,759,1).
txt_page(796,759,1).
txt_page(744,759,1).
txt_page(926,759,1).
txt_page(811,759,1).
txt_page(577,759,1).
txt_page(414,759,1).
txt_page(705,759,1).
txt_page(942,759,1).
txt_page(928,759,1).
txt_page(480,759,1).
txt_page(217,759,1).
txt_page(134,759,1).
txt_page(46,759,1).
txt_page(157,759,1).
txt_page(135,759,1).
txt_page(374,759,1).
txt_page(690,759,1).
txt_page(523,760,1).
txt_page(61,760,1).
txt_page(275,760,1).
txt_page(120,760,1).
txt_page(385,760,1).
txt_page(671,760,1).
txt_page(538,760,1).
txt_page(716,760,1).
txt_page(885,760,1).
txt_page(449,760,1).
txt_page(678,760,1).
txt_page(254,760,1).
txt_page(478,760,1).
txt_page(862,760,1).
txt_page(645,760,1).
txt_page(156,760,1).
txt_page(27,760,1).
txt_page(184,760,1).
txt_page(914,760,1).
txt_page(943,760,1).
txt_page(171,761,1).
txt_page(692,761,1).
txt_page(22,761,1).
txt_page(612,761,1).
txt_page(475,761,1).
txt_page(231,761,2).
txt_page(495,761,1).
txt_page(844,761,1).
txt_page(627,761,1).
txt_page(339,761,1).
txt_page(419,761,1).
txt_page(810,761,1).
txt_page(533,761,1).
txt_page(77,761,1).
txt_page(80,761,1).
txt_page(605,761,1).
txt_page(338,761,1).
txt_page(682,761,1).
txt_page(525,761,1).
txt_page(822,762,1).
txt_page(908,762,1).
txt_page(84,762,1).
txt_page(369,762,1).
txt_page(426,762,1).
txt_page(553,762,1).
txt_page(326,762,1).
txt_page(19,762,1).
txt_page(566,762,1).
txt_page(16,762,1).
txt_page(280,762,1).
txt_page(249,762,1).
txt_page(160,762,1).
txt_page(835,762,1).
txt_page(740,762,1).
txt_page(518,762,1).
txt_page(559,762,1).
txt_page(738,762,1).
txt_page(85,762,1).
txt_page(475,762,1).
txt_page(681,763,1).
txt_page(848,763,1).
txt_page(833,763,1).
txt_page(358,763,1).
txt_page(903,763,1).
txt_page(742,763,1).
txt_page(408,763,1).
txt_page(684,763,1).
txt_page(414,763,1).
txt_page(639,763,1).
txt_page(966,763,1).
txt_page(184,763,1).
txt_page(193,763,1).
txt_page(190,763,1).
txt_page(93,763,1).
txt_page(673,763,1).
txt_page(156,763,1).
txt_page(113,763,1).
txt_page(132,763,1).
txt_page(378,763,1).
txt_page(97,764,1).
txt_page(388,764,1).
txt_page(756,764,1).
txt_page(545,764,1).
txt_page(957,764,1).
txt_page(910,764,1).
txt_page(281,764,1).
txt_page(193,764,1).
txt_page(44,764,1).
txt_page(410,764,1).
txt_page(904,764,1).
txt_page(351,764,1).
txt_page(431,764,1).
txt_page(719,764,1).
txt_page(958,764,1).
txt_page(401,764,1).
txt_page(237,764,1).
txt_page(114,764,1).
txt_page(763,764,1).
txt_page(137,764,1).
txt_page(362,765,1).
txt_page(699,765,1).
txt_page(177,765,1).
txt_page(738,765,1).
txt_page(943,765,1).
txt_page(706,765,1).
txt_page(572,765,1).
txt_page(589,765,1).
txt_page(774,765,1).
txt_page(206,765,1).
txt_page(523,765,1).
txt_page(546,765,1).
txt_page(305,765,1).
txt_page(790,765,1).
txt_page(455,765,1).
txt_page(909,765,1).
txt_page(743,765,1).
txt_page(652,765,1).
txt_page(251,765,1).
txt_page(569,765,1).
txt_page(202,766,1).
txt_page(69,766,1).
txt_page(656,766,1).
txt_page(900,766,1).
txt_page(187,766,1).
txt_page(276,766,1).
txt_page(706,766,1).
txt_page(229,766,1).
txt_page(197,766,1).
txt_page(432,766,1).
txt_page(608,766,2).
txt_page(81,766,1).
txt_page(373,766,1).
txt_page(312,766,1).
txt_page(420,766,1).
txt_page(875,766,1).
txt_page(298,766,1).
txt_page(392,766,1).
txt_page(465,766,1).
txt_page(516,767,1).
txt_page(613,767,1).
txt_page(151,767,1).
txt_page(543,767,1).
txt_page(527,767,1).
txt_page(253,767,1).
txt_page(270,767,1).
txt_page(533,767,1).
txt_page(329,767,1).
txt_page(1,767,1).
txt_page(522,767,1).
txt_page(726,767,1).
txt_page(180,767,1).
txt_page(546,767,1).
txt_page(589,767,1).
txt_page(949,767,1).
txt_page(55,767,1).
txt_page(729,767,1).
txt_page(369,767,1).
txt_page(203,767,1).
txt_page(804,768,1).
txt_page(661,768,1).
txt_page(350,768,1).
txt_page(456,768,1).
txt_page(485,768,1).
txt_page(371,768,1).
txt_page(37,768,1).
txt_page(230,768,1).
txt_page(640,768,1).
txt_page(420,768,1).
txt_page(564,768,1).
txt_page(538,768,1).
txt_page(471,768,1).
txt_page(216,768,1).
txt_page(234,768,1).
txt_page(707,768,1).
txt_page(597,768,1).
txt_page(93,768,1).
txt_page(689,768,1).
txt_page(559,768,1).
txt_page(213,769,1).
txt_page(191,769,1).
txt_page(636,769,1).
txt_page(380,769,1).
txt_page(598,769,1).
txt_page(17,769,1).
txt_page(733,769,1).
txt_page(206,769,1).
txt_page(205,769,1).
txt_page(73,769,1).
txt_page(42,769,1).
txt_page(333,769,1).
txt_page(496,769,1).
txt_page(201,769,1).
txt_page(328,769,1).
txt_page(967,769,1).
txt_page(159,769,1).
txt_page(572,769,1).
txt_page(765,769,1).
txt_page(346,769,1).
txt_page(406,770,1).
txt_page(967,770,1).
txt_page(453,770,1).
txt_page(953,770,1).
txt_page(271,770,1).
txt_page(133,770,1).
txt_page(868,770,1).
txt_page(11,770,1).
txt_page(753,770,1).
txt_page(185,770,1).
txt_page(55,770,1).
txt_page(330,770,1).
txt_page(797,770,1).
txt_page(371,770,1).
txt_page(536,770,1).
txt_page(609,770,1).
txt_page(800,770,1).
txt_page(231,770,1).
txt_page(345,770,1).
txt_page(466,770,1).
txt_page(555,771,1).
txt_page(968,771,1).
txt_page(126,771,1).
txt_page(960,771,1).
txt_page(631,771,1).
txt_page(888,771,1).
txt_page(879,771,1).
txt_page(952,771,1).
txt_page(14,771,1).
txt_page(874,771,1).
txt_page(698,771,1).
txt_page(447,771,1).
txt_page(223,771,1).
txt_page(601,771,1).
txt_page(156,771,1).
txt_page(430,771,1).
txt_page(704,771,1).
txt_page(814,771,1).
txt_page(409,771,1).
txt_page(336,771,1).
txt_page(50,772,1).
txt_page(869,772,1).
txt_page(558,772,1).
txt_page(151,772,1).
txt_page(454,772,1).
txt_page(884,772,1).
txt_page(127,772,1).
txt_page(16,772,1).
txt_page(46,772,1).
txt_page(94,772,1).
txt_page(889,772,1).
txt_page(394,772,1).
txt_page(798,772,1).
txt_page(588,772,1).
txt_page(954,772,1).
txt_page(549,772,1).
txt_page(510,772,1).
txt_page(2,772,1).
txt_page(429,772,1).
txt_page(38,772,1).
txt_page(206,773,1).
txt_page(245,773,1).
txt_page(523,773,1).
txt_page(552,773,1).
txt_page(722,773,1).
txt_page(175,773,1).
txt_page(818,773,1).
txt_page(831,773,1).
txt_page(173,773,1).
txt_page(344,773,1).
txt_page(361,773,1).
txt_page(741,773,1).
txt_page(535,773,1).
txt_page(323,773,1).
txt_page(174,773,1).
txt_page(384,773,1).
txt_page(621,773,1).
txt_page(135,773,1).
txt_page(102,773,1).
txt_page(364,773,1).
txt_page(251,774,1).
txt_page(332,774,1).
txt_page(413,774,1).
txt_page(518,774,1).
txt_page(725,774,1).
txt_page(762,774,1).
txt_page(736,774,1).
txt_page(132,774,1).
txt_page(537,774,1).
txt_page(808,774,1).
txt_page(19,774,1).
txt_page(475,774,1).
txt_page(464,774,1).
txt_page(245,774,1).
txt_page(938,774,1).
txt_page(922,774,1).
txt_page(742,774,1).
txt_page(44,774,1).
txt_page(451,774,1).
txt_page(689,774,1).
txt_page(967,775,1).
txt_page(741,775,1).
txt_page(595,775,1).
txt_page(251,775,2).
txt_page(467,775,1).
txt_page(93,775,1).
txt_page(282,775,1).
txt_page(35,775,1).
txt_page(55,775,1).
txt_page(635,775,1).
txt_page(875,775,1).
txt_page(77,775,1).
txt_page(358,775,1).
txt_page(132,775,1).
txt_page(615,775,1).
txt_page(221,775,1).
txt_page(778,775,1).
txt_page(146,775,1).
txt_page(896,775,1).
txt_page(226,776,1).
txt_page(162,776,1).
txt_page(499,776,1).
txt_page(212,776,1).
txt_page(441,776,1).
txt_page(935,776,1).
txt_page(216,776,1).
txt_page(547,776,1).
txt_page(119,776,1).
txt_page(470,776,1).
txt_page(271,776,1).
txt_page(449,776,1).
txt_page(682,776,1).
txt_page(919,776,1).
txt_page(947,776,1).
txt_page(452,776,1).
txt_page(722,776,1).
txt_page(589,776,1).
txt_page(21,776,1).
txt_page(912,776,1).
txt_page(831,777,1).
txt_page(314,777,1).
txt_page(372,777,1).
txt_page(53,777,1).
txt_page(243,777,1).
txt_page(395,777,1).
txt_page(478,777,1).
txt_page(342,777,1).
txt_page(125,777,1).
txt_page(355,777,1).
txt_page(67,777,1).
txt_page(707,777,1).
txt_page(275,777,1).
txt_page(864,777,1).
txt_page(429,777,1).
txt_page(940,777,1).
txt_page(560,777,1).
txt_page(295,777,1).
txt_page(59,777,1).
txt_page(510,777,1).
txt_page(374,778,1).
txt_page(757,778,1).
txt_page(741,778,1).
txt_page(543,778,1).
txt_page(229,778,1).
txt_page(625,778,1).
txt_page(793,778,1).
txt_page(165,778,1).
txt_page(357,778,1).
txt_page(479,778,1).
txt_page(257,778,1).
txt_page(57,778,1).
txt_page(872,778,1).
txt_page(166,778,1).
txt_page(63,778,1).
txt_page(703,778,1).
txt_page(205,778,1).
txt_page(645,778,2).
txt_page(681,778,1).
txt_page(573,779,1).
txt_page(683,779,1).
txt_page(716,779,1).
txt_page(726,779,1).
txt_page(490,779,1).
txt_page(25,779,1).
txt_page(632,779,1).
txt_page(952,779,1).
txt_page(542,779,1).
txt_page(846,779,1).
txt_page(118,779,1).
txt_page(889,779,1).
txt_page(509,779,1).
txt_page(386,779,1).
txt_page(525,779,1).
txt_page(847,779,2).
txt_page(639,779,1).
txt_page(262,779,1).
txt_page(554,779,1).
txt_page(397,780,1).
txt_page(960,780,1).
txt_page(367,780,1).
txt_page(892,780,1).
txt_page(864,780,1).
txt_page(652,780,1).
txt_page(678,780,1).
txt_page(211,780,1).
txt_page(230,780,1).
txt_page(891,780,1).
txt_page(177,780,1).
txt_page(893,780,1).
txt_page(197,780,1).
txt_page(143,780,1).
txt_page(697,780,1).
txt_page(307,780,1).
txt_page(811,780,1).
txt_page(334,780,1).
txt_page(10,780,1).
txt_page(81,780,1).
txt_page(891,781,1).
txt_page(546,781,1).
txt_page(314,781,1).
txt_page(815,781,1).
txt_page(722,781,1).
txt_page(588,781,1).
txt_page(762,781,1).
txt_page(728,781,1).
txt_page(296,781,1).
txt_page(402,781,1).
txt_page(396,781,1).
txt_page(385,781,1).
txt_page(655,781,1).
txt_page(368,781,1).
txt_page(751,781,1).
txt_page(66,781,1).
txt_page(593,781,1).
txt_page(887,781,1).
txt_page(365,781,1).
txt_page(462,781,1).
txt_page(660,782,1).
txt_page(27,782,1).
txt_page(186,782,1).
txt_page(147,782,1).
txt_page(558,782,1).
txt_page(683,782,1).
txt_page(332,782,1).
txt_page(48,782,1).
txt_page(115,782,1).
txt_page(351,782,1).
txt_page(692,782,1).
txt_page(713,782,1).
txt_page(183,782,1).
txt_page(940,782,1).
txt_page(46,782,1).
txt_page(935,782,1).
txt_page(745,782,1).
txt_page(387,782,1).
txt_page(654,782,1).
txt_page(134,782,1).
txt_page(321,783,1).
txt_page(691,783,1).
txt_page(5,783,1).
txt_page(760,783,1).
txt_page(59,783,2).
txt_page(809,783,1).
txt_page(191,783,1).
txt_page(262,783,1).
txt_page(515,783,1).
txt_page(407,783,1).
txt_page(129,783,1).
txt_page(754,783,1).
txt_page(870,783,1).
txt_page(807,783,1).
txt_page(474,783,1).
txt_page(784,783,1).
txt_page(154,783,1).
txt_page(92,783,1).
txt_page(873,783,1).
txt_page(69,784,1).
txt_page(819,784,1).
txt_page(791,784,1).
txt_page(907,784,1).
txt_page(96,784,1).
txt_page(170,784,1).
txt_page(860,784,1).
txt_page(165,784,1).
txt_page(782,784,1).
txt_page(613,784,1).
txt_page(294,784,1).
txt_page(637,784,1).
txt_page(694,784,1).
txt_page(739,784,1).
txt_page(287,784,1).
txt_page(959,784,1).
txt_page(539,784,1).
txt_page(45,784,1).
txt_page(6,784,1).
txt_page(773,784,1).
txt_page(647,785,1).
txt_page(115,785,1).
txt_page(345,785,1).
txt_page(788,785,1).
txt_page(527,785,1).
txt_page(390,785,1).
txt_page(210,785,1).
txt_page(841,785,1).
txt_page(348,785,1).
txt_page(143,785,1).
txt_page(396,785,1).
txt_page(616,785,1).
txt_page(944,785,1).
txt_page(643,785,1).
txt_page(373,785,1).
txt_page(43,785,1).
txt_page(432,785,1).
txt_page(347,785,1).
txt_page(193,785,1).
txt_page(655,785,1).
txt_page(937,786,1).
txt_page(440,786,1).
txt_page(618,786,1).
txt_page(750,786,1).
txt_page(668,786,1).
txt_page(447,786,1).
txt_page(415,786,1).
txt_page(721,786,1).
txt_page(219,786,1).
txt_page(759,786,1).
txt_page(365,786,1).
txt_page(7,786,1).
txt_page(597,786,1).
txt_page(35,786,1).
txt_page(158,786,1).
txt_page(241,786,1).
txt_page(246,786,1).
txt_page(689,786,1).
txt_page(507,786,1).
txt_page(290,786,1).
txt_page(762,787,1).
txt_page(739,787,1).
txt_page(380,787,1).
txt_page(191,787,1).
txt_page(542,787,1).
txt_page(292,787,1).
txt_page(621,787,1).
txt_page(100,787,1).
txt_page(6,787,1).
txt_page(152,787,1).
txt_page(665,787,1).
txt_page(421,787,1).
txt_page(825,787,1).
txt_page(185,787,1).
txt_page(930,787,1).
txt_page(538,787,1).
txt_page(568,787,1).
txt_page(415,787,1).
txt_page(89,787,1).
txt_page(746,787,1).
txt_page(616,788,1).
txt_page(757,788,1).
txt_page(255,788,1).
txt_page(170,788,1).
txt_page(129,788,1).
txt_page(81,788,1).
txt_page(325,788,1).
txt_page(633,788,1).
txt_page(963,788,1).
txt_page(689,788,1).
txt_page(920,788,1).
txt_page(887,788,1).
txt_page(147,788,1).
txt_page(942,788,1).
txt_page(486,788,1).
txt_page(576,788,2).
txt_page(565,788,1).
txt_page(934,788,1).
txt_page(450,788,1).
txt_page(49,789,1).
txt_page(457,789,1).
txt_page(299,789,1).
txt_page(815,789,1).
txt_page(94,789,1).
txt_page(282,789,1).
txt_page(314,789,1).
txt_page(843,789,1).
txt_page(124,789,1).
txt_page(281,789,1).
txt_page(783,789,1).
txt_page(937,789,1).
txt_page(230,789,1).
txt_page(543,789,1).
txt_page(246,789,1).
txt_page(650,789,1).
txt_page(875,789,1).
txt_page(531,789,1).
txt_page(232,789,1).
txt_page(882,789,1).
txt_page(790,790,1).
txt_page(143,790,1).
txt_page(330,790,1).
txt_page(559,790,1).
txt_page(850,790,1).
txt_page(594,790,1).
txt_page(944,790,1).
txt_page(129,790,1).
txt_page(676,790,1).
txt_page(941,790,1).
txt_page(612,790,1).
txt_page(859,790,1).
txt_page(373,790,1).
txt_page(866,790,1).
txt_page(696,790,1).
txt_page(477,790,1).
txt_page(715,790,1).
txt_page(936,790,1).
txt_page(354,790,1).
txt_page(172,790,1).
txt_page(833,791,1).
txt_page(703,791,1).
txt_page(404,791,1).
txt_page(723,791,1).
txt_page(205,791,1).
txt_page(405,791,1).
txt_page(263,791,1).
txt_page(386,791,1).
txt_page(497,791,1).
txt_page(142,791,1).
txt_page(269,791,1).
txt_page(19,791,1).
txt_page(913,791,1).
txt_page(447,791,1).
txt_page(10,791,1).
txt_page(784,791,1).
txt_page(530,791,1).
txt_page(687,791,1).
txt_page(604,791,1).
txt_page(433,791,1).
txt_page(813,792,1).
txt_page(896,792,1).
txt_page(776,792,1).
txt_page(580,792,1).
txt_page(388,792,1).
txt_page(742,792,1).
txt_page(455,792,1).
txt_page(891,792,1).
txt_page(24,792,1).
txt_page(960,792,1).
txt_page(407,792,1).
txt_page(217,792,1).
txt_page(738,792,1).
txt_page(213,792,1).
txt_page(34,792,1).
txt_page(263,792,2).
txt_page(493,792,1).
txt_page(96,792,1).
txt_page(523,792,1).
txt_page(920,793,1).
txt_page(320,793,1).
txt_page(336,793,1).
txt_page(331,793,1).
txt_page(572,793,1).
txt_page(717,793,1).
txt_page(873,793,1).
txt_page(743,793,1).
txt_page(855,793,1).
txt_page(845,793,1).
txt_page(938,793,1).
txt_page(578,793,1).
txt_page(60,793,1).
txt_page(799,793,1).
txt_page(219,793,1).
txt_page(818,793,1).
txt_page(628,793,1).
txt_page(861,793,1).
txt_page(530,793,1).
txt_page(644,793,1).
txt_page(406,794,1).
txt_page(372,794,1).
txt_page(403,794,1).
txt_page(494,794,1).
txt_page(749,794,1).
txt_page(793,794,1).
txt_page(302,794,1).
txt_page(417,794,1).
txt_page(888,794,1).
txt_page(722,794,1).
txt_page(882,794,1).
txt_page(21,794,1).
txt_page(849,794,1).
txt_page(466,794,1).
txt_page(97,794,1).
txt_page(128,794,1).
txt_page(816,794,1).
txt_page(344,794,1).
txt_page(923,794,1).
txt_page(524,794,1).
txt_page(633,795,1).
txt_page(18,795,1).
txt_page(713,795,1).
txt_page(921,795,1).
txt_page(871,795,1).
txt_page(940,795,1).
txt_page(209,795,1).
txt_page(384,795,1).
txt_page(420,795,1).
txt_page(770,795,1).
txt_page(484,795,1).
txt_page(621,795,1).
txt_page(502,795,1).
txt_page(769,795,1).
txt_page(71,795,1).
txt_page(884,795,1).
txt_page(46,795,1).
txt_page(19,795,1).
txt_page(566,795,1).
txt_page(28,795,1).
txt_page(630,796,1).
txt_page(9,796,1).
txt_page(277,796,1).
txt_page(195,796,1).
txt_page(163,796,1).
txt_page(464,796,1).
txt_page(494,796,1).
txt_page(822,796,1).
txt_page(122,796,1).
txt_page(327,796,1).
txt_page(119,796,1).
txt_page(62,796,1).
txt_page(647,796,1).
txt_page(108,796,1).
txt_page(428,796,1).
txt_page(12,796,2).
txt_page(134,796,1).
txt_page(318,796,1).
txt_page(526,796,1).
txt_page(900,797,1).
txt_page(721,797,1).
txt_page(57,797,1).
txt_page(504,797,1).
txt_page(679,797,1).
txt_page(230,797,1).
txt_page(857,797,1).
txt_page(810,797,1).
txt_page(551,797,1).
txt_page(234,797,1).
txt_page(872,797,1).
txt_page(386,797,1).
txt_page(625,797,1).
txt_page(175,797,1).
txt_page(200,797,1).
txt_page(272,797,1).
txt_page(220,797,1).
txt_page(149,797,1).
txt_page(227,797,1).
txt_page(733,797,1).
txt_page(736,798,1).
txt_page(796,798,1).
txt_page(277,798,1).
txt_page(508,798,1).
txt_page(746,798,1).
txt_page(64,798,1).
txt_page(908,798,1).
txt_page(294,798,1).
txt_page(504,798,1).
txt_page(200,798,1).
txt_page(295,798,1).
txt_page(47,798,1).
txt_page(299,798,1).
txt_page(267,798,1).
txt_page(342,798,1).
txt_page(406,798,1).
txt_page(659,798,1).
txt_page(505,798,1).
txt_page(380,798,1).
txt_page(489,798,1).
txt_page(490,799,1).
txt_page(419,799,1).
txt_page(924,799,1).
txt_page(216,799,1).
txt_page(367,799,1).
txt_page(475,799,1).
txt_page(533,799,1).
txt_page(906,799,1).
txt_page(540,799,1).
txt_page(104,799,1).
txt_page(268,799,1).
txt_page(295,799,1).
txt_page(89,799,1).
txt_page(892,799,1).
txt_page(230,799,1).
txt_page(869,799,1).
txt_page(772,799,1).
txt_page(783,799,1).
txt_page(824,799,1).
txt_page(713,799,1).
txt_page(533,800,1).
txt_page(339,800,1).
txt_page(716,800,1).
txt_page(564,800,1).
txt_page(384,800,1).
txt_page(402,800,1).
txt_page(449,800,1).
txt_page(877,800,1).
txt_page(254,800,1).
txt_page(557,800,1).
txt_page(246,800,1).
txt_page(41,800,1).
txt_page(448,800,1).
txt_page(652,800,1).
txt_page(289,800,1).
txt_page(102,800,1).
txt_page(692,800,1).
txt_page(656,800,1).
txt_page(549,800,1).
txt_page(655,800,1).
txt_page(118,801,1).
txt_page(414,801,1).
txt_page(437,801,1).
txt_page(774,801,1).
txt_page(685,801,1).
txt_page(532,801,1).
txt_page(295,801,1).
txt_page(283,801,1).
txt_page(815,801,1).
txt_page(457,801,1).
txt_page(807,801,1).
txt_page(944,801,1).
txt_page(472,801,1).
txt_page(289,801,1).
txt_page(961,801,1).
txt_page(126,801,1).
txt_page(332,801,1).
txt_page(495,801,1).
txt_page(696,801,1).
txt_page(5,801,1).
txt_page(519,802,1).
txt_page(889,802,1).
txt_page(590,802,1).
txt_page(430,802,1).
txt_page(899,802,1).
txt_page(484,802,1).
txt_page(49,802,1).
txt_page(22,802,1).
txt_page(914,802,1).
txt_page(458,802,1).
txt_page(821,802,1).
txt_page(407,802,1).
txt_page(924,802,1).
txt_page(745,802,1).
txt_page(148,802,1).
txt_page(831,802,1).
txt_page(226,802,1).
txt_page(660,802,1).
txt_page(73,802,1).
txt_page(414,802,1).
txt_page(461,803,1).
txt_page(721,803,1).
txt_page(674,803,1).
txt_page(256,803,1).
txt_page(416,803,1).
txt_page(158,803,1).
txt_page(807,803,1).
txt_page(835,803,1).
txt_page(786,803,1).
txt_page(675,803,1).
txt_page(538,803,1).
txt_page(223,803,1).
txt_page(878,803,1).
txt_page(810,803,1).
txt_page(771,803,1).
txt_page(150,803,1).
txt_page(831,803,1).
txt_page(939,803,1).
txt_page(751,803,1).
txt_page(920,803,1).
txt_page(279,804,1).
txt_page(867,804,1).
txt_page(268,804,1).
txt_page(774,804,1).
txt_page(533,804,1).
txt_page(312,804,1).
txt_page(892,804,1).
txt_page(144,804,1).
txt_page(206,804,1).
txt_page(542,804,1).
txt_page(722,804,1).
txt_page(614,804,1).
txt_page(724,804,1).
txt_page(551,804,1).
txt_page(952,804,1).
txt_page(241,804,1).
txt_page(346,804,1).
txt_page(316,804,1).
txt_page(321,804,1).
txt_page(960,804,1).
txt_page(190,805,1).
txt_page(467,805,1).
txt_page(405,805,1).
txt_page(433,805,1).
txt_page(720,805,1).
txt_page(910,805,1).
txt_page(716,805,1).
txt_page(882,805,1).
txt_page(965,805,1).
txt_page(44,805,1).
txt_page(142,805,1).
txt_page(342,805,1).
txt_page(832,805,1).
txt_page(459,805,1).
txt_page(658,805,1).
txt_page(511,805,1).
txt_page(145,805,2).
txt_page(295,805,1).
txt_page(881,805,1).
txt_page(549,806,1).
txt_page(326,806,1).
txt_page(554,806,1).
txt_page(259,806,1).
txt_page(645,806,1).
txt_page(755,806,1).
txt_page(785,806,1).
txt_page(266,806,1).
txt_page(488,806,1).
txt_page(872,806,1).
txt_page(463,806,1).
txt_page(92,806,1).
txt_page(331,806,1).
txt_page(940,806,1).
txt_page(208,806,1).
txt_page(964,806,1).
txt_page(86,806,1).
txt_page(224,806,1).
txt_page(878,806,1).
txt_page(145,806,1).
txt_page(790,807,1).
txt_page(413,807,1).
txt_page(163,807,1).
txt_page(569,807,1).
txt_page(42,807,1).
txt_page(592,807,1).
txt_page(223,807,1).
txt_page(398,807,1).
txt_page(49,807,1).
txt_page(403,807,1).
txt_page(519,807,1).
txt_page(745,807,1).
txt_page(138,807,1).
txt_page(638,807,1).
txt_page(911,807,1).
txt_page(930,807,1).
txt_page(86,807,1).
txt_page(723,807,1).
txt_page(117,807,1).
txt_page(90,807,1).
txt_page(423,808,1).
txt_page(246,808,1).
txt_page(52,808,1).
txt_page(390,808,1).
txt_page(294,808,1).
txt_page(210,808,1).
txt_page(801,808,1).
txt_page(777,808,1).
txt_page(139,808,1).
txt_page(307,808,1).
txt_page(106,808,1).
txt_page(88,808,1).
txt_page(392,808,1).
txt_page(363,808,1).
txt_page(583,808,1).
txt_page(71,808,1).
txt_page(420,808,1).
txt_page(455,808,1).
txt_page(410,808,1).
txt_page(285,808,1).
txt_page(513,809,1).
txt_page(955,809,1).
txt_page(478,809,1).
txt_page(320,809,1).
txt_page(806,809,1).
txt_page(680,809,1).
txt_page(531,809,1).
txt_page(219,809,1).
txt_page(489,809,1).
txt_page(514,809,1).
txt_page(840,809,1).
txt_page(470,809,1).
txt_page(330,809,1).
txt_page(98,809,1).
txt_page(216,809,1).
txt_page(557,809,1).
txt_page(592,809,1).
txt_page(189,809,1).
txt_page(156,809,1).
txt_page(563,809,1).
txt_page(641,810,1).
txt_page(365,810,1).
txt_page(668,810,1).
txt_page(959,810,1).
txt_page(740,810,1).
txt_page(302,810,1).
txt_page(521,810,1).
txt_page(474,810,1).
txt_page(91,810,1).
txt_page(378,810,1).
txt_page(898,810,1).
txt_page(749,810,1).
txt_page(82,810,1).
txt_page(876,810,1).
txt_page(460,810,1).
txt_page(52,810,1).
txt_page(760,810,1).
txt_page(128,810,1).
txt_page(196,810,1).
txt_page(882,810,1).
txt_page(201,811,1).
txt_page(31,811,1).
txt_page(610,811,1).
txt_page(279,811,1).
txt_page(16,811,1).
txt_page(640,811,1).
txt_page(365,811,1).
txt_page(425,811,1).
txt_page(544,811,1).
txt_page(886,811,1).
txt_page(253,811,1).
txt_page(29,811,1).
txt_page(109,811,1).
txt_page(267,811,1).
txt_page(539,811,1).
txt_page(54,811,1).
txt_page(57,811,1).
txt_page(398,811,1).
txt_page(155,811,1).
txt_page(704,811,1).
txt_page(363,812,1).
txt_page(812,812,1).
txt_page(8,812,1).
txt_page(824,812,1).
txt_page(278,812,1).
txt_page(247,812,1).
txt_page(261,812,1).
txt_page(897,812,1).
txt_page(188,812,1).
txt_page(502,812,1).
txt_page(702,812,1).
txt_page(959,812,1).
txt_page(350,812,1).
txt_page(625,812,1).
txt_page(139,812,1).
txt_page(324,812,1).
txt_page(74,812,1).
txt_page(368,812,1).
txt_page(690,812,1).
txt_page(192,812,1).
txt_page(21,813,1).
txt_page(146,813,1).
txt_page(931,813,1).
txt_page(266,813,1).
txt_page(875,813,1).
txt_page(176,813,1).
txt_page(764,813,1).
txt_page(903,813,1).
txt_page(456,813,1).
txt_page(618,813,1).
txt_page(81,813,1).
txt_page(740,813,1).
txt_page(147,813,1).
txt_page(477,813,1).
txt_page(684,813,1).
txt_page(218,813,1).
txt_page(638,813,1).
txt_page(873,813,1).
txt_page(24,813,1).
txt_page(394,813,1).
txt_page(164,814,1).
txt_page(667,814,1).
txt_page(463,814,1).
txt_page(211,814,1).
txt_page(454,814,1).
txt_page(899,814,1).
txt_page(260,814,1).
txt_page(777,814,1).
txt_page(827,814,1).
txt_page(698,814,1).
txt_page(47,814,1).
txt_page(605,814,1).
txt_page(255,814,1).
txt_page(302,814,1).
txt_page(582,814,1).
txt_page(155,814,1).
txt_page(226,814,1).
txt_page(562,814,1).
txt_page(872,814,1).
txt_page(456,814,1).
txt_page(784,815,1).
txt_page(105,815,1).
txt_page(454,815,1).
txt_page(399,815,1).
txt_page(503,815,1).
txt_page(240,815,1).
txt_page(694,815,1).
txt_page(629,815,1).
txt_page(439,815,1).
txt_page(581,815,1).
txt_page(406,815,1).
txt_page(161,815,1).
txt_page(184,815,1).
txt_page(499,815,1).
txt_page(965,815,1).
txt_page(110,815,1).
txt_page(707,815,1).
txt_page(735,815,1).
txt_page(842,815,1).
txt_page(461,815,1).
txt_page(503,816,1).
txt_page(506,816,1).
txt_page(365,816,1).
txt_page(696,816,1).
txt_page(902,816,1).
txt_page(273,816,1).
txt_page(535,816,1).
txt_page(746,816,1).
txt_page(944,816,1).
txt_page(457,816,1).
txt_page(67,816,1).
txt_page(243,816,1).
txt_page(181,816,1).
txt_page(295,816,2).
txt_page(206,816,1).
txt_page(921,816,1).
txt_page(45,816,1).
txt_page(232,816,1).
txt_page(15,816,1).
txt_page(191,817,1).
txt_page(968,817,1).
txt_page(619,817,1).
txt_page(98,817,1).
txt_page(621,817,1).
txt_page(638,817,1).
txt_page(232,817,1).
txt_page(786,817,1).
txt_page(766,817,1).
txt_page(490,817,1).
txt_page(135,817,1).
txt_page(404,817,1).
txt_page(871,817,1).
txt_page(409,817,1).
txt_page(635,817,1).
txt_page(321,817,1).
txt_page(273,817,1).
txt_page(287,817,1).
txt_page(172,817,1).
txt_page(194,817,1).
txt_page(648,818,1).
txt_page(754,818,1).
txt_page(620,818,1).
txt_page(129,818,1).
txt_page(700,818,1).
txt_page(818,818,1).
txt_page(366,818,1).
txt_page(6,818,1).
txt_page(397,818,1).
txt_page(21,818,1).
txt_page(326,818,1).
txt_page(636,818,1).
txt_page(487,818,1).
txt_page(13,818,1).
txt_page(628,818,1).
txt_page(142,818,1).
txt_page(901,818,1).
txt_page(713,818,1).
txt_page(920,818,1).
txt_page(239,818,1).
txt_page(684,819,1).
txt_page(803,819,1).
txt_page(521,819,1).
txt_page(494,819,1).
txt_page(818,819,1).
txt_page(335,819,1).
txt_page(268,819,1).
txt_page(322,819,1).
txt_page(670,819,1).
txt_page(383,819,2).
txt_page(594,819,1).
txt_page(402,819,1).
txt_page(869,819,1).
txt_page(929,819,1).
txt_page(122,819,1).
txt_page(428,819,1).
txt_page(755,819,1).
txt_page(396,819,1).
txt_page(628,819,1).
txt_page(443,820,1).
txt_page(366,820,1).
txt_page(639,820,1).
txt_page(71,820,1).
txt_page(258,820,2).
txt_page(351,820,1).
txt_page(669,820,1).
txt_page(429,820,1).
txt_page(612,820,1).
txt_page(182,820,1).
txt_page(409,820,1).
txt_page(711,820,1).
txt_page(889,820,2).
txt_page(584,820,1).
txt_page(275,820,1).
txt_page(56,820,1).
txt_page(166,820,1).
txt_page(395,820,1).
txt_page(489,821,1).
txt_page(759,821,1).
txt_page(865,821,1).
txt_page(474,821,1).
txt_page(3,821,1).
txt_page(494,821,1).
txt_page(113,821,1).
txt_page(692,821,1).
txt_page(82,821,1).
txt_page(748,821,1).
txt_page(842,821,1).
txt_page(386,821,1).
txt_page(81,821,1).
txt_page(63,821,1).
txt_page(300,821,1).
txt_page(684,821,1).
txt_page(464,821,1).
txt_page(741,821,1).
txt_page(519,821,1).
txt_page(785,821,1).
txt_page(537,822,1).
txt_page(106,822,1).
txt_page(617,822,1).
txt_page(608,822,1).
txt_page(799,822,1).
txt_page(957,822,1).
txt_page(199,822,1).
txt_page(115,822,1).
txt_page(508,822,1).
txt_page(768,822,1).
txt_page(614,822,1).
txt_page(155,822,1).
txt_page(174,822,1).
txt_page(612,822,1).
txt_page(907,822,1).
txt_page(159,822,1).
txt_page(916,822,1).
txt_page(512,822,1).
txt_page(15,822,1).
txt_page(498,822,1).
txt_page(110,823,1).
txt_page(785,823,1).
txt_page(944,823,1).
txt_page(444,823,1).
txt_page(83,823,1).
txt_page(74,823,1).
txt_page(567,823,1).
txt_page(837,823,1).
txt_page(893,823,1).
txt_page(327,823,1).
txt_page(881,823,1).
txt_page(697,823,1).
txt_page(25,823,1).
txt_page(197,823,1).
txt_page(840,823,1).
txt_page(10,823,1).
txt_page(934,823,1).
txt_page(68,823,1).
txt_page(312,823,1).
txt_page(628,823,1).
txt_page(528,824,1).
txt_page(909,824,1).
txt_page(733,824,1).
txt_page(78,824,1).
txt_page(502,824,1).
txt_page(800,824,1).
txt_page(595,824,1).
txt_page(588,824,1).
txt_page(602,824,1).
txt_page(257,824,1).
txt_page(699,824,1).
txt_page(908,824,1).
txt_page(716,824,1).
txt_page(450,824,1).
txt_page(336,824,1).
txt_page(172,824,1).
txt_page(594,824,1).
txt_page(633,824,1).
txt_page(341,824,1).
txt_page(931,824,1).
txt_page(960,825,1).
txt_page(571,825,1).
txt_page(694,825,1).
txt_page(410,825,1).
txt_page(891,825,1).
txt_page(260,825,1).
txt_page(573,825,1).
txt_page(788,825,1).
txt_page(804,825,1).
txt_page(463,825,1).
txt_page(660,825,1).
txt_page(35,825,1).
txt_page(517,825,1).
txt_page(768,825,1).
txt_page(549,825,1).
txt_page(792,825,1).
txt_page(415,825,1).
txt_page(245,825,1).
txt_page(875,825,1).
txt_page(431,825,1).
txt_page(428,826,1).
txt_page(751,826,1).
txt_page(748,826,1).
txt_page(607,826,1).
txt_page(34,826,1).
txt_page(12,826,1).
txt_page(836,826,1).
txt_page(401,826,1).
txt_page(258,826,1).
txt_page(111,826,1).
txt_page(288,826,1).
txt_page(891,826,1).
txt_page(378,826,1).
txt_page(626,826,1).
txt_page(433,826,1).
txt_page(581,826,1).
txt_page(867,826,1).
txt_page(787,826,1).
txt_page(844,826,1).
txt_page(945,826,1).
txt_page(641,827,1).
txt_page(286,827,1).
txt_page(812,827,1).
txt_page(32,827,1).
txt_page(109,827,1).
txt_page(371,827,1).
txt_page(792,827,1).
txt_page(90,827,1).
txt_page(528,827,1).
txt_page(958,827,1).
txt_page(762,827,1).
txt_page(510,827,1).
txt_page(349,827,1).
txt_page(387,827,1).
txt_page(341,827,1).
txt_page(968,827,1).
txt_page(663,827,1).
txt_page(795,827,1).
txt_page(165,827,1).
txt_page(547,827,1).
txt_page(876,828,1).
txt_page(460,828,1).
txt_page(308,828,1).
txt_page(590,828,1).
txt_page(446,828,1).
txt_page(317,828,1).
txt_page(275,828,1).
txt_page(544,828,1).
txt_page(606,828,1).
txt_page(102,828,2).
txt_page(649,828,1).
txt_page(451,828,1).
txt_page(375,828,1).
txt_page(766,828,1).
txt_page(810,828,1).
txt_page(106,828,1).
txt_page(347,828,1).
txt_page(438,828,1).
txt_page(130,828,1).
txt_page(712,829,1).
txt_page(809,829,1).
txt_page(886,829,1).
txt_page(501,829,1).
txt_page(207,829,1).
txt_page(1,829,1).
txt_page(795,829,1).
txt_page(370,829,1).
txt_page(861,829,1).
txt_page(470,829,1).
txt_page(368,829,1).
txt_page(716,829,1).
txt_page(453,829,1).
txt_page(781,829,1).
txt_page(278,829,1).
txt_page(693,829,1).
txt_page(849,829,1).
txt_page(57,829,1).
txt_page(677,829,1).
txt_page(165,829,1).
txt_page(650,830,1).
txt_page(174,830,1).
txt_page(896,830,1).
txt_page(891,830,1).
txt_page(584,830,1).
txt_page(672,830,1).
txt_page(50,830,2).
txt_page(36,830,1).
txt_page(908,830,1).
txt_page(289,830,1).
txt_page(93,830,1).
txt_page(604,830,1).
txt_page(932,830,1).
txt_page(318,830,1).
txt_page(30,830,1).
txt_page(374,830,1).
txt_page(373,830,1).
txt_page(119,830,1).
txt_page(236,830,1).
txt_page(879,831,1).
txt_page(205,831,1).
txt_page(426,831,1).
txt_page(593,831,1).
txt_page(713,831,1).
txt_page(886,831,1).
txt_page(300,831,1).
txt_page(797,831,1).
txt_page(601,831,1).
txt_page(490,831,1).
txt_page(678,831,2).
txt_page(863,831,1).
txt_page(669,831,1).
txt_page(718,831,1).
txt_page(581,831,1).
txt_page(177,831,1).
txt_page(461,831,1).
txt_page(495,831,1).
txt_page(522,831,1).
txt_page(752,832,1).
txt_page(424,832,1).
txt_page(495,832,1).
txt_page(908,832,1).
txt_page(612,832,1).
txt_page(191,832,1).
txt_page(694,832,1).
txt_page(769,832,1).
txt_page(307,832,1).
txt_page(402,832,1).
txt_page(521,832,1).
txt_page(28,832,1).
txt_page(720,832,1).
txt_page(938,832,1).
txt_page(660,832,1).
txt_page(408,832,1).
txt_page(547,832,1).
txt_page(802,832,1).
txt_page(936,832,1).
txt_page(218,832,1).
txt_page(697,833,1).
txt_page(614,833,1).
txt_page(721,833,1).
txt_page(643,833,1).
txt_page(52,833,1).
txt_page(77,833,1).
txt_page(193,833,1).
txt_page(85,833,1).
txt_page(905,833,1).
txt_page(79,833,1).
txt_page(674,833,1).
txt_page(941,833,1).
txt_page(828,833,1).
txt_page(234,833,1).
txt_page(280,833,1).
txt_page(45,833,1).
txt_page(420,833,2).
txt_page(725,833,1).
txt_page(346,833,1).
txt_page(323,834,1).
txt_page(312,834,1).
txt_page(309,834,1).
txt_page(81,834,1).
txt_page(466,834,1).
txt_page(851,834,1).
txt_page(37,834,1).
txt_page(614,834,1).
txt_page(57,834,1).
txt_page(585,834,1).
txt_page(46,834,1).
txt_page(928,834,1).
txt_page(251,834,1).
txt_page(470,834,1).
txt_page(295,834,1).
txt_page(201,834,1).
txt_page(621,834,1).
txt_page(624,834,1).
txt_page(109,834,1).
txt_page(36,834,1).
txt_page(228,835,1).
txt_page(498,835,1).
txt_page(114,835,1).
txt_page(615,835,1).
txt_page(607,835,1).
txt_page(21,835,1).
txt_page(55,835,1).
txt_page(575,835,1).
txt_page(654,835,1).
txt_page(903,835,1).
txt_page(796,835,1).
txt_page(306,835,1).
txt_page(678,835,1).
txt_page(651,835,1).
txt_page(181,835,1).
txt_page(736,835,1).
txt_page(359,835,1).
txt_page(430,835,1).
txt_page(749,835,2).
txt_page(211,836,1).
txt_page(927,836,1).
txt_page(259,836,1).
txt_page(537,836,1).
txt_page(153,836,1).
txt_page(481,836,1).
txt_page(474,836,1).
txt_page(558,836,1).
txt_page(200,836,1).
txt_page(968,836,1).
txt_page(8,836,1).
txt_page(245,836,1).
txt_page(468,836,1).
txt_page(552,836,1).
txt_page(611,836,1).
txt_page(752,836,1).
txt_page(396,836,1).
txt_page(798,836,1).
txt_page(566,836,1).
txt_page(306,836,1).
txt_page(44,837,1).
txt_page(197,837,1).
txt_page(522,837,1).
txt_page(115,837,1).
txt_page(676,837,1).
txt_page(678,837,1).
txt_page(592,837,1).
txt_page(168,837,1).
txt_page(871,837,1).
txt_page(567,837,1).
txt_page(881,837,1).
txt_page(869,837,1).
txt_page(516,837,1).
txt_page(194,837,1).
txt_page(466,837,1).
txt_page(16,837,1).
txt_page(213,837,1).
txt_page(409,837,1).
txt_page(907,837,1).
txt_page(38,837,1).
txt_page(555,838,1).
txt_page(151,838,1).
txt_page(866,838,1).
txt_page(168,838,1).
txt_page(318,838,1).
txt_page(965,838,1).
txt_page(270,838,1).
txt_page(337,838,1).
txt_page(163,838,1).
txt_page(961,838,1).
txt_page(885,838,1).
txt_page(887,838,1).
txt_page(329,838,1).
txt_page(364,838,1).
txt_page(70,838,1).
txt_page(302,838,1).
txt_page(287,838,1).
txt_page(875,838,1).
txt_page(170,838,1).
txt_page(148,838,1).
txt_page(159,839,1).
txt_page(518,839,1).
txt_page(743,839,1).
txt_page(243,839,1).
txt_page(38,839,1).
txt_page(964,839,1).
txt_page(846,839,1).
txt_page(505,839,1).
txt_page(921,839,1).
txt_page(131,839,1).
txt_page(469,839,1).
txt_page(539,839,1).
txt_page(943,839,1).
txt_page(65,839,1).
txt_page(195,839,1).
txt_page(500,839,1).
txt_page(658,839,1).
txt_page(925,839,1).
txt_page(133,839,1).
txt_page(289,839,1).
txt_page(148,840,1).
txt_page(918,840,1).
txt_page(207,840,1).
txt_page(387,840,1).
txt_page(699,840,1).
txt_page(537,840,1).
txt_page(767,840,1).
txt_page(434,840,1).
txt_page(456,840,1).
txt_page(391,840,1).
txt_page(489,840,1).
txt_page(807,840,1).
txt_page(653,840,1).
txt_page(499,840,1).
txt_page(582,840,1).
txt_page(543,840,1).
txt_page(317,840,1).
txt_page(421,840,1).
txt_page(631,840,1).
txt_page(330,840,1).
txt_page(627,841,1).
txt_page(779,841,1).
txt_page(161,841,1).
txt_page(816,841,1).
txt_page(408,841,1).
txt_page(445,841,1).
txt_page(99,841,1).
txt_page(921,841,1).
txt_page(513,841,1).
txt_page(96,841,1).
txt_page(392,841,1).
txt_page(119,841,1).
txt_page(423,841,1).
txt_page(221,841,1).
txt_page(101,841,1).
txt_page(618,841,1).
txt_page(389,841,1).
txt_page(641,841,1).
txt_page(761,841,1).
txt_page(413,841,1).
txt_page(121,842,1).
txt_page(545,842,1).
txt_page(279,842,1).
txt_page(425,842,1).
txt_page(446,842,1).
txt_page(264,842,1).
txt_page(330,842,1).
txt_page(496,842,1).
txt_page(14,842,1).
txt_page(896,842,1).
txt_page(692,842,1).
txt_page(470,842,1).
txt_page(246,842,1).
txt_page(929,842,1).
txt_page(959,842,1).
txt_page(925,842,1).
txt_page(543,842,1).
txt_page(606,842,1).
txt_page(563,842,1).
txt_page(485,842,1).
txt_page(22,843,1).
txt_page(776,843,1).
txt_page(511,843,1).
txt_page(55,843,1).
txt_page(937,843,1).
txt_page(255,843,1).
txt_page(718,843,1).
txt_page(229,843,1).
txt_page(841,843,1).
txt_page(87,843,2).
txt_page(403,843,1).
txt_page(24,843,1).
txt_page(941,843,1).
txt_page(56,843,1).
txt_page(699,843,1).
txt_page(163,843,1).
txt_page(410,843,1).
txt_page(44,843,1).
txt_page(835,843,1).
txt_page(158,844,1).
txt_page(387,844,1).
txt_page(787,844,1).
txt_page(555,844,1).
txt_page(471,844,1).
txt_page(888,844,1).
txt_page(27,844,1).
txt_page(634,844,1).
txt_page(94,844,1).
txt_page(650,844,1).
txt_page(585,844,1).
txt_page(285,844,1).
txt_page(399,844,1).
txt_page(868,844,1).
txt_page(242,844,1).
txt_page(375,844,1).
txt_page(338,844,1).
txt_page(924,844,1).
txt_page(742,844,1).
txt_page(703,844,1).
txt_page(898,845,1).
txt_page(874,845,1).
txt_page(355,845,1).
txt_page(295,845,1).
txt_page(934,845,1).
txt_page(93,845,1).
txt_page(841,845,1).
txt_page(561,845,1).
txt_page(868,845,1).
txt_page(611,845,1).
txt_page(824,845,1).
txt_page(711,845,1).
txt_page(165,845,1).
txt_page(322,845,1).
txt_page(682,845,1).
txt_page(409,845,1).
txt_page(710,845,1).
txt_page(106,845,1).
txt_page(288,845,1).
txt_page(66,845,1).
txt_page(298,846,1).
txt_page(662,846,1).
txt_page(880,846,1).
txt_page(158,846,1).
txt_page(403,846,1).
txt_page(400,846,1).
txt_page(89,846,1).
txt_page(654,846,1).
txt_page(3,846,1).
txt_page(720,846,1).
txt_page(650,846,1).
txt_page(150,846,1).
txt_page(232,846,1).
txt_page(482,846,1).
txt_page(607,846,1).
txt_page(659,846,1).
txt_page(245,846,1).
txt_page(107,846,1).
txt_page(801,846,1).
txt_page(734,846,1).
txt_page(671,847,1).
txt_page(32,847,1).
txt_page(581,847,1).
txt_page(540,847,1).
txt_page(585,847,1).
txt_page(84,847,1).
txt_page(762,847,1).
txt_page(833,847,1).
txt_page(215,847,1).
txt_page(744,847,1).
txt_page(927,847,1).
txt_page(113,847,1).
txt_page(132,847,1).
txt_page(352,847,1).
txt_page(873,847,1).
txt_page(575,847,1).
txt_page(451,847,1).
txt_page(796,847,1).
txt_page(731,847,1).
txt_page(530,847,1).
txt_page(302,848,1).
txt_page(335,848,1).
txt_page(325,848,1).
txt_page(239,848,1).
txt_page(717,848,1).
txt_page(447,848,1).
txt_page(492,848,1).
txt_page(824,848,1).
txt_page(38,848,1).
txt_page(419,848,1).
txt_page(952,848,1).
txt_page(165,848,1).
txt_page(176,848,1).
txt_page(408,848,1).
txt_page(625,848,1).
txt_page(171,848,1).
txt_page(481,848,1).
txt_page(62,848,1).
txt_page(706,848,1).
txt_page(254,848,1).
txt_page(918,849,1).
txt_page(491,849,1).
txt_page(282,849,1).
txt_page(242,849,1).
txt_page(359,849,1).
txt_page(563,849,1).
txt_page(172,849,1).
txt_page(498,849,1).
txt_page(338,849,1).
txt_page(270,849,1).
txt_page(920,849,1).
txt_page(391,849,1).
txt_page(132,849,1).
txt_page(947,849,2).
txt_page(934,849,1).
txt_page(541,849,1).
txt_page(852,849,1).
txt_page(579,849,1).
txt_page(386,849,1).
txt_page(930,850,1).
txt_page(429,850,1).
txt_page(87,850,1).
txt_page(285,850,1).
txt_page(820,850,1).
txt_page(663,850,1).
txt_page(320,850,1).
txt_page(591,850,1).
txt_page(585,850,1).
txt_page(174,850,1).
txt_page(565,850,1).
txt_page(296,850,1).
txt_page(859,850,1).
txt_page(262,850,1).
txt_page(964,850,1).
txt_page(916,850,1).
txt_page(388,850,1).
txt_page(373,850,1).
txt_page(137,850,1).
txt_page(431,850,1).
txt_page(725,851,1).
txt_page(365,851,1).
txt_page(148,851,1).
txt_page(229,851,1).
txt_page(72,851,1).
txt_page(369,851,1).
txt_page(942,851,1).
txt_page(548,851,1).
txt_page(635,851,1).
txt_page(912,851,1).
txt_page(301,851,1).
txt_page(741,851,1).
txt_page(620,851,1).
txt_page(868,851,1).
txt_page(705,851,1).
txt_page(171,851,1).
txt_page(437,851,1).
txt_page(961,851,1).
txt_page(623,851,1).
txt_page(957,851,1).
txt_page(108,852,1).
txt_page(145,852,1).
txt_page(862,852,1).
txt_page(443,852,1).
txt_page(934,852,1).
txt_page(160,852,1).
txt_page(265,852,1).
txt_page(495,852,1).
txt_page(55,852,1).
txt_page(912,852,1).
txt_page(247,852,1).
txt_page(465,852,1).
txt_page(168,852,1).
txt_page(876,852,1).
txt_page(758,852,1).
txt_page(796,852,1).
txt_page(329,852,1).
txt_page(195,852,1).
txt_page(370,852,1).
txt_page(138,852,1).
txt_page(127,853,1).
txt_page(877,853,1).
txt_page(793,853,1).
txt_page(369,853,1).
txt_page(481,853,1).
txt_page(510,853,1).
txt_page(280,853,1).
txt_page(521,853,1).
txt_page(103,853,1).
txt_page(233,853,1).
txt_page(531,853,1).
txt_page(125,853,1).
txt_page(291,853,1).
txt_page(93,853,1).
txt_page(257,853,1).
txt_page(292,853,1).
txt_page(652,853,2).
txt_page(214,853,1).
txt_page(635,853,1).
txt_page(901,854,1).
txt_page(375,854,1).
txt_page(613,854,1).
txt_page(763,854,1).
txt_page(95,854,1).
txt_page(302,854,1).
txt_page(798,854,1).
txt_page(814,854,1).
txt_page(914,854,1).
txt_page(227,854,1).
txt_page(592,854,1).
txt_page(30,854,2).
txt_page(659,854,1).
txt_page(514,854,1).
txt_page(795,854,1).
txt_page(385,854,1).
txt_page(331,854,1).
txt_page(314,854,1).
txt_page(455,854,1).
txt_page(467,855,1).
txt_page(246,855,1).
txt_page(479,855,1).
txt_page(275,855,1).
txt_page(657,855,1).
txt_page(934,855,1).
txt_page(956,855,1).
txt_page(796,855,1).
txt_page(223,855,1).
txt_page(716,855,1).
txt_page(23,855,1).
txt_page(349,855,1).
txt_page(375,855,1).
txt_page(837,855,1).
txt_page(759,855,1).
txt_page(12,855,1).
txt_page(611,855,1).
txt_page(878,855,1).
txt_page(772,855,1).
txt_page(297,855,1).
txt_page(943,856,1).
txt_page(958,856,1).
txt_page(33,856,1).
txt_page(679,856,1).
txt_page(183,856,1).
txt_page(526,856,1).
txt_page(596,856,1).
txt_page(485,856,1).
txt_page(624,856,1).
txt_page(107,856,1).
txt_page(728,856,1).
txt_page(644,856,1).
txt_page(222,856,2).
txt_page(578,856,1).
txt_page(536,856,1).
txt_page(53,856,1).
txt_page(261,856,1).
txt_page(57,856,1).
txt_page(331,856,1).
txt_page(554,857,1).
txt_page(235,857,1).
txt_page(431,857,1).
txt_page(507,857,1).
txt_page(620,857,1).
txt_page(837,857,1).
txt_page(141,857,1).
txt_page(353,857,1).
txt_page(712,857,1).
txt_page(738,857,1).
txt_page(327,857,1).
txt_page(154,857,1).
txt_page(602,857,1).
txt_page(170,857,1).
txt_page(7,857,1).
txt_page(380,857,1).
txt_page(562,857,1).
txt_page(472,857,1).
txt_page(88,857,1).
txt_page(392,857,1).
txt_page(760,858,1).
txt_page(520,858,1).
txt_page(426,858,1).
txt_page(614,858,1).
txt_page(154,858,1).
txt_page(921,858,1).
txt_page(963,858,1).
txt_page(429,858,1).
txt_page(80,858,1).
txt_page(620,858,1).
txt_page(256,858,1).
txt_page(585,858,1).
txt_page(674,858,1).
txt_page(327,858,1).
txt_page(118,858,1).
txt_page(871,858,1).
txt_page(901,858,1).
txt_page(359,858,1).
txt_page(314,858,1).
txt_page(433,858,1).
txt_page(614,859,1).
txt_page(333,859,1).
txt_page(927,859,1).
txt_page(900,859,1).
txt_page(826,859,1).
txt_page(434,859,1).
txt_page(576,859,1).
txt_page(328,859,1).
txt_page(673,859,1).
txt_page(674,859,1).
txt_page(916,859,1).
txt_page(341,859,1).
txt_page(736,859,1).
txt_page(618,859,1).
txt_page(619,859,1).
txt_page(539,859,1).
txt_page(511,859,1).
txt_page(237,859,1).
txt_page(859,859,1).
txt_page(95,859,1).
txt_page(727,860,1).
txt_page(677,860,1).
txt_page(696,860,1).
txt_page(496,860,1).
txt_page(812,860,1).
txt_page(183,860,1).
txt_page(151,860,1).
txt_page(789,860,1).
txt_page(284,860,1).
txt_page(126,860,1).
txt_page(735,860,1).
txt_page(746,860,2).
txt_page(829,860,1).
txt_page(404,860,1).
txt_page(554,860,1).
txt_page(814,860,1).
txt_page(74,860,1).
txt_page(657,860,1).
txt_page(739,860,1).
txt_page(422,861,1).
txt_page(623,861,1).
txt_page(805,861,1).
txt_page(284,861,1).
txt_page(856,861,1).
txt_page(771,861,1).
txt_page(758,861,1).
txt_page(11,861,1).
txt_page(183,861,1).
txt_page(186,861,1).
txt_page(719,861,1).
txt_page(662,861,1).
txt_page(332,861,1).
txt_page(715,861,1).
txt_page(382,861,1).
txt_page(792,861,1).
txt_page(187,861,1).
txt_page(554,861,1).
txt_page(496,861,1).
txt_page(167,861,1).
txt_page(850,862,1).
txt_page(101,862,1).
txt_page(727,862,1).
txt_page(113,862,1).
txt_page(284,862,1).
txt_page(295,862,1).
txt_page(438,862,1).
txt_page(577,862,1).
txt_page(461,862,1).
txt_page(908,862,1).
txt_page(39,862,1).
txt_page(313,862,1).
txt_page(26,862,1).
txt_page(320,862,1).
txt_page(760,862,1).
txt_page(147,862,1).
txt_page(16,862,1).
txt_page(153,862,2).
txt_page(327,862,1).
txt_page(820,863,1).
txt_page(581,863,1).
txt_page(178,863,1).
txt_page(424,863,1).
txt_page(328,863,1).
txt_page(710,863,1).
txt_page(141,863,1).
txt_page(636,863,1).
txt_page(264,863,1).
txt_page(828,863,1).
txt_page(687,863,1).
txt_page(239,863,1).
txt_page(572,863,1).
txt_page(368,863,1).
txt_page(278,863,1).
txt_page(559,863,1).
txt_page(730,863,1).
txt_page(273,863,1).
txt_page(896,863,1).
txt_page(129,863,1).
txt_page(189,864,1).
txt_page(352,864,1).
txt_page(550,864,1).
txt_page(432,864,1).
txt_page(439,864,1).
txt_page(137,864,1).
txt_page(503,864,1).
txt_page(664,864,1).
txt_page(391,864,1).
txt_page(838,864,1).
txt_page(898,864,1).
txt_page(963,864,1).
txt_page(68,864,1).
txt_page(212,864,1).
txt_page(50,864,1).
txt_page(844,864,1).
txt_page(761,864,1).
txt_page(151,864,1).
txt_page(696,864,1).
txt_page(794,864,1).
txt_page(43,865,1).
txt_page(840,865,1).
txt_page(313,865,1).
txt_page(352,865,1).
txt_page(263,865,1).
txt_page(869,865,1).
txt_page(203,865,1).
txt_page(187,865,1).
txt_page(723,865,1).
txt_page(411,865,1).
txt_page(398,865,1).
txt_page(496,865,1).
txt_page(53,865,1).
txt_page(679,865,1).
txt_page(622,865,1).
txt_page(880,865,1).
txt_page(812,865,1).
txt_page(729,865,2).
txt_page(874,865,1).
txt_page(726,866,1).
txt_page(593,866,1).
txt_page(21,866,1).
txt_page(421,866,1).
txt_page(502,866,1).
txt_page(214,866,1).
txt_page(137,866,1).
txt_page(624,866,1).
txt_page(603,866,1).
txt_page(110,866,1).
txt_page(239,866,1).
txt_page(220,866,1).
txt_page(313,866,1).
txt_page(135,866,1).
txt_page(749,866,1).
txt_page(690,866,1).
txt_page(818,866,1).
txt_page(412,866,1).
txt_page(60,866,1).
txt_page(205,866,1).
txt_page(125,867,1).
txt_page(312,867,1).
txt_page(741,867,1).
txt_page(137,867,1).
txt_page(733,867,1).
txt_page(547,867,1).
txt_page(582,867,1).
txt_page(280,867,1).
txt_page(749,867,1).
txt_page(172,867,1).
txt_page(95,867,1).
txt_page(475,867,1).
txt_page(902,867,1).
txt_page(51,867,1).
txt_page(385,867,1).
txt_page(568,867,1).
txt_page(297,867,1).
txt_page(417,867,1).
txt_page(111,867,1).
txt_page(567,867,1).
txt_page(445,868,1).
txt_page(289,868,1).
txt_page(716,868,1).
txt_page(138,868,1).
txt_page(917,868,1).
txt_page(18,868,1).
txt_page(903,868,1).
txt_page(576,868,1).
txt_page(40,868,1).
txt_page(817,868,1).
txt_page(438,868,1).
txt_page(853,868,1).
txt_page(897,868,1).
txt_page(563,868,1).
txt_page(910,868,1).
txt_page(301,868,1).
txt_page(511,868,1).
txt_page(823,868,1).
txt_page(559,868,1).
txt_page(375,868,1).
txt_page(582,869,1).
txt_page(638,869,1).
txt_page(963,869,1).
txt_page(774,869,1).
txt_page(904,869,1).
txt_page(507,869,1).
txt_page(146,869,1).
txt_page(418,869,1).
txt_page(19,869,1).
txt_page(777,869,1).
txt_page(87,869,1).
txt_page(369,869,1).
txt_page(304,869,1).
txt_page(252,869,1).
txt_page(52,869,1).
txt_page(584,869,1).
txt_page(208,869,1).
txt_page(172,869,1).
txt_page(16,869,1).
txt_page(242,869,1).
txt_page(54,870,1).
txt_page(177,870,1).
txt_page(892,870,1).
txt_page(893,870,1).
txt_page(75,870,1).
txt_page(825,870,1).
txt_page(941,870,1).
txt_page(859,870,1).
txt_page(324,870,1).
txt_page(368,870,1).
txt_page(230,870,1).
txt_page(765,870,1).
txt_page(612,870,1).
txt_page(870,870,1).
txt_page(190,870,1).
txt_page(823,870,1).
txt_page(444,870,1).
txt_page(597,870,1).
txt_page(3,870,1).
txt_page(179,870,1).
txt_page(843,871,1).
txt_page(642,871,1).
txt_page(396,871,1).
txt_page(413,871,1).
txt_page(782,871,1).
txt_page(474,871,1).
txt_page(829,871,1).
txt_page(552,871,1).
txt_page(88,871,1).
txt_page(957,871,1).
txt_page(279,871,1).
txt_page(120,871,1).
txt_page(886,871,1).
txt_page(966,871,1).
txt_page(99,871,1).
txt_page(594,871,1).
txt_page(837,871,1).
txt_page(11,871,1).
txt_page(208,871,1).
txt_page(906,871,1).
txt_page(941,872,1).
txt_page(308,872,1).
txt_page(855,872,1).
txt_page(390,872,1).
txt_page(318,872,1).
txt_page(138,872,1).
txt_page(299,872,1).
txt_page(854,872,1).
txt_page(707,872,1).
txt_page(76,872,1).
txt_page(619,872,1).
txt_page(539,872,1).
txt_page(484,872,1).
txt_page(194,872,1).
txt_page(72,872,1).
txt_page(322,872,1).
txt_page(307,872,1).
txt_page(791,872,1).
txt_page(78,872,1).
txt_page(296,872,1).
txt_page(669,873,1).
txt_page(270,873,1).
txt_page(52,873,1).
txt_page(251,873,1).
txt_page(942,873,1).
txt_page(414,873,1).
txt_page(120,873,1).
txt_page(306,873,1).
txt_page(892,873,1).
txt_page(773,873,1).
txt_page(542,873,1).
txt_page(624,873,1).
txt_page(73,873,1).
txt_page(267,873,1).
txt_page(625,873,1).
txt_page(831,873,1).
txt_page(405,873,1).
txt_page(379,873,1).
txt_page(852,873,1).
txt_page(901,873,1).
txt_page(126,874,1).
txt_page(386,874,1).
txt_page(408,874,1).
txt_page(617,874,1).
txt_page(894,874,1).
txt_page(687,874,1).
txt_page(354,874,1).
txt_page(570,874,1).
txt_page(566,874,1).
txt_page(618,874,1).
txt_page(345,874,1).
txt_page(454,874,1).
txt_page(510,874,1).
txt_page(942,874,1).
txt_page(455,874,1).
txt_page(565,874,1).
txt_page(392,874,1).
txt_page(791,874,1).
txt_page(904,874,1).
txt_page(801,874,1).
txt_page(186,875,1).
txt_page(115,875,1).
txt_page(83,875,1).
txt_page(735,875,1).
txt_page(392,875,1).
txt_page(113,875,1).
txt_page(823,875,1).
txt_page(163,875,1).
txt_page(941,875,1).
txt_page(652,875,1).
txt_page(515,875,1).
txt_page(918,875,1).
txt_page(161,875,1).
txt_page(580,875,1).
txt_page(925,875,1).
txt_page(827,875,1).
txt_page(119,875,1).
txt_page(86,875,1).
txt_page(130,875,1).
txt_page(871,875,1).
txt_page(271,876,1).
txt_page(426,876,1).
txt_page(915,876,1).
txt_page(427,876,1).
txt_page(19,876,1).
txt_page(98,876,1).
txt_page(618,876,1).
txt_page(945,876,1).
txt_page(265,876,1).
txt_page(761,876,1).
txt_page(717,876,1).
txt_page(86,876,1).
txt_page(753,876,1).
txt_page(829,876,1).
txt_page(601,876,1).
txt_page(898,876,1).
txt_page(311,876,1).
txt_page(334,876,1).
txt_page(128,876,1).
txt_page(809,876,1).
txt_page(954,877,1).
txt_page(899,877,1).
txt_page(639,877,1).
txt_page(921,877,1).
txt_page(261,877,1).
txt_page(682,877,1).
txt_page(29,877,1).
txt_page(525,877,1).
txt_page(754,877,1).
txt_page(93,877,1).
txt_page(263,877,1).
txt_page(645,877,1).
txt_page(25,877,1).
txt_page(151,877,1).
txt_page(631,877,1).
txt_page(353,877,1).
txt_page(768,877,1).
txt_page(679,877,1).
txt_page(497,877,1).
txt_page(610,877,1).
txt_page(661,878,1).
txt_page(229,878,1).
txt_page(68,878,1).
txt_page(715,878,1).
txt_page(178,878,1).
txt_page(676,878,1).
txt_page(713,878,1).
txt_page(445,878,1).
txt_page(949,878,1).
txt_page(529,878,1).
txt_page(755,878,1).
txt_page(921,878,1).
txt_page(966,878,1).
txt_page(34,878,1).
txt_page(159,878,1).
txt_page(685,878,1).
txt_page(893,878,1).
txt_page(397,878,1).
txt_page(393,878,1).
txt_page(587,878,1).
txt_page(634,879,1).
txt_page(824,879,1).
txt_page(821,879,1).
txt_page(133,879,1).
txt_page(368,879,1).
txt_page(807,879,1).
txt_page(223,879,1).
txt_page(400,879,1).
txt_page(221,879,1).
txt_page(81,879,1).
txt_page(99,879,1).
txt_page(182,879,1).
txt_page(496,879,1).
txt_page(775,879,1).
txt_page(898,879,1).
txt_page(7,879,1).
txt_page(445,879,1).
txt_page(432,879,1).
txt_page(650,879,1).
txt_page(958,879,1).
txt_page(698,880,1).
txt_page(391,880,1).
txt_page(793,880,1).
txt_page(734,880,1).
txt_page(203,880,1).
txt_page(443,880,1).
txt_page(722,880,1).
txt_page(448,880,1).
txt_page(611,880,1).
txt_page(426,880,1).
txt_page(956,880,1).
txt_page(761,880,1).
txt_page(114,880,1).
txt_page(100,880,1).
txt_page(423,880,1).
txt_page(696,880,1).
txt_page(260,880,1).
txt_page(113,880,1).
txt_page(749,880,1).
txt_page(750,880,1).
txt_page(763,881,1).
txt_page(960,881,1).
txt_page(170,881,1).
txt_page(229,881,1).
txt_page(228,881,1).
txt_page(558,881,1).
txt_page(457,881,2).
txt_page(708,881,1).
txt_page(339,881,1).
txt_page(648,881,1).
txt_page(534,881,1).
txt_page(172,881,1).
txt_page(753,881,1).
txt_page(332,881,1).
txt_page(183,881,1).
txt_page(623,881,1).
txt_page(27,881,1).
txt_page(240,881,1).
txt_page(184,881,1).
txt_page(205,882,1).
txt_page(92,882,1).
txt_page(811,882,1).
txt_page(9,882,1).
txt_page(300,882,1).
txt_page(714,882,1).
txt_page(106,882,1).
txt_page(911,882,1).
txt_page(367,882,1).
txt_page(44,882,1).
txt_page(578,882,1).
txt_page(844,882,1).
txt_page(594,882,1).
txt_page(122,882,1).
txt_page(49,882,1).
txt_page(201,882,1).
txt_page(311,882,1).
txt_page(244,882,1).
txt_page(425,882,1).
txt_page(141,882,1).
txt_page(612,883,1).
txt_page(900,883,1).
txt_page(248,883,1).
txt_page(563,883,1).
txt_page(337,883,1).
txt_page(549,883,1).
txt_page(848,883,1).
txt_page(774,883,1).
txt_page(340,883,1).
txt_page(531,883,1).
txt_page(858,883,1).
txt_page(893,883,1).
txt_page(479,883,1).
txt_page(553,883,2).
txt_page(342,883,1).
txt_page(443,883,1).
txt_page(46,883,1).
txt_page(586,883,1).
txt_page(570,883,1).
txt_page(16,884,1).
txt_page(654,884,1).
txt_page(102,884,1).
txt_page(276,884,1).
txt_page(860,884,1).
txt_page(542,884,1).
txt_page(714,884,1).
txt_page(472,884,1).
txt_page(682,884,1).
txt_page(281,884,1).
txt_page(577,884,1).
txt_page(409,884,1).
txt_page(119,884,1).
txt_page(786,884,1).
txt_page(901,884,1).
txt_page(249,884,1).
txt_page(758,884,1).
txt_page(321,884,1).
txt_page(130,884,1).
txt_page(131,884,1).
txt_page(343,885,1).
txt_page(331,885,1).
txt_page(487,885,1).
txt_page(387,885,1).
txt_page(891,885,1).
txt_page(8,885,1).
txt_page(583,885,1).
txt_page(91,885,1).
txt_page(809,885,1).
txt_page(249,885,1).
txt_page(169,885,1).
txt_page(471,885,1).
txt_page(877,885,1).
txt_page(6,885,1).
txt_page(146,885,1).
txt_page(890,885,1).
txt_page(463,885,1).
txt_page(506,885,1).
txt_page(295,885,1).
txt_page(86,885,1).
txt_page(834,886,1).
txt_page(709,886,1).
txt_page(395,886,1).
txt_page(357,886,1).
txt_page(206,886,1).
txt_page(64,886,1).
txt_page(766,886,1).
txt_page(880,886,1).
txt_page(459,886,1).
txt_page(731,886,1).
txt_page(149,886,1).
txt_page(407,886,1).
txt_page(346,886,1).
txt_page(380,886,1).
txt_page(426,886,1).
txt_page(282,886,1).
txt_page(539,886,1).
txt_page(364,886,1).
txt_page(734,886,1).
txt_page(353,886,1).
txt_page(836,887,1).
txt_page(772,887,1).
txt_page(591,887,1).
txt_page(442,887,1).
txt_page(448,887,1).
txt_page(711,887,1).
txt_page(365,887,1).
txt_page(850,887,1).
txt_page(315,887,1).
txt_page(69,887,1).
txt_page(587,887,1).
txt_page(508,887,1).
txt_page(916,887,1).
txt_page(62,887,1).
txt_page(727,887,1).
txt_page(834,887,1).
txt_page(67,887,1).
txt_page(530,887,1).
txt_page(563,887,1).
txt_page(613,887,1).
txt_page(115,888,1).
txt_page(777,888,1).
txt_page(459,888,1).
txt_page(727,888,1).
txt_page(450,888,1).
txt_page(646,888,1).
txt_page(681,888,1).
txt_page(222,888,1).
txt_page(965,888,1).
txt_page(28,888,1).
txt_page(735,888,1).
txt_page(870,888,1).
txt_page(830,888,1).
txt_page(807,888,1).
txt_page(909,888,1).
txt_page(888,888,1).
txt_page(9,888,1).
txt_page(307,888,1).
txt_page(53,888,1).
txt_page(548,888,1).
txt_page(129,889,1).
txt_page(134,889,1).
txt_page(87,889,1).
txt_page(95,889,1).
txt_page(506,889,1).
txt_page(452,889,1).
txt_page(348,889,1).
txt_page(69,889,1).
txt_page(26,889,1).
txt_page(710,889,1).
txt_page(644,889,1).
txt_page(945,889,1).
txt_page(91,889,1).
txt_page(708,889,1).
txt_page(681,889,1).
txt_page(597,889,1).
txt_page(662,889,1).
txt_page(24,889,1).
txt_page(729,889,1).
txt_page(414,889,1).
txt_page(654,890,1).
txt_page(88,890,1).
txt_page(54,890,1).
txt_page(369,890,1).
txt_page(888,890,1).
txt_page(536,890,1).
txt_page(287,890,1).
txt_page(641,890,1).
txt_page(443,890,1).
txt_page(827,890,1).
txt_page(872,890,1).
txt_page(499,890,1).
txt_page(777,890,1).
txt_page(437,890,1).
txt_page(95,890,1).
txt_page(524,890,1).
txt_page(356,890,1).
txt_page(347,890,1).
txt_page(763,890,1).
txt_page(774,890,1).
txt_page(954,891,1).
txt_page(564,891,1).
txt_page(412,891,1).
txt_page(920,891,1).
txt_page(513,891,1).
txt_page(814,891,1).
txt_page(443,891,1).
txt_page(820,891,1).
txt_page(132,891,1).
txt_page(68,891,1).
txt_page(297,891,1).
txt_page(590,891,1).
txt_page(85,891,1).
txt_page(559,891,1).
txt_page(518,891,1).
txt_page(864,891,1).
txt_page(453,891,1).
txt_page(679,891,1).
txt_page(463,891,1).
txt_page(274,891,1).
txt_page(914,892,1).
txt_page(280,892,1).
txt_page(379,892,1).
txt_page(418,892,1).
txt_page(141,892,1).
txt_page(817,892,1).
txt_page(468,892,1).
txt_page(422,892,1).
txt_page(843,892,1).
txt_page(367,892,1).
txt_page(469,892,1).
txt_page(154,892,1).
txt_page(688,892,1).
txt_page(852,892,1).
txt_page(570,892,1).
txt_page(961,892,1).
txt_page(354,892,1).
txt_page(23,892,1).
txt_page(347,892,1).
txt_page(926,892,1).
txt_page(335,893,1).
txt_page(813,893,1).
txt_page(38,893,1).
txt_page(11,893,1).
txt_page(391,893,1).
txt_page(274,893,1).
txt_page(568,893,1).
txt_page(919,893,1).
txt_page(905,893,1).
txt_page(843,893,1).
txt_page(646,893,1).
txt_page(740,893,1).
txt_page(896,893,1).
txt_page(407,893,1).
txt_page(59,893,1).
txt_page(934,893,1).
txt_page(360,893,1).
txt_page(421,893,1).
txt_page(19,893,1).
txt_page(895,893,1).
txt_page(483,894,1).
txt_page(297,894,1).
txt_page(458,894,1).
txt_page(96,894,1).
txt_page(502,894,1).
txt_page(577,894,1).
txt_page(948,894,1).
txt_page(304,894,1).
txt_page(106,894,1).
txt_page(362,894,1).
txt_page(251,894,1).
txt_page(455,894,1).
txt_page(523,894,1).
txt_page(587,894,1).
txt_page(323,894,1).
txt_page(892,894,1).
txt_page(413,894,1).
txt_page(838,894,1).
txt_page(32,894,1).
txt_page(627,894,1).
txt_page(68,895,1).
txt_page(51,895,1).
txt_page(707,895,1).
txt_page(217,895,1).
txt_page(230,895,1).
txt_page(276,895,1).
txt_page(678,895,1).
txt_page(806,895,2).
txt_page(187,895,1).
txt_page(102,895,1).
txt_page(196,895,1).
txt_page(827,895,1).
txt_page(409,895,1).
txt_page(152,895,1).
txt_page(278,895,1).
txt_page(837,895,1).
txt_page(721,895,1).
txt_page(186,895,1).
txt_page(166,895,1).
txt_page(774,896,1).
txt_page(352,896,1).
txt_page(205,896,1).
txt_page(686,896,1).
txt_page(462,896,1).
txt_page(313,896,1).
txt_page(556,896,1).
txt_page(245,896,1).
txt_page(355,896,1).
txt_page(594,896,1).
txt_page(409,896,1).
txt_page(386,896,1).
txt_page(412,896,1).
txt_page(903,896,1).
txt_page(298,896,1).
txt_page(829,896,1).
txt_page(521,896,1).
txt_page(60,896,1).
txt_page(56,896,1).
txt_page(850,896,1).
txt_page(534,897,1).
txt_page(366,897,1).
txt_page(276,897,1).
txt_page(285,897,1).
txt_page(788,897,1).
txt_page(793,897,1).
txt_page(709,897,1).
txt_page(802,897,1).
txt_page(312,897,1).
txt_page(242,897,1).
txt_page(675,897,1).
txt_page(526,897,1).
txt_page(661,897,1).
txt_page(87,897,1).
txt_page(659,897,1).
txt_page(536,897,1).
txt_page(281,897,1).
txt_page(184,897,1).
txt_page(485,897,1).
txt_page(632,897,1).
txt_page(938,898,1).
txt_page(399,898,1).
txt_page(183,898,1).
txt_page(739,898,1).
txt_page(449,898,1).
txt_page(430,898,1).
txt_page(44,898,1).
txt_page(908,898,1).
txt_page(878,898,1).
txt_page(18,898,1).
txt_page(247,898,1).
txt_page(315,898,1).
txt_page(9,898,1).
txt_page(345,898,1).
txt_page(55,898,1).
txt_page(602,898,1).
txt_page(239,898,1).
txt_page(139,898,1).
txt_page(688,898,1).
txt_page(332,898,1).
txt_page(442,899,1).
txt_page(302,899,1).
txt_page(843,899,1).
txt_page(697,899,1).
txt_page(92,899,1).
txt_page(463,899,1).
txt_page(667,899,1).
txt_page(818,899,1).
txt_page(822,899,1).
txt_page(106,899,1).
txt_page(622,899,1).
txt_page(437,899,1).
txt_page(384,899,1).
txt_page(162,899,1).
txt_page(861,899,1).
txt_page(790,899,1).
txt_page(75,899,2).
txt_page(39,899,1).
txt_page(122,899,1).
txt_page(609,900,1).
txt_page(301,900,1).
txt_page(860,900,1).
txt_page(118,900,1).
txt_page(682,900,1).
txt_page(299,900,1).
txt_page(470,900,1).
txt_page(819,900,1).
txt_page(787,900,1).
txt_page(275,900,1).
txt_page(56,900,1).
txt_page(852,900,1).
txt_page(88,900,1).
txt_page(756,900,1).
txt_page(431,900,1).
txt_page(288,900,1).
txt_page(40,900,1).
txt_page(157,900,1).
txt_page(927,900,1).
txt_page(777,900,1).
txt_page(126,901,1).
txt_page(690,901,1).
txt_page(580,901,1).
txt_page(292,901,1).
txt_page(762,901,1).
txt_page(773,901,1).
txt_page(950,901,1).
txt_page(37,901,1).
txt_page(184,901,1).
txt_page(426,901,1).
txt_page(430,901,1).
txt_page(784,901,1).
txt_page(139,901,1).
txt_page(794,901,1).
txt_page(813,901,1).
txt_page(467,901,1).
txt_page(956,901,1).
txt_page(289,901,1).
txt_page(116,901,1).
txt_page(492,901,1).
txt_page(341,902,1).
txt_page(488,902,1).
txt_page(437,902,1).
txt_page(180,902,1).
txt_page(459,902,1).
txt_page(29,902,1).
txt_page(863,902,1).
txt_page(89,902,1).
txt_page(1,902,1).
txt_page(901,902,1).
txt_page(896,902,1).
txt_page(267,902,1).
txt_page(150,902,1).
txt_page(889,902,1).
txt_page(851,902,1).
txt_page(947,902,1).
txt_page(791,902,1).
txt_page(467,902,1).
txt_page(763,902,1).
txt_page(692,902,1).
txt_page(397,903,1).
txt_page(552,903,1).
txt_page(809,903,1).
txt_page(652,903,1).
txt_page(353,903,1).
txt_page(503,903,1).
txt_page(575,903,1).
txt_page(347,903,1).
txt_page(633,903,1).
txt_page(228,903,1).
txt_page(680,903,1).
txt_page(914,903,1).
txt_page(333,903,1).
txt_page(492,903,1).
txt_page(138,903,1).
txt_page(797,903,1).
txt_page(807,903,1).
txt_page(811,903,1).
txt_page(216,903,1).
txt_page(437,903,1).
txt_page(407,904,1).
txt_page(221,904,1).
txt_page(783,904,1).
txt_page(808,904,1).
txt_page(843,904,1).
txt_page(399,904,1).
txt_page(117,904,1).
txt_page(522,904,1).
txt_page(674,904,1).
txt_page(656,904,1).
txt_page(860,904,1).
txt_page(864,904,1).
txt_page(116,904,1).
txt_page(495,904,1).
txt_page(178,904,1).
txt_page(394,904,1).
txt_page(941,904,1).
txt_page(196,904,1).
txt_page(439,904,1).
txt_page(265,904,1).
txt_page(794,905,1).
txt_page(203,905,1).
txt_page(185,905,1).
txt_page(957,905,1).
txt_page(925,905,1).
txt_page(927,905,1).
txt_page(905,905,1).
txt_page(607,905,1).
txt_page(803,905,1).
txt_page(744,905,1).
txt_page(709,905,1).
txt_page(472,905,1).
txt_page(410,905,1).
txt_page(229,905,1).
txt_page(530,905,1).
txt_page(656,905,1).
txt_page(594,905,1).
txt_page(638,905,1).
txt_page(830,905,1).
txt_page(632,905,1).
txt_page(379,906,1).
txt_page(677,906,1).
txt_page(814,906,1).
txt_page(588,906,1).
txt_page(608,906,1).
txt_page(944,906,1).
txt_page(554,906,1).
txt_page(159,906,2).
txt_page(27,906,1).
txt_page(720,906,1).
txt_page(843,906,1).
txt_page(539,906,1).
txt_page(682,906,1).
txt_page(555,906,1).
txt_page(937,906,1).
txt_page(169,906,1).
txt_page(889,906,1).
txt_page(756,906,1).
txt_page(20,906,1).
txt_page(389,907,1).
txt_page(136,907,1).
txt_page(695,907,1).
txt_page(224,907,1).
txt_page(596,907,1).
txt_page(713,907,1).
txt_page(558,907,1).
txt_page(547,907,2).
txt_page(473,907,1).
txt_page(139,907,1).
txt_page(593,907,1).
txt_page(150,907,1).
txt_page(31,907,1).
txt_page(575,907,1).
txt_page(40,907,1).
txt_page(287,907,1).
txt_page(184,907,1).
txt_page(25,907,1).
txt_page(137,907,1).
txt_page(348,908,1).
txt_page(373,908,1).
txt_page(489,908,1).
txt_page(663,908,1).
txt_page(776,908,1).
txt_page(184,908,1).
txt_page(951,908,1).
txt_page(454,908,1).
txt_page(501,908,1).
txt_page(739,908,1).
txt_page(468,908,1).
txt_page(647,908,1).
txt_page(494,908,1).
txt_page(32,908,1).
txt_page(666,908,1).
txt_page(588,908,1).
txt_page(752,908,1).
txt_page(202,908,1).
txt_page(480,908,1).
txt_page(435,908,1).
txt_page(711,909,1).
txt_page(545,909,1).
txt_page(250,909,1).
txt_page(179,909,1).
txt_page(232,909,1).
txt_page(676,909,1).
txt_page(93,909,1).
txt_page(689,909,1).
txt_page(355,909,1).
txt_page(674,909,1).
txt_page(447,909,1).
txt_page(176,909,1).
txt_page(354,909,1).
txt_page(556,909,1).
txt_page(171,909,1).
txt_page(623,909,1).
txt_page(30,909,1).
txt_page(580,909,1).
txt_page(597,909,1).
txt_page(393,909,1).
txt_page(798,910,1).
txt_page(3,910,1).
txt_page(16,910,1).
txt_page(953,910,1).
txt_page(264,910,1).
txt_page(912,910,1).
txt_page(618,910,1).
txt_page(514,910,1).
txt_page(42,910,1).
txt_page(794,910,1).
txt_page(564,910,1).
txt_page(875,910,1).
txt_page(924,910,1).
txt_page(944,910,1).
txt_page(445,910,1).
txt_page(5,910,1).
txt_page(380,910,1).
txt_page(663,910,1).
txt_page(918,910,1).
txt_page(405,910,1).
txt_page(25,911,1).
txt_page(451,911,1).
txt_page(368,911,1).
txt_page(206,911,1).
txt_page(479,911,1).
txt_page(769,911,1).
txt_page(737,911,1).
txt_page(758,911,1).
txt_page(205,911,1).
txt_page(170,911,1).
txt_page(928,911,1).
txt_page(831,911,1).
txt_page(612,911,1).
txt_page(546,911,1).
txt_page(926,911,1).
txt_page(954,911,1).
txt_page(957,911,1).
txt_page(409,911,1).
txt_page(693,911,1).
txt_page(626,911,1).
txt_page(575,912,1).
txt_page(540,912,1).
txt_page(242,912,1).
txt_page(249,912,1).
txt_page(391,912,1).
txt_page(465,912,1).
txt_page(966,912,1).
txt_page(584,912,1).
txt_page(642,912,1).
txt_page(519,912,1).
txt_page(284,912,2).
txt_page(896,912,1).
txt_page(940,912,1).
txt_page(661,912,1).
txt_page(691,912,1).
txt_page(143,912,1).
txt_page(934,912,1).
txt_page(959,912,1).
txt_page(671,912,1).
txt_page(502,913,1).
txt_page(768,913,1).
txt_page(197,913,2).
txt_page(615,913,1).
txt_page(11,913,1).
txt_page(514,913,1).
txt_page(958,913,1).
txt_page(554,913,1).
txt_page(358,913,1).
txt_page(933,913,1).
txt_page(161,913,1).
txt_page(792,913,1).
txt_page(227,913,1).
txt_page(315,913,1).
txt_page(512,913,1).
txt_page(140,913,1).
txt_page(824,913,1).
txt_page(922,913,1).
txt_page(330,913,1).
txt_page(778,914,1).
txt_page(57,914,1).
txt_page(171,914,1).
txt_page(336,914,1).
txt_page(212,914,1).
txt_page(420,914,1).
txt_page(568,914,1).
txt_page(288,914,1).
txt_page(328,914,1).
txt_page(818,914,1).
txt_page(931,914,1).
txt_page(172,914,1).
txt_page(306,914,1).
txt_page(116,914,1).
txt_page(512,914,1).
txt_page(32,914,1).
txt_page(699,914,1).
txt_page(197,914,1).
txt_page(749,914,1).
txt_page(645,914,1).
txt_page(839,915,1).
txt_page(338,915,1).
txt_page(515,915,1).
txt_page(536,915,1).
txt_page(369,915,1).
txt_page(339,915,1).
txt_page(48,915,1).
txt_page(870,915,1).
txt_page(177,915,1).
txt_page(838,915,1).
txt_page(173,915,1).
txt_page(443,915,1).
txt_page(622,915,1).
txt_page(117,915,1).
txt_page(749,915,1).
txt_page(362,915,1).
txt_page(740,915,1).
txt_page(166,915,1).
txt_page(306,915,1).
txt_page(41,915,1).
txt_page(674,916,1).
txt_page(953,916,1).
txt_page(755,916,1).
txt_page(625,916,1).
txt_page(888,916,1).
txt_page(484,916,1).
txt_page(312,916,1).
txt_page(341,916,1).
txt_page(95,916,1).
txt_page(871,916,1).
txt_page(287,916,1).
txt_page(903,916,1).
txt_page(347,916,1).
txt_page(492,916,1).
txt_page(863,916,1).
txt_page(113,916,1).
txt_page(372,916,1).
txt_page(121,916,1).
txt_page(130,916,1).
txt_page(143,916,1).
txt_page(526,917,1).
txt_page(406,917,1).
txt_page(222,917,1).
txt_page(528,917,1).
txt_page(562,917,1).
txt_page(176,917,1).
txt_page(324,917,1).
txt_page(934,917,1).
txt_page(945,917,1).
txt_page(190,917,1).
txt_page(441,917,1).
txt_page(469,917,1).
txt_page(396,917,1).
txt_page(952,917,1).
txt_page(657,917,1).
txt_page(697,917,1).
txt_page(589,917,1).
txt_page(413,917,1).
txt_page(479,917,1).
txt_page(545,917,1).
txt_page(957,918,1).
txt_page(773,918,1).
txt_page(175,918,1).
txt_page(188,918,1).
txt_page(760,918,1).
txt_page(960,918,1).
txt_page(936,918,1).
txt_page(509,918,1).
txt_page(703,918,1).
txt_page(106,918,1).
txt_page(552,918,1).
txt_page(547,918,1).
txt_page(844,918,1).
txt_page(262,918,1).
txt_page(372,918,1).
txt_page(142,918,1).
txt_page(803,918,1).
txt_page(590,918,1).
txt_page(817,918,1).
txt_page(104,918,1).
txt_page(793,919,1).
txt_page(552,919,1).
txt_page(480,919,1).
txt_page(746,919,1).
txt_page(607,919,1).
txt_page(548,919,1).
txt_page(125,919,1).
txt_page(549,919,1).
txt_page(431,919,1).
txt_page(710,919,1).
txt_page(732,919,1).
txt_page(715,919,1).
txt_page(413,919,1).
txt_page(510,919,1).
txt_page(452,919,1).
txt_page(427,919,1).
txt_page(138,919,1).
txt_page(350,919,1).
txt_page(681,919,1).
txt_page(700,919,1).
txt_page(170,920,1).
txt_page(830,920,1).
txt_page(648,920,1).
txt_page(802,920,1).
txt_page(45,920,1).
txt_page(885,920,1).
txt_page(488,920,1).
txt_page(239,920,1).
txt_page(587,920,1).
txt_page(917,920,1).
txt_page(16,920,1).
txt_page(866,920,1).
txt_page(12,920,1).
txt_page(46,920,1).
txt_page(454,920,1).
txt_page(805,920,1).
txt_page(948,920,1).
txt_page(425,920,2).
txt_page(444,920,1).
txt_page(303,921,1).
txt_page(610,921,1).
txt_page(711,921,1).
txt_page(654,921,1).
txt_page(596,921,1).
txt_page(59,921,1).
txt_page(750,921,1).
txt_page(661,921,1).
txt_page(557,921,1).
txt_page(498,921,1).
txt_page(538,921,1).
txt_page(739,921,1).
txt_page(631,921,1).
txt_page(160,921,1).
txt_page(819,921,1).
txt_page(350,921,1).
txt_page(309,921,1).
txt_page(80,921,1).
txt_page(140,921,1).
txt_page(395,921,1).
txt_page(232,922,1).
txt_page(731,922,1).
txt_page(381,922,1).
txt_page(792,922,1).
txt_page(922,922,1).
txt_page(255,922,1).
txt_page(218,922,1).
txt_page(526,922,1).
txt_page(630,922,1).
txt_page(194,922,1).
txt_page(334,922,1).
txt_page(40,922,1).
txt_page(741,922,1).
txt_page(203,922,1).
txt_page(282,922,1).
txt_page(372,922,1).
txt_page(97,922,1).
txt_page(508,922,1).
txt_page(230,922,2).
txt_page(654,923,1).
txt_page(595,923,1).
txt_page(391,923,1).
txt_page(44,923,1).
txt_page(202,923,1).
txt_page(366,923,1).
txt_page(616,923,1).
txt_page(634,923,1).
txt_page(455,923,1).
txt_page(364,923,1).
txt_page(696,923,1).
txt_page(427,923,1).
txt_page(569,923,1).
txt_page(606,923,1).
txt_page(306,923,1).
txt_page(165,923,1).
txt_page(842,923,1).
txt_page(674,923,1).
txt_page(695,923,1).
txt_page(953,923,1).
txt_page(613,924,1).
txt_page(450,924,1).
txt_page(79,924,1).
txt_page(817,924,1).
txt_page(122,924,1).
txt_page(314,924,1).
txt_page(148,924,1).
txt_page(463,924,1).
txt_page(208,924,1).
txt_page(841,924,1).
txt_page(265,924,1).
txt_page(352,924,1).
txt_page(642,924,1).
txt_page(505,924,1).
txt_page(380,924,1).
txt_page(454,924,1).
txt_page(220,924,1).
txt_page(253,924,1).
txt_page(440,924,1).
txt_page(70,924,1).
txt_page(466,925,1).
txt_page(506,925,1).
txt_page(202,925,1).
txt_page(111,925,1).
txt_page(299,925,1).
txt_page(880,925,1).
txt_page(254,925,1).
txt_page(40,925,1).
txt_page(645,925,1).
txt_page(672,925,1).
txt_page(29,925,1).
txt_page(476,925,1).
txt_page(588,925,2).
txt_page(437,925,1).
txt_page(482,925,1).
txt_page(350,925,1).
txt_page(343,925,1).
txt_page(324,925,1).
txt_page(55,925,1).
txt_page(347,926,1).
txt_page(845,926,1).
txt_page(653,926,1).
txt_page(721,926,1).
txt_page(941,926,1).
txt_page(217,926,1).
txt_page(193,926,1).
txt_page(960,926,1).
txt_page(306,926,1).
txt_page(908,926,1).
txt_page(569,926,1).
txt_page(596,926,1).
txt_page(227,926,1).
txt_page(743,926,1).
txt_page(902,926,1).
txt_page(381,926,1).
txt_page(432,926,1).
txt_page(599,926,1).
txt_page(32,926,1).
txt_page(394,926,1).
txt_page(967,927,1).
txt_page(257,927,1).
txt_page(384,927,1).
txt_page(482,927,1).
txt_page(273,927,1).
txt_page(701,927,1).
txt_page(362,927,1).
txt_page(177,927,1).
txt_page(213,927,1).
txt_page(332,927,1).
txt_page(841,927,1).
txt_page(415,927,1).
txt_page(514,927,1).
txt_page(84,927,1).
txt_page(517,927,1).
txt_page(307,927,1).
txt_page(697,927,1).
txt_page(754,927,1).
txt_page(219,927,1).
txt_page(96,927,1).
txt_page(406,928,1).
txt_page(388,928,1).
txt_page(903,928,1).
txt_page(170,928,1).
txt_page(704,928,1).
txt_page(860,928,1).
txt_page(823,928,1).
txt_page(690,928,1).
txt_page(826,928,1).
txt_page(28,928,1).
txt_page(311,928,1).
txt_page(800,928,1).
txt_page(66,928,1).
txt_page(417,928,1).
txt_page(849,928,1).
txt_page(610,928,1).
txt_page(58,928,1).
txt_page(624,928,1).
txt_page(414,928,1).
txt_page(769,928,1).
txt_page(148,929,1).
txt_page(18,929,1).
txt_page(319,929,1).
txt_page(322,929,1).
txt_page(666,929,1).
txt_page(86,929,1).
txt_page(478,929,1).
txt_page(919,929,1).
txt_page(568,929,1).
txt_page(244,929,1).
txt_page(213,929,1).
txt_page(669,929,1).
txt_page(607,929,1).
txt_page(71,929,1).
txt_page(938,929,1).
txt_page(723,929,1).
txt_page(294,929,1).
txt_page(332,929,1).
txt_page(913,929,1).
txt_page(771,929,1).
txt_page(922,930,1).
txt_page(455,930,1).
txt_page(124,930,1).
txt_page(217,930,1).
txt_page(597,930,1).
txt_page(817,930,1).
txt_page(121,930,1).
txt_page(137,930,1).
txt_page(467,930,1).
txt_page(708,930,1).
txt_page(851,930,1).
txt_page(329,930,1).
txt_page(469,930,1).
txt_page(186,930,1).
txt_page(640,930,1).
txt_page(52,930,1).
txt_page(376,930,1).
txt_page(617,930,1).
txt_page(797,930,1).
txt_page(450,930,1).
txt_page(551,931,1).
txt_page(161,931,1).
txt_page(787,931,1).
txt_page(838,931,1).
txt_page(195,931,1).
txt_page(927,931,1).
txt_page(450,931,1).
txt_page(112,931,1).
txt_page(395,931,1).
txt_page(42,931,1).
txt_page(113,931,1).
txt_page(480,931,1).
txt_page(313,931,1).
txt_page(784,931,1).
txt_page(534,931,1).
txt_page(274,931,1).
txt_page(84,931,1).
txt_page(182,931,1).
txt_page(321,931,1).
txt_page(736,931,1).
txt_page(842,932,1).
txt_page(463,932,1).
txt_page(675,932,1).
txt_page(44,932,1).
txt_page(360,932,1).
txt_page(34,932,1).
txt_page(786,932,1).
txt_page(944,932,1).
txt_page(227,932,1).
txt_page(176,932,1).
txt_page(796,932,1).
txt_page(498,932,1).
txt_page(143,932,1).
txt_page(920,932,1).
txt_page(479,932,1).
txt_page(121,932,1).
txt_page(289,932,1).
txt_page(804,932,1).
txt_page(416,932,1).
txt_page(509,932,1).
txt_page(351,933,1).
txt_page(254,933,1).
txt_page(170,933,1).
txt_page(173,933,1).
txt_page(599,933,1).
txt_page(122,933,1).
txt_page(421,933,1).
txt_page(891,933,1).
txt_page(395,933,1).
txt_page(782,933,1).
txt_page(819,933,1).
txt_page(877,933,1).
txt_page(217,933,1).
txt_page(431,933,1).
txt_page(62,933,1).
txt_page(674,933,1).
txt_page(22,933,1).
txt_page(709,933,1).
txt_page(637,933,1).
txt_page(888,933,1).
txt_page(865,934,1).
txt_page(798,934,1).
txt_page(330,934,1).
txt_page(772,934,1).
txt_page(326,934,1).
txt_page(866,934,1).
txt_page(492,934,1).
txt_page(469,934,1).
txt_page(880,934,1).
txt_page(20,934,1).
txt_page(775,934,1).
txt_page(786,934,1).
txt_page(635,934,1).
txt_page(158,934,1).
txt_page(327,934,1).
txt_page(137,934,1).
txt_page(313,934,1).
txt_page(71,934,1).
txt_page(367,934,1).
txt_page(365,934,1).
txt_page(940,935,1).
txt_page(910,935,2).
txt_page(684,935,1).
txt_page(962,935,1).
txt_page(838,935,1).
txt_page(674,935,1).
txt_page(663,935,1).
txt_page(21,935,1).
txt_page(714,935,1).
txt_page(734,935,1).
txt_page(202,935,1).
txt_page(210,935,1).
txt_page(882,935,1).
txt_page(128,935,2).
txt_page(101,935,2).
txt_page(166,935,1).
txt_page(965,935,1).
txt_page(481,936,1).
txt_page(236,936,1).
txt_page(355,936,1).
txt_page(492,936,1).
txt_page(250,936,1).
txt_page(451,936,1).
txt_page(780,936,1).
txt_page(545,936,1).
txt_page(805,936,1).
txt_page(428,936,1).
txt_page(598,936,1).
txt_page(380,936,1).
txt_page(48,936,1).
txt_page(180,936,1).
txt_page(422,936,1).
txt_page(77,936,1).
txt_page(67,936,1).
txt_page(336,936,1).
txt_page(650,936,1).
txt_page(337,936,1).
txt_page(878,937,1).
txt_page(101,937,1).
txt_page(271,937,1).
txt_page(816,937,1).
txt_page(854,937,1).
txt_page(873,937,1).
txt_page(957,937,1).
txt_page(839,937,1).
txt_page(829,937,1).
txt_page(937,937,1).
txt_page(118,937,1).
txt_page(174,937,1).
txt_page(821,937,1).
txt_page(4,937,1).
txt_page(422,937,1).
txt_page(417,937,1).
txt_page(437,937,1).
txt_page(707,937,1).
txt_page(853,937,1).
txt_page(476,937,1).
txt_page(358,938,1).
txt_page(388,938,1).
txt_page(249,938,1).
txt_page(950,938,1).
txt_page(819,938,1).
txt_page(255,938,1).
txt_page(456,938,1).
txt_page(529,938,1).
txt_page(449,938,1).
txt_page(688,938,1).
txt_page(831,938,1).
txt_page(546,938,1).
txt_page(425,938,1).
txt_page(41,938,2).
txt_page(651,938,1).
txt_page(240,938,1).
txt_page(574,938,1).
txt_page(791,938,1).
txt_page(139,938,1).
txt_page(865,939,1).
txt_page(595,939,1).
txt_page(963,939,1).
txt_page(834,939,1).
txt_page(610,939,1).
txt_page(64,939,1).
txt_page(709,939,1).
txt_page(219,939,1).
txt_page(659,939,1).
txt_page(762,939,1).
txt_page(136,939,1).
txt_page(357,939,1).
txt_page(234,939,1).
txt_page(912,939,1).
txt_page(366,939,1).
txt_page(302,939,1).
txt_page(529,939,1).
txt_page(300,939,1).
txt_page(959,939,1).
txt_page(900,939,1).
txt_page(678,940,1).
txt_page(221,940,1).
txt_page(316,940,1).
txt_page(548,940,1).
txt_page(159,940,1).
txt_page(884,940,1).
txt_page(421,940,1).
txt_page(371,940,1).
txt_page(603,940,1).
txt_page(868,940,1).
txt_page(379,940,1).
txt_page(498,940,1).
txt_page(312,940,1).
txt_page(864,940,1).
txt_page(14,940,1).
txt_page(902,940,1).
txt_page(146,940,1).
txt_page(476,940,1).
txt_page(39,940,1).
txt_page(388,940,1).
txt_page(898,941,1).
txt_page(749,941,1).
txt_page(149,941,1).
txt_page(568,941,1).
txt_page(880,941,1).
txt_page(10,941,1).
txt_page(484,941,1).
txt_page(168,941,1).
txt_page(963,941,1).
txt_page(454,941,1).
txt_page(365,941,1).
txt_page(801,941,1).
txt_page(702,941,1).
txt_page(445,941,1).
txt_page(410,941,1).
txt_page(859,941,1).
txt_page(879,941,1).
txt_page(867,941,1).
txt_page(448,941,1).
txt_page(495,941,1).
txt_page(716,942,1).
txt_page(950,942,1).
txt_page(47,942,1).
txt_page(380,942,1).
txt_page(911,942,1).
txt_page(324,942,1).
txt_page(25,942,1).
txt_page(487,942,1).
txt_page(803,942,1).
txt_page(718,942,1).
txt_page(78,942,1).
txt_page(311,942,1).
txt_page(95,942,1).
txt_page(313,942,1).
txt_page(245,942,1).
txt_page(558,942,1).
txt_page(391,942,1).
txt_page(801,942,1).
txt_page(883,942,1).
txt_page(531,942,1).
txt_page(798,943,1).
txt_page(692,943,1).
txt_page(355,943,1).
txt_page(630,943,1).
txt_page(688,943,1).
txt_page(226,943,1).
txt_page(392,943,1).
txt_page(479,943,1).
txt_page(32,943,1).
txt_page(639,943,1).
txt_page(925,943,1).
txt_page(520,943,1).
txt_page(741,943,1).
txt_page(656,943,1).
txt_page(159,943,1).
txt_page(835,943,1).
txt_page(185,943,1).
txt_page(621,943,1).
txt_page(429,943,1).
txt_page(951,943,1).
txt_page(850,944,1).
txt_page(335,944,1).
txt_page(126,944,1).
txt_page(658,944,1).
txt_page(904,944,1).
txt_page(351,944,1).
txt_page(832,944,1).
txt_page(915,944,1).
txt_page(556,944,1).
txt_page(872,944,1).
txt_page(352,944,1).
txt_page(756,944,1).
txt_page(736,944,1).
txt_page(263,944,1).
txt_page(532,944,1).
txt_page(694,944,1).
txt_page(347,944,1).
txt_page(21,944,1).
txt_page(92,944,1).
txt_page(202,944,1).
txt_page(887,945,1).
txt_page(680,945,1).
txt_page(586,945,1).
txt_page(698,945,1).
txt_page(384,945,1).
txt_page(513,945,1).
txt_page(447,945,1).
txt_page(325,945,1).
txt_page(455,945,1).
txt_page(154,945,1).
txt_page(601,945,1).
txt_page(537,945,1).
txt_page(657,945,1).
txt_page(618,945,1).
txt_page(846,945,1).
txt_page(913,945,2).
txt_page(51,945,1).
txt_page(826,945,1).
txt_page(125,945,1).
txt_page(638,946,1).
txt_page(227,946,1).
txt_page(600,946,1).
txt_page(812,946,1).
txt_page(554,946,1).
txt_page(275,946,1).
txt_page(931,946,1).
txt_page(529,946,1).
txt_page(635,946,1).
txt_page(146,946,1).
txt_page(145,946,1).
txt_page(108,946,1).
txt_page(440,946,1).
txt_page(406,946,1).
txt_page(289,946,1).
txt_page(699,946,1).
txt_page(524,946,1).
txt_page(272,946,1).
txt_page(425,946,1).
txt_page(531,946,1).
txt_page(329,947,1).
txt_page(838,947,1).
txt_page(159,947,1).
txt_page(560,947,1).
txt_page(57,947,1).
txt_page(968,947,1).
txt_page(261,947,1).
txt_page(818,947,1).
txt_page(340,947,1).
txt_page(486,947,1).
txt_page(95,947,1).
txt_page(678,947,1).
txt_page(218,947,1).
txt_page(370,947,2).
txt_page(175,947,1).
txt_page(162,947,1).
txt_page(233,947,1).
txt_page(281,947,1).
txt_page(386,947,1).
txt_page(10,948,1).
txt_page(68,948,1).
txt_page(444,948,1).
txt_page(385,948,1).
txt_page(202,948,1).
txt_page(658,948,1).
txt_page(250,948,1).
txt_page(481,948,1).
txt_page(79,948,1).
txt_page(295,948,1).
txt_page(392,948,1).
txt_page(397,948,1).
txt_page(132,948,1).
txt_page(348,948,1).
txt_page(628,948,1).
txt_page(370,948,1).
txt_page(192,948,1).
txt_page(506,948,1).
txt_page(30,948,1).
txt_page(379,948,1).
txt_page(42,949,1).
txt_page(683,949,1).
txt_page(416,949,1).
txt_page(570,949,1).
txt_page(271,949,1).
txt_page(724,949,1).
txt_page(63,949,1).
txt_page(690,949,1).
txt_page(320,949,1).
txt_page(634,949,1).
txt_page(728,949,1).
txt_page(455,949,1).
txt_page(327,949,1).
txt_page(918,949,1).
txt_page(30,949,1).
txt_page(413,949,1).
txt_page(795,949,1).
txt_page(746,949,1).
txt_page(464,949,1).
txt_page(962,949,1).
txt_page(208,950,1).
txt_page(462,950,1).
txt_page(609,950,1).
txt_page(199,950,1).
txt_page(448,950,1).
txt_page(272,950,1).
txt_page(951,950,1).
txt_page(880,950,1).
txt_page(342,950,2).
txt_page(866,950,1).
txt_page(953,950,1).
txt_page(595,950,1).
txt_page(361,950,1).
txt_page(140,950,1).
txt_page(358,950,1).
txt_page(88,950,1).
txt_page(905,950,1).
txt_page(917,950,1).
txt_page(336,950,1).
txt_page(587,951,1).
txt_page(476,951,1).
txt_page(141,951,1).
txt_page(235,951,1).
txt_page(590,951,1).
txt_page(69,951,1).
txt_page(846,951,1).
txt_page(497,951,1).
txt_page(194,951,1).
txt_page(208,951,1).
txt_page(934,951,1).
txt_page(118,951,1).
txt_page(856,951,1).
txt_page(957,951,1).
txt_page(108,951,1).
txt_page(614,951,1).
txt_page(84,951,1).
txt_page(166,951,1).
txt_page(939,951,1).
txt_page(197,951,1).
txt_page(73,952,1).
txt_page(276,952,1).
txt_page(708,952,1).
txt_page(967,952,1).
txt_page(303,952,1).
txt_page(847,952,1).
txt_page(598,952,1).
txt_page(48,952,1).
txt_page(822,952,1).
txt_page(147,952,1).
txt_page(172,952,1).
txt_page(641,952,1).
txt_page(805,952,1).
txt_page(834,952,1).
txt_page(563,952,1).
txt_page(495,952,1).
txt_page(343,952,1).
txt_page(550,952,1).
txt_page(347,952,1).
txt_page(907,952,1).
txt_page(945,953,1).
txt_page(950,953,1).
txt_page(261,953,1).
txt_page(497,953,1).
txt_page(757,953,1).
txt_page(577,953,1).
txt_page(440,953,1).
txt_page(198,953,1).
txt_page(705,953,1).
txt_page(725,953,1).
txt_page(329,953,1).
txt_page(691,953,1).
txt_page(443,953,1).
txt_page(33,953,1).
txt_page(427,953,1).
txt_page(453,953,1).
txt_page(93,953,1).
txt_page(75,953,1).
txt_page(394,953,1).
txt_page(883,953,1).
txt_page(22,954,1).
txt_page(634,954,1).
txt_page(141,954,1).
txt_page(146,954,1).
txt_page(776,954,1).
txt_page(441,954,1).
txt_page(376,954,1).
txt_page(528,954,1).
txt_page(602,954,1).
txt_page(921,954,1).
txt_page(553,954,1).
txt_page(650,954,1).
txt_page(700,954,1).
txt_page(545,954,1).
txt_page(810,954,1).
txt_page(547,954,1).
txt_page(429,954,1).
txt_page(797,954,1).
txt_page(342,954,1).
txt_page(458,954,1).
txt_page(231,955,1).
txt_page(614,955,1).
txt_page(151,955,1).
txt_page(453,955,1).
txt_page(140,955,1).
txt_page(894,955,1).
txt_page(451,955,1).
txt_page(276,955,1).
txt_page(656,955,1).
txt_page(382,955,1).
txt_page(810,955,1).
txt_page(283,955,1).
txt_page(876,955,1).
txt_page(432,955,1).
txt_page(695,955,1).
txt_page(553,955,1).
txt_page(574,955,1).
txt_page(233,955,1).
txt_page(673,955,1).
txt_page(908,955,1).
txt_page(867,956,1).
txt_page(935,956,1).
txt_page(150,956,1).
txt_page(326,956,1).
txt_page(442,956,1).
txt_page(423,956,1).
txt_page(620,956,1).
txt_page(261,956,1).
txt_page(890,956,1).
txt_page(575,956,1).
txt_page(201,956,1).
txt_page(614,956,1).
txt_page(661,956,1).
txt_page(489,956,1).
txt_page(662,956,1).
txt_page(57,956,1).
txt_page(854,956,1).
txt_page(92,956,1).
txt_page(308,956,1).
txt_page(444,956,1).
txt_page(571,957,1).
txt_page(612,957,1).
txt_page(703,957,1).
txt_page(952,957,1).
txt_page(498,957,1).
txt_page(795,957,1).
txt_page(810,957,1).
txt_page(874,957,1).
txt_page(23,957,1).
txt_page(879,957,1).
txt_page(177,957,1).
txt_page(573,957,1).
txt_page(455,957,1).
txt_page(257,957,1).
txt_page(136,957,1).
txt_page(849,957,1).
txt_page(866,957,1).
txt_page(393,957,1).
txt_page(604,957,1).
txt_page(318,957,1).
txt_page(511,958,1).
txt_page(785,958,1).
txt_page(318,958,1).
txt_page(921,958,1).
txt_page(780,958,1).
txt_page(651,958,1).
txt_page(452,958,1).
txt_page(468,958,1).
txt_page(117,958,1).
txt_page(741,958,1).
txt_page(327,958,1).
txt_page(326,958,1).
txt_page(261,958,1).
txt_page(920,958,1).
txt_page(164,958,1).
txt_page(181,958,1).
txt_page(800,958,1).
txt_page(98,958,1).
txt_page(950,958,1).
txt_page(370,958,1).
txt_page(733,959,1).
txt_page(556,959,1).
txt_page(507,959,1).
txt_page(370,959,1).
txt_page(756,959,1).
txt_page(385,959,1).
txt_page(253,959,1).
txt_page(446,959,1).
txt_page(960,959,1).
txt_page(306,959,1).
txt_page(434,959,1).
txt_page(48,959,1).
txt_page(184,959,1).
txt_page(366,959,1).
txt_page(269,959,1).
txt_page(524,959,1).
txt_page(390,959,1).
txt_page(99,959,1).
txt_page(42,959,1).
txt_page(788,959,1).
txt_page(246,960,1).
txt_page(66,960,1).
txt_page(187,960,1).
txt_page(295,960,1).
txt_page(642,960,1).
txt_page(445,960,1).
txt_page(834,960,1).
txt_page(469,960,1).
txt_page(812,960,1).
txt_page(62,960,1).
txt_page(959,960,1).
txt_page(152,960,1).
txt_page(540,960,1).
txt_page(234,960,1).
txt_page(957,960,1).
txt_page(939,960,1).
txt_page(189,960,1).
txt_page(589,960,1).
txt_page(256,960,1).
txt_page(98,960,1).
txt_page(236,961,1).
txt_page(636,961,1).
txt_page(28,961,1).
txt_page(519,961,1).
txt_page(153,961,1).
txt_page(783,961,1).
txt_page(7,961,1).
txt_page(76,961,1).
txt_page(422,961,1).
txt_page(70,961,1).
txt_page(278,961,1).
txt_page(562,961,1).
txt_page(917,961,1).
txt_page(597,961,1).
txt_page(442,961,1).
txt_page(826,961,2).
txt_page(919,961,1).
txt_page(471,961,1).
txt_page(346,961,1).
txt_page(938,962,1).
txt_page(547,962,1).
txt_page(732,962,1).
txt_page(20,962,1).
txt_page(177,962,1).
txt_page(482,962,1).
txt_page(703,962,1).
txt_page(271,962,1).
txt_page(733,962,1).
txt_page(932,962,1).
txt_page(681,962,1).
txt_page(484,962,1).
txt_page(246,962,1).
txt_page(515,962,1).
txt_page(372,962,1).
txt_page(632,962,1).
txt_page(471,962,1).
txt_page(367,962,1).
txt_page(893,962,1).
txt_page(398,962,1).
txt_page(870,963,1).
txt_page(493,963,1).
txt_page(383,963,1).
txt_page(283,963,1).
txt_page(447,963,1).
txt_page(607,963,1).
txt_page(859,963,1).
txt_page(195,963,1).
txt_page(227,963,1).
txt_page(624,963,1).
txt_page(713,963,2).
txt_page(274,963,1).
txt_page(817,963,1).
txt_page(260,963,1).
txt_page(731,963,1).
txt_page(408,963,1).
txt_page(959,963,1).
txt_page(223,963,1).
txt_page(131,963,1).
txt_page(375,964,1).
txt_page(272,964,1).
txt_page(695,964,1).
txt_page(594,964,1).
txt_page(79,964,1).
txt_page(502,964,1).
txt_page(270,964,1).
txt_page(274,964,1).
txt_page(160,964,1).
txt_page(941,964,1).
txt_page(521,964,1).
txt_page(103,964,1).
txt_page(434,964,1).
txt_page(859,964,1).
txt_page(508,964,1).
txt_page(754,964,1).
txt_page(186,964,1).
txt_page(138,964,1).
txt_page(520,964,1).
txt_page(52,964,1).
txt_page(498,965,1).
txt_page(621,965,1).
txt_page(808,965,1).
txt_page(82,965,1).
txt_page(53,965,1).
txt_page(28,965,1).
txt_page(517,965,1).
txt_page(549,965,1).
txt_page(759,965,1).
txt_page(135,965,1).
txt_page(454,965,1).
txt_page(540,965,1).
txt_page(43,965,1).
txt_page(380,965,1).
txt_page(299,965,1).
txt_page(432,965,1).
txt_page(822,965,1).
txt_page(872,965,1).
txt_page(160,965,1).
txt_page(357,965,1).
txt_page(712,966,1).
txt_page(355,966,1).
txt_page(272,966,1).
txt_page(877,966,1).
txt_page(743,966,1).
txt_page(863,966,1).
txt_page(50,966,1).
txt_page(553,966,1).
txt_page(71,966,1).
txt_page(426,966,1).
txt_page(786,966,1).
txt_page(847,966,1).
txt_page(701,966,1).
txt_page(116,966,1).
txt_page(957,966,1).
txt_page(501,966,1).
txt_page(157,966,1).
txt_page(321,966,1).
txt_page(826,966,1).
txt_page(121,966,1).
txt_page(315,967,1).
txt_page(588,967,1).
txt_page(195,967,1).
txt_page(215,967,1).
txt_page(332,967,1).
txt_page(578,967,1).
txt_page(541,967,1).
txt_page(120,967,1).
txt_page(759,967,1).
txt_page(907,967,1).
txt_page(224,967,1).
txt_page(542,967,1).
txt_page(200,967,2).
txt_page(847,967,1).
txt_page(101,967,1).
txt_page(671,967,1).
txt_page(618,967,1).
txt_page(470,967,1).
txt_page(88,967,1).
txt_page(193,968,1).
txt_page(522,968,1).
txt_page(343,968,1).
txt_page(757,968,1).
txt_page(576,968,1).
txt_page(467,968,1).
txt_page(56,968,1).
txt_page(448,968,1).
txt_page(783,968,1).
txt_page(415,968,1).
txt_page(933,968,1).
txt_page(664,968,1).
txt_page(118,968,1).
txt_page(26,968,1).
txt_page(932,968,1).
txt_page(555,968,1).
txt_page(121,968,1).
txt_page(90,968,1).
txt_page(551,968,1).
txt_page(778,968,1).
txt_page(551,969,1).
txt_page(564,969,1).
txt_page(941,969,1).
txt_page(668,969,1).
txt_page(819,969,1).
txt_page(511,969,1).
txt_page(633,969,1).
txt_page(821,969,1).
txt_page(93,969,1).
txt_page(559,969,1).
txt_page(14,969,1).
txt_page(962,969,1).
txt_page(427,969,1).
txt_page(262,969,1).
txt_page(619,969,1).
txt_page(490,969,1).
txt_page(912,969,1).
txt_page(263,969,1).
txt_page(139,969,1).
txt_page(333,969,1).
txt_page(80,970,1).
txt_page(525,970,1).
txt_page(523,970,1).
txt_page(804,970,1).
txt_page(862,970,1).
txt_page(466,970,1).
txt_page(222,970,1).
txt_page(104,970,1).
txt_page(476,970,1).
txt_page(90,970,1).
txt_page(790,970,1).
txt_page(760,970,1).
txt_page(351,970,1).
txt_page(384,970,1).
txt_page(946,970,1).
txt_page(436,970,1).
txt_page(58,970,1).
txt_page(832,970,1).
txt_page(666,970,1).
txt_page(340,970,1).
txt_page(484,971,1).
txt_page(35,971,1).
txt_page(674,971,1).
txt_page(208,971,1).
txt_page(145,971,1).
txt_page(66,971,1).
txt_page(620,971,1).
txt_page(906,971,1).
txt_page(849,971,1).
txt_page(733,971,1).
txt_page(664,971,1).
txt_page(785,971,1).
txt_page(820,971,1).
txt_page(685,971,1).
txt_page(547,971,1).
txt_page(279,971,1).
txt_page(239,971,1).
txt_page(228,971,1).
txt_page(247,971,1).
txt_page(294,971,1).
txt_page(765,972,1).
txt_page(417,972,1).
txt_page(731,972,1).
txt_page(520,972,1).
txt_page(858,972,1).
txt_page(732,972,1).
txt_page(141,972,1).
txt_page(278,972,1).
txt_page(884,972,1).
txt_page(758,972,1).
txt_page(16,972,1).
txt_page(443,972,1).
txt_page(595,972,1).
txt_page(259,972,1).
txt_page(553,972,1).
txt_page(893,972,1).
txt_page(54,972,1).
txt_page(324,972,1).
txt_page(227,972,1).
txt_page(574,972,1).
txt_page(547,973,1).
txt_page(551,973,1).
txt_page(191,973,1).
txt_page(967,973,1).
txt_page(442,973,1).
txt_page(702,973,1).
txt_page(58,973,1).
txt_page(353,973,1).
txt_page(114,973,1).
txt_page(214,973,1).
txt_page(85,973,1).
txt_page(704,973,1).
txt_page(59,973,1).
txt_page(824,973,1).
txt_page(815,973,1).
txt_page(897,973,1).
txt_page(181,973,1).
txt_page(246,973,1).
txt_page(92,973,1).
txt_page(496,973,1).
txt_page(428,974,1).
txt_page(404,974,1).
txt_page(487,974,1).
txt_page(55,974,1).
txt_page(236,974,1).
txt_page(485,974,1).
txt_page(198,974,1).
txt_page(855,974,1).
txt_page(852,974,1).
txt_page(847,974,1).
txt_page(817,974,1).
txt_page(85,974,1).
txt_page(13,974,1).
txt_page(430,974,1).
txt_page(54,974,1).
txt_page(315,974,1).
txt_page(790,974,1).
txt_page(313,974,1).
txt_page(658,974,1).
txt_page(448,974,1).
txt_page(762,975,1).
txt_page(758,975,1).
txt_page(251,975,1).
txt_page(757,975,1).
txt_page(422,975,1).
txt_page(266,975,1).
txt_page(268,975,1).
txt_page(270,975,1).
txt_page(679,975,1).
txt_page(533,975,1).
txt_page(124,975,1).
txt_page(216,975,1).
txt_page(641,975,1).
txt_page(347,975,1).
txt_page(769,975,1).
txt_page(812,975,1).
txt_page(99,975,1).
txt_page(482,975,1).
txt_page(143,975,1).
txt_page(715,975,1).
txt_page(150,976,1).
txt_page(866,976,1).
txt_page(799,976,1).
txt_page(41,976,1).
txt_page(353,976,1).
txt_page(599,976,1).
txt_page(808,976,1).
txt_page(309,976,1).
txt_page(252,976,1).
txt_page(958,976,1).
txt_page(649,976,1).
txt_page(197,976,1).
txt_page(96,976,1).
txt_page(165,976,1).
txt_page(632,976,1).
txt_page(614,976,1).
txt_page(967,976,1).
txt_page(251,976,1).
txt_page(665,976,1).
txt_page(704,976,1).
txt_page(733,977,1).
txt_page(255,977,1).
txt_page(269,977,1).
txt_page(813,977,1).
txt_page(298,977,1).
txt_page(535,977,1).
txt_page(871,977,1).
txt_page(872,977,1).
txt_page(18,977,1).
txt_page(109,977,1).
txt_page(808,977,1).
txt_page(666,977,1).
txt_page(720,977,1).
txt_page(56,977,1).
txt_page(81,977,1).
txt_page(953,977,1).
txt_page(213,977,1).
txt_page(325,977,1).
txt_page(675,977,1).
txt_page(957,977,1).
txt_page(752,978,1).
txt_page(9,978,1).
txt_page(665,978,1).
txt_page(625,978,1).
txt_page(283,978,1).
txt_page(277,978,1).
txt_page(42,978,1).
txt_page(762,978,1).
txt_page(584,978,1).
txt_page(763,978,1).
txt_page(297,978,1).
txt_page(758,978,1).
txt_page(329,978,1).
txt_page(221,978,1).
txt_page(637,978,1).
txt_page(414,978,1).
txt_page(148,978,1).
txt_page(591,978,1).
txt_page(853,978,1).
txt_page(686,978,1).
txt_page(36,979,1).
txt_page(340,979,1).
txt_page(387,979,1).
txt_page(86,979,1).
txt_page(560,979,1).
txt_page(119,979,1).
txt_page(422,979,1).
txt_page(699,979,1).
txt_page(466,979,1).
txt_page(959,979,1).
txt_page(326,979,1).
txt_page(217,979,1).
txt_page(48,979,1).
txt_page(185,979,1).
txt_page(952,979,1).
txt_page(97,979,1).
txt_page(957,979,1).
txt_page(184,979,1).
txt_page(792,979,2).
txt_page(135,980,1).
txt_page(846,980,1).
txt_page(139,980,1).
txt_page(607,980,1).
txt_page(608,980,1).
txt_page(628,980,1).
txt_page(601,980,1).
txt_page(291,980,1).
txt_page(401,980,1).
txt_page(68,980,1).
txt_page(148,980,1).
txt_page(862,980,1).
txt_page(884,980,1).
txt_page(431,980,1).
txt_page(886,980,1).
txt_page(445,980,1).
txt_page(91,980,1).
txt_page(616,980,1).
txt_page(199,980,1).
txt_page(662,980,1).
txt_page(198,981,1).
txt_page(391,981,1).
txt_page(932,981,1).
txt_page(707,981,1).
txt_page(818,981,1).
txt_page(756,981,1).
txt_page(441,981,1).
txt_page(42,981,1).
txt_page(795,981,1).
txt_page(827,981,1).
txt_page(356,981,1).
txt_page(757,981,1).
txt_page(748,981,1).
txt_page(452,981,1).
txt_page(613,981,1).
txt_page(609,981,1).
txt_page(775,981,1).
txt_page(594,981,1).
txt_page(200,981,1).
txt_page(403,981,1).
txt_page(709,982,1).
txt_page(90,982,1).
txt_page(878,982,1).
txt_page(379,982,1).
txt_page(485,982,1).
txt_page(404,982,1).
txt_page(550,982,1).
txt_page(382,982,1).
txt_page(850,982,1).
txt_page(189,982,1).
txt_page(541,982,1).
txt_page(708,982,1).
txt_page(953,982,1).
txt_page(693,982,1).
txt_page(817,982,1).
txt_page(1,982,1).
txt_page(215,982,1).
txt_page(68,982,1).
txt_page(52,982,1).
txt_page(611,982,1).
txt_page(204,983,1).
txt_page(656,983,1).
txt_page(635,983,1).
txt_page(740,983,1).
txt_page(603,983,1).
txt_page(194,983,1).
txt_page(270,983,1).
txt_page(565,983,1).
txt_page(596,983,1).
txt_page(488,983,1).
txt_page(386,983,1).
txt_page(47,983,1).
txt_page(236,983,1).
txt_page(102,983,1).
txt_page(127,983,1).
txt_page(456,983,1).
txt_page(523,983,1).
txt_page(638,983,1).
txt_page(241,983,1).
txt_page(866,983,1).
txt_page(17,984,1).
txt_page(755,984,1).
txt_page(79,984,1).
txt_page(920,984,1).
txt_page(339,984,1).
txt_page(408,984,1).
txt_page(874,984,1).
txt_page(800,984,1).
txt_page(76,984,1).
txt_page(188,984,1).
txt_page(53,984,1).
txt_page(295,984,1).
txt_page(758,984,1).
txt_page(235,984,1).
txt_page(452,984,1).
txt_page(100,984,1).
txt_page(399,984,1).
txt_page(416,984,1).
txt_page(759,984,1).
txt_page(170,984,1).
txt_page(243,985,1).
txt_page(748,985,1).
txt_page(655,985,1).
txt_page(247,985,1).
txt_page(219,985,1).
txt_page(853,985,1).
txt_page(740,985,1).
txt_page(46,985,1).
txt_page(322,985,1).
txt_page(9,985,1).
txt_page(72,985,1).
txt_page(578,985,1).
txt_page(646,985,1).
txt_page(626,985,1).
txt_page(83,985,1).
txt_page(881,985,1).
txt_page(662,985,1).
txt_page(24,985,1).
txt_page(827,985,1).
txt_page(807,985,1).
txt_page(513,986,1).
txt_page(607,986,1).
txt_page(681,986,1).
txt_page(968,986,1).
txt_page(903,986,1).
txt_page(817,986,1).
txt_page(799,986,1).
txt_page(744,986,1).
txt_page(102,986,1).
txt_page(164,986,1).
txt_page(152,986,1).
txt_page(895,986,1).
txt_page(582,986,1).
txt_page(364,986,1).
txt_page(600,986,1).
txt_page(806,986,1).
txt_page(547,986,2).
txt_page(428,986,1).
txt_page(861,986,1).
txt_page(789,987,1).
txt_page(808,987,1).
txt_page(326,987,1).
txt_page(415,987,1).
txt_page(301,987,1).
txt_page(282,987,1).
txt_page(159,987,1).
txt_page(865,987,1).
txt_page(151,987,1).
txt_page(801,987,1).
txt_page(388,987,1).
txt_page(547,987,1).
txt_page(597,987,1).
txt_page(831,987,1).
txt_page(688,987,1).
txt_page(433,987,1).
txt_page(31,987,1).
txt_page(242,987,1).
txt_page(89,987,1).
txt_page(386,987,1).
txt_page(257,988,1).
txt_page(925,988,1).
txt_page(372,988,1).
txt_page(740,988,1).
txt_page(750,988,1).
txt_page(915,988,1).
txt_page(329,988,1).
txt_page(763,988,1).
txt_page(386,988,1).
txt_page(136,988,1).
txt_page(363,988,1).
txt_page(155,988,1).
txt_page(513,988,1).
txt_page(601,988,1).
txt_page(360,988,1).
txt_page(154,988,1).
txt_page(229,988,1).
txt_page(946,988,1).
txt_page(464,988,1).
txt_page(591,988,1).
txt_page(181,989,1).
txt_page(334,989,1).
txt_page(947,989,1).
txt_page(570,989,1).
txt_page(817,989,1).
txt_page(742,989,1).
txt_page(112,989,1).
txt_page(76,989,1).
txt_page(755,989,1).
txt_page(229,989,2).
txt_page(486,989,1).
txt_page(944,989,1).
txt_page(467,989,1).
txt_page(866,989,1).
txt_page(810,989,1).
txt_page(501,989,1).
txt_page(431,989,1).
txt_page(368,989,1).
txt_page(272,989,1).
txt_page(819,990,1).
txt_page(803,990,1).
txt_page(801,990,1).
txt_page(776,990,1).
txt_page(953,990,1).
txt_page(541,990,1).
txt_page(829,990,1).
txt_page(355,990,1).
txt_page(695,990,1).
txt_page(318,990,1).
txt_page(277,990,1).
txt_page(836,990,1).
txt_page(888,990,1).
txt_page(823,990,1).
txt_page(532,990,1).
txt_page(144,990,1).
txt_page(60,990,1).
txt_page(182,990,1).
txt_page(16,990,1).
txt_page(24,990,1).
txt_page(196,991,1).
txt_page(354,991,1).
txt_page(764,991,1).
txt_page(304,991,1).
txt_page(872,991,1).
txt_page(272,991,1).
txt_page(750,991,1).
txt_page(170,991,1).
txt_page(358,991,1).
txt_page(704,991,1).
txt_page(32,991,1).
txt_page(99,991,1).
txt_page(627,991,1).
txt_page(920,991,1).
txt_page(690,991,1).
txt_page(265,991,1).
txt_page(878,991,1).
txt_page(621,991,1).
txt_page(37,991,1).
txt_page(307,991,1).
txt_page(389,992,1).
txt_page(481,992,1).
txt_page(197,992,1).
txt_page(690,992,1).
txt_page(203,992,1).
txt_page(527,992,1).
txt_page(919,992,1).
txt_page(817,992,1).
txt_page(158,992,1).
txt_page(164,992,1).
txt_page(637,992,1).
txt_page(688,992,1).
txt_page(770,992,1).
txt_page(499,992,1).
txt_page(930,992,1).
txt_page(795,992,1).
txt_page(897,992,1).
txt_page(161,992,1).
txt_page(748,992,1).
txt_page(598,992,1).
txt_page(716,993,1).
txt_page(560,993,1).
txt_page(114,993,1).
txt_page(600,993,1).
txt_page(872,993,1).
txt_page(673,993,1).
txt_page(781,993,1).
txt_page(131,993,1).
txt_page(873,993,1).
txt_page(738,993,1).
txt_page(366,993,1).
txt_page(244,993,1).
txt_page(392,993,1).
txt_page(372,993,1).
txt_page(690,993,1).
txt_page(569,993,1).
txt_page(639,993,1).
txt_page(838,993,1).
txt_page(273,993,1).
txt_page(824,993,1).
txt_page(122,994,1).
txt_page(405,994,1).
txt_page(376,994,1).
txt_page(537,994,1).
txt_page(296,994,1).
txt_page(652,994,1).
txt_page(430,994,1).
txt_page(421,994,1).
txt_page(485,994,1).
txt_page(610,994,1).
txt_page(846,994,1).
txt_page(321,994,1).
txt_page(188,994,1).
txt_page(482,994,1).
txt_page(579,994,1).
txt_page(238,994,1).
txt_page(925,994,1).
txt_page(362,994,1).
txt_page(66,994,1).
txt_page(944,994,1).
txt_page(961,995,1).
txt_page(306,995,1).
txt_page(127,995,1).
txt_page(565,995,1).
txt_page(595,995,1).
txt_page(423,995,1).
txt_page(388,995,1).
txt_page(864,995,1).
txt_page(131,995,1).
txt_page(405,995,1).
txt_page(844,995,1).
txt_page(168,995,1).
txt_page(851,995,1).
txt_page(424,995,1).
txt_page(940,995,1).
txt_page(248,995,1).
txt_page(644,995,1).
txt_page(460,995,1).
txt_page(379,995,1).
txt_page(809,995,1).
txt_page(944,996,1).
txt_page(95,996,1).
txt_page(823,996,1).
txt_page(810,996,1).
txt_page(182,996,1).
txt_page(133,996,1).
txt_page(233,996,1).
txt_page(378,996,1).
txt_page(140,996,1).
txt_page(555,996,1).
txt_page(559,996,1).
txt_page(189,996,1).
txt_page(543,996,1).
txt_page(782,996,1).
txt_page(333,996,1).
txt_page(485,996,1).
txt_page(675,996,1).
txt_page(495,996,1).
txt_page(428,996,1).
txt_page(549,996,1).
txt_page(316,997,1).
txt_page(28,997,1).
txt_page(247,997,1).
txt_page(415,997,1).
txt_page(493,997,1).
txt_page(833,997,1).
txt_page(749,997,1).
txt_page(183,997,1).
txt_page(548,997,1).
txt_page(968,997,1).
txt_page(60,997,1).
txt_page(457,997,1).
txt_page(880,997,1).
txt_page(498,997,1).
txt_page(284,997,1).
txt_page(242,997,1).
txt_page(62,997,1).
txt_page(101,997,1).
txt_page(557,997,1).
txt_page(198,997,1).
txt_page(85,998,1).
txt_page(756,998,1).
txt_page(561,998,1).
txt_page(127,998,1).
txt_page(180,998,1).
txt_page(928,998,1).
txt_page(760,998,1).
txt_page(477,998,1).
txt_page(308,998,1).
txt_page(287,998,1).
txt_page(789,998,1).
txt_page(130,998,1).
txt_page(798,998,1).
txt_page(426,998,1).
txt_page(709,998,1).
txt_page(654,998,1).
txt_page(280,998,1).
txt_page(455,998,1).
txt_page(292,998,1).
txt_page(377,998,1).
txt_page(171,999,1).
txt_page(311,999,1).
txt_page(633,999,1).
txt_page(212,999,1).
txt_page(800,999,1).
txt_page(717,999,1).
txt_page(623,999,1).
txt_page(237,999,1).
txt_page(806,999,1).
txt_page(638,999,1).
txt_page(640,999,1).
txt_page(590,999,1).
txt_page(75,999,1).
txt_page(903,999,1).
txt_page(150,999,1).
txt_page(29,999,1).
txt_page(841,999,1).
txt_page(798,999,1).
txt_page(282,999,1).
txt_page(261,999,1).
txt_page(58,1000,1).
txt_page(780,1000,1).
txt_page(736,1000,1).
txt_page(690,1000,1).
txt_page(204,1000,1).
txt_page(125,1000,1).
txt_page(940,1000,1).
txt_page(747,1000,1).
txt_page(45,1000,1).
txt_page(400,1000,1).
txt_page(592,1000,1).
txt_page(617,1000,1).
txt_page(495,1000,1).
txt_page(301,1000,1).
txt_page(779,1000,1).
txt_page(679,1000,1).
txt_page(896,1000,1).
txt_page(259,1000,1).
txt_page(678,1000,1).
txt_page(239,1000,1).
txt_page(194,1001,1).
txt_page(761,1001,1).
txt_page(141,1001,1).
txt_page(109,1001,1).
txt_page(47,1001,1).
txt_page(403,1001,1).
txt_page(960,1001,1).
txt_page(361,1001,1).
txt_page(155,1001,1).
txt_page(399,1001,1).
txt_page(54,1001,1).
txt_page(588,1001,1).
txt_page(537,1001,1).
txt_page(430,1001,1).
txt_page(943,1001,1).
txt_page(485,1001,1).
txt_page(653,1001,1).
txt_page(726,1001,1).
txt_page(145,1001,1).
txt_page(738,1001,1).
txt_page(348,1002,1).
txt_page(98,1002,1).
txt_page(77,1002,1).
txt_page(143,1002,1).
txt_page(905,1002,1).
txt_page(189,1002,1).
txt_page(261,1002,1).
txt_page(748,1002,1).
txt_page(248,1002,1).
txt_page(721,1002,1).
txt_page(121,1002,1).
txt_page(328,1002,1).
txt_page(380,1002,1).
txt_page(132,1002,1).
txt_page(123,1002,1).
txt_page(872,1002,1).
txt_page(609,1002,1).
txt_page(2,1002,1).
txt_page(749,1002,1).
txt_page(39,1002,1).
txt_page(656,1003,1).
txt_page(290,1003,1).
txt_page(313,1003,1).
txt_page(570,1003,1).
txt_page(314,1003,1).
txt_page(123,1003,1).
txt_page(40,1003,1).
txt_page(235,1003,1).
txt_page(622,1003,1).
txt_page(694,1003,1).
txt_page(160,1003,1).
txt_page(791,1003,1).
txt_page(86,1003,1).
txt_page(305,1003,1).
txt_page(36,1003,1).
txt_page(329,1003,1).
txt_page(309,1003,1).
txt_page(353,1003,1).
txt_page(270,1003,1).
txt_page(634,1003,1).
txt_page(817,1004,1).
txt_page(806,1004,1).
txt_page(419,1004,1).
txt_page(382,1004,1).
txt_page(351,1004,1).
txt_page(312,1004,1).
txt_page(707,1004,1).
txt_page(750,1004,1).
txt_page(218,1004,2).
txt_page(663,1004,1).
txt_page(807,1004,1).
txt_page(499,1004,1).
txt_page(739,1004,1).
txt_page(24,1004,1).
txt_page(743,1004,1).
txt_page(794,1004,1).
txt_page(947,1004,1).
txt_page(591,1004,1).
txt_page(147,1004,1).
txt_page(901,1005,1).
txt_page(388,1005,1).
txt_page(369,1005,1).
txt_page(180,1005,1).
txt_page(713,1005,1).
txt_page(951,1005,1).
txt_page(737,1005,1).
txt_page(221,1005,1).
txt_page(325,1005,1).
txt_page(914,1005,1).
txt_page(63,1005,1).
txt_page(232,1005,1).
txt_page(270,1005,1).
txt_page(834,1005,1).
txt_page(936,1005,1).
txt_page(910,1005,1).
txt_page(926,1005,1).
txt_page(311,1005,1).
txt_page(29,1005,1).
txt_page(87,1005,1).
txt_page(467,1006,1).
txt_page(686,1006,1).
txt_page(510,1006,1).
txt_page(810,1006,1).
txt_page(58,1006,1).
txt_page(880,1006,1).
txt_page(806,1006,1).
txt_page(710,1006,1).
txt_page(172,1006,1).
txt_page(80,1006,1).
txt_page(772,1006,1).
txt_page(795,1006,1).
txt_page(890,1006,1).
txt_page(455,1006,1).
txt_page(419,1006,1).
txt_page(681,1006,1).
txt_page(137,1006,1).
txt_page(157,1006,1).
txt_page(331,1006,1).
txt_page(862,1006,1).
txt_page(390,1007,1).
txt_page(589,1007,1).
txt_page(69,1007,1).
txt_page(51,1007,1).
txt_page(247,1007,1).
txt_page(785,1007,1).
txt_page(355,1007,1).
txt_page(217,1007,1).
txt_page(396,1007,1).
txt_page(639,1007,1).
txt_page(383,1007,1).
txt_page(155,1007,1).
txt_page(86,1007,1).
txt_page(891,1007,1).
txt_page(913,1007,1).
txt_page(126,1007,1).
txt_page(661,1007,1).
txt_page(292,1007,1).
txt_page(139,1007,1).
txt_page(572,1007,1).
txt_page(167,1008,1).
txt_page(793,1008,1).
txt_page(331,1008,1).
txt_page(583,1008,1).
txt_page(234,1008,1).
txt_page(821,1008,1).
txt_page(561,1008,1).
txt_page(312,1008,1).
txt_page(7,1008,1).
txt_page(692,1008,1).
txt_page(872,1008,1).
txt_page(523,1008,1).
txt_page(746,1008,1).
txt_page(149,1008,1).
txt_page(322,1008,1).
txt_page(357,1008,1).
txt_page(551,1008,1).
txt_page(106,1008,1).
txt_page(23,1008,1).
txt_page(24,1008,1).
txt_page(826,1009,1).
txt_page(275,1009,1).
txt_page(463,1009,1).
txt_page(46,1009,1).
txt_page(628,1009,1).
txt_page(854,1009,1).
txt_page(258,1009,1).
txt_page(340,1009,1).
txt_page(815,1009,1).
txt_page(589,1009,1).
txt_page(466,1009,1).
txt_page(824,1009,1).
txt_page(859,1009,1).
txt_page(241,1009,1).
txt_page(952,1009,1).
txt_page(919,1009,1).
txt_page(479,1009,1).
txt_page(783,1009,1).
txt_page(106,1009,1).
txt_page(173,1009,1).
txt_page(732,1010,1).
txt_page(156,1010,1).
txt_page(659,1010,1).
txt_page(684,1010,1).
txt_page(245,1010,1).
txt_page(319,1010,1).
txt_page(517,1010,1).
txt_page(966,1010,1).
txt_page(79,1010,1).
txt_page(721,1010,1).
txt_page(23,1010,1).
txt_page(654,1010,1).
txt_page(164,1010,1).
txt_page(188,1010,1).
txt_page(602,1010,1).
txt_page(789,1010,1).
txt_page(871,1010,1).
txt_page(297,1010,1).
txt_page(456,1010,1).
txt_page(583,1010,1).
txt_page(72,1011,1).
txt_page(710,1011,1).
txt_page(809,1011,1).
txt_page(566,1011,1).
txt_page(69,1011,1).
txt_page(578,1011,1).
txt_page(240,1011,1).
txt_page(458,1011,1).
txt_page(672,1011,1).
txt_page(285,1011,1).
txt_page(680,1011,1).
txt_page(213,1011,1).
txt_page(229,1011,1).
txt_page(477,1011,1).
txt_page(963,1011,1).
txt_page(99,1011,1).
txt_page(400,1011,1).
txt_page(236,1011,1).
txt_page(218,1011,2).
txt_page(122,1012,1).
txt_page(184,1012,1).
txt_page(76,1012,1).
txt_page(778,1012,1).
txt_page(750,1012,1).
txt_page(209,1012,1).
txt_page(595,1012,1).
txt_page(760,1012,1).
txt_page(289,1012,1).
txt_page(672,1012,1).
txt_page(722,1012,1).
txt_page(195,1012,1).
txt_page(365,1012,1).
txt_page(939,1012,1).
txt_page(403,1012,1).
txt_page(811,1012,1).
txt_page(276,1012,1).
txt_page(633,1012,1).
txt_page(529,1012,1).
txt_page(588,1012,1).
txt_page(399,1013,1).
txt_page(743,1013,1).
txt_page(386,1013,1).
txt_page(245,1013,1).
txt_page(642,1013,1).
txt_page(618,1013,1).
txt_page(31,1013,1).
txt_page(928,1013,1).
txt_page(513,1013,1).
txt_page(51,1013,1).
txt_page(796,1013,1).
txt_page(398,1013,1).
txt_page(851,1013,1).
txt_page(636,1013,1).
txt_page(914,1013,1).
txt_page(298,1013,1).
txt_page(792,1013,1).
txt_page(95,1013,1).
txt_page(163,1013,1).
txt_page(620,1013,1).
txt_page(632,1014,1).
txt_page(822,1014,1).
txt_page(166,1014,1).
txt_page(768,1014,1).
txt_page(475,1014,1).
txt_page(622,1014,1).
txt_page(330,1014,1).
txt_page(334,1014,1).
txt_page(537,1014,1).
txt_page(599,1014,1).
txt_page(767,1014,1).
txt_page(126,1014,1).
txt_page(741,1014,1).
txt_page(664,1014,1).
txt_page(223,1014,1).
txt_page(772,1014,1).
txt_page(617,1014,1).
txt_page(441,1014,1).
txt_page(533,1014,1).
txt_page(543,1014,1).
txt_page(361,1015,1).
txt_page(821,1015,1).
txt_page(491,1015,1).
txt_page(733,1015,1).
txt_page(187,1015,1).
txt_page(802,1015,1).
txt_page(44,1015,1).
txt_page(145,1015,2).
txt_page(474,1015,1).
txt_page(173,1015,1).
txt_page(775,1015,1).
txt_page(800,1015,1).
txt_page(348,1015,1).
txt_page(18,1015,1).
txt_page(285,1015,1).
txt_page(556,1015,1).
txt_page(743,1015,1).
txt_page(722,1015,1).
txt_page(450,1015,1).
txt_page(967,1016,1).
txt_page(487,1016,1).
txt_page(221,1016,1).
txt_page(144,1016,1).
txt_page(955,1016,1).
txt_page(109,1016,1).
txt_page(732,1016,1).
txt_page(112,1016,1).
txt_page(158,1016,1).
txt_page(775,1016,1).
txt_page(389,1016,1).
txt_page(713,1016,1).
txt_page(390,1016,1).
txt_page(189,1016,1).
txt_page(171,1016,1).
txt_page(700,1016,1).
txt_page(364,1016,1).
txt_page(821,1016,1).
txt_page(282,1016,1).
txt_page(726,1016,1).
txt_page(537,1017,1).
txt_page(294,1017,1).
txt_page(295,1017,1).
txt_page(567,1017,1).
txt_page(28,1017,1).
txt_page(483,1017,1).
txt_page(153,1017,1).
txt_page(185,1017,1).
txt_page(749,1017,1).
txt_page(196,1017,1).
txt_page(794,1017,1).
txt_page(398,1017,1).
txt_page(932,1017,1).
txt_page(949,1017,1).
txt_page(735,1017,1).
txt_page(916,1017,1).
txt_page(817,1017,1).
txt_page(919,1017,1).
txt_page(243,1017,1).
txt_page(48,1017,1).
txt_page(731,1018,1).
txt_page(124,1018,1).
txt_page(910,1018,1).
txt_page(539,1018,1).
txt_page(753,1018,1).
txt_page(858,1018,1).
txt_page(274,1018,1).
txt_page(591,1018,1).
txt_page(419,1018,1).
txt_page(289,1018,1).
txt_page(121,1018,1).
txt_page(643,1018,1).
txt_page(737,1018,1).
txt_page(926,1018,1).
txt_page(320,1018,1).
txt_page(649,1018,1).
txt_page(180,1018,1).
txt_page(79,1018,1).
txt_page(78,1018,1).
txt_page(669,1018,1).
txt_page(176,1019,1).
txt_page(788,1019,1).
txt_page(259,1019,1).
txt_page(728,1019,1).
txt_page(906,1019,1).
txt_page(150,1019,1).
txt_page(598,1019,1).
txt_page(422,1019,1).
txt_page(684,1019,1).
txt_page(498,1019,1).
txt_page(273,1019,1).
txt_page(740,1019,1).
txt_page(719,1019,2).
txt_page(277,1019,1).
txt_page(462,1019,1).
txt_page(138,1019,1).
txt_page(926,1019,1).
txt_page(452,1019,1).
txt_page(780,1019,1).
txt_page(260,1020,1).
txt_page(899,1020,1).
txt_page(702,1020,1).
txt_page(393,1020,1).
txt_page(693,1020,1).
txt_page(946,1020,1).
txt_page(818,1020,1).
txt_page(308,1020,1).
txt_page(866,1020,1).
txt_page(733,1020,1).
txt_page(32,1020,1).
txt_page(203,1020,1).
txt_page(683,1020,1).
txt_page(134,1020,1).
txt_page(675,1020,1).
txt_page(158,1020,1).
txt_page(166,1020,1).
txt_page(401,1020,1).
txt_page(405,1020,1).
txt_page(314,1020,1).
txt_page(559,1021,1).
txt_page(475,1021,1).
txt_page(610,1021,1).
txt_page(147,1021,1).
txt_page(546,1021,1).
txt_page(2,1021,1).
txt_page(446,1021,1).
txt_page(195,1021,1).
txt_page(845,1021,1).
txt_page(281,1021,1).
txt_page(332,1021,1).
txt_page(180,1021,1).
txt_page(327,1021,1).
txt_page(361,1021,1).
txt_page(445,1021,1).
txt_page(526,1021,1).
txt_page(605,1021,1).
txt_page(494,1021,1).
txt_page(689,1021,1).
txt_page(870,1021,1).
txt_page(444,1022,1).
txt_page(237,1022,1).
txt_page(742,1022,1).
txt_page(694,1022,1).
txt_page(299,1022,1).
txt_page(49,1022,1).
txt_page(422,1022,1).
txt_page(579,1022,1).
txt_page(8,1022,1).
txt_page(342,1022,1).
txt_page(437,1022,1).
txt_page(232,1022,1).
txt_page(578,1022,1).
txt_page(128,1022,1).
txt_page(745,1022,1).
txt_page(7,1022,1).
txt_page(63,1022,1).
txt_page(685,1022,1).
txt_page(362,1022,1).
txt_page(611,1022,1).
txt_page(930,1023,1).
txt_page(462,1023,1).
txt_page(938,1023,1).
txt_page(166,1023,1).
txt_page(256,1023,1).
txt_page(111,1023,1).
txt_page(493,1023,1).
txt_page(393,1023,1).
txt_page(265,1023,1).
txt_page(184,1023,1).
txt_page(816,1023,1).
txt_page(193,1023,1).
txt_page(488,1023,1).
txt_page(6,1023,1).
txt_page(388,1023,1).
txt_page(411,1023,1).
txt_page(162,1023,1).
txt_page(908,1023,1).
txt_page(255,1023,1).
txt_page(361,1023,1).
txt_page(379,1024,1).
txt_page(257,1024,1).
txt_page(689,1024,1).
txt_page(325,1024,1).
txt_page(765,1024,1).
txt_page(905,1024,1).
txt_page(544,1024,1).
txt_page(566,1024,1).
txt_page(58,1024,1).
txt_page(186,1024,1).
txt_page(177,1024,1).
txt_page(400,1024,1).
txt_page(45,1024,1).
txt_page(414,1024,1).
txt_page(464,1024,1).
txt_page(928,1024,1).
txt_page(876,1024,1).
txt_page(920,1024,1).
txt_page(688,1024,1).
txt_page(593,1024,1).
txt_page(103,1025,1).
txt_page(862,1025,1).
txt_page(746,1025,1).
txt_page(8,1025,1).
txt_page(660,1025,1).
txt_page(839,1025,1).
txt_page(867,1025,1).
txt_page(311,1025,1).
txt_page(200,1025,1).
txt_page(851,1025,1).
txt_page(568,1025,1).
txt_page(631,1025,1).
txt_page(117,1025,1).
txt_page(676,1025,1).
txt_page(162,1025,1).
txt_page(844,1025,1).
txt_page(285,1025,1).
txt_page(539,1025,1).
txt_page(486,1025,1).
txt_page(910,1025,1).
txt_page(872,1026,1).
txt_page(573,1026,1).
txt_page(495,1026,1).
txt_page(574,1026,1).
txt_page(96,1026,1).
txt_page(227,1026,1).
txt_page(394,1026,1).
txt_page(386,1026,1).
txt_page(367,1026,1).
txt_page(739,1026,1).
txt_page(852,1026,1).
txt_page(791,1026,1).
txt_page(564,1026,1).
txt_page(833,1026,1).
txt_page(511,1026,1).
txt_page(483,1026,1).
txt_page(916,1026,1).
txt_page(70,1026,1).
txt_page(623,1026,1).
txt_page(58,1026,1).
txt_page(523,1027,1).
txt_page(308,1027,1).
txt_page(506,1027,1).
txt_page(823,1027,1).
txt_page(272,1027,1).
txt_page(556,1027,1).
txt_page(748,1027,1).
txt_page(252,1027,1).
txt_page(309,1027,1).
txt_page(504,1027,1).
txt_page(886,1027,1).
txt_page(382,1027,1).
txt_page(10,1027,1).
txt_page(792,1027,1).
txt_page(28,1027,1).
txt_page(827,1027,1).
txt_page(225,1027,1).
txt_page(770,1027,1).
txt_page(164,1027,1).
txt_page(39,1027,1).
txt_page(534,1028,1).
txt_page(114,1028,1).
txt_page(123,1028,1).
txt_page(180,1028,1).
txt_page(287,1028,1).
txt_page(712,1028,1).
txt_page(310,1028,1).
txt_page(856,1028,1).
txt_page(894,1028,1).
txt_page(280,1028,1).
txt_page(647,1028,1).
txt_page(664,1028,1).
txt_page(290,1028,1).
txt_page(730,1028,1).
txt_page(403,1028,1).
txt_page(427,1028,1).
txt_page(387,1028,1).
txt_page(368,1028,1).
txt_page(683,1028,1).
txt_page(84,1028,1).
txt_page(89,1029,1).
txt_page(824,1029,1).
txt_page(936,1029,1).
txt_page(32,1029,1).
txt_page(701,1029,1).
txt_page(687,1029,1).
txt_page(875,1029,1).
txt_page(169,1029,1).
txt_page(667,1029,1).
txt_page(231,1029,1).
txt_page(806,1029,1).
txt_page(460,1029,1).
txt_page(162,1029,1).
txt_page(637,1029,1).
txt_page(88,1029,1).
txt_page(361,1029,1).
txt_page(897,1029,1).
txt_page(489,1029,1).
txt_page(841,1029,2).
txt_page(824,1030,1).
txt_page(48,1030,1).
txt_page(912,1030,1).
txt_page(699,1030,1).
txt_page(348,1030,1).
txt_page(899,1030,1).
txt_page(280,1030,1).
txt_page(663,1030,1).
txt_page(776,1030,1).
txt_page(894,1030,1).
txt_page(347,1030,1).
txt_page(588,1030,1).
txt_page(312,1030,1).
txt_page(333,1030,1).
txt_page(226,1030,1).
txt_page(220,1030,1).
txt_page(932,1030,1).
txt_page(860,1030,1).
txt_page(585,1030,1).
txt_page(224,1030,1).
txt_page(849,1031,1).
txt_page(770,1031,1).
txt_page(468,1031,1).
txt_page(959,1031,1).
txt_page(277,1031,1).
txt_page(370,1031,1).
txt_page(548,1031,1).
txt_page(513,1031,1).
txt_page(608,1031,1).
txt_page(862,1031,1).
txt_page(353,1031,1).
txt_page(363,1031,1).
txt_page(589,1031,1).
txt_page(125,1031,1).
txt_page(301,1031,1).
txt_page(87,1031,1).
txt_page(444,1031,1).
txt_page(16,1031,1).
txt_page(676,1031,1).
txt_page(724,1031,1).
txt_page(651,1032,1).
txt_page(147,1032,1).
txt_page(539,1032,1).
txt_page(208,1032,1).
txt_page(668,1032,1).
txt_page(928,1032,1).
txt_page(303,1032,1).
txt_page(676,1032,1).
txt_page(236,1032,1).
txt_page(771,1032,1).
txt_page(559,1032,1).
txt_page(486,1032,1).
txt_page(743,1032,1).
txt_page(182,1032,1).
txt_page(183,1032,1).
txt_page(500,1032,1).
txt_page(58,1032,1).
txt_page(571,1032,1).
txt_page(109,1032,1).
txt_page(744,1032,1).
txt_page(1,1033,1).
txt_page(754,1033,1).
txt_page(185,1033,1).
txt_page(60,1033,1).
txt_page(319,1033,1).
txt_page(363,1033,1).
txt_page(524,1033,1).
txt_page(375,1033,1).
txt_page(207,1033,1).
txt_page(312,1033,1).
txt_page(523,1033,1).
txt_page(732,1033,1).
txt_page(321,1033,1).
txt_page(527,1033,1).
txt_page(506,1033,1).
txt_page(366,1033,1).
txt_page(816,1033,1).
txt_page(822,1033,1).
txt_page(963,1033,1).
txt_page(592,1033,1).
txt_page(495,1034,1).
txt_page(418,1034,1).
txt_page(274,1034,1).
txt_page(815,1034,1).
txt_page(149,1034,1).
txt_page(735,1034,1).
txt_page(241,1034,1).
txt_page(855,1034,1).
txt_page(122,1034,1).
txt_page(811,1034,1).
txt_page(957,1034,1).
txt_page(678,1034,1).
txt_page(731,1034,1).
txt_page(933,1034,1).
txt_page(155,1034,1).
txt_page(124,1034,1).
txt_page(88,1034,1).
txt_page(456,1034,1).
txt_page(939,1034,1).
txt_page(13,1034,1).
txt_page(514,1035,1).
txt_page(949,1035,1).
txt_page(253,1035,1).
txt_page(227,1035,1).
txt_page(198,1035,1).
txt_page(544,1035,1).
txt_page(68,1035,1).
txt_page(538,1035,1).
txt_page(411,1035,1).
txt_page(495,1035,1).
txt_page(458,1035,1).
txt_page(206,1035,1).
txt_page(305,1035,1).
txt_page(261,1035,1).
txt_page(85,1035,1).
txt_page(721,1035,1).
txt_page(254,1035,1).
txt_page(111,1035,1).
txt_page(39,1035,1).
txt_page(368,1035,1).
txt_page(320,1036,1).
txt_page(543,1036,1).
txt_page(75,1036,1).
txt_page(424,1036,1).
txt_page(917,1036,1).
txt_page(913,1036,1).
txt_page(496,1036,1).
txt_page(870,1036,1).
txt_page(180,1036,1).
txt_page(139,1036,1).
txt_page(267,1036,1).
txt_page(614,1036,1).
txt_page(877,1036,1).
txt_page(155,1036,1).
txt_page(11,1036,1).
txt_page(464,1036,1).
txt_page(15,1036,1).
txt_page(649,1036,1).
txt_page(529,1036,1).
txt_page(565,1036,1).
txt_page(596,1037,1).
txt_page(555,1037,1).
txt_page(112,1037,1).
txt_page(474,1037,1).
txt_page(781,1037,1).
txt_page(854,1037,1).
txt_page(521,1037,1).
txt_page(361,1037,1).
txt_page(417,1037,1).
txt_page(212,1037,1).
txt_page(446,1037,1).
txt_page(597,1037,1).
txt_page(743,1037,1).
txt_page(564,1037,1).
txt_page(187,1037,1).
txt_page(599,1037,1).
txt_page(649,1037,1).
txt_page(125,1037,1).
txt_page(518,1037,1).
txt_page(672,1037,1).
txt_page(467,1038,1).
txt_page(304,1038,1).
txt_page(831,1038,1).
txt_page(489,1038,1).
txt_page(119,1038,1).
txt_page(264,1038,1).
txt_page(204,1038,1).
txt_page(309,1038,1).
txt_page(652,1038,1).
txt_page(413,1038,1).
txt_page(2,1038,1).
txt_page(25,1038,1).
txt_page(673,1038,1).
txt_page(679,1038,1).
txt_page(276,1038,1).
txt_page(366,1038,1).
txt_page(74,1038,1).
txt_page(716,1038,1).
txt_page(459,1038,1).
txt_page(949,1038,1).
txt_page(909,1039,1).
txt_page(277,1039,1).
txt_page(642,1039,1).
txt_page(661,1039,1).
txt_page(446,1039,1).
txt_page(399,1039,1).
txt_page(306,1039,1).
txt_page(907,1039,1).
txt_page(543,1039,1).
txt_page(8,1039,1).
txt_page(816,1039,1).
txt_page(476,1039,1).
txt_page(346,1039,1).
txt_page(663,1039,1).
txt_page(326,1039,1).
txt_page(127,1039,1).
txt_page(822,1039,1).
txt_page(658,1039,1).
txt_page(46,1039,1).
txt_page(781,1039,1).
txt_page(363,1040,1).
txt_page(129,1040,1).
txt_page(106,1040,1).
txt_page(13,1040,1).
txt_page(121,1040,1).
txt_page(137,1040,1).
txt_page(681,1040,1).
txt_page(967,1040,1).
txt_page(310,1040,1).
txt_page(692,1040,1).
txt_page(511,1040,1).
txt_page(917,1040,1).
txt_page(954,1040,1).
txt_page(424,1040,1).
txt_page(910,1040,1).
txt_page(465,1040,1).
txt_page(488,1040,1).
txt_page(368,1040,1).
txt_page(783,1040,1).
txt_page(388,1040,1).
txt_page(77,1041,1).
txt_page(180,1041,1).
txt_page(443,1041,1).
txt_page(838,1041,1).
txt_page(700,1041,1).
txt_page(263,1041,1).
txt_page(632,1041,1).
txt_page(762,1041,1).
txt_page(601,1041,1).
txt_page(402,1041,1).
txt_page(390,1041,1).
txt_page(294,1041,1).
txt_page(394,1041,1).
txt_page(541,1041,1).
txt_page(304,1041,1).
txt_page(40,1041,1).
txt_page(629,1041,1).
txt_page(381,1041,1).
txt_page(849,1041,1).
txt_page(415,1041,1).
txt_page(325,1042,1).
txt_page(742,1042,1).
txt_page(455,1042,1).
txt_page(869,1042,1).
txt_page(302,1042,1).
txt_page(583,1042,1).
txt_page(298,1042,1).
txt_page(676,1042,1).
txt_page(151,1042,1).
txt_page(67,1042,1).
txt_page(126,1042,1).
txt_page(781,1042,1).
txt_page(56,1042,1).
txt_page(783,1042,1).
txt_page(191,1042,1).
txt_page(690,1042,1).
txt_page(602,1042,1).
txt_page(368,1042,1).
txt_page(286,1042,1).
txt_page(360,1042,1).
txt_page(702,1043,1).
txt_page(387,1043,1).
txt_page(676,1043,1).
txt_page(467,1043,1).
txt_page(157,1043,1).
txt_page(216,1043,1).
txt_page(700,1043,1).
txt_page(890,1043,1).
txt_page(490,1043,1).
txt_page(62,1043,1).
txt_page(491,1043,1).
txt_page(829,1043,1).
txt_page(315,1043,1).
txt_page(898,1043,1).
txt_page(412,1043,1).
txt_page(938,1043,1).
txt_page(173,1043,1).
txt_page(240,1043,2).
txt_page(935,1043,1).
txt_page(383,1044,1).
txt_page(559,1044,1).
txt_page(513,1044,1).
txt_page(908,1044,1).
txt_page(19,1044,1).
txt_page(834,1044,1).
txt_page(825,1044,1).
txt_page(262,1044,1).
txt_page(885,1044,1).
txt_page(497,1044,1).
txt_page(132,1044,1).
txt_page(116,1044,1).
txt_page(460,1044,1).
txt_page(646,1044,1).
txt_page(165,1044,1).
txt_page(721,1044,1).
txt_page(832,1044,1).
txt_page(726,1044,1).
txt_page(227,1044,1).
txt_page(643,1044,1).
txt_page(121,1045,1).
txt_page(486,1045,1).
txt_page(288,1045,1).
txt_page(178,1045,1).
txt_page(204,1045,1).
txt_page(838,1045,1).
txt_page(932,1045,1).
txt_page(883,1045,1).
txt_page(703,1045,1).
txt_page(916,1045,1).
txt_page(133,1045,1).
txt_page(631,1045,1).
txt_page(422,1045,1).
txt_page(87,1045,1).
txt_page(224,1045,1).
txt_page(961,1045,1).
txt_page(742,1045,1).
txt_page(826,1045,1).
txt_page(654,1045,1).
txt_page(882,1045,1).
txt_page(526,1046,1).
txt_page(712,1046,1).
txt_page(361,1046,1).
txt_page(637,1046,1).
txt_page(372,1046,1).
txt_page(86,1046,1).
txt_page(68,1046,1).
txt_page(520,1046,1).
txt_page(453,1046,1).
txt_page(761,1046,1).
txt_page(390,1046,1).
txt_page(902,1046,1).
txt_page(160,1046,1).
txt_page(428,1046,1).
txt_page(926,1046,1).
txt_page(863,1046,1).
txt_page(790,1046,1).
txt_page(163,1046,1).
txt_page(11,1046,1).
txt_page(917,1046,1).
txt_page(607,1047,1).
txt_page(709,1047,1).
txt_page(819,1047,1).
txt_page(806,1047,1).
txt_page(665,1047,1).
txt_page(259,1047,1).
txt_page(590,1047,1).
txt_page(699,1047,1).
txt_page(712,1047,1).
txt_page(357,1047,1).
txt_page(645,1047,1).
txt_page(217,1047,1).
txt_page(856,1047,1).
txt_page(933,1047,1).
txt_page(92,1047,1).
txt_page(879,1047,1).
txt_page(703,1047,1).
txt_page(105,1047,1).
txt_page(282,1047,1).
txt_page(84,1047,1).
txt_page(335,1048,1).
txt_page(958,1048,1).
txt_page(530,1048,1).
txt_page(965,1048,1).
txt_page(797,1048,1).
txt_page(582,1048,1).
txt_page(379,1048,1).
txt_page(448,1048,1).
txt_page(11,1048,1).
txt_page(750,1048,1).
txt_page(674,1048,1).
txt_page(814,1048,1).
txt_page(185,1048,1).
txt_page(323,1048,1).
txt_page(14,1048,1).
txt_page(885,1048,1).
txt_page(664,1048,1).
txt_page(927,1048,1).
txt_page(823,1048,1).
txt_page(779,1048,1).
txt_page(607,1049,1).
txt_page(9,1049,1).
txt_page(149,1049,1).
txt_page(859,1049,1).
txt_page(653,1049,1).
txt_page(683,1049,1).
txt_page(926,1049,1).
txt_page(327,1049,1).
txt_page(1,1049,1).
txt_page(943,1049,1).
txt_page(893,1049,1).
txt_page(42,1049,1).
txt_page(76,1049,1).
txt_page(561,1049,1).
txt_page(575,1049,1).
txt_page(272,1049,1).
txt_page(448,1049,1).
txt_page(281,1049,1).
txt_page(92,1049,1).
txt_page(71,1049,1).
txt_page(685,1050,1).
txt_page(133,1050,1).
txt_page(955,1050,1).
txt_page(713,1050,1).
txt_page(734,1050,1).
txt_page(184,1050,1).
txt_page(490,1050,1).
txt_page(406,1050,1).
txt_page(461,1050,1).
txt_page(686,1050,1).
txt_page(580,1050,1).
txt_page(901,1050,1).
txt_page(774,1050,1).
txt_page(202,1050,1).
txt_page(183,1050,1).
txt_page(790,1050,1).
txt_page(21,1050,1).
txt_page(722,1050,1).
txt_page(450,1050,1).
txt_page(61,1050,1).
txt_page(759,1051,1).
txt_page(948,1051,1).
txt_page(213,1051,1).
txt_page(516,1051,1).
txt_page(25,1051,1).
txt_page(922,1051,1).
txt_page(151,1051,1).
txt_page(789,1051,1).
txt_page(728,1051,1).
txt_page(770,1051,1).
txt_page(902,1051,1).
txt_page(710,1051,1).
txt_page(372,1051,1).
txt_page(923,1051,1).
txt_page(228,1051,1).
txt_page(797,1051,1).
txt_page(461,1051,1).
txt_page(303,1051,1).
txt_page(519,1051,1).
txt_page(420,1051,1).
txt_page(919,1052,1).
txt_page(598,1052,1).
txt_page(99,1052,1).
txt_page(115,1052,1).
txt_page(908,1052,1).
txt_page(414,1052,1).
txt_page(220,1052,1).
txt_page(105,1052,1).
txt_page(5,1052,1).
txt_page(252,1052,1).
txt_page(367,1052,1).
txt_page(281,1052,1).
txt_page(461,1052,1).
txt_page(197,1052,1).
txt_page(501,1052,1).
txt_page(113,1052,1).
txt_page(77,1052,1).
txt_page(844,1052,1).
txt_page(344,1052,1).
txt_page(830,1052,1).
txt_page(870,1053,1).
txt_page(920,1053,1).
txt_page(879,1053,1).
txt_page(772,1053,1).
txt_page(881,1053,1).
txt_page(194,1053,1).
txt_page(555,1053,1).
txt_page(138,1053,1).
txt_page(423,1053,1).
txt_page(938,1053,1).
txt_page(943,1053,1).
txt_page(203,1053,1).
txt_page(617,1053,1).
txt_page(951,1053,1).
txt_page(322,1053,1).
txt_page(726,1053,1).
txt_page(470,1053,1).
txt_page(223,1053,1).
txt_page(599,1053,1).
txt_page(252,1053,1).
txt_page(335,1054,1).
txt_page(855,1054,1).
txt_page(703,1054,1).
txt_page(594,1054,1).
txt_page(438,1054,1).
txt_page(254,1054,1).
txt_page(534,1054,1).
txt_page(166,1054,1).
txt_page(123,1054,1).
txt_page(694,1054,1).
txt_page(850,1054,1).
txt_page(456,1054,1).
txt_page(879,1054,1).
txt_page(812,1054,1).
txt_page(2,1054,1).
txt_page(25,1054,1).
txt_page(675,1054,1).
txt_page(339,1054,1).
txt_page(164,1054,1).
txt_page(480,1054,1).
txt_page(437,1055,1).
txt_page(595,1055,1).
txt_page(313,1055,1).
txt_page(418,1055,1).
txt_page(454,1055,1).
txt_page(263,1055,1).
txt_page(584,1055,1).
txt_page(801,1055,1).
txt_page(341,1055,1).
txt_page(708,1055,1).
txt_page(388,1055,1).
txt_page(175,1055,1).
txt_page(645,1055,1).
txt_page(202,1055,1).
txt_page(900,1055,1).
txt_page(637,1055,1).
txt_page(89,1055,1).
txt_page(538,1055,1).
txt_page(696,1055,1).
txt_page(159,1055,1).
txt_page(694,1056,1).
txt_page(290,1056,1).
txt_page(752,1056,1).
txt_page(565,1056,1).
txt_page(350,1056,1).
txt_page(119,1056,1).
txt_page(522,1056,1).
txt_page(647,1056,1).
txt_page(666,1056,1).
txt_page(276,1056,1).
txt_page(819,1056,1).
txt_page(913,1056,1).
txt_page(944,1056,1).
txt_page(567,1056,1).
txt_page(850,1056,1).
txt_page(622,1056,1).
txt_page(951,1056,1).
txt_page(886,1056,1).
txt_page(166,1056,1).
txt_page(926,1056,1).
txt_page(376,1057,1).
txt_page(905,1057,1).
txt_page(853,1057,1).
txt_page(470,1057,1).
txt_page(806,1057,1).
txt_page(232,1057,1).
txt_page(530,1057,1).
txt_page(549,1057,1).
txt_page(208,1057,1).
txt_page(508,1057,1).
txt_page(731,1057,1).
txt_page(343,1057,1).
txt_page(423,1057,1).
txt_page(725,1057,1).
txt_page(482,1057,1).
txt_page(21,1057,1).
txt_page(204,1057,1).
txt_page(487,1057,1).
txt_page(603,1057,1).
txt_page(86,1057,1).
txt_page(654,1058,1).
txt_page(390,1058,1).
txt_page(553,1058,1).
txt_page(399,1058,1).
txt_page(653,1058,1).
txt_page(45,1058,1).
txt_page(536,1058,1).
txt_page(322,1058,1).
txt_page(470,1058,1).
txt_page(398,1058,1).
txt_page(320,1058,1).
txt_page(532,1058,1).
txt_page(122,1058,1).
txt_page(424,1058,1).
txt_page(518,1058,1).
txt_page(552,1058,1).
txt_page(847,1058,1).
txt_page(262,1058,1).
txt_page(126,1058,1).
txt_page(812,1058,1).
txt_page(398,1059,1).
txt_page(394,1059,1).
txt_page(330,1059,1).
txt_page(375,1059,1).
txt_page(482,1059,1).
txt_page(933,1059,1).
txt_page(641,1059,1).
txt_page(7,1059,1).
txt_page(95,1059,1).
txt_page(779,1059,1).
txt_page(103,1059,1).
txt_page(413,1059,1).
txt_page(437,1059,1).
txt_page(396,1059,1).
txt_page(296,1059,1).
txt_page(860,1059,1).
txt_page(134,1059,1).
txt_page(659,1059,1).
txt_page(864,1059,1).
txt_page(173,1059,1).
txt_page(666,1060,1).
txt_page(234,1060,1).
txt_page(475,1060,1).
txt_page(725,1060,1).
txt_page(287,1060,1).
txt_page(28,1060,1).
txt_page(670,1060,1).
txt_page(952,1060,1).
txt_page(961,1060,1).
txt_page(697,1060,1).
txt_page(662,1060,1).
txt_page(15,1060,1).
txt_page(732,1060,1).
txt_page(3,1060,1).
txt_page(133,1060,1).
txt_page(310,1060,1).
txt_page(24,1060,1).
txt_page(122,1060,1).
txt_page(69,1060,1).
txt_page(614,1060,1).
txt_page(665,1061,1).
txt_page(189,1061,1).
txt_page(611,1061,1).
txt_page(766,1061,1).
txt_page(282,1061,1).
txt_page(296,1061,1).
txt_page(686,1061,1).
txt_page(697,1061,1).
txt_page(229,1061,1).
txt_page(628,1061,1).
txt_page(725,1061,1).
txt_page(929,1061,1).
txt_page(267,1061,1).
txt_page(111,1061,1).
txt_page(718,1061,1).
txt_page(563,1061,1).
txt_page(592,1061,1).
txt_page(625,1061,1).
txt_page(517,1061,1).
txt_page(550,1061,1).
txt_page(134,1062,1).
txt_page(384,1062,1).
txt_page(247,1062,1).
txt_page(691,1062,1).
txt_page(420,1062,1).
txt_page(909,1062,1).
txt_page(720,1062,1).
txt_page(516,1062,1).
txt_page(534,1062,1).
txt_page(762,1062,1).
txt_page(442,1062,1).
txt_page(792,1062,1).
txt_page(507,1062,1).
txt_page(29,1062,1).
txt_page(195,1062,1).
txt_page(473,1062,1).
txt_page(260,1062,1).
txt_page(66,1062,1).
txt_page(213,1062,1).
txt_page(486,1062,1).
txt_page(595,1063,1).
txt_page(283,1063,1).
txt_page(142,1063,1).
txt_page(594,1063,1).
txt_page(927,1063,1).
txt_page(16,1063,1).
txt_page(405,1063,1).
txt_page(797,1063,1).
txt_page(305,1063,1).
txt_page(226,1063,1).
txt_page(853,1063,1).
txt_page(873,1063,1).
txt_page(669,1063,1).
txt_page(691,1063,1).
txt_page(117,1063,1).
txt_page(487,1063,1).
txt_page(720,1063,1).
txt_page(910,1063,1).
txt_page(75,1063,1).
txt_page(70,1063,1).
txt_page(222,1064,1).
txt_page(16,1064,1).
txt_page(457,1064,1).
txt_page(209,1064,1).
txt_page(877,1064,1).
txt_page(800,1064,1).
txt_page(790,1064,1).
txt_page(763,1064,1).
txt_page(873,1064,1).
txt_page(497,1064,1).
txt_page(963,1064,1).
txt_page(77,1064,1).
txt_page(821,1064,1).
txt_page(90,1064,1).
txt_page(787,1064,1).
txt_page(939,1064,1).
txt_page(438,1064,1).
txt_page(564,1064,1).
txt_page(131,1064,1).
txt_page(98,1064,1).
txt_page(9,1065,1).
txt_page(931,1065,1).
txt_page(283,1065,1).
txt_page(498,1065,1).
txt_page(680,1065,1).
txt_page(329,1065,1).
txt_page(97,1065,1).
txt_page(132,1065,1).
txt_page(3,1065,1).
txt_page(582,1065,1).
txt_page(665,1065,1).
txt_page(862,1065,1).
txt_page(227,1065,1).
txt_page(113,1065,1).
txt_page(404,1065,1).
txt_page(598,1065,1).
txt_page(273,1065,1).
txt_page(837,1065,1).
txt_page(629,1065,1).
txt_page(699,1065,1).
txt_page(611,1066,1).
txt_page(836,1066,1).
txt_page(198,1066,1).
txt_page(123,1066,1).
txt_page(703,1066,1).
txt_page(373,1066,1).
txt_page(859,1066,1).
txt_page(617,1066,1).
txt_page(917,1066,1).
txt_page(275,1066,1).
txt_page(800,1066,1).
txt_page(681,1066,1).
txt_page(233,1066,1).
txt_page(326,1066,1).
txt_page(924,1066,1).
txt_page(379,1066,1).
txt_page(222,1066,1).
txt_page(768,1066,1).
txt_page(455,1066,1).
txt_page(608,1066,1).
txt_page(827,1067,1).
txt_page(165,1067,1).
txt_page(473,1067,1).
txt_page(488,1067,1).
txt_page(861,1067,1).
txt_page(394,1067,1).
txt_page(194,1067,1).
txt_page(304,1067,1).
txt_page(641,1067,1).
txt_page(897,1067,1).
txt_page(281,1067,1).
txt_page(408,1067,1).
txt_page(614,1067,1).
txt_page(482,1067,1).
txt_page(197,1067,1).
txt_page(815,1067,1).
txt_page(180,1067,1).
txt_page(316,1067,1).
txt_page(302,1067,1).
txt_page(528,1067,1).
txt_page(487,1068,1).
txt_page(901,1068,1).
txt_page(255,1068,1).
txt_page(603,1068,1).
txt_page(352,1068,1).
txt_page(661,1068,1).
txt_page(855,1068,1).
txt_page(259,1068,1).
txt_page(243,1068,1).
txt_page(462,1068,1).
txt_page(4,1068,1).
txt_page(648,1068,1).
txt_page(437,1068,1).
txt_page(799,1068,1).
txt_page(412,1068,1).
txt_page(533,1068,1).
txt_page(955,1068,1).
txt_page(554,1068,1).
txt_page(45,1068,1).
txt_page(903,1068,1).
txt_page(537,1069,1).
txt_page(628,1069,1).
txt_page(497,1069,1).
txt_page(467,1069,1).
txt_page(450,1069,1).
txt_page(382,1069,1).
txt_page(54,1069,1).
txt_page(217,1069,1).
txt_page(384,1069,1).
txt_page(598,1069,1).
txt_page(376,1069,1).
txt_page(634,1069,1).
txt_page(897,1069,1).
txt_page(959,1069,1).
txt_page(550,1069,1).
txt_page(308,1069,1).
txt_page(170,1069,1).
txt_page(488,1069,1).
txt_page(885,1069,1).
txt_page(191,1069,1).
txt_page(643,1070,1).
txt_page(637,1070,1).
txt_page(595,1070,1).
txt_page(881,1070,1).
txt_page(52,1070,1).
txt_page(930,1070,1).
txt_page(536,1070,1).
txt_page(749,1070,1).
txt_page(795,1070,1).
txt_page(76,1070,1).
txt_page(416,1070,1).
txt_page(243,1070,1).
txt_page(885,1070,1).
txt_page(18,1070,1).
txt_page(122,1070,1).
txt_page(850,1070,1).
txt_page(29,1070,1).
txt_page(741,1070,1).
txt_page(896,1070,1).
txt_page(449,1070,1).
txt_page(462,1071,2).
txt_page(18,1071,1).
txt_page(315,1071,1).
txt_page(569,1071,1).
txt_page(220,1071,1).
txt_page(197,1071,1).
txt_page(678,1071,1).
txt_page(783,1071,1).
txt_page(221,1071,1).
txt_page(5,1071,1).
txt_page(891,1071,1).
txt_page(832,1071,1).
txt_page(854,1071,1).
txt_page(114,1071,1).
txt_page(549,1071,1).
txt_page(493,1071,1).
txt_page(951,1071,1).
txt_page(56,1071,1).
txt_page(871,1071,1).
txt_page(421,1072,1).
txt_page(149,1072,1).
txt_page(740,1072,1).
txt_page(465,1072,1).
txt_page(601,1072,1).
txt_page(475,1072,1).
txt_page(334,1072,1).
txt_page(865,1072,1).
txt_page(381,1072,1).
txt_page(372,1072,1).
txt_page(807,1072,1).
txt_page(889,1072,1).
txt_page(358,1072,1).
txt_page(360,1072,1).
txt_page(111,1072,1).
txt_page(816,1072,1).
txt_page(683,1072,1).
txt_page(145,1072,1).
txt_page(937,1072,1).
txt_page(810,1072,1).
txt_page(395,1073,1).
txt_page(802,1073,1).
txt_page(316,1073,1).
txt_page(489,1073,1).
txt_page(374,1073,1).
txt_page(323,1073,1).
txt_page(586,1073,1).
txt_page(869,1073,1).
txt_page(123,1073,1).
txt_page(91,1073,1).
txt_page(667,1073,1).
txt_page(905,1073,1).
txt_page(538,1073,1).
txt_page(810,1073,1).
txt_page(301,1073,1).
txt_page(505,1073,1).
txt_page(934,1073,1).
txt_page(225,1073,1).
txt_page(227,1073,1).
txt_page(583,1073,1).
txt_page(892,1074,1).
txt_page(101,1074,1).
txt_page(264,1074,1).
txt_page(184,1074,1).
txt_page(335,1074,1).
txt_page(967,1074,1).
txt_page(670,1074,1).
txt_page(584,1074,1).
txt_page(563,1074,1).
txt_page(759,1074,1).
txt_page(449,1074,1).
txt_page(855,1074,1).
txt_page(62,1074,1).
txt_page(792,1074,1).
txt_page(648,1074,1).
txt_page(751,1074,1).
txt_page(186,1074,1).
txt_page(621,1074,1).
txt_page(900,1074,1).
txt_page(10,1074,1).
txt_page(443,1075,1).
txt_page(581,1075,1).
txt_page(122,1075,1).
txt_page(360,1075,1).
txt_page(732,1075,1).
txt_page(571,1075,1).
txt_page(812,1075,1).
txt_page(670,1075,1).
txt_page(645,1075,1).
txt_page(7,1075,1).
txt_page(672,1075,1).
txt_page(367,1075,1).
txt_page(241,1075,1).
txt_page(66,1075,1).
txt_page(187,1075,1).
txt_page(775,1075,1).
txt_page(649,1075,1).
txt_page(24,1075,1).
txt_page(127,1075,1).
txt_page(452,1075,1).
txt_page(288,1076,1).
txt_page(249,1076,1).
txt_page(544,1076,1).
txt_page(715,1076,1).
txt_page(512,1076,1).
txt_page(133,1076,1).
txt_page(616,1076,1).
txt_page(34,1076,1).
txt_page(917,1076,1).
txt_page(596,1076,1).
txt_page(781,1076,1).
txt_page(511,1076,1).
txt_page(437,1076,1).
txt_page(199,1076,1).
txt_page(747,1076,1).
txt_page(778,1076,1).
txt_page(304,1076,1).
txt_page(953,1076,1).
txt_page(504,1076,1).
txt_page(417,1076,1).
txt_page(168,1077,1).
txt_page(477,1077,1).
txt_page(914,1077,1).
txt_page(564,1077,1).
txt_page(101,1077,1).
txt_page(815,1077,1).
txt_page(937,1077,1).
txt_page(869,1077,2).
txt_page(72,1077,1).
txt_page(115,1077,1).
txt_page(735,1077,1).
txt_page(355,1077,1).
txt_page(677,1077,1).
txt_page(342,1077,1).
txt_page(888,1077,1).
txt_page(181,1077,1).
txt_page(392,1077,1).
txt_page(142,1077,1).
txt_page(542,1077,1).
txt_page(922,1078,1).
txt_page(420,1078,1).
txt_page(27,1078,1).
txt_page(576,1078,1).
txt_page(512,1078,1).
txt_page(795,1078,1).
txt_page(449,1078,1).
txt_page(437,1078,1).
txt_page(270,1078,1).
txt_page(914,1078,1).
txt_page(542,1078,1).
txt_page(456,1078,1).
txt_page(816,1078,1).
txt_page(312,1078,1).
txt_page(833,1078,1).
txt_page(432,1078,1).
txt_page(810,1078,1).
txt_page(799,1078,1).
txt_page(941,1078,1).
txt_page(940,1078,1).
txt_page(313,1079,1).
txt_page(462,1079,1).
txt_page(739,1079,1).
txt_page(431,1079,1).
txt_page(814,1079,1).
txt_page(928,1079,1).
txt_page(83,1079,1).
txt_page(145,1079,1).
txt_page(782,1079,1).
txt_page(52,1079,1).
txt_page(713,1079,1).
txt_page(377,1079,1).
txt_page(124,1079,1).
txt_page(743,1079,1).
txt_page(280,1079,1).
txt_page(715,1079,1).
txt_page(29,1079,1).
txt_page(362,1079,1).
txt_page(598,1079,1).
txt_page(92,1079,1).
txt_page(238,1080,1).
txt_page(56,1080,1).
txt_page(65,1080,1).
txt_page(955,1080,1).
txt_page(253,1080,1).
txt_page(40,1080,1).
txt_page(121,1080,1).
txt_page(73,1080,1).
txt_page(101,1080,1).
txt_page(878,1080,1).
txt_page(888,1080,1).
txt_page(605,1080,1).
txt_page(67,1080,1).
txt_page(255,1080,1).
txt_page(379,1080,1).
txt_page(920,1080,1).
txt_page(368,1080,1).
txt_page(622,1080,1).
txt_page(575,1080,1).
txt_page(196,1080,1).
txt_page(545,1081,1).
txt_page(826,1081,1).
txt_page(154,1081,1).
txt_page(868,1081,1).
txt_page(566,1081,1).
txt_page(334,1081,1).
txt_page(124,1081,1).
txt_page(64,1081,1).
txt_page(547,1081,1).
txt_page(115,1081,1).
txt_page(285,1081,1).
txt_page(895,1081,1).
txt_page(329,1081,1).
txt_page(448,1081,1).
txt_page(645,1081,1).
txt_page(208,1081,1).
txt_page(561,1081,1).
txt_page(296,1081,1).
txt_page(872,1081,1).
txt_page(694,1081,1).
txt_page(919,1082,1).
txt_page(62,1082,1).
txt_page(18,1082,1).
txt_page(329,1082,1).
txt_page(310,1082,1).
txt_page(892,1082,1).
txt_page(494,1082,1).
txt_page(153,1082,1).
txt_page(745,1082,1).
txt_page(930,1082,1).
txt_page(691,1082,1).
txt_page(408,1082,1).
txt_page(786,1082,1).
txt_page(285,1082,1).
txt_page(761,1082,1).
txt_page(624,1082,1).
txt_page(264,1082,1).
txt_page(449,1082,1).
txt_page(464,1082,1).
txt_page(316,1082,1).
txt_page(741,1083,1).
txt_page(925,1083,1).
txt_page(674,1083,1).
txt_page(766,1083,1).
txt_page(875,1083,1).
txt_page(767,1083,1).
txt_page(331,1083,1).
txt_page(3,1083,1).
txt_page(967,1083,1).
txt_page(322,1083,1).
txt_page(135,1083,1).
txt_page(41,1083,2).
txt_page(212,1083,1).
txt_page(902,1083,1).
txt_page(282,1083,1).
txt_page(202,1083,1).
txt_page(387,1083,1).
txt_page(440,1083,1).
txt_page(48,1083,1).
txt_page(393,1084,1).
txt_page(869,1084,1).
txt_page(180,1084,1).
txt_page(480,1084,1).
txt_page(154,1084,1).
txt_page(360,1084,1).
txt_page(308,1084,1).
txt_page(26,1084,1).
txt_page(215,1084,1).
txt_page(909,1084,1).
txt_page(330,1084,1).
txt_page(292,1084,1).
txt_page(886,1084,1).
txt_page(244,1084,1).
txt_page(563,1084,1).
txt_page(365,1084,1).
txt_page(380,1084,1).
txt_page(662,1084,1).
txt_page(24,1084,1).
txt_page(77,1084,1).
txt_page(137,1085,1).
txt_page(65,1085,1).
txt_page(855,1085,1).
txt_page(206,1085,1).
txt_page(604,1085,1).
txt_page(344,1085,1).
txt_page(169,1085,1).
txt_page(805,1085,1).
txt_page(843,1085,1).
txt_page(198,1085,1).
txt_page(501,1085,1).
txt_page(221,1085,1).
txt_page(531,1085,1).
txt_page(923,1085,1).
txt_page(270,1085,2).
txt_page(222,1085,1).
txt_page(268,1085,1).
txt_page(190,1085,2).
txt_page(828,1086,1).
txt_page(51,1086,1).
txt_page(58,1086,1).
txt_page(961,1086,2).
txt_page(648,1086,1).
txt_page(109,1086,1).
txt_page(745,1086,1).
txt_page(947,1086,1).
txt_page(351,1086,1).
txt_page(404,1086,1).
txt_page(666,1086,1).
txt_page(636,1086,1).
txt_page(765,1086,1).
txt_page(645,1086,1).
txt_page(290,1086,1).
txt_page(552,1086,1).
txt_page(171,1086,2).
txt_page(68,1086,1).
txt_page(896,1087,1).
txt_page(721,1087,1).
txt_page(854,1087,1).
txt_page(806,1087,1).
txt_page(597,1087,1).
txt_page(634,1087,1).
txt_page(502,1087,1).
txt_page(836,1087,1).
txt_page(15,1087,1).
txt_page(188,1087,1).
txt_page(796,1087,1).
txt_page(750,1087,1).
txt_page(861,1087,1).
txt_page(567,1087,1).
txt_page(670,1087,1).
txt_page(532,1087,1).
txt_page(574,1087,1).
txt_page(470,1087,1).
txt_page(604,1087,1).
txt_page(919,1087,1).
txt_page(758,1088,1).
txt_page(537,1088,1).
txt_page(482,1088,1).
txt_page(422,1088,1).
txt_page(677,1088,1).
txt_page(379,1088,1).
txt_page(854,1088,1).
txt_page(641,1088,1).
txt_page(788,1088,1).
txt_page(120,1088,1).
txt_page(867,1088,1).
txt_page(655,1088,1).
txt_page(420,1088,1).
txt_page(410,1088,1).
txt_page(266,1088,1).
txt_page(719,1088,1).
txt_page(237,1088,1).
txt_page(349,1088,2).
txt_page(270,1088,1).
txt_page(233,1089,1).
txt_page(664,1089,1).
txt_page(337,1089,1).
txt_page(826,1089,1).
txt_page(303,1089,1).
txt_page(672,1089,1).
txt_page(75,1089,1).
txt_page(37,1089,1).
txt_page(247,1089,1).
txt_page(907,1089,1).
txt_page(797,1089,1).
txt_page(661,1089,1).
txt_page(272,1089,1).
txt_page(821,1089,1).
txt_page(584,1089,1).
txt_page(860,1089,1).
txt_page(748,1089,1).
txt_page(587,1089,1).
txt_page(680,1089,1).
txt_page(214,1089,1).
txt_page(254,1090,1).
txt_page(796,1090,1).
txt_page(729,1090,1).
txt_page(932,1090,1).
txt_page(172,1090,1).
txt_page(370,1090,1).
txt_page(892,1090,1).
txt_page(406,1090,1).
txt_page(845,1090,1).
txt_page(529,1090,1).
txt_page(718,1090,1).
txt_page(859,1090,1).
txt_page(964,1090,1).
txt_page(491,1090,1).
txt_page(384,1090,1).
txt_page(747,1090,1).
txt_page(922,1090,1).
txt_page(23,1090,1).
txt_page(865,1090,1).
txt_page(882,1090,1).
txt_page(655,1091,1).
txt_page(44,1091,1).
txt_page(880,1091,1).
txt_page(323,1091,1).
txt_page(242,1091,1).
txt_page(175,1091,1).
txt_page(391,1091,1).
txt_page(914,1091,1).
txt_page(76,1091,1).
txt_page(265,1091,1).
txt_page(270,1091,1).
txt_page(374,1091,1).
txt_page(375,1091,1).
txt_page(714,1091,1).
txt_page(911,1091,1).
txt_page(125,1091,1).
txt_page(33,1091,1).
txt_page(808,1091,1).
txt_page(147,1091,1).
txt_page(570,1091,1).
txt_page(135,1092,1).
txt_page(960,1092,1).
txt_page(559,1092,1).
txt_page(380,1092,2).
txt_page(8,1092,1).
txt_page(348,1092,1).
txt_page(181,1092,1).
txt_page(112,1092,1).
txt_page(643,1092,1).
txt_page(402,1092,1).
txt_page(291,1092,1).
txt_page(321,1092,1).
txt_page(453,1092,1).
txt_page(116,1092,1).
txt_page(75,1092,1).
txt_page(942,1092,1).
txt_page(941,1092,1).
txt_page(237,1092,1).
txt_page(370,1092,1).
txt_page(586,1093,1).
txt_page(200,1093,1).
txt_page(313,1093,1).
txt_page(308,1093,1).
txt_page(187,1093,1).
txt_page(897,1093,1).
txt_page(763,1093,1).
txt_page(41,1093,1).
txt_page(590,1093,1).
txt_page(845,1093,1).
txt_page(746,1093,1).
txt_page(929,1093,1).
txt_page(252,1093,1).
txt_page(424,1093,1).
txt_page(358,1093,1).
txt_page(114,1093,1).
txt_page(760,1093,1).
txt_page(616,1093,1).
txt_page(473,1093,1).
txt_page(894,1093,1).
txt_page(791,1094,1).
txt_page(184,1094,1).
txt_page(353,1094,1).
txt_page(937,1094,1).
txt_page(808,1094,1).
txt_page(201,1094,1).
txt_page(27,1094,1).
txt_page(965,1094,1).
txt_page(381,1094,1).
txt_page(367,1094,1).
txt_page(922,1094,1).
txt_page(674,1094,1).
txt_page(77,1094,2).
txt_page(654,1094,1).
txt_page(127,1094,1).
txt_page(636,1094,1).
txt_page(470,1094,1).
txt_page(928,1094,1).
txt_page(966,1094,1).
txt_page(401,1095,1).
txt_page(187,1095,1).
txt_page(668,1095,1).
txt_page(808,1095,1).
txt_page(693,1095,1).
txt_page(527,1095,1).
txt_page(507,1095,1).
txt_page(317,1095,1).
txt_page(695,1095,1).
txt_page(347,1095,1).
txt_page(680,1095,1).
txt_page(114,1095,1).
txt_page(89,1095,1).
txt_page(674,1095,1).
txt_page(350,1095,1).
txt_page(667,1095,1).
txt_page(501,1095,1).
txt_page(97,1095,1).
txt_page(330,1095,1).
txt_page(751,1095,1).
txt_page(45,1096,1).
txt_page(68,1096,1).
txt_page(752,1096,1).
txt_page(789,1096,1).
txt_page(46,1096,1).
txt_page(276,1096,1).
txt_page(436,1096,1).
txt_page(558,1096,1).
txt_page(837,1096,1).
txt_page(905,1096,1).
txt_page(858,1096,1).
txt_page(846,1096,1).
txt_page(143,1096,1).
txt_page(145,1096,1).
txt_page(120,1096,1).
txt_page(554,1096,1).
txt_page(861,1096,1).
txt_page(458,1096,1).
txt_page(153,1096,1).
txt_page(753,1096,1).
txt_page(454,1097,1).
txt_page(680,1097,1).
txt_page(128,1097,1).
txt_page(852,1097,1).
txt_page(274,1097,1).
txt_page(487,1097,1).
txt_page(313,1097,1).
txt_page(859,1097,1).
txt_page(382,1097,1).
txt_page(895,1097,1).
txt_page(809,1097,1).
txt_page(563,1097,1).
txt_page(129,1097,2).
txt_page(599,1097,1).
txt_page(370,1097,1).
txt_page(478,1097,1).
txt_page(116,1097,1).
txt_page(234,1097,1).
txt_page(531,1097,1).
txt_page(127,1098,1).
txt_page(595,1098,1).
txt_page(326,1098,1).
txt_page(656,1098,1).
txt_page(844,1098,1).
txt_page(791,1098,1).
txt_page(651,1098,1).
txt_page(635,1098,1).
txt_page(387,1098,1).
txt_page(878,1098,1).
txt_page(954,1098,1).
txt_page(348,1098,1).
txt_page(264,1098,1).
txt_page(846,1098,1).
txt_page(744,1098,1).
txt_page(6,1098,1).
txt_page(462,1098,1).
txt_page(526,1098,2).
txt_page(927,1098,1).
txt_page(146,1099,1).
txt_page(852,1099,1).
txt_page(827,1099,1).
txt_page(485,1099,1).
txt_page(55,1099,1).
txt_page(311,1099,1).
txt_page(24,1099,1).
txt_page(135,1099,1).
txt_page(788,1099,1).
txt_page(87,1099,1).
txt_page(356,1099,1).
txt_page(119,1099,1).
txt_page(435,1099,1).
txt_page(249,1099,1).
txt_page(814,1099,1).
txt_page(181,1099,1).
txt_page(813,1099,1).
txt_page(648,1099,1).
txt_page(919,1099,1).
txt_page(516,1099,1).
txt_page(906,1100,1).
txt_page(343,1100,1).
txt_page(198,1100,1).
txt_page(696,1100,1).
txt_page(344,1100,1).
txt_page(139,1100,1).
txt_page(486,1100,1).
txt_page(253,1100,1).
txt_page(705,1100,1).
txt_page(89,1100,1).
txt_page(195,1100,1).
txt_page(699,1100,1).
txt_page(614,1100,1).
txt_page(109,1100,1).
txt_page(673,1100,1).
txt_page(596,1100,1).
txt_page(371,1100,1).
txt_page(644,1100,1).
txt_page(506,1100,1).
txt_page(905,1100,1).
txt_page(53,1101,1).
txt_page(826,1101,1).
txt_page(651,1101,1).
txt_page(608,1101,1).
txt_page(754,1101,1).
txt_page(255,1101,1).
txt_page(476,1101,1).
txt_page(180,1101,1).
txt_page(454,1101,1).
txt_page(110,1101,1).
txt_page(197,1101,1).
txt_page(345,1101,1).
txt_page(679,1101,1).
txt_page(830,1101,1).
txt_page(543,1101,1).
txt_page(210,1101,1).
txt_page(716,1101,1).
txt_page(456,1101,1).
txt_page(760,1101,1).
txt_page(535,1101,1).
txt_page(370,1102,1).
txt_page(35,1102,1).
txt_page(64,1102,1).
txt_page(126,1102,1).
txt_page(380,1102,1).
txt_page(392,1102,1).
txt_page(270,1102,1).
txt_page(725,1102,1).
txt_page(102,1102,1).
txt_page(219,1102,1).
txt_page(732,1102,1).
txt_page(239,1102,1).
txt_page(848,1102,1).
txt_page(279,1102,1).
txt_page(360,1102,1).
txt_page(248,1102,1).
txt_page(635,1102,1).
txt_page(162,1102,1).
txt_page(574,1102,1).
txt_page(777,1102,1).
txt_page(915,1103,1).
txt_page(785,1103,1).
txt_page(513,1103,1).
txt_page(694,1103,1).
txt_page(909,1103,1).
txt_page(864,1103,1).
txt_page(315,1103,1).
txt_page(603,1103,1).
txt_page(74,1103,1).
txt_page(943,1103,1).
txt_page(140,1103,1).
txt_page(349,1103,1).
txt_page(574,1103,1).
txt_page(542,1103,1).
txt_page(605,1103,1).
txt_page(628,1103,1).
txt_page(569,1103,1).
txt_page(670,1103,1).
txt_page(230,1103,1).
txt_page(276,1103,1).
txt_page(757,1104,1).
txt_page(395,1104,1).
txt_page(897,1104,1).
txt_page(890,1104,1).
txt_page(397,1104,1).
txt_page(182,1104,1).
txt_page(441,1104,1).
txt_page(681,1104,1).
txt_page(624,1104,1).
txt_page(760,1104,1).
txt_page(665,1104,1).
txt_page(424,1104,1).
txt_page(963,1104,1).
txt_page(820,1104,1).
txt_page(65,1104,1).
txt_page(290,1104,1).
txt_page(69,1104,1).
txt_page(662,1104,1).
txt_page(421,1104,1).
txt_page(144,1104,1).
txt_page(907,1105,1).
txt_page(878,1105,1).
txt_page(329,1105,1).
txt_page(507,1105,1).
txt_page(563,1105,1).
txt_page(426,1105,1).
txt_page(876,1105,1).
txt_page(113,1105,1).
txt_page(246,1105,1).
txt_page(395,1105,1).
txt_page(925,1105,1).
txt_page(890,1105,1).
txt_page(69,1105,1).
txt_page(323,1105,1).
txt_page(467,1105,1).
txt_page(842,1105,1).
txt_page(143,1105,1).
txt_page(87,1105,1).
txt_page(136,1105,1).
txt_page(339,1105,1).
txt_page(113,1106,1).
txt_page(550,1106,1).
txt_page(420,1106,1).
txt_page(776,1106,1).
txt_page(267,1106,1).
txt_page(508,1106,1).
txt_page(727,1106,1).
txt_page(273,1106,1).
txt_page(327,1106,1).
txt_page(766,1106,1).
txt_page(639,1106,1).
txt_page(442,1106,1).
txt_page(223,1106,1).
txt_page(611,1106,1).
txt_page(852,1106,1).
txt_page(131,1106,1).
txt_page(43,1106,1).
txt_page(593,1106,1).
txt_page(216,1106,1).
txt_page(496,1106,1).
txt_page(930,1107,1).
txt_page(830,1107,1).
txt_page(434,1107,1).
txt_page(67,1107,1).
txt_page(22,1107,1).
txt_page(760,1107,1).
txt_page(514,1107,1).
txt_page(265,1107,1).
txt_page(329,1107,1).
txt_page(703,1107,1).
txt_page(842,1107,1).
txt_page(461,1107,1).
txt_page(833,1107,1).
txt_page(613,1107,1).
txt_page(834,1107,1).
txt_page(521,1107,1).
txt_page(902,1107,1).
txt_page(170,1107,1).
txt_page(88,1107,1).
txt_page(862,1107,1).
txt_page(608,1108,1).
txt_page(28,1108,1).
txt_page(803,1108,1).
txt_page(740,1108,1).
txt_page(292,1108,1).
txt_page(76,1108,1).
txt_page(234,1108,1).
txt_page(935,1108,1).
txt_page(118,1108,1).
txt_page(716,1108,1).
txt_page(536,1108,1).
txt_page(550,1108,1).
txt_page(845,1108,1).
txt_page(577,1108,1).
txt_page(651,1108,1).
txt_page(667,1108,1).
txt_page(84,1108,1).
txt_page(211,1108,1).
txt_page(494,1108,1).
txt_page(693,1108,1).
txt_page(405,1109,1).
txt_page(591,1109,1).
txt_page(358,1109,1).
txt_page(9,1109,1).
txt_page(26,1109,1).
txt_page(964,1109,1).
txt_page(456,1109,1).
txt_page(922,1109,1).
txt_page(759,1109,1).
txt_page(226,1109,1).
txt_page(611,1109,1).
txt_page(589,1109,1).
txt_page(265,1109,1).
txt_page(667,1109,1).
txt_page(526,1109,1).
txt_page(171,1109,1).
txt_page(295,1109,1).
txt_page(560,1109,1).
txt_page(373,1109,1).
txt_page(641,1109,1).
txt_page(914,1110,1).
txt_page(218,1110,1).
txt_page(8,1110,1).
txt_page(711,1110,1).
txt_page(678,1110,1).
txt_page(308,1110,1).
txt_page(728,1110,1).
txt_page(3,1110,1).
txt_page(304,1110,1).
txt_page(319,1110,1).
txt_page(558,1110,1).
txt_page(797,1110,1).
txt_page(908,1110,1).
txt_page(244,1110,1).
txt_page(240,1110,1).
txt_page(297,1110,1).
txt_page(876,1110,1).
txt_page(474,1110,1).
txt_page(692,1110,1).
txt_page(322,1110,1).
txt_page(646,1111,1).
txt_page(49,1111,1).
txt_page(560,1111,1).
txt_page(533,1111,1).
txt_page(289,1111,1).
txt_page(573,1111,1).
txt_page(858,1111,1).
txt_page(710,1111,1).
txt_page(325,1111,1).
txt_page(562,1111,1).
txt_page(67,1111,1).
txt_page(233,1111,1).
txt_page(252,1111,1).
txt_page(160,1111,1).
txt_page(893,1111,1).
txt_page(957,1111,1).
txt_page(44,1111,1).
txt_page(713,1111,1).
txt_page(620,1111,1).
txt_page(619,1111,1).
txt_page(268,1112,1).
txt_page(685,1112,1).
txt_page(238,1112,1).
txt_page(833,1112,1).
txt_page(522,1112,1).
txt_page(477,1112,1).
txt_page(632,1112,1).
txt_page(465,1112,1).
txt_page(29,1112,1).
txt_page(850,1112,1).
txt_page(319,1112,1).
txt_page(364,1112,1).
txt_page(398,1112,1).
txt_page(949,1112,1).
txt_page(108,1112,1).
txt_page(474,1112,1).
txt_page(608,1112,1).
txt_page(433,1112,1).
txt_page(593,1112,1).
txt_page(662,1112,1).
txt_page(815,1113,1).
txt_page(451,1113,1).
txt_page(349,1113,1).
txt_page(628,1113,1).
txt_page(721,1113,2).
txt_page(68,1113,1).
txt_page(960,1113,1).
txt_page(624,1113,1).
txt_page(950,1113,1).
txt_page(677,1113,1).
txt_page(661,1113,1).
txt_page(183,1113,1).
txt_page(33,1113,1).
txt_page(745,1113,1).
txt_page(779,1113,1).
txt_page(408,1113,1).
txt_page(466,1113,1).
txt_page(461,1113,1).
txt_page(719,1113,1).
txt_page(118,1114,1).
txt_page(636,1114,1).
txt_page(37,1114,1).
txt_page(197,1114,1).
txt_page(935,1114,1).
txt_page(200,1114,1).
txt_page(839,1114,1).
txt_page(51,1114,1).
txt_page(721,1114,1).
txt_page(34,1114,1).
txt_page(321,1114,1).
txt_page(471,1114,1).
txt_page(780,1114,1).
txt_page(884,1114,1).
txt_page(567,1114,1).
txt_page(856,1114,1).
txt_page(787,1114,1).
txt_page(555,1114,1).
txt_page(461,1114,1).
txt_page(93,1114,1).
txt_page(3,1115,1).
txt_page(959,1115,1).
txt_page(690,1115,1).
txt_page(126,1115,1).
txt_page(713,1115,1).
txt_page(350,1115,1).
txt_page(720,1115,1).
txt_page(379,1115,1).
txt_page(827,1115,1).
txt_page(747,1115,1).
txt_page(589,1115,1).
txt_page(296,1115,1).
txt_page(348,1115,1).
txt_page(47,1115,1).
txt_page(109,1115,1).
txt_page(464,1115,1).
txt_page(166,1115,1).
txt_page(802,1115,1).
txt_page(91,1115,1).
txt_page(551,1115,1).
txt_page(309,1116,1).
txt_page(673,1116,1).
txt_page(171,1116,1).
txt_page(24,1116,1).
txt_page(176,1116,1).
txt_page(447,1116,1).
txt_page(959,1116,1).
txt_page(210,1116,1).
txt_page(440,1116,1).
txt_page(341,1116,1).
txt_page(675,1116,1).
txt_page(895,1116,1).
txt_page(401,1116,1).
txt_page(791,1116,1).
txt_page(255,1116,1).
txt_page(109,1116,1).
txt_page(924,1116,1).
txt_page(448,1116,2).
txt_page(611,1116,1).
txt_page(181,1117,1).
txt_page(50,1117,1).
txt_page(725,1117,1).
txt_page(291,1117,1).
txt_page(465,1117,1).
txt_page(665,1117,1).
txt_page(290,1117,1).
txt_page(117,1117,1).
txt_page(655,1117,1).
txt_page(675,1117,1).
txt_page(960,1117,1).
txt_page(81,1117,1).
txt_page(269,1117,1).
txt_page(626,1117,1).
txt_page(447,1117,1).
txt_page(534,1117,1).
txt_page(119,1117,1).
txt_page(516,1117,1).
txt_page(306,1117,1).
txt_page(916,1117,1).
txt_page(225,1118,1).
txt_page(139,1118,1).
txt_page(617,1118,1).
txt_page(420,1118,1).
txt_page(425,1118,1).
txt_page(836,1118,1).
txt_page(413,1118,1).
txt_page(738,1118,1).
txt_page(190,1118,1).
txt_page(492,1118,2).
txt_page(290,1118,1).
txt_page(863,1118,1).
txt_page(966,1118,1).
txt_page(252,1118,1).
txt_page(228,1118,1).
txt_page(215,1118,1).
txt_page(481,1118,1).
txt_page(592,1118,1).
txt_page(114,1118,1).
txt_page(478,1119,1).
txt_page(495,1119,1).
txt_page(699,1119,1).
txt_page(27,1119,1).
txt_page(819,1119,1).
txt_page(592,1119,1).
txt_page(803,1119,1).
txt_page(823,1119,1).
txt_page(446,1119,1).
txt_page(621,1119,1).
txt_page(86,1119,1).
txt_page(240,1119,1).
txt_page(291,1119,1).
txt_page(418,1119,1).
txt_page(643,1119,1).
txt_page(714,1119,1).
txt_page(599,1119,1).
txt_page(535,1119,1).
txt_page(682,1119,1).
txt_page(692,1119,1).
txt_page(324,1120,1).
txt_page(62,1120,1).
txt_page(75,1120,1).
txt_page(875,1120,1).
txt_page(117,1120,1).
txt_page(740,1120,1).
txt_page(560,1120,1).
txt_page(482,1120,1).
txt_page(566,1120,1).
txt_page(254,1120,1).
txt_page(461,1120,1).
txt_page(905,1120,1).
txt_page(683,1120,1).
txt_page(640,1120,1).
txt_page(573,1120,1).
txt_page(53,1120,1).
txt_page(31,1120,1).
txt_page(252,1120,1).
txt_page(370,1120,1).
txt_page(122,1120,1).
txt_page(932,1121,1).
txt_page(159,1121,1).
txt_page(343,1121,1).
txt_page(571,1121,1).
txt_page(902,1121,1).
txt_page(314,1121,1).
txt_page(412,1121,1).
txt_page(221,1121,1).
txt_page(766,1121,1).
txt_page(247,1121,1).
txt_page(877,1121,1).
txt_page(74,1121,1).
txt_page(324,1121,1).
txt_page(646,1121,1).
txt_page(531,1121,1).
txt_page(597,1121,1).
txt_page(280,1121,1).
txt_page(579,1121,1).
txt_page(648,1121,1).
txt_page(366,1121,1).
txt_page(423,1122,1).
txt_page(373,1122,1).
txt_page(734,1122,1).
txt_page(24,1122,1).
txt_page(789,1122,1).
txt_page(320,1122,1).
txt_page(891,1122,1).
txt_page(509,1122,1).
txt_page(78,1122,1).
txt_page(942,1122,1).
txt_page(301,1122,1).
txt_page(111,1122,1).
txt_page(832,1122,1).
txt_page(201,1122,1).
txt_page(572,1122,1).
txt_page(214,1122,1).
txt_page(695,1122,1).
txt_page(528,1122,1).
txt_page(931,1122,1).
txt_page(442,1122,1).
txt_page(632,1123,1).
txt_page(648,1123,1).
txt_page(373,1123,1).
txt_page(310,1123,1).
txt_page(965,1123,1).
txt_page(343,1123,1).
txt_page(760,1123,1).
txt_page(664,1123,1).
txt_page(158,1123,1).
txt_page(78,1123,1).
txt_page(42,1123,1).
txt_page(134,1123,1).
txt_page(912,1123,1).
txt_page(327,1123,1).
txt_page(131,1123,1).
txt_page(899,1123,1).
txt_page(225,1123,1).
txt_page(339,1123,1).
txt_page(72,1123,1).
txt_page(163,1123,1).
txt_page(367,1124,1).
txt_page(404,1124,1).
txt_page(484,1124,1).
txt_page(883,1124,1).
txt_page(130,1124,1).
txt_page(340,1124,1).
txt_page(424,1124,1).
txt_page(906,1124,1).
txt_page(833,1124,1).
txt_page(28,1124,1).
txt_page(347,1124,1).
txt_page(483,1124,1).
txt_page(741,1124,1).
txt_page(105,1124,1).
txt_page(457,1124,1).
txt_page(518,1124,1).
txt_page(822,1124,1).
txt_page(862,1124,1).
txt_page(656,1124,1).
txt_page(787,1124,1).
txt_page(955,1125,1).
txt_page(486,1125,1).
txt_page(786,1125,1).
txt_page(593,1125,1).
txt_page(316,1125,1).
txt_page(295,1125,1).
txt_page(329,1125,1).
txt_page(467,1125,1).
txt_page(406,1125,1).
txt_page(297,1125,1).
txt_page(177,1125,1).
txt_page(374,1125,1).
txt_page(306,1125,1).
txt_page(426,1125,1).
txt_page(608,1125,1).
txt_page(32,1125,1).
txt_page(94,1125,1).
txt_page(751,1125,1).
txt_page(695,1125,1).
txt_page(611,1125,1).
txt_page(32,1126,1).
txt_page(822,1126,1).
txt_page(832,1126,1).
txt_page(865,1126,1).
txt_page(927,1126,1).
txt_page(559,1126,1).
txt_page(873,1126,1).
txt_page(411,1126,1).
txt_page(478,1126,1).
txt_page(182,1126,1).
txt_page(414,1126,1).
txt_page(13,1126,1).
txt_page(108,1126,1).
txt_page(835,1126,1).
txt_page(290,1126,1).
txt_page(815,1126,1).
txt_page(456,1126,1).
txt_page(501,1126,1).
txt_page(745,1126,1).
txt_page(596,1126,1).
txt_page(163,1127,1).
txt_page(391,1127,1).
txt_page(25,1127,1).
txt_page(709,1127,1).
txt_page(583,1127,1).
txt_page(757,1127,1).
txt_page(250,1127,1).
txt_page(183,1127,1).
txt_page(683,1127,1).
txt_page(255,1127,1).
txt_page(254,1127,1).
txt_page(147,1127,1).
txt_page(321,1127,1).
txt_page(80,1127,1).
txt_page(573,1127,1).
txt_page(411,1127,1).
txt_page(112,1127,1).
txt_page(182,1127,1).
txt_page(222,1127,1).
txt_page(39,1127,1).
txt_page(93,1128,1).
txt_page(120,1128,1).
txt_page(62,1128,1).
txt_page(421,1128,1).
txt_page(637,1128,1).
txt_page(2,1128,1).
txt_page(686,1128,1).
txt_page(237,1128,1).
txt_page(820,1128,1).
txt_page(756,1128,1).
txt_page(923,1128,1).
txt_page(238,1128,1).
txt_page(513,1128,1).
txt_page(560,1128,1).
txt_page(583,1128,1).
txt_page(39,1128,1).
txt_page(535,1128,1).
txt_page(95,1128,1).
txt_page(398,1128,1).
txt_page(276,1128,1).
txt_page(176,1129,1).
txt_page(565,1129,1).
txt_page(31,1129,1).
txt_page(882,1129,1).
txt_page(785,1129,1).
txt_page(117,1129,1).
txt_page(392,1129,1).
txt_page(256,1129,1).
txt_page(842,1129,1).
txt_page(96,1129,1).
txt_page(163,1129,1).
txt_page(426,1129,1).
txt_page(71,1129,1).
txt_page(872,1129,1).
txt_page(679,1129,1).
txt_page(643,1129,1).
txt_page(663,1129,1).
txt_page(353,1129,1).
txt_page(320,1129,1).
txt_page(798,1129,1).
txt_page(259,1130,1).
txt_page(668,1130,1).
txt_page(370,1130,1).
txt_page(758,1130,1).
txt_page(853,1130,1).
txt_page(359,1130,1).
txt_page(843,1130,1).
txt_page(696,1130,1).
txt_page(786,1130,1).
txt_page(782,1130,1).
txt_page(126,1130,1).
txt_page(143,1130,1).
txt_page(362,1130,1).
txt_page(729,1130,1).
txt_page(47,1130,1).
txt_page(779,1130,1).
txt_page(90,1130,1).
txt_page(889,1130,1).
txt_page(311,1130,1).
txt_page(916,1130,1).
txt_page(653,1131,1).
txt_page(431,1131,1).
txt_page(349,1131,1).
txt_page(718,1131,1).
txt_page(480,1131,1).
txt_page(479,1131,1).
txt_page(34,1131,1).
txt_page(964,1131,1).
txt_page(380,1131,1).
txt_page(851,1131,1).
txt_page(782,1131,1).
txt_page(882,1131,1).
txt_page(953,1131,1).
txt_page(36,1131,1).
txt_page(205,1131,1).
txt_page(690,1131,1).
txt_page(627,1131,1).
txt_page(49,1131,1).
txt_page(159,1131,1).
txt_page(684,1131,1).
txt_page(49,1132,1).
txt_page(239,1132,1).
txt_page(516,1132,1).
txt_page(453,1132,1).
txt_page(394,1132,1).
txt_page(652,1132,1).
txt_page(599,1132,1).
txt_page(878,1132,1).
txt_page(547,1132,1).
txt_page(664,1132,1).
txt_page(97,1132,1).
txt_page(679,1132,1).
txt_page(761,1132,1).
txt_page(752,1132,1).
txt_page(65,1132,1).
txt_page(93,1132,1).
txt_page(673,1132,1).
txt_page(633,1132,1).
txt_page(560,1132,1).
txt_page(170,1132,1).
txt_page(803,1133,1).
txt_page(468,1133,1).
txt_page(923,1133,1).
txt_page(906,1133,1).
txt_page(594,1133,1).
txt_page(885,1133,1).
txt_page(304,1133,1).
txt_page(770,1133,1).
txt_page(151,1133,1).
txt_page(473,1133,1).
txt_page(762,1133,1).
txt_page(563,1133,1).
txt_page(110,1133,1).
txt_page(629,1133,1).
txt_page(827,1133,1).
txt_page(454,1133,1).
txt_page(86,1133,1).
txt_page(638,1133,1).
txt_page(350,1133,1).
txt_page(437,1133,1).
txt_page(634,1134,1).
txt_page(506,1134,1).
txt_page(407,1134,1).
txt_page(611,1134,1).
txt_page(608,1134,1).
txt_page(484,1134,1).
txt_page(276,1134,1).
txt_page(69,1134,1).
txt_page(558,1134,1).
txt_page(945,1134,1).
txt_page(125,1134,1).
txt_page(117,1134,1).
txt_page(614,1134,1).
txt_page(741,1134,1).
txt_page(592,1134,1).
txt_page(719,1134,1).
txt_page(375,1134,1).
txt_page(1,1134,1).
txt_page(199,1134,1).
txt_page(337,1134,1).
txt_page(139,1135,1).
txt_page(286,1135,1).
txt_page(269,1135,1).
txt_page(328,1135,1).
txt_page(116,1135,1).
txt_page(425,1135,1).
txt_page(942,1135,1).
txt_page(776,1135,1).
txt_page(313,1135,1).
txt_page(521,1135,1).
txt_page(578,1135,2).
txt_page(674,1135,1).
txt_page(476,1135,1).
txt_page(164,1135,1).
txt_page(228,1135,1).
txt_page(62,1135,1).
txt_page(803,1135,1).
txt_page(770,1135,1).
txt_page(907,1135,1).
txt_page(135,1136,1).
txt_page(523,1136,1).
txt_page(955,1136,1).
txt_page(105,1136,1).
txt_page(273,1136,1).
txt_page(235,1136,1).
txt_page(74,1136,1).
txt_page(907,1136,1).
txt_page(237,1136,1).
txt_page(533,1136,1).
txt_page(320,1136,1).
txt_page(134,1136,1).
txt_page(396,1136,1).
txt_page(680,1136,1).
txt_page(834,1136,1).
txt_page(309,1136,1).
txt_page(29,1136,1).
txt_page(323,1136,1).
txt_page(900,1136,1).
txt_page(146,1136,1).
txt_page(894,1137,1).
txt_page(171,1137,1).
txt_page(467,1137,1).
txt_page(837,1137,1).
txt_page(614,1137,1).
txt_page(334,1137,1).
txt_page(757,1137,1).
txt_page(45,1137,1).
txt_page(706,1137,1).
txt_page(610,1137,1).
txt_page(362,1137,1).
txt_page(685,1137,1).
txt_page(595,1137,1).
txt_page(383,1137,1).
txt_page(908,1137,1).
txt_page(353,1137,2).
txt_page(739,1137,1).
txt_page(540,1137,1).
txt_page(742,1137,1).
txt_page(246,1138,1).
txt_page(203,1138,1).
txt_page(348,1138,1).
txt_page(99,1138,1).
txt_page(679,1138,1).
txt_page(545,1138,2).
txt_page(783,1138,1).
txt_page(156,1138,1).
txt_page(750,1138,1).
txt_page(901,1138,1).
txt_page(79,1138,1).
txt_page(585,1138,1).
txt_page(698,1138,1).
txt_page(493,1138,1).
txt_page(58,1138,1).
txt_page(682,1138,1).
txt_page(414,1138,1).
txt_page(77,1138,1).
txt_page(31,1138,1).
txt_page(173,1139,1).
txt_page(360,1139,1).
txt_page(454,1139,1).
txt_page(75,1139,1).
txt_page(350,1139,1).
txt_page(900,1139,1).
txt_page(711,1139,1).
txt_page(548,1139,1).
txt_page(365,1139,1).
txt_page(524,1139,1).
txt_page(194,1139,1).
txt_page(34,1139,1).
txt_page(833,1139,1).
txt_page(628,1139,1).
txt_page(924,1139,1).
txt_page(213,1139,1).
txt_page(157,1139,1).
txt_page(912,1139,1).
txt_page(490,1139,1).
txt_page(762,1139,1).
/* (22571 entrees) */


/**
 * Table doc_txt
 */

doc_txt(1).
doc_txt(2).
doc_txt(3).
doc_txt(4).
doc_txt(5).
doc_txt(6).
doc_txt(7).
doc_txt(8).
doc_txt(9).
doc_txt(11).
doc_txt(12).
doc_txt(17).
doc_txt(18).
doc_txt(23).
doc_txt(30).
doc_txt(31).
doc_txt(32).
doc_txt(33).
doc_txt(34).
doc_txt(35).
doc_txt(37).
doc_txt(39).
doc_txt(40).
doc_txt(41).
doc_txt(42).
doc_txt(43).
doc_txt(44).
doc_txt(45).
doc_txt(46).
doc_txt(47).
doc_txt(48).
doc_txt(49).
doc_txt(51).
doc_txt(52).
doc_txt(53).
doc_txt(54).
doc_txt(55).
doc_txt(56).
doc_txt(57).
doc_txt(58).
doc_txt(59).
doc_txt(60).
doc_txt(61).
doc_txt(62).
doc_txt(63).
doc_txt(64).
doc_txt(65).
doc_txt(66).
doc_txt(67).
doc_txt(68).
/* (50 entrees) */


